//Use this class to keep track of a set of reads for a chromosome (one
//or two strands).

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class ABIReads {

    public static final int MAX_READS_PER_LOCUS = Integer.MAX_VALUE; //20000;

    protected int chr, strand;
    protected ABIRead[] reads;

    public ABIReads(int c, boolean forward, ABIRead rs[]) {
	chr = c;
	strand = forward ? 1 : -1;
	reads= rs;
    }

    ///////////////////////////////////////////////////////

    public static TreeSet<TranscribedRegion2> loadRepeatReadsChr(String mRNAFileName, boolean forward, int c, int readLen, boolean flatten, String repeatFileName) {
	//Parse the list of repeat regions
	String species = repeatFileName!=null ? repeatFileName.substring(0, 2) : null;
	TreeSet<TranscribedRegion2> repeatRNAs = repeatFileName!=null ? EstimateGenicTranscription3.parseRepeatFile(repeatFileName, species, c, forward, false) : new TreeSet<TranscribedRegion2>();
	//We can store the reads for each region in a dynamic data structure since there will be relatively few of them. This is not an elegant solution, but it is relatively fast.
	TreeMap<TranscribedRegion2, Vector<ABIRead>> repeatRegionReads = new TreeMap<TranscribedRegion2, Vector<ABIRead>>();
	int nReads = 0;
	BufferedReader br = null;
	try {
	    if (!mRNAFileName.endsWith(".tsv"))
		mRNAFileName += ".tsv";
	    String suffix = mRNAFileName.substring(mRNAFileName.lastIndexOf("."));
	    br = new BufferedReader(new FileReader(mRNAFileName.replace(suffix, "_" + c + suffix)));
	    String line = br.readLine();
	    //chr loc strand nReads [total_len]
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)\\s*(\\d*)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    boolean strand = m.group(3).equals("F");
		    int pos = Integer.parseInt(m.group(2));
		    if (chr==c && strand==forward) { //22 is the splice chromosome
			TranscribedRegion2 repeatReg = getOverlappingTranscript(chr, pos - (strand ? 0 : readLen), pos + (strand ? readLen : 0), strand, repeatRNAs);
			if (repeatReg!=null) { 
			    if (!repeatRegionReads.containsKey(repeatReg))
				repeatRegionReads.put(repeatReg, new Vector<ABIRead>());
			    nReads = Integer.parseInt(m.group(4));
			    if (forward) {
				for (int j=0; j<(flatten ? 1 : nReads); j++) {
				    if (m.group(5)!=null && !m.group(5).equals("") && Integer.parseInt(m.group(5))>0)
					repeatRegionReads.get(repeatReg).addElement(new ABIRead(chr, true, pos, Integer.parseInt(m.group(5))/nReads));
				    else
					repeatRegionReads.get(repeatReg).addElement(new ABIRead(chr, true, pos, readLen));
				}
			    }
			    else {
				for (int j=0; j<(flatten ? 1 : nReads); j++) {
				    if (m.group(5)!=null && !m.group(5).equals("") && Integer.parseInt(m.group(5))>0)
					repeatRegionReads.get(repeatReg).addElement(new ABIRead(chr, false, pos - Integer.parseInt(m.group(5))/nReads, Integer.parseInt(m.group(5))/nReads));
				    else
					repeatRegionReads.get(repeatReg).addElement(new ABIRead(chr, false, pos - readLen, readLen));
				}
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    //Now we need to create arrays of reads and store with the repeat regions
	    for (TranscribedRegion2 t : repeatRegionReads.keySet()) {
		ABIRead[] reads = new ABIRead[repeatRegionReads.get(t).size()];
		int i = 0;
		for (ABIRead r : repeatRegionReads.get(t)) 
		    reads[i++] = r;
		Arrays.sort(reads);
		t.setFactorReads(mRNAFileName, reads);
		//System.out.println("Added " + reads.length + " reads to " + t.toString());
	    }
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	return repeatRNAs;
    }

    public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, boolean forward, int c, int readLen) { return loadMRNAReadsChr(mRNAFileName, forward, c, readLen, true, null); }

    public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, boolean forward, int c, int readLen, boolean flatten) { return loadMRNAReadsChr(mRNAFileName, forward, c, readLen, flatten, null); }

    //public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, boolean forward, int c, int readLen, boolean flatten) { return loadMRNAReadsChr(mRNAFileName, forward, c, readLen, flatten, null); }

    //public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, int c, int readLen) { return loadMRNAReadsChr(mRNAFileName, forward, c, readLen, true); }
    
    public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, boolean forward, int c, int readLen, boolean flatten, String repeatFileName) {
	ABIRead[] reads = null;
	if (mRNAFileName.contains("Plasmodium"))
	    return loadNIHReads(mRNAFileName, c);
	try {
	    //Parse the list of repeat regions
	    String species = repeatFileName!=null ? repeatFileName.substring(0, 2) : null;
	    TreeSet<TranscribedRegion2> repeatRNAs = repeatFileName!=null ? EstimateGenicTranscription3.parseRepeatFile(repeatFileName, species, c, forward) : new TreeSet<TranscribedRegion2>();

	    int nReads = 0;
	    BufferedReader br = null;
	    try {
		if (!mRNAFileName.endsWith(".tsv"))
		    mRNAFileName += ".tsv";
		String suffix = mRNAFileName.substring(mRNAFileName.lastIndexOf("."));
		br = new BufferedReader(new FileReader(mRNAFileName.replace(suffix, "_" + c + suffix)));
	    }
	    catch (IOException ex) {
		System.err.println("Error reading file: " + ex.toString());
		StackTraceElement s[] = ex.getStackTrace();
		for(int i=0; i<s.length; i++)
		    System.err.println(s[i].toString());
		br = new BufferedReader(new FileReader(mRNAFileName));
	    }
	    String line = br.readLine();
	    //chr loc strand nReads [total_len]
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)\\s*(\\d*)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    boolean strand = m.group(3).equals("F");
		    int pos = Integer.parseInt(m.group(2));
		    if (chr==c && strand==forward) { //22 is the splice chromosome
			if (!overlapsTranscript(chr, pos - (strand ? 0 : readLen), pos + (strand ? readLen : 0), strand, repeatRNAs))
			    nReads += flatten ? 1 : Math.min(Integer.parseInt(m.group(4)), MAX_READS_PER_LOCUS);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    if (nReads==0) {
		if (mRNAFileName.contains("Plasmodium"))
		    return loadNIHReads(mRNAFileName, c);
		else
		    return loadBroadReads(mRNAFileName, c, flatten);
	    }
	    reads = new ABIRead[nReads];
	    try {
		br = new BufferedReader(new FileReader(mRNAFileName.replace(".tsv", "_" + c + ".tsv")));
		//System.out.println("Opened file " + mRNAFileName.replace(".tsv", "_" + c + ".tsv"));
	    }
	    catch (IOException ex) {
		br = new BufferedReader(new FileReader(mRNAFileName));
		StackTraceElement s[] = ex.getStackTrace();
		for(int i=0; i<s.length; i++)
		    System.err.println(s[i].toString());
		System.out.println("Opened file " + mRNAFileName);
	    }
	    line = br.readLine();
	    int i = 0;
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    boolean strand = m.group(3).equals("F");
		    if (chr==c && strand==forward) { //22 is the splice chromosome
			nReads = Math.min(Integer.parseInt(m.group(4)), MAX_READS_PER_LOCUS);
			int pos = Integer.parseInt(m.group(2));
			if (forward && !overlapsTranscript(chr, pos, pos + readLen, strand, repeatRNAs)) {
			    for (int j=0; j<(flatten ? 1 : nReads); j++) {
				if (m.group(5)!=null && !m.group(5).equals("") && Integer.parseInt(m.group(5))>0)
				    reads[i++] = new ABIRead(chr, true, pos, Integer.parseInt(m.group(5))/nReads);
				else
				    reads[i++] = new ABIRead(chr, true, pos, readLen);
			    }
			}
			else if (!forward && !overlapsTranscript(chr, pos - readLen, pos, strand, repeatRNAs)) {
			    //System.out.println(line + "\t" + nReads);
			    for (int j=0; j<(flatten ? 1 : nReads); j++) {
				if (m.group(5)!=null && !m.group(5).equals("") && Integer.parseInt(m.group(5))>0)
				    reads[i++] = new ABIRead(chr, false, pos - Integer.parseInt(m.group(5))/nReads, Integer.parseInt(m.group(5))/nReads);
				else
				    reads[i++] = new ABIRead(chr, false, pos - readLen, readLen);
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    Arrays.sort(reads);
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	return reads;
    }

    public static boolean overlapsTranscript(int chr, int start, int end, boolean strand, TreeSet<TranscribedRegion2> segs) {
	for (TranscribedRegion2 seg : segs) {
	    if (overlapsTranscript(chr, start, end, strand, seg))
		return true;
	}
	return false;
    }

    public static boolean overlapsTranscript(int chr, int start, int end, boolean strand, TranscribedRegion2 seg) {
	if (seg.getChromosome()==chr && seg.getStrand()==strand) {
	    if ((start<seg.getStart() && end>seg.getStart()) || (start<seg.getEnd() && end>seg.getEnd()) || (seg.getStart()<start && seg.getEnd()>end) || (start<seg.getStart() && end>seg.getEnd()))
		return true;
	}
	return false;
    }

    public static boolean overlapsTranscript(int chr, int start, int end, TreeSet<TranscribedRegion2> segs) {
	for (TranscribedRegion2 seg : segs) {
	    if (overlapsTranscript(chr, start, end, seg))
		return true;
	}
	return false;
    }

    public static boolean overlapsTranscript(int chr, int start, int end, TranscribedRegion2 seg) {
	if (seg.getChromosome()==chr) {
	    if ((start<seg.getStart() && end>seg.getStart()) || (start<seg.getEnd() && end>seg.getEnd()) || (seg.getStart()<start && seg.getEnd()>end) || (start<seg.getStart() && end>seg.getEnd()))
		return true;
	}
	return false;
    }

    public static TranscribedRegion2 getOverlappingTranscript(int chr, int start, int end, boolean strand, TreeSet<TranscribedRegion2> segs) {
	for (TranscribedRegion2 seg : segs) {
	    if (overlapsTranscript(chr, start, end, strand, seg))
		return seg;
	}
	return null;
    }

    public static ABIRead[] loadBroadReads(String mRNAFileName, int c, boolean flatten) { return loadBroadReads(mRNAFileName, c, 36, flatten); }
    
    public static ABIRead[] loadBroadReads(String mRNAFileName, int c, int readLen, boolean flatten) {
	//Use this method to load the Yeast files that we got from the
	//Broad people. The data is not strand specific, so we want to
	//read both at the same time.
	ABIRead[] reads = null;
	try {
	    System.out.println("Loading Broad data...");
	    int nReads = 0;
	    BufferedReader br = new BufferedReader(new FileReader(mRNAFileName));
	    String line = br.readLine();
	    //seq:nReads chr:loc:strand:quality
	    Pattern mRNAPattern = Pattern.compile("([A|C|G|T]+):(\\d+)\\s+chr([M|\\d]+):(\\d+):([F|R]):([-|\\d]*)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = m.group(3).equals("M") ? 17 : Integer.parseInt(m.group(3));
		    if (chr==c) 
			nReads += flatten ? 1 : Math.min(Integer.parseInt(m.group(2)), MAX_READS_PER_LOCUS);
		}
		//else
		    //System.err.println("Error matching line: " + line);
		line = br.readLine();
	    }
	    br.close();
	    reads = new ABIRead[nReads];
	    br = new BufferedReader(new FileReader(mRNAFileName));
	    line = br.readLine();
	    int i = 0;
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = m.group(3).equals("M") ? 17 : Integer.parseInt(m.group(3));
		    if (chr==c) { 
			nReads = Math.min(Integer.parseInt(m.group(2)), MAX_READS_PER_LOCUS);
			if ((m.group(5).equals("F"))) {
			    for (int j=0; j<(flatten ? 1 : nReads); j++) 
				reads[i++] = new ABIRead(chr, true, Integer.parseInt(m.group(4)), readLen);
			}
			else if (m.group(5).equals("R")) {
			    //System.out.println(line + "\t" + nReads);
			    for (int j=0; j<(flatten ? 1 : nReads); j++) 
				reads[i++] = new ABIRead(chr, false, Integer.parseInt(m.group(4)) - readLen, readLen);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	Arrays.sort(reads);
	return reads;
    }

    public static ABIRead[] loadNIHReads(String mRNAFileName, int c) { return loadNIHReads(mRNAFileName, c, 35); }
    
    public static ABIRead[] loadNIHReads(String mRNAFileName, int c, int readLen) {
	//Use this method to load the Plasmodium files that we got from the
	//NIH people. 
	ABIRead[] reads = null;
	try {
	    System.out.println("Loading .bed data...");
	    int nReads = 0;
	    BufferedReader br = new BufferedReader(new FileReader(mRNAFileName));
	    String line = br.readLine();
	    //chr:start:end:code:score:strand (ignore 4th and 5th column)
	    Pattern mRNAPattern = Pattern.compile("^chr(\\d+)\\s+(\\d+)\\s+(\\d+)\\s*.*\\s*\\d*\\s*([+|-])");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    nReads += chr==c ? 1 : 0;
		}
		line = br.readLine();
	    }
	    br.close();
	    reads = new ABIRead[nReads];
	    br = new BufferedReader(new FileReader(mRNAFileName));
	    line = br.readLine();
	    int i = 0;
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    boolean strand = m.group(4).equals("+");
		    if (chr==c)
			reads[i++] = new ABIRead(chr, strand, strand ? Integer.parseInt(m.group(2)) : Integer.parseInt(m.group(3)) - readLen, readLen);
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	System.out.println("Found a total of " + reads.length);
	Arrays.sort(reads);
	//saveBedFile(mRNAFileName.substring(0, mRNAFileName.lastIndexOf(".")) + "_starts_" + c + ".bed", reads, !forward, true, c!=1 ? null : false, mRNAFileName.substring(mRNAFileName.lastIndexOf("/") + 1, mRNAFileName.lastIndexOf(".")) + "_starts"), "pf");
	return reads;
    }

    public static ABIRead[] loadKleinsteinReads(String mRNAFileName, String masterTableFileName, boolean etx, boolean forward, int c) {
	//Use this method to load the files from the Yale experiment. The files were produced using Bowtie, and the format is not very straightforward as the information resides in two separate files.
	ABIRead[] reads = null;
	try {
	    TreeMap<Integer, ABIRead> seq2read = new TreeMap<Integer, ABIRead>();
	    BufferedReader br = new BufferedReader(new FileReader(masterTableFileName));
	    String line = br.readLine();
	    Pattern mastertablePattern = Pattern.compile("seq(\\d+)\\s+([0|16]+)\\s+chr(\\w+)\\s+(\\d+)\\s+\\d+\\s+\\d+M\\s+\\*\\s+0\\s+0\\s+(\\w+)");
	    while (line!=null) {
		Matcher m = mastertablePattern.matcher(line);
		if (m.find()) {
		    int chr = -1;
		    if (m.group(3).equals("X"))
			chr = 20;
		    else if (m.group(3).equals("Y"))
			chr = 21;
		    else if (!m.group(3).equals("M"))		       
			chr = Integer.parseInt(m.group(3));
		    //0 means forward strand and 16 means reverse
		    boolean strand = Integer.parseInt(m.group(2))==0;
		    if (chr==c && strand==forward)
			seq2read.put(Integer.parseInt(m.group(1)), new ABIRead(c, strand, Integer.parseInt(m.group(4)), m.group(5).length()));
			}
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("Found " + seq2read.size() + " different sequences " + seq2read.keySet().iterator().next());
	    //Now that we have found all the separate reads, we need to figure out how many of them there were in the given experiment
	    int nTotalReads = 0, nDifferentLoci = 0, j = 0;
	    TreeMap<ABIRead, Integer> reads2count = new TreeMap<ABIRead, Integer>();
	    br = new BufferedReader(new FileReader(mRNAFileName));
	    Pattern numberPattern = Pattern.compile("seq(\\d+)\\s+(\\w+)\\s+(\\d+)\\s+(\\d+)");
	    line = br.readLine();
	    while (line!=null) {
		Matcher m = numberPattern.matcher(line);
		if (m.find()) {
		    int seq = Integer.parseInt(m.group(1));
		    if (seq2read.containsKey(seq)) {
			int nReads = Integer.parseInt(m.group(etx ? 3 : 4));
			if (nReads>0) {
			    nTotalReads += nReads;
			    if (reads2count.containsKey(seq2read.get(seq)))
				reads2count.put(seq2read.get(seq), reads2count.get(seq2read.get(seq)) + nReads);
			    else {
				reads2count.put(seq2read.get(seq), nReads);
				nDifferentLoci++;
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("Found " + nTotalReads + " reads for " + reads2count.size() + " and " + nDifferentLoci);
	    //Now we can finally create all the reads
	    reads = new ABIRead[nTotalReads];
	    for (ABIRead r : reads2count.keySet()) {
		for (int i=0; i<reads2count.get(r); i++)
		    reads[j++] = new ABIRead(r);
	    }
	    if (j!=nTotalReads)
		System.out.println("nTotalReads = " + nTotalReads + " and j = " + j);
	}
	catch (IOException e) {
	    System.err.println("Error reading Kleinstein files " + e.toString());
	}
	return reads;
    }

    //////////////////////////////////////////////////////////////////

    //For extracting a subset of reads
    public static ABIRead[] getReads(int chr, int start, int end, ABIRead reads[]) {
	int ind0 = Arrays.binarySearch(reads, new ABIRead(chr, true, start), null);
	int ind = ind0<0 ? -ind0-1 : ind0;
	int nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    ind++;
	}
	ABIRead rs[] = new ABIRead[nReads];
	ind = ind0<0 ? -ind0-1 : ind0;
	for (int i=0; i<nReads; i++)
	    rs[i] = reads[ind + i];
	return rs;
    }

    public TreeMap<Integer, Integer> calculateReadLengthDistribution() { return calculateReadLengthDistribution(reads); }

    public static TreeMap<Integer, Integer> calculateReadLengthDistribution(ABIRead reads[]) {
	TreeMap<Integer, Integer> lens = new TreeMap<Integer, Integer>();
	for (int i=0; i<reads.length; i++)
	    lens.put(reads[i].getLength(), lens.containsKey(reads[i].getLength()) ? lens.get(reads[i].getLength()) + 1 : 1);
	return lens;
    }

    public static double meanReadLength(ABIRead reads[]) {
	double n = 0;
	for (int i=0; i<reads.length; i++)
	    n += (double)reads[i].getLength()/reads.length;
	return n;
    }

    public static TreeMap<Integer, Integer> calculateReadCounts(int chr, ABIRead reads[], int binSize, String species, boolean chip) {
	//Use non-overlapping bins to tile the genome and count the
	//number of reads in each
	TreeMap<Integer, Integer> hist = new TreeMap<Integer, Integer>();
	for (int i=0; i<TranscriptFinder.numberOfBins(binSize, chr, species); i++) {
	    int start = i*binSize, end = (i+1)*binSize;
	    int nRNA = chip ? FindReadsAcrossFeature.numberOfReads(chr, start, end, reads[0].getLength(), reads) : FindReadsAcrossFeature.numberOfReads(chr, start, end, reads);
	    hist.put(nRNA, hist.containsKey(nRNA) ? hist.get(nRNA) + 1 : 1);
	}
	return hist;
    }

    public static TreeMap<Double, Integer> calculateReadDensity(int chr, ABIRead reads[], int binSize, double meanReadLength, int maxReadLen, String species, boolean chip, boolean log) {
	//Use non-overlapping bins to tile the genome and count the
	//number of reads in each
	TreeMap<Double, Integer> dens = new TreeMap<Double, Integer>();
	for (int i=0; i<TranscriptFinder.numberOfBins(binSize, chr, species); i++) {
	    int start = i*binSize, end = (i+1)*binSize;
	    double nRNA = 0;
	    ABIRead rs[] = ABIReads.getReads(chr, start, end, reads);
	    if (rs.length>0) {
		double hist[] = TranscriptFinder.calculateHistogram(rs, maxReadLen, reads[0].getStrand()); //get the histogram
		for (int j=0; j<hist.length; j++)
		    nRNA += log ? Math.log(hist[j] + 1) : hist[j];
		//double n = meanReadLength*((double)(chip ? FindReadsAcrossFeature.numberOfReads(chr, start, end, (int)meanReadLength, reads) : FindReadsAcrossFeature.numberOfReads(chr, start, end, reads)));
		//double nRNA = log ? Math.log(1 + n)/binSize : n/binSize;
		//double nRNA = Math.log(1 + ((double)(chip ? FindReadsAcrossFeature.numberOfReads(chr, start, end, (int)meanReadLength, reads) : FindReadsAcrossFeature.numberOfReads(chr, start, end, reads)))) + Math.log(meanReadLength) - Math.log(binSize);
		
	    }
	    else
		nRNA = 0.0;
	    dens.put(nRNA/binSize, dens.containsKey(nRNA/binSize) ? dens.get(nRNA/binSize) + 1 : 1);
	}
	return dens;
    }

    public static TreeMap<Double, Integer> calculateReadDensityVariance(int chr, ABIRead reads[], int binSize, double meanReadLength, int maxReadLen, String species, boolean chip, boolean log) {
	//Use non-overlapping bins to tile the genome and count the
	//number of reads in each
	TreeMap<Double, Integer> dens = new TreeMap<Double, Integer>();
	for (int i=0; i<TranscriptFinder.numberOfBins(binSize, chr, species); i++) {
	    int start = i*binSize, end = (i+1)*binSize;
	    double nRNA = 0, varRNA = 0;
	    ABIRead rs[] = ABIReads.getReads(chr, start, end, reads);
	    if (rs.length>0) {
		double hist[] = TranscriptFinder.calculateHistogram(rs, maxReadLen, reads[0].getStrand()); //get the histogram
		for (int j=0; j<hist.length; j++) 
		    nRNA += log ? Math.log(hist[j] + 1) : hist[j];
		double d = nRNA/binSize;
		for (int j=0; j<hist.length; j++) 
		    varRNA += log ? (Math.log(hist[j] + 1) - d)*(Math.log(hist[j] + 1) - d) : (hist[j] - d)*(hist[j] - d);
	    }
	    else
		varRNA = 0.0;
	    varRNA /= binSize - 1;
	    dens.put(varRNA, dens.containsKey(varRNA) ? dens.get(varRNA) + 1 : 1);
	}
	return dens;
    }

    //////////////////////////////////////////////////////////////////
    
    public static String getChromosomeString(int c, String species) { return "chr" + getChromosomeNumber(c, species); }

    public static String getChromosomeNumber(int c, String species) {
	if (species.toLowerCase().equals("mm")) {
	    if (c==20)
		return "X";
	    else if (c==21)
		return "Y";
	    else if(c>0 && c<20)
		return String.valueOf(c);
	}
	if (species.toLowerCase().equals("hs")) {
	    if (c==23)
		return "X";
	    else if (c==24)
		return "Y";
	    else if(c>0 && c<23)
		return String.valueOf(c);
	}
	else if (species.toLowerCase().equals("sc")) {
	    if (c==17)
		return "M";
	    else if(c>0 && c<17)
		return String.valueOf(c);
	}
	else if (species.toLowerCase().equals("pf")) {
	    if(c>0 && c<15)
		return String.valueOf(c);
	}
	System.err.println("Chr number " + c + " is not allowed for " + species);
	return new String("-1");
    }

    public static String getChromosomeNumber(int c) {
	return getChromosomeNumber(c, "mm");
    }

    ////////////////////////////////////////////

    public static void saveBedFile(String fileName, ABIRead reads[], boolean reverse, boolean log, boolean extend, String desc, String species) {
	try {
	    String chrStr = getChromosomeString(reads[0].getChromosome(), species);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    if (desc!=null)
		bw.write("track type=wiggle_0 name=\"" + desc + "\" description=\"" + desc + "\" visibility=full autoScale=off viewLimits=" + (reverse ? "-5:0" : "0:5") + " color=\"255,127,0\" maxHeightPixels=70:40:10 priority=1\n");
	    if (extend) {
		
	    }
	    else {
		//This is easier since we only save the 5'-most location
		int ind = 0;
		while (ind<reads.length) {
		    int n = 0, pos = reads[ind].getPosition();
		    do {
			ind++;
			n++;
		    } while (ind<reads.length && pos==reads[ind].getPosition());
		    bw.write(chrStr + "\t" + pos + "\t" + (pos + 1) + "\t" + (log ? (reverse ? -Math.log(n + 1) : Math.log(n + 1)) : (reverse ? -n : n)) + "\n");
		    //System.out.print(chrStr + "\t" + pos + "\t" + (pos + 1) + "\t" + (log ? (reverse ? -Math.log(n + 1) : Math.log(n + 1)) : (reverse ? -n : n)) + "\n");
		}
	    }
	    bw.close();
	    System.out.println("Saving .bed-file to " + fileName);
	}
	catch (IOException e) {
	    System.err.println("Error saving ABIReads to " + fileName + " :" + e.toString());
	}
    }

}
