//Use this to run the analysis of the gene-enhancer network. Most of
//the work is done by the GeneEnhancerGraph.java class which contains
//all the info about the genes, enhancers and reads. This method
//allows us to look at interesting subsets and compute meaningful
//statistics, including:
//
//1. Correlation between the amounts of different factors at genes and enhancers
//
//2. Enrichment when responding to stimuli

//java -Xmx1024m AnalyzeGeneEnhancers -n PeakFiles/enhancer_factor_names.txt

//use analyze_gene_enhancers.m to post-process the data.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class AnalyzeGeneEnhancers {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static final String GENE_ENHANCER_DISTANCE_SUFFIX = "_gene_enhancer_distances";
    public static final String GENE_ENHANCER_NEAREST_DISTANCE_SUFFIX = "_gene_enhancer_nearetst_distances";
    public static final String GENE_ENHANCER_SUFFIX = "_gene_enhancer";

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    //Location of the enhancer list
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption('e', "enhancerFileName");
	    //Consider the density at TSSs as well
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option maxDistanceFlag = argParser.addIntegerOption('d', "maxDistance");
	    CmdLineParser.Option maxDistanceStepFlag = argParser.addIntegerOption('D', "maxDistanceStep");
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption('N', "nMC");
	    CmdLineParser.Option minReadsFlag = argParser.addIntegerOption('r', "minReads");
	    CmdLineParser.Option geneSubsetFlag = argParser.addStringOption('g', "geneSubset");
	    CmdLineParser.Option nonGenicFlag = argParser.addBooleanOption("nonGenic");
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption("isolated");
	    CmdLineParser.Option nearestTSSOnlyFlag = argParser.addBooleanOption("nearestTSSOnly");
	    //Use this flag when we want to analyze the lengths of the
	    //transcribed regions at enhancers. Only meaningful for WT-seq
	    CmdLineParser.Option transcribedRegionLengthsFlag = argParser.addIntegerOption("transcribedRegionLengths");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : "";
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1500;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/enhancers_null_KCl.txt";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : enhancerFileName.substring(0, enhancerFileName.lastIndexOf("."));
	    tmp = argParser.getOptionValue(maxDistanceFlag);
	    int maxDist = tmp!=null ? ((Integer)tmp).intValue() : 1000000;
	    tmp = argParser.getOptionValue(maxDistanceStepFlag);
	    int maxDistStep = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(minReadsFlag);
	    int minReads = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(geneSubsetFlag);
	    String geneSubset = tmp!=null ? (String)tmp : "all";
	    boolean nonGenic = argParser.getOptionValue(nonGenicFlag)!=null;
	    //use only the genes that are at least maxDist from nearest gene
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    //Use if we only want to assign each enhancer with a single (nearest) TSS
	    boolean nearestTSSOnly = argParser.getOptionValue(nearestTSSOnlyFlag)!=null;
	    tmp = argParser.getOptionValue(transcribedRegionLengthsFlag);
	    int minDistTranscribedRegion = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21, nTSSs = 0, fragmentLength = 120;
	    double qualityScore = 10;
	    ////////////////////////////////////////////
	    //Figure out which geneIDs to keep
	    int geneIDsKeep[] = null;
	    if (geneSubset.equals("up_KCl")) {
		//geneIDsKeep = ExpressionEnhancerCorrelation.parseSymbolicExpressionFile(new String[]{"LHH", "LLH", "LMH", "LHM", "LMM"}, orchestra);
		geneIDsKeep = parseGeneIDFile("mRNA/IndRep/induced_KCl.txt");
		Arrays.sort(geneIDsKeep);
		System.out.println("Found " + geneIDsKeep.length + " genes to analyze.");
	    }
	    if (geneSubset.equals("down_KCl")) {
		//geneIDsKeep = ExpressionEnhancerCorrelation.parseSymbolicExpressionFile(new String[]{"LHH", "LLH", "LMH", "LHM", "LMM"}, orchestra);
		geneIDsKeep = parseGeneIDFile("mRNA/IndRep/repressed_KCl.txt");
		Arrays.sort(geneIDsKeep);
		System.out.println("Found " + geneIDsKeep.length + " genes to analyze.");
	    }
	    ///////////////////////////////////////////////////////
	    Vector<String> fileNames = fileName.contains("names") ? new Vector<String>(FindOverlappingPeaks.getFileNames(fileName).keySet()) : new Vector<String>();
	    if (!fileName.contains("names") && !fileName.equals(""))
		fileNames.addElement(fileName);
	    //Keep track of the total number of reads for normalization purposes
	    TreeMap<String, Integer> nTotalReads = new TreeMap<String, Integer>();
	    for (String f : fileNames)
		nTotalReads.put(f, 0);
	    //For checking the number of enhancers near genes at
	    //different distances
	    Vector<TreeMap<Integer, Integer>> nEnhancersDist = new Vector<TreeMap<Integer, Integer>>(), nEnhancersDistMC = new Vector<TreeMap<Integer, Integer>>();
	    //Find the nearest TSS for each enhancer
	    int nearestTSSDist[] = new int[maxDist/maxDistStep], nearestTSSDistMC[] = new int[maxDist/maxDistStep];
	    for (int j=1; j<=maxDist/maxDistStep; j++) {
		nEnhancersDist.addElement(new TreeMap<Integer, Integer>());
		nEnhancersDistMC.addElement(new TreeMap<Integer, Integer>());
	    }
	    //For checking the amount of RNA/TF at enhancers
	    TreeMap<String, Vector<Vector<Double>>> factorGeneEnhancer = new TreeMap<String, Vector<Vector<Double>>>();
	    GeneEnhancerGraph graphs[] = new GeneEnhancerGraph[nChr];
	    for (int i=1; i<=nChr; i++) {
		if (i==chr || chr==-1) {
		    graphs[i-1] = new GeneEnhancerGraph(i, geneIDsKeep, isolated ? maxDist : 0, enhancerFileName, nonGenic, fileName, windowSize, orchestra);
		    //Find the enhancers within a given range
		    nEnhancersDist = findGeneEnhancerDist(graphs[i-1], maxDist, maxDistStep, nEnhancersDist);
		    //Find the histogram showing the nearest TSS distance
		    nearestTSSDist = findNearestTSSDistance(graphs[i-1], maxDist, maxDistStep, nearestTSSDist);
		    /////////////////////////////////
		    //Do the MC shuffling of the enhancer sites
		    //nEnhancersDistMC = findGeneEnhancerDist(new GeneEnhancerGraph(i, geneIDsKeep, isolated ? maxDist : 0, enhancerFileName, nonGenic, fileName, windowSize, "uniform", orchestra), maxDist, maxDistStep, nEnhancersDistMC);
		    for (int j=0; j<nMC; j++)
			nearestTSSDistMC = findNearestTSSDistance(new GeneEnhancerGraph(i, geneIDsKeep, isolated ? maxDist : 0, enhancerFileName, nonGenic, fileName, windowSize, "uniform", orchestra), maxDist, maxDistStep, nearestTSSDistMC);
		}
	    }
	    if (minDistTranscribedRegion<=0) {
		//Find the factors at enhancers and nearby genes
		for (String fName : fileNames) {
		    String saveName = saveFileName + "_" + geneSubset + (nonGenic ? "_non_genic" : "") + (isolated ? "_isolated" : "") + GENE_ENHANCER_SUFFIX + "_" + getShortFactorName(fName) + (chr==-1 ? "" : "_" + chr) + ".txt";
		    String saveNameMC = saveFileName + "_" + geneSubset + (nonGenic ? "_non_genic" : "") + (isolated ? "_isolated" : "") + "_MC" + GENE_ENHANCER_SUFFIX + "_" + getShortFactorName(fName) + (chr==-1 ? "" : "_" + chr) + ".txt";
		    System.out.println("Saving to " + saveName);
		    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
		    BufferedWriter bwMC = new BufferedWriter(new FileWriter(saveNameMC));
		    for (int i=1; i<=nChr; i++) 
			nTotalReads.put(fName, nTotalReads.get(fName) + (fName.contains("RNA") ? FindNongenicTranscripts.loadMRNAReadsChr(fName, i).length : PeakFinder2.parseFiles(i, fName, fragmentLength, true).length));
		    double normFactor = (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(fName);
		    for (int i=1; i<=nChr; i++) {
			if (i==chr || chr==-1) {
			    graphs[i-1].addEdges(0, maxDist);
			    for (RefSeqGene g : graphs[i-1].getGenes()) {
				bw.write(g.getID() + "\t" + format.format(normFactor*(double)g.getFactorReads(fName).length) + "\t");
				bwMC.write(g.getID() + "\t" + format.format(normFactor*(double)g.getFactorReads(fName).length) + "\t");
				//Get all the enhancers, write the enrichment and the distance
				for (Enhancer en : g.getNearbyEnhancers()) {
				    int dist = g.getStrand() ? en.getCenter() - g.getStart() : g.getStart() - en.getCenter();
				    if (!nearestTSSOnly || (nearestTSSOnly && en.getNearestTSS()!=null && en.getNearestTSS().equals(g))) {
					bw.write(format.format(normFactor*(double)en.getFactorReads(fName).length) + "\t" + dist + "\t");
					//Do the shuffle control
					bwMC.write(format.format(normFactor*(double)graphs[i-1].getRandomEnhancer().getFactorReads(fName).length) + "\t" + dist + "\t");
				    }
				}
				bw.write("\n");
				bwMC.write("\n");
			    }
			}
		    }
		    bw.close();
		    bwMC.close();
		}
	    }
	    if (minDistTranscribedRegion>0) {
		for (String fName : fileNames) 
		    findTranscribedRegionLengths(graphs, fName, minDistTranscribedRegion, saveFileName + "_" + geneSubset + (nonGenic ? "_non_genic" : "") + (isolated ? "_isolated" : "") + FindEnhancerSites.TRANSCRIBED_REGION_LENGTH_SUFFIX + "_" + getShortFactorName(fName) + (chr==-1 ? "" : "_" + chr) + ".txt");
	    }
	    ///////////////////////////////////////////////////////
	    //Save stuff
	    saveGeneEnhancerDistances(saveFileName + "_" + geneSubset + (nonGenic ? "_non_genic" : "") + (isolated ? "_isolated" : "") + GENE_ENHANCER_DISTANCE_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt" , nEnhancersDist, maxDistStep, maxDist);
	    String saveName = saveFileName + "_" + geneSubset + (nonGenic ? "_non_genic" : "") + (isolated ? "_isolated" : "") + GENE_ENHANCER_NEAREST_DISTANCE_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    System.out.println("Saving to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    if (chr==1 || chr==-1) {
		bw.write("#");
		for (int j=1; j<=maxDist/maxDistStep; j++) 
		    bw.write((maxDistStep*j) + "\t");
		bw.write("\n");
	    }
	    for (int j=0; j<nearestTSSDist.length; j++) 
		bw.write(nearestTSSDist[j] + "\t");
	    bw.write("\n");
	    for (int j=0; j<nearestTSSDistMC.length; j++) 
		bw.write(format.format((double)nearestTSSDistMC[j]/nMC) + "\t");
	    bw.write("\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static void findTranscribedRegionLengths(GeneEnhancerGraph[] graphs, String fName, int minDist, String saveFileName) throws IOException {
	//For keeping track of the transcribed region lengths
	Vector<Integer> transcribedRegionLengths = new Vector<Integer>();
	Vector transcribedRegionNumbers[] = new Vector[4];
	for (int i=0; i<graphs.length; i++) {
	    if (graphs[i]!=null) {
		transcribedRegionLengths.addAll(graphs[i].getTranscribedRegionLengths(fName, minDist));
		for (int j=0; j<4; j++) {
		    if (transcribedRegionNumbers[j]==null)
			transcribedRegionNumbers[j] = new Vector<Integer>();
		    transcribedRegionNumbers[j].addAll(graphs[i].getTranscribedRegionNumbers(fName, minDist*(j+1)/4));
		}						       
	    }
	}
    	BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	for (Integer i : transcribedRegionLengths)
	    bw.write(i + "\n");
	bw.close();
	bw = new BufferedWriter(new FileWriter(saveFileName.replace(FindEnhancerSites.TRANSCRIBED_REGION_LENGTH_SUFFIX, FindEnhancerSites.TRANSCRIBED_REGION_NUMBER_SUFFIX)));
	for (int i=0; i<transcribedRegionNumbers[0].size(); i++) {
	    for (int j=0; j<4; j++)
		bw.write(((Integer)transcribedRegionNumbers[j].elementAt(i)).toString() + "\t");
	    bw.write("\n");
	}
	bw.close();
    }

    public static void findTranscribedRegionLengthsMC(GeneEnhancerGraph[] graphs, String fName, int minDist, String saveFileName, int nMC) throws IOException {
	//For keeping track of the transcribed region lengths
	Vector<Integer> transcribedRegionLengthsMC = new Vector<Integer>();
	for (int i=0; i<graphs.length; i++) {
	    if (graphs[i]!=null) {
		for (int j=0; j<nMC; j++) {
		    GeneEnhancerGraph graphMC = new GeneEnhancerGraph(graphs[i]);
		    graphMC.removeAllEdges();
		    graphMC.removeEnhancers();
		    //graphMC.addRandomEnhancers(true);
		//graphMC.addEnhancerReads(i+1, fName, windowSize);
		//transcribedRegionLengthsMC.addAll(graphMC.getTranscribedRegionLengths(fName, minDist));
		}
	    }
	}
    	/*BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	for (Integer i : transcribedRegionLengths)
	    bw.write(i + "\n");
	    bw.close();*/
    }

    public static Vector<TreeMap<Integer, Integer>> findGeneEnhancerDist(GeneEnhancerGraph graph, int maxDist, int maxDistStep, Vector<TreeMap<Integer, Integer>> nEnhancersDist) {
	for (int j=1; j<=maxDist/maxDistStep; j++) {
	    graph.removeAllEdges();
	    graph.addEdges((j-1)*maxDistStep, j*maxDistStep);
	    //Add the histograms
	    for (int k : graph.getGeneEnhancerDistribution().keySet())
		nEnhancersDist.elementAt(j-1).put(k, graph.getGeneEnhancerDistribution().get(k) + (nEnhancersDist.elementAt(j-1).containsKey(k) ? nEnhancersDist.elementAt(j-1).get(k) : 0));
	}
	return nEnhancersDist;
    }

    public static int[] findNearestTSSDistance(GeneEnhancerGraph graph, int maxDist, int maxDistStep, int[] nearestTSSDist) {
	graph.removeAllEdges();
	graph.addEdges(0, maxDist);
	//Add the histograms
	for (Enhancer en : graph.getEnhancers()) {
	    int nearestTSS = Integer.MAX_VALUE;
	    RefSeqGene nearestGene = null;
	    for (RefSeqGene g : new Vector<RefSeqGene>(en.getNeighbors())) {
		if (Math.abs(en.getCenter() - g.getStart())<nearestTSS) {
		    nearestGene = g;
		    nearestTSS = Math.abs(en.getCenter() - g.getStart());
		}
	    }
	    if (nearestTSS<maxDist) {
		//System.out.println("Found gene " + nearestGene.toString() + " and enhancer at " + en.getCenter() + " with distance " + nearestTSS);
		nearestTSSDist[nearestTSS/maxDistStep]++;
	    }
	}
	return nearestTSSDist;
    }


    public static void saveGeneEnhancerDistances(String saveFileName, Vector<TreeMap<Integer, Integer>> nEnhancersDist, int maxDistStep, int maxDist) throws IOException {
	BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	bw.write("#\t"); //write the header
	for (int i=1; i<=maxDist/maxDistStep; i++)
	    bw.write((i*maxDistStep) + "\t");
	bw.write("\n");
	int nBinsMax = 0; //Find out which is the largest bin
	for (TreeMap<Integer, Integer> t : nEnhancersDist)
	    nBinsMax = Math.max(nBinsMax, t.size());
	for (int i=0; i<=nBinsMax; i++) {
	    bw.write(i + "\t");
	    for (TreeMap<Integer, Integer> t : nEnhancersDist)
		bw.write((t.containsKey(i) ? t.get(i) : 0) + "\t");
	    bw.write("\n");
	}
	bw.close();
    }

    public static int[] parseGeneIDFile(String fileName) throws IOException {
	int nGenes = 0;
	BufferedReader br = new BufferedReader(new FileReader(fileName));
	String line = br.readLine();
	Pattern idPattern = Pattern.compile("^(\\d+)\\s*");
	while (line!=null) {
	    Matcher m = idPattern.matcher(line);
	    if (m.find())
		nGenes++;
	    line = br.readLine();
	}
	br.close();
	int ids[] = new int[nGenes], i = 0;
	br = new BufferedReader(new FileReader(fileName));
	line = br.readLine();
	while (line!=null) {
	    Matcher m = idPattern.matcher(line);
	    if (m.find())
		ids[i++] = Integer.parseInt(m.group(1));
	    line = br.readLine();
	}
	br.close();
	return ids;
    }

    public static String getShortFactorName(String fName) {
	String shortFactorName = null;
	Pattern tfPattern = Pattern.compile("/([_|\\w]+)[\\.|\\w]*$");
	Matcher m = tfPattern.matcher(fName);
	if (m.find()) { shortFactorName = m.group(1); }
	return shortFactorName;
    }


}

