//Use this method to analyze the significance of the overlap between
//genes that are perturbed by either KCl or RNAi and the nearby
//binding. We are interested in both obtaining a list of high
//confidence targets that are affected by KCl, RNAi and Fos binding
//for experimental validation as well as finding the enrichment and
//increased probability of having fos-peaks near induced genes.
//
//java -Xmx2048m AnalyzePeaksKClRNAi -o -c 19
//
//

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class AnalyzePeaksKClRNAi {

    public static final int MIN_PEAK_DIST = 500;
    public static final String RNA_SEQ_NAMES[] = new String[]{"total0h", "total1h", "total6h"};


    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option maxDistanceFlag = argParser.addIntegerOption('d', "maxDist");	    
	    CmdLineParser.Option proximalDistanceFlag = argParser.addIntegerOption("proximal");	    
	    CmdLineParser.Option maxTFCombinationsFlag = argParser.addIntegerOption('t', "maxTFCombinations");	    
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option peakExpTableFlag = argParser.addBooleanOption("table");
	    CmdLineParser.Option targetListFlag = argParser.addBooleanOption("target");
	    CmdLineParser.Option bayesNetworkFlag = argParser.addBooleanOption("bayes");
	    CmdLineParser.Option nearestOnlyFlag = argParser.addBooleanOption('N', "nearestOnly");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option rnaSeqInductionFlag = argParser.addDoubleOption('i', "rnaSeqInduction");
	    CmdLineParser.Option rnaiInductionFlag = argParser.addDoubleOption('I', "rnaiInduction");
	    CmdLineParser.Option minExpressionFlag = argParser.addDoubleOption('e', "minExp");
	    CmdLineParser.Option pValFlag = argParser.addDoubleOption('p', "pVal");
	    CmdLineParser.Option edgePenaltyFlag = argParser.addDoubleOption('E', "edgePenalty");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean peakExpTable = argParser.getOptionValue(peakExpTableFlag)!=null;
	    boolean targetList = argParser.getOptionValue(targetListFlag)!=null;
	    boolean bayesNetwork = argParser.getOptionValue(bayesNetworkFlag)!=null;
	    boolean nearestGeneOnly = argParser.getOptionValue(nearestOnlyFlag)!=null;
	    //tmp = argParser.getOptionValue(saveFileNameFlag);
	    //String saveFileName = tmp!=null ? (String)tmp : fileName + "_input_normalization.txt";
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    tmp = argParser.getOptionValue(maxDistanceFlag);
	    int maxDist = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(proximalDistanceFlag);
	    int proximalDistance = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(maxTFCombinationsFlag);
	    int maxTFCombinations = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(rnaSeqInductionFlag);
	    double rnaSeqInduction = tmp!=null ? ((Double)tmp).doubleValue() : 4;
	    tmp = argParser.getOptionValue(rnaiInductionFlag);
	    double rnaiInduction = tmp!=null ? ((Double)tmp).doubleValue() : 2;
	    tmp = argParser.getOptionValue(minExpressionFlag);
	    double minExp = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(pValFlag);
	    double pVal = tmp!=null ? ((Double)tmp).doubleValue() : .01;
	    tmp = argParser.getOptionValue(edgePenaltyFlag);
	    double edgePenaltyCoeff = tmp!=null ? ((Double)tmp).doubleValue() : 1000;
	    boolean firstTSSOnly = false;
	    //////////////////////////////////////////////////////////////////
	    String fosRNAiFileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/Athar/Arrays/chip_data_genes_clean.txt";
	    String fosPeakProximalFileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/Matlab/Athar/ap1_fos_peaks_proximal_ids.txt";
	    String fosPeakDistalFileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/Matlab/Athar/ap1_fos_peaks_distal_ids.txt";
	    String intragenicEnhancersFileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/Matlab/Athar/intragenic_enhancers_ids.txt";
	    String extragenicEnhancersFileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/Matlab/Athar/extragenic_enhancers_ids.txt";
	    String proximalPeaksFileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/Matlab/Athar/tf_kcl_peak_locs_proximal_ids.txt";
	    String tfNs[] = new String[]{"cfos_sc7202_KCl_B2", "fosb_sc48_KCl_B2", "CBP_Millipore_KCl_B1", "Npas4_Npas4_KCl_B1", "Pol2_8WG16_KCl_B1", "CREB_upstate_KCl_B1", "SRF_H300_KCl_B1"};
	    TreeSet<String> tfNames = new TreeSet<String>(), targetTFNames = new TreeSet<String>();
	    for (int i=0; i<3/*tfNs.length*/; i++)
		tfNames.add(tfNs[i]);
	    //Use this list to extract a list of high confidence targets
	    targetTFNames.add(tfNs[0]);
	    targetTFNames.add(tfNs[2]);
	    targetTFNames.add(tfNs[3]);
	    targetTFNames.add(tfNs[4]);
	    int nChr = 21, nRNAiTimes = 4, nRNAiConditions = 5, proximalDist = 10000;
	    boolean force = false;
	    RefSeqGenes refSeqGenes[] = new RefSeqGenes[nChr];	    
	    RefSeqGenesPeakGraph genesPeaksGraph = new RefSeqGenesPeakGraph(nChr);
	    for (int i=1; i<=nChr; i++) {
		if (i==chr || chr==-1) {
		    refSeqGenes[i-1] = new RefSeqGenes(i, orchestra);
		    System.out.println("Loaded " + refSeqGenes[i-1].size());
		    //Load the RNA-Seq data
		    //loadRNASeqReads(refSeqGenes[i-1], i, force);
		    loadExonDensities(refSeqGenes[i-1], i, orchestra);
		    //Load the RNAi data
		    refSeqGenes[i-1].loadRNAiData(fosRNAiFileName);
		    //Load the peak data
		    TreeSet<Peak> fosPeaksProximal = parsePeakFile(fosPeakProximalFileName, i, tfNames);
		    TreeSet<Peak> fosPeaksDistal = parsePeakFile(fosPeakDistalFileName, i, tfNames);
		    //Load the enhancers
		    TreeSet<Peak> extragenicEnhancers = parsePeakFile(extragenicEnhancersFileName, i, tfNames);
		    TreeSet<Peak> intragenicEnhancers = parsePeakFile(intragenicEnhancersFileName, i, tfNames);
		    //Load the proximal peaks
		    TreeSet<Peak> proximalPeaks = parsePeakFile(proximalPeaksFileName, i, tfNames);
		    //Create a graph with peaks, enhancers and genes as nodes
		    genesPeaksGraph.addGenes(refSeqGenes[i-1], i);
		    genesPeaksGraph.addPeaks(i, fosPeaksDistal, maxDist, MIN_PEAK_DIST, firstTSSOnly, nearestGeneOnly);
		    genesPeaksGraph.addPeaks(i, fosPeaksProximal, maxDist, MIN_PEAK_DIST, firstTSSOnly, nearestGeneOnly);
		    genesPeaksGraph.addPeaks(i, extragenicEnhancers, maxDist, MIN_PEAK_DIST, firstTSSOnly, nearestGeneOnly);
		    genesPeaksGraph.addPeaks(i, intragenicEnhancers, maxDist, MIN_PEAK_DIST, firstTSSOnly, nearestGeneOnly);
		    genesPeaksGraph.addPeaks(i, proximalPeaks, maxDist, MIN_PEAK_DIST, firstTSSOnly, nearestGeneOnly);
		}
	    }
	    ////////////////////////////////////////////
	    //Now we can start extracting various subsets and
	    //ask how they compare to what one would expect by
	    //chance.
	    if (rnaiInduction>0 && rnaSeqInduction>0) {
		for (int j=0; j<nRNAiTimes; j++) {
		    for (int k=0; k<nRNAiConditions; k++) {
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, true, true, rnaiInduction, true, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, true, true, rnaiInduction, false, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, true, false, rnaiInduction, true, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, true, false, rnaiInduction, false, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, false, true, rnaiInduction, true, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, false, true, rnaiInduction, false, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, false, false, rnaiInduction, true, j, k, pVal);
			genesPeaksGraph.findInductionRNAiOverlap(minExp, rnaSeqInduction, false, false, rnaiInduction, false, j, k, pVal);
		    }
		}
	    }
	    if (rnaiInduction>0) {
		for (int j=0; j<nRNAiTimes; j++) {
		    for (int k=0; k<nRNAiConditions; k++) {
			genesPeaksGraph.findPeakRNAiOverlap(minExp, rnaiInduction, true, j, k, maxTFCombinations, tfNames, pVal);
			genesPeaksGraph.findPeakRNAiOverlap(minExp, rnaiInduction, false, j, k, maxTFCombinations, tfNames, pVal);
		    }
		}
	    }
	    if (rnaSeqInduction>0) {
		genesPeaksGraph.findPeakInductionOverlap(minExp, rnaSeqInduction, true, true, maxTFCombinations, tfNames, pVal);
		genesPeaksGraph.findPeakInductionOverlap(minExp, rnaSeqInduction, true, false, maxTFCombinations, tfNames, pVal);
		genesPeaksGraph.findPeakInductionOverlap(minExp, rnaSeqInduction, false, true, maxTFCombinations, tfNames, pVal);
		genesPeaksGraph.findPeakInductionOverlap(minExp, rnaSeqInduction, false, false, maxTFCombinations, tfNames, pVal);
	    }
	    if (peakExpTable) {
		genesPeaksGraph.peakInductionTable(minExp, Math.abs(rnaSeqInduction), true, true, tfNames);
		//genesPeaksGraph.peakInductionTable(minExp, Math.abs(rnaSeqInduction), false, true, tfNames);
		//genesPeaksGraph.peakInductionTable(minExp, Math.abs(rnaSeqInduction), true, false, tfNames);
		//genesPeaksGraph.peakInductionTable(minExp, Math.abs(rnaSeqInduction), false, false, tfNames);
	    }
	    if (targetList) { //Extract a list of high confidence fos-targets
		for (int j=1; j<nRNAiTimes; j++) {
		    LinkedHashSet<RefSeqGene> goodTargets = genesPeaksGraph.getBestTargets(minExp, Math.abs(rnaSeqInduction), Math.abs(rnaiInduction), true, true, false, j, 0, targetTFNames);
		    for (int k=1; k<nRNAiConditions; k++) 
			goodTargets.addAll(genesPeaksGraph.getBestTargets(minExp, Math.abs(rnaSeqInduction), Math.abs(rnaiInduction), true, true, false, j, k, targetTFNames));
		    System.out.println("Good IEG targets for time " + j + " are:");
		    for (RefSeqGene g : goodTargets)
			System.out.println(g.toString() + "\t" + g.getID() + "\t" + format.format(g.getInductionRatio(true)) + "\t" + format.format(g.getRNAi(RefSeqGenesPeakGraph.RNAI_NAME_B1[0][j]).elementAt(0)) + "\t" + format.format(g.getRNAi(RefSeqGenesPeakGraph.CONTROL_RNAI_NAME_B1[j]).elementAt(0)));
		    goodTargets = genesPeaksGraph.getBestTargets(minExp, Math.abs(rnaSeqInduction), Math.abs(rnaiInduction), false, true, false, j, 0, targetTFNames);
		    for (int k=1; k<nRNAiConditions; k++) 
			goodTargets.addAll(genesPeaksGraph.getBestTargets(minExp, Math.abs(rnaSeqInduction), Math.abs(rnaiInduction), false, true, false, j, k, targetTFNames));
		    System.out.println("Good LRG targets for time " + j + " are:");
		    for (RefSeqGene g : goodTargets)
			System.out.println(g.toString() + "\t" + g.getID() + "\t" + format.format(g.getInductionRatio(false)) + "\t" + format.format(g.getRNAi(RefSeqGenesPeakGraph.RNAI_NAME_B1[0][j]).elementAt(0)) + "\t" + format.format(g.getRNAi(RefSeqGenesPeakGraph.CONTROL_RNAI_NAME_B1[j]).elementAt(0)));
		}
	    }
	    //Create a classifier for predicting if a gene is going to be affected by KCl
	    if (bayesNetwork) {
		PeakGeneInductionNetwork bayesianNetwork = new PeakGeneInductionNetwork(genesPeaksGraph, maxDist, proximalDist, Math.abs(rnaSeqInduction), minExp, edgePenaltyCoeff);
		System.out.println(bayesianNetwork.toString() + "\nCalculating prediction probabilities:");
		TreeMap<RefSeqGene, Double> inductionProbs = null; //bayesianNetwork.calculateGeneInductionProbability(minExp, true, rnaSeqInduction);
		for (RefSeqGene g : inductionProbs.keySet()) 
		    System.out.println(g.toString() + "\t" + format.format(inductionProbs.get(g)) + "\t" + format.format(g.getInductionRatio(true)));
		bayesianNetwork.findBestNetwork(proximalDist);
		System.out.println(bayesianNetwork.toString() + "\nCalculating prediction probabilities:");
		//Search for the best model that can predict the induction. We do this iteratively, by first considering proximal peaks. Then we add another 
		/*for (int i=2; i<=maxDist/distBins; i++) {
		  bayesianNetwork.findBestNetwork(i*distBins);
		  System.out.println("Adding more nodes for " + (i*distBins));
		  }*/
		/*System.out.println(bayesianNetwork.toString() + "\nCalculating prediction probabilities:");
		//Find out how well the network predicts induction
		TreeMap<RefSeqGene, Double> inductionProbs = bayesianNetwork.calculateGeneInductionProbability(minExp, true, rnaSeqInduction);
		for (RefSeqGene g : inductionProbs.keySet()) {
		System.out.println(g.toString() + "\t" + format.format(inductionProbs.get(g)) + "\t" + format.format(g.getInductionRatio(true)));
		}*/
	    } 
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static TreeSet<Peak> parsePeakFile(String peakFileName, int chr, TreeSet<String> tfNames) {
	TreeSet<Peak> peaks = new TreeSet<Peak>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(peakFileName));
	    String line = br.readLine();
	    String headers[] = line.split("\t");
	    int colInds[] = new int[tfNames.size()], j = 0, chrColInd = 0, posColInd = 1;
	    //Find the headers that we are interested in
	    for (int i=0; i<headers.length; i++) {
		if (tfNames.contains(headers[i]))
		    colInds[j++] = i;
	    }
	    if (j<colInds.length)
		System.out.println("Warning! Could not find all TF names.");
	    line = br.readLine();
	    while (line!=null) {
		String tokens[] = line.split("\t");
		try {
		    int c = Integer.parseInt(tokens[chrColInd]);
		    if (c==chr) {
			Peak p = new Peak(c, Integer.parseInt(tokens[posColInd]), Integer.parseInt(tokens[posColInd]));
			for (int i=0; i<colInds.length; i++) {
			    double e = Double.parseDouble(tokens[colInds[i]]);
			    if (e>0)
				p.addFactor(headers[colInds[i]], e);
			}
			peaks.add(p);
		    }
		}
		catch (NumberFormatException e) {
		    System.err.println("Error parsing " + line);
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing peak file " + e.toString());
	}
	System.out.println("Found " + peaks.size() + " peaks from " + peakFileName);
	return peaks;
    }

    public static void loadExonDensities(RefSeqGenes refSeqGenes, int chr, boolean orchestra) {
	int t[] = new int[]{0, 1, 6};
	for (int i=0; i<t.length; i++) {
	    String fileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/mRNA/hr" + t[i] + "_unique__Genes-FEATURES-READS-r35_x500_N10M__ALL-GENES__EXN-ALL_INT-ALL.tsv";
	    //Parse the data from the file
	    try {
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		//Parse the header
		String line = br.readLine();
		Pattern headerPattern = Pattern.compile("([\\w|\\_|\\-|\\d|-|_]+)\\s+");
		Matcher m = headerPattern.matcher(line);
		int nColumns = 0;
		int exonDensCol = -1;
		TreeMap<String, Integer> headerNames2ColumnMap = new TreeMap<String, Integer>();
		while (m.find()) {
		    headerNames2ColumnMap.put(m.group(1), nColumns++);
		    if (m.group(1).equals(RNASeqData.SENSE_EXON_DENSITY))
			exonDensCol = nColumns - 2;
		}
		//Now parse the rest of the file
		line = br.readLine();
		Pattern columnPattern = Pattern.compile("([\\w|_|\\d|\\@|\\||\\.|-|\\+|\\(|\\)\\^]+)\\s+");
		while (line!=null) {
		    m = columnPattern.matcher(line);
		    m.find(); //zeroth column is index
		    if (m.find()) {//First column is the gene name
			String name = m.group(1);
			if (m.find()) {//Second column is the gene ID
			    int geneID = Integer.parseInt(m.group(1)), chrom;
			    if (m.find()) {//Third column is the chromosome
				if (m.group(1).equals("X")) { chrom = 20; }
				else if (m.group(1).equals("Y")) { chrom = 21; }
				else { chrom = Integer.parseInt(m.group(1)); }
				if (chr==chrom) {
				    RefSeqGene g = refSeqGenes!=null ? refSeqGenes.getGene(geneID) : null;
				    if (g!=null) {
					//Find the exon density
					nColumns = 3;
					while (m.find()) {
					    if (exonDensCol==nColumns++)
						g.setExonDensity(RNA_SEQ_NAMES[i], Double.parseDouble(m.group(1)));
					}
				    }
				}
			    }
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error opening " + fileName + ": " + e.toString());
	    }
	}
    }


    public static void loadRNASeqReads(RefSeqGenes refSeqGenes, int chr, boolean force) {
	System.out.println("Loading RNA-Seq data for chr " + chr);
	//First try to load from a file 
	try {
	    FileInputStream fis = new FileInputStream("refSeqGenesReads" + chr + ".txt");
	    ObjectInputStream in = new ObjectInputStream(fis);
	    refSeqGenes = (RefSeqGenes)in.readObject();
	    in.close();
	}
	catch(Exception e) {
	    System.out.println("Failed to load reads info, recomputing....");
	    force = true;
	}
	if (force) {
	    //Load all the reads and assign. This is slow....
	    String repeatFileName = EstimateGenicTranscription3.REPEAT_RNA_FILE_MM;
	    String rnaFileName[] = new String[]{"/groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr0c_b1/hr0c_b1", "/groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr1c_b1/hr1c_b1", "/groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c_b2/hr6c_b2"};
	    int readLength = 35;
	    for (int i=0; i<rnaFileName.length; i++) {
		ABIRead reads[] = ABIReads.loadMRNAReadsChr(rnaFileName[i], true, chr, readLength, false, repeatFileName);
		refSeqGenes.setFactorReads(RNA_SEQ_NAMES[i] + "_f", reads, 1);
		reads = ABIReads.loadMRNAReadsChr(rnaFileName[i], false, chr, readLength, false, repeatFileName);
		refSeqGenes.setFactorReads(RNA_SEQ_NAMES[i] + "_r", reads, -1);
	    }
	    try {
		FileOutputStream fos = new FileOutputStream("refSeqGenesReads" + chr + ".txt");
		ObjectOutputStream out = new ObjectOutputStream(fos);
		out.writeObject(refSeqGenes);
		out.close();
	    }
	    catch(IOException ex) {
		ex.printStackTrace();
	    }
	}
    }

}





