//Use this method to analyze the distribution of the enhancer sites
//with respect to TSSs. In particular, we are interested in finding
//out how much of the factor binding that we can explain based on the
//either proximity to the TSS or by the presence of an enhancer

//java -Xmx512m AnalyzeTFLocations -e Enhancers/enhancers_null_KCl.txt -p PeakFiles/all_tfs_names -n PeakFiles/enhancer_factor_names.txt

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.probdist.HypergeometricDist;
import umontreal.iro.lecuyer.probdist.BinomialDist;

public class AnalyzeTFLocations {

    public static final String TF_TSS_DISTRIBUTION = "_TFs_TSS";
    public static final String TF_ENHANCER_DISTRIBUTION = "_TFs_Enhancer";
    public static final String TF_NONGENIC_DISTRIBUTION = "_TFs_Nongenic";
    public static final String TF_INTRONIC_DISTRIBUTION = "_TFs_Intronic";

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option pValueFlag = argParser.addDoubleOption('P', "pValue");
	    //Location of the enhancer list
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption('e', "enhancerFileName");
	    CmdLineParser.Option minEnhancer2TSSDistanceFlag = argParser.addIntegerOption('d', "minEnhancer2TSSDistance");
	    CmdLineParser.Option peakFileFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option unstimulatedFlag = argParser.addStringOption('u', "unstimulated");
	    CmdLineParser.Option maxEnhancer2TSSDistanceFlag = argParser.addIntegerOption('d', "maxEnhancer2TSSDistance");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption('N', "nMC");
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption('C', "conservation");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    boolean kcl = argParser.getOptionValue(unstimulatedFlag)==null;
	    tmp = argParser.getOptionValue(minEnhancer2TSSDistanceFlag);
	    int minEnhancer2TSSDistance = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(pValueFlag);
	    double pValue = tmp!=null ? ((Double)tmp).doubleValue() : .005;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/enhancers_null_KCl.txt";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : enhancerFileName.substring(0, enhancerFileName.lastIndexOf("."));
	    tmp = argParser.getOptionValue(peakFileFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(maxEnhancer2TSSDistanceFlag);
	    int maxEnhancer2TSSDistance = tmp!=null ? ((Integer)tmp).intValue() : 100000;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 500;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21, nTSSs = 0;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    //Load the geneIDs of the genes that are upregulated
	    int upregulatedGeneIDs[] = ExpressionEnhancerCorrelation.parseSymbolicExpressionFile(new String[]{"LHH", "LLH", "LMH", "LHM"}, orchestra);
	    Arrays.sort(upregulatedGeneIDs);
	    //Find the list of TFs to compare
	    Vector<String> fileNames = new Vector<String>(), shortNames = new Vector<String>();
	    if (fileName.contains("names")) 
		fileNames = FindOverlappingPeaks.getFileNames(fileName, shortNames);
	    else 
		fileNames.addElement(fileName);
	    //Load the enhancer sites
	    Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	    int nUnidirectionalEnhancers = 0, nTested = 0, nEnhancerReads = 0, nUnidirectionalEnhancersReads = 0, nNongenic = 0;
	    for (Enhancer en : enhancers) {
		int n = en.numberOfReadsForward() + en.numberOfReadsReverse();
		if (en.isNonGenic()) {
		    nNongenic++;
		    if (n>=7) {
			nTested++;
			nEnhancerReads += n;
			double p = BinomialDist.cdf(n, .5, en.numberOfReadsReverse());
			p = p>.5 ? 1-p : p;
			nUnidirectionalEnhancers += p<pValue ? 1 : 0;
			nUnidirectionalEnhancersReads += p<pValue ? n : 0;
		    }
		}
	    }
	    System.out.println("There were " + nTested + " enhancers that could be tested and of those, " + nUnidirectionalEnhancers + " were unidirectional at p-value " + format.format(pValue));
	    System.out.println("The average number of reads for all non-genic enhancers is " + format.format((double)nEnhancerReads/nTested) + " and the average number of reads for uni-directional is " + format.format((double)nUnidirectionalEnhancersReads/nUnidirectionalEnhancers));
	    //Find the number of TSSs that have a given number of enhancers within a certain cut-off distance
	    
	    //Load all the TF binding sites
	    Vector<Peak> peaks = new Vector<Peak>();
	    for (int i=1; i<=nChr; i++) { 
		if (chr==-1 || chr==i) {
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
		    nTSSs += geneInfo[i-1].size();
		    peaks.addAll(FindEnhancerSites.parsePeakFile(peakFileName, i, conservation).values()); 
		}
	    }
	    String[] factorNames = peaks.firstElement().getFactorNames(false);
	    System.out.println("Found factors " + Arrays.toString(factorNames));
	    //Find out the total number of peaks in each category
	    TreeMap<String, Integer> nPeaksWithFactor = new TreeMap<String, Integer>(), nConsPeaksWithFactor = new TreeMap<String, Integer>();
	    for (int i=0; i<factorNames.length; i++) {
		nPeaksWithFactor.put(factorNames[i], 0);
		nConsPeaksWithFactor.put(factorNames[i], 0); 
	    }
	    for (Peak p : peaks) {
		for (String k : p.getFactorNames()) {
		    nPeaksWithFactor.put(k, nPeaksWithFactor.get(k) + 1);
		    if (p.isConserved())
			nConsPeaksWithFactor.put(k, nConsPeaksWithFactor.get(k) + 1);
		}
	    }
	    System.out.println("Number of peaks found:\nTF\tAll\tCons");
	    for (String k : nPeaksWithFactor.keySet())
		System.out.println(k + "\t" + nPeaksWithFactor.get(k) + "\t" + nConsPeaksWithFactor.get(k));
	    //Go through the TFBS and try to assign each to either a
	    //TSS or an exon
	    TreeMap<Peak, Enhancer> peak2EnhancerMap = new TreeMap<Peak, Enhancer>();
	    TreeMap<Peak, RefSeqGene> peak2GeneMap = new TreeMap<Peak, RefSeqGene>();
	    TreeMap<Peak, Integer> remainingPeaksIntronic = new TreeMap<Peak, Integer>(), remainingPeaksNongenic = new TreeMap<Peak, Integer>();
	    for (Peak p : peaks) {
		//if (p.hasStimulatedFactor(kcl)) {
		//nPeaks++;
		if (!conservation || p.isConserved()) {
		    int c = p.getChromosome();
		    Enhancer e = getNearestEnhancer(enhancers, p.getCenter());
		    int enhancerDist = e.getCenter() - p.getCenter();
		    if (Math.abs(enhancerDist)<minEnhancer2TSSDistance)
			peak2EnhancerMap.put(p, e);
		    else {
			RefSeqGene g = geneInfo[c-1].getNearestTSSGene(p.getCenter(), false);
			int tssDist = p.getCenter() - g.getStart();
			if (Math.abs(tssDist)<minEnhancer2TSSDistance)
			    peak2GeneMap.put(p, g);
			else if(geneInfo[c-1].overlapsGene(p.getCenter()))
			    remainingPeaksIntronic.put(p, tssDist);
			else 
			    remainingPeaksNongenic.put(p, tssDist);
		    }
		}
	    }
	    System.out.println("Found a total of " + peaks.size() + " different peaks and " + enhancers.size() + " enhancers for " + (kcl ? "KCl" : "un"));
	    System.out.println("Found " + peak2GeneMap.size() + " peaks near genes and " + peak2EnhancerMap.size() + " near enhancers.");
	    System.out.println("Of the remaining peaks, " + remainingPeaksIntronic.size() + " are intronic and " + remainingPeaksNongenic.size() + " are non-genic.");
	    /////////////////////////////////
	    TreeMap<String, Integer> tfsEnhancers = new TreeMap<String, Integer>();
	    TreeMap<String, Integer> enhancerCombos = new TreeMap<String, Integer>(), enhancerCombosCons = new TreeMap<String, Integer>();
	    for (Peak p : peak2EnhancerMap.keySet()) {
		String key = Arrays.toString(p.getFactorNames());
		tfsEnhancers.put(key, (tfsEnhancers.containsKey(key) ? tfsEnhancers.get(key) + 1 : 1));
		//Figure out which combinations are more or less unlikely
		//than expected. Consider only pairwise and use a
		//hypergeometric test to decide significance.
		for (int i=0; i<factorNames.length; i++) {
		    if (p.hasFactor(factorNames[i])) {
			enhancerCombos.put(factorNames[i], (enhancerCombos.containsKey(factorNames[i]) ? enhancerCombos.get(factorNames[i]) + 1 : 1));
			if (p.isConserved())
			    enhancerCombosCons.put(factorNames[i], (enhancerCombosCons.containsKey(factorNames[i]) ? enhancerCombosCons.get(factorNames[i]) + 1 : 1));
			//System.out.println("Found " + factorNames[i] + " for " + enhancerCombos.get(factorNames[i]));
			for (int j=i+1; j<factorNames.length; j++) {
			    key = factorNames[i] + ", " + factorNames[j];
			    if (p.hasFactor(factorNames[i]) && p.hasFactor(factorNames[j])) {
				enhancerCombos.put(key, (enhancerCombos.containsKey(key) ? enhancerCombos.get(key) + 1 : 1));
				if (p.isConserved())
				    enhancerCombosCons.put(key, (enhancerCombosCons.containsKey(key) ? enhancerCombosCons.get(key) + 1 : 1));
			    }
			}
		    }
		}
	    }	    
	    //findSignificantTFCorrelations(enhancerCombos, factorNames, enhancers.size(), pValue);
	    System.out.println("Total number of peaks at enhancers are:");
	    for (int i=0; i<factorNames.length; i++) {
		if (enhancerCombos.containsKey(factorNames[i]))
		    System.out.println(factorNames[i] + "\t" + enhancerCombos.get(factorNames[i]) + "\t" + enhancerCombosCons.get(factorNames[i]));
	    }
	    //////////////////////
	    TreeMap<String, Integer> tfsTSSs = new TreeMap<String, Integer>();
	    TreeMap<String, Integer> tssCombos = new TreeMap<String, Integer>(), tssCombosCons = new TreeMap<String, Integer>();
	    for (Peak p : peak2GeneMap.keySet()) {
		String key = Arrays.toString(p.getFactorNames());
		tfsTSSs.put(key, (tfsTSSs.containsKey(key) ? tfsTSSs.get(key) + 1 : 1));
		for (int i=0; i<factorNames.length; i++) {
		    if (p.hasFactor(factorNames[i])) {
			tssCombos.put(factorNames[i], (tssCombos.containsKey(factorNames[i]) ? tssCombos.get(factorNames[i]) + 1 : 1));
			if (p.isConserved())
			    tssCombosCons.put(factorNames[i], (tssCombosCons.containsKey(factorNames[i]) ? tssCombosCons.get(factorNames[i]) + 1 : 1));
			for (int j=i+1; j<factorNames.length; j++) {
			    key = factorNames[i] + ", " + factorNames[j];
			    if (p.hasFactor(factorNames[i]) && p.hasFactor(factorNames[j])) {
				tssCombos.put(key, (tssCombos.containsKey(key) ? tssCombos.get(key) + 1 : 1));
				if (p.isConserved())
				    tssCombosCons.put(key, (tssCombosCons.containsKey(key) ? tssCombosCons.get(key) + 1 : 1));
			    }
			}
		    }
		}
	    }
	    //findSignificantTFCorrelations(tssCombos, factorNames, nTSSs, pValue);
	    System.out.println("Total number of peaks at TSSs are:");
	    for (int i=0; i<factorNames.length; i++) {
		if (tssCombos.containsKey(factorNames[i]))
		    System.out.println(factorNames[i] + "\t" + tssCombos.get(factorNames[i]) + "\t" + tssCombosCons.get(factorNames[i]));
	    }
	    /////////////////////////////////
	    //Now we want to find out if there is a correlation
	    //between the presence of a TF at a TSS and if it seems to
	    //be present at nearby enhancers.
	    TreeMap<String, Integer> tfEnhancerTSSPairs = findTFEnhancerTSSPairs(peak2GeneMap, peak2EnhancerMap, maxEnhancer2TSSDistance, null);
	    //Shuffle the peaks for both the enhancers and the TSSs
	    TreeMap<Peak, RefSeqGene> peak2GeneMapMC = getRandomPeak2GeneMap(new Vector<Peak>(peak2GeneMap.keySet()), geneInfo, chr);
	    TreeMap<Peak, Enhancer> peak2EnhancerMapMC = getRandomPeak2EnhancerMap(new Vector<Peak>(peak2EnhancerMap.keySet()), enhancers);
	    TreeMap<String, Integer> tfEnhancerTSSPairsMC = null;
	    for (int i=0; i<nMC; i++)
		tfEnhancerTSSPairsMC = findTFEnhancerTSSPairs(peak2GeneMapMC, peak2EnhancerMapMC, maxEnhancer2TSSDistance, tfEnhancerTSSPairsMC);
	    System.out.println("The pairs that were found are:\nTSS\tEnhancer\tnObs\tnMC");
	    TreeSet<String> keys = new TreeSet<String>(tfEnhancerTSSPairs.keySet());
	    keys.addAll(tfEnhancerTSSPairsMC.keySet());
	    for (String k : keys)
		System.out.println(k + "\t" + (tfEnhancerTSSPairs.containsKey(k) ? tfEnhancerTSSPairs.get(k) : 0) + "\t" + format.format(tfEnhancerTSSPairsMC.containsKey(k) ? (double)tfEnhancerTSSPairsMC.get(k)/nMC : 0));
	    /////////////////////////////////
	    //Do the same, but now we consider the raw data instead
	    //(in case the peak-finding is too coarse-grained)
	    //The map is from TF_name:gene:nReads@TSS:nReads@nearbyEnhancers
	    TreeMap<String, TreeMap<RefSeqGene, Vector<Integer>>> tfEnhancerTSSCorrelation = findTFEnhancerTSSCorrelation(geneInfo, enhancers, chr, maxEnhancer2TSSDistance, windowSize, fileNames, inputFileName);
	    for (String fName : tfEnhancerTSSCorrelation.keySet()) {
		for (RefSeqGene g : tfEnhancerTSSCorrelation.get(fName).keySet()) {
		    int nReadsTSS = tfEnhancerTSSCorrelation.get(fName).get(g).firstElement();
		}
	    }

	    /////////////////////////////////
	    TreeMap<String, Integer> tfsIntronic = new TreeMap<String, Integer>();
	    for (Peak p : remainingPeaksIntronic.keySet()) {
		String key = Arrays.toString(p.getFactorNames());
		tfsIntronic.put(key, (tfsIntronic.containsKey(key) ? tfsIntronic.get(key) + 1 : 1));
	    }
	    ////////////////////////////////////////////
	    TreeMap<String, Integer> tfsNongenic = new TreeMap<String, Integer>();
	    for (Peak p : remainingPeaksNongenic.keySet()) {
		String key = Arrays.toString(p.getFactorNames());
		tfsNongenic.put(key, (tfsNongenic.containsKey(key) ? tfsNongenic.get(key) + 1 : 1));
	    }
	    ///////////////////////////////////////////////////////
	    //Save the distributions of TFs near enhancers and TSSs
	    String saveName = saveFileName + TF_ENHANCER_DISTRIBUTION + ".txt";
	    System.out.println("Saving to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    for (String k : tfsEnhancers.keySet()) {
		bw.write(k + "\t" + tfsEnhancers.get(k) + "\n");
	    }
	    bw.close();
	    //////////////////////
	    saveName = saveFileName + TF_TSS_DISTRIBUTION + ".txt";
	    System.out.println("Saving to " + saveName);
	    bw = new BufferedWriter(new FileWriter(saveName));
	    for (String k : tfsTSSs.keySet()) {
		bw.write(k + "\t" + tfsTSSs.get(k) + "\n");
	    }
	    bw.close();
	    //////////////////////
	    saveName = saveFileName + TF_INTRONIC_DISTRIBUTION + ".txt";
	    System.out.println("Saving to " + saveName);
	    bw = new BufferedWriter(new FileWriter(saveName));
	    for (String k : tfsIntronic.keySet()) {
		bw.write(k + "\t" + tfsIntronic.get(k) + "\n");
	    }
	    bw.close();
	    //////////////////////
	    saveName = saveFileName + TF_NONGENIC_DISTRIBUTION + ".txt";
	    System.out.println("Saving to " + saveName);
	    bw = new BufferedWriter(new FileWriter(saveName));
	    for (String k : tfsNongenic.keySet()) {
		bw.write(k + "\t" + tfsNongenic.get(k) + "\n");
	    }
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static TreeMap<Peak, RefSeqGene> getRandomPeak2GeneMap(Vector<Peak> peaks, RefSeqGenes geneInfo[], int chr) { return getRandomPeak2GeneMap(peaks, geneInfo, chr, peaks.size()); }

    public static TreeMap<Peak, RefSeqGene> getRandomPeak2GeneMap(Vector<Peak> peaks, RefSeqGenes geneInfo[], int chr, int n) {
	TreeMap<Peak, RefSeqGene> peak2genesMC = new TreeMap<Peak, RefSeqGene>();
	int nTSS[] = new int[geneInfo.length], sumTSS = 0;
	for (int i=0; i<geneInfo.length; i++) { 
	    if (chr==-1 || i+1==chr) {
		nTSS[i] = geneInfo[i].size(); 
		sumTSS += geneInfo[i].size(); 
	    }
	    else
		nTSS[i] = 0;
	}
	for (Peak p : peaks) {
	    //Draw a random gene and assign it to each peak
	    int r = random.nextInt(sumTSS), c = 0;
	    while (r>0) 
		r -= nTSS[c++];
	    peak2genesMC.put(p, geneInfo[--c].getRandomGene(random));
	    if (peak2genesMC.size()>=n) { break; }
	}
	return peak2genesMC;
    }

    public static TreeMap<Peak, Enhancer> getRandomPeak2EnhancerMap(Vector<Peak> peaks, Vector<Enhancer> enhancers) { return getRandomPeak2EnhancerMap(peaks, enhancers, peaks.size()); }

    public static TreeMap<Peak, Enhancer> getRandomPeak2EnhancerMap(Vector<Peak> peaks, Vector<Enhancer> enhancers, int n) {
	TreeMap<Peak, Enhancer> peak2EnhancerMapMC = new TreeMap<Peak, Enhancer>();
	for (Peak p : peaks) {
	    //Draw a random enhancer and assign it to each peak
	    peak2EnhancerMapMC.put(p, enhancers.elementAt(random.nextInt(enhancers.size())));
	    if (peak2EnhancerMapMC.size()>=n) { break; }
	}
	return peak2EnhancerMapMC;
    }

    public static TreeMap<String, Integer> findTFEnhancerTSSPairs(TreeMap<Peak, RefSeqGene> peak2GeneMap, TreeMap<Peak, Enhancer> peak2EnhancerMap, int maxEnhancer2TSSDistance, TreeMap<String, Integer> tfEnhancerTSSPairsOld) {
	TreeMap<String, Integer> tfEnhancerTSSPairs = tfEnhancerTSSPairsOld==null ? new TreeMap<String, Integer>() : tfEnhancerTSSPairsOld;
	int nPeakEnhancersNearTSS = 0, nMatchingPeakEnhancersNearTSS = 0;
	for (Peak p : peak2GeneMap.keySet()) {
	    RefSeqGene g = peak2GeneMap.get(p);
	    for (Peak pe : peak2EnhancerMap.keySet()) {
		if (pe.getChromosome()==p.getChromosome() && Math.abs(pe.getCenter() - g.getStart())<maxEnhancer2TSSDistance) {
		    nPeakEnhancersNearTSS++;
		    for (int j=0; j<p.getFactorNames().length; j++) {
			int ind = Arrays.binarySearch(pe.getFactorNames(), p.getFactorNames()[j], null);
			if (ind>=0) 
			    nMatchingPeakEnhancersNearTSS++;
			for (int i=0; i<pe.getFactorNames().length; i++) {
			    String k = p.getFactorNames()[j] + ", " + pe.getFactorNames()[i];
			    tfEnhancerTSSPairs.put(k, (tfEnhancerTSSPairs.containsKey(k) ? tfEnhancerTSSPairs.get(k) + 1 : 1));
			}
		    }
		    if (g.getName().equals("Fos") || g.getName().equals("Arc")) {
			System.out.println("Found enhancer " + pe.toString() + " at distance " + (pe.getCenter() - g.getStart()) + " near " + g.toString() + " with TFs " + Arrays.toString(p.getFactorNames()));
			System.out.println("TFs at enhancer are: " + Arrays.toString(pe.getFactorNames()));
		    }
		}
	    }
	}
	if (tfEnhancerTSSPairsOld==null)
	    System.out.println(nPeakEnhancersNearTSS + " enhancers with other TFs than CBP were within " + maxEnhancer2TSSDistance + " bps of a TSS with a TF other than CBP and " + nMatchingPeakEnhancersNearTSS + " were matching.");
	return tfEnhancerTSSPairs;
    }

    public static void findSignificantTFCorrelations(TreeMap<String, Integer> combos, String factorNames[], int n, double pValue) {
	for (int i=0; i<factorNames.length; i++) {
	    for (int j=i+1; j<factorNames.length; j++) {
		int m1 = 0, m2 = 0, m12 = 0; //Find the marginals and joint count
		for (String k : combos.keySet()) {
		    if (k.equals(factorNames[i]))
			m1 = combos.get(k);
		    if (k.equals(factorNames[j]))
			m2 = combos.get(k);
		    if (k.contains(factorNames[i]) && k.contains(factorNames[j]))
			m12 = combos.get(k);
		}
		if (m1>0 && m2>0) {
		    System.out.println("Found " + m1 + " peaks for " + factorNames[i] + " and " + m2 + " peaks for " + factorNames[j] + " with an overlap of " + m12);
		    double p = HypergeometricDist.barF(m1, n, m2, m12);
		    p = p>1-pValue ? 1-p : p;
		    if (p<pValue && (m1>1 && m2>1 && m12>0))
			System.out.println("This gives a p-value of " + format.format(p) + ". The expected overlap is " + format.format(HypergeometricDist.getMean(m1, n, m2)));
		}
	    }
	}
    }


    public static Enhancer getNearestEnhancer(Vector<Enhancer> enhancers, int c) {
	//Go through the list of enhancers and find the one which is closest to c
	Enhancer nearest = null;
	int nearestDist = Integer.MAX_VALUE;
	for (Enhancer e : enhancers) {
	    int dist = Math.abs(c - e.getCenter());
	    if (dist<nearestDist) {
		nearest = e;
		nearestDist = dist;
	    }
	}
	return nearest;
    }

    public static Peak getNearestPeak(Vector<Peak> peaks, int c) {
	Peak nearest = null;
	int nearestDist = Integer.MAX_VALUE;
	for (Peak p : peaks) {
	    int dist = Math.abs(c - p.getCenter());
	    if (dist<nearestDist) {
		nearest = p;
		nearestDist = dist;
	    }
	}
	return nearest;
    }

    public static TreeMap<String, TreeMap<RefSeqGene, Vector<Integer>>> findTFEnhancerTSSCorrelation(RefSeqGenes geneInfo[], Vector<Enhancer> enhancers, int chr, int maxEnhancer2TSSDistance, int windowSize, Vector<String> fileNames, String inputFileName) {
	//The map is from TF_name:gene:nReads@TSS:nReads@nearbyEnhancers
	TreeMap<String, TreeMap<RefSeqGene, Vector<Integer>>> tfEnhancerTSSCorrelation = new TreeMap<String, TreeMap<RefSeqGene, Vector<Integer>>>();

	int fragmentLength = 120;
	double qualityScore = 10;
	for (String fileName : fileNames) {
	    tfEnhancerTSSCorrelation.put(fileName , new TreeMap<RefSeqGene, Vector<Integer>>());
	    for (int c=1; c<=geneInfo.length; c++) {
		if (chr==-1 || c==chr) {
		    ABIRead[] reads = PeakFinder2.parseFiles(c, fileName, fragmentLength, true, 0, qualityScore);
		    ABIRead[] inputReads = inputFileName==null ? null : PeakFinder2.parseFiles(c, inputFileName, fragmentLength, true, 0, qualityScore);
		    for (RefSeqGene g : geneInfo[c-1].getGenes()) {
			//Find the enhancers near the gene
			Vector<Enhancer> nearbyEnhancers = new Vector<Enhancer>();
			for (Enhancer e : enhancers) {
			    if (c==e.getChromosome() && Math.abs(e.getCenter() - g.getStart())<maxEnhancer2TSSDistance)
				nearbyEnhancers.addElement(e);
			}
			if (nearbyEnhancers.size()>0) { //Get the number of reads surrounding the peak
			    tfEnhancerTSSCorrelation.get(fileName).put(g, new Vector<Integer>());
			    int nReadsTSS = FindReadsAcrossFeature.numberOfReads(c, g.getStart() - windowSize, g.getStart() + windowSize, fragmentLength, reads);
			    int nInputTSS = inputFileName==null ? 0 : FindReadsAcrossFeature.numberOfReads(c, g.getStart() - windowSize, g.getStart() + windowSize, fragmentLength, inputReads);
			    tfEnhancerTSSCorrelation.get(fileName).get(g).addElement(nReadsTSS - nInputTSS);
			    double meanEnhancerReads = 0;
			    for (Enhancer ne : nearbyEnhancers) {
				int nReads = FindReadsAcrossFeature.numberOfReads(c, ne.getCenter() - windowSize, ne.getCenter() + windowSize, fragmentLength, reads);
				int nInput = inputFileName==null ? 0 : FindReadsAcrossFeature.numberOfReads(c, ne.getCenter() - windowSize, ne.getCenter() + windowSize, fragmentLength, inputReads);;
				tfEnhancerTSSCorrelation.get(fileName).get(g).addElement(nReads - nInput);
				meanEnhancerReads += nReads - nInput;
			    }
			    meanEnhancerReads /= nearbyEnhancers.size();
			    if (nReadsTSS - nInputTSS>5 || meanEnhancerReads>5)
				System.out.println("Found " + (nReadsTSS - nInputTSS) + " " + fileName + " reads near the TSS and an average of " + format.format(meanEnhancerReads) + " at " + nearbyEnhancers.size() + " enhancers near " + g.toString());
			}
		    }
		}
	    }
	}
	return tfEnhancerTSSCorrelation;
    }

}

