//Use this class to analyze the read density related to the
//transcripts that we have obtained using FindTranscripts.java. This
//can also be used for other sets of regions, such as the
//annotation. We are interested in investigating what is going on
//around the starts and ends, but also consider a normalized
//(quantiled) representation of what is happening along the length of
//the region.
//
//We also want to be able to investigate subsets, such as regions
//which we believe are non-coding, or anti-sense to annotated genes.
//
//Example:
//
//java -Xmx1024m AnalyzeTranscriptReadProfiles -b /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr0c --minDens .001 -o -n ../MmDataPeakFiles/Pol2_8WG16_KCl_B1_names.txt -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -c 19 -w 1500 -B 20 


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.*;

public class AnalyzeTranscriptReadProfiles {

    public static final String CODING_START_END_DIST_SUFFIX = "_coding_start_end_dist";
    public static final int START_WINDOW_SIZE_HISTONES = 3000;
    public static final int START_WINDOW_SIZE_TFS = 300;
    public static final int START_WINDOW_SIZE_RNA = 1500;
    public static final int ASSOCIATION_STRENGTH_CHARACTERISTIC_LENGTH_HISTONES = 500;
    public static final int ASSOCIATION_STRENGTH_CHARACTERISTIC_LENGTH_TFS = 1000;


    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    //By default we want to investigate both strands together
	    CmdLineParser.Option strandFlag = argParser.addIntegerOption('r', "reverseStrand");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag if we want to parse the big-bed file instead of the chromosome specific
	    CmdLineParser.Option sameBedFileFlag = argParser.addBooleanOption("sameBedFile");
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option logFlag = argParser.addIntegerOption("log");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option sequenceFlag = argParser.addBooleanOption("sequence");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option bedFileNameFlag = argParser.addStringOption('b', "bedFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //For the bins before and after the transcribed regions
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    //Use to determine how many bins each segment should be split into
	    CmdLineParser.Option nBinsFlag = argParser.addIntegerOption('B', "nBins");
	    CmdLineParser.Option maxDivergentTSSDistanceFlag = argParser.addIntegerOption("maxDivergent");
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    CmdLineParser.Option rnaFileNameFlag = argParser.addStringOption("rnaFile");
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption("enhancers");
	    CmdLineParser.Option lncRNAFileNameFlag = argParser.addStringOption("lncRNA");
	    CmdLineParser.Option macroRNAFileNameFlag = argParser.addStringOption("macroRNA");
	    CmdLineParser.Option miRNAFileNameFlag = argParser.addStringOption("miRNA");
	    CmdLineParser.Option miRNAFileNamePrecursorsFlag = argParser.addStringOption("miRNAPrecursors");
	    //Calculate the average PhastConsScores as well
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption("conservation");
	    //Compare with EST-tags
	    CmdLineParser.Option estFileNameFlag = argParser.addBooleanOption("EST");
	    //Compare with CAGE tags
	    CmdLineParser.Option cageFileNameFlag = argParser.addBooleanOption("CAGE");
	    //Use only the cortex CAGE tags
	    CmdLineParser.Option cortexFlag = argParser.addBooleanOption('x', "cortex");
	    //Use only the CAGE tags which have the polyA mark
	    CmdLineParser.Option polyAFlag = argParser.addBooleanOption('p', "polyA");
	    CmdLineParser.Option promoterLocationCategoriesFlag = argParser.addBooleanOption("promoterLocation");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean sameBedFile = argParser.getOptionValue(sameBedFileFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null; 
	    boolean sequence = argParser.getOptionValue(sequenceFlag)!=null; 
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean reverseStrand = argParser.getOptionValue(strandFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(bedFileNameFlag);
	    String rnaFileStart = tmp!=null ? (String)tmp : null;
	    String bedFileName = rnaFileStart!=null ? rnaFileStart + FindTranscripts.BED_FILE_SUFFIX + "_" + (reverseStrand ? "r" : "f" ) + (chr==-1 || sameBedFile ? "" : "_" + chr) + ".bed" : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lMe3_real.tsv";
	    tmp = argParser.getOptionValue(lncRNAFileNameFlag);
	    String lncRNAFileName = tmp!=null ? (String)tmp : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    tmp = argParser.getOptionValue(macroRNAFileNameFlag);
	    String macroRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    tmp = argParser.getOptionValue(miRNAFileNameFlag);
	    String miRNAFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
	    tmp = argParser.getOptionValue(miRNAFileNamePrecursorsFlag);
	    String miRNAFileNamePrecursors = null; //tmp!=null ? (String)tmp : (orchestra ? "/home/mh168" : "/home/mhemberg" ) + "/Projects/ChIP-Seq/TKData/mRNA/miRNA/landgrafTuschl_2007_C_mm_miRNA_precursors_names.txt";
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    //tmp = argParser.getOptionValue(saveFileNameFlag);
	    //String saveFileName = tmp!=null ? (String)tmp : fileName.substring(0, fileName.lastIndexOf("."));
	    tmp = argParser.getOptionValue(logFlag);
	    int nLogs = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean log = nLogs>0;
	    tmp = argParser.getOptionValue(nBinsFlag);
	    int nBins = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    boolean promoterLocationCategories = argParser.getOptionValue(promoterLocationCategoriesFlag)!=null;
	    tmp = argParser.getOptionValue(maxDivergentTSSDistanceFlag);
	    int maxDivergentTSSDistance = tmp!=null ? ((Integer)tmp).intValue() : promoterLocationCategories ? 10000 : 2000;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(rnaFileNameFlag);
	    String rnaFileName = tmp!=null ? (String)tmp : rnaFileStart.replace("Random/", "") + ".tsv";
	    String rnaFileNameRandom = tmp!=null ? (String)tmp : rnaFileStart + ".tsv";
	    //EST and CAGE stuff
	    boolean cortex = argParser.getOptionValue(cortexFlag)!=null;
	    boolean polyA = argParser.getOptionValue(polyAFlag)!=null;
	    String cageFileName = argParser.getOptionValue(cageFileNameFlag)!=null ? (polyA ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/CAGE/mm9/" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "deepCAGE/mm9/") : null;
	    String estFileName = argParser.getOptionValue(estFileNameFlag)!=null ? (orchestra ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/" : null) : null;
	    ////////////////////////////////////////////
	    int nChr = FindTranscripts.numberOfChromosomes(species), nFixedBins = 2*windowSize + 1;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    //Load the enhancers
	    TreeSet<Enhancer2> enhancers = FindTranscripts2.parseEnhancerFile(enhancerFileName);
	    //Load the lncRNAs
	    TreeSet<LncRNA> lncRNAs = loadLncRNA(lncRNAFileName, species);
	    TreeSet<MacroRNA> macroRNAs = loadMacroRNA(macroRNAFileName, species);
	    //Load the miRNAs
	    TreeSet<MiRNA> miRNAs = loadMiRNA(miRNAFileName, miRNAFileNamePrecursors, species);
	    //Parse the file containing names of experiments to investigate
	    Vector<String> fileNames = new Vector<String>(), shortNames = new Vector<String>();
	    TreeMap<String, Integer> normFactors = null;
	    if (fileName!=null) {
		if (fileName.contains("names")) 
		    fileNames = FindOverlappingPeaks.getFileNames(fileName, shortNames);
		else 
		    fileNames.addElement(fileName);
	    }
	    else
		fileNames = null;
	    if (fileName!=null && !(fileName.contains("RNA") || fileName.contains("rna"))) 
		normFactors = FindReadsAcrossFeature.calculateTotalReads(fileNames, inputFileName, nChr);
	    else {
		normFactors = new TreeMap<String, Integer>();
		normFactors.put("test", 0); //to avoid getting errors later...
	    }
	    System.out.println("Calculated normalization factors for " + normFactors.size());
	    String[] arr = promoterLocationCategories ? TranscribedRegion.LOCATION_TYPES_ARRAY_STR : TranscribedRegion.TYPES_ARRAY_STR;
	    //[chr][fileName][type][pos]
	    double[][][][] startHist = new double[nChr][normFactors.size() - 1][arr.length][nFixedBins], endHist = new double[nChr][normFactors.size() - 1][arr.length][nFixedBins], binHist = new double[nChr][normFactors.size() - 1][arr.length][nBins];
	    //The densities for the RNA and phastConsScores
	    double[][][] startHistRNA = new double[nChr][arr.length][nFixedBins], endHistRNA = new double[nChr][arr.length][nFixedBins], binHistRNA = new double[nChr][arr.length][nBins];
	    double[][][] startHistCons = new double[nChr][arr.length][nFixedBins], endHistCons = new double[nChr][arr.length][nFixedBins], binHistCons = new double[nChr][arr.length][nBins];
	    TreeMap<String, Integer> typeCount = new TreeMap<String, Integer>();
	    //Keep track of the average distance to the start and end of annotated genes
	    TreeMap<RefSeqGene, TranscribedRegion> codingTrans = new TreeMap<RefSeqGene, TranscribedRegion>();
	    for (int i=1; i<=nChr-1; i++) {
		if (chr==-1 ||i==chr) {
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
		    //Find which genes are expressed highly enough to be included in the search for genes matching transcripts.
		    ABIRead[] rnaReadsForward = ABIReads.loadMRNAReadsChr(rnaFileName, true, i, readLength, false);
		    geneInfo[i-1].setFactorReads(rnaFileName, rnaReadsForward, 1);
		    ABIRead[] rnaReadsReverse = ABIReads.loadMRNAReadsChr(rnaFileName, false, i, readLength, false);
		    geneInfo[i-1].setFactorReads(rnaFileName, rnaReadsReverse, -1);
		    if (!promoterLocationCategories) {
			setMiRNAFactorReads(miRNAs, rnaFileName, rnaReadsForward, i, 1);
			setLncRNAFactorReads(lncRNAs, rnaFileName, rnaReadsForward, i);
			setMaRNAFactorReads(macroRNAs, fileName, rnaReadsForward, i, 1);
			setMiRNAFactorReads(miRNAs, rnaFileName, rnaReadsReverse, i, -1);
			setLncRNAFactorReads(lncRNAs, rnaFileName, rnaReadsReverse, i);
			setMaRNAFactorReads(macroRNAs, fileName, rnaReadsReverse, i, -1);
		    }
		    geneInfo[i-1].setGeneIDsKeep(true, rnaFileName, minDens);
		    geneInfo[i-1].setGeneIDsKeep(false, rnaFileName, minDens);
		    //Parse the .bed-file containing the location of the transcribed regions
		    ReducedModel r = null; //new ReducedModel(bedFileName, reverseStrand ? rnaReadsReverse : rnaReadsForward, i, !reverseStrand, species);
		    /*if (!promoterLocationCategories) 
			r.compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, enhancers);
		    else
		    r.compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1]);*/
		    //r.categorizeTranscribedRegions(promoterLocationCategories);
		    //typeCount = r.getTypeCount(typeCount);
		    if (chr==-1) {
			//save later since we want to find the average across all chromosomes
			if (normFactors.size()>1) {
			    //startHist[i-1] = r.extractReadsProfiles(normFactors, inputFileName, promoterLocationCategories, windowSize, true);
			    //endHist[i-1] = r.extractReadsProfiles(normFactors, inputFileName, promoterLocationCategories, windowSize, false);
			    //binHist[i-1] = r.extractReadsProfiles(normFactors, inputFileName, promoterLocationCategories, nBins);
			}
			//startHistRNA[i-1] = r.extractReadsProfiles(reverseStrand ? rnaReadsReverse : rnaReadsForward, windowSize, true, readLength, log, typeCount, promoterLocationCategories);
			//endHistRNA[i-1] = r.extractReadsProfiles(reverseStrand ? rnaReadsReverse : rnaReadsForward, windowSize, false, readLength, log, typeCount, promoterLocationCategories);
			//binHistRNA[i-1] = r.extractReadsProfiles(reverseStrand ? rnaReadsReverse : rnaReadsForward, nBins, readLength, log, typeCount, promoterLocationCategories);
		    }
		    else {
			//FindTranscripts.printTypeStats(r);
			if (normFactors.size()>1) {
			    //r.extractReadsProfiles(normFactors, inputFileName, windowSize, nBins, rnaFileStart, promoterLocationCategories, typeCount, fileName.contains("histone") ? START_WINDOW_SIZE_HISTONES : START_WINDOW_SIZE_TFS, fileName.contains("histone") ? ASSOCIATION_STRENGTH_CHARACTERISTIC_LENGTH_HISTONES : (fileName.contains("rna") ? -1 : ASSOCIATION_STRENGTH_CHARACTERISTIC_LENGTH_TFS), .1/*Double.NEGATIVE_INFINITY*/);
			    //r.saveAssociationStrengths(rnaFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) + "_", promoterLocationCategories, fileNames);
			}
			else if (fileName.contains("RNA") || fileName.contains("rna")) {
			    //r.extractReadsProfiles(fileNames, true, readLength, windowSize, nBins, log, rnaFileStart, promoterLocationCategories, typeCount, START_WINDOW_SIZE_RNA);
			    //r.extractReadsProfiles(fileNames, false, readLength, windowSize, nBins, log, rnaFileStart, promoterLocationCategories, typeCount, START_WINDOW_SIZE_RNA);
			}
			//Find the average RNA profiles surrounding start and end
			//r.extractReadsProfiles(reverseStrand ? rnaReadsReverse : rnaReadsForward, windowSize, readLength, nBins, rnaFileName.substring(0, rnaFileName.lastIndexOf(".")), promoterLocationCategories, log, typeCount);
			//Find the number of reads in fixed bins for more fine-grained analysis
			//r.extractBinProfiles(binSize, readLength, rnaFileName.substring(0, rnaFileName.lastIndexOf(".")));
			if (conservation)
			    //r.extractPhastConsProfiles(windowSize, nBins, rnaFileStart, promoterLocationCategories, typeCount, orchestra);
			if (sequence) {
			    //r.extractSequences(DetectSpliceJunctions.TRANSCRIPT_SEQUENCE_SAVE_DIR + "/" + rnaFileStart + "_", orchestra);
			    //r.extractGCProfiles(windowSize, nBins, rnaFileStart, promoterLocationCategories, typeCount, orchestra);
			    //Calculate the GC and CpG islands for a scatter-file as well
			    //r.saveTranscriptPromoterGC(rnaFileStart + "_GC_CpG_" + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".txt", chr==1 || chr==-1);
			    if (estFileName!=null) { //Get the EST data
				//r.getESTCategories(estFileName, i, reverseStrand, orchestra, rnaFileName.substring(0, rnaFileName.lastIndexOf(".")));
			    }			
			    //Key is position and value is number of tags
			    TreeMap<Integer, Integer> cageTagsForward = new TreeMap<Integer, Integer>(), cageTagsReverse = new TreeMap<Integer, Integer>();
			    TreeMap<Integer, Integer> estTagsForward = new TreeMap<Integer, Integer>(), estTagsReverse = new TreeMap<Integer, Integer>();
			    
			}
			//Save in files where each file contains transcripts of a given type. There each line corresponds to transcript and the columns contain information about the transcript.
			if (fileName!=null) {
			    for (int j=0; j<arr.length; j++) {
				//saveTranscriptFactorProperties(r.getSegments(arr[j]), normFactors, inputFileName, rnaFileStart + "_" + arr[j] + "_reads_" + (fileName.contains("names") ? fileName.substring(fileName.lastIndexOf("/")+ 1, fileName.lastIndexOf(".")) + "_" : "") + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".txt", chr==1 || chr==-1);
			    }
			}
			//codingTrans = r.findCodingTranscript(codingTrans);
			//Save a file with position and density info
			//for each transcript:
			if (fileName!=null) {
			    //if (inputFileName!=null)
				//r.saveTranscriptFactorDensities(normFactors, inputFileName, rnaFileStart + "_transcript_densities_" + (fileName.contains("names") ? fileName.substring(fileName.lastIndexOf("/")+ 1, fileName.lastIndexOf(".")) + "_" : "") + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".txt", chr==1 || chr==-1);
			    //else
				//r.saveTranscriptFactorDensities(fileNames, rnaFileStart + "_transcript_densities_" + (fileName.contains("names") ? fileName.substring(fileName.lastIndexOf("/")+ 1, fileName.lastIndexOf(".")) + "_" : "") + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".txt", chr==1 || chr==-1);
			}
		    }
		}
	    }
	    if (chr==-1) {
		//Now we can save the average for all chromosomes,
		//first we need to sum things up.
		if (normFactors.size()>1) {
		    double[][][] startSum = new double[startHist[0].length][startHist[0][0].length][startHist[0][0][0].length], endSum = new double[endHist[0].length][endHist[0][0].length][endHist[0][0][0].length], binSum = new double[binHist[0].length][binHist[0][0].length][binHist[0][0][0].length];
		    for (int i=1; i<=nChr-1; i++) {
			for (int j=0; j<startHist[i].length; j++) {
			    for (int k=0; k<startHist[i][j].length; k++) {
				for (int l=0; l<startHist[i][j][k].length; l++) {
				    startSum[j][k][l] += startHist[i-1][j][k][l];
				    endSum[j][k][l] += endHist[i-1][j][k][l];
				}
				for (int l=0; l<binHist[i][j][k].length; l++) 
				    binSum[j][k][l] += binHist[i-1][j][k][l];
			    }
			}
		    }
		    int j = 0;
		    for (String fName : normFactors.keySet()) {
			if (!fName.equals(inputFileName)) {
			    ReducedModel2.saveProfile(startSum[j], fName + ReducedModel2.START_HIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + ".txt", promoterLocationCategories, windowSize, false);
			    ReducedModel2.saveProfile(endSum[j], fName + ReducedModel2.END_HIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + ".txt", promoterLocationCategories, windowSize, false);
			    ReducedModel2.saveProfile(binSum[j], fName + ReducedModel2.BIN_HIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + ".txt", promoterLocationCategories, nBins, true);
			    j++;
			}
		    }
		}
		double[][] startSumRNA = new double[startHistRNA[0].length][startHistRNA[0][0].length], endSumRNA = new double[endHistRNA[0].length][endHistRNA[0][0].length], binSumRNA = new double[binHistRNA[0].length][binHistRNA[0][0].length];
		for (int i=1; i<=nChr-1; i++) {
		    for (int j=0; j<startHistRNA[i].length; j++) {
			for (int k=0; k<startHistRNA[i][j].length; k++) {
			    startSumRNA[j][k] += startHistRNA[i-1][j][k]/nChr;
			    endSumRNA[j][k] += endHistRNA[i-1][j][k]/nChr;
			}
			for (int k=0; k<binHistRNA[i][j].length; k++) 
			    binSumRNA[j][k] += binHistRNA[i-1][j][k]/nChr;
		    }
		}
		//Save the RNA densities
		ReducedModel2.saveProfile(startSumRNA, rnaFileNameRandom.substring(0, rnaFileNameRandom.lastIndexOf(".")) + ReducedModel2.START_HIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + ".txt", promoterLocationCategories, windowSize, false, typeCount);
		ReducedModel2.saveProfile(endSumRNA, rnaFileNameRandom.substring(0, rnaFileNameRandom.lastIndexOf(".")) + ReducedModel2.END_HIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + ".txt", promoterLocationCategories, windowSize, false, typeCount);
		ReducedModel2.saveProfile(binSumRNA, rnaFileNameRandom.substring(0, rnaFileNameRandom.lastIndexOf(".")) + ReducedModel2.BIN_HIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + ".txt", promoterLocationCategories, nBins, true, typeCount);
	    }
	    /////////////////////////////////
	    //Save the distance to the annotated TSS for the transcritps corresponding to coding sequences. Also save some other useful info, such as lengths and densities of genes and transcripts
	    String saveName = rnaFileName.substring(0, rnaFileName.lastIndexOf(".")) + CODING_START_END_DIST_SUFFIX + "_" + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    System.out.println("Saving distances for coding transcripts to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    if (chr==1 || chr==-1)
		bw.write("#ID\tChr\tTranscript_start\tTSS_dist\tTranscript_end\tTES_dist\tTranscript_dens\n");
	    for (RefSeqGene g : codingTrans.keySet()) {
		TranscribedRegion t = codingTrans.get(g);
		boolean forward = g.getStrand();
		int startDist, endDist;
		if (forward) {
		    startDist = t.getStart() - g.getStart();
		    endDist = t.getEnd() - g.getEnd();
		}
		else {
		    startDist = t.getStart() - g.getEnd();
		    endDist = t.getEnd() - g.getStart();
		}
		bw.write(g.getID() + "\t" + g.getChromosome() + "\t" + (forward ? t.getStart() : t.getEnd()) + "\t" + startDist + "\t" + (forward ? t.getEnd() : t.getStart()) + "\t" + endDist + "\t" + format.format(t.getLevel()) + "\n");
	    }
	    bw.close();
	    /////////////////////////////////
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static TreeSet<LncRNA> loadLncRNA(String lncRNAFileName, String species) { return loadLncRNA(lncRNAFileName, species, -1); }

    public static TreeSet<LncRNA> loadLncRNA(String lncRNAFileName, String species, int chr) {
	TreeSet<LncRNA> lncRNAs = new TreeSet<LncRNA>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(lncRNAFileName));
	    String line = br.readLine();
	    Pattern lncPattern = Pattern.compile("chr([\\d|X|Y]+):(\\d+)-(\\d+)");
	    while (line!=null) {
		Matcher m = lncPattern.matcher(line);
		if (m.find()) {
		    int c = 0;
		    if (m.group(1).equals("X"))
			c = 20;
		    else if (m.group(1).equals("Y"))
			c = 21;
		    else
			c = Integer.parseInt(m.group(1));
		    if (c==chr || chr==-1)
			lncRNAs.add(new LncRNA(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3))));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error trying to open " + lncRNAFileName);
	}
	catch (NullPointerException e) {
	    System.err.println("Could not find and lincRNAs.");
	}
	System.out.println("Loaded " + lncRNAs.size() + " lncRNAs");
	return lncRNAs;
    }

    public static TreeSet<MiRNA> loadMiRNA(String miRNAFileName, String species) { return loadMiRNA(miRNAFileName, null, species); }

    public static TreeSet<MiRNA> loadMiRNA(String miRNAFileName, String miRNAFileNamePrecursors, String species) {
	//Use the second argument if we have information about
	//precursors as well. We parse that information first.
	TreeSet<MiRNA> miRNAs = new TreeSet<MiRNA>();
	try {
	    if (miRNAFileNamePrecursors!=null) {
		BufferedReader br = new BufferedReader(new FileReader(miRNAFileNamePrecursors));
		String line = br.readLine();
		//Each line should have this pattern twice, the first
		//giving the location of the miRNA and the second
		//giving the coordinates for the precursor.
		Pattern miPattern = Pattern.compile("chr([\\d|X|Y]+)\\s+(\\d+)\\s+(\\d+)");
		Pattern namePattern = Pattern.compile("[mir|let]-[\\w|-]+");
		while (line!=null) {
		    if (line.startsWith("#")) {
			line = br.readLine();
			continue;
		    }
		    Matcher mName = namePattern.matcher(line);
		    if (mName.find()) {
			String name = mName.group(0);
			Matcher m = miPattern.matcher(line);
			if (m.find()) {
			    int c = 0;
			    if (m.group(1).equals("X"))
				c = 20;
			    else if (m.group(1).equals("Y"))
				c = 21;
			    else
				c = Integer.parseInt(m.group(1));
			    boolean strand = line.contains("+");
			    int start = Integer.parseInt(m.group(2)), end = Integer.parseInt(m.group(3));
			    if (m.find(m.end())) { //look for the precursor
				int c2 = 0;
				if (m.group(1).equals("X"))
				    c2 = 20;
				else if (m.group(1).equals("Y"))
				    c2 = 21;
				else
				    c2 = Integer.parseInt(m.group(1));
				if (c==c2)
				    miRNAs.add(new MiRNA(c, start, end, strand, name, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3))));
				else {
				    System.err.println("Warning! Could not find precursor info for " + line);
				    miRNAs.add(new MiRNA(c, start, end, strand, name));
				}
			    }
			    else {
				System.err.println("Warning! Could not find precursor info for " + line);
				miRNAs.add(new MiRNA(c, start, end, strand, name));
			    }
			}
			else
			    System.err.println("Warning! Could not find miRNA info for " + line);
		    }
		    else
			System.err.println("Warning! Could not find miRNA name for " + line);
		    line = br.readLine();
		}
		br.close();
		System.out.println("Found " + miRNAs.size() + " miRNAs with precursors.");
		    }
	    BufferedReader br = new BufferedReader(new FileReader(miRNAFileName));
	    String line = br.readLine();
	    //- chr start end name  - strand
	    Pattern mirnaPattern = Pattern.compile("\\d+\\s+chr([X|Y|\\d]+)\\s+(\\d+)\\s+(\\d+)\\s+mmu-([-|\\w]+)\\s+\\d+\\s+([-|+])");
	    while (line!=null) {
		Matcher m = mirnaPattern.matcher(line);
		if (m.find()) {
		    int c = 0;
		    if (m.group(1).equals("X"))
			c = 20;
		    else if (m.group(1).equals("Y"))
			c = 21;
		    else
			c = Integer.parseInt(m.group(1));
		    miRNAs.add(new MiRNA(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), m.group(5).equals("+"), m.group(4)));
		}
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("Found " + miRNAs.size() + " miRNAs in total.");
	}
	catch (IOException e) {
	    System.err.println("Error trying to open miRNA file " + e.toString());
	}
	catch (NullPointerException e) {
	    System.err.println("Could not find and miRNAs.");
	}
	return miRNAs;
    }

    public static TreeSet<MacroRNA> loadMacroRNA(String macroRNAFileName, String species) { return loadMacroRNA(macroRNAFileName, species, -1); }

    public static TreeSet<MacroRNA> loadMacroRNA(String macroRNAFileName, String species, int chr) {
	//Use the second argument if we have information about
	//precursors as well. We parse that information first.
	TreeSet<MacroRNA> macroRNAs = new TreeSet<MacroRNA>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(macroRNAFileName));
	    String line = br.readLine();
	    //bin    matches misMatches      repMatches      nCount  qNumInsert      qBaseInsert     tNumInsert      tBaseInsert     strand  qName   qSize   qStart qEnd     tName   tSize   tStart  tEnd    blockCount      blockSizes      qStarts tStarts
	    Pattern macrornaPattern = Pattern.compile("([+|-])\\s+(\\w+)\\s+\\d+\\s+\\d+\\s+\\d+\\s+chr([X|Y|\\d]+)\\s+\\d+\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		//System.out.println(line);
		Matcher m = macrornaPattern.matcher(line);
		if (m.find()) {
		    int c = 0;
		    if (m.group(3).equals("X"))
			c = 20;
		    else if (m.group(3).equals("Y"))
			c = 21;
		    else
			c = Integer.parseInt(m.group(3));
		    if (c==chr || chr==-1)
			macroRNAs.add(new MacroRNA(c, Integer.parseInt(m.group(4)), Integer.parseInt(m.group(5)), m.group(1).equals("+"), m.group(2)));
		}
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("Found " + macroRNAs.size() + " macroRNAs in total.");
	}
	catch (IOException e) {
	    System.err.println("Error trying to open macroRNA file " + e.toString());
	}
	catch (NullPointerException e) {
	    System.err.println("Could not find and macroRNAs.");
	}
	return macroRNAs;
    }


    public static int[][] calculateReadDensityDistribution(ABIRead[] reads, int chr, int binSize, String species) {
	//Use this method to calculate the density distribution for a
	//straightforward tiling of the genome of non-overlapping bins
	//sized binSize. Intended to be used from within Matlab.
	TreeMap<Integer, Integer> binReadCount = new TreeMap<Integer, Integer>();
	int i = 1;
	int[] chr_lens = species.equals("mm") ? RefSeqGenes.CHROMOSOME_LENGTH : RefSeqGenes.CHROMOSOME_LENGTH_HS;
	while (i*binSize<chr_lens[chr-1]) {
	    int nReads = FindReadsAcrossFeature.numberOfReads(chr, (i-1)*binSize, i*binSize, reads);
	    binReadCount.put(nReads, binReadCount.containsKey(nReads) ? binReadCount.get(nReads) + 1 : 1);
	    i++;
	}
	int brc[][] = new int[binReadCount.size()][2], j = 0;
	for (Integer r : binReadCount.keySet()) {
	    brc[j][0] = r;
	    brc[j++][1] = binReadCount.get(r);
	}
	return brc;
    }

    public static int[][] calculateReadDensityDistribution(ABIRead[] reads, ABIRead[] inputReads, int chr, int binSize, String species) {
	//Use this method to calculate the density distribution for a
	//straightforward tiling of the genome of non-overlapping bins
	//sized binSize. Intended to be used from within Matlab with
	//ChIP and input samples.
	TreeMap<Integer, Integer> binReadCount = new TreeMap<Integer, Integer>();
	int i = 1;
	int[] chr_lens = species.equals("mm") ? RefSeqGenes.CHROMOSOME_LENGTH : RefSeqGenes.CHROMOSOME_LENGTH_HS;
	while (i*binSize<chr_lens[chr-1]) {
	    int nReads = FindReadsAcrossFeature.numberOfReads(chr, (i-1)*binSize, i*binSize, reads);
	    int nInputReads = FindReadsAcrossFeature.numberOfReads(chr, (i-1)*binSize, i*binSize, inputReads);
	    int diffReads = nReads - nInputReads;
	    binReadCount.put(diffReads, binReadCount.containsKey(diffReads) ? binReadCount.get(diffReads) + 1 : 1);
	    i++;
	}
	int brc[][] = new int[binReadCount.size()][2], j = 0;
	for (Integer r : binReadCount.keySet()) {
	    brc[j][0] = r;
	    brc[j++][1] = binReadCount.get(r);
	}
	return brc;
    }

    public static int[][] calculateReadDensityDistribution(ABIRead[] reads, boolean strand, RefSeqGenes geneInfo, int binSize, boolean nonGenicOnly, String species) {
	//Use this method to calculate the density distribution for a
	//straightforward tiling of the genome of non-overlapping bins
	//sized binSize. Intended to be used from within Matlab.
	TreeMap<Integer, Integer> binReadCount = new TreeMap<Integer, Integer>();
	int i = 1, chr = geneInfo.getChromosome();
	int[] chr_lens = species.equals("mm") ? RefSeqGenes.CHROMOSOME_LENGTH : RefSeqGenes.CHROMOSOME_LENGTH_HS;
	while (i*binSize<chr_lens[chr-1]) {
	    boolean binOverlapsGene = geneInfo.overlapsGene((i-1)*binSize, i*binSize, strand);
	    if ((nonGenicOnly && !binOverlapsGene) || (!nonGenicOnly && binOverlapsGene)) {
		int nReads = FindReadsAcrossFeature.numberOfReads(chr, (i-1)*binSize, i*binSize, reads);
		binReadCount.put(nReads, binReadCount.containsKey(nReads) ? binReadCount.get(nReads) + 1 : 1);
	    }
	    i++;
	}
	int brc[][] = new int[binReadCount.size()][2], j = 0;
	for (Integer r : binReadCount.keySet()) {
	    brc[j][0] = r;
	    brc[j++][1] = binReadCount.get(r);
	}
	return brc;
    }

    public static int[][] calculateIntronExonReadDensityDistribution(ABIRead[] reads, boolean strand, RefSeqGenes geneInfo, int binSize, boolean intron, String species) {
	//Use this method to calculate the density distribution for a
	//straightforward tiling of the genome of non-overlapping bins
	//sized binSize. Intended to be used from within Matlab.
	TreeMap<Integer, Integer> binReadCount = new TreeMap<Integer, Integer>();
	int i = 1, chr = geneInfo.getChromosome();
	int[] chr_lens = species.equals("mm") ? RefSeqGenes.CHROMOSOME_LENGTH : RefSeqGenes.CHROMOSOME_LENGTH_HS;
	while (i*binSize<chr_lens[chr-1]) {	
	    RefSeqGene gene = geneInfo.getOverlappingGene((i-1)*binSize, i*binSize, strand);
	    if (gene!=null) {
		boolean overlapsExon = gene.overlapsExon((i-1)*binSize, i*binSize); 
		if ((intron && !overlapsExon) || (!intron && overlapsExon)) {
		    int nReads = FindReadsAcrossFeature.numberOfReads(chr, (i-1)*binSize, i*binSize, reads);
		    binReadCount.put(nReads, binReadCount.containsKey(nReads) ? binReadCount.get(nReads) + 1 : 1);
		}
	    }
	    i++;
	}
	int brc[][] = new int[binReadCount.size()][2], j = 0;
	for (Integer r : binReadCount.keySet()) {
	    brc[j][0] = r;
	    brc[j++][1] = binReadCount.get(r);
	}
	return brc;
    }

    public static double[] setMiRNAFactorReads(TreeSet<MiRNA> miRNAs, String rnaFileName, ABIRead[] reads, int chr, int strand) {
	//Use to find out the read densities across miRNAs
	Vector<Double> dens = new Vector<Double>();
	if (miRNAs!=null) {
	    for (MiRNA m : miRNAs) {
		if (m.getChromosome()==chr && (strand==0 || (strand>0 && m.getStrand()) || (strand<0 && !m.getStrand()))) {
		    int startInd = Arrays.binarySearch(reads, new ABIRead(chr, m.getStrand(), m.getStart()));
		    startInd = startInd<0 ? -startInd-1 : startInd;
		    int endInd = Arrays.binarySearch(reads, new ABIRead(chr, m.getStrand(), m.getEnd()));
		    endInd = endInd<0 ? -endInd-1 : endInd;
		    ABIRead rs[] = new ABIRead[endInd - startInd];
		    for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		    m.setFactorReads(rnaFileName, rs);
		    dens.addElement(m.getReadDensity(rnaFileName));
		    //if (rs.length>0)
		    //  System.out.println(m.toString() + "\t" + rs.length);
		}
	    }
	}
	double ds[] = new double[dens.size()];
	for (int i=0; i<ds.length; i++)
	    ds[i] = dens.elementAt(i);
	return ds;
    }

    public static double[] setMaRNAFactorReads(TreeSet<MacroRNA> macroRNAs, String rnaFileName, ABIRead[] reads, int chr, int strand) {
	//Use to find out the read densities across macroRNAs
	Vector<Double> dens = new Vector<Double>();
	if (macroRNAs!=null) {
	    for (MacroRNA m : macroRNAs) {
		if (m.getChromosome()==chr && (strand==0 || (strand>0 && m.getStrand()) || (strand<0 && !m.getStrand()))) {
		    int startInd = Arrays.binarySearch(reads, new ABIRead(chr, m.getStrand(), m.getStart()));
		    startInd = startInd<0 ? -startInd-1 : startInd;
		    int endInd = Arrays.binarySearch(reads, new ABIRead(chr, m.getStrand(), m.getEnd()));
		    endInd = endInd<0 ? -endInd-1 : endInd;
		    ABIRead rs[] = new ABIRead[endInd - startInd];
		    for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		    m.setFactorReads(rnaFileName, rs);
		    dens.addElement(m.getReadDensity(rnaFileName));
		}
	    }
	}
	double ds[] = new double[dens.size()];
	for (int i=0; i<ds.length; i++)
	    ds[i] = dens.elementAt(i);
	return ds;
    }

    public static double[] setLncRNAFactorReads(TreeSet<LncRNA> lncRNAs, String rnaFileName, ABIRead[] reads, int chr) {
	//Use to find out the read densities across lncRNAs, here we
	//don't have any strand info, so everything is merged
	Vector<Double> dens = new Vector<Double>();
	if (lncRNAs!=null) {
	    for (LncRNA l : lncRNAs) {
		if (l.getChromosome()==chr) {
		    int startInd = Arrays.binarySearch(reads, new ABIRead(chr, true, l.getStart()));
		    startInd = startInd<0 ? -startInd-1 : startInd;
		    int endInd = Arrays.binarySearch(reads, new ABIRead(chr, true, l.getEnd()));
		    endInd = endInd<0 ? -endInd-1 : endInd;
		    ABIRead rs[] = new ABIRead[endInd - startInd];
		    for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		    l.setFactorReads(rnaFileName, rs);
		    dens.addElement(l.getReadDensity(rnaFileName));
		}
	    }
	}
	double ds[] = new double[dens.size()];
	for (int i=0; i<ds.length; i++)
	    ds[i] = dens.elementAt(i);
	return ds;
    }

    public static void saveTranscriptFactorProperties(TreeSet<TranscribedRegion> segments, TreeMap<String, Integer> nTotalReads, String inputFileName, String saveFileName, boolean header) {
	try {
	    System.out.println("Saving read info to " + saveFileName);
	    Pattern winPattern = Pattern.compile("_(start)_w_(\\d+)");
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (segments.size()>0) {
		TreeSet<String> factors = segments.first().getFactors();
		if (header) {
		    bw.write("#Chr\tStart\tEnd");
		    for (String name : factors) {
			if (inputFileName==null || !name.contains(inputFileName)) {
			    bw.write("\t" + name.substring(name.lastIndexOf("/")+1));
			}
		    }
		    bw.write("\n");
		}
		String upDownSuffix[] = new String[]{"_u", "_d"};
		for (TranscribedRegion t : segments) {
		    bw.write(t.getChromosome() + "\t" + t.getStart() + "\t" + t.getEnd());
		    for (String name : factors) {
			if (inputFileName==null) 
			    bw.write("\t" + format.format(t.getReadDensity(name))); //*CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(name)));
			else if (!name.contains(inputFileName)) {
			    Matcher m = winPattern.matcher(name);
			    if (m.find()) {
				bw.write("\t" + format.format((double)t.getFactorReads(name).length*CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(name.substring(0, name.lastIndexOf("_" + m.group(1)))) - (double)t.getFactorReads(inputFileName + "_" + m.group(1) + "_w_" + m.group(2)).length*CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(inputFileName)));
			    }
			    else
				bw.write("\t" + format.format(t.getReadDensity(name)*CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(name) - t.getReadDensity(inputFileName)*CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(inputFileName)));
			}
		    }
		    bw.write("\n");
		}
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving reads info " + e.toString());
	}	
    }

}