//Use this to represent a bin of a fixed width when scanning the
//genome for enriched regions in terms of the number of reads found
//(compared to input). Once the enriched windows are found we can
//merge them and pull out the regions with contiguous overlapping
//fragments.

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.text.*;

public class ChromosomeWindow implements Comparable {

    public static final int MIN_CLUSTER_SIZE = 3;

    protected TreeSet reads = new TreeSet(), largestCluster = null, inputReads = new TreeSet();
    protected int nInputReads = 0, chr, potReads, start = Integer.MAX_VALUE, end = 0;
    //Keep track of where the part of the window which is above the
    //background level starts and ends
    protected int significantStart, significantEnd;
    protected int constRegions[] = null;
    protected double gcMultiplier, peakObservationProbability, lambda = 1;
    //Calculating the log of the p-value allows us to retain better
    //precision, also for calculating the ratio
    protected double logPValue, logInputPValue;
    protected String sequence = null;
    protected boolean sizeFixed = false;
    //Keep track of the number of reads which must be removed in order
    //to drop below the detection limit
    protected int dist2detectionLimit = 0;

    //Use this stuff to keep track of where the nearest genes are
    protected MRNAInfoEntry nearestGene = null;
    //and the closest miRNA
    protected MiRNAInfoEntry nearestMiRNA = null;

    protected DecimalFormat format = new DecimalFormat("#######.###");

    //////////////////////////////////////////////////////////////////

    public ChromosomeWindow(int c, int s, int e) {
	chr = c;
	start = s; 
	end = e;
    }

    public ChromosomeWindow(int c, int s, int e, TreeSet reads) {
	chr = c;
	start = s; 
	end = e;
	Iterator it = reads.iterator();
	while (it.hasNext()) { reads.add((ABIRead)it.next()); }
    }

    public ChromosomeWindow(ABIRead r, int c) { 
	reads.add(r); 
	chr = c; 
	start = r.getPosition();
    }

    public ChromosomeWindow(int c) { 
	chr = c; 
    }

    public ChromosomeWindow(ChromosomeWindow win) {
	chr = win.chr;
	start = win.start;
	end = win.end;
	lambda = win.lambda;
	Iterator it = win.reads.iterator();
	while (it.hasNext())
	    reads.add((ABIRead)it.next());
	it = win.inputReads.iterator();
	while (it.hasNext())
	    inputReads.add((ABIRead)it.next());
	potReads = win.potReads;
	sequence = win.sequence;
    }
    

    public ChromosomeWindow(ChromosomeWindow fWin, ChromosomeWindow rWin) {
	//Use this when combining forward and reverse windows
	chr = fWin.chr;
	start = fWin.start;
	end = fWin.end;
	Iterator it = fWin.reads.iterator();
	while (it.hasNext())
	    reads.add((ABIRead)it.next());
	it = rWin.reads.iterator();
	while (it.hasNext())
	    reads.add((ABIRead)it.next());
	potReads = fWin.potReads + rWin.potReads;
    }

    protected ChromosomeWindow(ChromosomeWindow win, int s, int e) {
	//Use this constructor to create a new window by trimming an
	//existing one. Find all reads (ChIP and input) which could
	//belong to a window from s to e.
	start = s;
	end = e;
	chr = win.getChromosome();
	/*for (int i=0; i<win.reads.length; i++) {
	    if (win.reads[i].overlaps(start, end))
		add(win.reads[i]);
		}*/
	sequence = win.getSequence().substring(start-win.getStart(), end - win.getStart());
	lambda = win.lambda;
    }
    
    //////////////////////////////////////////////////////////////////

    public void add(ABIRead r) { 
	reads.add(r); 
	if (r.chr!=chr)
	    System.err.println("Adding read with different chr:\t" + r.toString());
	if (!sizeFixed) {
	    if (r.getPosition()<start) { start = r.getPosition(); }
	    if (r.getPosition() + r.getLength()>end) { end = r.getPosition() + r.getLength(); }
	}
	//else if(r.pos<getStart() || r.pos>getEnd())
	//  System.err.println("Adding read outside window!");
    }

    public void add(ABIRead r[]) { for (int i=0; i<r.length; i++) { add(r[i]); } }

    public ChromosomeWindow merge(ChromosomeWindow win, double reads2SitesRatio, double inputReads2SitesRatio) {
	//Assume that win follows this, but check just to make sure
	if (getEnd()!=win.getStart() || getChromosome()!=win.getChromosome()) {
	    System.err.println("Windows not adjacent, cannot merge!");
	    return null;
	}
	ChromosomeWindow newWin = new ChromosomeWindow(getChromosome(), getStart(), win.getEnd());
	Iterator it = reads.iterator();
	while (it.hasNext()) { newWin.add((ABIRead)it.next()); }
	it = win.reads.iterator();
	while (it.hasNext()) { newWin.add((ABIRead)it.next()); }
	newWin.setInputReads(numberOfInputReads() + win.numberOfInputReads());
	newWin.setPotentialNumberOfReads(numberOfPotentialReads() + win.numberOfPotentialReads());
	newWin.setLambda(lambda);
	newWin.setLogPValue(PeakFinder2.calculateLogPValue(newWin, reads2SitesRatio));
	newWin.setLogInputPValue(PeakFinder2.calculateLogInputPValue(newWin, inputReads2SitesRatio));	    
	return newWin;
    }

    public ChromosomeWindow merge(ChromosomeWindow win) {
	//Assume that win overlaps by just one bp
	//if (getEnd()!=win.getEnd()-1 || getChromosome()!=win.getChromosome()) {
	//  System.err.println("Windows not adjacent, cannot merge!\n" + toString() + "\n" + win.toString());
	//  return null;
	//}
	ChromosomeWindow newWin = new ChromosomeWindow(getChromosome(), getStart(), win.getEnd());
	Iterator it = reads.iterator();
	newWin.setLambda(lambda);
	while (it.hasNext()) { newWin.add((ABIRead)it.next()); }
	it = win.reads.iterator();
	while (it.hasNext()) { newWin.add((ABIRead)it.next()); }
	it = inputReads.iterator();
	while (it.hasNext()) { newWin.addInputRead((ABIRead)it.next()); }
	it = win.inputReads.iterator();
	while (it.hasNext()) { newWin.addInputRead((ABIRead)it.next()); }
	if (sequence!=null && win.getSequence()!=null)
	    newWin.setSequence(getSequence() + win.getSequence().substring(win.getSequence().length()-(win.getEnd()-getEnd())));
	if (newWin.getSequence().length()!=newWin.getLength())
	    System.err.println("Error merging, seq has length " + newWin.getSequence().length() + " and window " + newWin.getLength() + ", " + getLength() + " and " + (win.getLength()) + " and " + win.getSequence().length());
	return newWin;
    }

    public TreeSet trimWindow(int windowSize) {
	//Use this method to reduce the window to a specific size
	//System.out.println("Trimming " + toString());
	int centerOfMass = getCenterOfMass(); //Find the center of mass
	start = centerOfMass - windowSize/2;
	end = centerOfMass + windowSize/2;
	//Remove all reads which are outside the window
	TreeSet removeSet = new TreeSet();
	Iterator it = reads.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (r.getPosition()<start || r.getPosition() + r.getLength()>end) 
		removeSet.add(r);
	}
	it = removeSet.iterator();
	while (it.hasNext()) 
	    reads.remove((ABIRead)it.next());
	constRegions = null;
	return removeSet;
    }

    public ChromosomeWindow[] splitWindow() {
	//Split based on the threshold computed from the background
	//intensity of average reads per bp.
	return splitWindow((int)Math.round(lambda*getLength()));
    }

    public ChromosomeWindow[] splitWindow(int minReads) {
	//Split the windows so that there are no portions containing
	//fewer than minReads reads. Morevoer, after the split,
	//windows must have at least a total of minReads reads. Return
	//a list of new windows.
	getConstantRegions();
	ChromosomeWindow newWins[] = new ChromosomeWindow[1];
	//Find the coordinates where the read count crosses beneath the threshold
	int upCrossings[] = getCrossings(minReads, true);
	int downCrossings[] = getCrossings(minReads, false);
	if (upCrossings.length>1 && downCrossings.length>1) {
	    //Create a new window containing all the reads which overlap the region between upCrossings[i] and downCrossings[i]
	    newWins = new ChromosomeWindow[upCrossings.length];
	    for (int i=0; i<upCrossings.length; i++) 
		newWins[i] = new ChromosomeWindow(this, upCrossings[i], downCrossings[i]);
	}
	else if (upCrossings.length!=downCrossings.length)
	    System.err.println("Different number of crossings up and down for window " + toString());
	else
	    newWins[0] = this;
	return newWins;
    }

    protected void splitReads(ChromosomeWindow newWins[], int cutOffLeft[], int cutOffRight[], boolean input) {
	Iterator it = input ? inputReads.iterator() : reads.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    for (int i=0; i<cutOffLeft.length; i++) {
		if (r.getPosition()<cutOffLeft[i]) {
		    if (input)
			newWins[i].addInputRead(r);
		    else
			newWins[i].add(r);
		    break; //Found peak where left end starts
		}
	    }
	    for (int i=cutOffRight.length-1; i>=0; i--) { //check for peak where right end starts
		if (r.getPosition() + r.getLength()>cutOffRight[i]) {
		    if (input)
			newWins[i+1].addInputRead(r);
		    else
			newWins[i+1].add(r);
		    break;
		}
	    }
	}
    }

    public void shiftRight() { shiftRight(1); }

    public void shiftRight(int n) {
	//Slide the window n basepairs to the right and discard all
	//reads which are no longer in the window.
	start += n;
	end += n;
	//System.out.println(sequence);
	if (sequence!=null) { sequence = sequence.substring(Math.min(n, sequence.length()-1)); }
	Iterator it = reads.iterator();
	TreeSet removeSet = new TreeSet();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (r.getPosition()<start)
		removeSet.add(r);
	    else
		break;
	}
	it = removeSet.iterator();
	while (it.hasNext()) 
	    reads.remove((ABIRead)it.next());
	//Remove the input reads
	it = inputReads.iterator();
	removeSet = new TreeSet();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (r.getPosition()<start)
		removeSet.add(r);
	    else
		break;
	}
	it = removeSet.iterator();
	while (it.hasNext()) 
	    inputReads.remove((ABIRead)it.next());
    }

    public boolean subsetOf(ChromosomeWindow win) {
	//Returns true if all the reads can be found in win
	if (numberOfReads()<win.numberOfReads()) {
	    Iterator it = reads.iterator();
	    while (it.hasNext()) {
		if (!win.reads.contains((ABIRead)it.next()))
		    return false;
	    }
	    return true;
	}
	return false;
    }

    
    //////////////////////////////////////////////////////////////////

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public boolean overlaps(int p) { return p>=start && p<=end; }

    public boolean overlaps(int p1, int p2) { return p1>=start && p2<=end; }

    public boolean overlapsSignificant(int p1, int p2) { return p1>=significantStart && p2<=significantEnd; }

    public int getCenter() { return (getStart() + getEnd())/2; }

    public int getLength() { return getEnd() - getStart(); }

    public TreeSet getReads() { return reads; }

    public ABIRead[] getReadsArray() {
	ABIRead r[] = new ABIRead[numberOfReads()];
	Iterator it = reads.iterator();
	int i = 0;
	while (it.hasNext())
	    r[i++] = (ABIRead)it.next();
	return r;
    }

    public int getChromosome() { return chr; }

    public void setInputReads(int n) { nInputReads = n; }

    //use this when adding anonymous input reads to acheive background coverage
    public void addInputRead() { nInputReads++; }

    public void addInputRead(ABIRead r) { inputReads.add(r); }

    public int numberOfInputReads() { return nInputReads + inputReads.size(); }

    public int numberOfInputReads(boolean actualOnly) { return actualOnly ? inputReads.size() : numberOfInputReads(); }

    public int numberOfReads() { return reads.size(); }

    public int numberOfNewReads(ChromosomeWindow win) { 
	//Calculate the number of reads which are in this but not in
	//win. This allows us to deduce the total number of reads when
	//merging adjacent windows.
	int nNew = 0;
	if (win!=null) { 
	    //System.out.println("Comparing " + toString() + "\n" + win.toString());
	    Iterator it = reads.iterator();
	    while (it.hasNext()) {
		ABIRead r = (ABIRead)it.next();
		nNew += win.reads.contains(r) ? 0 : 1;
	    }
	}
	else
	    nNew = reads.size();
	return nNew;
    }

    public int numberOfNewInputReads(ChromosomeWindow win) { 
	//Calculate the number of input reads which are in this but
	//not in win. This allows us to deduce the total number of
	//reads when merging adjacent windows.
	int nNew = 0;
	if (win!=null) { 
	    Iterator it = inputReads.iterator();
	    while (it.hasNext()) {
		ABIRead r = (ABIRead)it.next();
		nNew += win.inputReads.contains(r) ? 0 : 1;
	    }
	}
	else
	    nNew = inputReads.size();
	return nNew;
    }

    public void setPotentialNumberOfReads(int n) { potReads = n; }

    public int numberOfPotentialReads() { return potReads; }

    //public void setPValue(double p) { pValue = p; }

    public double getPValue() { return Math.exp(logPValue); }

    //public void setInputPValue(double p) { inputPValue = p; }

    public double getInputPValue() { return Math.exp(logInputPValue); }

    public void setLogPValue(double p) { logPValue = p; }

    public double getLogPValue() { return logPValue; }

    public void setLogInputPValue(double p) { logInputPValue = p; }

    public double getLogInputPValue() { return logInputPValue; }

    public double getPValueRatio() { return Math.exp(logPValue - logInputPValue); }

    public double getPValueLogRatio() { return logPValue - logInputPValue; }


    public double getPeakQualityScore() { 
	if (getPValue()>=0 && getInputPValue()>0)
	    return -(Math.log(getPValue()) - Math.log(getInputPValue())); 
	return 0;
    }

    public void setGCMultiplier(double m) { gcMultiplier = m; }

    public double getGCMultiplier() { return gcMultiplier; }

    public double getReadsMultiplier(int fragmentLength) { return gcMultiplier*((double)(getLength() - fragmentLength)*2)/numberOfPotentialReads(); }

    public int calculateEffectiveReadDiff(int fragmentLength) {
	if (numberOfReads()>MIN_CLUSTER_SIZE) {
	    double mappabilityMultiplier = ((double)(getLength() - fragmentLength)*2)/numberOfPotentialReads();
	    double nReadsMod = ((double)numberOfReads())*getGCMultiplier()*mappabilityMultiplier;
	    double nInputReadsMod = ((double)numberOfInputReads())*getGCMultiplier()*mappabilityMultiplier;
	    return (int)Math.round(nReadsMod - nInputReadsMod);
	}
	return 0;
    }

    public void setPeakObservationProbability(double p) { peakObservationProbability = p; }
    
    public double getPeakObservationProbability() { return peakObservationProbability; }

    public void setLambda(double l) { lambda = l; }

    public double getEnrichment() { return (double)numberOfReads()/(lambda*(double)numberOfPotentialReads()); }

    public double getEnrichment(double lambda) { return (double)numberOfReads()/(lambda*(double)numberOfPotentialReads()); }

    public double averageReadsMismatches(boolean chip) { return chip ? averageReadsMismatches() : averageInputReadsMismatches(); }

    public double averageReadsMismatches() {
	if (numberOfReads()==0) { return -1; }
	double s = 0;
	Iterator it = reads.iterator();
	while (it.hasNext())
	    s += (double)((ABIRead)it.next()).numberOfMismatches();
	return s/numberOfReads();
    }

    public double averageInputReadsMismatches() {
	if (numberOfInputReads()==0) { return -1; }
	double s = 0;
	Iterator it = inputReads.iterator();
	while (it.hasNext())
	    s += (double)((ABIRead)it.next()).numberOfMismatches();
	return s/numberOfInputReads();
    }

    public void setSequence(String seq) { sequence = seq;  }

    public void addSequence(String seq) { sequence += seq; }

    public String getSequence() { 
	if (sequence==null) { //patch the sequence together from the information contained in the reads
	    Iterator it = reads.iterator();
	    ABIRead r = (ABIRead)it.next();
	    sequence = r.getSequence();
	    int prevReadPos = r.getPosition();
	    while (it.hasNext()) {
		r = (ABIRead)it.next();
		if (r.getSequence()!=null)
		    sequence += r.getSequence().substring(r.getPosition() - prevReadPos);
		prevReadPos = r.getPosition();
	    }
	}
	return sequence; 
    }

    public void setDetectionLimitDistance(int d) { dist2detectionLimit = d; }

    public void setDetectionLimitDistance(int fragmentLength, int unweightedMinDistance) { 
	//Calculate the distance to the detection limit by applying
	//the weights from the mappability and the GC-bias
	//dist2detectionLimit = Math.max(0, (int)Math.round((double)((numberOfReads() - numberOfInputReads()) - unweightedMinDistance)*getReadsMultiplier(fragmentLength)));
	dist2detectionLimit = Math.max(0, (int)Math.round((double)((numberOfReads() - numberOfInputReads()))*getReadsMultiplier(fragmentLength) - unweightedMinDistance));
    }

    public int getDetectionLimitDistance() { return dist2detectionLimit; }

    public double gcRatio(){ return (double)gcContent()/sequence.length(); }

    public int gcContent() { return gcContent(sequence); }

    public static int gcContent(String s){
	//Calculate the proportion of bps which are G or C
	char b[] = s.toCharArray();
	int gc = 0;
	for (int i=0; i<b.length; i++) {
	    if (b[i]=='c' || b[i]=='C' || b[i]=='g' || b[i]=='G') { gc++; }
	}
	return gc;
    }

    public int getPeakHeight() {
	getConstantRegions();
	int max = 0;
	for (int i=0; i<constRegions.length; i++) 
	    max = max>constRegions[i] ? max : constRegions[i];
	return max;
    }

    public int getPeakLocation() {
	getConstantRegions();
	int max = 0, maxInd = 0;
	for (int i=0; i<constRegions.length; i++) {
	    if (max<constRegions[i]) {
		max = constRegions[i];
		maxInd = i;
	    }
	}
	return getStart() + maxInd;
    }

    public int[] getLocalMaxima() {
	int bedRegs[][] = getBedFormat(), nMax = 0;
	for (int i=1; i<bedRegs.length-1; i++) {
	    if (bedRegs[i][2]>bedRegs[i-1][2] && bedRegs[i][2]>bedRegs[i+1][2]) { nMax++; }
	}
	int max[] = new int[nMax], j = 0;
	for (int i=1; i<bedRegs.length-1; i++) {
	    if (bedRegs[i][2]>bedRegs[i-1][2] && bedRegs[i][2]>bedRegs[i+1][2]) { max[j++] = i; }
	}
	return max;
    }

    public int[] getLocalMinima() { return getLocalMinima(0); }

    public int[] getLocalMinima(int minReads) {
	int bedRegs[][] = getBedFormat(), nMin = 0;
	for (int i=1; i<bedRegs.length-1; i++) {
	    if (bedRegs[i][2]<bedRegs[i-1][2] && bedRegs[i][2]<bedRegs[i+1][2] && bedRegs[i][2]<=minReads) { nMin++; }
	}
	int min[] = new int[nMin], j = 0;
	for (int i=1; i<bedRegs.length-1; i++) {
	    if (bedRegs[i][2]<bedRegs[i-1][2] && bedRegs[i][2]<bedRegs[i+1][2] && bedRegs[i][2]<=minReads) { min[j++] = i; }
	}
	return min;
    }

    public int[] getCrossings(int minReads, boolean up) {
	//Find out where the number of reads become greater than 
	int nCrossings = 0, crossPoints[] = new int[0];
	getConstantRegions();
	for (int i=2; i<constRegions.length-1; i++) {
	    if (up)
		nCrossings += constRegions[i-2]<minReads && constRegions[i-1]<=minReads && constRegions[i]>minReads && constRegions[i+1]>minReads ? 1 : 0;
	    else if (!up)
		nCrossings += constRegions[i-2]>minReads && constRegions[i-1]>=minReads && constRegions[i]<minReads && constRegions[i+1]<minReads ? 1 : 0;
	}
	if (nCrossings>0) {
	    int j = 0;
	    crossPoints = new int[nCrossings];
	    for (int i=2; i<constRegions.length-1; i++) {
		if (up && constRegions[i-2]<minReads && constRegions[i-1]<=minReads && constRegions[i]>minReads && constRegions[i+1]>minReads)
		    crossPoints[j++] = i;
		else if(!up && constRegions[i-2]>minReads && constRegions[i-1]>=minReads && constRegions[i]<minReads && constRegions[i+1]<minReads)
		    crossPoints[j++] = i;
	    }
	}
	return crossPoints;
    }

    public int getCenterOfMass() {
	getConstantRegions();
	int cumMass[] = new int[getLength()], j = 0;
	cumMass[0] = constRegions[0];
	for (int i=1; i<cumMass.length; i++) { cumMass[i] = cumMass[i-1] + constRegions[i]; }
	//Find the index for half the total mass
	while (cumMass[j++]<cumMass[cumMass.length-1]/2) { }
	return getStart() + j;
    }

    public int getCenterOfMass(boolean overlapping) {
	if (!overlapping)
	    return getCenterOfMass();
	int largestConstRegion[] = getLargestConstantRegions();
	int cumMass[] = new int[getLength()], j = 0;
	if (largestConstRegion[0]>0 && largestConstRegion[AnalyzeControls.READ_LENGTH+1]>0)
	    cumMass[0] = largestConstRegion[0];
	else
	    cumMass[0] = 0;
	for (int i=1; i<cumMass.length; i++) { 
	    if (i+AnalyzeControls.READ_LENGTH+1<cumMass.length && largestConstRegion[i]>0 && largestConstRegion[i+AnalyzeControls.READ_LENGTH+1]>0)
		cumMass[i] = cumMass[i-1] + largestConstRegion[i]; 
	    else
		cumMass[i] = cumMass[i-1];
	}
	//Find the index for half the total mass
	while (cumMass[j++]<cumMass[cumMass.length-1]/2) { }
	return getStart() + j;
    }

    public int getSignificantStart() { return significantStart; }

    public void setSignificantStart(int s) { significantStart = s; }

    public int getSignificantEnd() { return significantEnd; }

    public void setSignificantEnd(int s) { significantEnd = s; }

    public boolean readsProtruding() {
	//Find out if there are any reads which stick out on the left
	//hand side of the window. We use this to detect windows which
	//may be overlapping peaks which sit in between two
	//windows. We need to discard these since they skew the
	//estimate for the peak distances.
	Iterator it = reads.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (r.getPosition() + r.getLength()>getEnd())
		return true;
	}
	return false;
    }

    public boolean readsProtruding(boolean largest) {
	//Find out if there are any reads which stick out on the left
	//hand side of the window. We use this to detect windows which
	//may be overlapping peaks which sit in between two
	//windows. We need to discard these since they skew the
	//estimate for the peak distances. If the argument is true,
	//then we only consider the largest cluster of reads.
	if (!largest)
	    return readsProtruding();
	getLargestCluster();
	Iterator it = largestCluster.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (r.getPosition() + r.getLength()>getEnd())
		return true;
	}
	return false;
    }

    public boolean readsOverlapping() {
	//Find out if all of the reads in the window are overlapping
	//to form a contiguous peak.
	Iterator it = reads.iterator();
	int prevReadPos = ((ABIRead)it.next()).getPosition();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (prevReadPos + r.getLength()<r.getPosition())
		return false;
	    prevReadPos = r.getPosition();
	}
	return true;
    }

    public int numberOfReadsProtruding() {
	//Find out if all of the reads in the window are overlapping
	//to form a contiguous peak.
	Iterator it = reads.iterator();
	int nOverlap = 0;
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (r.getPosition() + r.getLength()>getEnd())
		nOverlap++;
	}
	return nOverlap;
    }


    //////////////////////////////////////////////////////////////////

    public int[] getConstantRegions(){
	//Use this method to extract the information about how many
	//reads map to each basepair. This information is used for
	//creating .bed-files.
	if (constRegions!=null)
	    return constRegions;
	constRegions = new int[getLength()];
	Iterator it = reads.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    for (int i=0; i<r.getLength(); i++) { if (r.getPosition()-getStart()+i<getLength()) { constRegions[r.getPosition()-getStart()+i]++;	} }
	}
	return constRegions;
    }

    public int[][] getBedFormat(){ return getBedFormat(0); }

    public int[][] getBedFormat(int minReads){
	//Use this method to extract the information about how many
	//reads map to each basepair. It is essentially the same info
	//as getConstantRegions(), but formatted in a more compact way
	//suitable for .bed-files. The argument allows us to only show
	//the part of the peak which is above minReads.
	getConstantRegions();
	//Find out how many plateaus there are
	int nPlateaus = 1;
	for (int i=1; i<constRegions.length; i++) {
	    if (constRegions[i]!=constRegions[i-1] && constRegions[i]>minReads)
		nPlateaus++;
	}
	int plateaus[][] = new int[nPlateaus][3], j = 0;
	plateaus[j][0] = getStart() - 1;
	for (int i=1; i<constRegions.length; i++) {
	    if (constRegions[i]!=constRegions[i-1] && constRegions[i]>minReads) {
		plateaus[j][1] = getStart() + i - 1;
		plateaus[j][2] = constRegions[i-1];
		j++;
		plateaus[j][0] = getStart() + i - 1;
	    }
	}
	//Add the last plateau, find when it drops below the threshold
	int i = constRegions.length - 1;
	while (constRegions[i--]<minReads) { }
	plateaus[j][1] = getStart() + i - 1;
	plateaus[j][2] = constRegions[i];
	return plateaus;
    }

    public String printBedFormat() { return printBedFormat(0); }

    public String printBedFormat(int minReads) {
	int bedRegions[][] = getBedFormat();
	String str = new String();
	for (int i=0; i<bedRegions.length; i++)
	    str += bedRegions[i][0] + "\t" + bedRegions[i][1] + "\t" + bedRegions[i][2] + "\n";
	return str;
    }

    public TreeSet getLargestCluster() {
	//Use this method to locate the set of overlapping reads with
	//the largest number of reads in the window.
	if (largestCluster!=null)
	    return largestCluster;
	TreeSet cluster = new TreeSet();
	Vector clusters = new Vector();
	Iterator it = reads.iterator();
	int lastBase = Integer.MAX_VALUE;
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    if (!cluster.isEmpty() && r.getPosition()>lastBase) {
		clusters.addElement(cluster);
		cluster = new TreeSet();
	    }
	    if (cluster.isEmpty() || r.getPosition()<lastBase) {
		cluster.add(r);
		lastBase = r.getPosition() + r.getLength();
	    }
	}
	clusters.addElement(cluster);
	//Find which was the largest one
	largestCluster = (TreeSet)clusters.elementAt(0);
	if (clusters.size()>1) {
	    for (int i=1; i<clusters.size(); i++) {
		if (((TreeSet)clusters.elementAt(i)).size()>largestCluster.size())
		    largestCluster = (TreeSet)clusters.elementAt(i);
	    }
	}
	return largestCluster;
    }
    
    public int[] getLargestConstantRegions() {
	//Find the largest set of overlapping reads
	getLargestCluster();
	int largestConstRegion[] = new int[getLength()];
	Iterator it = largestCluster.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.next();
	    for (int i=0; i<r.getLength(); i++) { if (r.getPosition()-getStart()+i<getLength()) { largestConstRegion[r.getPosition()-getStart()+i]++;	} }
	}
	return largestConstRegion;
    }

    //////////////////////////////////////////////////////////////////

    public void setNearestGene(MRNAInfoEntry m) { nearestGene = m; }

    public MRNAInfoEntry getNearestGene() { return nearestGene; }

    public boolean isNearestGene(MRNAInfoEntry m) { 
	if (nearestGene!=null) { return nearestGene.equals(m); } 
	return false;
    }

    public int getNearestGeneDistance() { return Math.min(Math.abs(nearestGene.getStart() - getCenterOfMass()), Math.abs(nearestGene.getEnd() - getCenterOfMass())); }

    public void setNearestMiRNA(MiRNAInfoEntry m) { nearestMiRNA = m; }

    public MiRNAInfoEntry getNearestMiRNA() { return nearestMiRNA; }

    public boolean isNearestMiRNA(MiRNAInfoEntry m) { 
	if (nearestMiRNA!=null) { return nearestMiRNA.equals(m); } 
	return false;
    }

    public int getNearestMiRNADistance() { 
	if (nearestMiRNA!=null)
	    return Math.min(Math.abs(nearestMiRNA.getStart() - getCenterOfMass()), Math.abs(nearestMiRNA.getEnd() - getCenterOfMass())); 
	return Integer.MAX_VALUE;
    }

    //////////////////////////////////////////////////////////////////

    //Use this method to get the formatting for the toString() fcn
    public static String getHeaderString() { return "Chr\tStart\tEnd\tReads\tInput\tPotential\tQuality"; }

    public String toString() {
	//Trim the window so that 
	String str = "chr" + chr + "\t" + getStart() + "\t" + getEnd() + "\t" + numberOfReads() + "\t" + numberOfInputReads() + "\t" + numberOfPotentialReads() + "\t" + getDetectionLimitDistance() + "\t" + format.format((double)gcContent()/getLength()) /*+ "\t" + format.format(getReadsMultiplier(((ABIRead)reads.first()).getLength()))*/;
	//if (nearestGene!=null)
	//  str += "\t" + nearestGene.toString();
	return str;
    }
    
    public boolean equals(Object o) {
	if (!(o instanceof ChromosomeWindow))
	    System.err.println("ERROR comparing peaks!");
	ChromosomeWindow er = (ChromosomeWindow)o;
	return er.numberOfReads()==numberOfReads() && er.getStart()==getStart() && er.getEnd()==getEnd() && er.getChromosome()==getChromosome();
    }

    public int hashCode() { return getStart()*7 + numberOfReads()*31 + getChromosome()*37; }

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof ChromosomeWindow))
	    System.err.println("ERROR comparing peaks!");
	ChromosomeWindow er = (ChromosomeWindow)o;
	if (chr<er.chr)
	    return -1;
	else if (chr>er.chr)
	    return 1;
	else {
	    if (getStart()<er.getStart())
		return -1;
	    else if (getStart()>er.getStart())
		return 1;
	    else
		return 0;
	}
    }
}    
