//Use this to represent a bin of a fixed width when scanning the
//genome for enriched regions in terms of the number of reads found
//(compared to input). Once the enriched windows are found we can
//merge them and pull out the regions with contiguous overlapping
//fragments.
//
//v3 is to be used with PeakFinder3. It uses a fixed window
//size. After we have added all the reads, we find the center, using
//the center of mass method and remove the ones which are outside the
//region. This implies that we disregard from peaks which 


import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.text.*;

public class ChromosomeWindow3 extends ChromosomeWindow implements Comparable {

    protected TreeSet reads = new TreeSet(), largestCluster = null;
    protected int nInputReads = 0, chr, potReads, start, end, windowSize;
    protected int constRegions[] = null;
    protected double pValue, inputPValue;
    protected String sequence;

    protected DecimalFormat format = new DecimalFormat("#######.###");

    //////////////////////////////////////////////////////////////////

    public ChromosomeWindow3(ABIRead r, int c) { 
	reads.add(r); 
	chr = c; 
    }

    public ChromosomeWindow3(ChromosomeWindow fWin, ChromosomeWindow rWin) {
	//Use this when combining forward and reverse windows
	chr = fWin.chr;
	start = fWin.start;
	end = fWin.end;
	Iterator it = fWin.reads.iterator();
	while (it.hasNext())
	    reads.add((ABIRead)it.next());
	it = rWin.reads.iterator();
	while (it.hasNext())
	    reads.add((ABIRead)it.next());
	potReads = fWin.potReads + rWin.potReads;
    }
    
    //////////////////////////////////////////////////////////////////

    public void add(ABIRead r) { 
	reads.add(r); 
	if (r.chr!=chr)
	    System.err.println("Adding read with different chr:\t" + r.toString());
	else if(r.pos<getStart() || r.pos>getEnd())
	    System.err.println("Adding read outside window!");
    }

    public void add(ABIRead r[]) { for (int i=0; i<r.length; i++) { add(r[i]); } }

    public void trimWindow(int windowSize) {
	int centerOfMass = getCenterOfMass(); //Find the center of mass
	start = centerOfMass - windowSize/2;
	end = centerOfMass + windowSize/2;
	//Remove all reads which are outside the window
	TreeSet removeSet = new TreeSet();
	Iterator it = reads.iterator();
	while (it.hasNext()) {
	    ABIRead r = (ABIRead)it.hasNext();
	    if (r.getPosition()<start || r.getPosition()>end) 
		removeSet.add(r);
	}
	it = reads.iterator();
	while (it.hasNext()) 
	    reads.remove((ABIRead)it.next());
	int centerOfMass2 = getCenterOfMass(); //calculate the new center of mass
	if (Math.abs(centerOfMass2 - centerOfMass)>5)
	    System.err.println("Warning, center of mass moved " + (centerOfMass2 - centerOfMass));
    }

    //////////////////////////////////////////////////////////////////


    //////////////////////////////////////////////////////////////////


    //////////////////////////////////////////////////////////////////

    public String toString() {
	return "chr" + chr + "\t" + getStart() + "\t" + getEnd() + "\t" + numberOfReads() + "\t" + numberOfInputReads() + "\t" + format.format(getPeakQualityScore());
    }
    
    public boolean equals(Object o) {
	if (!(o instanceof ChromosomeWindow3))
	    System.err.println("ERROR comparing peaks!");
	ChromosomeWindow3 er = (ChromosomeWindow3)o;
	return er.numberOfReads()==numberOfReads() && er.getStart()==getStart() && er.getEnd()==getEnd() && er.getChromosome()==getChromosome();
    }

    public int hashCode() { return getStart()*7 + numberOfReads()*31 + getChromosome()*37; }

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof ChromosomeWindow3))
	    System.err.println("ERROR comparing peaks!");
	ChromosomeWindow3 er = (ChromosomeWindow3)o;
	if (chr<er.chr)
	    return -1;
	else if (chr>er.chr)
	    return 1;
	else {
	    if (getStart()<er.getStart())
		return -1;
	    else if (getStart()>er.getStart())
		return 1;
	    else
		return 0;
	}
    }
}    
