//Use this method for converting Chip-Seq data files to a .bed format
//suitable for use with the peak-caller sissrs. We want to include the
//possibility of extending the reads (default) as well as filtering
//for PCR duplicates (default). Also has the option of running sissrs
//right away
//
//Example:
//
//bsub -q shared_2h java Convert2Bed -n /groups/neuroduo/martin/Chip-Seq/MmData/CBP/Abcam/KCl/B1/CBP_Abcam_KCl_B1 -i /groups/neuroduo/martin/Chip-Seq/MmData/Input/Input/KCl/B1/Input_Input_KCl_B1 -c 19 -o --sissrs


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class Convert2Bed {

    public static DecimalFormat format = new DecimalFormat("###.######");

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    CmdLineParser.Option allowDuplicatesFlag = argParser.addBooleanOption('a', "allowDuplicates");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option sissrsFlag = argParser.addBooleanOption("sissrs");
	    //Minimum quality score required for a read to be included
	    CmdLineParser.Option qualityScoreFlag = argParser.addDoubleOption('q', "qualityScore");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('f', "fdr");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    int nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean sissrs = argParser.getOptionValue(sissrsFlag)!=null;
	    tmp = argParser.getOptionValue(qualityScoreFlag);
	    double qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .001;
	    //////////////////////////////////////////////////////////////////
	    for (int i=1; i<=21; i++) {
		if (chr==-1 || i==chr) { 
		    //Save as .bed format
		    ABIRead[] reads = PeakFinder2.parseFiles(i, fileName, nMismatches, true, -1, fragmentLength, true, 0, qualityScore, null);
		    String fileNameBed = fileName + "_" + i + ".bed";
		    BufferedWriter bw = new BufferedWriter(new FileWriter(fileNameBed));
		    for (int j=0; j<reads.length; j++) {
			if (reads[j].getStrand())
			    bw.write("chr" + i + "\t" + reads[j].getPosition() + "\t" + (reads[j].getPosition() + reads[j].getLength()) + "\tU0\t1\t+\n");
			else
			    bw.write("chr" + i + "\t" + (reads[j].getPosition()) + "\t" + (reads[j].getPosition() + reads[j].getLength()) + "\tU0\t1\t-\n");
		    }
		    bw.close();
		    ABIRead[] inputReads = PeakFinder2.parseFiles(i, inputFileName, nMismatches, true, -1, fragmentLength, true, 0, qualityScore, null);
		    String inputFileNameBed = inputFileName + "_" + i + ".bed";
		    bw = new BufferedWriter(new FileWriter(inputFileNameBed));
		    for (int j=0; j<inputReads.length; j++) {
			if (inputReads[j].getStrand())
			    bw.write("chr" + i + "\t" + inputReads[j].getPosition() + "\t" + (inputReads[j].getPosition() + inputReads[j].getLength()) + "\tU0\t1\t+\n");
			else
			    bw.write("chr" + i + "\t" + inputReads[j].getPosition() + "\t" + (inputReads[j].getPosition() + inputReads[j].getLength()) + "\tU0\t1\t-\n");
		    }
		    bw.close();
		    ////////////////////////////////////////////
		    if (sissrs) { //Run sissrs
			String outputFileName = fileName + "_" + i + ".sissrs";
			String sissrsCmd = "/home/mh168/Softwares/sissrs/sissrs.pl -i " + fileNameBed + " -o " + outputFileName + " -s " + RefSeqGenes.TOTAL_CHROMOSOME_LENGTH + " -p " + format.format(fdr) + " -F " + fragmentLength + " -b " + inputFileNameBed + " -r ";
			System.out.println(sissrsCmd);
			Runtime runtime = Runtime.getRuntime();
			Process sissrsProcess = runtime.exec(sissrsCmd);
			sissrsProcess.waitFor();
		    }
		    //Remove the .bed files used for input
		    (new File(fileNameBed)).delete();
		    (new File(inputFileNameBed)).delete();
		}
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()


}
