//Use this class to convert the raw read data files from color-space
//to sequence space. Also, extract the uniquely mapping reads and the
//unmappable reads (candidates for splice junctions) to separate files
//(one per chromosome for the mappable ones). We later want to use
//these files to detect splice-junctions in non-coding regions using
//tophat.
//
//Example:
//
//bsub -q shared_2h java -Xmx512m ConvertRawReadsFile -n /groups/neuroduo/jesse/SOLiD/Solid_20081007_1_KCl_0hr/hr0/Solid0034_20081007_1_KCl_0hr_RNA_F3.csfasta.ma.35.3 -s /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/ReadSequences/hr0c -o
//
//bsub -q shared_2h java -Xmx512m ConvertRawReadsFile -n /groups/neuroduo/jesse/SOLiD/Solid_20081219_2_KCl_MAG/h0_mag/h0_mag.csfasta.ma.35.3  -s /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/ReadSequences/hr0m -o


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class ConvertRawReadsFile {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "test";
	    ///////////////////////////////////////////////////////
	    int nChr = FindTranscripts.numberOfChromosomes(species);
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    BufferedWriter noMatchesBw = new BufferedWriter(new FileWriter(saveFileName + "_noMatches.fa")), spliceCodingForwardBw = new BufferedWriter(new FileWriter(saveFileName + "_spliceCoding_f.fa")), spliceCodingReverseBw = new BufferedWriter(new FileWriter(saveFileName + "_spliceCoding_r.fa"));
	    BufferedWriter bwsForward[] = new BufferedWriter[nChr], bwsReverse[] = new BufferedWriter[nChr];
	    for (int i=0; i<bwsReverse.length; i++) {
		bwsForward[i] = new BufferedWriter(new FileWriter(saveFileName + "_unique_f_" + (i+1) + ".fa"));
		bwsReverse[i] = new BufferedWriter(new FileWriter(saveFileName + "_unique_r_" + (i+1) + ".fa"));
	    }
	    int nMultiple = 0, nSpliceForward = 0, nSpliceReverse = 0, nNo = 0;
	    int nUniqueForward[] = new int[nChr], nUniqueReverse[] = new int[nChr];
	    String line = br.readLine();
	    while (line!=null) {
		if (line.startsWith(">")) {
		    if (line.contains("F3,")) {
			String strs[] = line.split(",");
			if (strs.length==2) { //unique match
			    int chr = Integer.parseInt(strs[1].substring(0, strs[1].indexOf("_")));
			    if (chr<=nChr) {
				if (strs[1].contains("-")) {
				    bwsReverse[chr - 1].write(line + "\n" + convertColorCode(br.readLine()) + "\n");
				    nUniqueReverse[chr - 1]++;
				}
				else {
				    bwsForward[chr - 1].write(line + "\n" + convertColorCode(br.readLine()) + "\n");
				    nUniqueForward[chr - 1]++;
				}
			    }
			    else { //Mapped to known splice junction
				if (strs[1].contains("-")) {
				    spliceCodingReverseBw.write(line + "\n" + convertColorCode(br.readLine()) + "\n");
				    nSpliceReverse++;
				}
				else {
				    spliceCodingForwardBw.write(line + "\n" + convertColorCode(br.readLine()) + "\n");
				    nSpliceForward++;
				}
			    }
			}
			else
			    nMultiple++;
		    }
		    else { //no matches
			noMatchesBw.write(line + "\n" + convertColorCode(br.readLine()) + "\n");
			nNo++;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    noMatchesBw.close();
	    spliceCodingForwardBw.close();
	    spliceCodingReverseBw.close();
	    for (int i=0; i<bwsForward.length; i++) {
		bwsForward[i].close();
		bwsReverse[i].close();
		System.out.println("Found " + (nUniqueForward[i] + nUniqueReverse[i]) + " reads matching to " + (i+1));
	    }
	    System.out.println("Found " + nNo + " reads with no matches.");
	    System.out.println("Found " + nMultiple + " reads with multiple matches.");
	    System.out.println("Found " + nSpliceForward + " reads matching to known splice junctions on the forward strand.");
	    System.out.println("Found " + nSpliceReverse + " reads matching to known splice junctions on the reverse strand.");
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    public static String convertColorCode(String c) {
	//Convert from color code to sequence
	String s = "";
	if (!c.startsWith("T"))
	    System.err.println("String to be converted does not start with T!");
	else {
	    char prevChar = 'T';
	    for (int i=1; i<c.length(); i++) {
		if (prevChar=='A' && c.charAt(i)=='0')
		    s += "A";
		else if (prevChar=='A' && c.charAt(i)=='1')
		    s += "C";
		else if (prevChar=='A' && c.charAt(i)=='2')
		    s += "G";
		else if (prevChar=='A' && c.charAt(i)=='3')
		    s += "T";
		else if (prevChar=='C' && c.charAt(i)=='0')
		    s += "C";
		else if (prevChar=='C' && c.charAt(i)=='1')
		    s += "A";
		else if (prevChar=='C' && c.charAt(i)=='2')
		    s += "T";
		else if (prevChar=='C' && c.charAt(i)=='3')
		    s += "G";
		else if (prevChar=='G' && c.charAt(i)=='0')
		    s += "G";
		else if (prevChar=='G' && c.charAt(i)=='1')
		    s += "T";
		else if (prevChar=='G' && c.charAt(i)=='2')
		    s += "A";
		else if (prevChar=='G' && c.charAt(i)=='3')
		    s += "C";
		else if (prevChar=='T' && c.charAt(i)=='0')
		    s += "T";
		else if (prevChar=='T' && c.charAt(i)=='1')
		    s += "G";
		else if (prevChar=='T' && c.charAt(i)=='2')
		    s += "C";
		else if (prevChar=='T' && c.charAt(i)=='3')
		    s += "A";
		prevChar = s.charAt(s.length() - 1);
	    }
	}
	return s;
    }

}