//Use this method to assign the reads to different categories based on
//features of the genome. At the moment, we consider:
//
//Coding regions
//Enhancers
//Divergent transcripts
//ncRNAs
//
//Example:
//
//java -Xmx1024m CountReadsCategories -e ../MmDataEnhancers/enhancers_b2_null.txt -n ../mRNA/hr0c.tsv

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class CountReadsCategories {

    public static final String CODING_SENSE_CATEGORY_STR = "coding_sense";
    public static final String CODING_ANTISENSE_CATEGORY_STR = "coding_antisense";
    public static final String CODING_SENSE_3P_PROXIMAL_CATEGORY_STR = "coding_sense_3'";
    public static final String CODING_ANTISENSE_3P_PROXIMAL_CATEGORY_STR = "coding_antisense_3'";
    public static final String CODING_SENSE_5P_PROXIMAL_CATEGORY_STR = "coding_sense_5'";
    public static final String CODING_ANTISENSE_5P_PROXIMAL_CATEGORY_STR = "coding_antisense_5'";

    public static final String ENHANCER_CATEGORY_STR = "enhancer";
    public static final String LINC_RNA_CATEGORY_STR = "lincRNA";
    public static final String MICRO_RNA_CATEGORY_STR = "miRNA";

    /////////////////////////////////

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option nonGenicFlag = argParser.addBooleanOption("nonGenic");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option lncFileNameFlag = argParser.addStringOption("lincFileName");
	    CmdLineParser.Option miFileNameFlag = argParser.addStringOption("miFileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    //Location of the enhancer list
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption('e', "enhancerFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option flattenFlag = argParser.addBooleanOption("flatten");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('f', "fragmentLength");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean nonGenic = argParser.getOptionValue(nonGenicFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(lncFileNameFlag);
	    String lincFileName = tmp!=null ? (String)tmp : "../mRNA/lncRNA/landerRegions.bed";
	    tmp = argParser.getOptionValue(miFileNameFlag);
	    String miFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/") + "Genomes/Mm/v9/miRNA.gff";
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (fileName.contains(".") ? fileName.substring(0, fileName.lastIndexOf(".")) : fileName);
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/enhancers_null_KCl.txt";
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    boolean flatten = argParser.getOptionValue(flattenFlag)!=null;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1500;
	    //////////////////////////////////////////////////////////////////
	    int nChr = FindTranscripts.numberOfChromosomes(species);
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    int geneIDsKeep[] = null;
	    //Get the names of the factors that we want to study
	    Vector<String> fileNames = new Vector<String>();
	    if (fileName.contains("names")) 
		fileNames = FindOverlappingPeaks.getFileNames(fileName, null);
	    else
		fileNames.addElement(fileName);
	    //Get the list of enhancers
	    Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	    //Get the list of lincRNAs
	    Vector<LncRNA> lincRNAs = lincFileName!=null ? parseLincRNA(lincFileName) : null;
	    //Get list of miRNAs
	    Vector<MiRNA> miRNAs = miFileName!=null ? parseMiRNA(miFileName) : null;
	    //Keep track of the categories that we want to consider
	    Vector<String> categoryStrings = new Vector<String>();
	    categoryStrings.addElement(CODING_SENSE_CATEGORY_STR);
	    categoryStrings.addElement(CODING_SENSE_5P_PROXIMAL_CATEGORY_STR);
	    categoryStrings.addElement(CODING_SENSE_3P_PROXIMAL_CATEGORY_STR);
	    categoryStrings.addElement(CODING_ANTISENSE_CATEGORY_STR);
	    categoryStrings.addElement(CODING_ANTISENSE_5P_PROXIMAL_CATEGORY_STR);
	    categoryStrings.addElement(CODING_ANTISENSE_3P_PROXIMAL_CATEGORY_STR);
	    if (enhancerFileName!=null)
		categoryStrings.addElement(ENHANCER_CATEGORY_STR);
	    if (lincFileName!=null)
		categoryStrings.addElement(LINC_RNA_CATEGORY_STR);
	    if (miFileName!=null)
		categoryStrings.addElement(MICRO_RNA_CATEGORY_STR);
	    TreeMap<String, TreeMap<String, int[]>> category2ReadsMap = new TreeMap<String, TreeMap<String, int[]>>();
	    TreeMap<String, Integer> nTotalReads = new TreeMap<String, Integer>();
	    for (String fName : fileNames) { nTotalReads.put(fName, 0); }
	    for (String categoryString : categoryStrings) {
		category2ReadsMap.put(categoryString, new TreeMap<String, int[]>());
		for (String fName : fileNames)
		    category2ReadsMap.get(categoryString).put(fName, new int[nChr]);
	    }
	    for (int i=1; i<=nChr; i++) {
		if (i==chr || chr==-1) {
		    for (String fName : fileNames) {
			geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
			ABIRead readsForward[] = ABIReads.loadMRNAReadsChr(fName, true, i, readLength, flatten), readsReverse[] = ABIReads.loadMRNAReadsChr(fName, false, i, readLength, flatten);
			nTotalReads.put(fName, nTotalReads.get(fName) + readsForward.length + readsReverse.length);
			GeneEnhancerGraph graph = new GeneEnhancerGraph(enhancers, geneInfo[i-1], nonGenic);
			if (miFileName!=null)
			    graph.addMiRNAInfo(miRNAs);
			if (lincFileName!=null)
			    graph.addLincRNAInfo(lincRNAs);
			for (String categoryString : categoryStrings) {
			    int nReadsCategory = 0;
			    if (categoryString.equals(CODING_SENSE_CATEGORY_STR)) 
				graph.addGeneReads(fName, readsForward, readsReverse);
			    else if (categoryString.equals(CODING_ANTISENSE_CATEGORY_STR)) 
				graph.addGeneReads(fName, readsForward, readsReverse, false);
			    else if (categoryString.equals(CODING_SENSE_3P_PROXIMAL_CATEGORY_STR)) 
				graph.addGeneReads(fName, readsForward, readsReverse, true, windowSize, 0);
			    else if (categoryString.equals(CODING_SENSE_5P_PROXIMAL_CATEGORY_STR)) 
				graph.addGeneReads(fName, readsForward, readsReverse, false, 0, windowSize);
			    else if (categoryString.equals(CODING_ANTISENSE_3P_PROXIMAL_CATEGORY_STR)) 
				graph.addGeneReads(fName, readsForward, readsReverse, true, windowSize, 0, false);
			    else if (categoryString.equals(CODING_ANTISENSE_5P_PROXIMAL_CATEGORY_STR)) 
				graph.addGeneReads(fName, readsForward, readsReverse, false, 0, windowSize, false);
			    else if (categoryString.equals(ENHANCER_CATEGORY_STR))
				graph.addEnhancerReads(fName, readsForward, readsReverse, windowSize);
			    else if (categoryString.equals(MICRO_RNA_CATEGORY_STR))
				graph.addMiRNAReads(fName, readsForward, readsReverse);
			    else if (categoryString.equals(LINC_RNA_CATEGORY_STR))
				graph.addLincRNAReads(fName, readsForward, readsReverse);
			    ////////////////////////////////////////////
			    if (categoryString.contains("coding")) {
				for (RefSeqGene g : graph.getGenes())
				    nReadsCategory += g.getFactorReads(fName).length;
			    }
			    else if (categoryString.contains("enhancer")) {
				for (Enhancer en : graph.getEnhancers())
				    nReadsCategory += en.getFactorReads(fName).length;
			    }
			    else if (categoryString.contains("lincRNA")) {
				for (LncRNA l : graph.getLincRNAs())
				    nReadsCategory += l.getFactorReads(fName).length;
			    }
			    else if (categoryString.contains("miRNA")) {
				for (MiRNA l : graph.getMiRNAs())
				    nReadsCategory += l.getFactorReads(fName).length;
			    }
			    int count[] = category2ReadsMap.get(categoryString).get(fName);
			    count[i-1] += nReadsCategory;
			    category2ReadsMap.get(categoryString).put(fName, count);
			}
		    }
		}
	    }
	    for (String fName : fileNames) {
		System.out.println("Found a total of " + nTotalReads.get(fName) + " reads for " + fName);
		for (String categoryString : categoryStrings) {
		    //System.out.println("Found " + Arrays.toString(category2ReadsMap.get(categoryString).get(fName)) + " reads factor " + fName + " for " + categoryString);
		    double prop[] = new double[nChr];
		    int nReads = 0;
		    for (int i=0; i<nChr; i++) {
			prop[i] = (double)category2ReadsMap.get(categoryString).get(fName)[i]/nTotalReads.get(fName);
			nReads += category2ReadsMap.get(categoryString).get(fName)[i];
		    }
		    //System.out.println("As a proportion of total " + Arrays.toString(prop) + " reads factor " + fName + " for " + categoryString);
		    System.out.println("Found " + nReads + " (" + format.format((double)nReads/nTotalReads.get(fName)) + ") reads from factor " + fName + " for " + categoryString);
		}
	    }
	    ///////////////////////////////////////////////////////
	    //Save stuff
	    if (saveFileName!=null) {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		System.out.println("Saving to " + saveFileName);
		if (chr==1 || chr==-1) {
		    bw.write("#Experiment");
		    for (String categoryString : categoryStrings)
			bw.write("\t" + categoryString);
		    bw.write("\n");
		}
		for (String fName : fileNames) {
		    bw.write(fName.substring(fName.lastIndexOf("/") + 1) + (chr==-1 ? "" : "_" + chr));

		}
		bw.close();
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////


    public static Vector<LncRNA> parseLincRNA(String fileName) {
	//Assume that it has .bed format
	Vector<LncRNA> lincRNAs = new Vector<LncRNA>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    Pattern lincRNAPattern = Pattern.compile("chr([\\d|X|Y]+):(\\d+)-(\\d+)");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = lincRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = m.group(1).equals("X") ? 20 : (m.group(1).equals("Y") ? 21 : Integer.parseInt(m.group(1)));
		    lincRNAs.addElement(new LncRNA(chr, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3))));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opening file " + fileName);
	}
	System.out.println("Parsed " + lincRNAs.size() + " lincRNAs.");
	return lincRNAs;
    }

    public static Vector<MiRNA> parseMiRNA(String fileName) {
	//Assume .gff format
	Vector<MiRNA> miRNAs = new Vector<MiRNA>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    Pattern miRNAPattern = Pattern.compile("([\\d|X|Y]+)\\s+\\.\\s+miRNA\\s+(\\d+)\\s+(\\d+)\\s+\\.\\s+([+|-])\\s+\\.\\s+ACC=\"MI\\d+\";\\s+ID=\"([\\w|-]+)\";");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = miRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = m.group(1).equals("X") ? 20 : (m.group(1).equals("Y") ? 21 : Integer.parseInt(m.group(1)));
		    boolean strand = m.group(4).equals("+");
		    miRNAs.addElement(new MiRNA(chr, Integer.parseInt(m.group(strand ? 2 : 3)), Integer.parseInt(m.group(strand ? 3 : 2)), strand, m.group(5)));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opening file " + fileName);
	}
	System.out.println("Parsed " + miRNAs.size() + " miRNAs.");
	return miRNAs;
    }

}

