import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import java.nio.channels.*;

//Example:
//
//java -Xmx1024m CreateWigFiles -n /groups/neuroduo/martin/ChIP-Seq/MmData/MeCP2/MeCP2/30WT_Joe/B1/MeCP2_MeCP2_30WT_Joe_B1 -u -o -l 120 -e
//
//For RNA-Seq files
//
//java -Xmx1024m CreateWigFiles -n /groups/neuroduo/martin/ChIP-Seq/KuerstenData/RNA/HeLa/pA/WT/hela_pA_WT -o -l 50 --hg19

public class CreateWigFiles{

    public static final int READS_NORMALIZATION_FACTOR = 1000000;

    public static String GENOME_LOC;
    public static final String TRACK_INFO_FILE_NAME = "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/track_info.txt";
    public static final String TRACK_INFO_MECP2_FILE_NAME = "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/MeCP2/track_info.txt";
    public static final String TRACK_INFO_RNA_FILE_NAME = "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/RNA/track_info.txt";

    protected static int nMismatches, fragmentLength, windowSize, downSample;
    protected static double normalizationFactor, qualityScore;
    protected static boolean verbose, unique, orchestra, extend, skipBed, split, noHeader, strand, hg19;
    protected static String fileName, inputFileName, species;
    
    public static DecimalFormat format = new DecimalFormat("###.######");

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    //Use this flag for determining the windowsize when
	    //scanning the genome for various properties
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option singletonFlag = argParser.addBooleanOption('s', "singletons");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption("strand");
	    CmdLineParser.Option hg19Flag = argParser.addBooleanOption("hg19");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //Use this flag if we want to extend the reads to length fragmentLength
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    CmdLineParser.Option noHeaderFlag = argParser.addBooleanOption("noHeader");
	    //Avoid writing the header, good for bigBedFiles
	    CmdLineParser.Option skipBedFlag = argParser.addBooleanOption("skipBed");
	    //Produce one .bed file per chr
	    CmdLineParser.Option splitFlag = argParser.addBooleanOption("split");
	    //Use this flag to normalize the count by the total number
	    //of reads and then mulitply by the number given as
	    //argument here.
	    CmdLineParser.Option normalizeFlag = argParser.addDoubleOption('N', "normalize");
	    //Use this flag when we want to downsample the number of
	    //tags. Use ful when we want to avoid generating a new set
	    //of randomized tags.
	    CmdLineParser.Option downSampleFlag = argParser.addIntegerOption('D', "downSampled");
	    //Minimum quality score required for a read to be included
	    CmdLineParser.Option qualityScoreFlag = argParser.addDoubleOption('q', "qualityScore");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(normalizeFlag);
	    normalizationFactor = tmp!=null ? ((Double)tmp).doubleValue() : READS_NORMALIZATION_FACTOR;
	    tmp = argParser.getOptionValue(downSampleFlag);
	    downSample = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    verbose = argParser.getOptionValue(verboseFlag)!=null;
	    unique = argParser.getOptionValue(uniqueFlag)!=null;
	    orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    strand = argParser.getOptionValue(strandFlag)==null;
	    hg19 = argParser.getOptionValue(hg19Flag)!=null;
	    extend = argParser.getOptionValue(extendFlag)!=null;
	    skipBed = argParser.getOptionValue(skipBedFlag)!=null;
	    noHeader = argParser.getOptionValue(noHeaderFlag)!=null;
	    split = argParser.getOptionValue(splitFlag)!=null;
	    tmp = argParser.getOptionValue(qualityScoreFlag);
	    qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(speciesFlag);
	    species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    //////////////////////////////////////////////////////////////////
	    GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	    
	    if (chr==-1) { //Look at all chromosomes
		if (split) {
		    try {
			Runtime runtime = Runtime.getRuntime();
			for (int i=1; i<=21; i++) {
			    analyze(i);
			    //zip
			    System.out.println("gzip " + fileName + (extend ? "_E" + fragmentLength : "") + "_" + i + ".bed");
			    runtime.exec("gzip " + fileName + (extend ? "_E" + fragmentLength : "") + "_" + i + ".bed");
			    //change permissions to make visible from www
			    System.out.println("chmod 775 " + fileName + (extend ? "_E" + fragmentLength : "") + "_" + i + ".bed.gz");
			    runtime.exec("chmod 775 " + fileName + (extend ? "_E" + fragmentLength : "") + "_" + i + ".bed.gz");
			    //recursively go through all the directories and change permissions
			    String dirs[] = fileName.split("/"), dir = dirs[0];
			    for (int j=1; j<dirs.length-1; j++) {
				dir += "/" + dirs[j];
				System.out.println("chmod 771 " + dir);
				runtime.exec("chmod 771 " + dir);
			    }
			    //add sym-link to www-directory
			    System.out.println("ln -s " + fileName + (extend ? "_E" + fragmentLength : "") + "_" + i + ".bed.gz /www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/");
			    runtime.exec("ln -s " + fileName + (extend ? "_E" + fragmentLength : "") + "_" + i + ".bed.gz /www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/");
			}
		    }
		    catch (IOException e) { }
		}
		else {
		    analyze(skipBed, fileName.contains("RNA/"), species, hg19);
		}
	    }
	    else
		analyze(chr);
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    public static void analyze(int c) {
	boolean hg19 = false;
	ABIRead reads[] = PeakFinder2.parseFiles(c, fileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", c, fragmentLength, "txt"), true);
	if (inputFileName==null)
	    PeakFinder2.savePositionsReads(reads, fileName.endsWith(".bed") ? fileName.replace(".bed", "_E" + fragmentLength + c + ".bed") : fileName + "_" + (extend ? "E" : "") + fragmentLength + "_" + c + ".bed", c, !noHeader, normalizationFactor*PeakFinder2.CHROMOSOME_LENGTH[c-1]/(reads.length*PeakFinder2.TOTAL_CHROMOSOME_LENGTH));
	else {
	    ABIRead inputReads[] = PeakFinder2.parseFiles(c, inputFileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", c, fragmentLength, "txt"), true);
	    PeakFinder2.savePositionsReads(reads, inputReads, fileName + "_input_" + (extend ? "E" : "") + fragmentLength + "_" + c + ".bed", c, false, normalizationFactor*PeakFinder2.CHROMOSOME_LENGTH[c-1]/(reads.length*PeakFinder2.TOTAL_CHROMOSOME_LENGTH), normalizationFactor*PeakFinder2.CHROMOSOME_LENGTH[c-1]/(inputReads.length*PeakFinder2.TOTAL_CHROMOSOME_LENGTH), c!=1);
	}
	//Create a file with track-info
	String name = fileName.substring(fileName.lastIndexOf("/") + 1) + (inputFileName==null ? "" : "_input_" + c);
	String colorStr = getColorString(name);
	int priority = getPriority(name);
	String trackInfoStr = "track type=bigWig name=\"" + name + "\" description=\"" + name + "\" priority=" + priority + " visibility=full autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=http://neuroduo.hms.harvard.edu/bigWig/mh/" + name + "_E120.bw";
	addTrackInfo(trackInfoStr, TRACK_INFO_FILE_NAME);
	createWWWLink(fileName + (inputFileName==null ? "" : "_input") + (extend ? "_E" + fragmentLength : "") + "_" + c + ".bed", "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/", species, hg19);
	//TKParser tkParser = new TKParser(fileName, c, nMismatches, unique, fragmentLength);
    }
    
    public static void analyze(boolean skipBed, boolean rnaSeq, String species, boolean hg19) {
	//Do all chromosomes in one big file
	if (!skipBed) {
	    int nChr = species.equals("hs") ? 24 : 21;
	    long totChrLen = species.equals("hs") ? RefSeqGenes.TOTAL_CHROMOSOME_LENGTH_HS : RefSeqGenes.TOTAL_CHROMOSOME_LENGTH;
	    for (int i=1; i<=nChr; i++) {
		int chrLen = species.equals("hs") ? RefSeqGenes.CHROMOSOME_LENGTH_HS[i-1] : RefSeqGenes.CHROMOSOME_LENGTH[i-1];
		if (rnaSeq) {
		    ABIRead forwardReads[] = PeakFinder2.parseFiles(i, fileName, nMismatches, unique, -1, fragmentLength, extend, 1, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats_f", i, fragmentLength, "txt"));
		    PeakFinder2.savePositionsReads(forwardReads, fileName + "_" + (extend ? "E" : "") + fragmentLength + "_f.bed", i, species, false, normalizationFactor*chrLen/(forwardReads.length*totChrLen), i!=1);
		    ABIRead reverseReads[] = PeakFinder2.parseFiles(i, fileName, nMismatches, unique, -1, fragmentLength, extend, -1, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats_r", i, fragmentLength, "txt"));
		    PeakFinder2.savePositionsReads(reverseReads, fileName + "_" + (extend ? "E" : "") + fragmentLength + "_r.bed", i, species, false, normalizationFactor*chrLen/(reverseReads.length*totChrLen), i!=1, true);
		}
		else {
		    ABIRead reads[] = fileName.endsWith("bed") ? parseBedFiles(i, fileName, unique, fragmentLength, extend, fileName.replace(".bed", "_read_stats_" + i + "_E" + fragmentLength + ".txt")) : PeakFinder2.parseFiles(i, fileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", i, fragmentLength, "txt"), true);
		    if (inputFileName==null)
			PeakFinder2.savePositionsReads(reads, fileName.endsWith(".bed") ? fileName.replace(".bed", "_E" + fragmentLength + ".bed") : fileName + "_" + (extend ? "E" : "") + fragmentLength + ".bed", i, species, false, normalizationFactor*chrLen/(reads.length*totChrLen), i!=1);
		    else {
			ABIRead inputReads[] = PeakFinder2.parseFiles(i, inputFileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", i, fragmentLength, "txt"));
			PeakFinder2.savePositionsReads(reads, inputReads, fileName + "_input_" + (extend ? "E" : "") + fragmentLength + ".bed", i, false, normalizationFactor*chrLen/(reads.length*totChrLen), normalizationFactor*PeakFinder2.CHROMOSOME_LENGTH[i-1]/(inputReads.length*totChrLen), i!=1);
		    }
		}
		System.out.println("Created .bed-file for chr " + i);
		//TKParser tkParser = new TKParser(fileName, c, nMismatches, unique, fragmentLength);
	    }
	}
	//Create a file with track-info
	if (rnaSeq) {
	    String name = fileName.substring(fileName.lastIndexOf("/") + 1);
	    String colorStr = getColorString(name + "_f");
	    int priority = getPriority(name + "_f");
	    boolean mecp2 = name.contains("MeCP2");
	    String trackInfoStr = "track type=bigWig name=\"" + name + "_f\" description=\"" + name + "_f\" priority=" + priority + " visibility=full autoScale=off maxHeightPixels=10:25:70 viewLimits=0:20 color=" + colorStr + " smoothingWindow=off bigDataUrl=http://neuroduo.hms.harvard.edu/bigWig/mh/" + (mecp2 ? "MeCP2/" : "") + name + "_" + fragmentLength + "_f.bw";
	    addTrackInfo(trackInfoStr, mecp2 ? TRACK_INFO_MECP2_FILE_NAME : rnaSeq ? TRACK_INFO_RNA_FILE_NAME : TRACK_INFO_FILE_NAME);
	    createWWWLink(fileName + "_" + fragmentLength + "_f.bed", "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/RNA/", species, hg19);
	    //reverse strand
	    priority = getPriority(name + "_r") + 1;
	    trackInfoStr = "track type=bigWig name=\"" + name + "_r\" description=\"" + name + "_r\" priority=" + priority + " visibility=full autoScale=off maxHeightPixels=10:25:70 viewLimits=0:-20 color=" + colorStr + " smoothingWindow=off bigDataUrl=http://neuroduo.hms.harvard.edu/bigWig/mh/" + (mecp2 ? "MeCP2/" : "") + name + "_" + fragmentLength + "_r.bw";
	    addTrackInfo(trackInfoStr, mecp2 ? TRACK_INFO_MECP2_FILE_NAME : rnaSeq ? TRACK_INFO_RNA_FILE_NAME : TRACK_INFO_FILE_NAME);
	    createWWWLink(fileName + "_" + fragmentLength + "_r.bed", "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/RNA/", species, hg19);
	}
	else {
	    String name = fileName.substring(fileName.lastIndexOf("/") + 1) + (inputFileName==null ? "" : "_input");
	    String colorStr = getColorString(name);
	    int priority = getPriority(name);
	    boolean mecp2 = name.contains("MeCP2");
	    String fName = (name.endsWith(".bed") ? name.replace(".bed", "_E" + fragmentLength + ".bed") : name + "_E" + fragmentLength + ".bed");
	    String trackInfoStr = "track type=bigWig name=\"" + name.replace(".bed", "") + "\" description=\"" + name.replace(".bed", "") + "\" priority=" + priority + " visibility=full autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=http://neuroduo.hms.harvard.edu/bigWig/mh/" + (mecp2 ? "MeCP2/" : "") + fName.replace(".bed", ".bw");
	    addTrackInfo(trackInfoStr, mecp2 ? TRACK_INFO_MECP2_FILE_NAME : TRACK_INFO_FILE_NAME);
	    createWWWLink(fileName.substring(0, fileName.lastIndexOf("/") + 1) + fName, "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/", species, hg19);
	}
    }

    public static int getPriority(String name) {
	boolean kcl = !name.contains("un");
	if (name.contains("H3") || name.contains("H4"))
	    return 100 + (int)name.charAt(2) + (int)name.charAt(3) + (kcl ? 1 : 0);
	else if (name.contains("transcripts"))
	    return 1400 + (int)name.charAt(0) + (int)name.charAt(1) + (name.endsWith("r") ? 1 : 0);
	else if (name.contains("regions"))
	    return 1800 + (int)name.charAt(0) + (int)name.charAt(1) + (kcl ? 1 : 0);
	else if (name.contains("hela"))
	    return 1300 + (int)name.charAt(name.length()-1) + (int)name.charAt(name.length()-2);
	else if (name.contains("hrb"))
	    return 1350 + (int)name.charAt(name.length()-1) + (int)name.charAt(name.length()-2);
	else if (name.contains("MeCP2"))
	    return 850 + (int)name.charAt(8) + (int)name.charAt(9);
	else
	    return 600 + (int)name.charAt(0) + (int)name.charAt(1) + (kcl ? 1 : 0);
    }

    public static String getColorString(String name) {
	String str = "";
	if (name.contains("H3K4Me1"))
	    str = "0,255,127";
	else if (name.contains("Input_Cherry"))
	    str = "38,113,25";
	else if (name.contains("CREB_upstate"))
	    str = "38,113,25";
	else if (name.contains("pCREB133_Millipore"))
	    str = "19,166,50";
	else if (name.contains("CRTC1_bethyl"))
	    str = "3,86,155";
	else if (name.contains("CBP"))
	    str = "148,0,211";
	else if (name.contains("pS86"))
	    str = "255,0,102";
	else if (name.contains("T308"))
	    str = "51,51,0";
	else if (name.contains("T308"))
	    str = "0,51,102";
	else if (name.contains("8WG16"))
	    str = "0,0,0";
	else if (name.contains("Npas4"))
	    str = "0,0,255";
	else if (name.contains("H3K27Me3"))
	    str = "136,69,19";
	else if (name.contains("H3K4Me3"))
	    str = "204,153,204";
	else if (name.contains("SRF"))
	    str = "51,102,0";
	else if (name.contains("CREB"))
	    str = "255,153,51";
	else if (name.contains("H3K4Ac"))
	    str = "153,204,255";
	else if (name.contains("p300"))
	    str = "102,255,255";
	else if (name.contains("hela"))
	    str = "102,255,51";
	else if (name.contains("hrb"))
	    str = "102,0,51";
	else if (name.contains("MEF2D_6066"))
	    str = "255,69,0";
	else if (name.contains("MEF2D_2373"))
	    str = "205,0,0";
	else
	    str = "255,0,0";
	return str;
    }

    public static void createWWWLink(String fileName, String wwwDirName, String species, boolean hg19) {
	try {
	    //create links etc
	    Runtime runtime = Runtime.getRuntime();
	    //create the bigWig file
	    String bwFileName = fileName.replace("bed", "bw");
	    String chrSizeStr = species.equals("mm") ? "/groups/neuroduo/genomes/mm_NCBI_37/chrom.sizes" : "/groups/neuroduo/genomes/" + (hg19 ? "human_37_1/hg19" : "hs_NCBI_36_2_jmg/hg18") + ".chrom.sizes";
	    String cmd = "./wigToBigWig " + fileName + " " + chrSizeStr + " " + bwFileName;
	    System.out.println(cmd);
	    Process process = runtime.exec(cmd);
	    try { process.waitFor(); }
	    catch ( InterruptedException e ) {
		System.err.println("Error waiting for wigToBigWig process " + e.toString());
	    }
	    //change permissions to make visible from www
	    cmd = "chmod 775 " + bwFileName;
	    System.out.println(cmd);
	    runtime.exec(cmd);
	    //recursively go through all the directories and change permissions
	    String dirs[] = bwFileName.split("/"), dir = dirs[0];
	    for (int i=1; i<dirs.length-1; i++) {
		dir += "/" + dirs[i];
		//System.out.println("chmod 771 " + dir);
		runtime.exec("chmod 771 " + dir);
	    }
	    //add sym-link to www-directory
	    cmd = "ln -s " + bwFileName + " " + wwwDirName + (bwFileName.contains("MeCP2") ? "MeCP2/" : "");
	    System.out.println(cmd);
	    runtime.exec(cmd);
	}
	catch (IOException e) {
	    System.err.println("Error saving to track info file : " + e.toString());
	}
    }

    public static void addTrackInfo(String trackInfoStr, String fileName) {
	try {
	    //Find the file containing track info. First we need to
	    //create a lock on it so that no other process can mess
	    //with it
	    File file = new File(fileName);
	    FileChannel channel = new RandomAccessFile(file, "rw").getChannel();
	    FileLock lock = channel.lock();
	    BufferedReader br = new BufferedReader(new FileReader(file));
	    //FileChannel channel = (new FileReader(file)).getChannel();
	    //FileLock lock = channel.lock();
	    //Check if it contains the track already
	    TreeSet<String> lines = new TreeSet<String>();
	    String line = br.readLine(), fileInfo = "";
	    boolean lineFound = false;
	    while (line!=null) {
		if (line.equals(trackInfoStr)) 
		    lineFound = true;
		lines.add(line);
		fileInfo += line + "\n";
		line = br.readLine();
	    }
	    br.close();
	    if (!lineFound) {
		fileInfo += trackInfoStr + "\n";
		lines.add(trackInfoStr);
		//Save all the info again
		BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
		for (String l : lines)
		    bw.write(l + "\n");
		bw.close();
	    }
	    //System.out.println("Saved track info:\n " + fileInfo);
	    lock.release();
	}
	catch (IOException e) {
	    System.err.println("Error saving to track info file : " + e.toString());
	}
    }

    //////////////////////////////////////////////////////////////////

    public static int getTotalNumberOfReads(String fileName, int nChr, int fragmentLength) {
	int n = 0;
	for (int i=0; i<nChr; i++)
	    n += getNumberOfReads(fileName, i+1, fragmentLength);
	return n;
    }

    public static int getNumberOfReads(String fileName, int chr, int fragmentLength) {
	//Find the total number of reads by parsing the read_stats-file
	int n = 0;
	try {
	    BufferedReader statsReader = new BufferedReader(new FileReader(fileName.endsWith(".bed") ? fileName.replace(".bed", WindowFinder.READ_STATS_SUFFIX + chr + "_E" + fragmentLength + ".txt") : fileName + WindowFinder.READ_STATS_SUFFIX + chr + "_E" + fragmentLength + ".txt"));
	    String line = statsReader.readLine();
	    Pattern statsPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = statsPattern.matcher(line);
		if (m.find()) { n = Integer.parseInt(m.group(3)); }
		line = statsReader.readLine();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error parsing stats-file " + e.toString());
	}
	return n;
    }


    public static ABIRead[] parseBedFiles(int chr, String fileName, boolean unique, int fragmentLength) { return fileName==null ? new ABIRead[0] : parseBedFiles(chr, fileName, unique, fragmentLength, true, fileName.replace(".bed", "_read_stats_" + chr + "_E" + fragmentLength + ".txt")); }

    public static ABIRead[] parseBedFiles(int chr, String fileName, boolean unique, int fragmentLength, boolean extend, String statsFileName) {
	//Use this for parsing .bed files (typically converted from
	//.bam files using BEDTools.
	if (fileName==null)
	    return new ABIRead[0];
	ABIRead[] reads = new ABIRead[0], rs = ABIReads.loadNIHReads(fileName.replace(".bed", "_" + chr + ".bed"), chr, fragmentLength);
	Arrays.sort(rs);
	int nPCRDuplicates = 0;
	if (unique) { //remove PCR-duplicates
	    TreeSet<Integer> removeInds = new TreeSet<Integer>();
	    for (int i=1; i<rs.length; i++) {
		if (rs[i].equals(rs[i-1]))
		    removeInds.add(i-1);
	    }
	    nPCRDuplicates = removeInds.size();
	    reads = new ABIRead[rs.length - nPCRDuplicates];
	    int j = 0;
	    for (int i=0; i<rs.length; i++) {
		if (!removeInds.contains(i)) {
		    reads[j++] = rs[i];
		    //if (Math.abs(rs[i].getPosition() - 14851900)<fragmentLength)
		    //System.out.println("Found read " + rs[i].toString() + " from " + fileName);
		}
	    }
	}
	else
	    reads = rs;
	if (statsFileName!=null) {
	    try { //Save the stats file
		BufferedWriter bw = new BufferedWriter(new FileWriter(statsFileName));
		if (chr==1)
		    bw.write("#Tot reads\t-Quality\tUnique\t0 mismatch\t1 mismatch\t2 mismatch\t3 mismatch\n");
		bw.write(rs.length + "\t" + rs.length + "\t" + reads.length + "\t" + reads.length + "\t0\t0\t0\n");
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving to " + statsFileName);
	    }
	}
	return reads;
    }

}
