//Use this class for scanning sequences and extracting all enriched words
//
//From Matlab:
//
//wf = javaObject('EnrichedWordFinder', '../../Athar/Sequences/ap1_fos_npas4_cbp_peaks_distal_w_100', '../../Athar/Sequences/enhancers_control_w_100');
//words6 = javaMethod('findEnrichedWords', wf, 6, .0001);

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.text.DecimalFormat;
import org.apache.commons.math.distribution.*;
import org.apache.commons.math.stat.descriptive.SummaryStatistics;
import umontreal.iro.lecuyer.stat.Tally;
import org.apache.commons.math.stat.regression.SimpleRegression;
import umontreal.iro.lecuyer.stat.list.ListOfTallies;

public class EnrichedWordFinder {

    public static final String TMP_CLUSTALW_FILE_IN = "tmp_clustalw.fa";
    public static final String TMP_CLUSTALW_FILE_OUT = "tmp_clustalw.aln";
    public static final int N_MC = 100;

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();
    
    protected Vector<String> seqs = new Vector<String>(), ctrlSeqs = new Vector<String>();
    protected boolean orchestra = false;
    //Background frequencies
    protected double a, c, g, t, aCtrl, cCtrl, gCtrl, tCtrl;
    protected int totSeqLen = 0, totCtrlSeqLen = 0;
    protected TreeMap<String, Double> enrichedWords = new TreeMap<String, Double>();
    protected String seqFileStart, ctrlSeqFileStart;

    public EnrichedWordFinder(String fileStart, String bgFileStart, int subsample, boolean o) {
	orchestra = o;
	seqFileStart = fileStart;
	ctrlSeqFileStart = bgFileStart;
	//Open the two .fa files and read sequences
	double tmp[] = parseSequenceFile(fileStart, seqs);
	a = tmp[0];
	c = tmp[1];
	g = tmp[2];
	t = tmp[3];
	totSeqLen = (int)(a + c + g + t);
	tmp = parseSequenceFile(bgFileStart, ctrlSeqs);
	aCtrl = tmp[0];
	cCtrl = tmp[1];
	gCtrl = tmp[2];
	tCtrl = tmp[3];
	totCtrlSeqLen = (int)(aCtrl + cCtrl + gCtrl + tCtrl);
	a /= totSeqLen;
	c /= totSeqLen;
	g /= totSeqLen;
	t /= totSeqLen;
	aCtrl /= totCtrlSeqLen;
	cCtrl /= totCtrlSeqLen;
	gCtrl /= totCtrlSeqLen;
	tCtrl /= totCtrlSeqLen;
	System.out.println("Base frequencies are " + format.format(a) + ", " + format.format(c) + ", " + format.format(g) + ", " + format.format(t) + ")");
	System.out.println("Base frequencies for control are " + format.format(aCtrl) + ", " + format.format(cCtrl) + ", " + format.format(gCtrl) + ", " + format.format(tCtrl) + ")");
	//Make sure that there are the same number of sequences in both samples
	if (subsample!=0) {
	    int nKeep;
	    if (subsample<0 || subsample>seqs.size() || subsample>ctrlSeqs.size()) 
		nKeep = Math.min(seqs.size(), ctrlSeqs.size());
	    else 
		nKeep = subsample;
	    while (nKeep<seqs.size()) {
		char[] s = seqs.remove(random.nextInt(seqs.size())).toCharArray();
		totSeqLen -= s.length;
		for (int i=0; i<s.length; i++)
		    totSeqLen -= s[i]=='N' ? 1 : 0;
	    }
	    while (nKeep<ctrlSeqs.size()) {
		char[] s = ctrlSeqs.remove(random.nextInt(ctrlSeqs.size())).toCharArray();
		totCtrlSeqLen -= s.length;
		for (int i=0; i<s.length; i++)
		    totCtrlSeqLen -= s[i]=='N' ? 1 : 0;
	    }
	}
	System.out.println("Found " + seqs.size() + " seqs and " + ctrlSeqs.size() + " controls.");
    }

    public static double[] parseSequenceFile(String fileStart, Vector<String> seqs) { return parseSequenceFile(fileStart, seqs, 21); }

    public static double[] parseSequenceFile(String fileStart, Vector<String> seqs, int nChr) {
	double a = 0, c = 0, g = 0, t = 0;
	try {
	    for (int i=1; i<=nChr; i++) {
		BufferedReader br = new BufferedReader(new FileReader(fileStart + (fileStart.endsWith("fa") ? "" : "_" + i + ".fa")));
		String line = br.readLine();
		while (line!=null) {
		    if (!line.startsWith(">")) {
			seqs.addElement(line);
			char[] array = line.toUpperCase().toCharArray();
			for (int j=0; j<array.length; j++) {
			    if (array[j]=='A')
				a++;
			    else if (array[j]=='C')
				c++;
			    else if (array[j]=='G')
				g++;
			    else if (array[j]=='T')
				t++;
			}
		    }
		    line = br.readLine();
		}
		br.close();
		if (fileStart.endsWith("fa"))
		    break;
	    }
	} 
	catch (IOException e) {
	    System.err.println("Error reading seqs " + e.toString());
	}
	return new double[]{a,c,g,t};
    }

    //////////////////////////////////////////////////////////////////

    public Vector<Vector<String>> clusterWords(Vector<String> words, int k) {
	//Do a hierarchical clustering to find the pseudo-motifs
	Vector<Vector<String>> clusters = new Vector<Vector<String>>();
	//Calculate the distance matrix
	LinkedHashMap<Vector<String>, LinkedHashMap<Vector<String>, Double>> distances = new LinkedHashMap<Vector<String>, LinkedHashMap<Vector<String>, Double>>();
	for (int i=0; i<words.size(); i++) {
	    Vector<String> c1 = new Vector<String>();
	    c1.addElement(words.elementAt(i));
	    clusters.addElement(c1);
	    if (!distances.containsKey(c1))
		distances.put(c1, new LinkedHashMap<Vector<String>, Double>());
	    for (Vector<String> c2 : clusters) {
		if (!distances.containsKey(c2))
		    distances.put(c2, new LinkedHashMap<Vector<String>, Double>());
		double d = calculateClusterDistance(c1, c2);
		distances.get(c1).put(c2, d);
		distances.get(c2).put(c1, d);
		//System.out.println("Distance btwn " + c1.toString() + " and " + c2.toString() + " is " + distances.get(c1).get(c2));
	    }
	}
	//Run the algorithm
	while (clusters.size()>k) {
	    //Merge the two nearest clusters
	    double minDist = Double.MAX_VALUE;
	    Vector<String> c1 = null, c2 = null;
	    for (Vector<String> c : clusters) {
		for (Vector<String> d : clusters) {
		    //System.out.println(c + ", " + d + ", " + distances.containsKey(c) + ", " + (distances.containsKey(c) ? distances.get(c).containsKey(d) : "false"));
		    if (!c.equals(d) && distances.get(c).get(d)<minDist) {
			minDist = distances.get(c).get(d);
			c1 = c;
			c2 = d;
			//System.out.println("Found " + c1.toString() + " and " + c2.toString() + " with distance " + format.format(minDist));
		    }
		}
	    }
	    try {
		//Remove one of the clusters that got merged
		clusters.remove(c2);
		//Update the distances
		distances.remove(c2);
		for (Vector<String> c : clusters) 
		    distances.get(c).remove(c2);
		c1.addAll(c2);
		distances.put(c1, new LinkedHashMap<Vector<String>, Double>());
		for (Vector<String> c : clusters) {
		    double d = calculateClusterDistance(c1, c);
		    distances.get(c1).put(c, d);
		    distances.get(c).put(c1, d);
		}
	    }
	    catch (NullPointerException e) {
		/*System.out.println("Failed merging " + c1.toString() + " and " + c2.toString() + " with clusters " + clusters.toString() + ": " + e.toString());
		StackTraceElement s[] = e.getStackTrace();
		for(int i=0; i<s.length; i++)
		    System.err.println(s[i].toString());
		    break;*/
	    }
	}
	return clusters;
    }

    public Vector<Vector<String>> calculateGapStatistic(Vector<String> words, int maxClusters, int nMC) {
	//Use this method to calculate the gap statistic. Instead of generating uniform data in the interval, we generate random words of the given length.
	double gaps[] = new double[maxClusters], withinClusterDistances[] = new double[maxClusters], randomizedWithinClusterDistances[][] = new double[maxClusters][nMC], stdDevs[] = new double[maxClusters];
	Vector<Vector<Vector<String>>> allClusters = new Vector<Vector<Vector<String>>>();
	int bestK = 0;
	double maxDist = maxDistance(words);
	for (int i=1; i<=maxClusters; i++) {
	    //Cluster using i clusters
	    Vector<Vector<String>> clusters = clusterWords(words, i);
	    allClusters.addElement(clusters);
	    //Calculate the within cluster point scatter
	    withinClusterDistances[i-1] = calculateWithinClusterDistances(clusters);
	    //Generate randomized reference data sets
	    Tally stats = new Tally();
	    for (int j=0; j<nMC; j++) {
		randomizedWithinClusterDistances[i-1][j] = calculateWithinClusterDistances(clusterWords(randomWords(words, maxDist), i));
		stats.add(Math.log(randomizedWithinClusterDistances[i-1][j]));
	    }
	    //Calculate the gap statistic
	    for (int j=0; j<nMC; j++) 
		gaps[i-1] += (Math.log(randomizedWithinClusterDistances[i-1][j]) - Math.log(withinClusterDistances[i-1]))/nMC;
	    //Estimate the std dev
	    stdDevs[i-1] = Math.sqrt(stats.variance());
	    System.out.println("Gap is " + format.format(gaps[i-1]) + " and std dev " + format.format(stdDevs[i-1]) + " and within cluster distance is " + format.format(withinClusterDistances[i-1]) + " for " + i);
	    if (i>1 && gaps[i-2]>gaps[i-1] - 1*stdDevs[i-1]) { //Find the optimal number of clusters
		bestK = i-2;
		//break;
	    }
	}
	return allClusters.elementAt(bestK);
    }

    public Vector<Vector<String>> calculateLStatistic(Vector<String> words, int maxClusters) {
	//Use this method to calculate the L statistic. 
	double gaps[] = new double[maxClusters], withinClusterDistances[] = new double[maxClusters], lStat[] = new double[maxClusters-4], minL = Double.MAX_VALUE;
	Vector<Vector<Vector<String>>> allClusters = new Vector<Vector<Vector<String>>>();
	int bestK = 0;
	for (int i=1; i<=maxClusters; i++) {
	    //Cluster using i clusters
	    Vector<Vector<String>> clusters = clusterWords(words, i);
	    allClusters.addElement(clusters);
	    //Calculate the within cluster point scatter
	    withinClusterDistances[i-1] = calculateWithinClusterDistances(clusters);
	}
	System.out.println("Within cluster distances are " + Arrays.toString(withinClusterDistances));
	//Find the best fit using two straight lines
	for (int i=2; i<maxClusters-2; i++) {
	    SimpleRegression leftLine = new SimpleRegression(), rightLine = new SimpleRegression();
	    for (int j=0; j<withinClusterDistances.length; j++) {
		if (j<i)
		    leftLine.addData(j + 1, withinClusterDistances[j]);
		else
		    rightLine.addData(j + 1, withinClusterDistances[j]);
	    }
	    lStat[i-2] = i*Math.sqrt(leftLine.getSumSquaredErrors()/i)/lStat.length + (lStat.length - i + 2)*Math.sqrt(rightLine.getSumSquaredErrors()/(lStat.length - i + 2))/lStat.length;
	}
	System.out.println("lStats are " + Arrays.toString(lStat));
	for (int i=0; i<lStat.length; i++) {
	    if (lStat[i]<minL) {
		minL = lStat[i];
		bestK = i;
	    }
	}
	return allClusters.elementAt(bestK);
    }
    

    public double maxDistance(Vector<String> words) {
	double dMax = 0;
	for (String w1 : words) {
	    for (String w2 : words) {
		double d = calculateDistance(w1, w2);
		dMax = dMax>d ? dMax : d;
	    }
	}
	return dMax;
    }

    public double maxDistance(String w1, Vector<String> words) {
	double dMax = 0;
	for (String w2 : words) {
	    double d = calculateDistance(w1, w2);
	    dMax = dMax>d ? dMax : d;
	}
	return dMax;
    }

    public Vector<String> randomWords(Vector<String> words, double maxD) {
	//Make sure that the spread is not too large and matches the
	//data set that we are trying to control for. The aim is that
	//the distances should be sampled uniformly from [1,
	//maxDistance]
	Vector<String> ws = new Vector<String>();
	int j = 1, l = words.elementAt(0).length();
	ws.addElement(new String(generateRandomWord(l)));
	while (ws.size()<words.size()) {
	    int targetDist = (int)maxD; //random.nextInt((int)maxD - 1) + 2; //(int)maxD;
	    String w;
	    do {
		w = new String(generateRandomWord(words.elementAt(j).length()));
	    }
	    while (ws.contains(w) || Math.round(maxDistance(w, ws))>targetDist);
	    ws.addElement(new String(w));
	    j++;
	}
	return ws;
    }

    public char[] generateRandomWord(int l) {
	char array[] = new char[l];
	for (int i=0; i<l; i++) {
	    double r = random.nextDouble();
	    if (r<a)
		array[i] = 'A';
	    else if (r<a+c)
		array[i] = 'C';
	    else if (r<a+c+g)
		array[i] = 'G';
	    else 
		array[i] = 'T';
	}
	return array;
    }

    public char[] generateRandomWord(Vector<String> words, int l) {
	//Use the frequency of the over-represented words instead
	char array[] = new char[l];
	for (int i=0; i<l; i++) 
	    array[i] = words.elementAt(random.nextInt(words.size())).toCharArray()[i];
	return array;
    }

    public double calculateWithinClusterDistances(Vector<Vector<String>> clusters) {
	double d = 0;
	for (Vector<String> c : clusters) {
	    for (String w1 : c) {
		for (String w2 : c)
		    d += calculateDistance(w1, w2)/2;
	    }
	}
	return d;
    }

    public double calculateClusterDistance(Vector<String> c1, Vector<String> c2) {
	//Calculate the average distance between all elements in the cluster
	double d = 0;
	for (String u : c1) {
	    for (String v : c2) 
		d += calculateDistance(u, v);
	}
	return d/(c1.size()*c2.size());
    }

    public double calculateDistance(String u, String v) { return motifDistance(u.toCharArray(), v.toCharArray()); /*return Math.min(motifDistance((MerTree.reverseComplement(u)).toCharArray(), v.toCharArray()), motifDistance(u.toCharArray(), v.toCharArray()));*/ }

    public double motifDistance(char[] u, char[] v) {
	//The distance between two words
	/*if (u.length==v.length) { //Hamming distance
	    double d = 0; //Number of bases that differ
	    for (int i=0; i<u.length; i++)
		d += u[i]==v[i] ? 0 : 1;
	    return d;
	}
	else*/
	    return levenshteinDistance(u, v);
    }

    public double levenshteinDistance(char[] s, char[] t) {
	// for all i and j, d[i,j] will hold the Levenshtein distance between
	// the first i characters of s and the first j characters of t;
	// note that d has (m+1)x(n+1) values
	int m = s.length, n = t.length;
	double d[][] = new double[m][n];
	for (int i=0; i<m; i++)
	    d[i][0] = i; // the distance of any first string to an empty second string
	for (int j=0; j<n; j++)
	    d[0][j] = j; // the distance of any second string to an empty first string
	for (int j=1; j<n; j++) {
	    for (int i=1; i<m; i++) {
		if (s[i]==t[j])
		    d[i][j] = d[i-1][j-1];       // no operation required
		else
		    d[i][j] = Math.min(d[i-1][j] + 1, Math.min(d[i][j-1] + 1, d[i-1][j-1] + 1));
	    }
	}
	return d[m-1][n-1];
    }

    ///////////////////////////////////////////////////////

    public int numberOfSequences() { return numberOfSequences(false); }

    public int numberOfSequences(boolean ctrl) { return ctrl ? ctrlSeqs.size() : seqs.size(); }

    public int numberOfClusterOccurences(Vector<String> c) { return numberOfClusterOccurences(c, false); }

    public int numberOfClusterOccurences(Vector<String> c, boolean ctrl) {
	//Find out how many sequences have at least one of the members in c
	int n = 0;
	Vector<String> ss = ctrl ? ctrlSeqs : seqs;
	for (String s : ss) {
	    for (String w : c) {
		boolean palindromic = w.equals(MerTree.reverseComplement(w));
		Pattern pattern = Pattern.compile(w), rcPattern = Pattern.compile(MerTree.reverseComplement(w));
		Matcher m = pattern.matcher(s);
		if (m.find()) {
		    n++;
		    break;
		}
		else if (!palindromic) {
		    m = rcPattern.matcher(s);
		    n += m.find() ? 1 : 0;
		    break;
		}
	    }
	}
	return n;
    }


    protected Vector<int[][]> findMotifs(int k) {
	//Get the clusters
	return findMotifs(clusterWords(new Vector<String>(enrichedWords.keySet()), k));
    }

    public Vector<int[][]> findMotifs(Vector<Vector<String>> clusters) {
	Vector<int[][]> motifs = new Vector<int[][]>();
	int len = clusters.firstElement().firstElement().length();
	for (Vector<String> c : clusters) {
	    int motif[][] = new int[len][4];
	    for (String w : c) {
		for (int i=0; i<len; i++) {
		    if (w.toCharArray()[i]=='A')
			motif[i][0] += enrichedWords.get(w);
		    else if (w.toCharArray()[i]=='C')
			motif[i][1] += enrichedWords.get(w);
		    else if (w.toCharArray()[i]=='G')
			motif[i][2] += enrichedWords.get(w);
		    else if (w.toCharArray()[i]=='T')
			motif[i][3] += enrichedWords.get(w);
		}
	    }
	    motifs.addElement(motif);
	}
	return motifs;
    }

    public Vector<int[][]> findEnrichedMotifs(int len, double pVal, int k) {
	//Extract the motifs from the clustered words
	if (k>0) {
	    findEnrichedWords(len, pVal);
	    return findMotifs(k);
	}
	return findEnrichedMotifs(len, pVal);
    }

    public Vector<int[][]> findEnrichedMotifs(int len, double pVal) { return findEnrichedMotifs(len, len, pVal); }

    public Vector<int[][]> findEnrichedMotifs(int minLen, int maxLen, double pVal, int k) {
	if (k>0) {
	    for (int i=minLen; i<=maxLen; i++)
		findEnrichedWords(i, pVal);
	    return findMotifs(k);
	}
	return findEnrichedMotifs(minLen, maxLen, pVal);
    }

    public Vector<int[][]> findEnrichedMotifs(int minLen, int maxLen, double pVal) {
	//Extract the motifs from the clustered words
	for (int i=minLen; i<=maxLen; i++)
	    findEnrichedWords(i, pVal);
	Vector<Vector<String>> clusters = calculateLStatistic(new Vector<String>(enrichedWords.keySet()), Math.min(enrichedWords.size() - 1, maxLen*2));
	return findEnrichedMotifs(clusters);
    }

    public Vector<int[][]> findEnrichedMotifs(Vector<Vector<String>> clusters) {
	Vector<int[][]> motifs = new Vector<int[][]>();
	int longestWordLength = 0;
	for (String w : enrichedWords.keySet())
	    longestWordLength = Math.max(w.length(), longestWordLength);
	for (Vector<String> c : clusters)
	    motifs.addElement(findEnrichedMotifs(c));
	return motifs;
    }

    public int[][] findEnrichedMotifs(Vector<String> c) {
	int matrix[][] = null;
	try { //Run clustalw to align the words
	    if (c.size()>1) {
		int i = 0;
		BufferedWriter bw = new BufferedWriter(new FileWriter(TMP_CLUSTALW_FILE_IN));
		for (String w : c)
		    bw.write(">word_" + (i++) + "\n" + w + "\n");
		bw.close();
		Runtime runtime = Runtime.getRuntime();
		String clustalwCmd = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Softwares/clustalw-2.1-linux-x86_64-libcppstatic/clustalw2 -infile=/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/Athar/" + TMP_CLUSTALW_FILE_IN + " -stats=test.stats -align -type=dna -ktuple=1 -window=5 -score=percent -topdiags=5 -pairgap=3 -pwmatrix=gonnet -pwdnamatrix=iub -pwgapopen=100 -pwgapext=10.0 -matrix=gonnet -dnamatrix=iub -gapopen=100 -gapext=10.0 -gapdist=10 -iteration=none -numiter=1 -clustering=NJ -seqnos=on -outorder=aligned";
		Process clustalw = runtime.exec(clustalwCmd);
		try { clustalw.waitFor(); }
		catch (InterruptedException e) {
		    System.err.println("Error waiting for clustalw process " + e.toString());
		}
		//Parse the alignment and create the motif matrix
		BufferedReader br = new BufferedReader(new FileReader(TMP_CLUSTALW_FILE_OUT));
		String line = br.readLine();
		while (line!=null) {
		    if (line.startsWith("word")) {
			String tokens[] = line.split("\\s+");
			if (tokens.length==3) {
			    char[] aligned = tokens[1].toCharArray();
			    if (matrix==null)
				matrix = new int[aligned.length][4];
			    //Find the number of occurrences of this word
			    String word = tokens[1].replace("-", "");
			    //Use a weight factor to penalize the shorter words
			    int weight = 1; //(int)Math.pow(4, longestWordLength - word.length());
			    for (int j=0; j<aligned.length; j++) {
				if (aligned[j]=='-') {
				    for (int k=0; k<3; k++)
					matrix[j][k] += 0; //enrichedWords.get(word)/(4*weight);
				}
				else if (aligned[j]=='A')
				    matrix[j][0] += enrichedWords.get(word)/weight;
				else if (aligned[j]=='C')
				    matrix[j][1] += enrichedWords.get(word)/weight;
				else if (aligned[j]=='G')
				    matrix[j][2] += enrichedWords.get(word)/weight;
				else if (aligned[j]=='T')
				    matrix[j][3] += enrichedWords.get(word)/weight;
			    }
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    else {
		String word = c.elementAt(0);
		char[] aligned = word.toCharArray();
		matrix = new int[aligned.length][4];
		int weight = 1;
		for (int j=0; j<aligned.length; j++) {
		    if (aligned[j]=='A')
			matrix[j][0] += enrichedWords.get(word)/weight;
		    else if (aligned[j]=='C')
			matrix[j][1] += enrichedWords.get(word)/weight;
		    else if (aligned[j]=='G')
			matrix[j][2] += enrichedWords.get(word)/weight;
		    else if (aligned[j]=='T')
			matrix[j][3] += enrichedWords.get(word)/weight;
		}
	    }
	}
	catch (IOException e) {
	    System.out.println("Error with clustalw files " + e.toString());
	}
	return matrix;
    }

    public static int[][] findEnrichedMotifs(HashMap<String, Double> wordWeights, boolean orchestra) {
	//Use this method when we have a set of words that we wish to cluster using clustalw
	int matrix[][] = null;
	try { //Run clustalw to align the words
	    if (wordWeights.size()>1) {
		int i = 0;
		BufferedWriter bw = new BufferedWriter(new FileWriter(TMP_CLUSTALW_FILE_IN));
		for (String w : wordWeights.keySet())
		    bw.write(">word_" + (i++) + "\n" + w + "\n");
		bw.close();
		Runtime runtime = Runtime.getRuntime();
		String clustalwCmd = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Softwares/clustalw-2.1-linux-x86_64-libcppstatic/clustalw2 -infile=/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/Athar/" + TMP_CLUSTALW_FILE_IN + " -stats=test.stats -align -type=dna -ktuple=1 -window=5 -score=percent -topdiags=5 -pairgap=3 -pwmatrix=gonnet -pwdnamatrix=iub -pwgapopen=100 -pwgapext=10.0 -matrix=gonnet -dnamatrix=iub -gapopen=100 -gapext=10.0 -gapdist=10 -iteration=none -numiter=1 -clustering=NJ -seqnos=on -outorder=aligned";
		Process clustalw = runtime.exec(clustalwCmd);
		try { clustalw.waitFor(); }
		catch (InterruptedException e) {
		    System.err.println("Error waiting for clustalw process " + e.toString());
		}
		//Parse the alignment and create the motif matrix
		BufferedReader br = new BufferedReader(new FileReader(TMP_CLUSTALW_FILE_OUT));
		String line = br.readLine();
		while (line!=null) {
		    if (line.startsWith("word")) {
			String tokens[] = line.split("\\s+");
			if (tokens.length==3) {
			    char[] aligned = tokens[1].toCharArray();
			    if (matrix==null)
				matrix = new int[aligned.length][4];
			    //Find the number of occurrences of this word
			    String word = tokens[1].replace("-", "");
			    //Use a weight factor to penalize the shorter words
			    int weight = 1; //(int)Math.pow(4, longestWordLength - word.length());
			    for (int j=0; j<aligned.length; j++) {
				if (aligned[j]=='-') {
				    for (int k=0; k<3; k++)
					matrix[j][k] += 0; //enrichedWords.get(word)/(4*weight);
				}
				else if (aligned[j]=='A')
				    matrix[j][0] += wordWeights.get(word)/weight;
				else if (aligned[j]=='C')
				    matrix[j][1] += wordWeights.get(word)/weight;
				else if (aligned[j]=='G')
				    matrix[j][2] += wordWeights.get(word)/weight;
				else if (aligned[j]=='T')
				    matrix[j][3] += wordWeights.get(word)/weight;
			    }
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    else {
		String word = wordWeights.keySet().iterator().next();
		char[] aligned = word.toCharArray();
		matrix = new int[aligned.length][4];
		int weight = 1;
		for (int j=0; j<aligned.length; j++) {
		    if (aligned[j]=='A')
			matrix[j][0] += wordWeights.get(word)/weight;
		    else if (aligned[j]=='C')
			matrix[j][1] += wordWeights.get(word)/weight;
		    else if (aligned[j]=='G')
			matrix[j][2] += wordWeights.get(word)/weight;
		    else if (aligned[j]=='T')
			matrix[j][3] += wordWeights.get(word)/weight;
		}
	    }
	}
	catch (IOException e) {
	    System.out.println("Error with clustalw files " + e.toString());
	}
	return matrix;
    }

    ////////////////////////////////////////////

    public Vector<Vector<String>> findEnrichedClusters(int len, double pVal, int k) { return findEnrichedClusters(len, len, pVal, k); }

    public Vector<Vector<String>> findEnrichedClusters(int minLen, int maxLen, double pVal, int k) {
	if (k>=0) {
	    TreeMap<String, Double> words = new TreeMap<String, Double>();
	    for (int i=minLen; i<=maxLen; i++)
		words.putAll(findEnrichedWords(i, pVal));
	    return clusterWords(new Vector<String>(words.keySet()), k);
	}
	return findEnrichedClusters(minLen, maxLen, pVal);
    }

    public Vector<Vector<String>> findEnrichedClusters(int len, double pVal) { return findEnrichedClusters(len, len, pVal); }

    public Vector<Vector<String>> findEnrichedClusters(int minLen, int maxLen, double pVal) {
	TreeMap<String, Double> words = new TreeMap<String, Double>();
	for (int i=minLen; i<=maxLen; i++)
	    words.putAll(findEnrichedWords(i, pVal));
	return calculateLStatistic(new Vector<String>(words.keySet()), Math.min(enrichedWords.size() - 1, maxLen*2)); //calculateGapStatistic(new Vector<String>(words.keySet()), maxLen*2, N_MC);
    }

    public TreeMap<String, Double> findEnrichedWords(int len, double pVal) {
	//Generate all permutations of len
	TreeSet<String> words = new TreeSet<String>(getAllWords(len));
	try {
	    BufferedWriter bwSeq = new BufferedWriter(new FileWriter(seqFileStart + "_words_" + len + ".txt"));
	    BufferedWriter bwCtrlSeq = new BufferedWriter(new FileWriter(ctrlSeqFileStart + "_words_" + len + ".txt"));
	    double[][] nWords = new double[words.size()][2];
	    int i = 0;
	    ListOfTallies tallies = ListOfTallies.createWithTallyStore(2);
	    for (String w : words) {
		//Find the number in the control set
		nWords[i][0] = (double)numberOfWords(w, seqs);
		bwSeq.write(w + "\t" + (int)nWords[i][0] + "\n");
		nWords[i][1] = (double)numberOfWords(w, ctrlSeqs);
		bwCtrlSeq.write(w + "\t" + (int)nWords[i][1] + "\n");
		tallies.add(nWords[i]);
		i++;
	    }
	    //Calculate the correlation
	    double rho = tallies.correlation(0, 1);
	    i = 0;
	    for (String w : words) {
		if (nWords[i][0]>0) {
		    //Calculate the expected number based on the 1st order Markov model
		    double expControl = numberOfExpectedWords(w.toUpperCase().toCharArray(), ctrlSeqs.size(), new double[]{aCtrl, cCtrl, gCtrl, tCtrl}, totCtrlSeqLen), expControlVar = expControl*(1 - expControl/ctrlSeqs.size());
		    double exp = numberOfExpectedWords(w.toUpperCase().toCharArray(), seqs.size(), new double[]{a, c, g, t}, totSeqLen), expVar = exp*(1 - exp/seqs.size());
		    //Calculate the Geary-Hinkly transformation
		    double z = (double)nWords[i][1]/nWords[i][0]; //To avoid dividing by zero
		    double t = (exp*z - expControl)/Math.sqrt(expVar*z*z - 2*rho*Math.sqrt(expVar*expControlVar)*z + expControlVar);
		    //Test if t is really small
		    NormalDistributionImpl normDist = new NormalDistributionImpl();
		    try { 
			if (normDist.cumulativeProbability(t)<pVal && /*n/exp>nControl/expControl &&*/ nWords[i][0]>exp/* && nWords[i][1]<expControl*/) {
			    System.out.println(w + " was found " + nWords[i][0] + " times in seqs and " + nWords[i][1] + " times in controls, expected was " + format.format(exp) + " in control " + format.format(expControl) + " with p-val " + format.format(normDist.cumulativeProbability(t)));
			    enrichedWords.put(w, (double)nWords[i][0]);
			}
		    }
		    catch (org.apache.commons.math.MathException e) {
		    }
		    i++;
		    //Calculate the p-val based on the binomial distribution
		    /*BinomialDistributionImpl bino = new BinomialDistributionImpl(seqs.size(), (double)nControl/ctrlSeqs.size());
		      double expected = ((double)(nControl*seqs.size())/ctrlSeqs.size());
		      double p = 1, enrichment = (double)n/expected;
		      try {
		      p = 1 - bino.cumulativeProbability(n);
		      }
		      catch (org.apache.commons.math.MathException e) {
		      p = 1;
		      }
		      if (p<pVal && enrichment>2) {
		      System.out.println(w + " was found " + n + " times in seqs and " + nControl + " timesin controls, expected was " + " with p-val " + format.format(p));
		      enrichedWords.put(w, (double)n);
		      }*/	    
		}
	    }
	    bwSeq.close();
	    bwCtrlSeq.close();
	}
	catch (IOException e) {
	    System.out.println("Error saving word counts " + e.toString());
	}
	return enrichedWords;
    }

    protected double numberOfExpectedWords(char[] w, int nSeqs, double[] freqs, int totLens) {
	double p = 1;
	for (int i=0; i<w.length; i++) {
	    if (w[i]=='A')
		p *= freqs[0];
	    else if (w[i]=='C')
		p *= freqs[1];
	    else if (w[i]=='G')
		p *= freqs[2];
	    else if (w[i]=='T')
		p *= freqs[3];
	}
	//We want to find the probability of having one or more words in each sequence. Hence, we make a Poisson approximation:
	double pInSeq = 1 - Math.exp(-p*((double)(totLens - nSeqs*w.length))/nSeqs);
	//System.out.println(Arrays.toString(w) + " has " + format.format(pInSeq));
	return pInSeq*nSeqs;
    }

    public static int[][] numberOfWords(Vector<String> ws, Vector<String> seqs) {
	int n[][] = new int[seqs.size()][ws.size()];
	for (int i=0; i<seqs.size(); i++) {
	    Vector<String> ss = new Vector<String>();
	    ss.addElement(seqs.elementAt(i));
	    for (int j=0; j<ws.size(); j++) {
		n[i][j] = numberOfWords(ws.elementAt(j), ss);
	    }
	}
	return n;
    }

    /*public static int[] numberOfWords(Vector<String> ws, Vector<String> seqs) {
	int n[] = new int[ws.size()], i = 0;
	for (String w : ws)
	    n[i++] = numberOfWords(w, seqs);
	return n;
	}*/

    public static int numberOfWords(String w, Vector<String> seqs) {
	//Check both strands
	int n = 0;
	boolean palindromic = w.equals(MerTree.reverseComplement(w));
	Pattern pattern = Pattern.compile(w), rcPattern = Pattern.compile(MerTree.reverseComplement(w));
	for (String seq : seqs) {
	    Matcher m = pattern.matcher(seq);
	    int s = 0;
	    while (m.find(s)) {
		n++;
		s = m.start() + 1;
	    }
	    if (!palindromic) {
		m = rcPattern.matcher(MerTree.reverseComplement(seq));
		s = 0;
		while (m.find(s)) {
		    n++;
		    s = m.start() + 1;
		}
	    }
	}
	return n;
    }

    public static LinkedHashSet<String> getAllWords(int len) {
	LinkedHashSet<String> words = new LinkedHashSet<String>();
	char w[] = new char[len];
	generateWords(w, 0, words);
	System.out.println("Generated " + words.size() + " words of length " + len);
	return words;
    }

    protected static void generateWords(char[] w, int j, LinkedHashSet<String> words) {
	if (j==w.length) {
	    String v = new String(w);
	    if (!words.contains(MerTree.reverseComplement(v)))
		words.add(v);
	}
	else {
	    for (int i=0; i<4; i++) {
		if (i==0)
		    w[j] = 'A';
		else if (i==1)
		    w[j] = 'C';
		else if (i==2)
		    w[j] = 'G';
		else if (i==3)
		    w[j] = 'T';
		generateWords(w, j + 1, words);
	    }
	}
    }

}