//Use this to estimate the fragment length from the ChIP-Seq
//experiments. The idea is to shift by different amounts and calculate
//the auto-correlation.
//
//Example:
//
//java -Xmx1024m EstimateFragmentLength -n PeakFiles/all_tfs_names.txt -o
//
//java -Xmx1024m EstimateFragmentLength -o -n MmDataPeakFiles/mecp2_hwg_names.txt

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
//import umontreal.iro.lecuyer.stat.Tally;



public class EstimateFragmentLength{

    public static String GENOME_LOC;
    public static final String FRAGMENT_LENGTH_ESTIMATE_SUFFIX = "_fragment_length_estimate.txt";

    protected static int nMismatches, windowSize;
    protected static boolean verbose, unique, orchestra;
    protected static String fileName, inputFileName, saveFileName;
    protected static double inputReadsRatio, pValue;
    
    public static DecimalFormat format = new DecimalFormat("###.#######");

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option singletonFlag = argParser.addBooleanOption('s', "singletons");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('f', "saveFileName");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    windowSize = tmp!=null ? ((Integer)tmp).intValue() : 240;
	    verbose = argParser.getOptionValue(verboseFlag)!=null;
	    unique = argParser.getOptionValue(uniqueFlag)!=null;
	    orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    saveFileName = tmp!=null ? (String)tmp : fileName.substring(0, fileName.lastIndexOf(".")) + FRAGMENT_LENGTH_ESTIMATE_SUFFIX;
	    //////////////////////////////////////////////////////////////////
	    String GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	    int nChr = 21, minShift = 70, maxShift = 200;
	    Vector<String> fileNames = new Vector<String>(), shortNames = new Vector<String>();
	    if (fileName.contains("names"))
		fileNames = FindOverlappingPeaks.getFileNames(fileName, shortNames);
	    else
		fileNames.addElement(fileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    bw.write("#TF");
	    for (int j=minShift/2; j<maxShift/2; j++) 
		bw.write("\t" + j);
	    bw.write("\n");
	    for (int k=0; k<fileNames.size(); k++) {
		//System.out.println("Analyzing fragment length for " + shortNames.elementAt(k));
		int nReads[] = new int[nChr], readsSum = 0;
		double crossCorr[][] = new double[nChr][maxShift/2 - minShift/2];
		for (int i=0; i<nChr; i++) {
		    if (chr==-1 || chr==i+1) {
			//Get all the reads
			ABIRead reads[] = PeakFinder2.parseFiles(i+1, fileNames.elementAt(k), 0, false);
			nReads[i] = reads.length;
			readsSum += reads.length;
			for (int j=minShift/2; j<maxShift/2; j++) {
			    //Shift all reads on the forward strand j bases
			    //forward and on the reverse strand j bases
			    //backward.
			    crossCorr[i][j-minShift/2] = calculateCrossCorrelation(i+1, shiftReads(reads, j));
			    //System.out.println(j + "\t" + format.format(crossCorr[j-minShift/2]) + "\n");
			}
		    }
		}
		//Calculate the weighted average
		bw.write(shortNames.elementAt(k));
		for (int j=minShift/2; j<maxShift/2; j++) {
		    double mean = 0;
		    for (int i=0; i<nChr; i++) {
			if (chr==-1 || chr==i+1) 
			    mean += crossCorr[i][j-minShift/2]*nReads[i];
		    }
		    bw.write("\t" + format.format(mean/readsSum));
		}
		bw.write("\n");
	    }
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    public static double calculateCrossCorrelation(int chr, ABIRead reads[]) {
	//Find the number of reads at each site
	TreeMap<Integer, Integer> chrVecForward = new TreeMap<Integer, Integer>(), chrVecReverse = new TreeMap<Integer, Integer>();
	for (int i=0; i<reads.length; i++) {
	    if (reads[i].getStrand()) 
		chrVecForward.put(reads[i].getPosition(), chrVecForward.containsKey(reads[i].getPosition()) ? chrVecForward.get(reads[i].getPosition()) + 1 : 1);
	    else 
		chrVecReverse.put(reads[i].getPosition(), chrVecReverse.containsKey(reads[i].getPosition()) ? chrVecReverse.get(reads[i].getPosition()) + 1 : 1);
	}
	//Now we can calculate the correlation between the two count vectors
	long forwardSum = 0, reverseSum = 0, forwardSum2 = 0, reverseSum2 = 0, prodSum = 0;
	for (int p : chrVecForward.keySet()) {
	    forwardSum += chrVecForward.get(p);
	    forwardSum2 += chrVecForward.get(p)*chrVecForward.get(p);
	    if (chrVecReverse.containsKey(p)) //Do the cross-product
		prodSum += chrVecForward.get(p)*chrVecReverse.get(p);
	}
	for (int p : chrVecReverse.keySet()) {
	    reverseSum += chrVecReverse.get(p);
	    reverseSum2 += chrVecReverse.get(p)*chrVecReverse.get(p);
	}
	int chrLen = PeakFinder2.CHROMOSOME_LENGTH[chr - 1];
	return (double)(chrLen*prodSum - forwardSum*reverseSum)/(Math.sqrt(chrLen*forwardSum2 - forwardSum*forwardSum)*Math.sqrt(chrLen*reverseSum2 - reverseSum*reverseSum));
    }

    public static ABIRead[] shiftReads(ABIRead reads[], int shift) {
	ABIRead shiftedReads[] = new ABIRead[reads.length];
	for (int i=0; i<reads.length; i++) {
	    shiftedReads[i] = new ABIRead(reads[i]);
	    if (reads[i].getStrand()) 
		shiftedReads[i].setPosition(shiftedReads[i].getPosition() + shift);
	    else
		shiftedReads[i].setPosition(shiftedReads[i].getPosition() - shift);
	}
	return shiftedReads;
    }

}
