//Use this method to estimate the amount of transcription in the
//repeat elements. We do this by checking if a non-uniquely mappable
//read maps to at least one repeat element.
//
//bsub -q shared_2h java -Xmx4096m -server EstimateRepeatRegionTranscription -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c_b1_z10/hr6c_b1_z10 -o


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class EstimateRepeatRegionTranscription {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option intronicFlag = argParser.addBooleanOption('I', "intronic");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option reverseFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option sumsFileFlag = argParser.addBooleanOption("sumsFile");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option bedFileNameFlag = argParser.addStringOption('b', "bedFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean intronic = argParser.getOptionValue(intronicFlag)!=null;
	    boolean forward = argParser.getOptionValue(reverseFlag)==null;
	    boolean sumsFile = argParser.getOptionValue(sumsFileFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(bedFileNameFlag);
	    String bedFileName = tmp!=null ? (String)tmp : fileName;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : fileName + "_repeat_regions" + (sumsFile ? (forward ? "_f" : "_r") : "") + (chr==-1 ? "" : "_" + chr) + ".txt";
	    String saveFileNameUnique = saveFileName.replace("regions", "regions_unique");
	    ///////////////////////////////////////////////////////
	    boolean rna = fileName.contains("RNA");
	    int nChr = 21, totReads = 0, totUniqueReads = 0, fragmentLength = rna ? 35 : 120;
	    String rnaFileName = fileName + ".tsv", rnaFileNameForward = fileName + "_f", rnaFileNameReverse = fileName + "_r";
	    //Use this data-structure to keep track of the number of reads that maps to a given type of repeat
	    RepeatInfo repeatInfo[] = new RepeatInfo[nChr];
	    TreeMap<String, String> name2class = new TreeMap<String, String>();
	    TreeMap<String, TreeMap<String, Integer>> repFamily2name2reads = new TreeMap<String, TreeMap<String, Integer>>(), repFamily2name2readsUnique = new TreeMap<String, TreeMap<String, Integer>>();
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    //Load the annotation
		    repeatInfo[i-1] = new RepeatInfo(i, orchestra);
		    name2class.putAll(repeatInfo[i-1].getName2classMap());
		    //Load reads. Here we don't have the sums files, so we treat the reads as ChIP reads
		    ABIRead reads[] = sumsFile ? ABIReads.loadMRNAReadsChr(rnaFileName, forward, i, fragmentLength, false) : parseFiles(i, fileName, fragmentLength, 0);
		    System.out.println("Found " + reads.length + " reads for chr " + i);
		    //totReads += reads.length;
		    //Check if each read will map to at least one repeat region
		    Iterator<RepeatRegion> repeatIt = repeatInfo[i-1].getRepeatRegions().iterator();
		    RepeatRegion r = repeatIt.next();
		    for (int j=0; j<reads.length; j++) {
			if (reads[j].numberOfMatches()>1)
			    totReads++;
			else
			    totUniqueReads++;
			if (reads[j].getPosition()>r.getEnd() && repeatIt.hasNext())
			    r = repeatIt.next();
			if (r.overlaps(reads[j].getPosition(), reads[j].getPosition() + reads[j].getLength())) {
			    TreeMap<String, TreeMap<String, Integer>> map = reads[j].numberOfMatches()>1 ? repFamily2name2reads : repFamily2name2readsUnique;
			    //if (totUniqueReads % 1000 == 0)
			    //System.out.println(reads[j].toString());
			    if (!map.containsKey(r.getFamilyName())) 
				map.put(r.getFamilyName(), new TreeMap<String, Integer>());
			    if (!map.get(r.getFamilyName()).containsKey(r.getName()))
				map.get(r.getFamilyName()).put(r.getName(), 0);
			    map.get(r.getFamilyName()).put(r.getName(), map.get(r.getFamilyName()).get(r.getName()) + 1);
			}
		    }
		}
	    }
	    //Invert the mapping
	    TreeMap<String, TreeSet<String>> class2name = new TreeMap<String, TreeSet<String>>();
	    for (String c : name2class.values())
		class2name.put(c, new TreeSet<String>());
	    for (String n : name2class.keySet())
		class2name.get(name2class.get(n)).add(n);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    System.out.println("Saving results to " + saveFileName);
	    int totalRepeatReads = 0;
	    for (String k : class2name.keySet()) {
		if (repFamily2name2reads.containsKey(k)) {
		    for (String n : class2name.get(k)) {
			bw.write(k + "\t" + n + "\t" + (repFamily2name2reads.get(k).containsKey(n) ? repFamily2name2reads.get(k).get(n) : 0) + "\n");
			totalRepeatReads += repFamily2name2reads.get(k).containsKey(n) ? repFamily2name2reads.get(k).get(n) : 0;
		    }
		}
	    }
	    bw.write("Non-repeat\tNon-repeat\t" + (totReads - totalRepeatReads) + "\n");
	    bw.close();
	    
	    bw = new BufferedWriter(new FileWriter(saveFileNameUnique));
	    System.out.println("Saving results to " + saveFileNameUnique);
	    int totalRepeatReadsUnique = 0;
	    for (String k : class2name.keySet()) {
		if (repFamily2name2readsUnique.containsKey(k)) {
		    for (String n : class2name.get(k)) {
			bw.write(k + "\t" + n + "\t" + (repFamily2name2readsUnique.get(k).containsKey(n) ? repFamily2name2readsUnique.get(k).get(n) : 0) + "\n");
			totalRepeatReadsUnique += repFamily2name2readsUnique.get(k).containsKey(n) ? repFamily2name2readsUnique.get(k).get(n) : 0;
		    }
		}
	    }
	    bw.write("Non-repeat\tNon-repeat\t" + (totUniqueReads - totalRepeatReadsUnique) + "\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, int strand) {
	//Always use the same random number generator to ensure that the same reads are picked. Also, we need to go through all of the files as there might be additional mappings in other files.
	random = new Random(987);
	ABIRead rs[] = null;
	try {
	    int nReads = 0, nChr = 21;
	    Vector<ABIRead> reads = new Vector<ABIRead>();
	    for (int j=1; j<=nChr; j++) {
		BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + j));
		String line = br.readLine();
		Pattern locPattern = Pattern.compile("([0-9]+)_(-*?)([0-9]+)\\.([0-3])");
		while (line!=null) {
		    if (line.startsWith(">")) {
			String tokens[] = line.split(",");
			if (tokens.length>1) {
			    //Find all the reads with the lowest number of mismatches
			    TreeMap<Integer, Vector<ABIRead>> mismatches2readsMap = new TreeMap<Integer, Vector<ABIRead>>();
			    for (int i=1; i<tokens.length; i++) {
				Matcher m = locPattern.matcher(tokens[i]);
				if (m.find()/* && Integer.parseInt(m.group(1))==chr*/) {
				    int nMismatches = Integer.parseInt(m.group(4));
				    if (!mismatches2readsMap.containsKey(nMismatches))
					mismatches2readsMap.put(nMismatches, new Vector<ABIRead>());
				    boolean forward = !m.group(2).equals("-");
				    mismatches2readsMap.get(nMismatches).addElement(new ABIRead(tokens.length - 1, chr, forward, Integer.parseInt(m.group(3)) + 1 + (forward ? 0 : -fragmentLength), fragmentLength));
				}
			    }
			    //get a random read which has the lowest number of mismatches
			    ABIRead r = mismatches2readsMap.get(mismatches2readsMap.firstKey()).elementAt(random.nextInt(mismatches2readsMap.get(mismatches2readsMap.firstKey()).size()));
			    if (r.getChromosome()==chr)
				reads.addElement(r);
			}
		    }
		    line = br.readLine(); 
		}
		br.close();
	    }
	    rs = new ABIRead[reads.size()];
	    for (int i=0; i<reads.size(); i++)
		rs[i] = reads.elementAt(i);
	    Arrays.sort(rs);
	}
	catch (IOException e) {
	    System.err.println("Error, could not open file: " + fileName);
	}
	return rs;
    }

}



