//Use this class to keep track of all the information about the read
//density surrounding a given feature.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import java.io.*;


public class FeatureInfo {

    //public static final double EPSILON = 1e-8;

    protected String feature;
    protected int nChr, sumLocs = 0;
    protected Vector<Vector<Integer>> locVecs = new Vector<Vector<Integer>>(); //chr, loc
    protected Vector<Vector<Integer>> lengthVecs = new Vector<Vector<Integer>>(); //chr, length
    protected Vector<Vector<double[]>> histVecs = new Vector<Vector<double[]>>(); //chr, loc, histogram
    protected Vector<Vector<Double>> nReadsInWindowsUpstream = new Vector<Vector<Double>>(), nReadsInWindowsDownstream = new Vector<Vector<Double>>(); //chr, nReads
    protected Vector<Vector<Boolean>> strandVecs = new Vector<Vector<Boolean>>();

    protected double sumHistMean[], epsilon;

    public FeatureInfo(String f, int n, int b) {
	feature = f;
	nChr = n;
	sumHistMean = new double[b];
	for (int i=0; i<nChr; i++) {
	    locVecs.addElement(new Vector<Integer>());
	    lengthVecs.addElement(new Vector<Integer>());
	    strandVecs.addElement(new Vector<Boolean>());
	    nReadsInWindowsDownstream.addElement(new Vector<Double>());
	    nReadsInWindowsUpstream.addElement(new Vector<Double>());
	    histVecs.addElement(new Vector<double[]>());
	}
    }

    //////////////////////////////////////////////////////////////////

    //public void setStrandVec(boolean strandVec[], int chr) { strandVecs.setElementAt(strandVec, chr-1); }

    public void addLocInfo(int chr, int loc, boolean strand, int[] hist, double normFactor, int windowSize, int len, boolean cons) { addLocInfo(chr, loc, strand, hist, normFactor, windowSize, len, windowSize, cons, 35); }

    public void addLocInfo(int chr, int loc, boolean strand, int[] hist, double normFactor, int windowSize, int len, int upstream, boolean cons, int fragmentLength) {
	if (cons) {
	    addStrand(strand, chr);
	    addLoc(loc, chr);
	    sumLocs++;
	    //Count the reads in the window
	    double nReadsInWindowUpstream = 0, nReadsInWindowDownstream = 0, h[] = new double[hist.length];
	    int firstDownStreamInd = -1, lastDownStreamInd = -1;
	    for (int j=0; j<hist.length; j++) {
		h[j] = hist[j]*normFactor;
		//if (strand) {
		if (j - upstream>0 && j - upstream<windowSize)
		    nReadsInWindowDownstream += h[j]/fragmentLength;
		//}
		//else {
		if (j - upstream>-windowSize && j - upstream<0) 
		    nReadsInWindowUpstream += h[j]/fragmentLength;
		if (h[j]>0 && firstDownStreamInd<0)
		    firstDownStreamInd = j;
		if (h[j]>0 && firstDownStreamInd>0)
		    lastDownStreamInd = j;
		//}
	    }
	    if (!strand) {
		double tmp = nReadsInWindowUpstream;
		nReadsInWindowUpstream = nReadsInWindowDownstream;
		nReadsInWindowDownstream = tmp;
	    }
	    //System.out.println("Found " + nReadsInWindow + " reads for " + feature + " at chr" + chr + ":" + (loc-upstream) + "-" + (loc-upstream+hist.length));
	    addReadsInWindow(nReadsInWindowDownstream, chr, true);
	    addReadsInWindow(nReadsInWindowUpstream, chr, false);
	    epsilon = normFactor/2;
	    addHistogram(h, chr);
	    if (len==-1) { //Estimate length from the histogram
		len = Math.abs(lastDownStreamInd - firstDownStreamInd);
	    }
	    addLength(len, chr);
	}
    }

    public void addStrand(boolean strand, int chr) { strandVecs.elementAt(chr-1).addElement(strand); }

    public void addLoc(int loc, int chr) { locVecs.elementAt(chr-1).addElement(loc); }

    public void addLength(int len, int chr) { lengthVecs.elementAt(chr-1).addElement(len); }

    public void addReadsInWindow(double r, int chr, boolean downstream) { 
	if (downstream)
	    nReadsInWindowsDownstream.elementAt(chr-1).addElement(r);
	else
	    nReadsInWindowsUpstream.elementAt(chr-1).addElement(r);
    }
	
    public void addHistogram(double hist[], int chr) { histVecs.elementAt(chr-1).addElement(hist); }

    public Vector<Boolean> getStrandVec(int chr) { return strandVecs.elementAt(chr-1);     }

    public Vector<Integer> getLocVec(int chr) { return locVecs.elementAt(chr-1); }
    
    public Vector<Integer> getLengthVec(int chr) { return lengthVecs.elementAt(chr-1); }

    public Vector<Double> getReadsInWindow(int chr, boolean downstream) { return downstream ? nReadsInWindowsDownstream.elementAt(chr-1) : nReadsInWindowsUpstream.elementAt(chr-1); }

    public void calculateMeanHist() { calculateMeanHist(Double.MAX_VALUE, Double.MAX_VALUE); }

    public void calculateMeanHist(double maxValueUpstream, double maxValueDownstream) {
	int nNonZero[] = new int[sumHistMean.length], n[] = new int[sumHistMean.length];;
	for (int i=0; i<nChr; i++) {
	    double mult = ((double)locVecs.elementAt(i).size())/sumLocs;
	    if (mult>0) {
		//System.out.println("found " + histVecs.elementAt(i).size() + " histograms for " + i);
		for (int j=0; j<histVecs.elementAt(i).size(); j++) {
		    if (nReadsInWindowsUpstream.elementAt(i).elementAt(j)<maxValueUpstream && nReadsInWindowsDownstream.elementAt(i).elementAt(j)<maxValueDownstream) {
			for (int k=0; k<histVecs.elementAt(i).elementAt(j).length; k++) {
			    sumHistMean[k] += histVecs.elementAt(i).elementAt(j)[k]*mult;
			    //Unfortunately, geomtric mean does not work since there are so many bins that have zero reads
			    //sumHistMean[k] += Math.log(histVecs.elementAt(i).elementAt(j)[k] + epsilon) + Math.log(mult);
			    n[k]++;
			    if (histVecs.elementAt(i).elementAt(j)[k]>0)
				nNonZero[k]++;
			}
		    }
		    else 
			System.out.println("Excluded chr" + (i+1) + ":"  + (locVecs.elementAt(i).elementAt(j) - 10000) + "-" + (locVecs.elementAt(i).elementAt(j) + 10000));
		}
	    }
	}
	for (int i=0; i<sumHistMean.length; i++) {
	    /*if (nNonZero[i]>0)
		//sumHistMean[i] = Math.max(0, Math.exp(sumHistMean[i]/nNonZero[i]) - epsilon);
		sumHistMean[i] = nNonZero[i]>0 ? sumHistMean[i]/nNonZero[i] : 0;
	    else
	    sumHistMean[i] = 0;*/
	    sumHistMean[i] = n[i]==0 ? 0 : sumHistMean[i]/n[i];
	}
    }

    public double[] getSumHistMean() { return sumHistMean; }

}