//Use this method to assemble all the reads across a given feature, eg
//TSSs, for all chromosomes.
//
//Example:
//
//java -Xmx512m FindReadsAcrossFeature -n Npas4/KCl/Npas4_KCl -u 1000 -d 1000 -f TSS
//
//For fig3 of paper:
//
//java -Xmx1024m FindReadsAcrossFeature -f TSS -u 2000 -d 2000 -n /groups/neuroduo/martin/ChIP-Seq/TKData/8WG16/un/8WG16_un -o -i /groups/neuroduo/martin/ChIP-Seq/TKData/Input/un/Input_un
//
//For generating histograms for the maturity paper:
//
//java -Xmx1024m FindReadsAcrossFeature -f TSS -u 10000 -d 10000 -n MmDataPeakFiles/histones_names.txt -o -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/MNase/B1/Input_Input_MNase_B1 -N 0
//
//For enhancers:
//
//java -Xmx1024m FindReadsAcrossFeature -f /groups/neuroduo/martin/ChIP-Seq/TKData/Enhancers/enhancers_null_KCl.txt -u 5000 -d 5000 -n PeakFiles/enhancer_factor_names.txt -o -N 0
//
//java -Xmx1024m FindReadsAcrossFeature -f /groups/neuroduo/martin/ChIP-Seq/TKData/Enhancers/enhancers_null_KCl.txt -u 5000 -d 5000 -n PeakFiles/enhancer_factor_names.txt -o -N 0
//
//For random control regions
//
//bsub -q shared_12h java -Xmx1024m FindReadsAcrossFeature -f /groups/neuroduo/martin/ChIP-Seq/TKData/ -u 5000 -d 5000 -n MmDataPeakFiles/tf_names.txt -o -N 0


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import org.apache.commons.math.distribution.BinomialDistributionImpl;
import org.apache.commons.math.MathException;

public class FindReadsAcrossFeature{

    public static final double HIGHLY_CONSERVED_THRESHOLD = .9;
    public static final double QUALITY_SCORE_CUTOFF = 10;
    public static final int MIN_CAGE_TAGS_PER_TSS = 5;


    public static final int SENSE_START = -300;
    public static final int ANTI_SENSE_START = 150;
    //where the big sense driving peak seems to be located based on aggregate data
    public static final int MAJOR_SENSE_LOCATION = 40;
    public static final int MAJOR_ANTI_SENSE_LOCATION = -180;
    //A limit for the number of 
    public static final int MAX_TSS_DISTANCE_TRANSCRIPT_START = 2000;
    //Limits for what we consider to be the main convergent region
    public static final int CONVERGENT_UPSTREAM_REGION_START = -360;
    public static final int CONVERGENT_DOWNSTREAM_REGION_START = 150;
    public static final int MAX_CONVERGENT_REGION = 2000;

    public static final String TSS_HISTOGRAM_PREFIX = "/groups/neuroduo/martin/ChIP-Seq/MmData/Histograms/tss";
    public static final String TES_HISTOGRAM_PREFIX = "/groups/neuroduo/martin/ChIP-Seq/MmData/Histograms/tes";
    public static final String HISTOGRAM_SUFFIX = "_histograms_";
    public static final String AGGREGATE_SUFFIX = "_aggregate_";
    public static final String SHUFFLE_FILE_NAME_SUFFIX = "_tss_shuffle_";

    public static final String INTERGENIC_STR = "intergenic";
    public static final String INTRONIC_STR = "intronic";

    public static String GENOME_LOC;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    //Keep track of all the peaks in this datastructure
    protected static TreeSet[] peaks = new TreeSet[21];


    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option featureFlag = argParser.addStringOption('f', "feature");
	    //Use this flag if we only want to consider the intronic
	    //or intergenic enhancers, default is to consider all
	    CmdLineParser.Option enhancerSubsetFlag = argParser.addStringOption("enhancerSubset");
	    CmdLineParser.Option upstreamFlag = argParser.addIntegerOption('u', "upstream");
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('d', "downstream");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //Use this flag to search for the GC content in the region as well
	    CmdLineParser.Option gcFlag = argParser.addBooleanOption('G', "gc");
	    //Find the conservation scores as well
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption('C', "conservation");
	    //Use this flag when trying to find multiple peaks near
	    //the TSS by looking for local maxima. This value is
	    //determined heuristically and it represents a minimum
	    //distance enforced by the size of the pol2-complex
	    CmdLineParser.Option minPeakDistFlag = argParser.addIntegerOption('p', "minPeakDist");
	    //This flag sets the threshold for finding local minima
	    CmdLineParser.Option minPeakHeightFlag = argParser.addIntegerOption('P', "minPeakHeight");
	    //This flag is used to find the files with the starting
	    //locations of transcripts.
	    CmdLineParser.Option transcriptStartsFlag = argParser.addStringOption('t', "transcriptStarts");
	    //The number of times for shuffling the controls
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption('N', "nMC");
	    //Location of the CAGE tags, use when looking for peaks near CAGE tags
	    CmdLineParser.Option cageStartsFlag = argParser.addStringOption('T', "cageStarts");
	    //File containing a list of all the mRNA data sets that we
	    //want to investigate
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    //Avoid producing the files containing individual patterns for each loci
	    CmdLineParser.Option noHistsFlag = argParser.addBooleanOption("noHists");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean gc = argParser.getOptionValue(gcFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    tmp = argParser.getOptionValue(featureFlag);
	    String feature = tmp==null ? "TSS" : tmp.toString();
	    tmp = argParser.getOptionValue(enhancerSubsetFlag);
	    String enhancerSubset = tmp==null ? "" : tmp.toString();
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 300;
	    tmp = argParser.getOptionValue(selectFlag);
	    String selectFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(minPeakDistFlag);
	    int minPeakDist = tmp!=null ? ((Integer)tmp).intValue() : 150;
	    tmp = argParser.getOptionValue(minPeakHeightFlag);
	    int minPeakHeight = tmp!=null ? ((Integer)tmp).intValue() : 5;
	    tmp = argParser.getOptionValue(transcriptStartsFlag);
	    String transcriptStartsFileName = tmp!=null ? (String)tmp : "mRNA/hr0c_starts_n_stops_TSS_13kb.tsv";
	    tmp = argParser.getOptionValue(cageStartsFlag);
	    String cageStartsDirName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "deepCAGE/";
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/hr";
	    boolean noHists = argParser.getOptionValue(noHistsFlag)!=null;
	    //////////////////////////////////////////////////////////////////
	    String species = "mm";
	    int nChr = 21, fragmentLength = 120;
	    double qualityScore = QUALITY_SCORE_CUTOFF;
	    int isolatedIDs[] = isolated ? parseIsolatedFile(orchestra) : null;
	    int geneIDsKeep[] = parseSelectFile(selectFileName);
	    if (geneIDsKeep!=null) {
		Arrays.sort(geneIDsKeep);
		System.out.println("Only looking at " + geneIDsKeep.length + " genes");
	    }
	    Vector<String> fileNames = new Vector<String>(), shortNames = new Vector<String>();
	    if (fileName.contains("names"))
		fileNames = FindOverlappingPeaks.getFileNames(fileName, shortNames);
	    else
		fileNames.addElement(fileName);
	    System.out.println("Feature is " + feature + " and there are " + fileNames.size() + " factors.");
	    TreeMap<String, Integer> nTotalReads = new TreeMap<String, Integer>();
	    double histsMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], histsVar[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], histsMeanMC[][] = new double[nChr][numberOfBins(downstream, upstream, 1)];
	    double consHistsMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], consHistsMeanMC[][] = new double[nChr][numberOfBins(downstream, upstream, 1)];
	    double gcHistsMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], gcHistsMeanMC[][] = new double[nChr][numberOfBins(downstream, upstream, 1)];
	    double sumHistMean[][] = new double[fileNames.size()][numberOfBins(downstream, upstream, 1)], sumHistVar[][] = new double[fileNames.size()][numberOfBins(downstream, upstream, 1)];
	    double sumGCMean[] = new double[numberOfBins(downstream, upstream, 1)], sumGCVar[] = new double[numberOfBins(downstream, upstream, 1)], sumGCMeanMC[] = new double[numberOfBins(downstream, upstream, 1)];
	    double sumConsMean[] = new double[numberOfBins(downstream, upstream, 1)], sumConsVar[] = new double[numberOfBins(downstream, upstream, 1)], sumConsMeanMC[] = new double[numberOfBins(downstream, upstream, 1)];
	    double histsMeanInput[][] = new double[nChr][numberOfBins(downstream, upstream, 1)];
	    //First find the total number of TSSs
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    int nTSS[] = new int[nChr], sumTSS = 0;
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    //Load the annotation info
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
		    nTSS[i-1] = geneInfo[i-1].getStarts(true, geneIDsKeep).length;
		    sumTSS += nTSS[i-1];
		}
	    }			
	    if (feature.equals("TSS")) {
		int nReadsInPeaks = 0, nReadsInRegion = 0;
		int transcriptStarts[][] = parseInitiationFile(transcriptStartsFileName, isolatedIDs, true);
		int nPeaksNearStartSense = 0, nPeaksNearStartAntiSense = 0, nPeaksNearStartSenseCons = 0, nPeaksNearStartAntiSenseCons = 0, nPeaks = 0, nPeaksConserved = 0;
		int nPeaksNearCageStartSense = 0, nPeaksNearCageStartAntiSense = 0, nPeaksNearCageStartSenseCons = 0, nPeaksNearCageStartAntiSenseCons = 0;
		//Find how many peaks that can be explained by CAGE tags
		int nCageStartSenseNearPeaks = 0, nCageStartSenseNearPeaksCons = 0, nCageStartAntiSenseNearPeaks = 0, nCageStartAntiSenseNearPeaksCons = 0, nPeaksInGenesWithCageSense = 0, nPeaksInGenesWithCageAntiSense = 0;
		//Distribution of peaks near starts
		TreeMap<Integer, Integer> nPeaksNearStartSenseHist = new TreeMap<Integer, Integer>(), nPeaksNearStartAntiSenseHist = new TreeMap<Integer, Integer>(), nPeaksNearStartSenseConsHist = new TreeMap<Integer, Integer>(), nPeaksNearStartAntiSenseConsHist = new TreeMap<Integer, Integer>();
		//Distribution of peaks near CAGE starts
		TreeMap<Integer, Integer> nPeaksNearCageStartSenseHist = new TreeMap<Integer, Integer>(), nPeaksNearCageStartAntiSenseHist = new TreeMap<Integer, Integer>(), nPeaksNearCageStartSenseConsHist = new TreeMap<Integer, Integer>(), nPeaksNearCageStartAntiSenseConsHist = new TreeMap<Integer, Integer>();
		//Keep track of the number of peaks per region distribution
		TreeMap<Integer, Integer> peaksPerTSS = new TreeMap<Integer, Integer>(), peaksPerTSSConvergentRegions = new TreeMap<Integer, Integer>(), peaksPerTSSnonGenicMC = new TreeMap<Integer, Integer>(), peaksPerTSSgenicMC = new TreeMap<Integer, Integer>();
		int nPeakLocsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int peakHeight[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocsConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pSenseHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pAntiSenseHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pBiHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pSenseShuffleHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pAntiSenseShuffleHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pSenseConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pAntiSenseConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pSenseShuffleConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocs5pAntiSenseShuffleConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		//Histogram for CAGE stuff
		int nPeakLocsCageHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocsCageShuffleHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocsCageConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeakLocsCageShuffleConsHist[] = new int[numberOfBins(downstream, upstream, 1)];
		int nPeaksNearStartSenseShuffle[] = new int[nMC], nPeaksNearStartAntiSenseShuffle[] = new int[nMC], nPeaksNearStartSenseConsShuffle[] = new int[nMC], nPeaksNearStartAntiSenseConsShuffle[] = new int[nMC];		
		//Likewise for CAGE
		int nPeaksNearCageStartSenseShuffle[] = new int[nMC], nPeaksNearCageStartAntiSenseShuffle[] = new int[nMC], nPeaksNearCageStartSenseConsShuffle[] = new int[nMC], nPeaksNearCageStartAntiSenseConsShuffle[] = new int[nMC];
		//Count the number of peaks in the convergent regions
		int nPeaksInUpstreamConvergentRegion = 0, nPeaksInDownstreamConvergentRegion = 0, nPeaksInUpstreamConvergentRegionWithCageData = 0, nPeaksInDownstreamConvergentRegionWithCageData = 0, nPeaksInUpstreamConvergentRegionWithTranscript = 0, nPeaksInDownstreamConvergentRegionWithTranscript = 0, nGenesWithPeakAndCageInConvergentUpstream = 0, nGenesWithPeakAndCageInConvergentDownstream = 0, nGenesWithPeakAndTranscriptInConvergentUpstream = 0, nGenesWithPeakAndTranscriptInConvergentDownstream = 0;
		//Keep track of how many unique genes have peaks in
		//the convergent regions.
		int nGenesWithPeakInConvergentUpstream = 0, nGenesWithPeakInConvergentDownstream = 0, nGenesWithPeakAnd5pEndInConvergentUpstream = 0, nGenesWithPeakAnd5pEndInConvergentDownstream = 0;
		int nSenseInConvergentUpstream = 0, nAntiSenseInConvergentDownstream = 0;
		/////////////////////////////////
		//Get the profiles surrounding the TSSs
		nTotalReads = calculateTotalReads(fileNames, inputFileName, nChr, fragmentLength, qualityScore);
		for (int k=0; k<fileNames.size(); k++) {
		    nTotalReads.put(fileNames.elementAt(k), 0);
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) {
			    boolean saveHist = false;
			    int ids[] = geneInfo[i-1].getIDs();
			    //System.out.println("Getting info for chr " + i + " for " + ids.length + " genes.");
			    /////////////////////////////////
			    //Load the reads
			    ABIRead reads[] = fileNames.elementAt(k).contains("RNA") ? PeakFinder2.parseFiles(i, fileNames.elementAt(k), fragmentLength, true, 0, qualityScore) : ABIReads.loadMRNAReadsChr(fileNames.elementAt(k), true, i, 35);
			    nTotalReads.put(fileNames.elementAt(k), nTotalReads.get(fileNames.elementAt(k)) + reads.length);
			    ABIRead readsInput[] = inputFileName!=null ? PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore) : null;
			    int tssVec[] = getTSSs(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    boolean strandVec[] = getStrands(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    //Get the histograms surrounding the TSSs
			    //int hists[][] = getHistograms(i, tssVec, strandVec, reads, null, downstream, upstream, fragmentLength, saveHist);
			    String histSaveName = null;
			    if (!noHists)
				histSaveName = (feature.equals("TSS") ? TSS_HISTOGRAM_PREFIX : TES_HISTOGRAM_PREFIX) + "_" + shortNames.elementAt(k) + HISTOGRAM_SUFFIX + (readsInput==null ? "" : "input_") + i + ".txt";
			    int hists[][] = getHistograms(i, tssVec, strandVec, reads, readsInput, downstream, upstream, fragmentLength, histSaveName, (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(fileNames.elementAt(k)), inputFileName==null ? 0 : (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(inputFileName));
			    int histsInput[][] = readsInput!=null ? getHistograms(i, tssVec, strandVec, reads, readsInput, downstream, upstream, fragmentLength, saveHist) : hists;
			    //Get the average histogram
			    double h[][] = findMeanReadHistograms(hists, numberOfBins(downstream, upstream, 1));
			    histsMean[i-1] = h[0];
			    histsVar[i-1] = h[1];
			    if (k==0 && readsInput!=null) 
				histsMeanInput[i-1] = findMeanReadHistograms(getHistograms(i, tssVec, strandVec, readsInput, null, downstream, upstream, fragmentLength, false), numberOfBins(downstream, upstream, 1))[0];
			    Vector<TreeMap<Integer, Integer>> peakLocs = findAllPeaks(hists, histsInput, minPeakDist, minPeakHeight);
			    //Get the GC distribution 
			    if (gc && k==0) { gcHistsMean[i-1] = getGCHistogram(i, orchestra, tssVec, strandVec, upstream, downstream); }
			    //Get the conservation scores
			    double pcs[][] = null;
			    Vector<TreeMap<Integer, Boolean>> peakConserved = null;
			    if (conservation && k==0) {
				PhastConsParser verterbateCons = new PhastConsParser(i, false, orchestra, species);
				pcs = new double[tssVec.length][numberOfBins(downstream, upstream, 1)];
				for (int j=0; j<tssVec.length; j++) {
				    if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, ids[j])>=0)
					pcs[j] = strandVec[j] ? verterbateCons.findScores(tssVec[j] - upstream, tssVec[j] + downstream + 1) : flipArray(verterbateCons.findScores(tssVec[j] - downstream, tssVec[j] + upstream + 1));
				}
				consHistsMean[i-1] = getConservationHistogram(pcs);
				peakConserved = findAllConservedPeaks(hists, histsInput, pcs, minPeakDist, minPeakHeight);
				for (TreeMap<Integer, Boolean> promoterPeaks : peakConserved) {
				    for (Integer p : promoterPeaks.keySet())
					nPeaksConserved += promoterPeaks.get(p) ? 1 : 0;
				}
			    }
			    boolean modeOnly = false;
			    //Key is geneID and value is a list of CAGE tag starts
			    //TreeMap<Integer, Vector<Integer>> cageStartsSense = parseCageTagFile(cageStartsDirName + CompareCageEstRNASeqStarts.CAGE_EST_RNA_COMPARISON_FILE_NAME + "_" + i + "_S_CAGE", i, isolatedIDs, true, true, modeOnly);
			    //TreeMap<Integer, Vector<Integer>> cageStartsAntiSense = parseCageTagFile(cageStartsDirName + CompareCageEstRNASeqStarts.CAGE_EST_RNA_COMPARISON_FILE_NAME + "_" + i + "_AS_CAGE", i, isolatedIDs, false, true, modeOnly);
			    /////////////////////////////////
			    //Get the locations, conservation and enrichment of all peaks
			    if (fileNames.size()==1) {
				for (int j=0; j<peakLocs.size(); j++) {
				    TreeMap<Integer, Integer> ps = peakLocs.get(j);
				    TreeMap<Integer, Boolean> peakCons = peakConserved!=null ? peakConserved.get(j) : null;
				    int peakInConvergentUpstream = 0, peakInConvergentDownstream = 0;
				    int ind = Arrays.binarySearch(transcriptStarts[0], ids[j]);
				    for (Integer p : ps.keySet()) {
					nPeaks++;
					nPeakLocsHist[p]++;
					if (peakCons!=null && peakCons.containsKey(p) && peakCons.get(p))
					    nPeakLocsConsHist[p]++;
					/////////////////////////////////
					//Check if the peak appears to be driving a transcript
					if (ind>=0) {
					    boolean found5pSense = false, found5pAntiSense = false;
					    if (!found5pSense && peakNearLocation(p, transcriptStarts[1][ind] + upstream, true, fragmentLength)>=0) { //sense direction
						nPeakLocs5pSenseHist[p]++;
						found5pSense = true;
						if (peakCons!=null && peakCons.containsKey(p) && peakCons.get(p)) { nPeakLocs5pSenseConsHist[p]++; }
					    }
					    if (!found5pAntiSense && peakNearLocation(p, transcriptStarts[2][ind] + upstream, false, fragmentLength)>=0) { //antisense direction
						nPeakLocs5pAntiSenseHist[p]++;
						found5pAntiSense = true;
						if (found5pSense) { nPeakLocs5pBiHist[p]++; }
						if (peakCons!=null && peakCons.containsKey(p) && peakCons.get(p)) { nPeakLocs5pAntiSenseConsHist[p]++; }
					    }
					}
					if (nMC>0) { //Do the shuffle control
					    nPeakLocs5pSenseShuffleHist[p] += numberOfPeaksNearRandomStart(p, transcriptStarts[1], upstream, true, fragmentLength, nMC);
					    nPeakLocs5pAntiSenseShuffleHist[p] += numberOfPeaksNearRandomStart(p, transcriptStarts[2], upstream, false, fragmentLength, nMC);
					    if (conservation) {
						nPeakLocs5pSenseShuffleConsHist[p] += numberOfPeaksNearRandomStart(p - upstream, transcriptStarts[1], upstream, pcs, true, fragmentLength, nMC);
						nPeakLocs5pSenseShuffleConsHist[p] += numberOfPeaksNearRandomStart(p - upstream, transcriptStarts[2], upstream, pcs, false, fragmentLength, nMC);
					    }
					}
					peakHeight[p] += ps.get(p);
					//Find the number of reads which belong to
					//the peaks
					nReadsInPeaks += ps.get(p);
					if (inConvergentRegion(true, p - upstream)) { peakInConvergentUpstream++; }
					if (inConvergentRegion(false, p - upstream)) { peakInConvergentDownstream++; }
				    }
				    /////////////////////////////////
				    //Check the overlap of peaks and ends in
				    //convergent regions
				    if (ind>=0 && inConvergentRegion(true, transcriptStarts[1][ind])) { 
					nSenseInConvergentUpstream++; 
					nGenesWithPeakAnd5pEndInConvergentUpstream += peakInConvergentUpstream>0 ? 1 : 0;
				    }
				    if (ind>=0 && inConvergentRegion(false, transcriptStarts[2][ind])) { 
					nAntiSenseInConvergentDownstream++;
					nGenesWithPeakAnd5pEndInConvergentDownstream += peakInConvergentDownstream>0 ? 1 : 0;
				    }
				    nGenesWithPeakInConvergentUpstream += peakInConvergentUpstream>0 ? 1 : 0;
				    nGenesWithPeakInConvergentDownstream += peakInConvergentDownstream>0 ? 1 : 0;
				    //Get the distribution of the number of peaks per TSS
				    peaksPerTSS.put(ps.size(), peaksPerTSS.containsKey(ps.size()) ? peaksPerTSS.get(ps.size()) + 1 : 1);
				    peaksPerTSSConvergentRegions.put(peakInConvergentDownstream + peakInConvergentUpstream, peaksPerTSSConvergentRegions.containsKey(peakInConvergentDownstream + peakInConvergentUpstream) ? peaksPerTSSConvergentRegions.get(peakInConvergentDownstream + peakInConvergentUpstream) + 1 : 1);
				}
				//Get a random non-genic region and find the number of pol2-peaks
				peaksPerTSSnonGenicMC = getRandomRegionPeaksHist(reads, readsInput, downstream, upstream, fragmentLength, nTSS[i-1], geneInfo[i-1], true, minPeakDist, minPeakHeight, peaksPerTSSnonGenicMC);
				peaksPerTSSgenicMC = getRandomRegionPeaksHist(reads, readsInput, downstream, upstream, fragmentLength, nTSS[i-1], geneInfo[i-1], false, minPeakDist, minPeakHeight, peaksPerTSSgenicMC);
				//Find the total number of reads in the region
				for (int j=0; j<hists.length; j++) {
				    for (int l=0; l<hists[j].length; l++)
					nReadsInRegion += hists[j][l];
				}
			    }
			}
		    }
		    //Add them all up and save to file
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    for (int j=0; j<histsMean[i - 1].length; j++) {
				double prop = ((double)nTSS[i-1])/sumTSS;
				sumHistMean[k][j] += histsMean[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/nTotalReads.get(fileNames.elementAt(k)) - (inputFileName!=null ? (histsMeanInput[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR))/nTotalReads.get(inputFileName) : 0);
				//sumHistVar[k][j] += histsVar[i - 1][j]*prop*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR*(double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/(nTotalReads.get(fileNames.elementAt(k))*nTotalReads.get(fileNames.elementAt(k)));
			    }
			}
		    }
		}
		for (int i=1; i<=nChr; i++) {
		    if (chr==-1 || i==chr) { 
			for (int j=0; j<histsMean[i - 1].length; j++) {
			    double prop = ((double)nTSS[i-1])/sumTSS;
			    if (gc && gcHistsMean[i-1]!=null) {
				sumGCMean[j] += gcHistsMean[i-1][j]*prop;
				sumGCVar[j] += gcHistsMean[i-1][j]*prop*prop;
			    }
			    if (conservation && consHistsMean[i-1]!=null) {
				sumConsMean[j] += consHistsMean[i-1][j]*prop;
				sumConsVar[j] += consHistsMean[i-1][j]*prop*prop;
			    }
			}
		    }
		}
		String saveFileName = null;
		if (fileNames.size()==1)
		    saveFileName = fileName + "_combined_" + feature + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (inputFileName!=null ? "_input" : "") + (chr!=-1 ? "_" + chr : "") + ".txt";
		else
		    saveFileName = fileName.substring(0, fileName.lastIndexOf(".")) + "_combined_" + feature + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (inputFileName!=null ? "_input" : "") + (chr!=-1 ? "_" + chr : "") + ".txt";
		System.out.println("Saving to " + saveFileName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		if (fileNames.size()==1) {
		    String chipName = fileName.substring(fileName.lastIndexOf("/") + 1);
		    bw.write("#Bin\tMean_" + chipName + "\tStd_" + chipName);
		}
		else {
		    bw.write("#pos");
		    for (String fName : shortNames) 
			bw.write("\t" + fName);
		}
		if (gc)
		    bw.write("\tMean_GC\tStd_GC");
		if (conservation)
		    bw.write("\tMean_cons\tStd_cons");
		if (fileNames.size()==1) {
		    bw.write("\tMean_height\t#peaks");
		    if (conservation)
			bw.write("\t#peaks_cons");
		    bw.write("\t#peaks_5p_S");
		    bw.write("\t#peaks_5p_AS");
		    bw.write("\t#peaks_5p_Bi");
		    if (conservation)
			bw.write("\t#peaks_5p_S_cons\t#peaks_5p_AS_cons");
		    if (nMC>0) {
			bw.write("\tpeaks_5p_S_MC\tpeaks_5p_AS_MC");
			if (conservation)
			    bw.write("\tpeaks_5p_S_MC_cons\tpeaks_5p_AS_MC_cons");
		    }
		}
		bw.write("\n");
		for (int i=0; i<sumHistMean[0].length; i++) {
		    bw.write("" + (-upstream + i));
		    if (fileNames.size()==1) 
			bw.write("\t" + format.format(sumHistMean[0][i]) + "\t" + format.format(Math.sqrt(sumHistVar[0][i])));
		    for (int j=0; j<fileNames.size(); j++)
			bw.write("\t" + format.format(sumHistMean[j][i]));
		    if (gc)
			bw.write("\t" + format.format(sumGCMean[i]) + "\t" + format.format(Math.sqrt(sumGCVar[i])));
		    if (conservation)
			bw.write("\t" + format.format(sumConsMean[i]) + "\t" + format.format(Math.sqrt(sumConsVar[i])));
		    if (fileNames.size()==1) {
			bw.write("\t" + (nPeakLocsHist[i]>0 ? format.format((double)peakHeight[i]/nPeakLocsHist[i]) : "0"));
			bw.write("\t" + nPeakLocsHist[i]);
			if (conservation) { bw.write("\t" + nPeakLocsConsHist[i]); }
			bw.write("\t" + nPeakLocs5pSenseHist[i]);
			bw.write("\t" + nPeakLocs5pAntiSenseHist[i]);
			bw.write("\t" + nPeakLocs5pBiHist[i]);
			if (conservation) { 
			    bw.write("\t" + nPeakLocs5pSenseConsHist[i]);
			    bw.write("\t" + nPeakLocs5pAntiSenseConsHist[i]);
			}
			if (nMC>0) {
			    bw.write("\t" + format.format((double)nPeakLocs5pSenseShuffleHist[i]/nMC));
			    bw.write("\t" + format.format((double)nPeakLocs5pAntiSenseShuffleHist[i]/nMC));
			    if (conservation) { 
				bw.write("\t" + format.format((double)nPeakLocs5pSenseShuffleConsHist[i]/nMC)); 
				bw.write("\t" + format.format((double)nPeakLocs5pAntiSenseShuffleConsHist[i]/nMC)); 
			    }
			}
		    }
		    bw.write("\n");
		}
		bw.close();
		System.out.println("Found " + (nReadsInRegion/120) + " reads in the region and " + nReadsInPeaks + " reads in peaks, the ratio is " + format.format((double)nReadsInPeaks/(nReadsInRegion/120)));
		System.out.println("The background denisty is " + format.format((((double)(nReadsInRegion/120) - nReadsInPeaks))/(sumTSS*(upstream + downstream + 1)/1000)) + " reads per kb\n/////////////////////////////////");
		/////////////////////////////////
		//Save the distribution of peaks near TSSs
		System.out.println("Found " + nGenesWithPeakInConvergentUpstream + " genes with peaks in convergent upstream region and " + nGenesWithPeakInConvergentDownstream + " genes with peaks in convergent downstream region.");
		System.out.println("Found " + nSenseInConvergentUpstream + " genes with 5' ends in convergent upstream region and " + nAntiSenseInConvergentDownstream + " genes with 5'-ends in convergent downstream region.");
		System.out.println("Found " + nGenesWithPeakAnd5pEndInConvergentUpstream + " genes with peaks and 5'-ends in convergent upstream region and " + nGenesWithPeakAnd5pEndInConvergentDownstream + " genes with peaks and 5'-ends in convergent downstream region.\n/////////////////////////////////");
		
		System.out.println("Found " + nPeaksInUpstreamConvergentRegion + " peaks in the upstream convergent region and " + nPeaksInDownstreamConvergentRegion + " peaks in the downstream convergent region out of a total of " + nPeaks);
		System.out.println("Found " + nPeaksNearStartSense + " peaks near sense transcripts and " + nPeaksNearStartAntiSense + " peaks near anti-sense out of a total of " + (nPeaksInUpstreamConvergentRegion + nPeaksInDownstreamConvergentRegion));
		System.out.println("Found " + nGenesWithPeakAndTranscriptInConvergentUpstream + " genes with both convergent upstream transcripts and peaks and " + nGenesWithPeakAndTranscriptInConvergentDownstream + " genes with both convergent downstream transcripts and peaks.");
		System.out.println("Found " + nPeaksNearCageStartSense + " peaks near sense CAGE starts and " + nPeaksNearCageStartAntiSense + " peaks near anti-sense CAGE starts out of " + nPeaksInUpstreamConvergentRegionWithCageData + " and " + nPeaksInDownstreamConvergentRegionWithCageData);
		System.out.println("Found " + nGenesWithPeakAndCageInConvergentUpstream + " genes with both convergent upstream Cage and peaks and " + nGenesWithPeakAndCageInConvergentDownstream + " genes with both convergent downstream Cage and peaks.");
		/////////////////////////////////
		if (fileNames.size()==1) {
		    saveFileName = fileName + "_peaks_near_TSS_dist" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt";
		    System.out.println("Saving histogram to " + saveFileName);
		    bw = new BufferedWriter(new FileWriter(saveFileName));
		    bw.write("#n\tTSS\tTSS_conv\tnon-genic\tgenic\n");
		    for (int k = 0; k<=Math.max(peaksPerTSS.lastKey(), Math.max(peaksPerTSSnonGenicMC.lastKey(), peaksPerTSSgenicMC.lastKey())); k++) 
			bw.write(k + "\t" + (peaksPerTSS.containsKey(k) ? peaksPerTSS.get(k) : 0) + "\t" + (peaksPerTSSConvergentRegions.containsKey(k) ? peaksPerTSSConvergentRegions.get(k) : 0) + "\t" + (peaksPerTSSnonGenicMC.containsKey(k) ? peaksPerTSSnonGenicMC.get(k) : 0) + "\t" + (peaksPerTSSgenicMC.containsKey(k) ? peaksPerTSSgenicMC.get(k) : 0) + "\n");
		    bw.close();
		}
		/////////////////////////////////
		if (conservation)
		    System.out.println("Found " + nPeaksNearStartSenseCons + " conserved peaks near sense transcripts and " + nPeaksNearStartAntiSenseCons + " conserved peaks near anti-sense out of a total of " + nPeaksConserved);
		if (conservation)
		    System.out.println("Found " + nPeaksNearCageStartSenseCons + " conserved peaks near CAGE sense starts and " + nPeaksNearCageStartAntiSenseCons + " conserved peaks near CAGE anti-sense out of a total of " + nPeaksConserved);
		if (nMC>0) {
		    saveShuffleHistogram(nPeaksNearStartSenseHist, fileName + "_tss_shuffle_s" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
		    saveShuffleHistogram(nPeaksNearStartAntiSenseHist, fileName + "_tss_shuffle_as" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
		    if (conservation) {
			saveShuffleHistogram(nPeaksNearStartSenseConsHist, fileName + "_tss_shuffle_s_cons" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
			saveShuffleHistogram(nPeaksNearStartAntiSenseConsHist, fileName + "_tss_shuffle_as_cons" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
		    }
		    //Repeat for CAGE results
		    saveShuffleHistogram(nPeaksNearCageStartSenseHist, fileName + "_tss_cage_shuffle_s" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
		    saveShuffleHistogram(nPeaksNearCageStartAntiSenseHist, fileName + "_tss_cage_shuffle_as" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
		    if (conservation) {
			saveShuffleHistogram(nPeaksNearCageStartSenseConsHist, fileName + "_tss_cage_shuffle_s_cons" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
			saveShuffleHistogram(nPeaksNearCageStartAntiSenseConsHist, fileName + "_tss_cage_shuffle_as_cons" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt");
		    }
		}
	    }
	    //////////////////////////////////////////////////////////////////
	    else if (feature.contains("convergent")) { //Find the densities surrounding the putative 5'-ends.
		boolean convergentUpstream = feature.contains("up");
		int transcriptStarts[][] = parseInitiationFile(transcriptStartsFileName, isolatedIDs, true), n5pEnds[] = new int[nChr], sum5pEnds = 0;
		double sumHistMeanMC[] = new double[numberOfBins(upstream, downstream, 1)];
		for (int k=0; k<fileNames.size(); k++) {
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    //Find the absolute locations of the 5p-ends
			    int ids[] = getIDs(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    int tssVec[] = getTSSs(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    System.out.println("Found " + tssVec.length + " genes.");
			    boolean strandVec[] = getStrands(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    //key is the absolute location and the value is the gene
			    TreeMap<Integer, RefSeqGene> endLocs = new TreeMap<Integer, RefSeqGene>();
			    //Key is the absolute location and the value is the distance to the TSS
			    TreeMap<Integer, Integer> endsTSSDist = new TreeMap<Integer, Integer>();
			    for (int j=0; j<ids.length; j++) {
				int ind = Arrays.binarySearch(transcriptStarts[0], ids[j]), col = convergentUpstream ? 1 : 2;
				if (ind>=0 && transcriptStarts[col][ind]!=Integer.MAX_VALUE && transcriptStarts[col][ind]!=Integer.MIN_VALUE && inConvergentRegion(convergentUpstream, transcriptStarts[col][ind])) {
				    int start5p = tssVec[j] + (strandVec[j] ? transcriptStarts[col][ind] : -transcriptStarts[col][ind]);
				    endLocs.put(start5p, geneInfo[i-1].getGene(ids[j]));
				    endsTSSDist.put(start5p, transcriptStarts[col][ind]);
				}
			    }
			    //find the total number of 5p-ends
			    n5pEnds[i-1] = endLocs.size();
			    sum5pEnds += endLocs.size();
			    //Get the GC distribution 
			    if (gc) { gcHistsMean[i-1] = getGCHistogram(i, orchestra, endLocs, endsTSSDist, upstream, downstream, convergentUpstream); }
			    //Get the conservation scores
			    double pcs[][] = null;
			    //Extract a larger chunk for all genes, then
			    //find the smaller region which surrounds the
			    //putative 5'-end
			    if (conservation) {
				PhastConsParser verterbateCons = new PhastConsParser(i, false, orchestra, species);
				pcs = new double[endLocs.size()][numberOfBins(downstream, upstream, 1)];
				Vector<Integer> ends = new Vector<Integer>(endLocs.keySet());
				Collections.sort(ends);
				for (int j=0; j<ends.size(); j++) {
				    pcs[j] = endLocs.get(ends.elementAt(j)).getStrand() ? verterbateCons.findScores(ends.elementAt(j) - upstream, ends.elementAt(j) + downstream + 1) : flipArray(verterbateCons.findScores(ends.elementAt(j) - downstream, ends.elementAt(j) + upstream + 1));
				    //Make sure that we only get the stuff in the convergent region
				    for (int l=0; l<pcs[j].length; l++) {
					RefSeqGene g = endLocs.get(ends.elementAt(j));
					if (!inConvergentRegion(convergentUpstream, endsTSSDist.get(ends.elementAt(j)) + l - upstream) || (!convergentUpstream && g.overlapsExon(ends.elementAt(j) + (g.getStrand() ? -upstream + l : downstream - l)))) //Find out if we are in the convergent region or in an exon
					    pcs[j][l] = -1;
					//if ((!convergentUpstream && g.overlapsExon(ends.elementAt(j) + (g.getStrand() ? -upstream + l : downstream - l))))
					//  System.out.println("Overlaps exon at " + l + ", " + ((g.getStrand() ? -upstream + l : downstream - l) + ends.elementAt(j)) + " for " + g.toString() + " with end " + ends.elementAt(j));
				    }
				}
				consHistsMean[i-1] = getConservationHistogram(pcs);
			    }
			    /////////////////////////////////
			    //Get the profiles surrounding the 5p-ends
			    ABIRead reads[] = PeakFinder2.parseFiles(i, fileName, fragmentLength, true, 0, qualityScore);			    
			    //Get the histograms surrounding the peaks
			    int hists[][] = getHistograms(i, endLocs, endsTSSDist, convergentUpstream, reads, null, downstream, upstream, fragmentLength, false);
			    //Get the average histogram
			    double h[][] = findMeanReadHistograms(hists, numberOfBins(downstream, upstream, 1));
			    histsMean[i-1] = h[0];
			    histsVar[i-1] = h[1];
			    if (nMC>0) { //Shuffle the 5'-ends with respect to the promoters
				TreeMap<Integer, RefSeqGene> endLocsMC = new TreeMap<Integer, RefSeqGene>();
				TreeMap<Integer, Integer> endsTSSDistMC = new TreeMap<Integer, Integer>();
				for (int d : endsTSSDist.values()) {
				    //Draw a 5'-end, assign it to a random gene 
				    int randInd = random.nextInt(ids.length);
				    endLocsMC.put(tssVec[randInd] + d, geneInfo[i-1].getGene(ids[randInd]));
				    endsTSSDistMC.put(tssVec[randInd] + d, d);
				}
				int histsMC[][] = getHistograms(i, endLocsMC, endsTSSDistMC, convergentUpstream, reads, null, downstream, upstream, fragmentLength, false);
				//Get the average histogram
				double hMC[][] = findMeanReadHistograms(histsMC, numberOfBins(downstream, upstream, 1));
				histsMeanMC[i-1] = hMC[0];
				//histsVarMC[i-1] = hMC[1];
				if (gc) { gcHistsMeanMC[i-1] = getGCHistogram(i, orchestra, endLocsMC, endsTSSDistMC, upstream, downstream, convergentUpstream); }
				if (conservation) {
				    PhastConsParser verterbateCons = new PhastConsParser(i, false, orchestra, species);
				    double pcsMC[][] = new double[endLocsMC.size()][numberOfBins(downstream, upstream, 1)];
				    Vector<Integer> endsMC = new Vector<Integer>(endLocsMC.keySet());
				    Collections.sort(endsMC);
				    for (int j=0; j<endsMC.size(); j++) {
					pcsMC[j] = endLocsMC.get(endsMC.elementAt(j)).getStrand() ? verterbateCons.findScores(endsMC.elementAt(j) - upstream, endsMC.elementAt(j) + downstream + 1) : flipArray(verterbateCons.findScores(endsMC.elementAt(j) - downstream, endsMC.elementAt(j) + upstream + 1));
					//Make sure that we only get the stuff in the convergent region
					for (int l=0; l<pcsMC[j].length; l++) {
					    RefSeqGene g = endLocsMC.get(endsMC.elementAt(j));
					    if (!inConvergentRegion(convergentUpstream, endsTSSDistMC.get(endsMC.elementAt(j)) + l - upstream) || (!convergentUpstream && g.overlapsExon(endsMC.elementAt(j) + (g.getStrand() ? -upstream + l : downstream - l)))) //Find out if we are in the convergent region
						pcsMC[j][l] = -1;
					}
					consHistsMeanMC[i-1] = getConservationHistogram(pcsMC);
				    }
				}
			    }
			}
		    }
		}
		//Add them all up and save to file
		for (int k=0; k<fileNames.size(); k++) {
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    for (int j=0; j<histsMean[i - 1].length; j++) {
				double prop = ((double)n5pEnds[i-1])/sum5pEnds;
				sumHistMean[k][j] += histsMean[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/nTotalReads.get(fileNames.elementAt(k));
				//sumHistVar[k][j] += histsVar[i - 1][j]*prop*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/nTotalReads.get(fileNames.elementAt(k))*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/nTotalReads.get(fileNames.elementAt(k));
			    }
			}
		    }
		}
		for (int i=1; i<=nChr; i++) {
		    if (chr==-1 || i==chr) { 
			for (int j=0; j<histsMean[i - 1].length; j++) {
			    double prop = ((double)n5pEnds[i-1])/sum5pEnds;
			    if (gc && gcHistsMean[i-1]!=null) {
				sumGCMean[j] += gcHistsMean[i-1][j]*prop;
				sumGCVar[j] += gcHistsMean[i-1][j]*prop*prop;
			    }
			    if (conservation && consHistsMean[i-1]!=null) {
				sumConsMean[j] += consHistsMean[i-1][j]*prop;
				sumConsVar[j] += consHistsMean[i-1][j]*prop*prop;
			    }
			    if (nMC>0) {
				sumHistMeanMC[j] += histsMeanMC[i - 1][j]*prop;
				if (gc && gcHistsMeanMC[i-1]!=null) 
				    sumGCMeanMC[j] += gcHistsMeanMC[i-1][j]*prop;
				if (conservation && consHistsMeanMC[i-1]!=null)
				    sumConsMeanMC[j] += consHistsMeanMC[i-1][j]*prop;
			    }
			}
		    }
		}
		String saveFileName = fileName + "_combined_5p_" + (convergentUpstream ? "up" : "down") + (isolated ? "_isolated" : "") + (chr==-1 ? "" : "_" + chr) + ".txt";
		System.out.println("Saving to " + saveFileName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		String chipName = fileName.substring(fileName.lastIndexOf("/") + 1);
		bw.write("#Bin\tMean_" + chipName + "\tStd_" + chipName);
		if (gc) { bw.write("\tmean_GC\tstd_dev_GC"); }
		if (conservation) { bw.write("\tmean_cons\tstd_dev_cons"); }
		if (nMC>0) { 
		    bw.write("\tMean_" + chipName + "_MC");
		    if (gc) { bw.write("\tmean_GC_MC"); }
		    if (conservation) { bw.write("\tmean_cons_MC"); }
		}
		bw.write("\n");
		for (int i=0; i<sumHistMean[0].length; i++) {
		    bw.write((-upstream + i) + "");
		    bw.write("\t" + format.format(sumHistMean[0][i]) + "\t" + format.format(Math.sqrt(sumHistVar[0][i])));
		    if (gc)
			bw.write("\t" + format.format(sumGCMean[i]) + "\t" + format.format(Math.sqrt(sumGCVar[i])));
		    if (conservation)
			bw.write("\t" + format.format(sumConsMean[i]) + "\t" + format.format(Math.sqrt(sumConsVar[i])));
		    if (nMC>0) {
			bw.write("\t" + format.format(sumHistMeanMC[i]));
			if (gc)
			    bw.write("\t" + format.format(sumGCMeanMC[i]));
			if (conservation)
			    bw.write("\t" + format.format(sumConsMeanMC[i]));
		    }
		    bw.write("\n");
		}
		bw.close();
	    }
	    //////////////////////////////////////////////////////////////////
	    else if (feature.equals("TES")) {
		//Get the profiles surrounding the TSSs
		nTotalReads = calculateTotalReads(fileNames, inputFileName, nChr, fragmentLength, qualityScore);
		for (int k=0; k<fileNames.size(); k++) {
		    nTotalReads.put(fileNames.elementAt(k), 0);
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) {
			    boolean saveHist = false;
			    int ids[] = geneInfo[i-1].getIDs();
			    //System.out.println("Getting info for chr " + i + " for " + ids.length + " genes.");
			    /////////////////////////////////
			    //Load the reads
			    ABIRead reads[] = PeakFinder2.parseFiles(i, fileNames.elementAt(k), fragmentLength, true, 0, qualityScore);
			    nTotalReads.put(fileNames.elementAt(k), nTotalReads.get(fileNames.elementAt(k)) + reads.length);
			    ABIRead readsInput[] = inputFileName!=null ? PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore) : null;
			    int tssVec[] = getTESs(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    boolean strandVec[] = getStrands(geneInfo[i-1], geneIDsKeep, isolatedIDs);
			    //Get the histograms surrounding the TSSs
			    //int hists[][] = getHistograms(i, tssVec, strandVec, reads, null, downstream, upstream, fragmentLength, saveHist);
			    String histSaveName = null;
			    if (!noHists)
				histSaveName = TES_HISTOGRAM_PREFIX + "_" + shortNames.elementAt(k) + HISTOGRAM_SUFFIX + (readsInput==null ? "" : "input_") + i + ".txt";
			    int hists[][] = getHistograms(i, tssVec, strandVec, reads, readsInput, downstream, upstream, fragmentLength, histSaveName, (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(fileNames.elementAt(k)), inputFileName==null ? 0 : (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(inputFileName));
			    int histsInput[][] = readsInput!=null ? getHistograms(i, tssVec, strandVec, reads, readsInput, downstream, upstream, fragmentLength, saveHist) : hists;
			    //Get the average histogram
			    double h[][] = findMeanReadHistograms(hists, numberOfBins(downstream, upstream, 1));
			    histsMean[i-1] = h[0];
			    histsVar[i-1] = h[1];
			    if (k==0 && readsInput!=null) 
				histsMeanInput[i-1] = findMeanReadHistograms(getHistograms(i, tssVec, strandVec, readsInput, null, downstream, upstream, fragmentLength, false), numberOfBins(downstream, upstream, 1))[0];
			    Vector<TreeMap<Integer, Integer>> peakLocs = findAllPeaks(hists, histsInput, minPeakDist, minPeakHeight);
			    //Get the GC distribution 
			    if (gc && k==0) { gcHistsMean[i-1] = getGCHistogram(i, orchestra, tssVec, strandVec, upstream, downstream); }
			    //Get the conservation scores
			    double pcs[][] = null;
			    Vector<TreeMap<Integer, Boolean>> peakConserved = null;
			    if (conservation && k==0) {
				PhastConsParser verterbateCons = new PhastConsParser(i, false, orchestra, species);
				pcs = new double[tssVec.length][numberOfBins(downstream, upstream, 1)];
				for (int j=0; j<tssVec.length; j++) {
				    if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, ids[j])>=0)
					pcs[j] = strandVec[j] ? verterbateCons.findScores(tssVec[j] - upstream, tssVec[j] + downstream + 1) : flipArray(verterbateCons.findScores(tssVec[j] - downstream, tssVec[j] + upstream + 1));
				}
				consHistsMean[i-1] = getConservationHistogram(pcs);
			    }
			}
		    }
		    //Add them all up and save to file
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    for (int j=0; j<histsMean[i - 1].length; j++) {
				double prop = ((double)nTSS[i-1])/sumTSS;
				sumHistMean[k][j] += histsMean[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/nTotalReads.get(fileNames.elementAt(k)) - (inputFileName!=null ? (histsMeanInput[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR))/nTotalReads.get(inputFileName) : 0);
				//sumHistVar[k][j] += histsVar[i - 1][j]*prop*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR*(double)CreateBedFiles.READS_NORMALIZATION_FACTOR)/(nTotalReads.get(fileNames.elementAt(k))*nTotalReads.get(fileNames.elementAt(k)));
			    }
			}
		    }
		}
		for (int i=1; i<=nChr; i++) {
		    if (chr==-1 || i==chr) { 
			for (int j=0; j<histsMean[i - 1].length; j++) {
			    double prop = ((double)nTSS[i-1])/sumTSS;
			    if (gc && gcHistsMean[i-1]!=null) {
				sumGCMean[j] += gcHistsMean[i-1][j]*prop;
				sumGCVar[j] += gcHistsMean[i-1][j]*prop*prop;
			    }
			    if (conservation && consHistsMean[i-1]!=null) {
				sumConsMean[j] += consHistsMean[i-1][j]*prop;
				sumConsVar[j] += consHistsMean[i-1][j]*prop*prop;
			    }
			}
		    }
		}
		String saveFileName = null;
		if (fileNames.size()==1)
		    saveFileName = fileName + "_combined_" + feature + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (inputFileName!=null ? "_input" : "") + (chr!=-1 ? "_" + chr : "") + ".txt";
		else
		    saveFileName = fileName.substring(0, fileName.lastIndexOf(".")) + "_combined_" + feature + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (inputFileName!=null ? "_input" : "") + (chr!=-1 ? "_" + chr : "") + ".txt";
		System.out.println("Saving to " + saveFileName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		if (fileNames.size()==1) {
		    String chipName = fileName.substring(fileName.lastIndexOf("/") + 1);
		    bw.write("#Bin\tMean_" + chipName + "\tStd_" + chipName);
		}
		else {
		    bw.write("#pos");
		    for (String fName : shortNames) 
			bw.write("\t" + fName);
		}
		if (gc)
		    bw.write("\tMean_GC\tStd_GC");
		if (conservation)
		    bw.write("\tMean_cons\tStd_cons");
		if (fileNames.size()==1) {
		    bw.write("\tMean_height\t#peaks");
		    if (conservation)
			bw.write("\t#peaks_cons");
		    bw.write("\t#peaks_5p_S");
		    bw.write("\t#peaks_5p_AS");
		    bw.write("\t#peaks_5p_Bi");
		    if (conservation)
			bw.write("\t#peaks_5p_S_cons\t#peaks_5p_AS_cons");
		    if (nMC>0) {
			bw.write("\tpeaks_5p_S_MC\tpeaks_5p_AS_MC");
			if (conservation)
			    bw.write("\tpeaks_5p_S_MC_cons\tpeaks_5p_AS_MC_cons");
		    }
		}
		bw.write("\n");
		for (int i=0; i<sumHistMean[0].length; i++) {
		    bw.write("" + (-upstream + i));
		    if (fileNames.size()==1) 
			bw.write("\t" + format.format(sumHistMean[0][i]) + "\t" + format.format(Math.sqrt(sumHistVar[0][i])));
		    for (int j=0; j<fileNames.size(); j++)
			bw.write("\t" + format.format(sumHistMean[j][i]));
		    if (gc)
			bw.write("\t" + format.format(sumGCMean[i]) + "\t" + format.format(Math.sqrt(sumGCVar[i])));
		    if (conservation)
			bw.write("\t" + format.format(sumConsMean[i]) + "\t" + format.format(Math.sqrt(sumConsVar[i])));
		    bw.write("\n");
		}
	    }
	    //////////////////////////////////////////////////////////////////
	    else if (feature.contains("enhancers")) { //Find the features surrounding a set of peaks  sitting around TSSs
		//We want to be able to look at more than one 
		int timePoints[] = new int[]{0, 1, 6};
		double histsRNAForwardMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], histsRNAForwardVar[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], histsRNAForwardSmallMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)];
		double histsRNAReverseMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], histsRNAReverseVar[][] = new double[nChr][numberOfBins(downstream, upstream, 1)], histsRNAReverseSmallMean[][] = new double[nChr][numberOfBins(downstream, upstream, 1)];
		int nEnhancers[] = new int[nChr], sumEnhancers = 0;
		Vector<int[]> peakVecs = new Vector<int[]>();
		Vector<Enhancer> enhancers = feature.contains("rand") ? parseRandFile(feature) : parseEnhancerFile(feature);
		Collections.sort(enhancers);
		//Keep track of howm many reads there are for each individual loci
		TreeMap<String, TreeMap<Enhancer, Integer>> tfPeakReads = new TreeMap<String, TreeMap<Enhancer, Integer>>();
		/////////////////////////////////
		if (inputFileName!=null) { nTotalReads.put(inputFileName, 0); }
		for (int i=1; i<=nChr; i++) {
		    peakVecs.addElement(new int[0]);
		    if (chr==-1 || i==chr) { 
			peakVecs.setElementAt(getEnhancerLocations(enhancers, i, enhancerSubset, geneInfo), i-1);
			//find the total number of enhancers
			nEnhancers[i-1] = peakVecs.elementAt(i-1).length;
			sumEnhancers += peakVecs.elementAt(i-1).length;
			//Get the GC distribution 
			if (gc) { gcHistsMean[i-1] = getGCHistogram(i, orchestra, peakVecs.elementAt(i-1), upstream, downstream); }
			//Get the conservation scores
			double pcs[][] = null;
			if (conservation) {
			    PhastConsParser verterbateCons = new PhastConsParser(i, false, orchestra, species);
			    pcs = new double[peakVecs.elementAt(i-1).length][numberOfBins(downstream, upstream, 1)];
			    for (int j=0; j<peakVecs.elementAt(i-1).length; j++) 
				pcs[j] = verterbateCons.findScores(peakVecs.elementAt(i-1)[j] - upstream, peakVecs.elementAt(i-1)[j] + downstream + 1);
			    consHistsMean[i-1] = getConservationHistogram(pcs);
			}
		    }
		}
		/////////////////////////////////
		//Get the profiles surrounding the enahncers
		for (int k=0; k<fileNames.size(); k++) {
		    nTotalReads.put(fileNames.elementAt(k), 0);
		    tfPeakReads.put(fileNames.elementAt(k), new TreeMap<Enhancer, Integer>());
		    for (int i=1; i<=nChr; i++) { //For normalization purposes, we want to consider all chromosomes
			ABIRead reads[] = PeakFinder2.parseFiles(i, fileNames.elementAt(k), fragmentLength, true, 0, qualityScore);			    
			nTotalReads.put(fileNames.elementAt(k), nTotalReads.get(fileNames.elementAt(k)) + reads.length);
			if (inputFileName!=null && k==0) 
			    nTotalReads.put(inputFileName, nTotalReads.get(inputFileName) + PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore).length);
		    }
		}
		//Get the histograms surrounding the peaks
		for (int k=0; k<fileNames.size(); k++) {
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    ABIRead reads[] = PeakFinder2.parseFiles(i, fileNames.elementAt(k), fragmentLength, true, 0, qualityScore);			    
			    ABIRead readsInput[] = inputFileName!=null ? PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore) : null;
			    System.out.println("Found " + nTotalReads.get(fileNames.elementAt(k)) + " reads for " + fileNames.elementAt(k) + " and " + nTotalReads.get(inputFileName) + " for input " + inputFileName);
			    int hists[][] = getHistograms(i, peakVecs.elementAt(i-1), reads, readsInput, downstream, upstream, fragmentLength, (noHists ? null : feature.substring(0, feature.lastIndexOf(".")) + "_" + shortNames.elementAt(k) + AGGREGATE_SUFFIX + (enhancerSubset.length()>0 ? enhancerSubset + "_" : "") + (readsInput==null ? "" : "input_") + "w_" + windowSize + "_" + i + ".txt"), (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(fileNames.elementAt(k)), inputFileName==null ? 0 : (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/nTotalReads.get(inputFileName), windowSize);
			    //Get the average histogram
			    double h[][] = findMeanReadHistograms(hists, numberOfBins(downstream, upstream, 1));
			    histsMean[i-1] = h[0];
			    histsVar[i-1] = h[1];
			    if (k==0 && readsInput!=null) 
				histsMeanInput[i-1] = findMeanReadHistograms(getHistograms(i, peakVecs.elementAt(i-1), readsInput, null, downstream, upstream, fragmentLength, null, inputFileName==null ? 0 : (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/*(double)PeakFinder2.CHROMOSOME_LENGTH[i - 1]/(PeakFinder2.TOTAL_CHROMOSOME_LENGTH*//nTotalReads.get(inputFileName), 0, 0), numberOfBins(downstream, upstream, 1))[0];
			}
		    }
		    //Add them all up
		    for (int i=1; i<=nChr; i++) {
			if (chr==-1 || i==chr) { 
			    for (int j=0; j<histsMean[i - 1].length; j++) {
				double prop = ((double)nEnhancers[i-1])/sumEnhancers;
				sumHistMean[k][j] += (histsMean[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR))/nTotalReads.get(fileNames.elementAt(k)) - (inputFileName!=null ? (histsMeanInput[i - 1][j]*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR))/nTotalReads.get(inputFileName) : 0);
				//sumHistVar[k][j] += (histsVar[i - 1][j]*prop*prop*((double)CreateBedFiles.READS_NORMALIZATION_FACTOR*(double)CreateBedFiles.READS_NORMALIZATION_FACTOR))/(nTotalReads.get(fileNames.elementAt(k))*nTotalReads.get(fileNames.elementAt(k)));
			    }
			}
		    }
		    System.out.println("Found " + nTotalReads.get(fileNames.elementAt(k)) + " reads for " + fileNames.elementAt(k));
		}
		//Add them all up and save to file
		for (int i=1; i<=nChr; i++) {
		    if (chr==-1 || i==chr) { 
			for (int j=0; j<histsMean[i - 1].length; j++) {
			    double prop = ((double)nEnhancers[i-1])/sumEnhancers;
			    if (gc && gcHistsMean[i-1]!=null) {
				sumGCMean[j] += gcHistsMean[i-1][j]*prop;
				sumGCVar[j] += gcHistsMean[i-1][j]*prop*prop;
			    }
			    if (conservation && consHistsMean[i-1]!=null) {
				sumConsMean[j] += consHistsMean[i-1][j]*prop;
				sumConsVar[j] += consHistsMean[i-1][j]*prop*prop;
			    }
			}
		    }
		}
		String saveFileName = feature.substring(0, feature.lastIndexOf(".")) + "_combined_peaks" + (fileNames.size()>1 ? "_" + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf("names") - 1) : "") +  (enhancerSubset.length()>0 ? "_" + enhancerSubset : "") + (inputFileName!=null ? "_input" : "") + (chr==-1 ? "" : "_" + chr) + ".txt";
		//String saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/" + FindEnhancerSites.ENHANCER_SUFFIX + "_combined_peaks" + (chr==-1 ? "" : "_" + chr) + ".txt";
		System.out.println("Saving to " + saveFileName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		bw.write("#pos");
		for (String fName : shortNames) 
		    bw.write("\t" + fName);
		if (gc) { bw.write("\tmean_GC\tstd_dev_GC"); }
		if (conservation) { bw.write("\tmean_cons\tstd_dev_cons"); }
		bw.write("\n");
		for (int i=0; i<sumHistMean[0].length; i++) {
		    bw.write((-upstream + i) + "");
		    for (int j=0; j<fileNames.size(); j++)
			bw.write("\t" + format.format(sumHistMean[j][i]));
		    if (gc)
			bw.write("\t" + format.format(sumGCMean[i]) + "\t" + format.format(Math.sqrt(sumGCVar[i])));
		    if (conservation)
			bw.write("\t" + format.format(sumConsMean[i]) + "\t" + format.format(Math.sqrt(sumConsVar[i])));
		    bw.write("\n");
		}
		bw.close();
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////
    //Functions for analyzing aligned enhancer sites

    public static Vector<Enhancer> parseRandFile(String fileName) {
	Vector<Enhancer> enhancers = new Vector<Enhancer>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //chr start end nForward nReverse
	    Pattern randPattern = Pattern.compile("^(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = randPattern.matcher(line);
		if (m.find()) 
		    enhancers.addElement(new Enhancer(Integer.parseInt(m.group(1)), Integer.parseInt(m.group(2))));
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading enhancers from " + fileName + ": " + e.toString());
	}
	return enhancers;
    }

    public static Vector<Enhancer> parseEnhancerFile(String fileName) { return parseEnhancerFile(fileName, "Mm"); }

    public static Vector<Enhancer> parseEnhancerFile(String fileName, String species) { return parseEnhancerFile(fileName, species, -1); }

    public static Vector<Enhancer> parseEnhancerFile(String fileName, int chr) { return parseEnhancerFile(fileName, "Mm", chr); }

    public static Vector<Enhancer> parseEnhancerFile(String fileName, String species, int chr) {
	Vector<Enhancer> enhancers = new Vector<Enhancer>();
	if (fileName!=null) {
	    try {
		boolean bed = fileName.endsWith(".bed");
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		//chr start end nForward nReverse
		Pattern enhancerPattern = Pattern.compile("^chr([\\d|X|Y]+):(\\d+)-(\\d+)\\s*(\\d*)\\s*(\\d*)");
		Pattern shortPattern = Pattern.compile("^chr([\\d|X|Y]+):(\\d+)");
		while (line!=null) {
		    Matcher m = enhancerPattern.matcher(line);
		    if (m.find()) {
			int c;
			if (m.group(1).equals("X"))
			    c = species.equalsIgnoreCase("Mm") ? 20 : 23;
			else if (m.group(1).equals("Y"))
			    c = species.equalsIgnoreCase("Mm") ? 21 : 24;
			else
			    c = Integer.parseInt(m.group(1));
			if (c==chr || chr==-1) {
			    if (bed || m.group(4)==null || m.group(4).equals(""))
				enhancers.addElement(new Enhancer(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3))));
			    else
				enhancers.addElement(new Enhancer(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4)), Integer.parseInt(m.group(5))));
			}
		    }
		    else { 
			m = shortPattern.matcher(line);
			if (m.find()) {
			    int c;
			    if (m.group(1).equals("X"))
				c = species.equalsIgnoreCase("Mm") ? 20 : 23;
			    else if (m.group(1).equals("Y"))
				c = species.equalsIgnoreCase("Mm") ? 21 : 24;
			    else
				c = Integer.parseInt(m.group(1));
			    if (c==chr || chr==-1) 
				enhancers.addElement(new Enhancer(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(2))));
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading enhancers from " + fileName + ": " + e.toString());
	    }
	}
	System.out.println("Found " + enhancers.size() + " enhancers");
	return enhancers;
    }

    public static int[] getEnhancerLocations(Vector<Enhancer> enhancers, int chr, String enhancerSubset, RefSeqGenes geneInfo[]) {
	int nEnhancersChr = 0;
	for (Enhancer e : enhancers) { 
	    if (enhancerOK(e, chr, enhancerSubset, geneInfo))
		nEnhancersChr += e.getChromosome()==chr ? 1 : 0; 
	}
	int peakLocs[] = new int[nEnhancersChr], i = 0;
	for (Enhancer e : enhancers) {
	    if (enhancerOK(e, chr, enhancerSubset, geneInfo))
		peakLocs[i++] = e.getCenter();
	}
	return peakLocs;
    }

    public static boolean enhancerOK(Enhancer e, int chr, String enhancerSubset, RefSeqGenes geneInfo[]) {
	if (e.getChromosome()==chr) {
	    if (enhancerSubset.equalsIgnoreCase(INTERGENIC_STR)) {
		//System.out.println("Found intergenic enhancer " + e.toString());
		return !geneInfo[chr-1].overlapsGene(e.getCenter());
	    }
	    else if (enhancerSubset.equalsIgnoreCase(INTRONIC_STR)) {
		//System.out.println("Found intronic enhancer " + e.toString());
		return geneInfo[chr-1].overlapsGene(e.getCenter());
	    }
	    return true;
	}
	return false;
    }

    public static int[] calculateBalance(int hists[][], int upstream, int fragmentLength, double pVal) {
	//For each enhancer, return -1 if there are more reads in the
	//upstream regions, 0 if they are equally distributed and 1 if
	//there are more downstream. Use a binomial test to determine
	//if the imbalance is significant.
	int balance[] = new int[hists.length], nBalanced = 0;
	for (int i=0; i<hists.length; i++) {
	    int nReadsUp = 0, nReadsDown = 0;
	    for (int j=0; j<hists[i].length; j++) {
		if (j<upstream)
		    nReadsUp += hists[i][j];
		else
		    nReadsDown += hists[i][j];
	    }
	    nReadsUp /= fragmentLength;
	    nReadsDown /= fragmentLength;
	    //Do the binomial test
	    BinomialDistributionImpl binDist = new BinomialDistributionImpl(nReadsDown + nReadsUp, .5);
	    try {
		if (binDist.cumulativeProbability(nReadsUp)<pVal/2)
		    balance[i] = 1;
		else if (binDist.cumulativeProbability(nReadsUp)>1-pVal/2)
		    balance[i] = -1;
		else {
		    balance[i] = 0;
		    nBalanced++;
		}
		//System.out.println("Found " + nReadsUp + " reads up, " + nReadsDown + " reads down and p-val " + format.format(binDist.cumulativeProbability(nReadsUp)));
	    }
	    catch (MathException e) {
		System.err.println("Error with binomial dist.");
	    }
	}
	System.out.println("Found " + nBalanced + " balanced out of " + hists.length);
	return balance;
    }

    public static int[][] getTranscripts(int chr, int peakVec[], ABIRead allReads[], boolean forward, int downstream, int upstream, int minTranscriptReadDistance) {
	//Find the extent of the transcript in this region by
	//connecting the 5'-most and 3'most RNA reads.
	int transcripts[][] = new int[peakVec.length][2];
	for (int i=0; i<peakVec.length; i++) {
	    //Extract the reads
	    ABIRead reads[] = getReads(chr, peakVec[i], downstream, upstream, allReads);
	    if (reads.length>1) {
		transcripts[i][0] = -peakVec[i] + (forward ? reads[0].getPosition() : reads[reads.length - 1].getPosition());// + (forward ? 0 : FindNongenicTranscripts.MRNA_READ_LENGTH));
		int j = forward ? 1 : reads.length-2;
		if (forward) {
		    while (j<reads.length && reads[j].getPosition() - reads[j-1].getPosition()>minTranscriptReadDistance) 
			j++;
		}
		else {
		    while (j>0 && reads[j+1].getPosition() - reads[j].getPosition()>minTranscriptReadDistance) 
			j--;
		}
		if (j==reads.length && forward)
		    transcripts[i][1] = downstream;
		else if (j==0 && !forward)
		    transcripts[i][1] = -upstream;
		else
		    transcripts[i][1] = -peakVec[i] + reads[j].getPosition();// + (forward ? FindNongenicTranscripts.MRNA_READ_LENGTH : 0);
		//System.out.println("Found transcript from " + transcripts[i][0] + " to " + transcripts[i][1] + " containing " + reads.length + " reads at " + peakVec[i]);
	    }
	    else {
		//System.out.println("Could not find any reads at " + peakVec[i]);
	    }
	}
	return transcripts;
    }
    
    //////////////////////////////////////////////////////////////////
    
    public static int numberOfBins(int up, int down, int step) { return (1 + up + down)/step; }

    public static int numberOfBins(int up, int down) { return numberOfBins(up, down, 1); }

    //////////////////////////////////////////////////////////////////

    public static int[] getShuffledStarts(int nPeaksNearStart[], Vector<TreeMap<Integer, Integer>> peakLocs, int ids[], int upstream, int fragmentLength, int transcriptStarts[][], boolean sense) {
	return getShuffledStarts(nPeaksNearStart, peakLocs, null, ids, upstream, fragmentLength, transcriptStarts, sense, null); 
    }

    public static int[] getShuffledStarts(int nPeaksNearStart[], Vector<TreeMap<Integer, Integer>> peakLocs, Vector<TreeMap<Integer, Boolean>> peakConserved, int ids[], int upstream, int fragmentLength, int transcriptStarts[][], boolean sense, double pcs[][]) {
	//Returns a vector containing all the nMC simulations of how many peaks are near start sites.
	for (int j=0; j<nPeaksNearStart.length; j++) {
	    int nPeaksNear = pcs!=null ? numberOfPeaksNearTranscriptStarts(peakLocs, peakConserved, ids, upstream, fragmentLength, transcriptStarts, sense, pcs, true) : numberOfPeaksNearTranscriptStarts(peakLocs, ids, upstream, fragmentLength, transcriptStarts, sense, true);
	    nPeaksNearStart[j] += nPeaksNear;
	}
	return nPeaksNearStart;
    }

    public static TreeMap<Integer, Integer> getShuffledStartsHistogram(int nPeaksNearStart[]) {
	//Transform the vector into a histogram
	TreeMap<Integer, Integer> nPeaksNearStartHist = new TreeMap<Integer, Integer>();
	for (int j=0; j<nPeaksNearStart.length; j++) 
	    nPeaksNearStartHist.put(nPeaksNearStart[j], nPeaksNearStartHist.containsKey(nPeaksNearStart[j]) ? nPeaksNearStartHist.get(nPeaksNearStart[j]) + 1 : 1);
	return nPeaksNearStartHist;
    }

    public static TreeMap<Integer, Integer> getRandomRegionPeaksHist(ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, int nMC, RefSeqGenes geneInfo, boolean nongenic, int minPeakDist, int minPeakHeight, TreeMap<Integer, Integer> nPeaksHist) {
	//Find nMC regions and see how many Pol2 peaks there were in
	//random regions.
	for (int i=0; i<nMC; i++) {
	    int randStart = geneInfo.getRandomRegion(downstream, upstream, nongenic, random);
	    int hi[] = getHistogram(reads, readsInput, geneInfo.getChromosome(), randStart, true, downstream, upstream, fragmentLength);
	    int h[] = getHistogram(reads, null, geneInfo.getChromosome(), randStart, true, downstream, upstream, fragmentLength);
	    TreeMap<Integer, Integer> peaks = findAllPeaks(hi, h, minPeakDist, minPeakHeight);
	    int nPeaks = peaks.size();
	    nPeaksHist.put(nPeaks, nPeaksHist.containsKey(nPeaks) ? nPeaksHist.get(nPeaks) + 1 : 1);
	}
	return nPeaksHist;
    }

    //////////////////////////////////////////////////////////////////

    public static Vector<TreeMap<Integer, Integer>> findAllPeaks(int hi[][], int h[][], int minPeakDist, int minPeakHeight) {
	Vector<TreeMap<Integer, Integer>> peakLocs = new Vector<TreeMap<Integer, Integer>>();
	for (int i=0; i<h.length; i++)
	    peakLocs.addElement(findAllPeaks(hi[i], h[i], minPeakDist, minPeakHeight));
	return peakLocs;
    }

    public static TreeMap<Integer, Integer> findAllPeaks(int hi[], int minPeakDist, int minPeakHeight) { return findAllPeaks(hi, hi, minPeakDist, minPeakHeight); }

    public static TreeMap<Integer, Integer> findAllPeaks(int hi[], int h[], int minPeakDist, int minPeakHeight) { return findAllPeaks(hi, hi, minPeakDist, minPeakHeight, true); }

    public static TreeMap<Integer, Integer> findAllPeaks(int hi[], int h[], int minPeakDist, int minPeakHeight, boolean heights) {
	//Returns either the center and the height or the start and end of each peak.
	//Map has key location and value is the height of the peak. 
	TreeMap<Integer, Integer> peaks = new TreeMap<Integer, Integer>();
	//Key is start and value is end of the peak
	TreeMap<Integer, Integer> peakInterval = new TreeMap<Integer, Integer>();
	int prevPeakEnd = Integer.MIN_VALUE + hi.length + 1, prevPeakHeight = -1, prevPeakHeightInput = -1, prevPeakStart = -1, prevPeakCenter = 0;
	boolean belowThres = true; //Allow new peak if we drop either below hlaf height or threshold
	for (int i=0; i<hi.length; i++) {
	    if (hi[i]>minPeakHeight) { //To be significant, must be larger when compensating for input
		if (i-prevPeakEnd>minPeakDist && belowThres) {
		    prevPeakStart = i;
		    belowThres = false;
		}
		else if (!belowThres) {
		    prevPeakEnd = i;
		    prevPeakHeight = prevPeakHeight>h[i] ? prevPeakHeight : h[i];
		    prevPeakHeightInput = prevPeakHeightInput>hi[i] ? prevPeakHeightInput : hi[i];
		    prevPeakCenter = (prevPeakStart + prevPeakEnd)/2;
		}
	    }
	    if (!belowThres && (peaks.size()==0 || prevPeakCenter - peaks.lastKey()>minPeakDist) && (hi[i]<minPeakHeight || hi[i]<prevPeakHeightInput/2)) {
		//System.out.println("Found peak at " + prevPeakCenter + " with height " + prevPeakHeight);
		peaks.put(prevPeakCenter, prevPeakHeight);
		peakInterval.put(prevPeakStart, prevPeakEnd);
		belowThres = true;
		prevPeakHeight = -1;
		prevPeakHeightInput = -1;
	    }
	}
	return heights ? peaks : peakInterval;
    }

    //////////////////////////////////////////////////////////////////

    public static Vector<TreeMap<Integer, Boolean>> findAllConservedPeaks(int hi[][], int h[][], double pcs[][], int minPeakDist, int minPeakHeight) {
	if (pcs==null)
	    return null;
	Vector<TreeMap<Integer, Boolean>> peakConserved = new Vector<TreeMap<Integer, Boolean>>();
	for (int i=0; i<h.length; i++)
	    peakConserved.addElement(findAllConservedPeaks(hi[i], h[i], pcs[i], minPeakDist, minPeakHeight));
	return peakConserved;
    }

    public static TreeMap<Integer, Boolean> findAllConservedPeaks(int hi[], int h[], double pcs[], int minPeakDist, int minPeakHeight) {
	//Map has key location and value is the height of the peak
	//System.out.println(Arrays.toString(pcs));
	TreeMap<Integer, Integer> peakInterval = findAllPeaks(hi, h, minPeakDist, minPeakHeight, false);
	TreeMap<Integer, Boolean> peakCons = new TreeMap<Integer, Boolean>();
	for (Integer s : peakInterval.keySet()) {
	    int e = peakInterval.get(s);
	    boolean conserved = false;
	    for (int i=s; i<e; i++) {
		if (pcs[i]>HIGHLY_CONSERVED_THRESHOLD) {
		    conserved = true;
		    //System.out.println("Found conserved peak at " + ((s + e)/2));
		    break;
		}
	    }
	    peakCons.put((s + e)/2, new Boolean(conserved));
	}
	return peakCons;
    }

    //////////////////////////////////////////////////////////////////
    
    public static double[][] findMeanReadHistograms(Vector<int[][]> hists, int nBins) { 
	//Concatenate into one large set of arrays before using old version
	int nHists = 0;
	for (int[][] h : hists) 
	    nHists += h.length;
	int concatenatedHists[][] = new int[nHists][nBins], n = 0;
	for (int[][] h : hists) {
	    for (int i=n; i<h.length; i++)
		concatenatedHists[i+n] = h[i];
	    n += h.length;
	}
	return findMeanReadHistograms(concatenatedHists, nBins);
    }


    public static double[][] findMeanReadHistograms(int hists[][], int nBins) { 
	if (hists.length>0) {
	    //We want to be able to account for the situation when
	    //some of the bins should not be included for one or more
	    //of the histograms. These will typically be represented
	    //by negative values.
	    int nTSS[] = new int[nBins];
	    double hist[] = new double[nBins], hist2[] = new double[nBins], var[] = new double[nBins];
	    for (int j=0; j<hists.length; j++) {
		int h[] = hists[j];
		for (int i=0; i<h.length; i++) 
		    if (h[i]>=0) { nTSS[i]++; }
	    }
	    for (int j=0; j<hists.length; j++) {
		int h[] = hists[j];
		for (int i=0; i<h.length; i++) {
		    if (h[i]>=0) { 
			hist[i] += (double)h[i]/nTSS[i];
			hist2[i] += ((double)h[i])*((double)h[i]);
		    }
		}
	    }
	    //Calculate the sample variance
	    for (int i=0; i<hist.length; i++) {
		if (nTSS[i]>1) 
		    var[i] = (hist2[i] - hist[i]*hist[i]*nTSS[i])/(nTSS[i] - 1);
		else
		    var[i] = 0;
	    }
	    return new double[][]{hist, var};
	}
	return new double[2][nBins];
    }

    public static double[][] findMeanReadHistograms(double hists[][], int nBins) { 
	if (hists.length>0) {
	    //We want to be able to account for the situation when
	    //some of the bins should not be included for one or more
	    //of the histograms. These will typically be represented
	    //by negative values.
	    int nTSS[] = new int[nBins];
	    double hist[] = new double[nBins], hist2[] = new double[nBins], var[] = new double[nBins];
	    for (int j=0; j<hists.length; j++) {
		double h[] = hists[j];
		for (int i=0; i<h.length; i++) 
		    if (h[i]>=0) { nTSS[i]++; }
	    }
	    for (int j=0; j<hists.length; j++) {
		double h[] = hists[j];
		for (int i=0; i<h.length; i++) {
		    if (h[i]>=0) { 
			hist[i] += (double)h[i]/nTSS[i];
			hist2[i] += ((double)h[i])*((double)h[i]);
		    }
		}
	    }
	    //Calculate the sample variance
	    for (int i=0; i<hist.length; i++) {
		if (nTSS[i]>1) 
		    var[i] = (hist2[i] - hist[i]*hist[i]*nTSS[i])/(nTSS[i] - 1);
		else
		    var[i] = 0;
	    }
	    return new double[][]{hist, var};
	}
	return new double[2][nBins];
    }

    public static int[] getTSSs(RefSeqGenes geneInfo, int geneIDsKeep[], int isolatedIDs[]) { return getTSSs(geneInfo, geneIDsKeep, isolatedIDs, false); }

    public static int[] getTSSs(RefSeqGenes geneInfo, int geneIDsKeep[], int isolatedIDs[], boolean internal) {
	int prevTSS = -1, ids[] = geneInfo.getIDs(), nTSS = findNumberOfTSS(ids, geneIDsKeep, isolatedIDs), tssCount = 0, tssVec[] = new int[nTSS];
	TreeSet genes = geneInfo.getGenes();
	Iterator geneIt = genes.iterator();
	if (internal) {
	    Vector<Integer> as = new Vector<Integer>();
	    while (geneIt.hasNext()) {
		RefSeqGene g = (RefSeqGene)geneIt.next();
		if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0)) {
		    for (int i=0; i<g.getAlternativeStarts(true).length; i++)
			as.addElement(g.getAlternativeStarts(true)[i]);
		}
	    }
	    int i = 0;
	    tssVec = new int[as.size()];
	    for (Integer s : as)
		tssVec[i++] = s;
	}
	else {
	    if (nTSS>0) {
		while (geneIt.hasNext()) {
		    RefSeqGene g = (RefSeqGene)geneIt.next();
		    if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0)) {
			//System.out.println("Checking gene with ID " + g.getRefSeq());
			//Iterator tssIt = g.getStarts().iterator();
			//while (tssIt.hasNext()) {
			//int tss = ((Integer)tssIt.next()).intValue();
			TreeSet tssSort = new TreeSet(g.getStarts());
			tssVec[tssCount++] = g.getStrand() ? ((Integer)tssSort.first()).intValue() : ((Integer)tssSort.last()).intValue();
		    }
		}
	    }
	}
	return tssVec;
    }

    public static int[] getTESs(RefSeqGenes geneInfo, int geneIDsKeep[], int isolatedIDs[]) {
	TreeSet genes = geneInfo.getGenes();
	Iterator geneIt = genes.iterator();
	int prevTSS = -1, ids[] = geneInfo.getIDs(), nTES = findNumberOfTSS(ids, geneIDsKeep, isolatedIDs), tesCount = 0, tesVec[] = new int[nTES];
	if (nTES>0) {
	    while (geneIt.hasNext()) {
		RefSeqGene g = (RefSeqGene)geneIt.next();
		if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0)) {
		    TreeSet tesSort = new TreeSet(g.getEnds());
		    tesVec[tesCount++] = g.getStrand() ? ((Integer)tesSort.last()).intValue() : ((Integer)tesSort.first()).intValue();
		}
	    }
	}
	return tesVec;
    }

    public static boolean[] getStrands(RefSeqGenes geneInfo, int geneIDsKeep[], int isolatedIDs[]) { return getStrands(geneInfo, geneIDsKeep, isolatedIDs, false); }

    public static boolean[] getStrands(RefSeqGenes geneInfo, int geneIDsKeep[], int isolatedIDs[], boolean internal) {
	TreeSet genes = geneInfo.getGenes();
	Iterator geneIt = genes.iterator();
	int prevTSS = -1, ids[] = geneInfo.getIDs(), nTSS = findNumberOfTSS(ids, geneIDsKeep, isolatedIDs), tssCount = 0;
	boolean strandVec[] = new boolean[nTSS];
	if (internal) {
	    Vector<Boolean> as = new Vector<Boolean>();
	    while (geneIt.hasNext()) {
		RefSeqGene g = (RefSeqGene)geneIt.next();
		if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0)) {
		    for (int i=0; i<g.getAlternativeStarts(true).length; i++)
			as.addElement(g.getStrand());
		}
	    }
	    int i = 0;
	    strandVec = new boolean[as.size()];
	    for (Boolean s : as)
		strandVec[i++] = s;
	}
	else {
	    if (nTSS>0) {
		while (geneIt.hasNext()) {
		    RefSeqGene g = (RefSeqGene)geneIt.next();
		    if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0)) {
			//System.out.println("Checking gene with ID " + g.getRefSeq());
			//Iterator tssIt = g.getStarts().iterator();
			//while (tssIt.hasNext()) {
			//int tss = ((Integer)tssIt.next()).intValue();
			strandVec[tssCount++] = g.getStrand();
		    }
		}
	    }
	}
	return strandVec;
    }

    public static int[] getIDs(RefSeqGenes geneInfo, int geneIDsKeep[], int isolatedIDs[]) {
	TreeSet genes = geneInfo.getGenes();
	Iterator geneIt = genes.iterator();
	int prevTSS = -1, ids[] = geneInfo.getIDs(), nTSS = findNumberOfTSS(ids, geneIDsKeep, isolatedIDs), idCount = 0, idVec[] = new int[nTSS];
	if (nTSS>0) {
	    while (geneIt.hasNext()) {
		RefSeqGene g = (RefSeqGene)geneIt.next();
		if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0)) 
		    idVec[idCount++] = g.getID();
	    }
	}
	return idVec;
    }

    ////////////////////////////////////////////

    public static int[][] getHistograms(int chr, TreeMap<Integer, RefSeqGene> endsStrand, TreeMap<Integer, Integer> endsTSSDist, boolean sense, ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, boolean saveHist) /*throws IOException*/ {
	//Extract the histogram, but only include the part that is in
	//the convergent region when computing the average.
	int pvec[] = new int[endsStrand.size()], i = 0;
	boolean[] svec = new boolean[endsStrand.size()];
	for (int e : endsStrand.keySet()) { 
	    pvec[i] = e;
	    svec[i++] = endsStrand.get(e).getStrand();
	}
	int h[][] = getHistograms(chr, pvec, svec, reads, readsInput, downstream, upstream, fragmentLength, saveHist);
	//Now we want to remove everything which is not in the convergent regions
	i = 0;
	for (int e : endsTSSDist.keySet()) { 
	    for (int j=0; j<h[i].length; j++) {
		if (!inConvergentRegion(sense, endsTSSDist.get(e) + j - upstream)) //Find out if we are in the convergent region
		    h[i][j] = -1;
	    }
	    i++;
	}
	return h;
    }

    public static int[][] getHistograms(int chr, Vector<Integer> tssVec, Vector<Boolean> strandVec, ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, boolean saveHist) /*throws IOException*/ {
	int h[][] = new int[tssVec.size()][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<tssVec.size(); i++) 
	    h[i] = getHistogram(reads, readsInput, chr, tssVec.elementAt(i), strandVec.elementAt(i), downstream, upstream, fragmentLength);
	return h;
    }

    public static int[][] getHistograms(int chr, Vector<Integer> tssVec, ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, boolean saveHist) /*throws IOException*/ {
	int h[][] = new int[tssVec.size()][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<tssVec.size(); i++) 
	    h[i] = getHistogram(reads, readsInput, chr, tssVec.elementAt(i), true, downstream, upstream, fragmentLength);
	return h;
    }

    public static int[][] getHistograms(int chr, int tssVec[], boolean strandVec[], ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, boolean saveHist) /*throws IOException*/ {
	//Get the read histograms for all genes
	int h[][] = new int[tssVec.length][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<tssVec.length; i++) 
	    h[i] = getHistogram(reads, readsInput, chr, tssVec[i], strandVec[i], downstream, upstream, fragmentLength);
	return h;
    }

    public static int[] getHistogram(ABIRead reads[], ABIRead readsInput[], int chr, int tss, boolean forward, int downstream, int upstream, int fragmentLength) {
	int h[] = getReads(chr, tss, forward, downstream, upstream, fragmentLength, reads);
	//System.out.println("Around " + tss + " found histogram " + Arrays.toString(h));
	if (readsInput==null)
	    return h;
	int hDiff[] = new int[h.length];
	int hInput[] = getReads(chr, tss, forward, downstream, upstream, fragmentLength, readsInput);
	for (int i=0; i<h.length; i++) 
	    hDiff[i] = h[i] - hInput[i]; 
	return hDiff;
    }


    public static double[] getNormalizedHistogram(ABIRead reads[], ABIRead readsInput[], int chr, int tss, boolean forward, int downstream, int upstream, int fragmentLength, double normFactor, double normFactorInput) {
	int h[] = getReads(chr, tss, forward, downstream, upstream, fragmentLength, reads);
	double hDiff[] = new double[h.length];
	int hInput[] = readsInput!=null ? getReads(chr, tss, forward, downstream, upstream, fragmentLength, readsInput) : null;
	for (int i=0; i<h.length; i++) 
	    hDiff[i] = ((double)h[i])*normFactor - (hInput==null ? 0 : ((double)hInput[i])*normFactorInput); 
	return hDiff;
    }

    public static int[][] getHistograms(int chr, int peakVec[], ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, String saveHistName, double normFactor, double normFactorInput, int aggregateWindow) throws IOException {
	//Get the read histograms for all peaks. This version also saves the information for the scatter-plots
	BufferedWriter bw = saveHistName!=null ? new BufferedWriter(new FileWriter(saveHistName)) : null;
	System.out.println("Saving to " + saveHistName);
	int h[][] = new int[peakVec.length][numberOfBins(upstream, downstream, 1)], center = upstream + 1;
	int hInput[][] = new int[peakVec.length][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<peakVec.length; i++) {
	    h[i] = getHistogram(reads, null, chr, peakVec[i], true, downstream, upstream, fragmentLength);
	    if (readsInput!=null)
		hInput[i] = getHistogram(readsInput, null, chr, peakVec[i], true, downstream, upstream, fragmentLength);
	    if (bw!=null) {
		bw.write(peakVec[i] + "\t");
		double sum = 0, cSum = 0, iSum = 0;
		for (int j=0; j<h[i].length; j++) {
		    if (Math.abs(j - center)<aggregateWindow) {
			sum += ((double)h[i][j])*normFactor - ((double)hInput[i][j])*normFactorInput;
			cSum += h[i][j];
			iSum += hInput[i][j];
		    }
		    //bw.write(format.format(((double)h[i][j])*normFactor - ((double)hInput[i][j])*normFactorInput) + "\t");
		}
		//bw.write("\n");
		//System.out.println(peakVec[i] + "\t" + format.format(sum) + "\t" + format.format(cSum) + "\t" + format.format(iSum));
		bw.write(format.format(sum/fragmentLength) + "\n");
	    }
	}
	if (bw!=null) 
	    bw.close();
	return h;
    }

    public static int[][] getHistograms(int chr, int tssVec[], boolean strandVec[], ABIRead reads[], ABIRead readsInput[], int downstream, int upstream, int fragmentLength, String saveHistName, double normFactor, double normFactorInput) throws IOException {
	//Get the read histograms for all peaks
	BufferedWriter bw = saveHistName!=null ? new BufferedWriter(new FileWriter(saveHistName)) : null;
	System.out.println("Saving to " + saveHistName);
	int h[][] = new int[tssVec.length][numberOfBins(upstream, downstream, 1)], center = upstream + 1;
	int hInput[][] = new int[tssVec.length][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<tssVec.length; i++) {
	    h[i] = getHistogram(reads, null, chr, tssVec[i], strandVec[i], downstream, upstream, fragmentLength);
	    if (readsInput!=null)
		hInput[i] = getHistogram(readsInput, null, chr, tssVec[i], strandVec[i], downstream, upstream, fragmentLength);
	    if (bw!=null) {
		bw.write(tssVec[i] + "\t");
		for (int j=0; j<h[i].length; j++) 
		    bw.write(format.format(((double)h[i][j])*normFactor - ((double)hInput[i][j])*normFactorInput) + "\t");
		bw.write("\n");
	    }
	}
	if (bw!=null) 
	    bw.close();
	return h;
    }

    public static int[][] getHistograms(int chr, Vector<Enhancer> enhancers, ABIRead reads[], int downstream, int upstream, int fragmentLength) /*throws IOException*/ {
	//Get the read histograms for all peaks, exclude the intronic ones
	int nNonGenic = 0;
	for (Enhancer e : enhancers) 
	    nNonGenic += e.isNonGenic() ? 1 : 0;
	int h[][] = new int[nNonGenic][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<enhancers.size(); i++) {
	    if (enhancers.elementAt(i).isNonGenic())
		h[i] = getHistogram(reads, null, chr, enhancers.elementAt(i).getCenter(), downstream, upstream);
	}
	return h;
    }

    public static int[][] getHistograms(int chr, int peakVec[], ABIRead reads[], ABIRead readsInput[], int downstream, int upstream) {
	//Get the read histograms for all peaks
	int h[][] = new int[peakVec.length][numberOfBins(upstream, downstream, 1)];
	for (int i=0; i<peakVec.length; i++) 
	    h[i] = getHistogram(reads, readsInput, chr, peakVec[i], downstream, upstream);
	return h;
    }

    public static int[] getHistogram(ABIRead reads[], ABIRead readsInput[], int chr, int tss, int downstream, int upstream) {
	int h[] = getReadsHistogram(chr, tss, downstream, upstream, reads);
	//System.out.println("Around " + tss + " found histogram " + Arrays.toString(h));
	if (readsInput==null)
	    return h;
	int hDiff[] = new int[h.length];
	int hInput[] = getReadsHistogram(chr, tss, downstream, upstream, readsInput);
	for (int i=0; i<h.length; i++) 
	    hDiff[i] = h[i] - hInput[i]; 
	return hDiff;
    }

    ///////////////////////////////////////////////////////

    public static int[] numberOfPeaksNearCageStarts(Vector<TreeMap<Integer, Integer>> peakLocs, Vector<TreeMap<Integer, Boolean>> peakConserved, int ids[], int upstream, int fragmentLength, TreeMap<Integer, Vector<Integer>> cageStarts, boolean sense, double pcs[][], boolean shuffle) {
	//Find out how many of the CAGE starts appear to be driven by pol2
	int nPeaksNearStart = 0, nPeaksNearStartCons = 0;
	for (int i=0; i<peakLocs.size(); i++) {//Find all the reads in the region surrounding this TSS
	    int ind = shuffle ? random.nextInt(cageStarts.size()) : cageStarts.containsKey(ids[i]) ? 1 : -1;
	    if (ind>=0) {
		Vector<Integer> cageStart = new Vector<Integer>();
		if (shuffle) {
		    int j = 0; //Find a random gene
		    for (Integer k : cageStarts.keySet()) {
			cageStart = cageStarts.get(k);
			if (j++>ind) { break; }
		    }
		    if (cageStart.size()==0)
			System.err.println("Error, found length zero for ind " + ind);
		    //while ((sense && s>=0) || (!sense && s<=0)) 
		    int s = cageStart.elementAt(random.nextInt(cageStart.size())); //Find a random CAGE start
		    //System.out.println("Found random CAGE start " + s + " for " + ind);
		    cageStart = new Vector<Integer>();
		    cageStart.addElement(s);
		}
		else
		    cageStart = cageStarts.get(ids[i]);
		for (Integer s : cageStart) {
		    if ((sense && s<0) || (!sense && s>0)) {
			int near = peakNearLocation(peakLocs.elementAt(i), s + upstream, sense, fragmentLength);
			if (near>=0) {
			    nPeaksNearStart++;
			    if (pcs!=null && shuffle) { //Get a random conservation profile and check if there appears to be conservation
				double pc[] = pcs[random.nextInt(pcs.length)];
				//Find out if there is a highly conserved region around the 
				for (int j=Math.max(0, near - fragmentLength/2); j<Math.min(pc.length, near + fragmentLength/2); j++) {
				    if (pc[j]>HIGHLY_CONSERVED_THRESHOLD) {
					nPeaksNearStartCons++;
					break;
				    }
				}
			    }
			    else
				nPeaksNearStartCons += peakConserved!=null && peakConserved.elementAt(i).get(near) ? 1 : 0;
			}
		    }
		}
	    }
	}
	return new int[]{nPeaksNearStart, nPeaksNearStartCons};
    }

    /////////////////////////////////

    public static int[] numberOfPeaksNearConvergentCageStarts(Vector<TreeMap<Integer, Integer>> peakLocs, Vector<TreeMap<Integer, Boolean>> peakConserved, int ids[], int upstream, int fragmentLength, TreeMap<Integer, Vector<Integer>> cageStarts, boolean sense, double pcs[][], boolean shuffle) {
	//Find out how many peaks appear to be driving CAGE
	//transcripts in the convergent regions
	int nPeakWithCageStart = 0, nConsPeakWithCageStart = 0, nPeaksInConvergentRegion = 0, nPeaksInConvergentRegionWithCageData = 0, nGenesWithPeaksInConvergentRegion = 0;
	Vector<Integer> nCageStarts = getCageDistribution(cageStarts, ids);
	for (int i=0; i<peakLocs.size(); i++) {//Find out if there are a CAGE tag with a 5' end close to this peak
	    Vector<Integer> peakLoc = new Vector<Integer>(peakLocs.elementAt(i).keySet());
	    Vector<Integer> cageStart = new Vector<Integer>();
	    if (shuffle) 
		cageStart = getRandomCageStart(cageStarts, nCageStarts.elementAt(random.nextInt(nCageStarts.size())));
	    else if(cageStarts.containsKey(ids[i]))
		cageStart = cageStarts.get(ids[i]);
	    boolean peakFound = false;
	    for (Integer p : peakLoc) {
		if (inConvergentRegion(sense, p - upstream)) {
		    //if (!shuffle)
			//System.out.println("Found peak at " + (p - upstream) + " for " + ids[i]);
		    nPeaksInConvergentRegion++;
		    peakFound = true;
		    boolean cageStartFound = false;
		    for (Integer s : cageStart) {
			if (inConvergentRegion(sense, s)) { //((sense && s<0) || (!sense && s>0)) {
			    cageStartFound = true;
			    int near = peakNearLocation(p, s + upstream, sense, fragmentLength);
			    if (near>=0) {
				nPeakWithCageStart++;
				if (pcs!=null && shuffle) { //Get a random conservation profile and check if there appears to be conservation
				    double pc[] = pcs[random.nextInt(pcs.length)];
				    //Find out if there is a highly conserved region around the 
				    for (int j=Math.max(0, near - fragmentLength/2); j<Math.min(pc.length, near + fragmentLength/2); j++) {
					if (pc[j]>HIGHLY_CONSERVED_THRESHOLD) {
					    nConsPeakWithCageStart++;
					    break;
					}
				    }
				}
				else if (peakConserved!=null && peakConserved.elementAt(i)!=null && peakConserved.elementAt(i).containsKey(near) && peakConserved.elementAt(i).get(near))
				    nConsPeakWithCageStart++;
				break;
			    }
			}
		    }
		    if (!cageStartFound)
			nPeaksInConvergentRegionWithCageData++;
		}
	    }
	    if (peakFound) 
		nGenesWithPeaksInConvergentRegion++;
	}
	if (!shuffle) {
	    //System.out.println("Found " + nPeaksInConvergentRegion + " peaks in convergent region for " + nGenesWithPeaksInConvergentRegion + " unique genes.");
	    //System.out.println("Found " + nPeaksInConvergentRegionWithCageData + " peaks in convergent region with CAGE tags.");
	}
	return new int[]{nPeakWithCageStart, nConsPeakWithCageStart, nPeaksInConvergentRegion, nGenesWithPeaksInConvergentRegion, nPeaksInConvergentRegionWithCageData};
    }

    public static boolean inConvergentRegion(boolean sense, int p) {
	if (sense) 
	    return p>-MAX_CONVERGENT_REGION && p<CONVERGENT_UPSTREAM_REGION_START;
	return p<MAX_CONVERGENT_REGION && p>CONVERGENT_DOWNSTREAM_REGION_START;
    }

    public static Vector<Integer> getCageDistribution(TreeMap<Integer, Vector<Integer>> cageStarts, int ids[]) {
	Vector<Integer> nCageStarts = new Vector<Integer>();
	for (int i=0; i<ids.length; i++)
	    nCageStarts.addElement(cageStarts.containsKey(ids[i]) ? cageStarts.get(ids[i]).size() : 0);
	return nCageStarts;
    }

    public static Vector<Integer> getRandomCageStart(TreeMap<Integer, Vector<Integer>> cageStarts) { return getRandomCageStart(cageStarts, 1); }

    public static Vector<Integer> getRandomCageStart(TreeMap<Integer, Vector<Integer>> cageStarts, int n) {
	Vector<Integer> vec = new Vector<Integer>();
	for (int j=0; j<n; j++) {
	    int r = random.nextInt(cageStarts.size()), i = 0, id = cageStarts.lastKey();
	    for (Integer k : cageStarts.keySet()) {
		if (++i>r) {
		    id = k;
		    break;
		}
	    }
	    vec.addElement(cageStarts.get(id).elementAt(random.nextInt(cageStarts.get(id).size())));
	}
	return vec;
    }

    public static Vector<Integer> getRandomPeakLocation(Vector<TreeMap<Integer, Integer>> peakLocs) {
	Vector<Integer> vec = new Vector<Integer>();
	TreeMap<Integer, Integer> randLoc = peakLocs.elementAt(random.nextInt(peakLocs.size()));
	int r = random.nextInt(randLoc.size()), i = 0;
	for (Integer k : randLoc.keySet()) {
	    if (++i>r) {
		vec.addElement(k);
		return vec;
	    }
	}
	vec.addElement(randLoc.lastKey());
	return vec;
    }

    ///////////////////////////////////////////////////////

    public static int numberOfPeaksNearTranscriptStarts(Vector<TreeMap<Integer, Integer>> peakLocs, int ids[], int upstream, int fragmentLength, int transcriptStarts[][], boolean sense, boolean shuffle) { return numberOfPeaksNearTranscriptStarts(peakLocs, null, ids, upstream, fragmentLength, transcriptStarts, sense, null, shuffle); }

    public static int numberOfPeaksNearTranscriptStarts(Vector<TreeMap<Integer, Integer>> peakLocs, Vector<TreeMap<Integer, Boolean>> peakConserved, int ids[], int upstream, int fragmentLength, int transcriptStarts[][], boolean sense, double pcs[][], boolean shuffle) {
	int nPeaksNearStart = 0, nPeaksNearStartCons = 0, colInd = sense ? 1 : 2;
	for (int i=0; i<peakLocs.size(); i++) {//Find all the reads in the region surrounding this TSS
	    int ind = shuffle ? random.nextInt(transcriptStarts[0].length) : Arrays.binarySearch(transcriptStarts[0], ids[i]);
	    if (ind>=0) {
		int transcriptStart = transcriptStarts[colInd][ind];
		if (transcriptStart>Integer.MIN_VALUE && transcriptStart<MAX_TSS_DISTANCE_TRANSCRIPT_START) {
		    if ((sense && transcriptStart<0) || (!sense && transcriptStart>0)) {
			int near = peakNearLocation(peakLocs.elementAt(i), transcriptStart + upstream, sense, fragmentLength);
			if (near>=0) {
			    nPeaksNearStart++;
			    if (pcs!=null && pcs.length>0 && shuffle) { //Get a random conservation profile and check if there appears to be conservation
				double pc[] = pcs[random.nextInt(pcs.length)];
				//Find out if there is a highly conserved region around the peak
				for (int j=Math.max(0, near - fragmentLength/2); j<Math.min(pc.length, near + fragmentLength/2); j++) {
				    if (pc[j]>HIGHLY_CONSERVED_THRESHOLD) {
					nPeaksNearStartCons++;
					break;
				    }
				}
			    }
			    else if (peakConserved!=null && peakConserved.elementAt(i)!=null && peakConserved.elementAt(i).containsKey(near) && peakConserved.elementAt(i).get(near))
				nPeaksNearStartCons++;
			}
		    }
		}
	    }
	}
	return pcs==null ? nPeaksNearStart : nPeaksNearStartCons;
    }

    public static int peakNearLocation(TreeMap<Integer, Integer> peakLocs, int transcriptStart, boolean sense, int fragmentLength) {
	int pos = -1;
	for (int p : peakLocs.keySet()) {
	    pos = peakNearLocation(p, transcriptStart, sense, fragmentLength);
	    if (pos>=0)
		break;
	}
	return pos;
    }

    public static int peakNearLocation(int p, int transcriptStart, boolean sense, int fragmentLength) {
	int winLen = fragmentLength;
	if (transcriptStart<Integer.MAX_VALUE && transcriptStart>Integer.MIN_VALUE) {
	    if (sense) {
		if (p>transcriptStart - winLen/2 && p<transcriptStart + winLen/2) 
		    return p;
	    }
	    else {
		if (p>transcriptStart - winLen/2 && p<transcriptStart + winLen/2) 
		    return p;
	    }
	}
	return -1;
    }

    public static int numberOfPeaksNearRandomStart(int p, int transcriptStarts[], int upstream, boolean sense, int fragmentLength, int nMC) { return numberOfPeaksNearRandomStart(p, transcriptStarts, upstream, null, sense, fragmentLength, nMC); }

    public static int numberOfPeaksNearRandomStart(int p, int transcriptStarts[], int upstream, double pcs[][], boolean sense, int fragmentLength, int nMC) {
	int nStarts = 0, nStartsCons = 0;
	for (int i=0; i<nMC; i++) {
	    int near = peakNearLocation(p, transcriptStarts[random.nextInt(transcriptStarts.length)] + upstream, sense, fragmentLength);
	    if (near>=0 && pcs!=null) { //Found starting location
		double pc[] = pcs[random.nextInt(pcs.length)];
		//Find out if there is a highly conserved region around the 
		for (int j=Math.max(0, near - fragmentLength/2); j<Math.min(pc.length, near + fragmentLength/2); j++) {
		    if (pc[j]>HIGHLY_CONSERVED_THRESHOLD) {
			nStartsCons++;
			break;
		    }
		}
	    }
	    if (near>=0)
		nStarts++;
	}
	//if (pcs!=null && nStarts>0)
	//System.out.println("Found " + nStartsCons + " conserved random starts near " + p + " and " + nStarts);
	return pcs==null ? nStarts : nStartsCons;
    }

    public static int numberOfPeaksNearRandomCage(int p, TreeMap<Integer, Vector<Integer>> cageStarts, Vector<Integer> cageStartDistribution, int upstream, boolean sense, int fragmentLength, int nMC) { return numberOfPeaksNearRandomCage(p, cageStarts, cageStartDistribution, upstream, null, sense, fragmentLength, nMC); }

    public static int numberOfPeaksNearRandomCage(int p, TreeMap<Integer, Vector<Integer>> cageStarts, Vector<Integer> cageStartDistribution, int upstream, double pcs[][], boolean sense, int fragmentLength, int nMC) {
	int nStarts = 0, nStartsCons = 0;
	for (int i=0; i<nMC; i++) {
	    //Get a random number of CAGE tags based on background distribution
	    Vector<Integer> cageStart = getRandomCageStart(cageStarts, cageStartDistribution.elementAt(random.nextInt(cageStartDistribution.size())));
	    for (Integer c : cageStart) {
		int near = peakNearLocation(p, c + upstream, sense, fragmentLength);
		if (near>=0 && pcs!=null) { //Found starting location
		    double pc[] = pcs[random.nextInt(pcs.length)];
		    //Find out if there is a highly conserved region around the 
		    boolean foundConserved = false;
		    for (int j=Math.max(0, near - fragmentLength/2); j<Math.min(pc.length, near + fragmentLength/2); j++) {
			if (pc[j]>HIGHLY_CONSERVED_THRESHOLD) {
			    nStartsCons++;
			    foundConserved = true;
			    break;
			}
		    }
		    if (foundConserved) { break; }
		}
		else if (near>=0) {
		    nStarts++;
		    break;
		}
	    }
	}
	//if (pcs!=null && nStarts>0)
	//  System.out.println("Found " + format.format(nStarts) + " conserved random CAGE starts near " + p);
	return nStarts;
    }

    public static int[] numberOfPeaksNearConvergentTranscriptStarts(Vector<TreeMap<Integer, Integer>> peakLocs, Vector<TreeMap<Integer, Boolean>> peakConserved, int ids[], int upstream, int fragmentLength, int transcriptStarts[][], boolean sense, double pcs[][], boolean shuffle) {
	//Find out how many peaks appear to be driving transcripts in
	//the convergent regions
	int nPeakWithTranscriptStart = 0, nConsPeakWithTranscriptStart = 0, nPeaksInConvergentRegion = 0, nPeaksInConvergentRegionWithTranscriptData = 0, nGenesWithPeaksInConvergentRegion = 0, colInd = sense ? 1 : 2;
	for (int i=0; i<peakLocs.size(); i++) {//Find out if there are a CAGE tag with a 5' end close to this peak
	    int ind = shuffle ? random.nextInt(transcriptStarts[0].length) : Arrays.binarySearch(transcriptStarts[0], ids[i]);
	    Vector<Integer> peakLoc = new Vector<Integer>(peakLocs.elementAt(i).keySet());
	    int t = ind>=0 ? transcriptStarts[colInd][ind] : Integer.MIN_VALUE;
	    boolean peakFound = false;
	    for (Integer p : peakLoc) {
		if (inConvergentRegion(sense, p - upstream)) {
		    nPeaksInConvergentRegion++;
		    peakFound = true;
		    if (inConvergentRegion(sense, t)) { 
			nPeaksInConvergentRegionWithTranscriptData++;
			int near = peakNearLocation(p, t + upstream, sense, fragmentLength);
			if (near>=0) {
			    nPeakWithTranscriptStart++;
			    if (pcs!=null && shuffle) { //Get a random conservation profile and check if there appears to be conservation
				double pc[] = pcs[random.nextInt(pcs.length)];
				//Find out if there is a highly conserved region around the 
				for (int j=Math.max(0, near - fragmentLength/2); j<Math.min(pc.length, near + fragmentLength/2); j++) {
				    if (pc[j]>HIGHLY_CONSERVED_THRESHOLD) {
					nConsPeakWithTranscriptStart++;
					break;
				    }
				}
			    }
			    else if (peakConserved!=null && peakConserved.elementAt(i)!=null && peakConserved.elementAt(i).containsKey(near) && peakConserved.elementAt(i).get(near)) {
				nConsPeakWithTranscriptStart++;

			    }
			}
		    }
		}
	    }
	    if (peakFound) 
		nGenesWithPeaksInConvergentRegion++;
	}
	if (!shuffle) {
	    System.out.println("Found " + nPeaksInConvergentRegion + " peaks in convergent region and " + nConsPeakWithTranscriptStart + " conserved ones.");
	    //System.out.println("Found " + nPeaksInConvergentRegionWithTranscriptData + " peaks in convergent region with transcripts.");
	}
	return new int[]{nPeakWithTranscriptStart, nConsPeakWithTranscriptStart, nPeaksInConvergentRegion, nGenesWithPeaksInConvergentRegion, nPeaksInConvergentRegionWithTranscriptData};
    }

    public static TreeMap<String, Integer> calculateTotalReads(Vector<String> fileNames, String inputFileName, int nChr) { return calculateTotalReads(fileNames, inputFileName, nChr, 120, 10); }

    public static TreeMap<String, Integer> calculateTotalReads(Vector<String> fileNames, String inputFileName, int nChr, int fragmentLength, double qualityScore) {
	//Try opening the file where the stats have been
	//pre-computed. It is located in /home/mh168/Projects/ChIP-Seq/TKData/MmData_read_stats.txt with columns:
	//#Experiment     fragment_length   Uniquely mappable       QS filtered     PCR filtered    Mismatches0     Mismatches1     Mismatches2     Mismatches3
	String statsFileName = "/home/mh168/Projects/ChIP-Seq/TKData/MmData_read_stats.txt";
	TreeMap<String, Integer> nTotalReads = new TreeMap<String, Integer>(), nReads = new TreeMap<String, Integer>();;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(statsFileName));
	    String line = br.readLine();
	    Pattern statsPattern = Pattern.compile("^([\\w|/|\\-|_]+)\\s+\\d+\\s+\\d+\\s+\\d+\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = statsPattern.matcher(line);
		if (m.find() && Integer.parseInt(m.group(2))==fragmentLength) {
		    nTotalReads.put(m.group(1), Integer.parseInt(m.group(3)));
		    //System.out.println("Found " + m.group(3) + " reads for " + m.group(1));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opening stats file " + statsFileName);
	}
	if (inputFileName!=null) {
	    if (nTotalReads.containsKey(inputFileName.substring(0, inputFileName.lastIndexOf("/"))))
		nReads.put(inputFileName, nTotalReads.get(inputFileName.substring(0, inputFileName.lastIndexOf("/"))));
	    else {
		for (int i=1; i<=nChr; i++) { //For normalization purposes, we want to consider all chromosomes
		    ABIRead reads[] = PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore);
		    nTotalReads.put(inputFileName, nTotalReads.get(inputFileName) + reads.length);
		}
	    }
	}
	for (String fName : fileNames) {
	    if (nTotalReads.containsKey(fName.substring(0, fName.lastIndexOf("/")))) {
		nReads.put(fName, nTotalReads.get(fName.substring(0, fName.lastIndexOf("/"))));
		System.out.println("From file, found" + nReads.get(fName) + " reads for " + fName);
	    }
	    else {
		nReads.put(fName, 0);
		for (int i=1; i<=nChr; i++) { //For normalization purposes, we want to consider all chromosomes
		    //System.out.println(fName);
		    int n = !fName.contains("RNA") ? PeakFinder2.parseFiles(i, fName, fragmentLength, true, 0, qualityScore).length : ABIReads.loadMRNAReadsChr(fName, true, i, 35).length + ABIReads.loadMRNAReadsChr(fName, false, i, 35).length;
		    nReads.put(fName, nReads.get(fName) + n);
		}
		System.out.println("Counting, found " + nReads.get(fName) + " reads for " + fName);
	    }
	}
	return nReads;
    }

    /*public static TreeMap<String, Integer> calculateTotalReads(Vector<String> fileNames, String inputFileName, int nChr, int fragmentLength, double qualityScore) {
	TreeMap<String, Integer> nTotalReads = new TreeMap<String, Integer>();
	if (inputFileName!=null)
	    nTotalReads.put(inputFileName, 0);
	for (int k=0; k<fileNames.size(); k++) {
	    nTotalReads.put(fileNames.elementAt(k), 0);
	    for (int i=1; i<=nChr; i++) { //For normalization purposes, we want to consider all chromosomes
		ABIRead reads[] = !fileNames.elementAt(k).contains("RNA") ? PeakFinder2.parseFiles(i, fileNames.elementAt(k), fragmentLength, true, 0, qualityScore) : ABIReads.loadMRNAReadsChr(fileNames.elementAt(k), true, i, 35, false);
		nTotalReads.put(fileNames.elementAt(k), nTotalReads.get(fileNames.elementAt(k)) + reads.length);
		if (inputFileName!=null && k==0) 
		    nTotalReads.put(inputFileName, nTotalReads.get(inputFileName) + PeakFinder2.parseFiles(i, inputFileName, fragmentLength, true, 0, qualityScore).length);
	    }
	}
	return nTotalReads;
	}*/



    //////////////////////////////////////////////////////////////////
    //Older stuff

    public static double[] getGCHistogram(int chr, boolean orchestra, TreeMap<Integer, RefSeqGene> endLocs, TreeMap<Integer, Integer> endsTSSDist, int upstream, int downstream, boolean sense) {
	int pVec[] = new int[endLocs.size()], i = 0;
	boolean[] sVec = new boolean[endLocs.size()];
	for (int e : endLocs.keySet()) { 
	    pVec[i] = e;
	    sVec[i++] = endLocs.get(e).getStrand();
	}
	double gcHist[] = new double[numberOfBins(upstream, downstream, 1)];
	FastaSequence2 fastaSeq = new FastaSequence2(chr, orchestra);
	int ps[] = new int[pVec.length], n[] = new int[numberOfBins(upstream, downstream, 1)];
	for (i=0; i<pVec.length; i++) {
	    if (sVec[i])
		ps[i] = pVec[i] - upstream;
	    else
		ps[i] = pVec[i] - downstream;
	}
	String seqs[] = fastaSeq.getSequences(ps, upstream + downstream + 1);
	if (seqs.length>0) {
	    char seq[][] = new char[seqs.length][seqs[0].length()];
	    i = 0;
	    for (int e : endsTSSDist.keySet()) {
		if (seqs[i]!=null) {
		    seq[i] = sVec[i] ? seqs[i].toCharArray() : flipArray(seqs[i].toCharArray());
		    for (int j=0; j<seq[i].length; j++) {
			
			if (!inConvergentRegion(sense, endsTSSDist.get(e) + j - upstream)) //Find out if we are in the convergent region
			    seq[i][j] = 'N';
			else
			    n[j]++;
		    }
		}
		i++;
	    }
	    i = 0;
	    for (int e : endsTSSDist.keySet()) {
		for (int j=0; j<seq[i].length; j++) {
		    if (seq[i][j]=='c' || seq[i][j]=='C' || seq[i][j]=='g' || seq[i][j]=='G') 
			gcHist[j] += n[j]>0 ? 1.0/n[j] : 0;
		}
		i++;
	    }
	}
	return gcHist;
    }

    public static double[] getGCHistogram(int chr, boolean orchestra, int peakVec[], int upstream, int downstream) { return getGCHistogram(chr, orchestra, peakVec, upstream, downstream, true, null); }

    public static double[] getGCHistogram(int chr, boolean orchestra, int peakVec[], int upstream, int downstream, boolean strand, String saveFileStart) {
	boolean strandVec[] = new boolean[peakVec.length];
	Arrays.fill(strandVec, strand);
	return getGCHistogram(chr, orchestra, peakVec, strandVec, upstream, downstream, saveFileStart);
    }

    public static double[] getGCHistogram(int chr, boolean orchestra, int tssVec[], boolean strandVec[], int upstream, int downstream) { return getGCHistogram(chr, orchestra, tssVec, strandVec, upstream, downstream, null); }
    
    public static double[] getGCHistogram(int chr, boolean orchestra, int tssVec[], boolean strandVec[], int upstream, int downstream, String saveFileStart) {
	double gcHist[] = new double[numberOfBins(upstream, downstream, 1)];
	FastaSequence2 fastaSeq = new FastaSequence2(chr, orchestra);
	int tsss[] = new int[tssVec.length];
	for (int i=0; i<tssVec.length; i++) {
	    if (strandVec[i])
		tsss[i] = tssVec[i] - upstream;
	    else
		tsss[i] = tssVec[i] - downstream;
	}
	try {
	    BufferedWriter bw = saveFileStart==null ? null : new BufferedWriter(new FileWriter(saveFileStart));
	    if (bw!=null)
		System.out.println("Saving sequence to " + saveFileStart);
	    String seqs[] = fastaSeq.getSequences(tsss, upstream + downstream + 1);
	    for (int j=0; j<seqs.length; j++) {
		if (seqs[j]!=null) {
		    char seq[] = strandVec[j] ? seqs[j].toCharArray() : flipArray(seqs[j].toCharArray());
		    for (int i=0; i<seq.length; i++) {
			if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') 
			    gcHist[i] += 1.0/tssVec.length;
		    }
		    if (bw!=null)
			bw.write(">chr" + chr + ":" + tssVec[j] + "\n" + (new String(seq)) + "\n");
		}
	    }
	    if (bw!=null)
		bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving sequences " + e.toString());
	}
	return gcHist;
    }

    public static double[] getCpGHistogram(int chr, boolean orchestra, int tssVec[], boolean strandVec[], int upstream, int downstream) {
	double gcHist[] = new double[numberOfBins(upstream, downstream, 1)], cpgHist[] = new double[numberOfBins(upstream, downstream, 1)];
	FastaSequence2 fastaSeq = new FastaSequence2(chr, orchestra);
	int tsss[] = new int[tssVec.length], n = tssVec.length;;
	for (int i=0; i<tssVec.length; i++) {
	    if (strandVec[i])
		tsss[i] = tssVec[i] - upstream;
	    else
		tsss[i] = tssVec[i] - downstream;
	}
	String seqs[] = fastaSeq.getSequences(tsss, upstream + downstream + 1);
	for (int j=0; j<seqs.length; j++) {
	    if (seqs[j]!=null) {
		char seq[] = strandVec[j] ? seqs[j].toCharArray() : flipArray(seqs[j].toCharArray());
		for (int i=0; i<seq.length; i++) {
		    if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') 
			gcHist[i] += 1.0/n;
		}
	    }
	}
	for (int j=0; j<seqs.length; j++) {
	    if (seqs[j]!=null) {
		char seq[] = seqs[j].toCharArray();
		for (int i=1; i<seq.length; i++) {
		    //Count the number of CpGs and normalize by the
		    //expected CpG which is defined as (GC/2)^2
		    if (((seq[i-1]=='c' || seq[i-1]=='C') && (seq[i]=='g' || seq[i]=='G'))) {// || ((seq[i-1]=='g' || seq[i-1]=='G') && (seq[i]=='c' || seq[i]=='C'))) {
			cpgHist[i] += (1.0/n)/(((gcHist[i]+gcHist[i-1])/4)*((gcHist[i]+gcHist[i-1])/4)); //Normalize by the average GC-content at this dinucleotide
		    }
		}
	    }
	}
	return cpgHist;
    }

    /*public static double[] getCpGHistogram(int chr, boolean orchestra, int tssVec[], boolean strandVec[], int upstream, int downstream, String saveFileStart) {
	//Calculate the normalized frequency of CpGs as in the paper
	//by Saxonov et al (2006). This is intended to be done after
	//the GC calculation and hence we first try to read the
	//sequences from file.
	double cpgHist[] = new double[numberOfBins(upstream, downstream, 1)];
	String seqs[];
	try {

	}
	catch (IOException e) {
	    FastaSequence2 fastaSeq = new FastaSequence2(chr, orchestra);
	    int tsss[] = new int[tssVec.length];
	    for (int i=0; i<tssVec.length; i++) {
		if (strandVec[i])
		    tsss[i] = tssVec[i] - upstream;
		else
		    tsss[i] = tssVec[i] - downstream;
	    }
	    try {
		BufferedWriter bw = saveFileStart==null ? null : new BufferedWriter(new FileWriter(saveFileStart));
		if (bw!=null)
		    System.out.println("Saving sequence to " + saveFileStart);
		String seqs[] = fastaSeq.getSequences(tsss, upstream + downstream + 1);
		for (int j=0; j<seqs.length; j++) {
		    if (seqs[j]!=null) {
			if (bw!=null)
			    bw.write(">chr" + chr + ":" + tssVec[j] + "\n" + (new String(seq)) + "\n");
		    }
		}
		if (bw!=null)
		    bw.close();
	    }
	    catch (IOException ex) {

	    }
	}
	//Go through the list of sequences and calculate the number of CG dinucleotides
	for (int j=0; j<seqs.length; j++) {
	    if (seqs[j]!=null) {
		char seq[] = strandVec[j] ? seqs[j].toCharArray() : flipArray(seqs[j].toCharArray());
		int nCG = 0;
		for (int i=0; i<seq.length; i++) {
		    if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') 
			gcHist[i] += 1.0/tssVec.length;
		}
		for (int i=1; i<seq.length; i++) {
		    //Count the number of CpGs and normalize by the
		    //expected CpG which is defined as (GC/2)^2
		    if (((seq[i-1]=='c' || seq[i-1]=='C') && (seq[i]=='g' || seq[i]=='G')) || ((seq[i-1]=='g' || seq[i-1]=='G') && (seq[i]=='c' || seq[i]=='C')))
			cpgHist[i] += (1.0/tssVec.length)/((gcHist[i]/2)*(gcHist[i]/2));
		    }
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving sequences " + e.toString());
	}
	return cpgHist;
	}*/

    ///////////////////////////////////////////////////////

    public static double getConservation(double pcs[][], int[][] transcriptStarts, int ids[], int upstream, int fragmentLength, boolean sense, int nMC) {
	//Use this for the conservation of the 5'-ends
	double cons = 0;
	if (pcs!=null && pcs.length>0) {
	    int col = sense ? 1 : 2;
	    for (int j=0; j<pcs.length; j++) {
		int ind = Arrays.binarySearch(transcriptStarts[0], ids[j]);
		if (true) {
		    for (int i=upstream + transcriptStarts[col][ind] - fragmentLength/2; i<upstream + transcriptStarts[col][ind] + fragmentLength/2; i++) {
			if (pcs[j][i]>HIGHLY_CONSERVED_THRESHOLD)
			    cons += 1.0/pcs.length;
		    }
		}
	    }
	}
	return cons;
    }

    public static double[] getConservationHistogram(double pcs[][]) {
	double consHist[] = null;
	if (pcs!=null && pcs.length>0) {
	    int n[] = new int[pcs[0].length];
	    /*for (int j=0; j<pcs.length; j++) {
		for (int i=0; i<pcs[j].length; i++) 
		    n[i] += pcs[j][i]>=0 ? 1 : 0;
		    }*/
	    consHist = new double[pcs[0].length];
	    for (int j=0; j<pcs.length; j++) {
		for (int i=0; i<pcs[j].length; i++) {
		    //if (pcs[j][i]>HIGHLY_CONSERVED_THRESHOLD)
		    //consHist[i] += n[i]>0 ? 1.0/n[i] : 0;
		    consHist[i] += pcs[j][i]/pcs.length;
		}
	    }
	}
	else
	    consHist = new double[0];
	return consHist;
    }
    
    public static double[] shuffleTranscriptStarts(int transcriptStarts[], boolean sense, int hDiff[], int upstream, int fragmentLength, int minPeakHeight, double pcs[], int nMC) {
	double nShuffle = 0, nShuffleConserved = 0;
	for (int i=0; i<nMC; i++) {
	    int randInd = 0;
	    //do { //Make sure that we get something which actually has a start
		randInd = random.nextInt(transcriptStarts.length);
		//}
		//while (transcriptStarts[randInd]==Integer.MIN_VALUE || transcriptStarts[randInd]==Integer.MAX_VALUE);
		if (transcriptStarts[randInd]>Integer.MIN_VALUE && transcriptStarts[randInd]<Integer.MAX_VALUE && Math.abs(transcriptStarts[randInd])<MAX_TSS_DISTANCE_TRANSCRIPT_START) {
		int peakLoc = findTranscriptInitiationPeak(hDiff, transcriptStarts[randInd] + upstream, sense, fragmentLength, minPeakHeight);
		if (peakLoc>0 && peakLoc<hDiff.length && hDiff[peakLoc]>minPeakHeight) {
		    nShuffle++;
		    if (pcs!=null && peakHighlyConserved(hDiff, peakLoc, pcs))
			nShuffleConserved++;
		}
	    }
	}
	return new double[]{nShuffle/nMC, nShuffleConserved/nMC};
    }


    public static double[] flipArray(double a[]) {
	double b[] = new double[a.length];
	for (int i=0; i<a.length; i++)
	    b[i] = a[a.length - i - 1];
	return b;
    }

    public static int[] flipArray(int a[]) {
	int b[] = new int[a.length];
	for (int i=0; i<a.length; i++)
	    b[i] = a[a.length - i - 1];
	return b;
    }

    public static char[] flipArray(char a[]) {
	char b[] = new char[a.length];
	for (int i=0; i<a.length; i++)
	    b[i] = a[a.length - i - 1];
	return b;
    }

    public static boolean peakHighlyConserved(int h[], int peakLoc, double pcs[]) {
	//Use this method to figure out if there are one or more
	//highly conserved bases for the peak. First we find the width
	//of the peak which is defined as the stretch between the
	//half-height points.
	int peakStartEnd[] = findPeakWidth(h, peakLoc);
	for (int i=0; i<peakStartEnd[1] - peakStartEnd[0]; i++) {
	    if (pcs[peakStartEnd[0] + i]>HIGHLY_CONSERVED_THRESHOLD) 
		return true;
	}
	return false;
    }

    public static int findTranscriptInitiationPeak(int h[], int startInd, boolean sense, int fragmentLength, int minPeakHeight) {
	//Look for a peak of sufficient height near the 5' end of the transcript
	int winLen = fragmentLength + 1;
	int peakLoc = findClosestMax(h, startInd + (sense ? -winLen : 0), true, 0, winLen, 0);
	//Make sure that the max is not on the boundary, but at an
	//interior point.
	while ((sense && (peakLoc==startInd-winLen || peakLoc==startInd)) || (!sense && (peakLoc==startInd || peakLoc==startInd+winLen))) {
	    winLen -= 2;
	    if (winLen<=0) {
		peakLoc = -1;
		break;
	    }
	    peakLoc = findClosestMax(h, startInd + (sense ? -winLen : 0), true, 0, winLen, 0);
	}
	//System.out.println("Found peak at " + peakLoc + " close to " + startInd);
	return peakLoc;
    }

    public static int numberOfPeaks(int h[], int start, int end, int minPeakDist, int minPeakHeight) {
	int nPeaks = 0, prevPeak = Integer.MIN_VALUE + h.length + 1, prevPeakHeight = -1;
	boolean belowThres = true; //Allow new peak if we drop either below hlaf height or threshold
	for (int i=start; i<end; i++) {
	    if (h[i]>minPeakHeight) {
		if (i-prevPeak>minPeakDist && belowThres) 
		    nPeaks++;
		prevPeak = i;
		prevPeakHeight = prevPeakHeight>h[i] ? prevPeakHeight : h[i];
		belowThres = false;
	    }
	    if (h[i]<minPeakHeight || h[i]<prevPeakHeight/2) {
		belowThres = true;
		prevPeakHeight = -1;
	    }
	}
	return nPeaks;
    }

    public static int[] findPeakWidth(int hist[], int peakLoc) {
	//Find the points where we have reached half the height of the
	//maximum on either side. The distance between these points is
	//the width of the peak.
	int upstreamHalf = peakLoc, downstreamHalf = peakLoc;
	while (upstreamHalf>=0 && hist[upstreamHalf]>hist[peakLoc]/2)
	    upstreamHalf--;
	while (downstreamHalf<hist.length && hist[downstreamHalf]>hist[peakLoc]/2)
	    downstreamHalf++;
	return new int[]{upstreamHalf, downstreamHalf};
    }

    public static int findClosestMax(int hist[], int minPeakHeight) { return findClosestMax(hist, 0, true, 0, hist.length, minPeakHeight); }

    public static int findClosestMax(int hist[], int center, boolean direction, int minPeakDist, int maxPeakDist, int minPeakHeight) {
	//Find the global max in the regions which is restricted by
	//the center, minPeakDist and direction.
	int i = center + (direction ? minPeakDist : -minPeakDist), end = center + (direction ? maxPeakDist : -maxPeakDist), max = -1, maxInd = direction ? hist.length : -1;
	if (i>=0 && i<hist.length) {
	    if (direction) {
		while (i<hist.length && i<end) {
		    if (hist[i]>=max && hist[i]>=minPeakHeight) {
			max = hist[i];
			maxInd = i;
		    }
		    i++;
		}
	    }
	    else {
		while (i>=0 && i>=end) {
		    if (hist[i]>max && hist[i]>=minPeakHeight) {
			max = hist[i];
			maxInd = i;
		    }
		    i--;
		}
	    }
	    //if (maxInd>=0 && maxInd<hist.length)
	    //System.out.println("Found peak at " + maxInd + ", " + max);
	}
	return maxInd;
    }

    //public static Vector<int[]> findAllPeaks(double h[], int minPeakHeight) {
	//Use this method to detect all peaks in the window

    public static double[][] getPeakLocations(Vector peakPos, int nBins) {
	//Get histograms with the number of peaks fonud in each location
	double peakLocs[][] = new double[4][nBins], peakDist[] = new double[3], peakDist2[] = new double[3], meanPeakLoc[] = new double[4], meanPeakLoc2[] = new double[4];
	double meanPeakWidth[] = new double[4], meanPeakWidth2[] = new double[4];
	int center = nBins/2 + 1, nPeaks[] = new int[4];
	for (int i=0; i<peakPos.size(); i++) {
	    int vec[][] = (int[][])peakPos.elementAt(i);
	    int pos[] = vec[0];
	    int width[] = vec[1];
	    for (int j=0; j<pos.length; j++) {
		if (pos[j]>Integer.MIN_VALUE && pos[j]<Integer.MAX_VALUE) {
		    peakLocs[j][center + pos[j]]++;
		    nPeaks[j]++;
		    meanPeakLoc[j] += pos[j];
		    meanPeakLoc2[j] += pos[j]*pos[j];
		    meanPeakWidth[j] += width[j];
		    meanPeakWidth2[j] += width[j]*width[j];
		}
	    }
	    for (int j=0; j<pos.length - 1; j++) { //Calculate the distances between the peaks
		if (pos[j]>Integer.MIN_VALUE && pos[j]<Integer.MAX_VALUE && pos[j+1]>Integer.MIN_VALUE && pos[j+1]<Integer.MAX_VALUE) {
		    peakDist[j] += pos[j+1] - pos[j];
		    peakDist2[j] += (pos[j+1] - pos[j])*(pos[j+1] - pos[j]);
		}
	    }
	}
	for (int i=0; i<meanPeakLoc.length; i++) { //Look at the average locations of the peaks
	    System.out.println("Mean peak loc is " + format.format(meanPeakLoc[i]/nPeaks[i]) + " for " + nPeaks[i] + " with std err " + format.format(Math.sqrt((meanPeakLoc2[i] - meanPeakLoc[i]*meanPeakLoc[i]/nPeaks[i])/((nPeaks[i] - 1)*nPeaks[i]))));
	    System.out.println("Mean peak width is " + format.format(meanPeakWidth[i]/nPeaks[i]) + " for " + nPeaks[i] + " with std err " + format.format(Math.sqrt((meanPeakWidth2[i] - meanPeakWidth[i]*meanPeakWidth[i]/nPeaks[i])/((nPeaks[i] - 1)*nPeaks[i]))));
	}
	//Look at the average distances between pairs of peaks
	int nP[] = new int[3];
	for (int j=0; j<peakDist.length; j++) {
	    nP[j] = Math.min(nPeaks[j], nPeaks[j+1]);
	    System.out.println("Average peak distance is " + format.format(peakDist[j]/nP[j]) + " for " + nP[j] + " with std err " + format.format(Math.sqrt((peakDist2[j] - peakDist[j]*peakDist[j]/nP[j])/((nP[j] - 1)*nP[j]))));
	}
	return peakLocs;
    }

    //////////////////////////////////////////////////////////////////

    public static int getRegionStart(boolean forward, int center, int upstream, int downstream, int fragmentLength) { return forward ? center - upstream - fragmentLength : center - downstream - fragmentLength; }

    public static int getRegionEnd(boolean forward, int center, int upstream, int downstream, int fragmentLength) { return forward ? center + downstream + fragmentLength : center + upstream + fragmentLength; }

    public static int[] getReads(int chr, int tss, boolean forward, int downstream, int upstream, int fragmentLength, ABIRead reads[]) {
	//Use this method for finding ChIP-reads in a given region.
	int start = getRegionStart(forward, tss, upstream, downstream, fragmentLength); 
	int end = getRegionEnd(forward, tss, upstream, downstream, fragmentLength); 
	int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start, fragmentLength), null);
	ind = ind<0 ? -ind-1 : ind;
	int hist[] = new int[numberOfBins(downstream, upstream)];
	int nReads = 0;
	//System.out.println("Found peak " + tss + " starting " + start + "-" + end + " with " + ind);
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    for (int i=0; i<fragmentLength; i++) {
		if (reads[ind].getPosition() + i>=start + fragmentLength && reads[ind].getPosition() + i<=end - fragmentLength) {
		    if (forward) 
			hist[-tss + upstream + reads[ind].getPosition() + i]++;
		    else
			hist[hist.length - 1 -(-tss + downstream + reads[ind].getPosition() + i)]++;
		}
	    }
	    ind++;
	}
	//System.out.println("Found " + hist[hist.length/2] + " for " + tss);
	return hist;
    }

    public static ABIRead[] getReads(int chr, int tss, int downstream, int upstream, ABIRead reads[]) {
	//Use this method for finding RNA-reads in a given region.
	int start = tss - upstream;
	int end = tss + downstream;
	int fragmentLength = FindNongenicTranscripts.MRNA_READ_LENGTH;
	int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start), null);
	ind = ind<0 ? -ind-1 : ind;
	int ind0 = ind, nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    /*for (int i=0; i<fragmentLength; i++) {
		if (reads[ind].getPosition() + i>=start + fragmentLength && reads[ind].getPosition() + i<=end - fragmentLength) {
		    hist[-tss + downstream + reads[ind].getPosition() + i]++;
		}
		}*/
	    ind++;
	}
	ABIRead subsetReads[] = new ABIRead[nReads];
	for (int i=0; i<nReads; i++)
	    subsetReads[i] = reads[ind0 + i];
	//System.out.println("Found " + hist[hist.length/2] + " for " + tss);
	return subsetReads;
    }

    public static int[] getReadsHistogram(int chr, int tss, int downstream, int upstream, ABIRead reads[]) { return getReadsHistogram(chr, tss, downstream, upstream, reads, null); }

    public static int[] getReadsHistogram(int chr, int tss, int downstream, int upstream, ABIRead reads[], TreeMap<Integer, Integer> downSampleDistribution) {
	//Use this method for finding RNA-reads in a given region.
	int fragmentLength = FindNongenicTranscripts.MRNA_READ_LENGTH;
	int start = tss - upstream - fragmentLength;
	int end = tss + downstream + fragmentLength;
	int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start), null);
	ind = ind<0 ? -ind-1 : ind;
	int hist[] = new int[numberOfBins(downstream, upstream)], nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    for (int i=0; i<fragmentLength; i++) {
		//if (reg==0 || (reg!=0 && inConvergentRegion(reg<0, -tss + reads[ind].getPosition() + i))) {
		    if (reads[ind].getPosition() + i>=start + fragmentLength && reads[ind].getPosition() + i<=end - fragmentLength) {
			hist[-tss + upstream + reads[ind].getPosition() + i]++;
		    }
		    //}
	    }
	    ind++;
	}
	return hist;
    }

    public static int numberOfReads(int chr, int tss, boolean forward, int downstream, int upstream, int fragmentLength, ABIRead reads[]) {
	//Use this method for finding the number of ChIP-reads in a given region.
	int start = getRegionStart(forward, tss, upstream, downstream, fragmentLength); 
	int end = getRegionEnd(forward, tss, upstream, downstream, fragmentLength); 
	return numberOfReads(chr, start, end, fragmentLength, reads);
    }

    public static int numberOfReads(int chr, int start, int end, int fragmentLength, ABIRead reads[]) {
	int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start, fragmentLength), null);
	ind = ind<0 ? -ind-1 : ind;
	int nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    ind++;
	}
	return nReads;
    }

    public static int numberOfReads(int chr, int start, int end, ABIRead reads[]) {
	int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start), null);
	ind = ind<0 ? -ind-1 : ind;
	int nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    ind++;
	}
	return nReads;
    }

    public static ABIRead[] getReads(int chr, int start, int end, ABIRead reads[]) {
	int ind0 = Arrays.binarySearch(reads, new ABIRead(chr, true, start), null);
	int ind = ind0<0 ? -ind0-1 : ind0;
	int nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    ind++;
	}
	ABIRead rs[] = new ABIRead[nReads];
	ind = ind0<0 ? -ind0-1 : ind0;
	for (int i=0; i<nReads; i++)
	    rs[i] = reads[ind + i];
	return rs;
    }

    public static double[][] analyzeChromosome(int chr, String fileName, int upstream, int downstream, String peakFile, int nFactors, boolean isolated, int geneIDsKeep[], boolean orchestra) {
	double hists[][] = new double[nFactors][numberOfBins(downstream, upstream, 1)];
	int fragmentLength = 120, nTSS[] = new int[nFactors];
	double qualityScore = 10;
	ABIRead reads[] = PeakFinder2.parseFiles(chr, fileName, fragmentLength, true, 0, qualityScore);
	//System.out.println("Found " + reads.length + " reads.");
	TreeMap histograms = new TreeMap();
	try {
	    double hist[] = new double[numberOfBins(downstream, upstream, 1)];
	    Arrays.fill(hist, 0);
	    BufferedReader br = new BufferedReader(new FileReader(peakFile));
	    Pattern peakInfoPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+[\\d|\\.]+\\s+[\\d|\\.]+\\s+[\\d|\\.]+\\s+[\\d|\\.]+\\s+[\\d|\\.]+\\s+[\\d|\\.]+\\s+[\\d|\\.]+\\s+");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = peakInfoPattern.matcher(line);
		if (m.find()) {
		    int c = Integer.parseInt(m.group(1));
		    if (c==chr) {
			int center = (Integer.parseInt(m.group(2)) + Integer.parseInt(m.group(3)))/2;
			//Find all the reads in the region surrounding this TSS
			int start = center - upstream - fragmentLength;
			int end = center + downstream + fragmentLength;
			int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start, fragmentLength), null);
			ind = ind<0 ? -ind-1 : ind;
			//System.out.println("Found peak at " + center + " with index " + ind + " start = " + start + " and end = " + end);
			while (ind<reads.length && reads[ind].getPosition()<end) {
			    for (int i=0; i<fragmentLength; i++) {
				if (reads[ind].getPosition() + i>=start + fragmentLength && reads[ind].getPosition() + i<=end - fragmentLength) 
				    hist[-center + downstream + reads[ind].getPosition() + i]++;
			    }
			    ind++;
			}
			//System.out.println("Found reads around peak at " + center);
			//Next are the columns containing the
			//enrichment values for each factor in the
			//peak
			line = line.substring(m.end());
			String tokens[] = line.split("\\s");
			for (int i=0; i<nFactors; i++) {
			    double enrichment = Double.parseDouble(tokens[i]);
			    if (enrichment>0) {
				//System.out.println("Found enrichment " + enrichment + " at " + i);
				nTSS[i]++;
				for (int j=0; j<hists[i].length; j++)
				    hists[i][j] += hist[j];
			    }
			}
		    }
		    else if (c>chr)
			break;
		}
		line = br.readLine();
	    }
	}
	catch (IOException e) {

	}
	for (int i=0; i<nFactors; i++) {
	    if (nTSS[i]>0) {
		for (int j=0; j<hists[i].length; j++)
		    hists[i][j] /= nTSS[i];
	    }
	}
	//String saveFileName = fileName + "_combined_" + feature + "_" + chr + ".txt";
	/*BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	  for (int i=0; i<hist.length; i++)
	  bw.write((-upstream + i) + "\t" + format.format(hist[i]) + "\n");
	  bw.close();*/
	return hists;
    }
    
    
    public static int[] parseSelectFile(String fileName) {
	int geneIDsKeep[] = null;
	//String geneIDsKeep[] = null;
	if (fileName!=null) {
	    try {
		Vector vec = new Vector();
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		//GeneID should alwasy be first
		//GeneID name chr strand 
		Pattern selectPattern = Pattern.compile("^(\\d+)\\s*");
		//Pattern selectPattern = Pattern.compile("^([\\w|_\\d]+)\\s+");
		while (line!=null) {
		    Matcher m = selectPattern.matcher(line);
		    if (m.find() && !vec.contains(new Integer(m.group(1))))
			vec.addElement(new Integer(m.group(1)));
			//vec.addElement(m.group(1));
		    line = br.readLine();
		}
		br.close();
		geneIDsKeep = new int[vec.size()];
		//geneIDsKeep = new String[vec.size()];
		for (int i=0; i<vec.size(); i++)
		    geneIDsKeep[i] = ((Integer)vec.elementAt(i)).intValue();
		//geneIDsKeep[i] = (String)vec.elementAt(i);
		Arrays.sort(geneIDsKeep);
	    }
	    catch (IOException e) {
		System.err.println("Error parsing " + fileName);
	    }
	}
	return geneIDsKeep;
    }

    public static LinkedHashMap parseHeader(String fileName) throws IOException {
	BufferedReader br = new BufferedReader(new FileReader(fileName));
	String line = br.readLine();
	Pattern factorPattern = Pattern.compile("([\\d|\\w]+_[KCl|un]+)");
	Matcher m = factorPattern.matcher(line);
	LinkedHashMap factor2ColumnMap = new LinkedHashMap();
	int i = 0;
	while (m.find()) {
	    System.out.println("Found header " + m.group(1) + " at " + i);
	    factor2ColumnMap.put(m.group(1), new Integer(i++));
	}
	br.close();
	return factor2ColumnMap;
    }

    public static int[] parseIsolatedFile(boolean orchestra) {
	return parseIsolatedFile((orchestra ? "/groups/neuroduo/martin/" : "/media/exthd/") + "Genomes/Mm/v9/isolated_genes.txt");
    }
    
    public static int[] parseIsolatedFile(String fileName) {
	int ids[] = null;
	try {
	    Vector vec = new Vector();
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //name GeneID
	    Pattern selectPattern = Pattern.compile("^.+\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = selectPattern.matcher(line);
		if (m.find())
		    vec.addElement(new Integer(m.group(1)));
		line = br.readLine();
	    }
	    br.close();
	    ids = new int[vec.size()];
	    for (int i=0; i<vec.size(); i++)
		ids[i] = ((Integer)vec.elementAt(i)).intValue();
	    Arrays.sort(ids);
	}
	catch (IOException e) {
	    System.err.println("Error parsing " + fileName);
	}
	return ids;
    }

    public static int[][] parseInitiationFile(String fileName) { return parseInitiationFile(fileName, null, false); }

    public static int[][] parseInitiationFile(String fileName, int isolatedIDs[], boolean convergentOnly) { return parseInitiationFile(fileName, isolatedIDs, convergentOnly, false); }

    public static int[][] parseInitiationFile(String fileName, int isolatedIDs[], boolean convergentOnly, boolean bothStrandsOnly) {
	//Use this method to parse the file with initiation locations
	//as defined by JMG.
	int ret[][] = null;
	int nSenseConvergentStarts = 0, nAntiSenseConvergentStarts = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern initPattern = Pattern.compile("(\\d+)\\s+([NA|\\d|-]+)\\s+([NA|\\d|-]+)");
	    Vector<Integer> ids = new Vector<Integer>(), sStart = new Vector<Integer>(), asStart = new Vector<Integer>();
	    while (line!=null) {
		Matcher m = initPattern.matcher(line);
		if (m.find()) {
		    if (bothStrandsOnly) {
			if (!m.group(2).equals("NA") && !m.group(3).equals("NA")) {
			    if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, Integer.parseInt(m.group(1)))>=0)
				ids.addElement(Integer.parseInt(m.group(1)));
			}
		    }
		    else {
			if (!m.group(2).equals("NA") || !m.group(3).equals("NA")) {
			    if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, Integer.parseInt(m.group(1)))>=0)
				ids.addElement(Integer.parseInt(m.group(1)));
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    ret = new int[3][ids.size()];
	    for (int i=0; i<ids.size(); i++) 
		ret[0][i] = ids.elementAt(i);
	    Arrays.sort(ret[0]);
	    br = new BufferedReader(new FileReader(fileName));
	    line = br.readLine();
	    while (line!=null) {
		Matcher m = initPattern.matcher(line);
		if (m.find()) {
		    if (bothStrandsOnly) {
			if (!m.group(2).equals("NA") && !m.group(3).equals("NA")) {
			    int ind = Arrays.binarySearch(ret[0], Integer.parseInt(m.group(1)));
			    if (ind>=0) {
				ret[1][ind] = Integer.parseInt(m.group(2));
				nSenseConvergentStarts += !convergentOnly || ret[1][ind]<0 ? 1 : 0;
				ret[2][ind] = Integer.parseInt(m.group(3));
				nAntiSenseConvergentStarts += !convergentOnly || ret[2][ind]<0 ? 1 : 0;
			    }
			}
		    }
		    else {
			if (!m.group(2).equals("NA") || !m.group(3).equals("NA")) {
			    int ind = Arrays.binarySearch(ret[0], Integer.parseInt(m.group(1)));
			    if (ind>=0) {
				if (m.group(2).equals("NA"))
				    ret[1][ind] = Integer.MIN_VALUE;
				else {
				    ret[1][ind] = Integer.parseInt(m.group(2));
				    nSenseConvergentStarts += !convergentOnly || ret[1][ind]<0 ? 1 : 0;
				}
				if (m.group(3).equals("NA"))
				    ret[2][ind] = Integer.MAX_VALUE;
				else {
				    ret[2][ind] = Integer.parseInt(m.group(3));
				    nAntiSenseConvergentStarts += !convergentOnly || ret[2][ind]<0 ? 1 : 0;
				}
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	}
	catch (IOException e) {
	    System.out.println("Error " + e.toString() + " reading file " + fileName);
	}
	if (convergentOnly)
	    System.out.println("Found " + nSenseConvergentStarts + " convergent sense starts and " + nAntiSenseConvergentStarts + " convergent anti-sense.");
	return ret;
    }

    public static TreeMap<Integer, Vector<Integer>> parseCageTagFile(String fileStart, int chr, int isolatedIDs[], boolean sense, boolean convergentOnly, boolean modeOnly) {
	//Parse the file with 5'-nds for cage-tags centered around the
	//TSS. These were generated using CompareCageEstRNASeqStarts.java
	TreeMap<Integer, Vector<Integer>> fivePrimeEnds = new TreeMap<Integer, Vector<Integer>>();
	String fileName = fileStart + CompareCageEstRNASeqStarts.FIVE_PRIME_END_START_SUFFIX + chr + ".txt";
	int nUnique = 0, nGenes = 0, nTags = 0, nTagsInConvergentRegion = 0;
	//Count the number of genes with tags in the convergent region
	int nInConvergentRegion = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern fivePrimeEndPattern = Pattern.compile("^(\\d+)\\s+\\d+\\s+");
	    Pattern startPattern = Pattern.compile("([\\d|-]+)");
	    while(line!=null) {
		Matcher m = fivePrimeEndPattern.matcher(line);
		if (m.find()) {
		    int id = Integer.parseInt(m.group(1)), end = m.end();
		    Matcher m2 = startPattern.matcher(line);
		    Vector<Integer> ends = new Vector<Integer>();
		    boolean foundInConvergentRegion = false;
		    while (m2.find(end)) {
			int en = Integer.parseInt(m2.group(1));
			nTags++;
			if (!ends.contains(en)) {
			    if (convergentOnly) {
				if ((en<0 && sense) || (en>0 && !sense)) {
				    nUnique++; //Only add unique elements
				    ends.addElement(en);
				    if (inConvergentRegion(sense, en)) {
					nTagsInConvergentRegion++;
					if (!foundInConvergentRegion) { 
					    nInConvergentRegion++;
					    foundInConvergentRegion = true;
					}
				    }
				}
			    }
			    else {
				nUnique++; //Only add unique elements
				ends.addElement(en);
			    }
			}
			end = m2.end();
		    }
		    if (ends.size()>MIN_CAGE_TAGS_PER_TSS) { //Require a certain number of reads per TSS in order to be able to make the call
			nGenes++;
			if (modeOnly) {
			    int mostCommon = CompareCageEstRNASeqStarts.findMode(ends);
			    ends = new Vector<Integer>();
			    ends.addElement(mostCommon);
			}
			fivePrimeEnds.put(id, ends);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.out.println("Error " + e.toString() + " reading file " + fileName);
	}
	System.out.println("Found " + nUnique + " unique CAGE tags with " + nTagsInConvergentRegion + " in convergent region for " + nGenes + " genes with " + nInConvergentRegion + " genes having at least one tag in the convergent region.");
	return fivePrimeEnds;
    }
    
    public static int findNumberOfTSS(int ids[], int geneIDsKeep[], int isolatedIDs[]) {
	int nTSS = 0;
	if (geneIDsKeep!=null && isolatedIDs!=null)
	    for (int i=0; i<ids.length; i++) { if (Arrays.binarySearch(geneIDsKeep, ids[i])>=0 && Arrays.binarySearch(isolatedIDs, ids[i])>=0) { nTSS++; } }
	else if (geneIDsKeep!=null)
	    for (int i=0; i<ids.length; i++) { if (Arrays.binarySearch(geneIDsKeep, ids[i])>=0) { nTSS++; } }
	else if (isolatedIDs!=null)
	    for (int i=0; i<ids.length; i++) { if (Arrays.binarySearch(isolatedIDs, ids[i])>=0) { nTSS++; } }
	else
	    nTSS = ids.length;
	return nTSS;
    }

    public static void saveShuffleHistogram(TreeMap<Integer, Integer> hist, String saveFileName) throws IOException {
	System.out.println("Saving to " + saveFileName);
	BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	for (Integer i : hist.keySet()) 
	    bw.write(i + "\t" + hist.get(i) + "\n");
	bw.close();
    }


}
