//Use this class for a graph which contains genes and enhancers. Use
//one graph for each chromosome.


import edu.uci.ics.jung.graph.impl.*;
import edu.uci.ics.jung.graph.impl.BipartiteGraph;
import edu.uci.ics.jung.exceptions.*;
import edu.uci.ics.jung.utils.*;

import java.util.*;
import java.io.*;

public class GeneEnhancerGraph {

    protected int chr;
    //Stupid BipartiteGraph class does not work .....
    protected UndirectedSparseGraph graph = new UndirectedSparseGraph();

    protected Random random = new Random();

    ///////////////////////////////////////////////////////

    public GeneEnhancerGraph(GeneEnhancerGraph g) {
	//Copy constructor
	chr = g.chr;
	graph = (UndirectedSparseGraph)g.graph.copy();
    }

    public GeneEnhancerGraph(Vector<Enhancer> enhancers, RefSeqGenes geneInfo) {
	chr = geneInfo.getChromosome();
	for (Enhancer en : enhancers) {
	    if (en.getChromosome()==chr)
		graph.addVertex( en );
	}
	for (RefSeqGene g : geneInfo.getGenes())
	    graph.addVertex( g );
    }

    public GeneEnhancerGraph(Vector<Enhancer> enhancers, RefSeqGenes geneInfo, boolean nonGenic) {
	chr = geneInfo.getChromosome();
	for (Enhancer en : enhancers) {
	    if (en.getChromosome()==chr) {
		if ((!nonGenic && geneInfo.overlapsGene(en.getCenter())) || (nonGenic && !geneInfo.overlapsGene(en.getCenter())))
		    graph.addVertex( en );
	    }
	}
	for (RefSeqGene g : geneInfo.getGenes())
	    graph.addVertex( g );
    }

    public GeneEnhancerGraph(int chr, int geneIDsKeep[], int isolatedDist, String enhancerFileName, boolean nonGenic, String fileName, int windowSize, boolean orchestra) throws IOException {
	//Parse the enhancers and genes and add the read info in one fell swoop
	RefSeqGenes geneInfo = addGenes(chr, isolatedDist, geneIDsKeep, orchestra);
	//Add the enhancers
	Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	for (Enhancer en : enhancers) {
	    RefSeqGene g = geneInfo.getOverlappingGene(en.getCenter());
	    en.setOverlappingGene(g); //Set the genic status
	    if (en.getChromosome()==chr && !graph.getVertices().contains(en) && (!nonGenic || en.isNonGenic()))
		graph.addVertex( en );
	}
	System.out.println("Found " + getGenes().size() + " genes and " + getEnhancers().size() + " enhancers.");
	//Add the factors
	addReads(chr, fileName, windowSize);
    }

    public GeneEnhancerGraph(int chr, int geneIDsKeep[], int isolatedDist, String enhancerFileName, boolean nonGenic, String fileName, int windowSize, String randomization, boolean orchestra) throws IOException {
	//Parse the enhancers and genes and add the read info in one fell swoop
	RefSeqGenes geneInfo = addGenes(chr, isolatedDist, geneIDsKeep, orchestra);
	//Add enhancers based on some randomization scheme.
	Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	for (Enhancer en : enhancers) {
	    if (en.getChromosome()==chr) {
		if (randomization.equals("uniform")) { 
		    boolean success = false;
		    while (!success) {
			try {
			    graph.addVertex( new Enhancer(chr, geneInfo.getRandomRegion(windowSize, windowSize, random)) );
			    success = true;
			}
			catch (Exception e) {
			    System.err.println("Tried to add illegal enhancer.");
			}
		    }
		    //if (!nonGenic || en.isNonGenic())
		    //graph.addVertex( new Enhancer(chr, geneInfo.getRandomRegion(windowSize, windowSize, true, random)) );
		    //else if (nonGenic && en.isNonGenic())
		    //graph.addVertex( new Enhancer(chr, geneInfo.getRandomRegion(windowSize, windowSize, false, random)) );
		}
	    }
	}
	//System.out.println("Found " + getGenes().size() + " genes and " + getEnhancers().size() + " enhancers.");
	addReads(chr, fileName, windowSize); //Add the factors
    }

    ///////////////////////////////////////////////////////

    public RefSeqGenes addGenes(int chr, int isolatedDist, int[] geneIDsKeep, boolean orchestra) {
	RefSeqGenes geneInfo = new RefSeqGenes(chr, orchestra);
	int fragmentLength = 120;
	for (RefSeqGene g : geneInfo.getGenes()) {
	    if (geneIsolated(geneInfo, g, isolatedDist) && (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0))
		graph.addVertex( g );
	}
	return geneInfo;
    }

    public void addMiRNAInfo(Vector<MiRNA> miRNAs) {
	int n = graph.getVertices().size();
	for (MiRNA m : miRNAs) {
	    if (m.getChromosome()==chr) 
		graph.addVertex(m);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " miRNAs."); 
    }

    public void addLincRNAInfo(Vector<LncRNA> lincRNAs) {
	int n = graph.getVertices().size();
	for (LncRNA l : lincRNAs) {
	    if (l.getChromosome()==chr)
		graph.addVertex(l);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " lincRNAs.");
    }

    /////////////////////////////////

    public void addMiRNAReads(String fName, ABIRead readsForward[], ABIRead readsReverse[]) {
	for (MiRNA m : getMiRNAs()) {
	    if (m.getStrand())
		m.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, m.getStart(), m.getEnd(), readsForward));
	    else
		m.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, m.getEnd(), m.getStart(), readsReverse));
	    if (m.getFactorReads(fName).length>100)
		System.out.println(m.toString() + " has " + m.getFactorReads(fName).length + " reads for " + fName);
	}
    }

    public void addLincRNAReads(String fName, ABIRead readsForward[], ABIRead readsReverse[]) {
	for (LncRNA l : getLincRNAs()) {
	    l.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, l.getStart(), l.getEnd(), readsForward));
	    l.addFactorReads(fName, FindReadsAcrossFeature.getReads(chr, l.getStart(), l.getEnd(), readsReverse));
	    if (l.getFactorReads(fName).length>100)
		System.out.println(l.toString() + " has " + l.getFactorReads(fName).length + " reads for " + fName);
	}
    }

    public void addGeneReads(String fName, ABIRead readsForward[], ABIRead readsReverse[]) { addGeneReads(fName, readsForward, readsReverse, true); }

    public void addGeneReads(String fName, ABIRead readsForward[], ABIRead readsReverse[], boolean sense) {
	for (RefSeqGene g : getGenes()) {
	    if (g.getStrand())
		g.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, g.getStart(), g.getEnd(), sense ? readsForward : readsReverse));
	    else
		g.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, g.getEnd(), g.getStart(), sense ? readsReverse : readsForward));
	}
    }

    public void addGeneReads(String fName, ABIRead readsForward[], ABIRead readsReverse[], boolean tss, int upstream, int downstream) { addGeneReads(fName, readsForward, readsReverse, tss, upstream, downstream, true); }

    public void addGeneReads(String fName, ABIRead readsForward[], ABIRead readsReverse[], boolean tss, int upstream, int downstream, boolean sense) {
	for (RefSeqGene g : getGenes()) {
	    if (g.getStrand())
		g.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, (tss ? g.getStart() : g.getEnd()) - downstream, (tss ? g.getStart() : g.getEnd()) + upstream, sense ? readsForward : readsReverse));
	    else
		g.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, (tss ? g.getStart() : g.getEnd()) - upstream, (tss ? g.getStart() : g.getEnd()) + downstream, sense ? readsReverse : readsForward));
	}
    }

    public void addEnhancerReads(String fName, ABIRead readsForward[], ABIRead readsReverse[], int windowSize) {
	for (Enhancer en : getEnhancers()) {
	    en.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, en.getCenter() - windowSize, en.getCenter() + windowSize, readsForward));
	    en.addFactorReads(fName, FindReadsAcrossFeature.getReads(chr, en.getCenter() - windowSize, en.getCenter() + windowSize, readsReverse));
	}
    }

    public void addReads(int chr, String fileName, int windowSize) throws IOException {
	int fragmentLength = 120;
	Vector<String> fileNames = fileName.contains("names") ? new Vector<String>(FindOverlappingPeaks.getFileNames(fileName).keySet()) : new Vector<String>();
	if (!fileName.contains("names") && !fileName.equals(""))
	    fileNames.addElement(fileName);
	for (String fName : fileNames) {
	    boolean rna = fName.contains("RNA");
	    ABIRead reads[] = rna ? null : PeakFinder2.parseFiles(chr, fName, fragmentLength, true), readsForward[] = null, readsReverse[] = null;
	    if (rna) {
		readsForward = FindNongenicTranscripts.loadMRNAReadsChr(fName, true, chr);
		readsReverse = FindNongenicTranscripts.loadMRNAReadsChr(fName, false, chr);
	    }
	    for (Enhancer en : getEnhancers()) {
		if (rna) {
		    en.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, en.getCenter() - windowSize, en.getCenter() + windowSize, readsForward));
		    en.addFactorReads(fName, FindReadsAcrossFeature.getReads(chr, en.getCenter() - windowSize, en.getCenter() + windowSize, readsReverse));
		}
		else
		    en.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, en.getCenter() - windowSize, en.getCenter() + windowSize, reads));
	    }
	    if (rna) 
		addGeneReads(fName, readsForward, readsReverse);
	    else {
		for (RefSeqGene g : getGenes()) 
		    g.setFactorReads(fName, FindReadsAcrossFeature.getReads(chr, g.getStart() - windowSize, g.getStart() + windowSize, reads));
	    }
	    System.out.println("Loaded reads from chr " + chr + " for " + fName);
	}
    }

    public void removeAllEdges() { graph.removeAllEdges(); }

    public Vector<Enhancer> getEnhancers() { 
	Vector<Enhancer> ens = new Vector<Enhancer>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof Enhancer)
		ens.addElement((Enhancer)v);
	}
	return ens;
    }

    public Vector<RefSeqGene> getGenes(int geneIDs[]) {
	//only get the genes that are in the subset geneIDs
	Vector<RefSeqGene> genes = getGenes(), gs = new Vector<RefSeqGene>();
	if (geneIDs==null)
	    gs = genes;
	else {
	    for (RefSeqGene g : genes) {
		if (Arrays.binarySearch(geneIDs, g.getID())>=0)
		    gs.addElement(g);
	    }
	}
	return gs;
    }

    public Vector<RefSeqGene> getGenes() { 
	Vector<RefSeqGene> gs = new Vector<RefSeqGene>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof RefSeqGene)
		gs.addElement((RefSeqGene)v);
	}
	return gs;
    }

    public void addEdges(int minDist, int maxDist) {
	Vector<Enhancer> enhancers = getEnhancers();
	Collections.sort(enhancers);
	Vector<RefSeqGene> genes = getGenes();
	Collections.sort(genes);
	for (RefSeqGene g : getGenes()) {
	    for (Enhancer en : getEnhancers()) { 
		int dist = Math.abs(g.getStart() - en.getCenter());
		if (dist>minDist && dist<maxDist && g.findEdge(en)==null) {
		    try {
			graph.addEdge( new UndirectedSparseEdge( g, en ));
		    }
		    catch (ConstraintViolationException e) {
			System.out.println("Failed to add edge between\n" + g.toString() + "\n" + en.toString() + "\n" + e.toString());
			Map m = PredicateUtils.evaluateNestedPredicates(e.getViolatedConstraint(), new UndirectedSparseEdge( g, en ));
			Iterator it = m.keySet().iterator();
			while (it.hasNext()) {
			    Object o = it.next();
			    System.out.println(o.toString() + " -> " + m.get(o).toString());
			}
		    }
		}
	    }
	}
    }

    public Vector<Enhancer> getNearbyEnhancers(int geneIDs[]) {
	//get the enhancers that are connected to genes
	Vector<Enhancer> enhancers = new Vector<Enhancer>();
	for (RefSeqGene g : getGenes(geneIDs)) 
	    enhancers.addAll(new Vector<Enhancer>(g.getNeighbors()));
	return enhancers;
    }

    public Vector<RefSeqGene> getNearbyGenes(int geneIDs[]) {
	//get the genes that are connected to enhancers
	Vector<RefSeqGene> genes = new Vector<RefSeqGene>();
	for (Enhancer e : getEnhancers()) {
	    Vector<RefSeqGene> gs = new Vector<RefSeqGene>(e.getNeighbors());
	    if (geneIDs==null)
		genes.addAll(gs);
	    else {
		for (RefSeqGene g : gs) {
		    if (Arrays.binarySearch(geneIDs, g.getID())>=0)
			genes.addElement(g);
		}
	    }
	}
	return genes;
    }

    public TreeMap<Integer, Integer> getGeneEnhancerDistribution() { return getGeneEnhancerDistribution(null); }

    public TreeMap<Integer, Integer> getGeneEnhancerDistribution(int ids[]) {
	//Find the number of enhancers connected to each gene.
	TreeMap<Integer, Integer> dist = new TreeMap<Integer, Integer>();
	for (RefSeqGene g : getGenes()) {
	    if (ids==null || Arrays.binarySearch(ids, g.getID())>=0)
		dist.put(g.numNeighbors(), dist.containsKey(g.numNeighbors()) ? dist.get(g.numNeighbors()) + 1 : 1);
	}
	return dist;
    }

    public Enhancer getRandomEnhancer() { 
	Vector<Enhancer> enhancers = getEnhancers();
	return enhancers.elementAt(random.nextInt(enhancers.size()));
    }

    public static boolean geneIsolated(RefSeqGenes geneInfo, RefSeqGene g, int isolatedDist) {
	if (isolatedDist>0)
	    return geneInfo.getNearestGeneDistance(g.getStart())<isolatedDist && geneInfo.getNearestGeneDistance(g.getEnd())<isolatedDist;
	return true;
    }

    public Vector<Integer> getTranscribedRegionLengths(String fName, int minDist) {
	Vector<Integer> lens = new Vector<Integer>();
	for (Enhancer en : getEnhancers()) {
	    //if (en.numberOfReadsInTranscribedRegion(fName, true, minDist)>1)
		lens.addElement(en.getTranscribedRegionLength(fName, true, minDist));
		//if (en.numberOfReadsInTranscribedRegion(fName, false, minDist)>1)
		lens.addElement(en.getTranscribedRegionLength(fName, false, minDist));
	}
	return lens;
    }

    public Vector<Integer> getTranscribedRegionNumbers(String fName, int dist) {
	Vector<Integer> lens = new Vector<Integer>();
	for (Enhancer en : getEnhancers()) 
	    lens.addElement(en.numberOfReadsWithin(fName, dist));
	return lens;
    }

    public void removeEnhancers() {
	for (Enhancer en : getEnhancers())
	    graph.removeVertex(en);
    }

    public void addRandomEnhancers(boolean nonGenic, RefSeqGenes geneInfo) {

    }

    public Vector<LncRNA> getLincRNAs() {
	Vector<LncRNA> lincRNAs = new Vector<LncRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof LncRNA)
		lincRNAs.addElement((LncRNA)v);
	}
	return lincRNAs;
    }

    public Vector<MiRNA> getMiRNAs() {
	Vector<MiRNA> miRNAs = new Vector<MiRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof MiRNA)
		miRNAs.addElement((MiRNA)v);
	}
	return miRNAs;
    }

}