//Use this class for a graph which contains genes and transcripts. Use
//one graph for each chromosome.


import edu.uci.ics.jung.graph.impl.*;
import edu.uci.ics.jung.exceptions.*;
import edu.uci.ics.jung.utils.*;

import java.util.*;
import java.io.*;
import java.text.DecimalFormat;
import java.util.regex.*;

public class GeneTranscriptsGraph2 {

    public static final double MIN_READS_PROPORTION = .8;
    public static final double LIMIT_READS_PROPORTION = .1;

    //Use only one critical distance for all purposes
    protected int chr, cutOffDistance;
    protected boolean forward;
    protected RefSeqUCSCEnsemblGenes geneInfo;
    //Stupid BipartiteGraph class does not work .....
    protected UndirectedSparseGraph graph = new UndirectedSparseGraph();
    //Keep track of the genes that are not uniquely assigned to transcripts
    protected TreeSet<RefSeqUCSCEnsemblGene> mergedGenes = new TreeSet<RefSeqUCSCEnsemblGene>(), splitGenes = new TreeSet<RefSeqUCSCEnsemblGene>(), missedGenes = new TreeSet<RefSeqUCSCEnsemblGene>(), uniqueGenes = new TreeSet<RefSeqUCSCEnsemblGene>();
    protected TreeSet<Enhancer2> enhancers;

    protected Random random = new Random();
    public static DecimalFormat format = new DecimalFormat("###.#####");

    ///////////////////////////////////////////////////////

    public GeneTranscriptsGraph2(TreeSet<TranscribedRegion2> segments, RefSeqUCSCEnsemblGenes gi, boolean f, int maxDist) {
	chr = gi.getChromosome();
	geneInfo = gi;
	forward = f;
	cutOffDistance = maxDist;
	for (TranscribedRegion2 t : segments) 
	    addVertex(t);
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes())
	    addVertex( g );
    }

    public GeneTranscriptsGraph2(RefSeqUCSCEnsemblGenes gi, int c, boolean f, int d, TreeSet<TranscribedRegion2> segments, TreeSet<LncRNA> lncRNAs, TreeSet<MiRNA> miRNAs, TreeSet<MacroRNA> macroRNAs, TreeSet<Enhancer2> ens) { 
	geneInfo = gi;
	chr = c; 
	forward = f;
	cutOffDistance = d;
	for (TranscribedRegion2 t : segments) 
	    addVertex(t);
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes()) 
	    addVertex(g);
	if (lncRNAs!=null)
	    addLncRNAInfo(lncRNAs);
	if (miRNAs!=null)
	    addMiRNAInfo(miRNAs);
	if (macroRNAs!=null)
	    addMacroRNAInfo(macroRNAs);
	enhancers = ens;
	if (enhancers!=null)
	  addEnhancerInfo(enhancers);
    }

    ///////////////////////////////////////////////////////

    public TreeSet<RefSeqUCSCEnsemblGene> getSplitGenes() { return splitGenes; }

    public TreeSet<RefSeqUCSCEnsemblGene> getMergedGenes() { return mergedGenes; }

    public TreeSet<RefSeqUCSCEnsemblGene> getMissedGenes() { return missedGenes; }

    public TreeSet<RefSeqUCSCEnsemblGene> getUniqueGenes() { return uniqueGenes; }

    public boolean addVertex(UndirectedSparseVertex v) {
	HashSet<UndirectedSparseGraph> verts = new HashSet<UndirectedSparseGraph>(graph.getVertices());
	if (!verts.contains(v)) {
	    graph.addVertex(v);
	    //Add connections
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		for (TranscribedRegion2 t : getTranscripts()) {
		    if (assignEdge(g, t)) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(g, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + g.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof LncRNA) {
		LncRNA l = (LncRNA)v;
		for (TranscribedRegion2 t : getTranscripts()) {
		    if (l.overlaps(t.getStart(), t.getEnd())) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(l, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + l.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof MiRNA) {
		MiRNA m = (MiRNA)v;
		for (TranscribedRegion2 t : getTranscripts()) {
		    if (m.overlaps(t.getStart(), t.getEnd())) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(m, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + m.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof MacroRNA) {
		MacroRNA m = (MacroRNA)v;
		for (TranscribedRegion2 t : getTranscripts()) {
		    if (m.overlaps(t.getStart(), t.getEnd())) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(m, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + m.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof Enhancer2) {
		Enhancer2 e = (Enhancer2)v;
		for (TranscribedRegion2 t : getTranscripts()) {
		    if (e.overlaps(-cutOffDistance + (forward ? t.getStart() : t.getEnd()), cutOffDistance + (forward ? t.getStart() : t.getEnd()))) { //should start near enhancer
			try {
			    graph.addEdge(new UndirectedSparseEdge(e, t));
			}
			catch (Exception ex) { 
			    System.err.println("Failed to add edge between " + e.toString() + " and " + t.toString() + " because " + ex.toString()); 
			}
		    }
		}
	    }
	    return true;
	}
	return false;
    }

    public boolean assignEdge(RefSeqUCSCEnsemblGene g, TranscribedRegion2 t) {
	//Return true if we should put an edge between the gene and the transcript
	if (g.getStrand()==forward && g.getFactorReads().length>0)
	    return g.overlaps(t.getStart(), t.getEnd());
	else { //To find divergent transcripts, we must allow for transcripts starting ahead of the AS gene as well
	    if (forward) //Gene is on reverse strand
		return g.overlaps(t.getStart() - cutOffDistance, t.getEnd());
	    else
		return g.overlaps(t.getStart(), t.getEnd() + cutOffDistance);
	}
    }

    //////////////////////

    public void addMiRNAInfo(TreeSet<MiRNA> miRNAs) {
	int n = graph.getVertices().size();
	for (MiRNA m : miRNAs) {
	    if (m.getChromosome()==chr) 
		addVertex(m);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " miRNAs."); 
    }

    public void addMacroRNAInfo(TreeSet<MacroRNA> macroRNAs) {
	int n = graph.getVertices().size();
	for (MacroRNA m : macroRNAs) {
	    if (m.getChromosome()==chr) 
		addVertex(m);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " miRNAs."); 
    }

    public void addEnhancerInfo(TreeSet<Enhancer2> enhancers) {
	int n = graph.getVertices().size();
	for (Enhancer2 e : enhancers) {
	    if (e.getChromosome()==chr) 
		addVertex(e);
	}
	System.out.println("Added " + (graph.getVertices().size() - n) + " enhancers."); 
    }

    public void addLncRNAInfo(TreeSet<LncRNA> lncRNAs) {
	int n = graph.getVertices().size();
	for (LncRNA l : lncRNAs) {
	    if (l.getChromosome()==chr)
		addVertex(l);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " lincRNAs.");
    }

    /////////////////////////////////

    public void removeAllEdges() { graph.removeAllEdges(); }

    public void removeAllVertices() { graph.removeAllVertices(); }

    public void removeAll() { graph.removeAllEdges(); graph.removeAllVertices(); }

    public boolean contains(UndirectedSparseVertex v) { return graph.getVertices().contains(v); }

    public int numberOfGenesWithNTranscripts(int n, boolean ineq, boolean unique) {
	//Find the number of genes which are overlapped by n
	//transcripts. If the boolean is set, then we require that the
	//matching transcripts all have out-degree 1.
	int m = 0;
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		if ((!ineq && v.inDegree()==n) || (ineq && v.inDegree()>n)) {
		    if (!unique)
			m++;
		    else {
			Iterator<UndirectedSparseVertex> it = v.getNeighbors().iterator();
			boolean uniqueTranscripts = true;
			while (it.hasNext()) 
			    uniqueTranscripts = uniqueTranscripts && it.next().numNeighbors()==1;
			m += uniqueTranscripts ? 1 : 0;
		    }
		}
	    }
	}
	return m;
    }
    
    public int numberOfGenesWithOneTranscript() {
	//Make sure that the the majority of the gene is covered by
	//only one transcript and that the transcript in turn only
	//covers the minority of other genes.
	int m = 0, nGenes = 0;
	for (TranscribedRegion2 t : getTranscripts()) {
	    if (t.getType().equals(TranscribedRegion2.CODING_STR)) 
		m++;
	    /*if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		//check how much overlap each transcript has
		TranscribedRegion2 t = g.getUniqueTranscript();
		m += t!=null && t.inDegree()==1 ? 1 : 0;
		nGenes++;
		}*/
	}
	//System.out.println("Found " + nGenes + " genes and " + m + " unique.");
	return m;
    }

    public static int overlap(RefSeqUCSCEnsemblGene g, TranscribedRegion2 t) {
	int over = g.getStrand() ? Math.min(g.getEnd(), t.getEnd()) - Math.max(g.getStart(), t.getStart()) : Math.min(g.getStart(), t.getEnd()) - Math.max(g.getEnd(), t.getStart());
	//System.out.println("Checking overlap: " + g.toString() + "\t" + Math.min(g.getEnd(), t.getEnd()) + "-" + Math.max(g.getStart(), t.getStart()) + "=" + over + "\t" + t.toString());
	return over>0 ? over : 0;
    }

    public int numberOfTranscriptsWithOneGene() { return numberOfTranscriptsWithOneGene(true); }

    public int numberOfTranscriptsWithOneGene(boolean unique) { return numberOfTranscriptsWithNGenes(1, false, unique); }

    public int numberOfTranscriptsWithNGenes(int n, boolean ineq, boolean unique) {
	//Find the number of transcripts which are overlapped by n
	//genes. If the boolean unique is set, then we require that the
	//matching genes all have out-degree 1. If ineq is true, we
	//are interested in the case >=n.
	int m = 0;
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof TranscribedRegion2) {
		if ((!ineq && v.inDegree()==n) || (ineq && v.inDegree()>n)) {
		    //System.out.println(v.toString() + " overlaps " + v.outDegree());
		    if (!unique)
			m++;
		    else {
			Iterator<UndirectedSparseVertex> it = v.getNeighbors().iterator();
			boolean uniqueTranscripts = true;
			while (it.hasNext()) 
			    uniqueTranscripts = uniqueTranscripts && it.next().numNeighbors()==1;
			m += uniqueTranscripts ? 1 : 0;
		    }
		}
	    }
	}
	//System.out.println(m + " transcripts overlap only one gene.");
	return m;
    }

    public boolean overlapsMiRNA(TranscribedRegion2 t) { 
	for (MiRNA m : getMiRNAs()) {
	    if (m.getStrand()==forward && m.overlaps(t.getStart(), t.getEnd()))
		return true;
	}
	return false;
    }

    public Vector<MiRNA> getMiRNAs() { 
	Vector<MiRNA> ms = new Vector<MiRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof MiRNA)
		ms.addElement((MiRNA)v);
	}
	return ms;
    }

    public void findGenesPerTranscript() { 
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof TranscribedRegion2) {
		TranscribedRegion2 t = (TranscribedRegion2)v;
		if (t.numNeighbors()==0)
		    System.out.println(t.toString() + " does not overlap any genes.");
		else {
		    Iterator<UndirectedSparseVertex> it = t.getNeighbors().iterator();
		    while (it.hasNext()) {
			RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)it.next();
			System.out.println(t.toString() + " is overlapped for " + overlap(g, t) + "(" + format.format((double)overlap(g, t)/g.getLength()) + ") by " + g.toString());
		    }
		}
	    }
	}
    }

    public void findTranscriptsPerGene() {
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		if (g.numNeighbors()==0)
		    System.out.println(g.toString() + " does not overlap any transcripts.");
		else {
		    Iterator<UndirectedSparseVertex> it = g.getNeighbors().iterator();
		    while (it.hasNext()) {
			TranscribedRegion2 t = (TranscribedRegion2)it.next();
			System.out.println((g.hasUniqueTranscript() ? "* " : "  ") + g.toString() + " is overlapped for " + overlap(g, t) + "(" + format.format((double)overlap(g, t)/t.getLength()) + ") by " + t.toString());
		    }
		}
	    }
	}
    }
    
    public TreeSet<TranscribedRegion2> getTranscripts() { 
	TreeSet<TranscribedRegion2> ts = new TreeSet<TranscribedRegion2>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof TranscribedRegion2)
		ts.add((TranscribedRegion2)v);
	}
	return ts;
    }

    public Vector<RefSeqUCSCEnsemblGene> getGenes(int geneIDs[]) {
	//only get the genes that are in the subset geneIDs
	Vector<RefSeqUCSCEnsemblGene> genes = getGenes(), gs = new Vector<RefSeqUCSCEnsemblGene>();
	if (geneIDs==null)
	    gs = genes;
	else {
	    for (RefSeqUCSCEnsemblGene g : genes) {
		if (Arrays.binarySearch(geneIDs, g.getID())>=0)
		    gs.addElement(g);
	    }
	}
	return gs;
    }

    public Vector<RefSeqUCSCEnsemblGene> getGenes() { 
	Vector<RefSeqUCSCEnsemblGene> gs = new Vector<RefSeqUCSCEnsemblGene>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof RefSeqUCSCEnsemblGene)
		gs.addElement((RefSeqUCSCEnsemblGene)v);
	}
	return gs;
    }


    public static boolean geneIsolated(RefSeqUCSCEnsemblGenes geneInfo, RefSeqUCSCEnsemblGene g, int isolatedDist) {
	if (isolatedDist>0)
	    return geneInfo.getNearestGeneDistance(g.getStart())<isolatedDist && geneInfo.getNearestGeneDistance(g.getEnd())<isolatedDist;
	return true;
    }

    public boolean nearEnhancer(RefSeqUCSCEnsemblGenes geneInfo, TranscribedRegion2 t, int maxDist) { return nearEnhancer(this.chr, geneInfo, this.enhancers, t, maxDist); }

    public static boolean nearEnhancer(int chr, RefSeqUCSCEnsemblGenes geneInfo, TreeSet<Enhancer2> enhancers, TranscribedRegion2 t, int maxDist) {
	if (enhancers==null || t==null)
	    return false;
	for (Enhancer2 e : enhancers) {
	    RefSeqUCSCEnsemblGene overlappingGene = geneInfo.getOverlappingGene(e.getCenter());
	    //chromosome, distance and make sure anti-sense strand for intragenic ones
	    if (e.getChromosome()==chr && Math.abs(t.getStart() - e.getCenter())<maxDist) {
		if (overlappingGene==null) {
		    //System.out.println("Found eRNA " + t.toString() + " near enhancer " + e.toString());
		    return true;
		}
		else if (overlappingGene.getStrand()!=t.getStrand())
		    return true;
	    }
	}
	return false;
    }

    public void categorizeTranscribedRegions() {
	//Go through all the transcripts and try to classify them
	for (TranscribedRegion2 t : getTranscripts()) {
	    TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads = t.getOverlappingGenes(true);
	    TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps = t.getOverlappingGeneBps();
	    if (isUniqueCoding(genes2reads, genes2bps, t, false)) 
		t.setType(TranscribedRegion2.CODING_STR);
	    else if (isUniqueCoding(genes2reads, genes2bps, t, true)) 
		t.setType(TranscribedRegion2.NON_CODING_ANNOTATED_STR);
	    else if (isMultiGene(genes2reads, genes2bps, t)) {
		t.setType(TranscribedRegion2.MULTI_GENE_STR);
		mergedGenes.addAll(genes2reads.keySet());
		//System.out.println(t.toString() + " is " + TranscribedRegion2.MULTI_GENE_STR);
	    }
	    else if (isCodingFragment(genes2reads, genes2bps, t)) {
		t.setType(TranscribedRegion2.CODING_FRAGMENT_STR);
		splitGenes.addAll(genes2reads.keySet());
	    }
	    else { //Non-coding, try to determine what type
		//Only need to find out if we are starting near a TSS
		TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS = t.getOverlappingGenes(false);
		RefSeqUCSCEnsemblGene asGene = getDivergent(genes2readsAS, t, true, cutOffDistance);
		if (asGene!=null) {
		    t.setType(TranscribedRegion2.PROMOTER_PROXIMAL_NCRNA_STR);
		    t.setDivergentGene(asGene);
		}
		else
		    t.setType(TranscribedRegion2.PROMOTER_DISTAL_NCRNA_STR);
	    }
	    //System.out.println(t.toString() + " has type " + t.getType() + " and " + genes2reads.size());
	}
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes(true, forward)) { //Find out which genes were missed and store those in a list
	    if (!uniqueGenes.contains(g) && !mergedGenes.contains(g) && !splitGenes.contains(g))
		missedGenes.add(g);
	}
    }

    public void categorizeUnannotatedTranscribedRegions(int maxDist, boolean allowERNA) {
	//Go through all the transcripts and try to classify them
	for (TranscribedRegion2 t : getTranscripts()) {
	    TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads = t.getOverlappingGenes(true);
	    TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps = t.getOverlappingGeneBps();
	    if (isUniqueCoding(genes2reads, genes2bps, t, false)) 
		t.setType(TranscribedRegion2.CODING_STR);
	    else if (isUniqueCoding(genes2reads, genes2bps, t, true)) 
		t.setType(TranscribedRegion2.NON_CODING_ANNOTATED_STR);
	    else if (isMultiGene(genes2reads, genes2bps, t)) {
		t.setType(TranscribedRegion2.MULTI_GENE_STR);
		mergedGenes.addAll(genes2reads.keySet());
		//System.out.println(t.toString() + " is " + TranscribedRegion2.MULTI_GENE_STR);
	    }
	    else if (isCodingFragment(genes2reads, genes2bps, t)) {
		t.setType(TranscribedRegion2.CODING_FRAGMENT_STR);
		splitGenes.addAll(genes2reads.keySet());
	    }
	    else { //Non-coding, try to determine what type
		TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS = t.getOverlappingGenes(false);
		RefSeqUCSCEnsemblGene asGene = getDivergent(genes2readsAS, t, true, maxDist);
		if (asGene!=null) {
		    t.setType(TranscribedRegion2.DIVERGENT_STR);
		    t.setDivergentGene(asGene);
		}
		else if (overlapsMiRNA(t)) 
		    t.setType(TranscribedRegion2.MI_RNA_STR);
		else {
		    TreeMap<MacroRNA, Double> macroRNA2reads = t.getOverlappingMacroRNAs();
		    if (isUniqueMacroRNA(macroRNA2reads, t) || isMacroRNAFragment(macroRNA2reads, t))
			t.setType(TranscribedRegion2.MACRO_RNA_STR);
		    else {
			TreeMap<LncRNA, Double> lncRNA2reads = t.getOverlappingLncRNAs();
			if (isUniqueLncRNA(lncRNA2reads, t) || isLncRNAFragment(lncRNA2reads, t))
			    t.setType(TranscribedRegion2.LNC_RNA_STR);
			else if (/*allowERNA && */nearEnhancer(geneInfo, t, maxDist)) 
			    t.setType(TranscribedRegion2.ERNA_STR);
			else {
			    asGene = getOverlappingAS(genes2readsAS, t);
			    if (asGene!=null) {
				t.setType(TranscribedRegion2.NON_CODING_AS_STR);
				t.setASGene(asGene);
			    }
			    else {
				t.setType(TranscribedRegion2.NON_CODING_OTHER_RNA_STR);
				System.out.println(t.toString());
			    }
			}
		    }
		}
	    }
	    //if (overlapsMiRNA(t)) 
		//System.out.println(t.toString() + " has type " + t.getType() + " and overlaps miRNA");
	    if (t.getStart()==24125292)
		System.out.println("Found " + t.toString());
	}
	for (RefSeqUCSCEnsemblGene g : geneInfo.getGenes(true, forward)) { //Find out which genes were missed and store those in a list
	    if (!uniqueGenes.contains(g) && !mergedGenes.contains(g) && !splitGenes.contains(g))
		missedGenes.add(g);
	}
    }

    public boolean isDivergent(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS, TranscribedRegion2 t, int maxDist) { return isDivergent(genes2readsAS, t, true, maxDist); }

    public boolean isDivergent(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS, TranscribedRegion2 t, boolean start, int maxDist) { return getDivergent(genes2readsAS, t, true, maxDist)!=null; }

    public RefSeqUCSCEnsemblGene getDivergent(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS, TranscribedRegion2 t, boolean start, int maxDist) {
	for (RefSeqUCSCEnsemblGene g : genes2readsAS.keySet()) {
	    if (forward) {
		if (start) {
		    if (Math.abs(t.getStart() - g.getStart())<maxDist)
			return g;
		}
		else {
		    if (Math.abs(t.getStart() - g.getEnd())<maxDist)
			return g;
		}
	    }
	    else {
		if (start) {
		    if (Math.abs(t.getEnd() - g.getStart())<maxDist)
			return g;
		}
		else {
		    if (Math.abs(t.getEnd() - g.getEnd())<maxDist)
			return g;
		}
	    }
	}
	return null;
    }

    public static boolean isERNA(TranscribedRegion2 t) {
	Iterator it = t.getNeighbors().iterator();
	while (it.hasNext()) {
	    if (it.next() instanceof Enhancer2)
		return true;
	}
	return false;
    }

    public static boolean isOverlappingAS(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS, TranscribedRegion2 t) { return genes2readsAS.size()>0;    }

    public static RefSeqUCSCEnsemblGene getOverlappingAS(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2readsAS, TranscribedRegion2 t) { return genes2readsAS.isEmpty() ? null : genes2readsAS.firstKey();    }
    
    public static boolean isUniqueCoding(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads, TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps, TranscribedRegion2 t, boolean nonCoding) {
	//Determine if t overlaps a gene on the sense strand and
	//fulfils all the additional criteria for being a unique
	//transcript.
	if (genes2reads.size()==1) {
	    //Make sure that we cover at least half of the reads or
	    //half the length, and that no other transcripts cover the
	    //gene
	    RefSeqUCSCEnsemblGene g = genes2reads.keySet().iterator().next();
	    if (genes2reads.get(g)>MIN_READS_PROPORTION && genes2bps.get(g)>MIN_READS_PROPORTION) 
		return g.numNeighbors()==1 && g.isNonCoding()==nonCoding;
	    return false;
	}
	else if (genes2reads.size()>1 && !nonCoding)
	    return !coversMultipleGenes(genes2reads, genes2bps, t, nonCoding);
	return false;
    }

    public static boolean coversMultipleGenes(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads, TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps, TranscribedRegion2 t, boolean nonCoding) {
	if (genes2reads.size()>1) {
	    //Make sure that it includes most of the reads and only a fraction for the others
	    Double[] readFractions = genes2reads.values().toArray(new Double[genes2reads.size()]);
	    Arrays.sort(readFractions);
	    Double[] bpFractions = genes2bps.values().toArray(new Double[genes2bps.size()]);
	    Arrays.sort(bpFractions);
	    //Make sure that at least one of them is the main transcript-gene pair
	    if (readFractions[readFractions.length-1]>MIN_READS_PROPORTION && bpFractions[bpFractions.length-1]>MIN_READS_PROPORTION) {
		//Make sure that all of the genes are overlapping and
		//that the amount of reads is meaningful
		Iterator it = genes2reads.keySet().iterator();
		RefSeqUCSCEnsemblGene g = null;
		while (it.hasNext()) {
		    g = (RefSeqUCSCEnsemblGene)it.next();
		    if (genes2reads.get(g)>MIN_READS_PROPORTION && genes2bps.get(g)>MIN_READS_PROPORTION)
			break;
		}
		while (it.hasNext()) {
		    RefSeqUCSCEnsemblGene g2 = (RefSeqUCSCEnsemblGene)it.next();
		    if (!g.overlaps(g2) && genes2reads.get(g2)>LIMIT_READS_PROPORTION && genes2bps.get(g2)>LIMIT_READS_PROPORTION && g.isNonCoding()==nonCoding)
			return true;
		}
		return false;
	    }
	}
	return false;
    }

    public static boolean isUniqueLncRNA(TreeMap<LncRNA, Double> lncRNA2reads, TranscribedRegion2 t) {
	return lncRNA2reads.size()==1 && lncRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
    }

    public static boolean isUniqueMacroRNA(TreeMap<MacroRNA, Double> macroRNA2reads, TranscribedRegion2 t) {
	return macroRNA2reads.size()==1 && macroRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
    }

    public static boolean isMultiGene(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads, TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps, TranscribedRegion2 t) { return coversMultipleGenes(genes2reads, genes2bps, t, false); }

    public static boolean isCodingFragment(TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads, TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps, TranscribedRegion2 t) {
	//Check if any of the overlapped genes has
	//more fragments
	for (RefSeqUCSCEnsemblGene g : genes2reads.keySet()) {
	    if (/*(genes2reads.get(g)<MIN_READS_PROPORTION && genes2bps.get(g)<MIN_READS_PROPORTION) ||*/ g.getNeighbors().size()>1)
		return true;
	}
	return false;
    }

    public static boolean isLncRNAFragment(TreeMap<LncRNA, Double> lncRNA2reads, TranscribedRegion2 t) {
	//Check if at most one gene is overlapped and if that gene has
	//more fragments
	if (lncRNA2reads.size()==1) 
	    return lncRNA2reads.keySet().iterator().next().numNeighbors()>1 || lncRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
	return false;
    }

    public static boolean isMacroRNAFragment(TreeMap<MacroRNA, Double> macroRNA2reads, TranscribedRegion2 t) {
	//Check if at most one gene is overlapped and if that gene has
	//more fragments
	if (macroRNA2reads.size()==1) 
	    return macroRNA2reads.keySet().iterator().next().numNeighbors()>1 || macroRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
	return false;
    }

}