//Use this method to generate a null distribution for trying to
//determine whether or not there is initiation at a given ChIP
//peak. The strategy is to generate a large number of randomized peaks
//and then calculate the statistic D = reads_to_the_left -
//reads_to_the_right for each strand. The background generated by this
//scheme will be used when classifying the actual peaks. 

//In the first version we will not account for exons, introns or
//non-genic regions. Moreover, we will place the peaks completely at
//random (only conserving the relative number per chromosome) and hope
//that the large number will provide enough coverage.

//Example:

//java -Xmx512m GeneratePeakmRNANullDist -n 100 -p all_tfs_names_combos.txt -m mrna_names.txt -w 1000

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class GeneratePeakmRNANullDist{

    public static String GENOME_LOC;
    public static final String FORWARD_NULL_DIST_SUFFIX = "_f_null_w_";
    public static final String REVERSE_NULL_DIST_SUFFIX = "_r_null_w_";

    public static DecimalFormat format = new DecimalFormat("###.##");
    public static Random random = new Random();

    protected static int medianPeakWidth;
    protected static double peakChrDistribution[] = new double[21];
    //We want the null distribution to compensate for these two constraints
    protected static Vector peak2PeakDists = new Vector(), peak2TSSDists = new Vector();
    protected static RefSeqGenes geneInfo[] = new RefSeqGenes[21]; //Information about all TSSs


    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    //If this flag is not used, the default is to create as
	    //many peaks as there are in the peaks-file.
	    CmdLineParser.Option nMCFlag = argParser.addIntegerOption('n', "nMC");
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option nullPeakFileNameFlag = argParser.addStringOption('N', "nullPeakFileName");
	    //File containing a list of all the mRNA data sets that we
	    //want to investigate.
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag if we want to explicitly conserve the
	    //peak-to-peak distance
	    CmdLineParser.Option conservePeak2PeakDistFlag = argParser.addBooleanOption('C', "conservePeak2PeakDist");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean conservePeak2PeakDist = argParser.getOptionValue(conservePeak2PeakDistFlag)!=null;
	    tmp = argParser.getOptionValue(nullPeakFileNameFlag);
	    String nullPeakFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/NullDistributions/" + peakFileName.substring(peakFileName.lastIndexOf("/") + 1, peakFileName.lastIndexOf("."));
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(nMCFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    //////////////////////////////////////////////////////////////////
	    for (int i=0; i<geneInfo.length; i++) { geneInfo[i] = new RefSeqGenes(i + 1, orchestra); }
	    calculatePeakStatistics(peakFileName);
	    if (nMC==-1) { nMC = peak2TSSDists.size(); }
	    //Save all the peaks which were used to compute the
	    //null-distribution in a format similar to that for the
	    //real peaks
	    if (chr==-1) {
		for (int j=1; j<=21; j++) { //Do one chromosome at a time
		    int nC = (int)Math.round(peakChrDistribution[j-1]*(double)nMC/2);
		    analyzeChromosome(j, nC, conservePeak2PeakDist, orchestra, windowSize, mRNAFileName, nullPeakFileName);
		}
	    }
	    else {
		int nC = (int)Math.round(peakChrDistribution[chr-1]*(double)nMC/2);
		analyzeChromosome(chr, nC, conservePeak2PeakDist, orchestra, windowSize, mRNAFileName, nullPeakFileName);
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()
    
    public static void analyzeChromosome(int chr, int nC, boolean conservePeak2PeakDist, boolean orchestra, int windowSize, String mRNAFileName, String nullPeakFileName) throws IOException {
	//Store the mRNA info in a treemap for quick access,
	//use an arraay with one map for each time point
	BufferedWriter nullPeakWriter = new BufferedWriter(new FileWriter(nullPeakFileName + "_" + chr + ".txt"));
	TreeMap rnaReadsMapForward = readMRNAData(mRNAFileName, true, chr);
	TreeMap rnaReadsMapReverse = readMRNAData(mRNAFileName, false, chr);
	//Store the null distribution
	TreeMap forwardNullDists = new TreeMap(), reverseNullDists = new TreeMap();
	if (chr==1) {
	    nullPeakWriter.write("#chr\tStart\tEnd\tGC\tMean cons 11\tStd dev cons 11\tMax cons 11\tMean cons 30\tStd dev cons 30\tMax cons 30\t");
	    Iterator it = rnaReadsMapForward.keySet().iterator(); //Write the mRNA headers
	    while (it.hasNext()) {
		String name = (String)it.next();
		nullPeakWriter.write(name + "_FL\t" + name + "_FP\t" + name + "_FR\t" + name + "_RL\t" + name + "_RP\t" + name + "_RR\t");
	    }
	    nullPeakWriter.write("\n");
	}
	//Choose location, make sure to avoid the unanotated
	//telomeric regions.
	int ret[][] = getRandomPeaks(chr, nC*2, conservePeak2PeakDist, orchestra);
	int peakLocs[] = ret[0], nGC[] = ret[1];
	String peakLocInfo[] = getPeakLocInfo(chr, peakLocs, nGC);
	//Also calculate the conservation score statistics for
	//random peaks.
	String peakConsInfo[] = getConservationData(chr, peakLocs, orchestra);
	//Find the number of reads within the cut-off, for
	//simplicity, we assume that the peak has width equal
	//to the median peak width.
	String peakmRNAInfo[] = getPeakmRNAData(rnaReadsMapForward, rnaReadsMapReverse, peakLocs, windowSize, forwardNullDists, reverseNullDists);
	for (int k=0; k<peakLocs.length; k++) {
	    if (nGC[k]>=0) {
		nullPeakWriter.write(peakLocInfo[k] + "\t" + peakConsInfo[k] + peakmRNAInfo[k] + "\n");
		//System.out.println(peakLocInfo[k] + "\t" + peakConsInfo[k] + "\t" + peakmRNAInfo[k] + "\n");
	    }
	}
	nullPeakWriter.close();
	System.out.println("Found " + (2*nC) + " peaks for chromosome " + chr);
	//saveNullDistribution("mRNA/NullDistributions/", windowSize, forwardNullDists, true);
	//saveNullDistribution("mRNA/NullDistributions/", windowSize, reverseNullDists, false);
    }
    
    //////////////////////////////////////////////////////////////////
    
    public static String[] getPeakmRNAData(TreeMap rnaReadsMapForward, TreeMap rnaReadsMapReverse, int peakLocs[], int windowSize, TreeMap forwardNullDists, TreeMap reverseNullDists) {
	//Extract the number of reads under the peak as well as the
	//reads located upstream and downstream on each strand. Save
	//the relevant statistic in the nullDist variables and return
	//a string containing info about the peak which can be saved.
	String peakmRNAData[] = new String[peakLocs.length];
	Arrays.fill(peakmRNAData, new String(""));
	Iterator it = rnaReadsMapForward.keySet().iterator();
	while (it.hasNext()) {
	    String mRNAKey = (String)it.next();
	    TreeMap mRNAReadsForward = (TreeMap)rnaReadsMapForward.get(mRNAKey);
	    TreeMap mRNAReadsReverse = (TreeMap)rnaReadsMapReverse.get(mRNAKey);
	    for (int i=0; i<peakmRNAData.length; i++) {
		int nLeftForward = numberOfReads(mRNAReadsForward, peakLocs[i] - windowSize, peakLocs[i]), nRightForward = numberOfReads(mRNAReadsForward, peakLocs[i], peakLocs[i] + medianPeakWidth + windowSize);
		int nLeftReverse = numberOfReads(mRNAReadsReverse, peakLocs[i] - windowSize, peakLocs[i]), nRightReverse = numberOfReads(mRNAReadsReverse, peakLocs[i], peakLocs[i] + medianPeakWidth + windowSize);
		int nPeakForward = numberOfReads(mRNAReadsForward, peakLocs[i], peakLocs[i] + windowSize), nPeakReverse = numberOfReads(mRNAReadsReverse, peakLocs[i], peakLocs[i] + windowSize);
		//Create a string for saving
		peakmRNAData[i] += "\t" + nLeftForward + "\t" + nPeakForward + "\t" + nRightForward + "\t" + nLeftReverse + "\t" + nPeakReverse + "\t" + nRightReverse;
		//Calculate the difference between the right and left
		//hand sides and store in the histogram
		Integer diffForward = new Integer(nLeftForward - nRightForward), diffReverse = new Integer(nLeftReverse - nRightReverse);
		//Save for the appropriate null distribution
		/*TreeMap nullDistForward = forwardNullDists.containsKey(mRNAKey) ? (TreeMap)forwardNullDists.get(mRNAKey) : new TreeMap();
		TreeMap nullDistReverse = reverseNullDists.containsKey(mRNAKey) ? (TreeMap)reverseNullDists.get(mRNAKey) : new TreeMap();
		int n = nullDistForward.containsKey(diffForward) ? ((Integer)nullDistForward.get(diffForward)).intValue() : 0;
		nullDistForward.put(diffForward, new Integer(n + 1));
		forwardNullDists.put(mRNAKey, nullDistForward);
		n = nullDistReverse.containsKey(diffReverse) ? ((Integer)nullDistReverse.get(diffReverse)).intValue() : 0;
		nullDistReverse.put(diffReverse, new Integer(n + 1));
		reverseNullDists.put(mRNAKey, nullDistReverse);*/
	    }
	}
	return peakmRNAData;
    }
    
    public static String[] getConservationData(int chr, int peakLocs[], boolean orchestra) {
	//Returns strings with data to be saved for each peak
	PhastConsParser euarchontogliresCons = new PhastConsParser(chr, false, orchestra);
	double euarchontogliresConsScores[][] = euarchontogliresCons.findScores(peakLocs, medianPeakWidth);
	PhastConsParser verterbateCons = new PhastConsParser(chr, true, orchestra);
	double verterbateConsScores[][] = verterbateCons.findScores(peakLocs, medianPeakWidth);
	String peakConsInfo[] = new String[peakLocs.length];
	double euarchontogliresConsStats[][] = calculateConservationStats(euarchontogliresConsScores);
	double vertebrateConsStats[][] = calculateConservationStats(verterbateConsScores);
	for (int i=0; i<peakLocs.length; i++) 
	    peakConsInfo[i] = format.format(euarchontogliresConsStats[i][0]) + "\t" + format.format(euarchontogliresConsStats[i][1]) + "\t" + format.format(euarchontogliresConsStats[i][2]) + "\t" + format.format(vertebrateConsStats[i][0]) + "\t" + format.format(vertebrateConsStats[i][1]) + "\t" + format.format(vertebrateConsStats[i][2]);
	return peakConsInfo;
    }

    public static double[][] calculateConservationStats(double consScores[][]) {
	//Return the mean, std dev and max. For the mean and std dev,
	//only consider the bases which have a non-zero value.
	double ret[][] = new double[consScores.length][3];
	for (int i=0; i<consScores.length; i++) {
	    double m = 0, s = 0, max = 0, nNonZero = 0;
	    for (int j=0; j<consScores[i].length; j++) {
		if (consScores[i][j]>0) {
		    m += consScores[i][j];
		    nNonZero++;
		}
		max = max>consScores[i][j] ? max : consScores[i][j];
	    }
	    m /= nNonZero>0 ? nNonZero : 1;
	    for (int j=0; j<consScores[i].length; j++) 
		s += consScores[i][j]>0 ? (consScores[i][j] - m)*(consScores[i][j] - m) : 0;
	    s /= nNonZero>1 ? nNonZero - 1 : 1;
	    ret[i] = new double[]{m, Math.sqrt(s), max};
	}
	return ret;
    }

    public static String[] getPeakLocInfo(int chr, int peakLocs[], int nGC[]) {
	String peakLocInfo[] = new String[peakLocs.length];
	for (int i=0; i<peakLocInfo.length; i++) {
	    peakLocInfo[i] = chr + "\t" + peakLocs[i] + "\t" + (peakLocs[i] + medianPeakWidth) + "\t" + format.format((double)nGC[i]/medianPeakWidth);
	    //System.out.println(chr + "\t" + peakLocs[i] + "\t" + (peakLocs[i] + medianPeakWidth) + "\t" + format.format((double)nGC[i]/medianPeakWidth));
	}
	return peakLocInfo;
    }
    
    public static int[][] getRandomPeaks(int chr, int nPeaks, boolean conservePeak2PeakDist, boolean orchestra) {
	int ret[][] = new int[2][nPeaks], peakLocs[] = new int[nPeaks], nInvalid = 0;
	for (int i=0; i<nPeaks/2; i++) {
	    int tmp[] = getRandomPeakLocation(chr, conservePeak2PeakDist);
	    peakLocs[2*i] = tmp[0];
	    peakLocs[2*i + 1] = tmp[1];
	}
	Arrays.sort(peakLocs); //Sort the peaks so that they are in ascending order
	//Shift the peaks so that they do not overlap, move by a
	//random number of bps in the regime 10-100
	for (int i=1; i<peakLocs.length; i++) {
	    if (peakLocs[i]-peakLocs[i-1]<medianPeakWidth) {
		peakLocs[i] = peakLocs[i-1] + medianPeakWidth + random.nextInt(90) + 10;
		System.out.println("Moved peak to " + peakLocs[i]);
	    }
	}
	//Find the GC contents, some of the peaks may be invalid
	//having ended up in regions where reads cannot be
	//mapped. These will be kept for now and we make sure to track
	//how many there are.
	ret[0] = peakLocs;
	FastaSequence2 fastaSeq = new FastaSequence2(chr, orchestra);
	String seqs[] = fastaSeq.getSequences(peakLocs, medianPeakWidth);
	for (int i=0; i<seqs.length; i++) {
	    if (seqs[i].indexOf("N")==-1 && seqs[i].indexOf("n")==-1) 
		ret[1][i] = ChromosomeWindow.gcContent(seqs[i]);
	    else {
		nInvalid++;
		ret[1][i] = -1;
	    }
	}
	System.err.println("There were " + nInvalid + " invalid peaks.");
	return ret;
    }

    public static int[] getRandomPeakLocation(int chr, boolean conservePeak2PeakDist) {
	//We generate the peaks pairwise to compensate for the
	//interpeak distance distribution (to the first order at
	//least). For the first peak we want to control for the TSS
	//distance as well.
	int peakLocs[] = new int[2];
	while (!peaksWithinChromosome(peakLocs, chr)) {
	    RefSeqGene g = geneInfo[chr - 1].getRandomGene(random);
	    //System.out.println("Random gene is " + g.toString());
	    peakLocs[0] = g.getRandomTSS(random) + ((Integer)peak2TSSDists.elementAt(random.nextInt(peak2TSSDists.size()))).intValue();
	    if (conservePeak2PeakDist) //The second peak location is drawn from the peak2peak distance
		peakLocs[1] = peakLocs[0] + ((Integer)peak2PeakDists.elementAt(random.nextInt(peak2PeakDists.size()))).intValue();
	    else {
		g = geneInfo[chr - 1].getRandomGene(random);
		peakLocs[1] = g.getRandomTSS(random) + ((Integer)peak2TSSDists.elementAt(random.nextInt(peak2TSSDists.size()))).intValue();
	    }
	}
	return peakLocs;
    }

    protected static boolean peaksWithinChromosome(int peakLocs[], int chr) {
	for (int i=0; i<peakLocs.length; i++) {
	    if (peakLocs[i]<=0 || peakLocs[i]>PeakFinder2.CHROMOSOME_LENGTH[chr - 1])
		return false;
	}
	return true;
    }

    public static int getRandomChromosome(int p[]) {
	//p is an array with increasing numbers
	int r = random.nextInt(p[p.length - 1]), i = 0, c = 0;
	while (c<r && i<p.length) 
	    c += p[i++];
	return i;
    }

    //////////////////////////////////////////////////////////////////

    public static void saveNullDistribution(String fileStart, int windowSize, TreeMap nullDists, boolean forward) {
	Iterator distIt = nullDists.keySet().iterator();
	while (distIt.hasNext()) {
	    String name = (String)distIt.next();
	    TreeMap nullDist = (TreeMap)nullDists.get(name);
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(fileStart + name + (forward ? FORWARD_NULL_DIST_SUFFIX : REVERSE_NULL_DIST_SUFFIX) + windowSize + ".txt"));
		Iterator it = nullDist.keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    bw.write(key + "\t" + (Integer)nullDist.get(key) + "\n");
		}
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving file : " + fileStart);
	    }
	}
    }

    //////////////////////////////////////////////////////////////////

    public static void calculatePeakStatistics(String fileName) {
	//Count the total number of peaks for each chromosome
	int lengths[] = new int[21], prevPos = -1, prevPrevPos = -1, prevChr = -1;
	Vector widths = new Vector();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern peakWidthPattern = Pattern.compile("^(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = peakWidthPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1)), start = Integer.parseInt(m.group(2)), end = Integer.parseInt(m.group(3));
		    lengths[chr - 1]++;
		    widths.addElement(new Integer(end - start));
		    int center = (start + end)/2;
		    int tss = geneInfo[chr - 1].getNearestTSS(center), peak2peakDist = 0;
		    peak2TSSDists.addElement(new Integer(center - tss));
		    if (chr!=prevChr) {
			prevPos = -1;
			prevPrevPos = -1;
		    }
		    if (prevPos!=-1) {
			if (prevPrevPos==-1)
			    peak2peakDist = prevPos - center;
			else 
			    peak2peakDist = Math.abs(prevPos - center)<Math.abs(prevPos - prevPrevPos) ? prevPos - center : prevPos - prevPrevPos;
			peak2PeakDists.addElement(new Integer(peak2peakDist));
		    }
		    prevPrevPos = prevPos;
		    prevPos = center;
		    prevChr = chr;
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file : " + fileName);
	}
	for (int i=0; i<lengths.length; i++) //Calculate the proportion of peaks
	    peakChrDistribution[i] = (double)lengths[i]/widths.size();
	Collections.sort(widths);
	medianPeakWidth = ((Integer)widths.elementAt(widths.size()/2)).intValue();
	System.out.println("Median width is " + medianPeakWidth);
	try { //Save the peak-2-peak distances and the peak2TSS distances
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + "_peak2peak_dists.txt"));
	    for (int i=0; i<peak2PeakDists.size(); i++)
		bw.write(peak2PeakDists.elementAt(i) + "\n");
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(fileName.substring(0, fileName.lastIndexOf(".")) + "_peak2tss_dists.txt"));
	    for (int i=0; i<peak2TSSDists.size(); i++)
		bw.write(peak2TSSDists.elementAt(i) + "\n");
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving peak distances");
	}
    }

    public static int numberOfReads(TreeMap readsMap, int start, int end) {
	//Count the number of reads in the given interval
	int nReads = 0;
	TreeMap subMap = new TreeMap(readsMap.subMap(new Integer(start), new Integer(end)));
	Iterator it = subMap.keySet().iterator();
	while (it.hasNext()) 
	    nReads += ((Integer)subMap.get((Integer)it.next())).intValue();
	return nReads;
    }

    public static TreeMap readMRNAData(String fileName, boolean forward, int chr) {
	//Return an array with one map for each chromosome. For now,
	//we ignore the splices since these are only a minor
	//proportion and not very important for what we are trying to
	//detect.
	TreeMap mRNAReads = new TreeMap();
	Vector fileNames = parseMRNAFileList(fileName);
	Iterator it = fileNames.iterator();
	while (it.hasNext()) {
	    String fName = (String)it.next();
	    mRNAReads.put(fName.substring(fName.lastIndexOf("/") + 1), readMRNAInfo(fName, forward, chr));
	}
	return mRNAReads;
    }

    public static TreeMap readMRNAInfo(String fName, boolean forward, int chr) {
	TreeMap reads = new TreeMap();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fName + ".tsv"));
	    String line = br.readLine();
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int c = Integer.parseInt(m.group(1));
		    if (chr==c) { //22 is the splice chromosome
			if ((m.group(3).equals("F") && forward) || (m.group(3).equals("R") && !forward)) {
			    //System.out.println(line);
			    reads.put(new Integer(m.group(2)), new Integer(m.group(4)));
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file : " + fName);
	}
	return reads;
    }

    public static Vector parseMRNAFileList(String fileListName) {
	System.out.println("Parsing " + fileListName);
	Vector fileNames = new Vector();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileListName));
	    String line = br.readLine();
	    while (line!=null) {
		fileNames.addElement(line);
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing mrna file names : " + fileListName + ": " + e.toString());
	}
	return fileNames;
    }

}
