//Use this method to calculate the correlation between the amount of
//Pol2 at the initiation (fragments overlapping TSS) and the observed
//amount of transcribed RNA.

//Example:

//java -Xmx512m InitiationExpressionCorrelation -n 8WG16/un/8WG16_un -s mRNA/hr0_segregated_by_tRNA/Convergent.tsv -m mRNA/combined_hr0_20080805.unique__Genes-FEATURES-READS-r35_x500_N10M__ALL-GENES__EXN-ALL_INT-ALL_SPL

//Use initiation_expression_correlation.pl to run the batch job on orchestra.

//Use initiation_expression_correlation.m to analyze the data


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import org.apache.commons.math.stat.regression.SimpleRegression;

public class InitiationExpressionCorrelation{

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('i', "isolated");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //Use this flag if we want to normalize the Pol2 counts
	    //with respect to the GC-content
	    CmdLineParser.Option gcNormalizeFlag = argParser.addBooleanOption('g', "gcNormalize");
	    //Use this flag if we want to look for the max donwstream
	    //of the TSS instead of using the height above the TSS.
	    CmdLineParser.Option findPeakFlag = argParser.addBooleanOption('p', "findPeak");
	    //For determining the length of the region where we count
	    //the number of pol2 reads
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('d', "downstream");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(selectFlag);
	    String selectFileName = tmp!=null ? (String)tmp : null;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    boolean findPeak = argParser.getOptionValue(findPeakFlag)!=null;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21;
	    int fragmentLength = 120;
	    int isolatedIDs[] = isolated ? FindReadsAcrossFeature.parseIsolatedFile(orchestra) : null;
	    int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	    if (geneIDsKeep!=null) {
		Arrays.sort(geneIDsKeep);
		System.out.println("Only looking at " + geneIDsKeep.length);
	    }
	    String saveFileName = fileName + "_correlation_init_xpt" + (isolated ? "_isolated" : "") + (selectFileName!=null ? "_" + selectFileName.substring(selectFileName.lastIndexOf("/") + 1, selectFileName.lastIndexOf(".")) : "") + (chr!=-1 ? "_" + chr : "") + ".txt";
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    bw.write("#ID\tnTSS\txptDens\tnCoding\n");
	    SimpleRegression simpleRegression = new SimpleRegression();
	    if (chr==-1) {
		RefSeqGenes annotationInfo[] = new RefSeqGenes[nChr];
		for (int i=1; i<=21; i++) 
		    annotationInfo[i - 1] = new RefSeqGenes(i, orchestra);
		RNASeqData rnaSeqData = new RNASeqData(mRNAFileName, annotationInfo);
		Vector readsDens = new Vector();
		for (int i=1; i<=nChr; i++) {
		    double ret[][] = analyzeChromosome(i, fileName, annotationInfo[i-1], rnaSeqData, isolatedIDs, geneIDsKeep, fragmentLength, findPeak, downstream);
		    for (int j=0; j<ret[0].length; j++) {
			simpleRegression.addData(ret[1][j], ret[2][j]);
			bw.write(((int)ret[0][j]) + "\t" + ((int)ret[1][j]) + "\t" + format.format(ret[2][j]) + "\t" + ((int)ret[3][j]) + "\n");
		    }
		}
	    }
	    else {
		RefSeqGenes annotationInfo = new RefSeqGenes(chr, orchestra);
		RNASeqData rnaSeqData = new RNASeqData(mRNAFileName, annotationInfo, chr);
		double ret[][] = analyzeChromosome(chr, fileName, annotationInfo, rnaSeqData, isolatedIDs, geneIDsKeep, fragmentLength, findPeak, downstream);
		for (int j=0; j<ret[0].length; j++) {
		    simpleRegression.addData(ret[1][j], ret[2][j]);
		    bw.write(((int)ret[0][j]) + "\t" + ((int)ret[1][j]) + "\t" + format.format(ret[2][j]) + "\t" + ((int)ret[3][j]) + "\n");
		}
	    }
	    //Now we can calculate the correlation between the two vectors.
	    System.out.println("The correlation is " + format.format(simpleRegression.getR()));
	    bw.close();
	    System.out.println("Saved data to " + saveFileName);
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static double[][] analyzeChromosome(int chr, String fileName, RefSeqGenes annotationInfo, RNASeqData rnaSeqData, int isolatedIDs[], int geneIDsKeep[], int fragmentLength, boolean findPeak, int downstream) {
	TreeSet rnaSeq = rnaSeqData.getRNASeqData();
	int ids[] = annotationInfo.getIDs(), nTSS = FindReadsAcrossFeature.findNumberOfTSS(ids, geneIDsKeep, isolatedIDs);
	System.out.println("Found " + ids.length + " genes and " + nTSS);
	double readsOverlappingTSS[] = new double[nTSS], transcriptDensity[] = new double[nTSS];
	double geneIDs[] = new double[nTSS], nReads[] = new double[nTSS]; 
	double qualityScore = FindReadsAcrossFeature.QUALITY_SCORE_CUTOFF;
	ABIRead reads[] = PeakFinder2.parseFiles(chr, fileName, fragmentLength, true, 0, qualityScore);
	int tssCount = 0;
	Iterator it = rnaSeq.iterator();
	while (it.hasNext()) {
	    RNASeqGene g = (RNASeqGene)it.next();
	    if (g.getAnnotationInfo().getChromosome()==chr) {
		if ((geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getAnnotationInfo().getID())>=0) && (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getAnnotationInfo().getID())>=0)) {
		    TreeSet tssSort = new TreeSet(g.getAnnotationInfo().getStarts());
		    int tss = g.getAnnotationInfo().getStrand() ? ((Integer)tssSort.first()).intValue() : ((Integer)tssSort.last()).intValue();
		    //Find all the reads in the region surrounding this TSS
		    if (findPeak) {
			//try finding the maximum slightly downstream
			int h[] = FindReadsAcrossFeature.getReads(chr, tss, g.getAnnotationInfo().getStrand(), fragmentLength/2, fragmentLength/2, fragmentLength, reads);
			int ind = FindReadsAcrossFeature.findClosestMax(h, 0);
			readsOverlappingTSS[tssCount] = h[ind];
			//System.out.println("found " + readsOverlappingTSS[tssCount] + " and " + format.format(g.getTranscriptDensity()) + " for " + g.getAnnotationInfo().getName() + ", " + g.getAnnotationInfo().getID() + " at " + (ind - fragmentLength)); 
		    }
		    else
			readsOverlappingTSS[tssCount] = getReadsOverlapping(chr, tss, g.getAnnotationInfo().getStrand(), fragmentLength, reads);
		    if (downstream>0) { //Find the number of reads in a region downstream of the TSS
			nReads[tssCount] = (double)FindReadsAcrossFeature.numberOfReads(chr, tss + 1 + fragmentLength, g.getAnnotationInfo().getStrand(), 0, downstream, fragmentLength, reads);
		    }
		    transcriptDensity[tssCount] = g.getTranscriptDensity();
		    geneIDs[tssCount] = g.getAnnotationInfo().getID();
		    tssCount++;
		}
	    }
	}
	return new double[][]{geneIDs, readsOverlappingTSS, transcriptDensity, nReads};
    }

    public static int getReadsOverlapping(int chr, int tss, boolean forward, int fragmentLength, ABIRead reads[]) {
	//Find all the fragments which are physically overlapping the TSS
	int start = tss - fragmentLength, end = tss + 1;
	int ind = Arrays.binarySearch(reads, new ABIRead(chr, true, start, fragmentLength), null);
	ind = ind<0 ? -ind-1 : ind;
	int nReads = 0;
	while (ind<reads.length && reads[ind].getPosition()<end) {
	    nReads++;
	    ind++;
	}
	return nReads;
    }

}
