//Use this class to represent the tree spanned by enumerating all
//possible k-kers in the genome. This should be an efficient way for
//calculating the minimal distance between any pair of strings. It is
//relatively fast to figure out if a mer is unique (using bbbwt). Then
//we can use this scheme to figure out the distance between all unique
//mers.

import java.io.*;
import java.util.*;

public class MerTree {

    protected MerNode root = null;
    protected int maxDepth;
    protected boolean orchestra;


    public MerTree() { }

    public MerTree(boolean o) {
	orchestra = o;
    }

    //////////////////////////////////////////////////////////////////

    public int populateTree(int k, int chr, int start, int blockSize) {
	return populateTree(k, getGenomeFileName(chr), start, blockSize);
    }

    public String getGenomeFileName(int c) { 
	String GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	return GENOME_LOC + AnalyzeControls.getChromosomeString(c);
    }

    public int populateTree(int k, String fileName, int start, int blockSize) {
	//Create a tree from a subset of the genome [start, start+blockSize]
	System.out.println("Building tree with " + blockSize + " nodes for file " + fileName + " from " + start);
	maxDepth = k;
	root = new MerNode();
	int basesRead = 0, treeSize = 1, nwc = 0, nrep = 0, lineNo = 1; //count the repeated elements and wildcards
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //skip until we reach start
	    while (basesRead<start) { 
		char b = removeComments(br); 
		if (b=='\n') { lineNo++; }
		else { basesRead++; }
	    }
	    //Initialize
	    char mer[] = new char[k]; 
	    basesRead += initialize(mer, br);
	    for (int i=0; i<k; i++) {
		if (mer[i]=='n' || mer[i]=='N') { nwc++; }
		if (mer[i]=='a' || mer[i]=='c' || mer[i]=='g' || mer[i]=='t') { nrep++; }
	    }
	    //char rcmer[] = reverseComplement(mer);
	    while (br.ready() && treeSize<blockSize) {
		char b = removeComments(br);
		if (b!='\n') {
		    if (b=='n' || b=='N') { nwc++; }
		    if (b=='a' || b=='c' || b=='g' || b=='t') { nrep++; }
		    mer[k-1] = b; //Read a k-mer
		    //rcmer[0] = complement(b);
		    if (nwc==0/* && nrep<k*/) { //no wild cards and not all repetitive
			treeSize += addMer(mer); //Add it to the tree
			//addMer(rcmer);
		    }
		    if (mer[0]=='n' || mer[0]=='N') { nwc--; } //remove the first element
		    if (mer[0]=='a' || mer[0]=='c' || mer[0]=='g' || mer[0]=='t') { nrep--; }
		    basesRead++; 
		    for (int i=1; i<k; i++) { mer[i-1] = mer[i]; }
		    //remove the last element
		    //for (int i=k-1; i>0; i--) { rcmer[i] = mer[i-1]; }
		}
		else
		    lineNo++;
	    }
	    br.close();
	}
	catch(IOException e) { }
	catch(OutOfMemoryError e) {
	    System.err.println("Error after reading " + basesRead);
	}
	System.out.println("Read " + basesRead + " and " + lineNo + " lines.");
	return basesRead;
    }

    protected static int initialize(char[] mer, BufferedReader br) throws IOException {
	int basesRead = 0;
	while (basesRead<mer.length-1) {
	    char b = removeComments(br);
	    if (b!='\n') 
		mer[basesRead++] = b;
	}
	return basesRead;
    }

    public static char removeComments(BufferedReader br) throws IOException {
	char b = (char)br.read();
	if (b=='>') { //remove comments
	    while (br.ready() && b!='\n') { b = (char)br.read(); } 
	    b = (char)br.read(); //Read the char after \n
	}
	return b;
    }


    protected int addMer(char[] mer) {
	//returns the number of new nodes which were created
	root.nMers++;
	if (mer.length>0) {//Look at first base and send to that child
	    char nmer[] = new char[mer.length - 1];
	    for (int i=1; i<mer.length; i++) { nmer[i-1] = mer[i]; }
	    if (mer[0]=='a' || mer[0]=='A') 
		return root.aChild.addMer(nmer);
	    else if (mer[0]=='c' || mer[0]=='C') 
		return root.cChild.addMer(nmer);
	    else if (mer[0]=='g' || mer[0]=='G') 
		return root.gChild.addMer(nmer);
	    else if (mer[0]=='t' || mer[0]=='T') 
		return root.tChild.addMer(nmer);
	}
	return 0;
    }

    public static char complement(char b) {
	if (b=='a' || b=='A') { return 'T'; }
	else if (b=='c' || b=='C') { return 'G'; }
	else if (b=='g' || b=='G') { return 'C'; }
	else if (b=='t' || b=='T') { return 'A'; }
	return 'N';
    }

    public static char[] reverseComplement(char[] b) {
	char rc[] = new char[b.length];
	for (int i=0; i<b.length; i++) { rc[b.length-i-1] = complement(b[i]); }
	return rc;
    }

    public static String reverseComplement(String b) { return new String(reverseComplement(b.toCharArray())); }

    public HashMap findMinimumDistances(int k, int errorsAllowed, int chr, int start, int blockSize, HashMap minDists) { return findMinimumDistances(k, errorsAllowed, getGenomeFileName(chr), start, blockSize, minDists); }

    public HashMap findMinimumDistances(int k, int errorsAllowed, String fileName) { return findMinimumDistances(k, errorsAllowed, fileName, -Integer.MAX_VALUE, Integer.MAX_VALUE, new HashMap()); }

    public HashMap findMinimumDistances(int k, int errorsAllowed, String fileName, int start, int blockSize, HashMap minDists) {
	return findMinimumDistances(k, errorsAllowed, start, readGenome(fileName, start, blockSize), minDists);
    }
    public HashMap findMinimumDistances(int k, int errorsAllowed, int start, char bps[], HashMap minDists) {
	//Go through part of the genome between s and e and check the distance for each k-mer.
	char[] mer = new char[k], rcmer = new char[k];
	int repMers = 0, dist, rcdist, nwc = 0, nrep = 0, lineNo = 1;
	//Initialize
	for (int i=0; i<k-1; i++)
	    mer[i] = bps[i];
	for (int i=0; i<k-1; i++) {
	    if (mer[i]=='n' || mer[i]=='N') { nwc++; }
	    if (mer[i]=='a' || mer[i]=='c' || mer[i]=='g' || mer[i]=='t') { nrep++; }
	}
	rcmer = reverseComplement(mer);
	System.out.println(new String(mer));
	System.out.println(new String(rcmer));
	for (int j=k; j<bps.length; j++) {
	    //Read a k-mer
	    char b = bps[j];
	    if (b=='n' || b=='N') { nwc++; }
	    if (b=='a' || b=='c' || b=='g' || b=='t') { nrep++; }
	    mer[k-1] = b; //Read a k-mer
	    rcmer[0] = complement(b);
	    Integer key = new Integer(start+j-k);
	    MerInfo info = minDists.containsKey(key) ? (MerInfo)minDists.get(key) : new MerInfo();
	    if (nwc>0 || info.dist==0) //no need to compute
		dist = 0;
	    else {
		if (nrep<k) {
		    dist = root.minDistance(mer, 0, info.dist, errorsAllowed); //Find the distance and save
		    //System.out.println("Calculating min-dist for " + new String(mer) + " as " + dist);
		    if (dist==0 && info.foundOnce) 
			info.foundOnce = false;
		    else
			dist = dist<info.dist ? dist : info.dist;
		    rcdist = root.minDistance(rcmer, 0, dist, errorsAllowed);
		    //System.out.println("Calculating min-dist for " + new String(rcmer) + " as " + rcdist);
		    if (rcdist==0 && info.foundOnce) 
			info.foundOnce = false;
		    else
			dist = dist<rcdist ? dist : rcdist;
		    info.dist = dist;
		}
		else { repMers++; info.dist = 0; info.foundOnce = false; }
	    }
	    minDists.put(key, info);
	    //shift the array
	    if (mer[0]=='n' || mer[0]=='N') { nwc--; }
	    if (mer[0]=='a' || mer[0]=='c' || mer[0]=='g' || mer[0]=='t') {nrep--; }
	    for (int i=1; i<k; i++) { mer[i-1] = mer[i]; }
	    for (int i=k-1; i>0; i--) { rcmer[i] = rcmer[i-1]; }
	}
	return minDists;
    }

    public static char[] readGenome(String fileName, int start, int blockSize) {
	//Read blockSize bases from fileName, skipping everything until start
	//First we need to find out if we can read the maximal amount of chars
	int nBases = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //skip until we reach start
	    while (nBases<start && br.ready()) { 
		char b = removeComments(br); 
		if (b!='\n') { nBases++; }
	    }
	    while (nBases<blockSize+start && br.ready()) {
		char b = removeComments(br); 
		if (b!='\n') { nBases++; }
	    }
	    br.close();
	}
	catch (IOException e) { }
	char bs[] = new char[nBases];
	nBases = 0;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //skip until we reach start
	    while (nBases<start) { 
		char b = removeComments(br); 
		if (b!='\n') { nBases++; }
	    }
	    int i = 0;
	    while (i<bs.length) {
		char b = removeComments(br); 
		if (b!='\n') { bs[i++] = b; }
	    }
	    br.close();
	}
	catch (IOException e) { }
	return bs;
    }



    public int numberOfLeaves() { return root.numberOfLeaves(); }
    
    public int numberOfNodes() { return root.numberOfNodes(); }

    public int numberOfMers(String mer) { return numberOfMers(mer.toCharArray()); }

    public int numberOfMers(char[] mer) {
	//use this method to find how many instances there are of a given mer
	if (mer.length>maxDepth) {
	    System.err.println("Mer is too long! Max length is " + maxDepth + " and the length is " + mer.length);
	    return -1;
	}
	return root.numberOfMers(mer);
    }

    public String printBranch(char[] mer) {
	//Print the part of the tree which surrounds the branch of mer
	String as = "", cs = "", gs = "", ts = "";
	return root.printBranch(mer, as, cs, gs, ts);
    }


}



class MerNode {

    protected char b = 'r';
    public int nMers = 0; //Count the number of instances of the given k-mer
    public MerNode aChild = null, cChild = null, gChild = null, tChild = null;

    //Use this constructor for the root
    public MerNode() { 
	aChild = new MerNode('A');
	cChild = new MerNode('C');
	gChild = new MerNode('G');
	tChild = new MerNode('T');
    }

    public MerNode(char a) { b = a; }

    public boolean isLeaf() { return aChild==null && cChild==null && gChild==null && tChild==null; }


    public int addMer(char[] mer) {
	//go through tree recursively
	nMers++;
	//System.out.println("Adding " + new String(mer) + " to " + toString());
	if (mer.length>0) {//Look at first base and send to that child
	    char nmer[] = new char[mer.length - 1];
	    for (int i=1; i<mer.length; i++) { nmer[i-1] = mer[i]; }
	    if (mer[0]=='a' || mer[0]=='A') {
		if (aChild==null) { aChild = new MerNode('A'); } 
		return (nMers==1 ? 1 : 0) + aChild.addMer(nmer);
	    }
	    else if (mer[0]=='c' || mer[0]=='C') {
		if (cChild==null) { cChild = new MerNode('C'); } 
		return (nMers==1 ? 1 : 0) + cChild.addMer(nmer);
	    }
	    else if (mer[0]=='g' || mer[0]=='G') {
		if (gChild==null) { gChild = new MerNode('G'); } 
		return (nMers==1 ? 1 : 0) + gChild.addMer(nmer);
	    }
	    else if (mer[0]=='t' || mer[0]=='T') {
		if (tChild==null) { tChild = new MerNode('T'); } 
		return (nMers==1 ? 1 : 0) + tChild.addMer(nmer);
	    }
	}
	return nMers==1 ? 1 : 0;
    }

    public int numberOfChildren() {
	if (isLeaf()) { return 0; }
	else if (aChild!=null && cChild!=null && gChild!=null && tChild!=null)
	    return 4;
	else if ((aChild==null && cChild==null && gChild==null && tChild!=null) || (aChild==null && cChild==null && gChild!=null && tChild==null) || (aChild==null && cChild!=null && gChild==null && tChild==null) || (aChild!=null && cChild==null && gChild==null && tChild==null))
	    return 1;
	else if ((aChild==null && cChild!=null && gChild!=null && tChild!=null) || (aChild!=null && cChild==null && gChild!=null && tChild!=null) || (aChild!=null && cChild!=null && gChild==null && tChild!=null) || (aChild!=null && cChild!=null && gChild!=null && tChild==null))
	    return 3;
	else
	    return 2;
    }

    public int minDistance(char[] mer, int errors, int minErrors, int errorsAllowed){
	//go down the tree using a depth-first search. Stop when
	//there are more mismatches than for the best match so far.
	//System.out.println("Calculating min-dist for " + new String(mer) + " for " + toString());
	if (errors>=minErrors || errors>=errorsAllowed)
	    return errors; //abandon search
	else if (isLeaf()) {
	    if (errors==0)
		return nMers>1 ? 0 : Integer.MAX_VALUE; //mer is not unique or mer is unique
	    return errors;
	}
	else {
	    int aErr = Integer.MAX_VALUE, cErr = Integer.MAX_VALUE, gErr = Integer.MAX_VALUE, tErr = Integer.MAX_VALUE; //the errors for different choices of nucleotides at this position
	    char nmer[] = new char[mer.length - 1];
	    for (int i=1; i<mer.length; i++) { nmer[i-1] = mer[i]; }
	    //make sure to go for the most likely path first, that is
	    //the correct one.
	    if ((mer[0]=='a' || mer[0]=='A') && aChild!=null)
		aErr = aChild.minDistance(nmer, errors, minErrors, errorsAllowed);
	    else if ((mer[0]=='c' || mer[0]=='C') && cChild!=null)
		cErr = cChild.minDistance(nmer, errors, minErrors, errorsAllowed);
	    else if ((mer[0]=='g' || mer[0]=='G') && gChild!=null)
		gErr = gChild.minDistance(nmer, errors, minErrors, errorsAllowed);
	    else if ((mer[0]=='t' || mer[0]=='T') && tChild!=null)
		tErr = tChild.minDistance(nmer, errors, minErrors, errorsAllowed);
	    //do the other paths
	    minErrors = findMinError(aErr, cErr, gErr, tErr, minErrors);
	    if (aChild!=null && !(mer[0]=='a' || mer[0]=='A')) 
		aErr = aChild.minDistance(nmer, errors + 1, minErrors, errorsAllowed);
	    minErrors = minErrors>aErr ? aErr : minErrors;
	    if (cChild!=null && !(mer[0]=='c' || mer[0]=='C')) 
		cErr = cChild.minDistance(nmer, errors + 1, minErrors, errorsAllowed);
	    minErrors = minErrors>cErr ? cErr : minErrors;
	    if (gChild!=null && !(mer[0]=='g' || mer[0]=='G')) 
		gErr = gChild.minDistance(nmer, errors + 1, minErrors, errorsAllowed);
	    minErrors = minErrors>gErr ? gErr : minErrors;
	    if (tChild!=null && !(mer[0]=='t' || mer[0]=='T')) 
		tErr = tChild.minDistance(nmer, errors + 1, minErrors, errorsAllowed);
	    return minErrors>tErr ? tErr : minErrors;
	}
    }

    protected static int findMinError(int aErr, int cErr, int gErr, int tErr, int minErrors) {
	int min = aErr<cErr ? aErr : cErr; //Find the best candidate
	min = min<gErr ? min : gErr;
	min = min<tErr ? min : tErr;
	return min<minErrors ? min : minErrors;
    }


    public int numberOfLeaves() {
	if (isLeaf()) { return 1; }
	else { 
	    int n = 0;
	    if (aChild!=null) { n += aChild.numberOfLeaves(); }
	    if (cChild!=null) { n += cChild.numberOfLeaves(); }
	    if (gChild!=null) { n += gChild.numberOfLeaves(); }
	    if (tChild!=null) { n += tChild.numberOfLeaves(); }
	    return n;
	}
    }

    public int numberOfNodes() {
	if (isLeaf()) { return 1; }
	else { 
	    int n = 1;
	    if (aChild!=null) { n += aChild.numberOfNodes(); }
	    if (cChild!=null) { n += cChild.numberOfNodes(); }
	    if (gChild!=null) { n += gChild.numberOfNodes(); }
	    if (tChild!=null) { n += tChild.numberOfNodes(); }
	    return n;
	}
    }

    public int numberOfMers(char[] mer) {
	if (mer.length>0) {//Look at first base and send to that child
	    MerNode child = null;
	    if (mer[0]=='a' || mer[0]=='A') 
		child = aChild;
	    else if (mer[0]=='c' || mer[0]=='C') 
		child = cChild;
	    else if (mer[0]=='g' || mer[0]=='G') 
		child = gChild;
	    else if (mer[0]=='t' || mer[0]=='T') 
		child = tChild;
	    char nmer[] = new char[mer.length - 1];
	    for (int i=1; i<mer.length; i++) { nmer[i-1] = mer[i]; }
	    if (child!=null)
		return child.numberOfMers(nmer);
	    return -1; //No mers found
	}
	else
	    return nMers;
    }

    public String printBranch(char[] mer, String as, String cs, String gs, String ts) {
	as += aChild!=null ? "A" : " ";
	cs += cChild!=null ? "C" : " ";
	gs += gChild!=null ? "G" : " ";
	ts += tChild!=null ? "T" : " ";
	if (mer.length>0) {
	    char nmer[] = new char[mer.length - 1];
	    for (int i=1; i<mer.length; i++) { nmer[i-1] = mer[i]; }
	    if (mer[0]=='a' || mer[0]=='A') 
		return aChild.printBranch(nmer, as, cs, gs, ts);
	    else if (mer[0]=='c' || mer[0]=='C')
		return cChild.printBranch(nmer, as, cs, gs, ts);
	    else if (mer[0]=='g' || mer[0]=='G')
		return gChild.printBranch(nmer, as, cs, gs, ts);
	    else if (mer[0]=='t' || mer[0]=='T')
		return tChild.printBranch(nmer, as, cs, gs, ts);
	}
	return "";
    }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = "" + b + "( ";
	if (aChild!=null) { str += "a "; }
	if (cChild!=null) { str += "c "; }
	if (gChild!=null) { str += "g "; }
	if (tChild!=null) { str += "t "; }
	str += ") " + nMers;
	return str;
    }

}
