import java.util.*;

public class MiRNAInfoEntry implements Comparable {

    //Use this class to represent one miRNA gene. 

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    public int chr, start, end, id;
    public boolean strand;
    public String name;
    //Keep track of the peaks which are close to this gene
    protected TreeSet nearbyPeaks = new TreeSet();

    //////////////////////////////////////////////////////////////////

    public MiRNAInfoEntry(int c, int s, int e, boolean st, String n) {
	chr = c;
	start = s;
	end = e;
	strand = st;
	name = n;
    }

    public TreeSet findInternalPeaks(TreeSet regs) {
	TreeSet internal = new TreeSet();
	Iterator it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow er = (ChromosomeWindow)it.next();
	    if (peakInternal(er.getCenter()))
		internal.add(er);
	}
	return internal;
    }

    public boolean peakInternal(int peakPos) { return strand ? peakPos>start && peakPos<end : peakPos<start && peakPos>end; }

    public boolean peakUpstream(int peakPos) { return strand ? peakPos<start : peakPos>end; }

    public boolean peakDownstream(int peakPos) { return strand ? peakPos>start : peakPos<end; }

    public TreeSet findProximalPeaks(TreeSet regs, int cutOff) {
	TreeSet proximal = new TreeSet();
	Iterator it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow er = (ChromosomeWindow)it.next();
	    if (isProximal(er.getCenterOfMass(), cutOff))
		proximal.add(er);
	}
	return proximal;
    }

    public boolean isProximal(int peakPos, int cutOff) {
	//Proximal either to the 5' or 3' end
	if (strand)
	    return (start - peakPos>0 && start - peakPos<cutOff) || (peakPos - end>0 && peakPos - end<cutOff);
	else
	    return (end - peakPos>0 && end - peakPos<cutOff) || (peakPos - start>0 && peakPos - start<cutOff);
    }

    /*public TreeSet findNearestPeaks(TreeSet regs) {
	//Find those peaks which list this gene as their nearest one
	TreeSet set = new TreeSet();
	Iterator it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow er = (ChromosomeWindow)it.next();
	    if (er.getNearestGene().equals(this)) {
		set.add(er);
		System.out.println(er.toString() + " is close to " + name);
	    }
	}
	return set;
	}*/

    public int peakDistance(int peakPos) {
	//Returns the shortest distance to the peak from either the
	//transcription start site or stop site, depending on which
	//one is smaller. Negative values for genes which are upstream
	//and positive values for genes which are downstream.
	int startDist = strand ? peakPos - start : end - peakPos;
	int endDist = strand ? end - peakPos : peakPos - start;
	return Math.abs(startDist)<Math.abs(endDist) ? startDist : endDist;
    }

    public int getStart() { return start; }
    
    public int getEnd() { return end; }


    //////////////////////////////////////////////////////////////////

    public void addNearbyPeak(ChromosomeWindow p) { nearbyPeaks.add(p); }

    public TreeSet getNearbyPeaks() { return nearbyPeaks; }

    public boolean getStrand() { return strand; }

    public String getName() { return name; }

    public int getChromosome() { return chr; }

    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof MiRNAInfoEntry))
	    System.err.println("ERROR comparing genes!");
	MiRNAInfoEntry mi = (MiRNAInfoEntry)o;
	if (chr<mi.chr)
	    return -1;
	else if (chr>mi.chr)
	    return 1;
	else {
	    if (getStart()<mi.getStart())
		return -1;
	    else if (getStart()>mi.getStart())
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return name + "\t" + chr + "(" + (strand ? "+" : "-") + ")\t" + start + "-" + end;
    }

    public static String getHeaderString() {
	return "Name\tChr(Strand)\tLocation";
    }

    public boolean equals(Object o) {
	if (!(o instanceof MiRNAInfoEntry))
	    System.err.println("ERROR comparing genes!");
	MiRNAInfoEntry mi = (MiRNAInfoEntry)o;
	return mi.chr==chr && mi.start==start && mi.end==end && mi.name.equals(name) && mi.id==id;
    }

    public int hashCode() { return 13*start + 29*name.hashCode(); }

}
