//Use this class to represent a region in the genome which has one or
//more TFBS.

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;

public class OverlapRegion implements Comparable {
    //Use this class to represent a region where there are TFBS for
    //more than one factor.

    public static final String FORWARD_SUFFIX = "_F";
    public static final String REVERSE_SUFFIX = "_R";

    public static DecimalFormat format = new DecimalFormat("###.##");

    protected int start, end, chr;
    protected double gcContent, conservationScoresVertebrates[] = null, conservationScoresEuarchontoglires[] = null;
    //Store the sequence under the peak
    protected String sequence = null;
    //Key is the factor name and value is the number of reads. Ignore
    //input for now.
    protected TreeMap<String, Double> tfs = new TreeMap<String, Double>();
    //Use this to keep track of the mRNA reads
    protected TreeMap<String, Double> mRNAs = new TreeMap<String, Double>();
    //Keep track of the centers of each TF before merging
    protected TreeMap<String, Integer> tfCenters = new TreeMap<String, Integer>();

    //////////////////////////////////////////////////////////////////

    public OverlapRegion(Window win, String tfName, Vector ts, Vector ms) {
	chr = win.getChromosome();
	start = win.getStart();
	end = win.getEnd();
	gcContent = win.getGCContent();
	Iterator it = ts.iterator();
	while (it.hasNext())
	    tfs.put((String)it.next(), 0.0);
	tfs.put(tfName, new Double((double)win.numberOfReads()));
	tfs.put(tfName + "_input", new Double((double)win.numberOfInput()));
	//tfs.put(tfName, new Double(win.readsEnrichment()));
	tfCenters.put(tfName, win.getCenter());
	it = ms.iterator();
	while (it.hasNext()) { //Add one entry for each strand
	    String mRNA = (String)it.next();
	    mRNAs.put(mRNA + FORWARD_SUFFIX, 0.0);
	    mRNAs.put(mRNA + REVERSE_SUFFIX, 0.0);
	}
    }

    protected OverlapRegion(OverlapRegion r, Window win, String tfName) {
	//Copy-constructor
	chr = r.chr;
	start = r.start<win.getStart() ? r.start : win.getStart();
	end = r.end>win.getEnd() ? r.end : win.getEnd();
	gcContent = r.gcContent;
	conservationScoresEuarchontoglires = r.conservationScoresEuarchontoglires;
	conservationScoresVertebrates = r.conservationScoresVertebrates;
	sequence = r.sequence;
	Iterator it = r.tfs.keySet().iterator();
	while (it.hasNext()) {
	    String t = (String)it.next();
	    tfs.put(t, (Double)r.tfs.get(t));
	}
	//tfs.put(tfName, new Double(win.readsEnrichment()));
	tfs.put(tfName, new Double((tfs.containsKey(tfName) ? tfs.get(tfName) : 0) + win.numberOfReads()));
	tfs.put(tfName + "_input", new Double((tfs.containsKey(tfName + "_input") ? tfs.get(tfName + "_input") : 0) + (double)win.numberOfInput()));
	for (String k : r.tfCenters.keySet())
	    tfCenters.put(k, r.tfCenters.get(k));
	tfCenters.put(tfName, win.getCenter());
	it = r.mRNAs.keySet().iterator();
	while (it.hasNext()) {
	    String t = (String)it.next();
	    mRNAs.put(t, (Double)r.mRNAs.get(t));
	}
    }

    //dummy constructor, use when we want to find head or tail sets
    public OverlapRegion(int c, int s, int e, double g, Vector ts, Vector ms) {
	chr = c;
	start = s;
	end = e;
	gcContent = g;
	Iterator it = ts.iterator();
	while (it.hasNext())
	    tfs.put((String)it.next(), new Double(0));
	it = ts.iterator();
	while (it.hasNext())
	    tfCenters.put((String)it.next(), (s + e)/2);
	it = ms.iterator();
	while (it.hasNext()) {
	    String mRNA = (String)it.next();
	    mRNAs.put(mRNA + FORWARD_SUFFIX, new Double(0));
	    mRNAs.put(mRNA + REVERSE_SUFFIX, new Double(0));
	}
    }
    
    //////////////////////////////////////////////////////////////////

    public void merge(OverlapRegion r) {
	//System.out.println("Merging " + toString() + " and " + r.toString());
	start = r.getStart()<start ? r.getStart() : start;
	end = r.getEnd()>end ? r.getEnd() : end;
	for (String tfName : r.tfs.keySet()) {
	    tfs.put(tfName, new Double((tfs.containsKey(tfName) ? tfs.get(tfName) : 0) + r.tfs.get(tfName)));
	    if (tfs.containsKey(tfName + "_input") && r.tfs.containsKey(tfName + "_input"))
		tfs.put(tfName + "_input", new Double((tfs.containsKey(tfName + "_input") ? tfs.get(tfName + "_input") : 0) +  + r.tfs.get(tfName + "_input")));
	}
	for (String tfName : r.tfCenters.keySet()) 
	    tfCenters.put(tfName, r.tfCenters.get(tfName));
	//System.out.println("New is " + toString());
    }

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public int getCenter() { return (start + end)/2; }

    public int getLength() { return Math.abs(getEnd() - getStart()); }

    public double getGCContent() { return gcContent; }

    public void addTF(Window win, String tfName) { addTF(win, tfName, true); }

    public void addTF(Window win, String tfName, boolean significant) {
	if (chr==win.getChromosome()) {
	    gcContent = (gcContent*getLength() + win.getGCContent()*win.getLength())/(getLength() + win.getLength());
	    start = start<win.getStart() ? start : win.getStart();
	    end = end>win.getEnd() ? end : win.getEnd();
	    if (readsEnrichment(tfName)>0)
		System.out.println("Trying to add " + format.format(win.readsEnrichment()) + " " + tfName + " peaks to " + toString());
	    //tfs.put(tfName, new Double((significant ? win.readsEnrichment() : -win.readsEnrichment())));
	    tfs.put(tfName, new Double((significant ? win.numberOfReads() : -win.numberOfReads())));
	    tfs.put(tfName + "_input", new Double(significant ? (double)win.numberOfInput() : -(double)win.numberOfInput()));
	    tfCenters.put(tfName, win.getCenter());
	}
    }

    public int getTFBSDistance(String tf1, String tf2) {
	if (tfCenters.containsKey(tf1) && tfCenters.containsKey(tf2))
	    return Math.abs(tfCenters.get(tf1) - tfCenters.get(tf2));
	System.err.println("Error, trying to get distance for TF not found in peak");
	return 0;
    }

    public boolean overlaps(Window win) { return win.getChromosome()==chr && ((win.getStart()<=end && win.getStart()>=start) || (win.getEnd()<=end && win.getEnd()>=start) || (win.getEnd()>=end && win.getStart()<=start)); }

    public boolean overlaps(OverlapRegion r) { return r.chr==chr && ((r.start<=end && r.end>=start)); }

    public boolean overlaps(int s) { return start<=s && end>=s; }

    public int numberOfTFs() { 
	int n = 0;
	Iterator it = tfs.keySet().iterator();
	while (it.hasNext()) {
	    String tfName = (String)it.next();
	    n += !tfName.contains("input") && ((Double)tfs.get(tfName)).doubleValue()>0 ? 1 : 0;
	}
	return n;
    }

    //public int numberOfReads(String tfName) { return ((Integer)tfs.get(tfName)).intValue(); }

    public double readsEnrichment(String tfName) { return ((Double)tfs.get(tfName)).doubleValue(); }


    public void setConservationScores(double s[], boolean vertebrate) { 
	if (vertebrate) 
	    conservationScoresVertebrates = s;
	else
	    conservationScoresEuarchontoglires = s;
    }

    public double meanConservationScore(boolean vertebrate) {
	double[] conservationScores = vertebrate ? conservationScoresVertebrates : conservationScoresEuarchontoglires;
	double m = 0;
	int nNonZero = 0;
	for (int i=0; i<conservationScores.length; i++) {
	    if (conservationScores[i]>0) {
		m += conservationScores[i];
		nNonZero++;
	    }
	}
	return nNonZero>0 ? m/nNonZero : 0;
    }

    public double stdDevConservationScore(boolean vertebrate) {
	double[] conservationScores = vertebrate ? conservationScoresVertebrates : conservationScoresEuarchontoglires;
	double s = 0, m = meanConservationScore(vertebrate);
	int nNonZero = 0;
	for (int i=0; i<conservationScores.length; i++) {
	    if (conservationScores[i]>0) {
		s += (conservationScores[i] - m)*(conservationScores[i] - m);
		nNonZero++;
	    }
	}
	return nNonZero>1 ? Math.sqrt(s/(nNonZero - 1)) : Math.sqrt(s);
    }

    public double maxConservationScore(boolean vertebrate) {
	double[] conservationScores = vertebrate ? conservationScoresVertebrates : conservationScoresEuarchontoglires;
	double m = 0;
	for (int i=0; i<conservationScores.length; i++) 
	    m = Math.max(m, conservationScores[i]);
	return m;
    }

    //////////////////////////////////////////////////////////////////
    //mRNA stuff
    
    public void addMRNAReads(String mRNA, double n, boolean forward) {
	//System.out.println("Adding " + n + " to " + mRNA + (forward ? FORWARD_SUFFIX : REVERSE_SUFFIX) + ": " + toString());
	mRNA += forward ? FORWARD_SUFFIX : REVERSE_SUFFIX;
	double m = mRNAs.containsKey(mRNA) ? ((Double)mRNAs.get(mRNA)).doubleValue() : 0;
	mRNAs.put(mRNA, new Double(m + n));
	//System.out.println(toString());
    }

    //////////////////////////////////////////////////////////////////
    //Sequence stuff
    
    public void setSequence(String seq) { sequence = seq; }

    public String getSequence() { return sequence; }

    public boolean overlapsLocus(Vector<int[]> excludeLoci) {
	//excludeLoci are chr, start, end. Check if overlaps any
	for (int i=0; i<excludeLoci.size(); i++) {
	    if (excludeLoci.elementAt(i)[0]==chr && excludeLoci.elementAt(i)[1]<=end && start<=excludeLoci.elementAt(i)[2])
		return true;
	    if (excludeLoci.elementAt(i)[0]>chr)
		break;
	}
	return false;
    }
	    

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = chr + "\t" + start + "\t" + end + "\t" + format.format(gcContent);
	if (conservationScoresVertebrates!=null) 
	    str += "\t" + format.format(meanConservationScore(true)) + "\t" + format.format(stdDevConservationScore(true)) + "\t" + format.format(maxConservationScore(true));
	if (conservationScoresEuarchontoglires!=null) 
	    str += "\t" + format.format(meanConservationScore(false)) + "\t" + format.format(stdDevConservationScore(false)) + "\t" + format.format(maxConservationScore(false));
	Iterator it = tfs.keySet().iterator();
	while (it.hasNext()) 
	    str += "\t" + format.format(((Double)tfs.get((String)it.next())).doubleValue());
	it = mRNAs.keySet().iterator();
	while (it.hasNext()) 
	    str += "\t" + format.format(((Double)mRNAs.get((String)it.next())).doubleValue());
	return str;
    }

    public String toString(TreeMap<String, Double> normFactors) { return toString(normFactors, new TreeMap<String, String>()); }

    public String toString(TreeMap<String, Double> normFactors, TreeMap<String, String> tf2InputMap) { return toString(normFactors, tf2InputMap, false); }

    public String toString(TreeMap<String, Double> normFactors, TreeMap<String, String> tf2InputMap, boolean includeCenter) {
	String str = chr + "\t" + start + "\t" + end + "\t" + format.format(gcContent);
	if (conservationScoresVertebrates!=null) 
	    str += "\t" + format.format(meanConservationScore(true)) + "\t" + format.format(stdDevConservationScore(true)) + "\t" + format.format(maxConservationScore(true));
	if (conservationScoresEuarchontoglires!=null) 
	    str += "\t" + format.format(meanConservationScore(false)) + "\t" + format.format(stdDevConservationScore(false)) + "\t" + format.format(maxConservationScore(false));
	Iterator it = tfs.keySet().iterator();
	while (it.hasNext()) {
	    String tfName = (String)it.next();
	    if (!tfName.endsWith("_input")) {
		//System.out.println("Looking for normFactor for " + tfName + ": " + normFactors.get(tf2InputMap.get(tfName)) + " with " + tfs.get(tfName + "_input"));
		double inputNormalizedReads = 0, normalizedReads = ((Double)tfs.get(tfName)).doubleValue()*normFactors.get(tfName);
		//System.out.println(tfs.keySet().toString());
		//System.out.println(tf2InputMap.toString());
		if (tf2InputMap.containsKey(tfName)) {
		    if (tfs.containsKey(tfName + "_input") && tf2InputMap.get(tfName)!=null)
			inputNormalizedReads = ((double)tfs.get(tfName + "_input"))*normFactors.get(tf2InputMap.get(tfName));
			//else
			//System.err.println("Warning! Could not find input reads for " + tfName + " at " + chr + ":" + start + "-" + end);
		}
		else
		    System.err.println("Warning! Could not find input reads for " + tfName);
		str += "\t" + format.format(normalizedReads - inputNormalizedReads) + "\t" + (readsEnrichment(tfName)>0 ? tfCenters.get(tfName) : -1);
	    }
	}
	it = mRNAs.keySet().iterator();
	while (it.hasNext()) 
	    str += "\t" + format.format(((Double)mRNAs.get((String)it.next())).doubleValue());
	return str;
    }
    
    public boolean equals(Object o) {
	if (!(o instanceof OverlapRegion))
	    return false;
	OverlapRegion r = (OverlapRegion)o;
	//Check that they span the same region and contain the same
	//number of TFs
	return chr==r.chr && start==r.start && end==r.end; // && numberOfTFs()==r.numberOfTFs();
    }

    public int compareTo(Object o) {
	if (!(o instanceof OverlapRegion))
	    return 0;
	OverlapRegion r = (OverlapRegion)o;
	if (chr<r.chr)
	    return -1;
	else if (chr>r.chr)
	    return 1;
	else {
	    if (end<r.start)
		return -1;
	    else if (start>r.end)
		return 1;
	    else
		return 0;
	}
    }
    
}
    
