import java.util.*;
import edu.uci.ics.jung.graph.impl.*;

public class Peak extends UndirectedSparseVertex implements Comparable{

    public static final String CODING = "coding";
    public static final String NON_CODING = "noncoding";
    public static final String INTRAGENIC_CODING = "intragenic_coding";
    public static final String INTRAGENIC_NON_CODING = "intragenic_non_coding";
    public static final String UNANNOTATED_CODING = "unannotated_coding";
    public static final String UNANNOTATED_NON_CODING = "unannotated_non_coding";
    public static final String OTHER = "other";

    public static final double PEAK_CONSERVED_THRESHOLD = .5;
    
    protected int chr, start, end;
    protected boolean highlyCons;
    protected TreeMap<String, Double> factors = new TreeMap<String, Double>();
    //Use to keep track of the nearby transcripts
    protected TreeSet<TranscribedRegion2> nearbyTranscripts = new TreeSet<TranscribedRegion2>();
    protected String category;
    
    
    public Peak(int c, int s, int e, double cons) {
	chr = c;
	start = s;
	end = e;
	highlyCons = cons>PEAK_CONSERVED_THRESHOLD; //FindReadsAcrossFeature.HIGHLY_CONSERVED_THRESHOLD;
    }

    public Peak(int c, int s, int e) {
	chr = c;
	start = s;
	end = e;
	highlyCons = false;
    }

    ///////////////////////////////////////////////////////

    public void merge(Peak p) {
	//Use this method when merging two overlapping Encode peaks
	start = Math.min(p.getStart(), start);
	end = Math.max(p.getEnd(), end);
	for (String f : p.factors.keySet()) {
	    if (p.factors.get(f)>0)
		addFactor(f, p.factors.get(f));
	}
    }

    public void addFactor(String name, double e) { factors.put(name, e); }

    public int getCenter() { return (start + end)/2; }

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public void setConserved(boolean c) { highlyCons = c; }

    public boolean isConserved() { return highlyCons; }

    public int getChromosome() { return chr; }

    public double getFactor(String s) { 
	if (factors.containsKey(s)) //check for exact name
	    return factors.get(s);
	double bestEnrichment = 0; //check for partial matches
	for (String k : factors.keySet()) {
	    if (k.toLowerCase().contains(s.toLowerCase()))
		bestEnrichment = Math.abs(factors.get(k))>bestEnrichment ? factors.get(k) : bestEnrichment;
	}
	return bestEnrichment;
    }

    public boolean hasFactor(String s) { return hasFactor(s, false); }

    public boolean hasFactor(String s[]) { 
	for (int i=0; i<s.length; i++) {
	    if (hasFactor(s[i], false))
		return true;
	}
	return false;
    }

    public boolean hasAllFactors(LinkedHashSet<String> ss) { return hasAllFactors(ss, false); }
    
    public boolean hasAllFactors(String t, boolean exactMatch) { 
	LinkedHashSet<String> ss = new LinkedHashSet<String>();
	ss.add(t);
	return hasAllFactors(ss, exactMatch);
    }

    public boolean hasAllFactors(LinkedHashSet<String> ss, boolean exactMatch) { 
	for (String s : ss) {
	    if (!hasFactor(s, false))
		return false;
	}
	if (exactMatch) {
	    for (String k : factors.keySet()) {
		if (factors.get(k)>0 && !ss.contains(k))
		    return false;
	    }
	}
	return true;
    }

    public boolean hasAnyFactor(LinkedHashSet<String> ss) { 
	for (String s : ss) {
	    if (hasFactor(s, false))
		return true;
	}
	return false;
    }

    public boolean hasFactor(String s, boolean exact) {
	for (String k : factors.keySet()) {
	    if (!exact && k.toLowerCase().contains(s.toLowerCase()) && factors.get(k)>0)
		return true;
	    else if (exact && k.toLowerCase().equals(s.toLowerCase()) && factors.get(k)>0)
		return true;
	}
	return false;
    }

    public boolean hasFactor() {
	for (String k : factors.keySet()) {
	    if (factors.get(k)>0)
		return true;
	}
	return false;
    }

    public String[] getFactorNames() { return getFactorNames(true); }

    public String[] getFactorNames(boolean presentOnly) {
	TreeSet<String> names = new TreeSet<String>();
	for (String k : factors.keySet()) {
	    if (!presentOnly || factors.get(k)>0)
		names.add(k);
	}
	String ns[] = new String[names.size()];
	int i = 0;
	for (String n : names) { ns[i++] = n; }
	Arrays.sort(ns);
	return ns;
    }

    public TreeMap<String, Double> getFactors() { return factors; }

    public TreeMap<String, Double> getFactors(boolean presentOnly) { 
	if (!presentOnly)
	    return factors;
	else {
	    TreeMap<String, Double> fs = new TreeMap<String, Double>();
	    for (String tf : factors.keySet()) {
		if (factors.get(tf)>0)
		    fs.put(tf, factors.get(tf));
	    }
	    return fs;
	}
    }

    public void addFactors(TreeMap<String, Double> facs) { factors.putAll(facs); }

    public void removeFactor(String factor) { factors.remove(factor); }

    public boolean overlaps(int c) { return c>start && c<end; }

    public boolean overlaps(int st, int en) { 
	if (start==end)
	    return st<getCenter() && en>getCenter(); 
	else
	    return (start<=st && end>=st) || (start<=en && end>=en) || (start>=st && end<=en) || (st>=start && en<=end);
    }

    public boolean overlaps(Vector<Peak> peaks) {
	for (Peak p : peaks) {
	    if (p.getChromosome()==chr && p.overlaps(start, end))
		return true;
	}
	return false;
    }

    public boolean overlaps(int st, int en, int maxDist) { return st-maxDist<getCenter() && en+maxDist>getCenter(); }

    public boolean overlaps(TreeMap<Integer, Peak> peaks, int width) { return overlaps(peaks, null, width); }

    public boolean overlaps(TreeMap<Integer, Peak> peaks, String factor, int width) {
	//Use this method to check if this Peak overlaps with any of
	//the ones in peaks. We allow the peak to extend by width from
	//its center.
	if (factor==null)
	    return (new TreeMap<Integer, Peak>(peaks.subMap(getCenter() - width, getCenter() + width))).size()>0;
	TreeSet<Peak> sub = new TreeSet((new TreeMap<Integer, Peak>(peaks.subMap(getCenter() - width, getCenter() + width))).values());
	for (Peak p : sub) {
	    if (p.hasFactor(factor))
		return true;
	}
	return false;
    }

    public Peak getOverlappingPeak(TreeMap<Integer, Peak> peaks, String factor, int chr) { return getOverlappingPeak(peaks, factor, (end - start)/2, chr); }

    public Peak getOverlappingPeak(TreeSet<Peak> peaks) { return getOverlappingPeak(peaks, 0); }

    public Peak getOverlappingPeak(TreeSet<Peak> peaks, int maxDist) { 
	//Use this method to check if this Peak overlaps with any of
	//the ones in peaks. We allow the peak to extend by width from
	//its center.
	for (Peak p : peaks) {
	    if (p.getChromosome()==chr) {
		//if ((maxDist<=0 && p.overlaps(start, end)) || (maxDist>0 && p.overlaps(getCenter()-maxDist, getCenter()+maxDist)))
		if (p.overlaps(start - maxDist, end + maxDist))
		    return p;
	    }
	}
	return null;
    }

    public TreeSet<Peak> getAllOverlappingPeaks(TreeSet<Peak> peaks) { return getAllOverlappingPeaks(peaks, 0); }

    public TreeSet<Peak> getAllOverlappingPeaks(TreeSet<Peak> peaks, int maxDist) { 
	//Use this method to check if this Peak overlaps with any of
	//the ones in peaks. We allow the peak to extend by width from
	//its center.
	TreeSet<Peak> overlapping = new TreeSet<Peak>();
	for (Peak p : peaks) {
	    if (p.getChromosome()==chr) {
		//if ((maxDist<=0 && p.overlaps(start, end)) || (maxDist>0 && p.overlaps(getCenter()-maxDist, getCenter()+maxDist)))
		//if (p.getCenter()>2755000 && p.getCenter()<2756000)
		    //System.out.println("Merging " + p.toString() + " with " + Arrays.toString(p.getFactorNames(true)) + " overlapping " + start + "-" + end);
		if (p.overlaps(start - maxDist, end + maxDist))
		    overlapping.add(p);
	    }
	}
	return overlapping;
    }

    public Peak getOverlappingPeak(TreeMap<Integer, Peak> peaks, String factor, int width, int chr) { 
	//Use this method to check if this Peak overlaps with any of
	//the ones in peaks. We allow the peak to extend by width from
	//its center.
	if (factor==null) {
	    for (Peak p : peaks.values()) {
		if (p.getChromosome()==chr && p.overlaps(start, end))
		    return p;
	    }
	    /*if (peaks.size()==0) 
		return null;
	    TreeSet<Peak> sub = new TreeSet<Peak>((new TreeMap<Integer, Peak>(peaks.subMap(getCenter() - width, getCenter() + width))).values());
	    return sub!=null && sub.size()>0 ? sub.first() : null;*/
	}
	else {
	    TreeSet<Peak> sub = new TreeSet((new TreeMap<Integer, Peak>(peaks.subMap(getCenter() - width, getCenter() + width))).values());
	    for (Peak p : sub) {
		if (p.hasFactor(factor) && p.getChromosome()==chr)
		    return p;
	    }
	}
	return null;
    }

    public Peak getOverlappingPeak(TreeMap<Integer, Peak> peaks, String factor, int width, boolean up) {
	//Use this method to check if this Peak overlaps with any of
	//the ones in peaks. We allow the peak to extend by width from
	//its center.
	if (factor==null)
	    return (new TreeSet<Peak>((new TreeMap<Integer, Peak>(peaks.subMap(getCenter() - (up ? 0 : width), getCenter() + (up ? width : 0)))).values())).first();
	TreeSet<Peak> sub = new TreeSet((new TreeMap<Integer, Peak>(peaks.subMap(getCenter() - (up ? 0 : width), getCenter() + (up ? width : 0)))).values());
	for (Peak p : sub) {
	    if (p.hasFactor(factor))
		return p;
	}
	return null;
    }

    ////////////////////////////////////////////

    public void addTranscript(TranscribedRegion2 t) { nearbyTranscripts.add(t); }

    public TreeSet<TranscribedRegion2> getTranscripts() { return nearbyTranscripts; }

    public boolean hasBidirectionalTranscription() { return hasBidirectionalTranscription(Integer.MAX_VALUE); }

    public boolean hasBidirectionalTranscription(int maxDist) {
	boolean hasForward = false, hasReverse = false;
	for (TranscribedRegion2 t : nearbyTranscripts) {
	    if (t.getStrand() && Math.abs(getCenter() - t.getStart())<maxDist)
		hasForward = true;
	    else if (!t.getStrand() && Math.abs(getCenter() - t.getEnd())<maxDist)
		hasReverse = true;
	}
	return hasForward && hasReverse;
    }

    public boolean hasBidirectionalTranscription(int maxDist, boolean forwardAnnotated, boolean reverseAnnotated) {
	boolean hasForward = false, hasReverse = false;
	for (TranscribedRegion2 t : nearbyTranscripts) {
	    if (t.getStrand() && Math.abs(getCenter() - t.getStart())<maxDist && t.isAnnotated()==forwardAnnotated)
		hasForward = true;
	    else if (!t.getStrand() && Math.abs(getCenter() - t.getEnd())<maxDist && t.isAnnotated()==reverseAnnotated)
		hasReverse = true;
	}
	return hasForward && hasReverse;
    }

    public void setCategory(String cat) { category = cat; }

    public void setCategory(RefSeqUCSCEnsemblGenes geneInfo, int maxDist) { setCategory(geneInfo, maxDist, new HashSet<EST>()); }

    public void setCategory(RefSeqUCSCEnsemblGenes geneInfo, int maxDist, HashSet<EST> ests) {
	RefSeqUCSCEnsemblGene tssGene = geneInfo.getOverlappingTSS(getCenter() - maxDist, getCenter() + maxDist);
	if (tssGene!=null) {
	    if (tssGene.isNonCoding()) 
		category = NON_CODING;
	    else 
		category = CODING;
	}
	else {
	    RefSeqUCSCEnsemblGene overlappingGene = geneInfo.getOverlappingGene(getCenter(), getCenter());
	    if (overlappingGene!=null) {
		if (overlappingGene.isNonCoding()) 
		    category = INTRAGENIC_NON_CODING;
		else
		    category = INTRAGENIC_CODING;
	    }
	    else {
		for (EST est : ests) { //Check if the EST overlaps a TSS, that would suggest alternative promoter
		    tssGene = geneInfo.getOverlappingTSS(est.getStart(), est.getEnd());
		    if (tssGene!=null && tssGene.getStrand()==est.getStrand()) {
			if (tssGene.isNonCoding()) 
			    category = UNANNOTATED_NON_CODING;
			else {
			    category = UNANNOTATED_CODING;
			    break;
			}
		    }
		}
		if (category==null || category.length()==0) 
		    category = OTHER;
	    }
	}
    }

    public String getCategory() { return category; }

    public int categoryNumber() {
	if (category.equals(CODING))
	    return 1;
	else if (category.equals(NON_CODING))
	    return 2;
	else if (category.equals(INTRAGENIC_CODING))
	    return 3;
	else if (category.equals(INTRAGENIC_NON_CODING))
	    return 4;
	else if (category.equals(UNANNOTATED_CODING))
	    return 5;
	else if (category.equals(UNANNOTATED_NON_CODING))
	    return 6;
	else if (category.equals(OTHER))
	    return 7;
	else 
	    return 8;
    }

    ///////////////////////////////////////////////////////

    public boolean equals(Object o) {
	if (!(o instanceof Peak))
	    return false;
	Peak p = (Peak)o;
	return p.chr==chr && p.start==start && p.end==end;
    }

    public int compareTo(Object o) {
	if (!(o instanceof Peak))
	    return 0;
	Peak p = (Peak)o;
	if (this.chr<p.chr)
	    return -1;
	else if (this.chr>p.chr)
	    return 1;
	else {
	    if (getCenter()<p.getCenter())
		return -1;
	    else if (getCenter()>p.getCenter())
		return 1;
	    else 
		return 0;
	}
    }

    public String toString() {
	return "chr" + chr + ":" + start + "-" + end;
    }

    public int hashCode() { return 5*chr + 13*start + 29*end; }

}
	
