//Use this class to represent a Bayesian network for predicting
//whether or not a gene will be induced as a result of KCl binding. It
//has a pointer to a RefSeqGenesPeakGraph object which can be used to
//compute the likelihood.

import java.util.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;
import org.apache.commons.math.special.Gamma;


public class PeakGeneInductionNetwork {

    public static DecimalFormat format = new DecimalFormat("###.######");

    protected DirectedSparseGraph graph = new DirectedSparseGraph();
    protected TreeSet<String> tfNames = new TreeSet<String>();
    protected Vector<Integer> maxDists = new Vector<Integer>();
    //The graph representation of the peaks and the genes
    protected RefSeqGenesPeakGraph genesPeaksGraph;
    //Keep track of the genes that have RNAi and sufficiently expressed
    protected LinkedHashSet<RefSeqGene> genesWithExpAndRNAi = new LinkedHashSet<RefSeqGene>();
    //Keep track of how good of a model of the data the network provides
    protected double score, penalty, edgePenaltyCoeff, prior = 1, inductionThres, minExp;
    protected int distInc, maxDist;

    public PeakGeneInductionNetwork(RefSeqGenesPeakGraph g, int maxPeakDist, int proximalDist, double thres, double m, double penCost) {
	genesPeaksGraph = g;
	//This node represents the gene and it will not be the parent of any other node
	graph.addVertex(new TFPeakNode("", 0, 0));
	maxDists.addElement(0);
	maxDists.addElement(proximalDist);
	maxDist = maxPeakDist;
	distInc = proximalDist;
	inductionThres = thres;
	minExp = m;
	genesWithExpAndRNAi = RefSeqGenesPeakGraph.intersect(genesPeaksGraph.genesWithExpression(minExp), genesPeaksGraph.genesWithRNAi());
	System.out.println(genesWithExpAndRNAi.size() + " genes in the graph.");
	edgePenaltyCoeff = penCost;
	//Create one node for each TF
	for (String tf : genesPeaksGraph.getTFNames()) 
	    graph.addVertex(new TFPeakNode(tf, 0, proximalDist));
	//Find the best network for predicting
	findBestTopology();
	System.out.println(format.format(calculateLogLikelihood(inductionThres)) + "\n" + toString());
    }

    //////////////////////////////////////////////////////////////////

    public boolean findBestNetwork(int maxPeakDist) {
	//Do a greedy search. While there is still improvement keep
	//adding edges between nodes. Older edges are still viable
	//candidates for removal at each step. We sweep the distance
	//span incrementally, to find the best location for adding a
	//new distance cut-off. We start by creating one node for each TF

	double bestScore = score;
	int bestScorePartition = -1;
	LinkedHashSet<TFPeakEdge> bestEdges2add = new LinkedHashSet<TFPeakEdge>();
	//if (maxPeakDist<=maxDist && !maxDists.contains(maxPeakDist)) {
	    //Try adding a partition here
	    //Modify all the available nodes that will be affected by this partition
	    double oldScore = score;
	    DirectedSparseGraph oldGraph = insertDistancePartition(maxPeakDist);
	    LinkedHashSet<TFPeakEdge> edges2add = new LinkedHashSet<TFPeakEdge>();
	    while (findBestTopology(edges2add))
		System.out.println(format.format(bestScore) + "\n" + toString());
	    //}
	return true;
    }

    protected DirectedSparseGraph insertDistancePartition(int dist) {
	//Returns a copy of the old graph for easy re-construction
	DirectedSparseGraph oldGraph = (DirectedSparseGraph)graph.copy();
	//Go through all the vertices and split the ones that are
	//bisected by dist. The old nodes are assigned to the bin that
	//is closer to the TSS. We will have the option of pruning the
	//nodes later.
	Iterator it = graph.getVertices().iterator();
	LinkedHashSet<TFPeakNode> nodes2add = new LinkedHashSet<TFPeakNode>();
	while (it.hasNext()) {
	    TFPeakNode p = (TFPeakNode)it.next();
	    if (!p.isGene()) {
		if (p.getMinDistance()<dist && p.getMaxDistance()>dist) {
		    nodes2add.add(new TFPeakNode(p.getTFName(), dist, p.getMaxDistance()));
		    p.setMaxDistance(dist);
		    //re-compute the score for the edges to the node
		    Iterator edgeIt = p.getInEdges().iterator();
		    while (edgeIt.hasNext()) {
			TFPeakEdge e = (TFPeakEdge)edgeIt.next();
			e.setScore(-calculateDeltaLogLikelihood(e, inductionThres));
			System.out.println("Found score " + format.format(e.getScore()) + " for " + p.toString());
		    }
		    edgeIt = p.getOutEdges().iterator();
		    while (edgeIt.hasNext()) {
			TFPeakEdge e = (TFPeakEdge)edgeIt.next();
			e.setScore(-calculateDeltaLogLikelihood(e, inductionThres));
			System.out.println("Found score " + format.format(e.getScore()) + " for " + p.toString());
		    }
		}
	    }
	}
	for (String tf : genesPeaksGraph.getTFNames()) {
	    TFPeakNode n = new TFPeakNode(tf, dist, maxDist);
	    graph.addVertex(n);
	    System.out.println("Added " + n.toString());
	}
	for (TFPeakNode v : nodes2add)
	    graph.addVertex(v);
	return oldGraph;
    }

    protected void removeEdges(LinkedHashSet<TFPeakEdge> edges) {
	for (TFPeakEdge e : edges)
	    graph.removeEdge(e);
    }

    protected void addEdges(LinkedHashSet<TFPeakEdge> edges) {
	for (TFPeakEdge e : edges)
	    graph.addEdge(e);
    }

    public boolean findBestTopology() { return findBestTopology(new LinkedHashSet<TFPeakEdge>()); }

    public boolean findBestTopology(LinkedHashSet<TFPeakEdge> edges2add) {
	//Do a greedy search for the best topology. First we add the
	//best edge if that improves the performance. Then we consider
	//removing an edge. When no improvements are possible, we
	//return
	TreeMap<Double, TFPeakEdge> allPossibleEdges = new TreeMap<Double, TFPeakEdge>();
	Iterator nodeIt = graph.getVertices().iterator();
	DirectedSparseVertex exp = getGeneNode();
	score = 0;
	while (nodeIt.hasNext()) {
	    TFPeakNode p = (TFPeakNode)nodeIt.next();
	    if (!p.equals(exp)) {
		//Add to the gene node as well as to every other node (at the same distance)
		TFPeakEdge e = new TFPeakEdge(p, exp);
		//Calculate the score for this edge
		if (!hasEdge(e)) {
		    e.setScore(-calculateDeltaLogLikelihood(e, inductionThres));
		    System.out.println("Found score " + format.format(e.getScore()) + " for " + p.toString());
		    allPossibleEdges.put(e.getScore(), e);
		}
		Iterator nodeIt2 = graph.getVertices().iterator();
		while (nodeIt2.hasNext()) {
		    TFPeakNode o = (TFPeakNode)nodeIt2.next();
		    if (!o.isGene() && !p.equals(o)) {
			e = new TFPeakEdge(o, p);
			if (!hasEdge(e)) {
			    e.setScore(-calculateDeltaLogLikelihood(e, inductionThres));
			    System.out.println("Found score " + format.format(e.getScore()) + " for " + p.toString() + " to " + o.toString());
			    allPossibleEdges.put(e.getScore(), e);
			}
		    }
		}
	    }
	}
	//Go through the list of possible edges and add the most promising ones
	boolean modified = true, mod = false;
	while (!isConnected() || modified) {
	    modified = false;
	    if (!allPossibleEdges.isEmpty()) {
		TFPeakEdge e = allPossibleEdges.get(allPossibleEdges.firstKey());
		if (!isAcyclic(e)) {
		    graph.addEdge(e);
		    penalty = calculatePenaltyTerm();
		    score += e.getScore();
		    modified = true;
		    mod = true;
		    System.out.println("Added " + e.toString() + " with score " + format.format(e.getScore()) + " and penalty " + format.format(calculatePenaltyTerm()));
		}
		allPossibleEdges.remove(allPossibleEdges.firstKey());
	    }
	    else 
		break;
	    //Check if it would help to flip the direction of the lowest scoring edge

	    if (!modified) { //Check if it would help to remove the lowest scoring edge
		System.out.println("Removing an edge gives penalty reduction " + format.format(calculatePenaltyTermDec()));
		if (calculatePenaltyTermDec()-getLowestScore()<0) {
		    TFPeakEdge e = findLowestScoreEdge();
		    graph.removeEdge(e);
		    if (isConnected()) {
			penalty = calculatePenaltyTerm();
			modified = true;
			System.out.println("Removed " + e.toString() + " with score " + format.format(e.getScore()));
		    }
		    else
			graph.addEdge(e);
		}
	    }
	}
	return mod;
    }

    public boolean isConnected() {
	//Start from the gene node and visit all parents recursively and check if all of them can be reached
	TFPeakNode v = getGeneNode();
	HashSet<TFPeakNode> nodesFound = new HashSet<TFPeakNode>();
	getAllParents(v, nodesFound);
	//System.out.println("Found " + nodesFound.size() + " nodes connected to gene in graph with " + graph.numVertices());
	return nodesFound.size()==graph.numVertices();
    }

    public void getAllParents(TFPeakNode v, HashSet<TFPeakNode> vs) {
	Iterator it = v.getPredecessors().iterator();
	while (it.hasNext()) {
	    TFPeakNode n = (TFPeakNode)it.next();
	    if (!vs.contains(n)) {
		vs.add(n);
		getAllParents(n, vs);
	    }
	}
    }

    public boolean isAcyclic(TFPeakEdge e) {
	//Use Tarjan's algorithm to find strongly connected components. If such a component exists, then the graph has a cycle.
	boolean removed = containsEdge(e);
	if (!removed)
	    graph.addEdge(e);
	else
	    graph.removeEdge(e);
	Tarjan tarjan = new Tarjan();
	boolean cycle = tarjan.getLargestSCC(graph).size()>1;
	if (!removed)
	    graph.removeEdge(e);
	else
	    graph.addEdge(e);
	//System.out.println("Adding " + e.toString() + " gives cycle " + cycle);
	return cycle;
    }

    protected boolean containsEdge(TFPeakEdge e) {
	Iterator it = graph.getEdges().iterator();
	while (it.hasNext()) {
	    if (((TFPeakEdge)it.next()).equals(e))
		return true;
	}
	return false;
    }

    public TFPeakNode getGeneNode() {
	Iterator it = graph.getVertices().iterator();
	while (it.hasNext()) {
	    TFPeakNode p = (TFPeakNode)it.next();
	    if (p.isGene())
		return p;
	}
	return null;
    }

    public boolean hasEdge(TFPeakEdge e) {
	Iterator it = graph.getEdges().iterator();
	while (it.hasNext()) {
	    TFPeakEdge e2 = (TFPeakEdge)it.next();
	    if (e2.equals(e))
		return true;
	}
	return false;
    }

    public double getLowestScore() {
	//The score is the neg-log likelihood, hence we are looking for the largest positive value
	double minScore = 0;
	Iterator it = graph.getEdges().iterator();
	while (it.hasNext()) {
	    TFPeakEdge e = (TFPeakEdge)it.next();
	    minScore = e.getScore()>minScore ? e.getScore() : minScore;
	}
	return minScore;
    }

    public TFPeakEdge findLowestScoreEdge() {
	//The score is the neg-log likelihood, hence we are looking for the largest positive value
	double minScore = 0;
	TFPeakEdge minEdge = null;
	Iterator it = graph.getEdges().iterator();
	while (it.hasNext()) {
	    TFPeakEdge e = (TFPeakEdge)it.next();
	    if (e.getScore()>minScore) {
		minScore = e.getScore();
		minEdge = e;
	    }
	}
	//graph.removeEdge(minEdge);
	return minEdge;
    }

    //////////////////////////////////////////////////////////////////

    public TreeMap<RefSeqGene, Double> calculateGeneInductionProbability(boolean hasRNAi, double inductionThres) {
	TreeMap<RefSeqGene, Double> inductionProbs = new TreeMap<RefSeqGene, Double>();
	for (RefSeqGene g : genesPeaksGraph.getGenes(minExp, hasRNAi)) {
	    Vector<Peak> peaks = new Vector<Peak>();
	    Iterator it = g.getNeighbors().iterator();
	    while (it.hasNext()) {
		Object o = it.next();
		if (o instanceof Peak)
		    peaks.addElement((Peak)o);
	    }
	    inductionProbs.put(g, calculateGeneInductionProbability(g, peaks, inductionThres));
	}
	return inductionProbs;
    }

    public double calculateGeneInductionProbability(RefSeqGene g, Vector<Peak> peaks, double inductionThres) {
	//Find the predictions of the model for a given set of
	//peaks. 
	/*double ll = 0;
	HashMap<Integer, HashSet<String>> state = findState(peaks, g);
	//Now we can evaluate the probability for this event given the Bayesian network
	//First we evaluate the state at the gene-node.

	Iterator it = v.getPredecessors().iterator();
	while (it.hasNext()) {
	    TFPeakNode p = (TFPeakNode)it.next();
	    if (p.isGene()) {

	    }
	    else {
		theta = (prior + genes.size())/(prior*Math.pow(2, s.length) + genesWithExpAndRNAi.size());
	    }
	}
	return Math.exp(ll);*/
	return 0;
    }

    public boolean[] findState(TFPeakNode v, Vector<Peak> peaks, RefSeqGene g) {
	//Find out which state the configuration of peaks corresponds to
	int nParents = v.getPredecessors().size();
	boolean state[] = new boolean[nParents + 1];
	Arrays.fill(state, false);
	if (!v.isGene()) {
	    TFPeakNode p = (TFPeakNode)v;
	    for (Peak peak : peaks) {
		for (Integer s : g.getStarts()) {
		    if (peak.hasFactor(p.getTFName()) && (s - peak.getCenter())>=p.getMinDistance() && (s - peak.getCenter())<p.getMaxDistance())
			state[0] = true;
		}
	    }
	}
	else 
	    state[0] = true;
	Iterator it = v.getPredecessors().iterator();
	int i = 1;
	while (it.hasNext()) {
	    TFPeakNode p = (TFPeakNode)it.next();
	    for (Peak peak : peaks) {
		for (Integer s : g.getStarts()) {
		    if (peak.hasFactor(p.getTFName()) && (s - peak.getCenter())>=p.getMinDistance() && (s - peak.getCenter())<p.getMaxDistance())
			state[i] = true;
		}
	    }
	    i++;
	}
	return state;
    }

    //////////////////////////////////////////////////////////////////

    public double calculatePenaltyTerm() {
	return Math.log(graph.numEdges())*edgePenaltyCoeff;
    }

    public double calculatePenaltyTermInc() { return Math.log(graph.numEdges() + 1)*edgePenaltyCoeff; }

    public double calculatePenaltyTermDec() { return Math.log(graph.numEdges() - 0)*edgePenaltyCoeff; }

    public double calculateDeltaLogLikelihood(TFPeakEdge e, double inductionThres) {
	//Calculate the change from adding one edge
	/*double ll = 0;
	DirectedSparseVertex v = (DirectedSparseVertex)e.getDest();
	LinkedHashSet<TFPeakNode> parents = new LinkedHashSet<TFPeakNode>(v.getPredecessors());
	parents.add((TFPeakNode)e.getSource());
	int nParents = parents.size();
	HashMap<HashSet<TFPeakNode>, HashSet<TFPeakNode>> allStates = v.getAllStates();
	TFPeakNode p = (TFPeakNode)e.getDest();
	if (!p.isGene()) {
	    //Go through the parents
	    //Find the total number of genes with the given combination of TFs
	    int nGenesWithTFs = genesWithExpAndRNAi.size(); //RefSeqGenesPeakGraph.intersect(genesWithExpAndRNAi, genesPeaksGraph.genesWithPeakTFs(p.getTFNames(true))).size();
	    ll += org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size()) - org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size() + nGenesWithTFs);
	    //Go through the states
	    for (HashSet<TFPeakNode> tfsPresent : allStates) {
		HashSet<TFPeakNode> tfsAbsent = allStates.get(tfsPresent);
		int n = numberOfTFTFOccurences(p, parents, tfsPresent, tfsAbsent);
		ll += org.apache.commons.math.special.Gamma.logGamma(prior + n) - org.apache.commons.math.special.Gamma.logGamma(prior);
		//v.setTheta(tfsPresent, tfsAbsent, (double)(prior + n)/(prior*Math.pow(2, state.length) + genesWithExpAndRNAi.size())
	    }
	}
	else {
	    //Go through the parents
	    ll += org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size()) - org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size() + genesWithExpAndRNAi.size());
	    //Go through the states
	    for (boolean[] state : allStates) {
		int n = numberOfTFGeneOccurences(parents, state, inductionThres);
		ll += org.apache.commons.math.special.Gamma.logGamma(prior + n) - org.apache.commons.math.special.Gamma.logGamma(prior);
		//v.setTheta(state, (double)(prior + n)/(prior*Math.pow(2, state.length) + genesWithExpAndRNAi.size())
	    }
	}
	return ll;*/
	return 0;
    }

    public double calculateLogLikelihood(double inductionThres) {
	//Calculate the likelihood of the data for a fixed
	//network. This function implements eq (35) in Heckerman
	//(1996).
	/*double ll = 0;
	//Sum the likelihood for each node
	Iterator it = graph.getVertices().iterator();
	while (it.hasNext()) {
	    TFPeakNode p = (TFPeakNode)it.next();
	    LinkedHashSet<TFPeakNode> parents = new LinkedHashSet<TFPeakNode>(p.getPredecessors());
	    int nParents = parents.size();
	    if (!p.isGene()) {
		LinkedHashSet<boolean[]> allStates = getAllStates(nParents + 1);
		ll += org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size()) - org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size() + genesWithExpAndRNAi.size());
		//Go through the states
		for (boolean[] state : allStates) {
		    ll += org.apache.commons.math.special.Gamma.logGamma(prior + numberOfTFTFOccurences(p, parents, state)) - org.apache.commons.math.special.Gamma.logGamma(prior);
		}
	    }
	    else { //the node representing the gene
		LinkedHashSet<boolean[]> allStates = getAllStates(nParents + 1);
		ll += org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size()) - org.apache.commons.math.special.Gamma.logGamma(prior*allStates.size() + genesWithExpAndRNAi.size());
		//Go through the states
		for (boolean[] state : allStates) 
		    ll += org.apache.commons.math.special.Gamma.logGamma(prior + numberOfTFGeneOccurences(parents, state, inductionThres)) - org.apache.commons.math.special.Gamma.logGamma(prior);
	    }
	}
	return ll;*/
	return 0;
    }
    
    public HashMap<HashSet<TFPeakNode>, HashSet<TFPeakNode>> getAllStates(TFPeakNode v) {
	//Recursively generate all permutations of boolean vectors with length n
	boolean s[] = null; //new boolean[n];
	int i = 0;
	LinkedHashSet<boolean[]> states = null; //fillArray(s, i, new LinkedHashSet<boolean[]>());
	return null;
    }

    protected LinkedHashSet<boolean[]> fillArray(boolean[] s, int i, LinkedHashSet<boolean[]> states) {
	if (i<s.length) {
	    boolean[] st = new boolean[s.length], sf = new boolean[s.length];
	    System.arraycopy(s, 0, st, 0, i);
	    st[i] = true;
	    states = fillArray(st, i + 1, states);
	    System.arraycopy(s, 0, sf, 0, i);
	    sf[i] = false;
	    states = fillArray(sf, i + 1, states);
	}
	else 
	    states.add(s);
	return states;
    }

    public int numberOfTFTFOccurences(TFPeakNode p, Set parents, HashSet<TFPeakNode> tfsPresent, HashSet<TFPeakNode> tfsAbsent) {
	//Find out how many times we observed the given configuration with state s.
	/*LinkedHashSet<RefSeqGene> genes = RefSeqGenesPeakGraph.intersect(genesWithExpAndRNAi, genesPeaksGraph.genesWithPeakTFs(s[0] ? p.getTFName() : null, s[0] ? null : p.getTFName(), p.getMinDistance(), p.getMaxDistance()));
	//System.out.println("Found " + genes.size() + " for node " + p.toString() + " with state " + Arrays.toString(s) + " and " + parents.size() + " parents.");
	Iterator it = parents.iterator();
	while (it.hasNext()) {
	    TFPeakNode pa = (TFPeakNode)it.next();
	    genes = RefSeqGenesPeakGraph.intersect(genesPeaksGraph.genesWithPeakTFs(tfsPresent.contains(pa) ? pa.getTFName() : null, tfsAbsent.contains(pa) ? pa.getTFName() : null, pa.getMinDistance(), pa.getMaxDistance()), genes);
	}
	//System.out.println("Found " + genes.size() + " genes with " + p.toString() + " and " + parents.toString() + " with state " + Arrays.toString(s) + " and theta = " + format.format((double)(prior + genes.size())/(prior*Math.pow(2, s.length) + genesWithExpAndRNAi.size())));
	return genes.size();*/
	return 0;
    }

    
    public int numberOfTFGeneOccurences(Set parents, boolean geneInd, double minInd, HashSet<TFPeakNode> tfsPresent, HashSet<TFPeakNode> tfsAbsent) {
	//Find out how many times we observed the given configuration with state s.
	/*LinkedHashSet<RefSeqGene> genes = RefSeqGenesPeakGraph.intersect(genesWithExpAndRNAi, genesPeaksGraph.genesWithInduction(minInd, true, true, s[0]));
	Iterator it = parents.iterator();
	int i = 1;
	while (it.hasNext()) {
	    TFPeakNode pa = (TFPeakNode)it.next();
	    if (s[i++])
		genes = RefSeqGenesPeakGraph.intersect(genesPeaksGraph.genesWithPeakTFs(pa.getTFName(), null, pa.getMinDistance(), pa.getMaxDistance()), genes);
	    else
		genes = RefSeqGenesPeakGraph.intersect(genesPeaksGraph.genesWithPeakTFs(null, pa.getTFName(), pa.getMinDistance(), pa.getMaxDistance()), genes);
	}
	//System.out.println("Found " + genes.size() + " genes with " + parents.toString() + " with state " + Arrays.toString(s) + " and theta = " + format.format((double)(prior + genes.size())/(prior*Math.pow(2, s.length) + genesWithExpAndRNAi.size())));
	return genes.size();*/
	return 0;
    }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = "Bayesian network has nodes: \n";
	Iterator it = graph.getVertices().iterator();
	while (it.hasNext()) {
	    Object o = (Object)it.next();
	    if (o instanceof TFPeakNode) {
		TFPeakNode v = (TFPeakNode)o;
		if (v.numNeighbors()>0) {
		    str += v.toString();
		    Iterator edgeIt = v.getOutEdges().iterator();
		    while (edgeIt.hasNext()) {
			TFPeakEdge e = (TFPeakEdge)edgeIt.next();
			if (e.getDest() instanceof TFPeakNode)
			    str += "\t" + e.getDest().toString();
		    }
		    str += "\n";
		}
	    }
	}
	return str;
    }


}


//Use this class to represent a node in the Bayesian network for
//predicting gene expression induction. Each node corresponds to a
//specific TF and it is also associated with a distance interval wrt
//the TSS

class TFPeakNode extends DirectedSparseVertex {

    protected String tfName;
    protected int maxDist, minDist;
    protected HashMap<HashSet<TFPeakNode>, HashMap<HashSet<TFPeakNode>, Double>> theta = new HashMap<HashSet<TFPeakNode>, HashMap<HashSet<TFPeakNode>, Double>>();
    public int index = -1, lowlink; //Use for traversal

    public TFPeakNode(String tf, int min, int max) {
	tfName = tf;
	minDist = min;
	maxDist = max;
	//System.out.println(toString());
    }

    public String getTFName() { return tfName; }

    public LinkedHashSet<String> getTFNames(boolean includeThis) { 
	//Does not take the distance constraint into account
	LinkedHashSet<String> tfs = new LinkedHashSet<String>();
	if (includeThis)
	    tfs.add(tfName);
	Iterator it = getPredecessors().iterator();
	while (it.hasNext()) {
	    TFPeakNode p = (TFPeakNode)it.next();
	    tfs.add(p.getTFName());
	}
	return tfs;
    }

    public int getMaxDistance() { return maxDist; }

    public void setMaxDistance(int d) { maxDist = d; }

    public int getMinDistance() { return minDist; }

    public boolean hasPeak(RefSeqGene g, Vector<Peak> peaks) {
	//Returns true if there is a peak in a position relative to the TSS which also contains the TF of this node
	for (Peak p : peaks) {
	    if (p.hasFactor(tfName)) {
		for (Integer t : g.getStarts()) {
		    if (Math.abs(p.getCenter() - t)>=minDist && Math.abs(p.getCenter() - t)<maxDist)
			return true;
		}
	    }
	}
	return false;
    }

    /*public void setTheta(HashSet<TFPeakNode> tfsPresent, HashSet<TFPeakNode> tfsAbsent, double t) {
	if (!theta.contains(tfsPresent))
	    theta.put(tfsPresent, new HashMap<HashSet<TFPeakNode>, Double>());
	theta.get(tfsPresent).put(tfsAbsent, t);
	}*/

    public double getTheta(HashMap<Integer, HashSet<String>> tfsPresent) {
	//Determine the state
	Iterator it = getPredecessors().iterator();
	while (it.hasNext()) {
	    TFPeakNode parent = (TFPeakNode)it.next();
	    for (Integer d : tfsPresent.keySet()) {
	    if (d>=minDist && d<maxDist) {
		for (String tf : tfsPresent.get(d)) {
		    //Check if any of the parents correspond to this peak location

		    }
		}
	    }
	}
	return -1;
    }

    public boolean isGene() { return getTFName().equals(""); }

    public String toString() {
	return tfName + "[" + minDist + ", " + maxDist + "]";
    }

    public boolean equals(Object o) {
	if (!(o instanceof TFPeakNode))
	    return false;
	TFPeakNode v = (TFPeakNode)o;
	return v.tfName.equals(tfName) && v.minDist==minDist && v.maxDist==maxDist;
    }

    public int hashCode() { return 13*tfName.hashCode() + 37*minDist + 93*maxDist; }

}

class TFPeakEdge extends DirectedSparseEdge { //implements Comparable {

    protected double score = 0;

    public TFPeakEdge(DirectedSparseVertex p, DirectedSparseVertex s) {
	super(p, s);
    }

    public void setScore(double s) { score = s; }

    public double getScore() { return score; }

    public boolean equals(Object o) {
	if (!(o instanceof TFPeakEdge))
	    return false;
	TFPeakEdge e = (TFPeakEdge)o;
	return e.getSource().equals(getSource()) && e.getDest().equals(getDest());
    }

    public int hashCode() { return 3*getDest().hashCode() + 11*getSource().hashCode(); }

}


class Tarjan {

    private int index = 0;
    private ArrayList<TFPeakNode> stack = new ArrayList<TFPeakNode>();
    private ArrayList<ArrayList<TFPeakNode>> SCC = new ArrayList<ArrayList<TFPeakNode>>();

    public ArrayList<TFPeakNode> getLargestSCC(DirectedSparseGraph graph) {
	ArrayList<ArrayList<TFPeakNode>> sccs = executeTarjan(graph);
	ArrayList<TFPeakNode> largest = new ArrayList<TFPeakNode>();
	for (ArrayList<TFPeakNode> scc : sccs) {
	    if (scc.size()>largest.size())
		largest = scc;
	}
	//System.out.println("Found SCC : " + largest.toString());
	return largest;
    }

    /* The funtion tarjan has to be called for every unvisited node of the graph */
    public ArrayList<ArrayList<TFPeakNode>> executeTarjan(DirectedSparseGraph graph){
	//Reset all the nodes
	Iterator it = graph.getVertices().iterator();
	while (it.hasNext()) 
	    ((TFPeakNode)it.next()).index = -1;
	SCC.clear();
	index = 0;
	stack.clear();
	if(graph != null){
	    List<TFPeakNode> nodeList = new ArrayList<TFPeakNode>(graph.getVertices());
	    if(nodeList != null) {
                for (TFPeakNode node : nodeList) {
		    if (node.index==-1)
			tarjan(node, graph);
		}
	    }
	}
	return SCC;
    }
    
    
    private ArrayList<ArrayList<TFPeakNode>> tarjan(TFPeakNode v, DirectedSparseGraph list){
	v.index = index;
	v.lowlink = index;
	index++;
	stack.add(0, v);
	Iterator predIt = v.getPredecessors().iterator();
	while (predIt.hasNext()) {
	    TFPeakNode n = (TFPeakNode)predIt.next();
	    if(n.index == -1){
		tarjan(n, list);
		v.lowlink = Math.min(v.lowlink, n.lowlink);
	    }
	    else if(stack.contains(n))
		v.lowlink = Math.min(v.lowlink, n.index);
	}
	if(v.lowlink == v.index){
	    TFPeakNode n;
	    ArrayList<TFPeakNode> component = new ArrayList<TFPeakNode>();
	    do{
		n = stack.remove(0);
		component.add(n);
	    }while(!n.equals(v));
	    SCC.add(component);
	    //System.out.println("Found SCC : " + component.toString());
	}
	return SCC;
    }
}
