//Use this method to assign peaks to nearby transcripts and/or
//annotated regions. The output is a latex table as well as lists of
//peak locations that can be used as input for
//FindReadsAcrossFeature2.java when investigating the read density
//profiles.
//
//bsub -q shared_12h java -Xmx4096m -server PeaksNearTranscripts -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/merged_total/merged_total -o -p ../MmDataPeakFiles/H3K4Me3_KCl_peaks_names -c 19
//
//For HeLa:
//
//bsub -q shared_12h java -Xmx4096m -server PeaksNearTranscripts -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac -o -p ../../Encode/PeakFiles/wgEncodeUwChIPSeqPeaksRep1Helas3H3k4me3.narrowPeak --species hs
//
//bsub -q shared_12h java -Xmx4096m -server PeaksNearTranscripts -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac -o -p ../../Encode/PeakFiles/wgEncodeUwDnaseSeqPeaksRep1Helas3 --species hs -c 22
//
//bsub -q shared_12h java -Xmx4096m -server PeaksNearTranscripts -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac -o -p ../../Encode/PeakFiles/hela_tf_names.txt --species hs -c 4


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class PeaksNearTranscripts {

    public static final String CODING_ACTIVE_STR = "Promoter_genic,_active";
    public static final String CODING_PROMOTER_STR = "Promoter_genic";
    public static final String NON_CODING_PROMOTER_STR = "Promoter_ncRNA";
    public static final String CODING_INACTIVE_STR = "Promoter_genic,_inactive";
    public static final String NON_CODING_ACTIVE_STR = "Promoter_ncRNA,_active";
    public static final String NON_CODING_INACTIVE_STR = "Promoter_ncRNA,_inactive";
    public static final String EXTRAGENIC_ENHANCER_STR = "Extragenic_enhancer";
    public static final String INTRAGENIC_ENHANCER_STR = "Intragenic_enhancer";
    //public static final String INACTIVE_EXTRAGENIC_ENHANCER_STR = "Extragenic_enhancer_inactive";
    //public static final String INACTIVE_INTRAGENIC_ENHANCER_STR = "Intragenic_enhancer_inactive";
    public static final String NON_CODING_OTHER_STR = "Promoter_ncRNA,_novel";
    public static final String NON_CODING_OTHER_PROXIMAL_STR = "Unannotated_proximal";
    //public static final String CODING_ACTIVE_END_STR = "Coding_end";
    //public static final String CODING_INACTIVE_END_STR = "Coding_inactive_end";
    //public static final String NON_CODING_ACTIVE_END_STR = "Noncoding_end";
    //public static final String NON_CODING_INACTIVE_END_STR = "Noncoding_inactive_end";
    public static final String CODING_ACTIVE_INTRONIC_STR = "Intragenic,_active";
    public static final String CODING_INACTIVE_INTRONIC_STR = "Intragenic,_inactive";
    public static final String NON_CODING_ACTIVE_INTRONIC_STR = "Intragenic_ncRNA,_active";
    public static final String NON_CODING_INACTIVE_INTRONIC_STR = "Intragenic_ncRNA,_inactive";
    public static final String CODING_ACTIVE_EXONIC_STR = "Intragenic_exon,_active";
    public static final String CODING_INACTIVE_EXONIC_STR = "Intragenic_exon,_inactive";
    public static final String NON_CODING_ACTIVE_EXONIC_STR = "Intragenic_exon_ncRNA,_active";
    public static final String NON_CODING_INACTIVE_EXONIC_STR = "Intragenic_exon_ncRNA,_inactive";
    public static final String CODING_EXONIC_STR = "Exon";
    public static final String NON_CODING_EXONIC_STR = "Exon_ncRNA";
    public static final String CODING_INTRONIC_STR = "Intron";
    public static final String NON_CODING_INTRONIC_STR = "Intron_ncRNA";
    public static final String EXTRAGENIC_OTHER_STR = "Extragenic_other";
    public static final String EXTRAGENIC_EST_STR = "Extragenic_EST";
    public static final String MAR_STR = "MAR";
    //public static final String EXTRAGENIC_REPEAT_STR = "Extragenic_repeat";

    //public static final String[] PEAK_CATEGORIES = new String[]{CODING_ACTIVE_STR, CODING_INACTIVE_STR, NON_CODING_ACTIVE_STR, NON_CODING_INACTIVE_STR, EXTRAGENIC_ENHANCER_STR, INTRAGENIC_ENHANCER_STR/*, INACTIVE_EXTRAGENIC_ENHANCER_STR, INACTIVE_INTRAGENIC_ENHANCER_STR*/, NON_CODING_OTHER_STR/*, NON_CODING_OTHER_PROXIMAL_STR/*, CODING_ACTIVE_END_STR, CODING_INACTIVE_END_STR, NON_CODING_ACTIVE_END_STR, NON_CODING_INACTIVE_END_STR*/, CODING_ACTIVE_INTRONIC_STR, CODING_INACTIVE_INTRONIC_STR, NON_CODING_ACTIVE_INTRONIC_STR, NON_CODING_INACTIVE_INTRONIC_STR, CODING_ACTIVE_EXONIC_STR, CODING_INACTIVE_EXONIC_STR, NON_CODING_ACTIVE_EXONIC_STR, NON_CODING_INACTIVE_EXONIC_STR/*, EXTRAGENIC_EST_STR/*, EXTRAGENIC_REPEAT_STR*/, EXTRAGENIC_OTHER_STR};
    public static final String[] PEAK_CATEGORIES = new String[]{CODING_PROMOTER_STR, NON_CODING_PROMOTER_STR, EXTRAGENIC_ENHANCER_STR, INTRAGENIC_ENHANCER_STR, NON_CODING_OTHER_STR, CODING_INTRONIC_STR, NON_CODING_INTRONIC_STR, CODING_EXONIC_STR, NON_CODING_EXONIC_STR, EXTRAGENIC_OTHER_STR, MAR_STR};

    public static final int PROXIMAL_CUT_OFF = 10000;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option intronicFlag = argParser.addBooleanOption('I', "intronic");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option reverseFlag = argParser.addBooleanOption('r', "reverse");
	    CmdLineParser.Option saveRepeatsFlag = argParser.addBooleanOption("saveRepeats");
	    CmdLineParser.Option estsFlag = argParser.addBooleanOption("EST");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption('p', "peakFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option maxDistanceFlag = argParser.addIntegerOption('d', "maxDistance");	    
	    CmdLineParser.Option maxEnhancerDistFlag = argParser.addIntegerOption('e', "maxEnhancerDist");	    
	    CmdLineParser.Option maxPeakDistFlag = argParser.addIntegerOption('P', "maxPeakDist");	    
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean intronic = argParser.getOptionValue(intronicFlag)!=null;
	    boolean saveRepeats = argParser.getOptionValue(saveRepeatsFlag)!=null;
	    boolean ests = argParser.getOptionValue(estsFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/Encode/PeakFiles/wgEncodeUwDnaseSeqPeaksRep1Helas3.narrowPeak" : "/home/" + (orchestra ? "mh168" : "mhemberg") + "/Projects/ChIP-Seq/TKData/MmDataPeakFiles/tf_kcl_names";
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(maxDistanceFlag);
	    int maxDistance = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(maxEnhancerDistFlag);
	    int maxEnhancerDist = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(maxPeakDistFlag);
	    int maxPeakDist = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : .001;
	    ///////////////////////////////////////////////////////
	    String extragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_extragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lRNA_lMe3_real.tsv";
	    TreeSet<Enhancer2> extragenicEnhancers = FindTranscripts2.parseEnhancerFile(extragenicEnhancerFileName);
	    String intragenicEnhancerFileName = !species.equals("mm") ? "/groups/neuroduo/martin/ChIP-Seq/HeLa/Enhancers/heintzman_intragenic_enhancer_list_hg18.bed" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_in_lMe3_real.tsv";
	    TreeSet<Enhancer2> intragenicEnhancers = FindTranscripts2.parseEnhancerFile(intragenicEnhancerFileName);
	    TreeSet<Enhancer2> allEnhancers = new TreeSet<Enhancer2>();
	    allEnhancers.addAll(extragenicEnhancers);
	    allEnhancers.addAll(intragenicEnhancers);
	    //Get validated enhancers that were not called in our experiment
	    //String inactiveExtragenicEnhancerFileName = "/groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/Pennacchio/inactive_extragenic_enhancers.txt";
	    //TreeSet<Enhancer2> inactiveExtragenicEnhancers = FindTranscripts2.parseEnhancerFile(inactiveExtragenicEnhancerFileName);
	    //String inactiveIntragenicEnhancerFileName = "/groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/Pennacchio/inactive_intragenic_enhancers.txt";
	    //TreeSet<Enhancer2> inactiveIntragenicEnhancers = FindTranscripts2.parseEnhancerFile(inactiveIntragenicEnhancerFileName);
	    ////////////////////////////////////////////
	    String lncRNAFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species);
	    String macroRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species);
	    String miRNAFileName = !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
	    TreeSet<MiRNA> miRNAs = AnalyzeTranscriptReadProfiles.loadMiRNA(miRNAFileName, species);

	    //Get the list of exons
	    String exonFileName = "/groups/neuroduo/martin/Genomes/" + (species.equals("mm") ? "Mm/v9" : "Hs/v36") + "/all_EXNs.tsv";
	    TreeSet<Exon> exons = FindConservationIslands.parseExons(exonFileName, species, chr);
	    TreeSet<Exon> macroRNAExons = FindConservationIslands.parseMacroRNAExons(macroRNAFileName, chr);
	    String lncRNAExonFileName = !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/exonsNPC.gff";
	    TreeSet<Exon> lincRNAExons = FindConservationIslands.parseLincRNAExons(lncRNAExonFileName, chr, lncRNAs);

	    String savePeaksSuffix = "", tfNamesInStudy[] = null; //new String[]{};
	    if (peakFileName.contains("tf_kcl_names")) {
		savePeaksSuffix = "tfbs";
		tfNamesInStudy = new String[]{"CBP_Millipore_KCl_B1","CREB_SC_KCl_B1","Npas4_Npas4_KCl_B1","SRF_H300_KCl_B1"};
	    }
	    else if (peakFileName.contains("H3K4Me3_KCl")) {
		savePeaksSuffix = "h3k4me3_kcl";
		tfNamesInStudy = new String[]{"H3K4Me3_H3K4Me3_KCl_B2"};
	    }
	    else if (peakFileName.contains("H3k4me3")) {
		savePeaksSuffix = "h3k4me3";
		//tfNamesInStudy = new String[]{"H3k4me3"};
	    }
	    else if (peakFileName.contains("Ctcf")) {
		savePeaksSuffix = "hela_ctcf";
		//tfNamesInStudy = new String[]{"Ctcf"};
	    }
	    else if (peakFileName.contains("Dnase")) {
		savePeaksSuffix = "hela_dnase";
		//tfNamesInStudy = new String[]{"Ctcf"};
	    }
	    else if (peakFileName.contains("hela_tf_names")) {
		//File contains the name of the .narrowPeak files that we want to include. Find the union of all these files.
		savePeaksSuffix = "hela_tfbs";
		//tfNamesInStudy = new String[]{"Tr4", "E2f4", "Cmyc", "Cfos", "Bdp1", "Max", "Ap2gamma", "Ap2alpha", "ifngStat1", "E2f6"};
	    }
	    Vector<Peak> peaks = !species.equals("mm") ? new Vector<Peak>(EncodePeaks.parseEncodePeakFileTF(peakFileName, "Helas3", chr)) : new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values());
	    Collections.sort(peaks);
	    System.out.println("Found " + peaks.size() + " peaks.");

	    String estFileName = !species.equals("mm") ? "/groups/neuroduo/jesse/human_annot_for_mh/human_ESTs/" : "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/";
	    LinkedHashMap<String, EST> estTagsMapForward = null, estTagsMapReverse = null;

	    String latexTableFileName = savePeaksSuffix + "_peaks_categories" + (chr==-1 ? "" : "_" + chr) + ".tex";
	    int nChr = FindTranscripts.numberOfChromosomes(species), readLength = 35;
	    String rnaFileName = fileName + ".tsv", rnaFileNameForward = fileName + "_f", rnaFileNameReverse = fileName + "_r";
	    //Writers for storing the locations of the peaks in the different categories
	    TreeMap<String, BufferedWriter> peakWriters = new TreeMap<String, BufferedWriter>(), peakConsWriters = new TreeMap<String, BufferedWriter>();
	    //Keep track of the number of peaks in each category
	    TreeMap<String, TreeSet<Peak>> category2peakMap = new TreeMap<String, TreeSet<Peak>>();
	    for (int i=0; i<PEAK_CATEGORIES.length; i++) {
		peakWriters.put(PEAK_CATEGORIES[i], new BufferedWriter(new FileWriter("PeakLocations/" + PEAK_CATEGORIES[i] + "_" + savePeaksSuffix + "_locations" + (chr==-1 ? "" : "_" + chr) + ".txt")));
		peakConsWriters.put(PEAK_CATEGORIES[i], new BufferedWriter(new FileWriter("PeakLocations/" + PEAK_CATEGORIES[i] + "_" + savePeaksSuffix + "_cons_locations" + (chr==-1 ? "" : "_" + chr) + ".txt")));
		category2peakMap.put(PEAK_CATEGORIES[i], new TreeSet<Peak>());
	    }
	    int prevChr = 0, nPeaksCategorized = 0;
	    RefSeqUCSCEnsemblGenes geneInfo = null;
	    RepeatInfo repeatInfo = null;
	    String repeatFileName = species.toLowerCase().equals("mm") ? EstimateGenicTranscription3.REPEAT_RNA_FILE_MM : EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_36;
	    TreeSet<TranscribedRegion2> segmentsForward = null, segmentsReverse = null;
	    for (Peak p : peaks) {
		if (chr==-1 || p.getChromosome()==chr) {
		    if (p.getChromosome()!=prevChr) {
			prevChr = p.getChromosome();
			//if (prevChr==2) { break; }
			geneInfo = new RefSeqUCSCEnsemblGenes(prevChr, orchestra, species);
			//repeatInfo = new RepeatInfo(prevChr, orchestra, species);
			//Load reads and assign
			ABIRead[] rnaReadsForward = ABIReads.loadMRNAReadsChr(rnaFileName, true, prevChr, readLength, false, repeatFileName);
			ABIRead[] rnaReadsReverse = ABIReads.loadMRNAReadsChr(rnaFileName, false, prevChr, readLength, false, repeatFileName);
			System.out.println("Found " + rnaReadsForward.length + " reads forward and " + rnaReadsReverse.length + " reverse reads for chr " + prevChr);
			if ((prevChr==21 && species.equals("mm")) && (prevChr==24 && species.equals("hs"))) { //no transcript calling for this chr
			    segmentsForward = new TreeSet<TranscribedRegion2>();
			    segmentsReverse = new TreeSet<TranscribedRegion2>();
			}
			else {
			    //Set reads to genes and transcripts
			    geneInfo.setFactorReads(rnaFileName, rnaReadsReverse, -1);
			    geneInfo.setFactorReads(rnaFileName, rnaReadsForward, 1);
			    geneInfo.setGeneIDsKeep(true, rnaFileName, minDens);
			    geneInfo.setGeneIDsKeep(false, rnaFileName, minDens);
			    AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, rnaFileName, rnaReadsForward, prevChr, 1);
			    AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, rnaFileName, rnaReadsForward, prevChr);
			    AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsForward, prevChr, 1);
			    //Get the transcribed regions. We are only interested in the list of segments and since we cannot have ReducedModel2 objects for both strands at the same time, we create the in succession and retain only the list of transcribed regions.
			    ReducedModel2 reducedModel = new ReducedModel2(fileName  + FindTranscripts.BED_FILE_SUFFIX + "_f.bed", rnaReadsForward, rnaFileName, prevChr, true, species);
			    //Assign categories
			    reducedModel.compare2Annotation(maxDistance, geneInfo, lncRNAs, miRNAs, macroRNAs, allEnhancers);
			    reducedModel.categorizeUnannotatedTranscribedRegions(maxEnhancerDist, species.equals("mm"));
			    reducedModel.removeGenes();
			    segmentsForward = reducedModel.getSegments();
			    ////////////////////////////////////////////
			    //Again for the reverse strand
			    geneInfo = new RefSeqUCSCEnsemblGenes(prevChr, orchestra, species);
			    //Set reads to genes and transcripts
			    geneInfo.setFactorReads(rnaFileName, rnaReadsReverse, -1);
			    geneInfo.setFactorReads(rnaFileName, rnaReadsForward, 1);
			    geneInfo.setGeneIDsKeep(true, rnaFileName, minDens);
			    geneInfo.setGeneIDsKeep(false, rnaFileName, minDens);
			    AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, rnaFileName, rnaReadsReverse, prevChr, -1);
			    AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, rnaFileName, rnaReadsReverse, prevChr);
			    AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsReverse, prevChr, -1);
			    reducedModel = new ReducedModel2(fileName  + FindTranscripts.BED_FILE_SUFFIX + "_r.bed", rnaReadsReverse, rnaFileName, prevChr, true, species);
			    //Assign categories
			    reducedModel.compare2Annotation(maxDistance, geneInfo, lncRNAs, miRNAs, macroRNAs, allEnhancers);
			    reducedModel.categorizeUnannotatedTranscribedRegions(maxEnhancerDist, species.equals("mm"));
			    reducedModel.removeGenes();
			    segmentsReverse = reducedModel.getSegments();
			    if (ests) { //Get ESTs
				estTagsMapForward = CompareCageEstRNASeqStarts.parseEstTags(estFileName, prevChr, true, species);
				estTagsMapReverse = CompareCageEstRNASeqStarts.parseEstTags(estFileName, prevChr, false, species);
			    }
			}
			///////////////////////////////////////////////////////
			geneInfo = new RefSeqUCSCEnsemblGenes(prevChr, orchestra, species);
			//Set reads to genes and transcripts
			geneInfo.setFactorReads(rnaFileName, rnaReadsReverse, -1);
			geneInfo.setFactorReads(rnaFileName, rnaReadsForward, 1);
			geneInfo.setGeneIDsKeep(true, rnaFileName, minDens);
			geneInfo.setGeneIDsKeep(false, rnaFileName, minDens);
			//Set reads to the ncRNAs
			AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, rnaFileName, rnaReadsForward, prevChr, 1);
			AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, rnaFileName, rnaReadsForward, prevChr);
			AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsForward, prevChr, 1);
			System.out.println("Read data and annotation for chr " + p.getChromosome());
		    }
		    //System.out.println(p.toString() + " has factors " + Arrays.toString(p.getFactorNames(true)));
		    if (tfNamesInStudy==null || p.hasFactor(tfNamesInStudy)) {
			//Compare against annotation
			RefSeqUCSCEnsemblGene nearbyGene = geneInfo.getNearestTSSGene(p.getCenter());
			int nearestGeneDist = nearbyGene!=null ? Math.abs(p.getCenter() - nearbyGene.getStart()) : Integer.MAX_VALUE;
			if (nearbyGene!=null && nearestGeneDist<maxPeakDist) {
			    if (!nearbyGene.isNonCoding()) {
				//System.out.println(p.toString() + " starts near coding " + nearbyGene.toString());
				if (nearbyGene.getReadDensity(rnaFileName)>minDens) 				 
				    category2peakMap.get(CODING_ACTIVE_STR).add(p);
				else
				    category2peakMap.get(CODING_INACTIVE_STR).add(p);
			    }
			    else { //Non-coding annotated
				if (nearbyGene.getReadDensity(rnaFileName)>minDens)
				    category2peakMap.get(NON_CODING_ACTIVE_STR).add(p);
				else {
				    category2peakMap.get(NON_CODING_INACTIVE_STR).add(p);
				    //System.out.println(p.toString() + " starts near non-coding " + nearbyGene.toString());
				}
			    }
			}
			else { //Look for macroRNA and lncRNA promoters
			    MacroRNA m = getNearestMacroRNA(macroRNAs, p, maxPeakDist);
			    if (m!=null) {
				if (m.getReadDensity(rnaFileName)>minDens)
				    category2peakMap.get(NON_CODING_ACTIVE_STR).add(p);
				else {
				    category2peakMap.get(NON_CODING_INACTIVE_STR).add(p);
				    //System.out.println(p.toString() + " starts near macroRNA " + m.toString());
				}
			    }
			    else {
				LncRNA l = getNearestLncRNA(lncRNAs, p, maxPeakDist);
				if (l!=null) {
				    if (l.getReadDensity(rnaFileName)>minDens)
					category2peakMap.get(NON_CODING_ACTIVE_STR).add(p);
				    else {
					category2peakMap.get(NON_CODING_INACTIVE_STR).add(p);
					//System.out.println(p.toString() + " starts near lncRNA " + l.toString());
				    }
				}
				else {	//Check if near an annotated promoter, either defined by the annotation or the transcript caller.
				    nPeaksCategorized++;
				    TranscribedRegion2 tForwardNearestStart = ReducedModel2.getNearestTranscript(segmentsForward, p.getCenter(), true, true);
				    TranscribedRegion2 tReverseNearestStart = ReducedModel2.getNearestTranscript(segmentsReverse, p.getCenter(), true, false);
				    int nearestForwardTranscriptDistance = tForwardNearestStart!=null ? Math.abs(p.getCenter() - tForwardNearestStart.getStart()) : Integer.MAX_VALUE;
				    int nearestReverseTranscriptDistance = tReverseNearestStart!=null ? Math.abs(p.getCenter() - tReverseNearestStart.getStart()) : Integer.MAX_VALUE;
				    if (tForwardNearestStart!=null && nearestForwardTranscriptDistance<maxPeakDist) {
					if (tForwardNearestStart.isCodingAnnotated()) 
					    category2peakMap.get(CODING_ACTIVE_STR).add(p);
					else if (tForwardNearestStart.isNonCodingAnnotated()) 
					    category2peakMap.get(NON_CODING_ACTIVE_STR).add(p);
					else if (nearestGeneDist>PROXIMAL_CUT_OFF) 
					    category2peakMap.get(NON_CODING_OTHER_STR).add(p);
						//else
						//category2peakMap.get(NON_CODING_OTHER_PROXIMAL_STR).add(p);
					//}
				    }
				    else if (tReverseNearestStart!=null && nearestReverseTranscriptDistance<maxPeakDist) {
					if (tReverseNearestStart.isCodingAnnotated())
					    category2peakMap.get(CODING_ACTIVE_STR).add(p);
					else if (tReverseNearestStart.isNonCodingAnnotated())
					    category2peakMap.get(NON_CODING_ACTIVE_STR).add(p);
					else if (nearestGeneDist>PROXIMAL_CUT_OFF) 
					    category2peakMap.get(NON_CODING_OTHER_STR).add(p);
					//else
					//category2peakMap.get(NON_CODING_OTHER_PROXIMAL_STR).add(p);
					//}
				    }
				    else { //Look for enhancers
					if (nearEnhancer(intragenicEnhancers, p, maxEnhancerDist))
					    category2peakMap.get(INTRAGENIC_ENHANCER_STR).add(p);
					else if (nearEnhancer(extragenicEnhancers, p, maxEnhancerDist))
					    category2peakMap.get(EXTRAGENIC_ENHANCER_STR).add(p);
					else { //Compare against annotation
					    RefSeqUCSCEnsemblGene overlappingGene = geneInfo.getOverlappingGene(p.getCenter());
					    if (overlappingGene!=null) {
						if (!overlappingGene.isNonCoding()) {
						    if (overlappingGene.getReadDensity(rnaFileName)>minDens) {
							if (FindConservationIslands.overlapsExon(p.getCenter(), exons)) 
							    category2peakMap.get(CODING_ACTIVE_EXONIC_STR).add(p);
							else
							    category2peakMap.get(CODING_ACTIVE_INTRONIC_STR).add(p);
						    }
						    else {
							if (FindConservationIslands.overlapsExon(p.getCenter(), exons))
							    category2peakMap.get(CODING_INACTIVE_EXONIC_STR).add(p);
							else
							    category2peakMap.get(CODING_INACTIVE_INTRONIC_STR).add(p);
						    }
						}
						else { //Non-coding annotated
						    if (overlappingGene.getReadDensity(rnaFileName)>minDens) {
							if (FindConservationIslands.overlapsExon(p.getCenter(), exons))
							    category2peakMap.get(NON_CODING_ACTIVE_EXONIC_STR).add(p);
							else
							    category2peakMap.get(NON_CODING_ACTIVE_INTRONIC_STR).add(p);
						    }
						    else {
							if (FindConservationIslands.overlapsExon(p.getCenter(), exons))
							    category2peakMap.get(NON_CODING_INACTIVE_EXONIC_STR).add(p);
							else
							    category2peakMap.get(NON_CODING_INACTIVE_INTRONIC_STR).add(p);
						    }
						}
					    }
					    else { //macro or lncRNAs
						if (FindConservationIslands.overlapsExon(p.getCenter(), macroRNAExons) || FindConservationIslands.overlapsExon(p.getCenter(), lincRNAExons))
						    category2peakMap.get(NON_CODING_INACTIVE_EXONIC_STR).add(p);
						else { //Other extragenic
						    //See how many of these that have any EST evidence
						    if (ests && (FindEnhancerSites.findNearbyESTs(estTagsMapForward, p.getChromosome(), true, p.getCenter(), maxEnhancerDist, -1, null)>0 || FindEnhancerSites.findNearbyESTs(estTagsMapReverse, p.getChromosome(), false, p.getCenter(), maxEnhancerDist, -1, null)>0))
							category2peakMap.get(EXTRAGENIC_EST_STR).add(p);
						    //else if (repeatInfo.overlaps(p.getCenter())) //check if overlapping a repeat region
						    //    category2peakMap.get(EXTRAGENIC_REPEAT_STR).add(p);
						    else {
							category2peakMap.get(EXTRAGENIC_OTHER_STR).add(p);
							//System.out.println(p.toString() + " has category " + EXTRAGENIC_OTHER_STR);
						    }
						}
					    }
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	    //Load the conservation islands so that we can study the overlap
	    ConservationIslands conservationIslands = new ConservationIslands(chr, orchestra, species);
	    //Now we can save the results as a table in a .tex file 
	    BufferedWriter bw = new BufferedWriter(new FileWriter(latexTableFileName));
	    System.out.println("Saving " + nPeaksCategorized + " categorized peaks to " + latexTableFileName);
	    bw.write("\\begin{tabular}{l|cc}\n");
	    for (int i=0; i<PEAK_CATEGORIES.length; i++) {
		int nPeaks = category2peakMap.get(PEAK_CATEGORIES[i]).size();
		bw.write(PEAK_CATEGORIES[i] + " & " + nPeaks + " & " + numberOverlappingConservationIsland(category2peakMap.get(PEAK_CATEGORIES[i]), conservationIslands) + "\\\\\n");
	    }
	    bw.write("\\end{tabular}\n");
	    bw.close();
	    //Do a table where the number of peaks associated with
	    //repeat elements are broken down by category
	    //if (saveRepeats)
	    //saveRepeatRegionPeaks(category2peakMap.get(EXTRAGENIC_REPEAT_STR), latexTableFileName.replace("peaks", "peaks_repeat"), species);
	    //Repeat, but create one file for each TF
	    /*for (int j=0; j<tfNamesInStudy.length; j++) {
		bw = new BufferedWriter(new FileWriter(latexTableFileName.replace("peaks", tfNamesInStudy[j] + "_peaks")));
		bw.write("\\begin{tabular}{l|cc}\n");
		for (int i=0; i<PEAK_CATEGORIES.length; i++) {
		    int nPeaks = peaksWithTF(category2peakMap.get(PEAK_CATEGORIES[i]), tfNamesInStudy[j]).size();
		    bw.write(PEAK_CATEGORIES[i] + " & " + nPeaks + " & " + format.format((double)numberOverlappingConservationIsland(peaksWithTF(category2peakMap.get(PEAK_CATEGORIES[i]), tfNamesInStudy[j]), conservationIslands)/nPeaks) + "\\\\\n");
		}
		bw.write("\\end{tabular}\n");
		bw.close();
		//if (saveRepeats)
		//    saveRepeatRegionPeaks(peaksWithTF(category2peakMap.get(EXTRAGENIC_REPEAT_STR), tfNamesInStudy[j]), latexTableFileName.replace("peaks", tfNamesInStudy[j] + "repeat_peaks"), species);
		}*/
	    //And we can also save the locations of all the peaks so that we can investigate the histone, Pol2 and RNA profiles
	    for (int i=0; i<PEAK_CATEGORIES.length; i++) {
		System.out.println("Saving " + category2peakMap.get(PEAK_CATEGORIES[i]).size() + " peaks to " + PEAK_CATEGORIES[i]);
		for (Peak p : category2peakMap.get(PEAK_CATEGORIES[i])) {
		    peakWriters.get(PEAK_CATEGORIES[i]).write(p.getChromosome() + "\t" + p.getCenter() + "\n");
		    if (p.isConserved())
			peakConsWriters.get(PEAK_CATEGORIES[i]).write(p.getChromosome() + "\t" + p.getCenter() + "\n");
		}
		peakWriters.get(PEAK_CATEGORIES[i]).close();
		peakConsWriters.get(PEAK_CATEGORIES[i]).close();
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    ///////////////////////////////////////////////////////

    public static MacroRNA getNearestMacroRNA(TreeSet<MacroRNA> macroRNAs, Peak p, int maxDist) {
	//Get the one with a start nearest p
	return getNearestMacroRNA(macroRNAs, p.getChromosome(), p.getCenter(), maxDist);
    }

    public static MacroRNA getNearestMacroRNA(TreeSet<MacroRNA> macroRNAs, int chr, int pos, int maxDist) {
	//Get the one with a start nearest pos
	if (macroRNAs==null || macroRNAs.size()==0)
	    return null;
	MacroRNA nearest = macroRNAs.first();
	int minDist = Math.abs(pos - (nearest.getStrand() ? nearest.getStart() : nearest.getEnd()));
	for (MacroRNA m : macroRNAs) {
	    if (m.getChromosome()==chr) {
		int dist = Math.abs(pos - (m.getStrand() ? m.getStart() : m.getEnd()));
		if (dist<minDist) {
		    nearest = m;
		    minDist = dist;
		}
	    }
	}
	return minDist<maxDist ? nearest : null;
    }

    public static LncRNA getNearestLncRNA(TreeSet<LncRNA> lncRNAs, Peak p, int maxDist) {
	//Get the one with a start nearest p
	return getNearestLncRNA(lncRNAs, p.getChromosome(), p.getCenter(), maxDist);
    }

    public static LncRNA getNearestLncRNA(TreeSet<LncRNA> lncRNAs, int chr, int pos, int maxDist) {
	if (lncRNAs==null || lncRNAs.size()==0)
	    return null;
	LncRNA nearest = lncRNAs.first();
	int minDist = Math.min(Math.abs(pos - nearest.getStart()), Math.abs(pos - nearest.getEnd()));
	for (LncRNA m : lncRNAs) {
	    if (m.getChromosome()==chr) {
		int dist = Math.min(Math.abs(pos - m.getStart()), Math.abs(pos - m.getEnd()));
		if (dist<minDist) {
		    nearest = m;
		    minDist = dist;
		}
	    }
	}
	return minDist<maxDist ? nearest : null;
    }

    public static boolean nearEnhancer(TreeSet<Enhancer2> enhancers, Peak p, int maxDist) {
	for (Enhancer2 e : enhancers) {
	    if (p.getChromosome()==e.getChromosome() && Math.abs(p.getCenter() - e.getCenter())<maxDist)
		return true;
	}
	return false;
    }

    public static TreeSet<Peak> peaksWithTF(TreeSet<Peak> peaks, String tf) {
	TreeSet<Peak> peaks2keep = new TreeSet<Peak>();
	for (Peak p : peaks) {
	    if (p.hasFactor(tf))
		peaks2keep.add(p);
	}
	return peaks2keep;
    }

    public static void saveRepeatRegionPeaks(TreeSet<Peak> peaks, String fileName, String species) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    RepeatInfo repeatInfo = new RepeatInfo(true, species);
	    for (Peak p : peaks) {
		RepeatRegion r = repeatInfo.getOverlappingRegion(p.getChromosome(), p.getCenter());
		if (r!=null) 
		    bw.write(p.getChromosome() + "\t" + p.getCenter() + "\t" + r.toString() + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving repeat region peaks: " + e.toString());
	}
    }

    public static int numberOverlappingConservationIsland(TreeSet<Peak> peaks, ConservationIslands conservationIslands) {
	int nOverlap = 0;
	for (Peak p : peaks) {
	    if (conservationIslands.overlaps(p)) {
		nOverlap++;
		p.setConserved(true);
	    }
	    //System.out.println("found overlapping island at " + p.toString());
	}
	System.out.println("Found " + nOverlap + " overlapping islands.");
	return nOverlap;
    }

}

///////////////////////////////////////////////////////

