//Use this class for parsing the phastCons files for a genome.
//
//In Matlab we can use it as:
//
//psp = javaObject('PhastConsParser', '/media/My Book/Genomes/Mm/v9/PhastConsScores/chr2.data');
//cons = double(javaMethod('findScores', starts, 35));



import java.util.*;
import java.util.regex.*;
import java.io.*;

public class PhastConsParser {

    protected String fileName;
    protected BufferedReader phastConsReader = null;
    //Keep track of where in the genome we are when reading from the file
    protected int index  = 0;
    //Keep track of where the buffer is starting
    protected int bufferStart = 0;
    //Where the next set of non-zero scores start
    protected int blockEnd = -1;
    protected double buffer[];


    public static final Pattern phastConsHeaderPattern = Pattern.compile("fixedStep\\s+chrom=chr([\\d|X|Y]+)\\s+start=(\\d+)\\s+step=(\\d+)");

    //////////////////////////////////////////////////////////////////

    public PhastConsParser(String fName) { fileName = fName; }

    public PhastConsParser(int chr, boolean vertebrate, boolean orchestra, String species) {
	fileName = orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/";
	if (species.equals("mm")) {
	    if (chr==20)
		fileName += "Genomes/Mm/v9/PhastConsScores/chrX";
	    else if (chr==21) 
		fileName += "Genomes/Mm/v9/PhastConsScores/chrY";
	    else
		fileName += "Genomes/Mm/v9/PhastConsScores/chr" + chr;
	    fileName += vertebrate ? ".vertebrates.pp.data" : ".euarchontoglires.pp.data";
	}
	else {
	    if (chr==23)
		fileName += "Genomes/Hs/v36/PhastConsScores/chrX";
	    else if (chr==24) 
		fileName += "Genomes/Hs/v36/PhastConsScores/chrY";
	    else
		fileName += "Genomes/Hs/v36/PhastConsScores/chr" + chr;
	    fileName += ".pp";
	}
    }

    public PhastConsParser(int chr, boolean vertebrate, boolean orchestra, String species, boolean hg19) {
	fileName = orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/";
	if (species.equals("mm")) {
	    if (chr==20)
		fileName += "Genomes/Mm/v9/PhastConsScores/chrX";
	    else if (chr==21) 
		fileName += "Genomes/Mm/v9/PhastConsScores/chrY";
	    else
		fileName += "Genomes/Mm/v9/PhastConsScores/chr" + chr;
	    fileName += vertebrate ? ".vertebrates.pp.data" : ".euarchontoglires.pp.data";
	}
	else {
	    if (chr==23)
		fileName += "Genomes/Hs/v36/PhastConsScores/chrX";
	    else if (chr==24) 
		fileName += "Genomes/Hs/v36/PhastConsScores/chrY";
	    else
		fileName += "Genomes/Hs/v36/PhastConsScores/chr" + chr;
	    if (hg19) {
		fileName = fileName.replace("v36", "v37");
		fileName += ".phastCons46way.wigFix";
	    }
	    else
		fileName += ".pp";
	}
    }

    public PhastConsParser(int chr, boolean vertebrate, boolean orchestra, int len) {
	fileName = orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/";
	if (chr==20)
	    fileName += "Genomes/Mm/v9/PhastConsScores/chrX";
	else if (chr==21) 
	    fileName += "Genomes/Mm/v9/PhastConsScores/chrY";
	else
	    fileName += "Genomes/Mm/v9/PhastConsScores/chr" + chr;
	fileName += vertebrate ? ".vertebrates.pp.data" : ".euarchontoglires.pp.data";
	try {
	    initializeParser(len);
	}
	catch (IOException e) {
	    System.err.println("Error finding conservation scores from " + fileName);
	}
    }

    //////////////////////////////////////////////////////////////////

    public double[] findScores(int start, int end) throws ArrayIndexOutOfBoundsException {
	int len = end - start, i = 0;
	double ret[] = new double[len];
	try {
	    if (phastConsReader==null || len>buffer.length) 
		initializeParser(len);
	    //moveBuffer(start);
	    while (start>bufferStart + buffer.length) { 
		readScore();
		bufferStart++;
	    }
	    //Now we have reached the point where the buffer is
	    //aligned with region of interest.
	    while (start>bufferStart) { 
		buffer = shiftArrayLeft(buffer, readScore());
		bufferStart++;
	    }
	    //Copy the stuff that we need from the buffer to ret
	    for (int j=0; j<len; j++)
		ret[j] = buffer[j];
	}
	catch (IOException e) {
	    System.err.println("Error finding conservation scores from " + fileName + " for " + start + ", " + end);
	    return null;
	}
	//System.out.println("Found scores for " + start);
	return ret;
    }

    
    protected void initializeParser(int len) throws IOException {
	System.out.println("Initializing PhastConsParser for " + fileName + " with buffer length " + len);
	phastConsReader = new BufferedReader(new FileReader(fileName));
	String line = phastConsReader.readLine();
	Matcher m = phastConsHeaderPattern.matcher(line);
	index = 0;
	bufferStart = 0;
	buffer = new double[len];
	if (m.find()) {
	    //Read all the subsequent scores, they will be
	    //stored in the array scores
	    while (index<buffer.length) 
		buffer[index++] = readScore();
	}
	//System.out.println("Initialized parser, index " + index);
    }

    //protected void moveBuffer(int start) {
	//Read a block

	//If it includes the starting position, we should save the
	//contents in the buffer.

    protected double readScore() throws IOException {
	if (index<blockEnd) {
	    index++;
	    return 0;
	}
	else {
	    String line = phastConsReader.readLine();
	    if (line!=null) {
		Matcher m = phastConsHeaderPattern.matcher(line);
		if (m.find()) {
		    blockEnd = Integer.parseInt(m.group(2));
		    index++;
		    return 0;
		}
		else {
		    index++;
		    return Double.parseDouble(line);
		}
	    }
	}
	return 0;
    }

    public static double[] shiftArrayLeft(double a[], double d) {
	double b[] = new double[a.length];
	for (int i=1; i<a.length; i++)
	    b[i-1] = a[i];
	b[b.length - 1] = d;
	return b;
    }

    protected double[] parseBlock(BufferedReader br, int start) throws IOException {
	String line = br.readLine();
	Vector s = new Vector();
	while (line!=null) {
	    Matcher m = phastConsHeaderPattern.matcher(line);
	    if (m.find()) {
		int end = Integer.parseInt(m.group(2));
		return getDoubleArray(s, end - start);
	    }
	    else
		s.addElement(new Double(Double.parseDouble(line)));
	    line = br.readLine();
	}
	return getDoubleArray(s, s.size());
    }

    protected double[] getDoubleArray(Vector s, int len) {
	double v[] = new double[len];
	Arrays.fill(v, 0);
	for (int i=0; i<s.size(); i++)
	    v[i] = ((Double)s.elementAt(i)).doubleValue();
	return v;
    }

    //public double[] findScores(int start, int end) {
    //	double[][] s = findScores(new int[]{start}, end-start);
    //return s[0];
    //}

    public double[][] findScores(int starts[], int len) {
	double consScores[][] = new double[starts.length][len];
	//try {
	//if (phastConsReader==null)
	//phastConsReader = new BufferedReader(new FileReader(fileName));
	int i=0;
	try {
	    for (i=0; i<starts.length; i++) 
		consScores[i] = findScores(starts[i], starts[i] + len);
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    System.out.println("Failed to read scores at " + i);
	    buffer = null;
	    try {
		phastConsReader.close();
	    }
	    catch (IOException ex) {
		System.err.println("Error closing conservation file!");
	    }
	    phastConsReader = null;
	    PhastConsParser pcp = new PhastConsParser(fileName);
	    int startsLeft[] = new int[starts.length - i];
	    for (int j=0; j<startsLeft.length; j++) 
		startsLeft[j] = starts[i + j];
	    double scoresLeft[][] = pcp.findScores(startsLeft, len);
	    for (int j=0; j<scoresLeft.length; j++) 
		consScores[j + i] = scoresLeft[j];
	}
	return consScores;
    }

}
