//Use this method to parse and keep track of the mRNA-seq data
//collected from the files prepared by David Harmin.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import java.io.*;


public class RNASeqData {

    public static final String SENSE_EXON_DENSITY = "S_EXN_Density";
    public static final String SENSE_INTRON_DENSITY = "S_INT_Density";
    public static final String SENSE_TRANSCRIPT_DENSITY = "S_XPT_Density";
    
    //List of the features that we are interested in
    public static String interestingHeaderNames[] = new String[]{"Gene", "GeneID", "xStop", "S_EXN_Density", SENSE_EXON_DENSITY, SENSE_TRANSCRIPT_DENSITY, SENSE_INTRON_DENSITY, "Chrom", "Orientation", "xStart"};



    //Map column name -> column number
    protected TreeMap headerNames2ColumnMap = new TreeMap();
    //Focus only on the subset of interesting columns
    protected int interestingHeaderColumns[] = new int[interestingHeaderNames.length];
    protected TreeSet<RNASeqGene> rnaSeqData = new TreeSet<RNASeqGene>();

    public RNASeqData(String fileName, boolean orchestra) {
	//Get the annotation information
	RefSeqGenes annotationInfo[] = new RefSeqGenes[21];
	for (int i=1; i<=21; i++)
	    annotationInfo[i-1] = new RefSeqGenes(i, orchestra);
	//Parse the data from the file
	parseDataFile(fileName, annotationInfo);
    }

    public RNASeqData(int t, boolean orchestra) {
	if (t==0 || t==1 || t==6) {
	    //Get the annotation information
	    RefSeqGenes annotationInfo[] = new RefSeqGenes[21];
	    for (int i=1; i<=21; i++)
		annotationInfo[i-1] = new RefSeqGenes(i, orchestra);
	    String fileName = (orchestra ? "/home/mh168" : "/home/mhemberg") + "/Projects/ChIP-Seq/TKData/mRNA/hr" + t + "_unique__Genes-FEATURES-READS-r35_x500_N10M__ALL-GENES__EXN-ALL_INT-ALL.tsv";
	    //Parse the data from the file
	    parseDataFile(fileName, annotationInfo);
	}
	else
	    System.out.println("Error, no data available for time " + t);
    }

    public RNASeqData(String fileName, RefSeqGenes annotationInfo[]) {
	//Parse the data from the file
	parseDataFile(fileName, annotationInfo);
    }

    public RNASeqData(String fileName, RefSeqGenes annotationInfo, int chr) {
	//Parse the data from the file
	parseDataFile(fileName, annotationInfo, chr);
    }

    public RNASeqData(String fileName, int chr, boolean orchestra) {
	//Get the annotation information
	RefSeqGenes annotationInfo = new RefSeqGenes(chr, orchestra);
	//parseDataFile(fileName, chr, annotationInfo);
    }

    //////////////////////////////////////////////////////////////////

    public void parseDataFile(String fileName, RefSeqGenes annotationInfo[]) { 
	for (int i=1; i<=21; i++)
	    parseDataFile(fileName, annotationInfo[i-1], i); 
    }

    public void parseDataFile(String fileName, RefSeqGenes annotationInfo, int chr) {
	Arrays.sort(interestingHeaderNames);
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Parse the header
	    String line = br.readLine();
	    Pattern headerPattern = Pattern.compile("([\\w|\\_|\\-|\\d|-|_]+)\\s+");
	    Matcher m = headerPattern.matcher(line);
	    int nColumns = 0;
	    while (m.find()) {
		headerNames2ColumnMap.put(m.group(1), new Integer(nColumns++));
		int ind = Arrays.binarySearch(interestingHeaderNames, m.group(1), null);
		//System.out.println("Found header " + m.group(1) + " at " + nColumns);
		if (ind>=0) {
		    interestingHeaderColumns[ind] = nColumns - 1 - 2;
		    //System.out.println("Found header " + m.group(1) + " at " + nColumns);
		}
	    }
	    //Now parse the rest of the file
	    line = br.readLine();
	    Pattern columnPattern = Pattern.compile("([\\w|_|\\d|\\@|\\||\\.|-|\\+|\\(|\\)\\^]+)\\s+");
	    while (line!=null) {
		int i = 0;
		m = columnPattern.matcher(line);
		m.find(); //zeroth column is index
		if (m.find()) {//First column is the gene name
		    String name = m.group(1);
		    if (m.find()) {//Second column is the gene ID
			int geneID = Integer.parseInt(m.group(1)), chrom;
			if (m.find()) {//Third column is the chromosome
			    if (m.group(1).equals("X")) { chrom = 20; }
			    else if (m.group(1).equals("Y")) { chrom = 21; }
			    else { chrom = Integer.parseInt(m.group(1)); }
			    if (chr==chrom) {
				RefSeqGene g = annotationInfo!=null ? annotationInfo.getGene(geneID) : null;
				if (g!=null) {
				    RNASeqGene rg = new RNASeqGene(g);
				    nColumns = 3;
				    while (m.find()) {
					int ind = Arrays.binarySearch(interestingHeaderColumns, nColumns++);
					if (ind>=0) { //Found something interesting
					    if (interestingHeaderNames[ind].equals(SENSE_EXON_DENSITY))
						rg.setExonDensity(Double.parseDouble(m.group(1)));
					    else if (interestingHeaderNames[ind].equals(SENSE_TRANSCRIPT_DENSITY))
						rg.setTranscriptDensity(Double.parseDouble(m.group(1)));
					    //else if (interestingHeaderNames[ind].equals(SENSE_INTRON_DENSITY))
					    //	rg.setIntronDensity(Double.parseDouble(m.group(1)));
					}
				    }
				    rnaSeqData.add(rg);
				}
				else
				    System.err.println("Error, could not find annotation info for " + name + ", ID = " + geneID);
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error opening " + fileName);
	}
    }

    public TreeSet<RNASeqGene> getRNASeqData() { return rnaSeqData; }

    public double getExonDensity(int geneID) {
	for (RNASeqGene g : rnaSeqData) {
	    if (g.getAnnotationInfo().getID()==geneID)
		return g.getExonDensity();
	}
	return -1;
    }

    public double getTranscriptDensity(int geneID) {
	for (RNASeqGene g : rnaSeqData) {
	    if (g.getAnnotationInfo().getID()==geneID)
		return g.getTranscriptDensity();
	}
	return -1;
    }

    public RefSeqGene getAnnotationInfo(int geneID) {
	for (RNASeqGene g : rnaSeqData) {
	    if (g.getAnnotationInfo().getID()==geneID)
		return g.getAnnotationInfo();
	}
	return null;
    }

    public int[] getGeneIDs() {
	int ids[] = new int[rnaSeqData.size()], i = 0;
	for (RNASeqGene g : rnaSeqData) 
	    ids[i++] = g.getAnnotationInfo().getID();
	return ids;
    }

}

