import java.util.*;
import edu.uci.ics.jung.graph.impl.*;
import java.io.Serializable;

public class RefSeqGene extends UndirectedSparseVertex implements Comparable, Serializable {

    //Use this class to represent one gene. start will always be
    //smaller than end and the boolean strand tells us from which
    //direction the gene will be transcribed.

    public static final String UNIQUE_TRANSCRIPT_GENE_STR = "Unique";
    public static final String MERGED_TRANSCRIPT_GENE_STR = "Merged";
    public static final String SPLIT_TRANSCRIPT_GENE_STR = "Split";
    public static final String MISSED_TRANSCRIPT_GENE_STR = "Missed";

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    protected int chr, id, maxStart, maxEnd;
    protected Vector<Integer> starts = new Vector<Integer>(), ends = new Vector<Integer>();
    protected boolean strand, coding = true;
    protected String refSeq, name;
    protected Vector<Integer> exonStarts = null, exonStops = null;
    protected Vector<Integer> intronStarts = null, intronStops = null;
    protected Vector<Integer> utr5pStarts = null, utr5pStops = null;
    protected Vector<Integer> utr3pStarts = null, utr3pStops = null;
    //Keep track of the number of reads, can either be across coding
    //region or just at TSS
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();
    //For keeping track of the RNAi data, <condition, <time, value>>
    protected TreeMap<String, Vector<Double>> rnaiData = new TreeMap<String, Vector<Double>>();
    //For keeping track of the exon densities as calculated by DH
    protected LinkedHashMap<String, Double> exonDensities = new LinkedHashMap<String, Double>();

    //////////////////////////////////////////////////////////////////

    public RefSeqGene(int c, int s, int e, boolean st, String n, int i, String r) {
	chr = c;
	starts.addElement(new Integer(s));
	ends.addElement(new Integer(e));
	maxStart = s;
	maxEnd = e;
	strand = st;
	name = n;
	refSeq = r;
	id = i;
	coding = refSeq.contains("M");
    }

    public RefSeqGene(int c, int s, int e, boolean st, String n, int i) {
	chr = c;
	starts.addElement(new Integer(s));
	ends.addElement(new Integer(e));
	maxStart = s;
	maxEnd = e;
	strand = st;
	name = n;
	id = i;
    }

    ///////////////////////////////////////////////////////

    public int peakDistance(int peakPos) {
	//Returns the shortest distance to the peak from either the
	//transcription start site or stop site, depending on which
	//one is smaller. Negative values for genes which are upstream
	//and positive values for genes which are downstream.
	int startDist = strand ? peakPos - maxStart : maxEnd - peakPos;
	int endDist = strand ? maxEnd - peakPos : peakPos - maxStart;
	return Math.abs(startDist)<Math.abs(endDist) ? startDist : endDist;
    }

    public int getMinTSSDistance(int pos) { return getMinTSSDistance(pos, false); }

    public int getMinTSSDistance(int pos, boolean firstTSSOnly) {
	int minDist = firstTSSOnly ? maxStart - pos : Integer.MAX_VALUE;
	if (!firstTSSOnly) {
	    Iterator it = starts.iterator();
	    while (it.hasNext()) {
		int start = ((Integer)it.next()).intValue();
		minDist = Math.abs(start - pos)<Math.abs(minDist) ? start - pos : minDist;
	    }
	}
	return minDist;
    }

    public int getMinTESDistance(int pos) { return getMinTESDistance(pos, false); }

    public int getMinTESDistance(int pos, boolean lastTESOnly) {
	int minDist = lastTESOnly ? maxEnd - pos : Integer.MAX_VALUE;
	if (!lastTESOnly) {
	    Iterator it = ends.iterator();
	    while (it.hasNext()) {
		int end = ((Integer)it.next()).intValue();
		minDist = Math.abs(end - pos)<Math.abs(minDist) ? end - pos : minDist;
	    }
	}
	return minDist;
    }

    public int getRandomTSS(Random random) {
	return ((Integer)starts.elementAt(random.nextInt(starts.size()))).intValue();
    }

    public int numberOfTSS() { return starts.size(); }

    public int numberOfTSE() { return ends.size(); }

    public void addStart(int s) { 
	starts.addElement(new Integer(s)); 
	if (getStrand())
	    maxStart = maxStart<s ? maxStart : s;
	else
	    maxStart = maxStart>s ? maxStart : s;
    }

    public void addEnd(int e) { 
	ends.addElement(new Integer(e)); 
	if (getStrand())
	    maxEnd = maxEnd>e ? maxEnd : e;
	else
	    maxEnd = maxEnd<e ? maxEnd : e;
    }

    public int getStart() { return maxStart; }

    public Vector<Integer> getStarts() { return starts; }
    
    public int getEnd() { return maxEnd; }

    public Vector<Integer> getEnds() { return ends; }

    public int getStart(int i) { return ((Integer)starts.elementAt(i)).intValue(); }
    
    public int[] getAlternativeStarts() { return getAlternativeStarts(false); }

    public int[] getAlternativeStarts(boolean unique) {
	int as[] = null, i = 0;
	if (!unique) {
	    TreeSet<Integer> ss = new TreeSet<Integer>();
	    Iterator it = starts.iterator();
	    while (it.hasNext()) {
		int s = ((Integer)it.next()).intValue();
		if (s!=maxStart)
		    ss.add(s);
	    }
	    as = new int[ss.size()];
	    for (Integer s : ss)
		as[i++] = s;
	}
	else {
	    Vector<Integer> ss = new Vector<Integer>();
	    Iterator it = starts.iterator();
	    while (it.hasNext()) {
		int s = ((Integer)it.next()).intValue();
		if (s!=maxStart)
		    ss.addElement(s);
	    }
	    as = new int[ss.size()];
	    for (Integer s : ss)
		as[i++] = s;
	}
	return as;
    }

    public int getEnd(int i) { return ((Integer)ends.elementAt(i)).intValue(); }

    public int getLength() { return Math.abs(maxStart - maxEnd); }

    public boolean isNoncoding() { return !coding; }

    //////////////////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : null; }

    public ABIRead[] getFactorReads() { return factorReads.size()>0 ? factorReads.get(factorReads.firstKey()) : null; }

    public ABIRead[] getExonicFactorReads(String tfName) { 
	Vector<ABIRead> exonReads = new Vector<ABIRead>();
	ABIRead[] rs = factorReads.get(tfName);
	for (int i=0; i<numberOfExons(); i++) {
	    int s = getStrand() ? getExonStart(i) : getExonStop(i), e = getStrand() ? getExonStop(i) : getExonStart(i);
	    //Find the first read that starts inside the exon
	    int ind = Arrays.binarySearch(rs, new ABIRead(chr, true, s), null);
	    ind = ind<0 ? -ind-1 : ind;
	    while (ind<rs.length && rs[ind].getPosition()<=e)
		exonReads.addElement(rs[ind++]);
	}
	return rs;
    }

    public TreeSet<String> getFactors() { return new TreeSet<String>(factorReads.keySet()); }

    public double getReadDensity(String tfName) { return (double)getFactorReads(tfName).length/getLength(); }

    //////////////////////////////////////////////////////////////////

    public void setExonDensity(String k, double v) { exonDensities.put(k, v); }

    public double getExonDensity(String k) { return exonDensities.containsKey(k) ? exonDensities.get(k) : -1; }

    public double getInductionRatio(boolean ieg) { 
	//Assume that we have total RNA stored
	if (exonDensities.containsKey(AnalyzePeaksKClRNAi.RNA_SEQ_NAMES[0]) && exonDensities.containsKey(AnalyzePeaksKClRNAi.RNA_SEQ_NAMES[1]) && exonDensities.containsKey(AnalyzePeaksKClRNAi.RNA_SEQ_NAMES[2]))
	    return (ieg ? exonDensities.get(AnalyzePeaksKClRNAi.RNA_SEQ_NAMES[1]) : exonDensities.get(AnalyzePeaksKClRNAi.RNA_SEQ_NAMES[2]))/exonDensities.get(AnalyzePeaksKClRNAi.RNA_SEQ_NAMES[0]);
	return -1;
    }

    public boolean hasInductionPattern(double inductionThres, boolean iegUp, boolean lrgUp, boolean iegDown, boolean lrgDown, boolean aboveThres) {
	boolean ok = true;
	if (iegUp)
	    ok = ok && aboveThres==getInductionRatio(true)>inductionThres;
	if (iegDown)
	    ok = ok && aboveThres==getInductionRatio(true)<1/inductionThres;
	if (lrgUp)
	    ok = ok && aboveThres==getInductionRatio(false)>inductionThres;
	if (lrgDown)
	    ok = ok && aboveThres==getInductionRatio(false)<1/inductionThres;
	return ok;
    }

    public LinkedHashSet<Double> getExonDensities() { return new LinkedHashSet<Double>(exonDensities.values()); }

    
    
    //////////////////////////////////////////////////////////////////

    public void setRNAi(String k, double v) { 
	if (!rnaiData.containsKey(k))
	    rnaiData.put(k, new Vector<Double>());
	rnaiData.get(k).addElement(v); 
    }

    public Vector<Double> getRNAi(String k) { return rnaiData.containsKey(k) ? rnaiData.get(k) : new Vector<Double>(); }

    public boolean hasRNAi() { return rnaiData!=null && !rnaiData.isEmpty(); }

    public boolean hasRNAi(String header) { return rnaiData!=null && !rnaiData.isEmpty() && rnaiData.containsKey(header); }

    //////////////////////////////////////////////////////////////////

    public boolean getStrand() { return strand; }

    public String getName() { return name; }

    public int getID() { return id; }

    public String getRefSeq() { return refSeq; }

    public int getChromosome() { return chr; }

    /////////////////////////////////

    public void setNumberOfExons(int n) {
	if (exonStarts==null) {
	    exonStarts = new Vector<Integer>();
	    exonStops = new Vector<Integer>();
	}
	if (numberOfExons()<n) {
	    exonStarts.setSize(n);
	    exonStops.setSize(n);
	    for (int i=0; i<n; i++) {
		exonStarts.setElementAt(exonStarts.elementAt(i)==null ? (strand ? Integer.MAX_VALUE : 0) : exonStarts.elementAt(i), i);
		exonStops.setElementAt(exonStops.elementAt(i)==null ? (strand ? 0 : Integer.MAX_VALUE) : exonStops.elementAt(i), i);
	    }
	}
    }

    public void addExon(int i, int start, int stop) {
	exonStarts.setElementAt(strand ? Math.min(start, exonStarts.elementAt(i)) : Math.max(start, exonStarts.elementAt(i)), i);
	exonStops.setElementAt(strand ? Math.max(stop, exonStops.elementAt(i)) : Math.min(stop, exonStops.elementAt(i)), i);
    }

    public int numberOfExons() { return hasExonInfo() ? exonStops.size() : 1; }

    public int getExonStart(int i) { 
	if (!hasExonInfo() || i<0) { return getStart(); }
	return exonStarts.elementAt(i); 
    }

    public int getExonStop(int i) { 
	if (!hasExonInfo() || i>=numberOfExons()) { return getEnd(); }
	return exonStops.elementAt(i); 
    }

    public int getExonLength(int i) { return Math.abs(getExonStop(i) - getExonStart(i)); }

    public Integer[] getExonStarts() { return (Integer[])exonStarts.toArray(); }

    public Integer[] getExonStops() { return (Integer[])exonStops.toArray(); } 

    public int getExonStart(int i, boolean relative) { 
	if (relative)
	    return Math.abs(getExonStart(i) - maxStart);
	return getExonStart(i);
    }

    public int getExonStop(int i, boolean relative) { 
	if (relative)
	    return Math.abs(getExonStop(i) - maxStart);
	return getExonStop(i);
    }

    public boolean hasExonInfo() { return exonStops!=null; }

    /////////////////////////////////

    public void setNumberOfIntrons(int n) {
	if (intronStarts==null) {
	    intronStarts = new Vector<Integer>();
	    intronStops = new Vector<Integer>();
	}
	if (numberOfIntrons()<n) {
	    intronStarts.setSize(n);
	    intronStops.setSize(n);
	    for (int i=0; i<n; i++) {
		intronStarts.setElementAt(intronStarts.elementAt(i)==null ? (strand ? Integer.MAX_VALUE : 0) : intronStarts.elementAt(i), i);
		intronStops.setElementAt(intronStops.elementAt(i)==null ? (strand ? 0 : Integer.MAX_VALUE) : intronStops.elementAt(i), i);
	    }
	}
    }

    public void addIntron(int i, int start, int stop) {
	intronStarts.setElementAt(strand ? Math.min(start, intronStarts.elementAt(i)) : Math.max(start, intronStarts.elementAt(i)), i);
	intronStops.setElementAt(strand ? Math.max(stop, intronStops.elementAt(i)) : Math.min(stop, intronStops.elementAt(i)), i);
    }

    public int numberOfIntrons() { return hasIntronInfo() ? intronStops.size() : 0; }

    public int getIntronStart(int i) { 
	if (!hasIntronInfo() || i<0) { return getStart(); }
	return intronStarts.elementAt(i); 
    }

    public int getIntronStop(int i) { 
	if (!hasIntronInfo() || i>=numberOfIntrons()) { return getEnd(); }
	return intronStops.elementAt(i); 
    }

    public Integer[] getIntronStarts() { return (Integer[])intronStarts.toArray(); }

    public Integer[] getIntronStops() { return (Integer[])intronStops.toArray(); } 

    public int getIntronStart(int i, boolean relative) { 
	if (relative)
	    return Math.abs(getIntronStart(i) - maxStart);
	return getIntronStart(i);
    }

    public int getIntronStop(int i, boolean relative) { 
	if (relative)
	    return Math.abs(getIntronStop(i) - maxStart);
	return getIntronStop(i);
    }

    public int getIntronLength(int i) { return Math.abs(getIntronStop(i) - getIntronStart(i)); }

    public boolean hasIntronInfo() { return intronStops!=null; }

    /////////////////////////////////

    public void setNumberOfUTRs(int n, boolean fiveP) {
	if (fiveP) {
	    if (utr5pStarts==null) {
		utr5pStarts = new Vector<Integer>();
		utr5pStops = new Vector<Integer>();
	    }
	    if (numberOfUTRs(fiveP)<n) {
		utr5pStarts.setSize(n);
		utr5pStops.setSize(n);
		for (int i=0; i<n; i++) {
		    utr5pStarts.setElementAt(utr5pStarts.elementAt(i)==null ? (strand ? Integer.MAX_VALUE : 0) : utr5pStarts.elementAt(i), i);
		    utr5pStops.setElementAt(utr5pStops.elementAt(i)==null ? (strand ? 0 : Integer.MAX_VALUE) : utr5pStops.elementAt(i), i);
		}
	    }
	}
	else {
	    if (utr3pStarts==null) {
		utr3pStarts = new Vector<Integer>();
		utr3pStops = new Vector<Integer>();
	    }
	    if (numberOfUTRs(fiveP)<n) {
		utr3pStarts.setSize(n);
		utr3pStops.setSize(n);
		for (int i=0; i<n; i++) {
		    utr3pStarts.setElementAt(utr3pStarts.elementAt(i)==null ? (strand ? Integer.MAX_VALUE : 0) : utr3pStarts.elementAt(i), i);
		    utr3pStops.setElementAt(utr3pStops.elementAt(i)==null ? (strand ? 0 : Integer.MAX_VALUE) : utr3pStops.elementAt(i), i);
		}
	    }
	}
    }

    public void addUTR(int i, int start, int stop, boolean fiveP) {
	if (fiveP) {
	    utr5pStarts.setElementAt(strand ? Math.min(start, utr5pStarts.elementAt(i)) : Math.max(start, utr5pStarts.elementAt(i)), i);
	    utr5pStops.setElementAt(strand ? Math.max(stop, utr5pStops.elementAt(i)) : Math.min(stop, utr5pStops.elementAt(i)), i);
	}
	else {
	    utr3pStarts.setElementAt(strand ? Math.min(start, utr3pStarts.elementAt(i)) : Math.max(start, utr3pStarts.elementAt(i)), i);
	    utr3pStops.setElementAt(strand ? Math.max(stop, utr3pStops.elementAt(i)) : Math.min(stop, utr3pStops.elementAt(i)), i);
	}
    }

    public int numberOfUTRs(boolean fiveP) { return hasUTRInfo(fiveP) ? fiveP ? utr5pStops.size() : utr3pStops.size() : 1; }

    public int getUTRStart(boolean fiveP) { return getUTRStart(0, fiveP); }

    public int getUTRStart(int i, boolean fiveP) { 
	if (!hasUTRInfo(fiveP) || i<0) { return getStart(); }
	return fiveP ? utr5pStarts.elementAt(i) : utr3pStarts.elementAt(i); 
    }

    public int getUTRStop(boolean fiveP) { return getUTRStop(0, fiveP); }

    public int getUTRStop(int i, boolean fiveP) { 
	if (!hasUTRInfo(fiveP) || i>=numberOfUTRs(fiveP)) { return getEnd(); }
	return fiveP ? utr5pStops.elementAt(i) : utr3pStops.elementAt(i); 
    }

    public boolean hasUTRInfo(boolean fiveP) { return fiveP ? utr5pStops!=null : utr3pStops!=null; }

    /////////////////////////////////

    public boolean overlaps(int start, int end) {
	//if (strand)
	//  return end>maxStart && start<maxEnd;
	//return start<maxStart && end>maxEnd;
	if (strand) {
	    if ((start>maxStart && start<maxEnd) || (end>maxStart && end<maxEnd)) //Start or end of region inside gene
		return true;
	    else if ((maxStart>start && maxStart<end) || (maxEnd>start && maxEnd<end)) //Start or end of exon inside region
		return true;
	}
	else {
	    if ((start<maxStart && start>maxEnd) || (end<maxStart && end>maxEnd)) //Start or end of region inside gene
		return true;
	    else if ((maxStart>start && maxStart<end) || (maxEnd>start && maxEnd<end)) //Start or end of exon inside region
		return true;
	}
	return false;
    }

    public boolean overlapsExon(int center) { return overlapsExon(center, center); }

    public boolean overlapsExon(int start, int end) {
	//Returns true if any part of the interval overlaps an exon
	if (hasUTRInfo(true)) {
	    for (int i=0; i<utr5pStops.size(); i++) {
		if (strand) {
		    if ((start>=utr5pStarts.elementAt(i) && start<=utr5pStops.elementAt(i)) || (end>=utr5pStarts.elementAt(i) && end<=utr5pStops.elementAt(i))) //Start or end of region inside utr5p
			return true;
		    else if ((utr5pStarts.elementAt(i)>=start && utr5pStarts.elementAt(i)<=end) || (utr5pStops.elementAt(i)>=start && utr5pStops.elementAt(i)<=end)) //Start or end of utr5p inside region
			return true;
		}
		else {
		    if ((start<=utr5pStarts.elementAt(i) && start>=utr5pStops.elementAt(i)) || (end<=utr5pStarts.elementAt(i) && end>=utr5pStops.elementAt(i))) //Start or end of region inside utr5p
			return true;
		    else if ((utr5pStarts.elementAt(i)>=start && utr5pStarts.elementAt(i)<=end) || (utr5pStops.elementAt(i)>=start && utr5pStops.elementAt(i)<=end)) //Start or end of utr5p inside region
			return true;
		}
	    }
	}
	if (hasExonInfo()) {
	    for (int i=0; i<exonStarts.size(); i++) {
		if (strand) {
		    if ((start>=exonStarts.elementAt(i) && start<=exonStops.elementAt(i)) || (end>=exonStarts.elementAt(i) && end<=exonStops.elementAt(i))) //Start or end of region inside exon
			return true;
		    else if ((exonStarts.elementAt(i)>=start && exonStarts.elementAt(i)<=end) || (exonStops.elementAt(i)>=start && exonStops.elementAt(i)<=end)) //Start or end of exon inside region
			return true;
		}
		else {
		    if ((start<=exonStarts.elementAt(i) && start>=exonStops.elementAt(i)) || (end<=exonStarts.elementAt(i) && end>=exonStops.elementAt(i))) //Start or end of region inside exon
			return true;
		    else if ((exonStarts.elementAt(i)>=start && exonStarts.elementAt(i)<=end) || (exonStops.elementAt(i)>=start && exonStops.elementAt(i)<=end)) //Start or end of exon inside region
			return true;
		}
	    }
	}
	if (hasUTRInfo(false)) {
	    for (int i=0; i<utr3pStops.size(); i++) {
		if (strand) {
		    if ((start>=utr3pStarts.elementAt(i) && start<=utr3pStops.elementAt(i)) || (end>=utr3pStarts.elementAt(i) && end<=utr3pStops.elementAt(i))) //Start or end of region inside utr3p
			return true;
		    else if ((utr3pStarts.elementAt(i)>=start && utr3pStarts.elementAt(i)<=end) || (utr3pStops.elementAt(i)>=start && utr3pStops.elementAt(i)<=end)) //Start or end of utr3p inside region
			return true;
		}
		else {
		    if ((start<=utr3pStarts.elementAt(i) && start>=utr3pStops.elementAt(i)) || (end<=utr3pStarts.elementAt(i) && end>=utr3pStops.elementAt(i))) //Start or end of region inside utr3p
			return true;
		    else if ((utr3pStarts.elementAt(i)>=start && utr3pStarts.elementAt(i)<=end) || (utr3pStops.elementAt(i)>=start && utr3pStops.elementAt(i)<=end)) //Start or end of utr3p inside region
			return true;
		}
	    }
	}
	if (!hasUTRInfo(true) && !hasExonInfo() && !hasUTRInfo(false)) { 
	    //System.err.println("Could not find exon info for " + toString());
	    if (strand) {
		if ((start>=maxStart && start<=maxEnd) || (end>=maxStart && end<=maxEnd))
		    return true;
		else if ((maxStart>=start && maxEnd<=end) || (maxEnd>=start && maxEnd<=end))
		    return true;
	    }
	    else {
		if ((start<=maxStart && start>=maxEnd) || (end<=maxStart && end>=maxEnd))
		    return true;
		else if ((maxStart<=end && maxEnd>=start) || (maxEnd>=start && maxEnd<=end))
		    return true;
	    }
	}
	return false;
    }

    public boolean exonOnly(int start, int end) {
	//Returns true if any part of the interval overlaps an exon
	if (hasUTRInfo(true)) {
	    for (int i=0; i<utr5pStops.size(); i++) {
		if (strand) {
		    if ((start>utr5pStarts.elementAt(i) && end<utr5pStops.elementAt(i))) //inside utr5p
			return true;
		}
		else {
		    if ((end<utr5pStarts.elementAt(i) && start>utr5pStops.elementAt(i))) //inside utr5p
			return true;
		}
	    }
	}
	if (hasExonInfo()) {
	    for (int i=0; i<exonStarts.size(); i++) {
		if (strand) {
		    if ((start>exonStarts.elementAt(i) && end<exonStops.elementAt(i))) //region inside exon
			return true;
		}
		else {
		    if ((end<exonStarts.elementAt(i) && start>exonStops.elementAt(i)))
			return true;
		}
	    }
	}
	if (hasUTRInfo(false)) {
	    for (int i=0; i<utr3pStops.size(); i++) {
		if (strand) {
		    if ((start>utr3pStarts.elementAt(i) && end<utr3pStops.elementAt(i))) //inside utr3p
			return true;
		}
		else {
		    if ((end<utr3pStarts.elementAt(i) && start>utr3pStops.elementAt(i))) //inside utr3p
			return true;
		}
	    }
	}
	else if (!hasExonInfo())
	    //System.err.println("Could not find exon info for " + toString());
	    return strand ? start>maxStart && end<maxEnd : start>maxEnd && end<maxStart;
	return false;
    }

    public boolean intronOnly(int start, int end) {
	//Returns true if the entire interval overlaps introns
	if (hasIntronInfo()) {
	    for (int i=0; i<intronStarts.size(); i++) {
		if (strand) {
		    if ((start>intronStarts.elementAt(i) && end<intronStops.elementAt(i))) //region inside exon
			return true;
		}
		else {
		    if ((end<intronStarts.elementAt(i) && start>intronStops.elementAt(i)))
			return true;
		}
	    }
	}
	return false;
    }

    //////////////////////////////////////////////////////////////////
    //Stuff for the JUNG package
    
    public Vector<Enhancer> getNearbyEnhancers() { return new Vector<Enhancer>(getNeighbors()); }

    //public boolean hasUniqueTranscript() { return getUniqueTranscript()!=null; }

    /*public TranscribedRegion2 getUniqueTranscript() {
	if (numNeighbors()==1) { //one overlapping
	    TranscribedRegion2 mainOverlappingRegion = (TranscribedRegion2)getNeighbors().iterator().next();
	    //if (mainOverlappingRegion.numNeighbors()==1 && (double)ReducedModel.overlap(this, mainOverlappingRegion)>.5)
		return mainOverlappingRegion; //One transcript covering more than half the gene and no other genes
	    else { //make sure that overlap of other genes is small
		Iterator<UndirectedSparseVertex> it2 = mainOverlappingRegion.getNeighbors().iterator();
		while (it2.hasNext()) {
		    UndirectedSparseVertex v = it2.next();
		    if (v instanceof RefSeqGene) {
			RefSeqGene g = (RefSeqGene)v;
			if (!this.equals(g) && (double)GeneTranscriptsGraph.overlap(g, mainOverlappingRegion)/g.getLength()>.5)//ReducedModel.MAX_OTHER_GENE_OVERLAP)
			    return null;
		    }
		}
		return mainOverlappingRegion;
	    }
	}
	else { //more than one transcript overlapping. make sure that one of them covers the majority
	    Iterator<UndirectedSparseVertex> it = getNeighbors().iterator();
	    boolean uniqueTranscripts = true;
	    TranscribedRegion2 mainOverlappingRegion = null;
	    while (it.hasNext()) {
		TranscribedRegion2 t = (TranscribedRegion2)it.next();
		if ((double)GeneTranscriptsGraph.overlap(this, t)/getLength()>.5)  //transcript mostly overlapping
		    mainOverlappingRegion = t;
	    }
	    if (mainOverlappingRegion!=null) { //make sure that it does not cover much of any other gene
		Iterator<UndirectedSparseVertex> it2 = mainOverlappingRegion.getNeighbors().iterator();
		while (it2.hasNext()) {
		    UndirectedSparseVertex v = it2.next();
		    if (v instanceof RefSeqGene) {
			RefSeqGene g = (RefSeqGene)v;
			if (!this.equals(g) && (double)GeneTranscriptsGraph.overlap(g, mainOverlappingRegion)/g.getLength()>.5)//ReducedModel.MAX_OTHER_GENE_OVERLAP)
			    return null;
		    }
		}
		return mainOverlappingRegion;
	    }
	    else
		return null;
	}
    }*/

    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof RefSeqGene))
	    System.err.println("ERROR comparing genes!");
	RefSeqGene mi = (RefSeqGene)o;
	if (chr<mi.chr)
	    return -1;
	else if (chr>mi.chr)
	    return 1;
	else {
	    if (getStart()<mi.getStart())
		return -1;
	    else if (getStart()>mi.getStart())
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return name + " " + chr + "(" + (strand ? "+" : "-") + ")" + maxStart + "-" + maxEnd;
    }

    public static String getHeaderString() { return "Name\tRefSeq\tGeneID\tStart\tEnd\tStrand"; }

    public String geneInfo() { return name + "\t" + refSeq + "\t" + id + "\t" + maxStart + "\t" + maxEnd + "\t" + (strand ? "+" : "-") + "\t"; }

    public boolean equals(Object o) {
	if (!(o instanceof RefSeqGene))
	    System.err.println("ERROR comparing genes!");
	RefSeqGene mi = (RefSeqGene)o;
	return mi.chr==chr && mi.maxStart==maxStart && mi.maxEnd==maxEnd && mi.name.equals(name) && mi.id==id;
    }

    public int hashCode() { return 5*id + 13*maxStart + 29*name.hashCode(); }

}
