//use this class to represent all the RefSeq genes in a chromosome.

import java.util.*;
import java.util.regex.*;
import java.io.*;
import umontreal.iro.lecuyer.stat.*;
import java.io.Serializable;

public class RefSeqGenes implements Serializable {

    public static final int[] CHROMOSOME_LENGTH = new int[]{197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255, 121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430, 166650296, 15902555};

    public static final int[] CHROMOSOME_LENGTH_HS = new int[]{247249724, 242951154, 199501832, 191273068, 180857871, 170899997, 158821429, 146274831, 140273257, 135374743, 134452390, 132349540, 114142986, 106368591, 100338921, 88827260, 78774748, 76117159, 63811657, 62435970, 46944329, 49691438, 154913759, 57772959};

    public static final int[] CHROMOSOME_LENGTH_SC = new int[]{230208, 813178, 316617, 1531919, 576869, 270148, 1090947, 562643, 439885, 745741, 666454, 1078175, 924429, 784333, 1091289, 948062, 85779};

    public static final int[] CHROMOSOME_LENGTH_PF = new int[]{643292, 947102,1060087,1204112  ,1343552  ,1418244  ,1501719  ,1419563  ,1541723  ,1687655  ,2038337  ,2271477  ,2895605  ,3291871  };

    /////////////////////////////////

    public static final long TOTAL_CHROMOSOME_LENGTH = (long)(265489521.8*10);
    public static final long TOTAL_CHROMOSOME_LENGTH_HS = (long)(3.080419613e+09);
    public static final long TOTAL_CHROMOSOME_LENGTH_PF = 23264339;

    public static final int MIN_INTRON_LENGTH = 1000;
    public static final int MIN_EXON_LENGTH = 100;

    protected TreeSet<RefSeqGene> genes = new TreeSet<RefSeqGene>();
    protected int chr;
    protected String species = "Mm";
    //Keep track of the genes that have a high enough density for us
    //to be interested in (or some other criterion)
    protected int[] geneIDsKeepForward, geneIDsKeepReverse;

    public RefSeqGenes(int c, boolean orchestra) {
	chr = c;
	readData(orchestra);
    }

    public RefSeqGenes(int c, boolean orchestra, String sp) {
	chr = c;
	species = sp;
	if (species.toLowerCase().equals("sc") || species.toLowerCase().equals("yeast"))
	    readYeastData(orchestra);
	else if (species.toLowerCase().equals("pf") || species.toLowerCase().equals("plasmodium")) {
	    //readYeastData(orchestra);
	}
	else
	    readData(orchestra);
    }

    public RefSeqGenes(int c, boolean orchestra, String sp, boolean hg19) {
	chr = c;
	species = sp;
	if (species.toLowerCase().equals("sc") || species.toLowerCase().equals("yeast"))
	    readYeastData(orchestra);
	else if (species.toLowerCase().equals("pf") || species.toLowerCase().equals("plasmodium")) {
	    //readYeastData(orchestra);
	}
	else
	    readData(orchestra, hg19);
    }


    //////////////////////////////////////////////////////////////////

    protected void readData(boolean orchestra) { readData(orchestra, false); }

    protected void readData(boolean orchestra, boolean hg19) {
	String genomeLoc = null;
	boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	if (species==null || species.toLowerCase().equals("mm") || species.toLowerCase().equals("mouse"))
	    genomeLoc = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/My Book/Genomes/Mm/v9/";
	else if (hs)
	    genomeLoc = (orchestra ? "/groups/neuroduo/martin/Genomes/Hs" : "/media/My Book/Genomes/Hs") + (hg19 ? "/v37/" : "/v36/");
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "refSeqGenes/mrna_list.txt"));
	    String line = br.readLine();
	    //- gene geneID refseq - chr strand start end length n_coding n_intronic
	    Pattern mrnaPattern = Pattern.compile("\\d+\\s+([\\w|\\d|\\@|\\(|\\)|/|'|\\.|-]+)\\s+(\\d+)\\s+([X|N][R|M]_\\d+\\.\\d+)\\s+\\d+\\s+([X|Y|\\d]+)\\s+([+|-])\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = mrnaPattern.matcher(line);
		if (m.find()) {
		    int c2;
		    if (m.group(4).equals("X"))
			c2 = hs ? 23 : 20;
		    else if (m.group(4).equals("Y"))
			c2 = hs ? 24 : 21;
		    else
			c2 = Integer.parseInt(m.group(4));
		    if (chr==c2) {
			boolean strand = m.group(5).equals("+");
			int start = strand ? Integer.parseInt(m.group(6)) : Integer.parseInt(m.group(7));
			int end = strand ? Integer.parseInt(m.group(7)) : Integer.parseInt(m.group(6));
			//check if a version of the gene already
			//exists. If so, we merge the two versions
			//into a super-gene which is contains the
			//union of the two versions
			int id = Integer.parseInt(m.group(2));
			RefSeqGene g = getGene(id);
			if (g==null)
			    add(new RefSeqGene(chr, start, end, strand, m.group(1), id, m.group(3)));
			else { //found version already parsed, add an alternative start and end site
			    g.addStart(start);
			    g.addEnd(end);
			}
		    }
		}
		else if (!line.startsWith("#"))
		    System.out.println("Unable to parse " + line);
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/mrna_list.txt\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	//System.out.println("Found " + mRNAInfo.size() + " genes for chromosome " + c);
	addExonInfo(genomeLoc, hs);
	addIntronInfo(genomeLoc, hs);
	addUTRInfo(genomeLoc, true, hs);
	addUTRInfo(genomeLoc, false, hs);
    }
    
    public void addExonInfo(String genomeLoc, boolean hs) {
	//Use this method to parse information about the location of
	//the exons for each gene.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "Features/exon_list.txt"));
	    //# FeatureSORT	(Gene)	(GeneID)	RefSeq	Feature	(RefSeq_which_ftr)	(nFeatures)	(Chrom)	Orientation	(FeatureStart)	(FeatureStop)	Length
	    Pattern exonPattern = Pattern.compile("\\d+\\s+([\\d|\\w]+)\\s+(\\d+)\\s+\\w{2}_\\d+\\.\\d+\\s+EXON\\s+(\\d+)\\s+(\\d+)+\\s+([\\w|\\d]+)\\s+[+|-]\\s+(\\d+)\\s+(\\d+)\\s+\\d+");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = exonPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(5).equals("X")) { c = hs ? 23 : 20; }
		    else if (m.group(5).equals("Y")) { c = hs ? 24 : 21; }
		    else { c = Integer.parseInt(m.group(5)); }
		    if (c==chr) {
			int id = Integer.parseInt(m.group(2));
			//Find the corresponding gene
			RefSeqGene gene = getGene(Integer.parseInt(m.group(2)));
			if (gene!=null) {
			    //System.out.println("Set " + Integer.parseInt(m.group(4)) + " exons for " + gene.getName());
			    gene.setNumberOfExons(Integer.parseInt(m.group(4)));
			    if (gene.getStrand()) //Add the exon info
				gene.addExon(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(6)), Integer.parseInt(m.group(7)));
			    else
				gene.addExon(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(7)), Integer.parseInt(m.group(6)));
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/exon_list.txt");
	}
    }

    public void addIntronInfo(String genomeLoc, boolean hs) {
	//Use this method to parse information about the location of
	//the exons for each gene.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "Features/intron_list.txt"));
	    //# FeatureSORT	(Gene)	(GeneID)	RefSeq	Feature	(RefSeq_which_ftr)	(nFeatures)	(Chrom)	Orientation	(FeatureStart)	(FeatureStop)	Length
	    Pattern intronPattern = Pattern.compile("\\d+\\s+([\\d|\\w]+)\\s+(\\d+)\\s+\\w{2}_\\d+\\.\\d+\\s+INtron\\s+(\\d+)\\s+(\\d+)+\\s+([\\w|\\d]+)\\s+[+|-]\\s+(\\d+)\\s+(\\d+)\\s+\\d+");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = intronPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(5).equals("X")) { c = hs ? 23 : 20; }
		    else if (m.group(5).equals("Y")) { c = hs ? 24 : 21; }
		    else { c = Integer.parseInt(m.group(5)); }
		    if (c==chr) {
			int id = Integer.parseInt(m.group(2));
			//Find the corresponding gene
			RefSeqGene gene = getGene(Integer.parseInt(m.group(2)));
			if (gene!=null) {
			    gene.setNumberOfIntrons(Integer.parseInt(m.group(4)));
			    if (gene.getStrand()) //Add the intron info
				gene.addIntron(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(6)), Integer.parseInt(m.group(7)));
			    else
				gene.addIntron(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(7)), Integer.parseInt(m.group(6)));
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/intron_list.txt");
	}
    }

    public void addUTRInfo(String genomeLoc, boolean fiveP, boolean hs) {
	//Use this method to parse information about the location of
	//the exons for each gene.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "Features/" + (fiveP ? "5" : "3") + "putr_list.txt"));
	    //# FeatureSORT	(Gene)	(GeneID)	RefSeq	Feature	(RefSeq_which_ftr)	(nFeatures)	(Chrom)	Orientation	(FeatureStart)	(FeatureStop)	Length
	    Pattern utrPattern = Pattern.compile("\\d+\\s+([\\d|\\w]+)\\s+(\\d+)\\s+\\w{2}_\\d+\\.\\d+\\s+[3|5]pUTR\\s+(\\d+)\\s+(\\d+)+\\s+([\\w|\\d]+)\\s+[+|-]\\s+(\\d+)\\s+(\\d+)\\s+\\d+");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = utrPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(5).equals("X")) { c = hs ? 23 : 20; }
		    else if (m.group(5).equals("Y")) { c = hs ? 24 : 21; }
		    else { c = Integer.parseInt(m.group(5)); }
		    if (c==chr) {
			int id = Integer.parseInt(m.group(2));
			//Find the corresponding gene
			RefSeqGene gene = getGene(Integer.parseInt(m.group(2)));
			if (gene!=null) {
			    //System.out.println("Set " + Integer.parseInt(m.group(4)) + " 5' UTRs for " + gene.getName());
			    gene.setNumberOfUTRs(Integer.parseInt(m.group(4)), fiveP);
			    if (gene.getStrand()) //Add the exon info
				gene.addUTR(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(6)), Integer.parseInt(m.group(7)), fiveP);
			    else
				gene.addUTR(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(7)), Integer.parseInt(m.group(6)), fiveP);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/5putr_list.txt");
	}
    }

    /////////////////////////////////
    
    public void readYeastData(boolean orchestra) {
	//For now we are going to ignore the exons
	String genomeLoc = orchestra ? "/groups/neuroduo/martin" : "/media/My Book";
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "/Genomes/Sc/sacCer1/saccharomyces_cerevisiae.gff"));
	    String line = br.readLine();
	    //chr start end strand name ID
	    Pattern yeastGenePattern = Pattern.compile("chr(\\w+)\\s+SGD\\s+gene\\s+(\\d+)\\s+(\\d+)[\\.|\\s]+([\\+|-]).+Name=(\\w+).+SGD:S(\\d+)");
	    while (line!=null) {
		Matcher m = yeastGenePattern.matcher(line);
		if (m.find()) {
		    int c2 = parseYeastGeneNumber(m.group(1));
		    if (chr==c2) {
			boolean strand = m.group(4).equals("+");
			int start = strand ? Integer.parseInt(m.group(2)) : Integer.parseInt(m.group(3));
			int end = strand ? Integer.parseInt(m.group(3)) : Integer.parseInt(m.group(2));
			//check if a version of the gene already
			//exists. If so, we merge the two versions
			//into a super-gene which is contains the
			//union of the two versions
			int id = Integer.parseInt(m.group(6));
			RefSeqGene g = getGene(id);
			if (g==null)
			    add(new RefSeqGene(chr, start, end, strand, m.group(5), id));
			else { //found version already parsed, add an alternative start and end site
			    g.addStart(start);
			    g.addEnd(end);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "/Genomes/Sc/sacCer1/saccharomyces_cerevisiae.gff");
	}
    }

    public int parseYeastGeneNumber(String str) {
	if (str.equals("I")) { return 1; }
	else if (str.equals("II")) { return 2; }
	else if (str.equals("III")) { return 3; }
	else if (str.equals("IV")) { return 4; }
	else if (str.equals("V")) { return 5; }
	else if (str.equals("VI")) { return 6; }
	else if (str.equals("VII")) { return 7; }
	else if (str.equals("VIII")) { return 8; }
	else if (str.equals("IX")) { return 9; }
	else if (str.equals("X")) { return 10; }
	else if (str.equals("XI")) { return 11; }
	else if (str.equals("XII")) { return 12; }
	else if (str.equals("XIII")) { return 13; }
	else if (str.equals("XIV")) { return 14; }
	else if (str.equals("XV")) { return 15; }
	else if (str.equals("XVI")) { return 16; }
	else if (str.equals("Mito")) { return 17; }
	System.err.println("Error could not parse yeast chromosome number " + str);
	return -1;
    }

    ///////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { setFactorReads(tfName, reads, 0); }

    public void setFactorReads(String tfName, ABIRead[] reads, int strand) {
	for (RefSeqGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand())) {
		int startInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), g.getStrand() ? g.getStart() : g.getEnd()));
		startInd = startInd<0 ? -startInd-1 : startInd;
		int endInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), g.getStrand() ? g.getEnd() : g.getStart()));
		endInd = endInd<0 ? -endInd-1 : endInd;
		ABIRead rs[] = new ABIRead[endInd - startInd];
		for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		g.setFactorReads(tfName, rs);
		//System.out.println(g.toString() + "\t" + rs.length + "\t" + startInd + "\t" + endInd);
	    }
	    else
		g.setFactorReads(tfName, new ABIRead[0]); 
	}
    }

    public void setFactorReads(String tfName, ABIRead[] reads, int windowSize, boolean tss) { setFactorReads(tfName, reads, windowSize, tss, 0); }

    public void setFactorReads(String tfName, ABIRead[] reads, int windowSize, boolean tss, int strand) {
	//Use this functions when we want to set reads found in a
	//window surrounding the TSS.
	for (RefSeqGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand())) {
		int startInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), tss ? g.getStart() - windowSize : g.getEnd() - windowSize));
		startInd = startInd<0 ? -startInd-1 : startInd;
		int endInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), tss ? g.getStart() + windowSize : g.getEnd() + windowSize));
		endInd = endInd<0 ? -endInd-1 : endInd;
		ABIRead rs[] = new ABIRead[endInd - startInd];
		for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		g.setFactorReads(tfName, rs);
		//System.out.println(g.toString() + "\t" + rs.length + "\t" + startInd + "\t" + endInd);
	    }
	}
    }

    public int[] getNumberOfFactorReads(String tfName, int strand) {
	int i = 0, nGenes = strand==0 ? size() : size(strand>0);
	int nReads[] = new int[nGenes];
	for (RefSeqGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand())) 
		nReads[i++] = g.getFactorReads(tfName).length;
	}
	return nReads;
    }

    public int[] getNumberOfExonicFactorReads(String tfName, int strand) {
	int i = 0, nGenes = strand==0 ? size() : size(strand>0);
	int nReads[] = new int[nGenes];
	for (RefSeqGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand())) 
		nReads[i++] = g.getExonicFactorReads(tfName).length;
	}
	return nReads;
    }

    public void setGeneIDsKeep(boolean forward, String name, double cutOff) { 
	Vector<Integer> ids = new Vector<Integer>();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward && (double)g.getFactorReads(name).length/g.getLength()>cutOff)
		ids.addElement(g.getID());
	}
	Collections.sort(ids);
	if (forward)
	    geneIDsKeepForward = new int[ids.size()];
	else
	    geneIDsKeepReverse = new int[ids.size()];
	int j = 0;
	for (Integer i : ids) {
	    if (forward)
		geneIDsKeepForward[j++] = i;
	    else
		geneIDsKeepReverse[j++] = i;
	}
    }

    public void setGeneIDsKeep(boolean forward, int[] ids) { 
	if (forward)
	    geneIDsKeepForward = ids;
	else
	    geneIDsKeepReverse = ids;
    }

    public int[] getGeneIDsKeep(boolean forward) { return forward ? geneIDsKeepForward : geneIDsKeepReverse; }

    ///////////////////////////////////////////////////////

    public void loadRNAiData(String fileName) {
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //First line is the header, assume that the 1st column is
	    //geneID, and the remaining columns are alternating
	    //intensities and present/absent calls. Use the column
	    //headers as keys for the array intensities when storing.
	    String line = br.readLine();
	    String headers[] = line.split("\t");
	    //System.out.println(headers.length + "\t" + Arrays.toString(headers));
	    line = br.readLine();
	    while (line!=null) {
		String tokens[] = line.split("\t");
		RefSeqGene g = getGene(Integer.parseInt(tokens[0]));
		if (g!=null) {
		    for (int i=1; i<=(tokens.length-1)/2; i++) {
			//Assign a negative value to indicate the probe was not called as present
			//if ()
			g.setRNAi(headers[2*i-1], (Double.parseDouble(tokens[2*i])>0 ? 1 : -1)*Double.parseDouble(tokens[2*i-1]));
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
    }

    //////////////////////////////////////////////////////////////////

    public int getChromosome() { return chr; }

    public int size() { return genes.size(); }

    public int size(boolean forward) { 
	int n = 0;
	for (RefSeqGene g : genes)
	    n += g.getStrand()==forward ? 1 : 0;
	return n;
    }

    public void add(RefSeqGene m) { genes.add(m); }

    public TreeSet<RefSeqGene> getGenes() { return genes; }

    public TreeSet<RefSeqGene> getGenes(boolean keep) { 
	TreeSet<RefSeqGene> gs = new TreeSet<RefSeqGene>();
	if (keep) {
	    for (RefSeqGene g : getGenes()) {
		if (Arrays.binarySearch(geneIDsKeepForward, g.getID())>=0 || Arrays.binarySearch(geneIDsKeepReverse, g.getID())>=0)
		    gs.add(g);
	    }
	}
	else
	    gs = genes;
	return gs;
    }

    public TreeSet<RefSeqGene> getGenes(boolean keep, boolean forward) { 
	TreeSet<RefSeqGene> gs = new TreeSet<RefSeqGene>();
	int[] gk = forward ? geneIDsKeepForward : geneIDsKeepReverse;
	if (keep) {
	    for (RefSeqGene g : getGenes()) {
		if (Arrays.binarySearch(gk, g.getID())>=0)
		    gs.add(g);
	    }
	}
	else
	    gs = genes;
	return gs;
    }

    public int[] getLengths() {
	int lens[] = new int[size()], i = 0;
	for (RefSeqGene g : genes)
	    lens[i++] = g.getLength();
	return lens;
    }

    public int[] getLengths(boolean forward) {
	int lens[] = new int[size(forward)], i = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward)
		lens[i++] = g.getLength();
	}
	return lens;
    }

    public int[] getLengths(int ids[]) {
	int lens[] = new int[ids.length], i = 0;
	for (RefSeqGene g : genes) {
	    if (Arrays.binarySearch(ids, g.getID())>-1)
		lens[i++] = g.getLength();
	}
	return lens;
    }

    public String[] getNames() {
	String names[] = new String[size()];
	int i = 0;
	for (RefSeqGene g : genes)
	    names[i++] = g.getName();
	return names;
    }

    public RefSeqGene getNearestGene(int pos) { return getNearestGene(pos, null); }

    public RefSeqGene getNearestGene(int pos, int ids[]) {
	int nearestDist = Integer.MAX_VALUE;
	RefSeqGene nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (ids==null || Arrays.binarySearch(ids, m.getID())>=0) {
		int dist = m.peakDistance(pos);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		    nearestGene = m;
		}
	    }
	}
	//System.out.println("Found distance " + nearestDist + " and gene " + nearestGene.toString() + " for peak " + pos);
	return nearestGene;
    }

    ///////////////////////////////////////////////////////

    public boolean nearTSS(int pos, int maxDist, boolean strand) {
	//Check if pos is within maxDist of a TSS on strand
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==strand) {
		for (Integer s : new Vector<Integer>(g.getStarts())) {
		    if (pos>s - maxDist && pos<s + maxDist)
			return true;
		}
	    }
	}
	return false;
    }

    public boolean overlapsTSS(int start, int end) { return overlapsTSS(start, end, false); }

    public boolean overlapsTSS(int start, int end, boolean firstTSSOnly) { return getOverlappingTSS(start, end, firstTSSOnly)!=null; }

    public RefSeqGene getOverlappingTSS(int start, int end, boolean firstTSSOnly) { 
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (firstTSSOnly && start<m.getStart() && end>m.getStart())
		return m;
	    else {
		for (int i=0; i<m.getStarts().size(); i++) {
		    if (start<((Integer)m.getStarts().elementAt(i)).intValue() && end>((Integer)m.getStarts().elementAt(i)).intValue())
			return m;
		}
	    }
	}
	return null;
    }

    public TreeSet<RefSeqGene> getOverlappingTSSs(int start, int end, boolean firstTSSOnly) { 
	TreeSet<RefSeqGene> overlapping = new TreeSet<RefSeqGene>();
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (firstTSSOnly && start<m.getStart() && end>m.getStart())
		overlapping.add(m);
	    else {
		for (int i=0; i<m.getStarts().size(); i++) {
		    if (start<((Integer)m.getStarts().elementAt(i)).intValue() && end>((Integer)m.getStarts().elementAt(i)).intValue())
			overlapping.add(m);
		}
	    }
	}
	return overlapping;
    }

    public int getNearestGeneDistance(int pos) {
	//Returns the distance to the nearest TSS or TES
	return Math.min(getNearestTSS(pos, true, null), getNearestTES(pos, true)); 
    }

    public int getNearestTSS(int pos) {	return getNearestTSS(pos, false, null); }

    public int getNearestTSS(int pos, boolean firstTSSOnly, int ids[]) {
	int nearestForward = getNearestTSS(pos, firstTSSOnly, true, ids);
	int nearestReverse = getNearestTSS(pos, firstTSSOnly, false, ids);
	return Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos) ? nearestForward : nearestReverse;
    }

    public int getNearestTSS(int pos, boolean firstTSSOnly, boolean strand) { return getNearestTSS(pos, firstTSSOnly, strand, null); }

    public int getNearestTSS(int pos, boolean firstTSSOnly, boolean strand, int ids[]) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (m.getStrand()==strand && (ids==null || Arrays.binarySearch(ids, m.getID())>=0)) {
		int dist = m.getMinTSSDistance(pos, firstTSSOnly);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		}
	    }
	}
	return Math.min(pos + nearestDist, Integer.MAX_VALUE);
    }

    public RefSeqGene getNearestTSSGene(int pos, boolean firstTSSOnly) { return getNearestTSSGene(pos, firstTSSOnly, null); }

    public RefSeqGene getNearestTSSGene(int pos, boolean firstTSSOnly, int ids[]) {
	int nearestForward = getNearestTSS(pos, firstTSSOnly, true, ids);
	int nearestReverse = getNearestTSS(pos, firstTSSOnly, false, ids);
	boolean forward = Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos) || nearestReverse==Integer.MAX_VALUE;
	return getNearestTSSGene(pos, firstTSSOnly, forward, ids);
    }

    public RefSeqGene getNearestTSSGene(int pos, boolean firstTSSOnly, boolean strand, int ids[]) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	RefSeqGene nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (m.getStrand()==strand && (ids==null || Arrays.binarySearch(ids, m.getID())>=0)) {
		int dist = m.getMinTSSDistance(pos, firstTSSOnly);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		    nearestGene = m;
		}
	    }
	}
	return nearestGene;
    }

    public TreeSet<RefSeqGene> getNearbyGenes(int pos, int maxDist, boolean firstTSSOnly) {
	//Returns a list of all genes with a TSS within maxDist of pos
	TreeSet<RefSeqGene> nearby = new TreeSet<RefSeqGene>();
	for (RefSeqGene g : getGenes()) {
	    if (firstTSSOnly && Math.abs(pos - g.getStart())<maxDist) 
		nearby.add(g);
	    else {
		for (Integer s : g.getStarts()) {
		    if (Math.abs(s - pos)<maxDist) 
			nearby.add(g);
		}
	    }
	}
	return nearby;
    }

    /////////////////////////////////

    public int getNearestTES(int pos) {	return getNearestTES(pos, false); }

    public int getNearestTES(int pos, boolean lastTESOnly) {
	int nearestForward = getNearestTES(pos, lastTESOnly, true);
	int nearestReverse = getNearestTES(pos, lastTESOnly, false);
	return Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos) ? nearestForward : nearestReverse;
    }

    public int getNearestTES(int pos, boolean lastTESOnly, boolean strand) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (m.getStrand()==strand) {
		int dist = m.getMinTESDistance(pos, lastTESOnly);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		}
	    }
	}
	return pos + nearestDist;
    }

    public RefSeqGene getNearestTESGene(int pos, boolean lastTESOnly) {
	int nearestForward = getNearestTES(pos, lastTESOnly, true);
	int nearestReverse = getNearestTES(pos, lastTESOnly, false);
	return getNearestTESGene(pos, lastTESOnly, Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos));
    }

    public RefSeqGene getNearestTESGene(int pos, boolean lastTESOnly, boolean strand) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	RefSeqGene nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (m.getStrand()==strand) {
		int dist = m.getMinTESDistance(pos, lastTESOnly);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		    nearestGene = m;
		}
	    }
	}
	return nearestGene;
    }

    /////////////////////////////////

    public RefSeqGene getGene(String name) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    String n = m.name;
	    if (n.toLowerCase().equals(name.toLowerCase()))
		return m;
	}
	return null;
    }

    public int[] getGeneIDs() {
	int ids[] = new int[genes.size()], i = 0;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    ids[i++] = m.getID();
	}
	return ids;
    }

    public RefSeqGene getGene(int id) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene m = (RefSeqGene)it.next();
	    if (m.id==id)
		return m;
	}
	return null;
    }

    public RefSeqGene getRandomGene(Random random) {
	//Use this when we are creating a null distribution where the
	//peak distances are constrained to correspond to the observed
	//ones.
	int r = random.nextInt(size()), i = 0;
	Iterator it = genes.iterator();
	RefSeqGene g = (RefSeqGene)it.next();
	while (it.hasNext() && i++<r) { g = (RefSeqGene)it.next(); }
	return g;
    }

    public int[] calculateNearestTSSDistances(int peakCenters[]) {
	int dists[] = new int[peakCenters.length];
	Vector tss = new Vector();
	Iterator it = genes.iterator();
	while (it.hasNext()) { tss.addAll(new Vector(((RefSeqGene)it.next()).getStarts())); }
	int tssLoc[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++) { tssLoc[i] = ((Integer)tss.elementAt(i)).intValue(); }
	for (int i=0; i<peakCenters.length; i++) {
	    dists[i] = Integer.MAX_VALUE;
	    for (int j=0; j<tssLoc.length; j++) 
		dists[i] = Math.min(Math.abs(peakCenters[i] - tssLoc[j]), dists[i]);
	}
	return dists;
    }

    public int[] getStarts() {
	Vector tss = new Vector();
	Iterator it = genes.iterator();
	while (it.hasNext()) { tss.addAll(new Vector(((RefSeqGene)it.next()).getStarts())); }
	int tssLoc[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++) { tssLoc[i] = ((Integer)tss.elementAt(i)).intValue(); }
	return tssLoc;
    }

    public int[] getStarts(boolean unique) {
	if (!unique) { return getStarts(); }
	//Only get the unique TSSs
	int starts[] = getStarts();
	Vector tss = new Vector();
	tss.addElement(new Integer(starts[0]));
	for (int i=1; i<starts.length; i++) {
	    if (starts[i]!=starts[i-1])
		tss.addElement(new Integer(starts[i]));
	}
	int u[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++)
	    u[i] = ((Integer)tss.elementAt(i)).intValue();
	return u;
    }

    public int[] getStarts(int[] geneIDsKeep) { return getStarts(true, geneIDsKeep); }

    public int[] getStarts(boolean unique, int[] geneIDsKeep) {
	if (geneIDsKeep==null) { return getStarts(unique); }
	//Only look at the genes which are in the list
	Vector tss = new Vector();
	Iterator it = genes.iterator();
	while (it.hasNext()) { 
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
		if (!unique)
		    tss.addAll(new Vector(g.getStarts())); 
		else {
		    tss.addElement(g.getStarts().elementAt(0));
		    for (int i=1; i<g.getStarts().size(); i++) {
			if (g.getStarts().elementAt(i-1)!=g.getStarts().elementAt(i))
			    tss.addElement(g.getStarts().elementAt(i));
		    }
		}
	    }
	}
	int u[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++)
	    u[i] = ((Integer)tss.elementAt(i)).intValue();
	return u;
    }

    public int[] getIDs() {
	int ids[] = new int[size()], i = 0;
	Iterator it = genes.iterator();
	while (it.hasNext()) { 
	    RefSeqGene g = (RefSeqGene)it.next();
	    ids[i++] = g.getID();
	}
	return ids;
    }

    public int[] getIDs(boolean internal) {
	Vector<Integer> idVec = new Vector<Integer>();
	Iterator it = genes.iterator();
	while (it.hasNext()) { 
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (g.getAlternativeStarts(true).length>0)
		idVec.addElement(g.getID());
	}
	int ids[] = new int[idVec.size()], i = 0;
	for (Integer id : idVec)
	    ids[i++] = id;
	return ids;
    }

    public int[] numberOfPeaksWithinDistance(int peakCenters[], int cutOff) {
	//Find out how many TSSs have a given number of peaks within
	//cutOff. Returns a vector where each value represents the
	//number of peaks within cutOff.
	int tssLoc[] = getStarts();
	int nPeaks[] = new int[tssLoc.length];
	for (int i=0; i<tssLoc.length; i++) {
	    nPeaks[i] = 0;
	    for (int j=0; j<peakCenters.length; j++) {
		if (Math.abs(peakCenters[j] - tssLoc[i])<cutOff)
		    nPeaks[i]++;
	    }
	}
	return nPeaks;
    }

    public int[] calculateNearestTSSDistances(boolean notSameGene) {
	//Use this method to calculate the distance to the nearest TSS
	//for all annotated TSSs.
	int dists[] = null;
	if (!notSameGene) {
	    Vector tss = new Vector();
	    Iterator it = genes.iterator();
	    while (it.hasNext()) { tss.addAll(new Vector(((RefSeqGene)it.next()).getStarts())); }
	    dists = new int[tss.size() - 1];
	    int tssLoc[] = new int[tss.size()];
	    for (int i=0; i<tss.size(); i++) { tssLoc[i] = ((Integer)tss.elementAt(i)).intValue(); }
	    dists[0] = Math.abs(tssLoc[1] - tssLoc[0]);
	    for (int i=2; i<tssLoc.length; i++) {
		int prevDist = Math.abs(tssLoc[i-1] - tssLoc[i-2]);
		int nextDist = Math.abs(tssLoc[i] - tssLoc[i-1]);
		prevDist = prevDist==0 ? Integer.MAX_VALUE : prevDist;
		nextDist = nextDist==0 ? Integer.MAX_VALUE : nextDist;
		dists[i-1] = Math.min(nextDist, prevDist);
	    }
	}
	else {
	    System.err.println("Not implemented yet!");
	}
	return dists;
    }

    public int[] interGeneDistances(boolean sameStrand) { return interGeneDistances(sameStrand, null); }

    public int[] interGeneDistances(boolean sameStrand, int ids[]) {
	//Calculate the distances between all pairs of genes. If ids[]
	//is not null, then only include genes that are contained in
	//that subset.
	Vector<Integer> dists = new Vector<Integer>();
	if (sameStrand) {
	    int prevEndForward = -1, prevStartReverse = -1;
	    for (RefSeqGene g : genes) {
		if (g.getStrand() && (ids==null || Arrays.binarySearch(ids, g.getID())>-1)) {
		    if (prevEndForward>0 && g.getStart()>prevEndForward) {
			dists.addElement(g.getStart() - prevEndForward);
			prevEndForward = g.getEnd();
		    }
		    else
			prevEndForward = Math.max(prevEndForward, g.getEnd());
		}
		else if (ids==null || Arrays.binarySearch(ids, g.getID())>-1){
		    if (prevStartReverse>0 && g.getEnd()>prevStartReverse) {
			dists.addElement(g.getEnd() - prevStartReverse);
			prevStartReverse = g.getStart();
		    }
		    else
			prevStartReverse = Math.max(prevStartReverse, g.getStart());
		}
	    }
	}
	else {
	    int prevStartForward = -1, prevStartReverse = -1;
	    for (RefSeqGene g : genes) {
		if (ids==null || Arrays.binarySearch(ids, g.getID())>-1)
		    dists.addElement(Math.abs(g.getStart() - getNearestTSS(g.getStart(), true, !g.getStrand())));
	    }
	}
	int ds[] = new int[dists.size()], i = 0;
	for (Integer d : dists)
	    ds[i++] = d;
	return ds;
    }

    public boolean overlapsGene(int start) { return overlapsGene(start, start, true) || overlapsGene(start, start, false); }

    public boolean overlapsGene(int start, int end) { return overlapsGene(start, end, true) || overlapsGene(start, end, false); }

    public boolean overlapsGene(int start, int end, boolean forward) {
	//Returns true if any part of the interval overlaps a coding region
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (g.getStrand()==forward && g.overlaps(start, end))
		return true;
	}
	return false;
    }

    public RefSeqGene getOverlappingGene(int center) { return getOverlappingGene(center, center); }

    public RefSeqGene getOverlappingGene(int start, int end) { 
	for (RefSeqGene g : genes) {
	    if (g.overlaps(start, end))
		return g;
	}
	return null;
    }

    public RefSeqGene getOverlappingGene(int start, int end, boolean forward) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (g.getStrand()==forward && g.overlaps(start, end))
		return g;
	}
	return null;
    }

    //////////////////////////////////////////////////////////////////
    
    public int getRandomRegion(int downstream, int upstream, boolean nongenic, Random random) {
	if (nongenic) {
	    //Do rejection sampling until we find a region which does
	    //not overlap any exonic or intronic region.
	    boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	    //System.out.println("Species is " + species + " and hs is " + hs);
	    int len = hs ? PeakFinder2.CHROMOSOME_LENGTH_HS[chr - 1] : PeakFinder2.CHROMOSOME_LENGTH[chr - 1];
	    int randLoc = upstream + random.nextInt(len - upstream - downstream - 1);
	    while (overlapsGene(randLoc - upstream, randLoc + downstream))
		randLoc = random.nextInt(len);
	    return randLoc;
	}
	else {
	    //Get a random gene and pull out a random part of it
	    RefSeqGene randGene = null;
	    int len = -1, geneStart = 0, geneEnd = 0;
	    while (len<=0) {
		randGene = getRandomGene(random);
		geneStart = randGene.getStart();
		geneEnd = randGene.getEnd();
		len = randGene.getLength() - downstream - upstream - 1;
	    }
	    return randGene.getStrand() ? geneStart + upstream + random.nextInt(len) : geneStart - upstream - random.nextInt(len);
	}
    }

    public int getRandomRegion(int downstream, int upstream, Random random) {
	//Do rejection sampling until we find a region which does
	//not overlap any TSS
	boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	//System.out.println("Species is " + species + " and hs is " + hs);
	int len = hs ? PeakFinder2.CHROMOSOME_LENGTH_HS[chr - 1] : PeakFinder2.CHROMOSOME_LENGTH[chr - 1];
	int randLoc = upstream + random.nextInt(len - upstream - downstream - 1);
	while (overlapsTSS(randLoc - upstream, randLoc + downstream))
	  randLoc = random.nextInt(len);
	return randLoc;
    }

    //////////////////////////////////////////////////////////////////

    //public double averageCodingRegionDensity(ABIRead[] reads, boolean forward, int nLogs) { return averageCodingRegionDensity(reads, forward, nLogs, true); }

    //public double averageCodingRegionDensity(ABIRead[] reads, boolean forward, int nLogs, boolean excludeZeros) { return averageCodingRegionDensity(reads, forward, false, nLogs, true); }
	
    public double averageCodingRegionDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		int s = forward ? g.getStart() : g.getEnd(), e = forward ? g.getEnd() : g.getStart();
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, forward ? g.getStart() : g.getEnd(), forward ? g.getEnd() : g.getStart(), reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			    d += n;
			    s += bin;
			}
			l += g.getLength()/bin;
		    }
		}
	    }
	}
	return d/l;
    }

    public double averageCodingRegionDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition();
	for (RefSeqGene g : genes) {
	    int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
	    //Check if the gene overlaps the region where we have reads
	    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		double n = FindReadsAcrossFeature.numberOfReads(chr, g.getStrand() ? g.getStart() : g.getEnd(), g.getStrand() ? g.getEnd() : g.getStart(), reads);
		if (n>0 || !excludeZeros) { //exclude genes without any reads
		    while (s<e) {
			n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			d += n;
			s += bin;
		    }
		    l += g.getLength()/bin;
		}
	    }
	}
	System.out.println("Found " + d + " reads for " + l + " coding bins.");
	return d/l;
    }

    public double averageIntergenicDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs) {
	//if subset is true, we only consider the region covered by the reads
	double d = 0, l = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition() + reads[reads.length-1].getLength();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		int s = forward ? g.getStart() : g.getEnd(), e = forward ? g.getEnd() : g.getStart();
		if (!subset || (s>=s0 && e<=e0)) {
		    while (s0<s) {
			double n = (double)FindReadsAcrossFeature.numberOfReads(chr, s0, s0 + bin, reads)/bin;
			for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			d += n;
			l += bin;
			s0 += bin; //move region forward
		    }
		    s0 = e; //skip the gene
		    if (s0>e0)
			break;
		}
	    }
	}
	return d/l;
    }

    public double averageIntergenicDensity(ABIRead[] reads, boolean subset, int bin, int nLogs) {
	//if subset is true, we only consider the region covered by the reads
	double d = 0, l = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition() + reads[reads.length-1].getLength();
	for (RefSeqGene g : genes) {
	    int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
	    if (!subset || (s>=s0 && e<=e0)) {
		while (s0<s) {
		    double n = (double)FindReadsAcrossFeature.numberOfReads(chr, s0, s0 + bin, reads)/bin;
		    for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
		    d += n;
		    l += bin;
		    s0 += bin; //move region forward
		}
		s0 = e; //skip the gene
		if (s0>e0)
		    break;
	    }
	}
	System.out.println("Found " + d + " reads for " + l + " bins.");
	return d/l;
    }

    public double[] averageExonDensities(ABIRead[] reads, boolean forward, boolean subset, int minReads) { return averageExonDensities(reads, forward, 0, subset, minReads); }

    public double[] averageExonDensities(ABIRead[] reads, boolean forward, int buffer, boolean subset, int minReads) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	TallyStore d = new TallyStore();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++) {
		    int s = forward ? g.getExonStart(i) : g.getExonStop(i), e = forward ? g.getExonStop(i) : g.getExonStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s + buffer, e - buffer, reads);
			if (n>minReads && g.getExonLength(i)>MIN_EXON_LENGTH) { //exclude genes without any reads
			    d.add(n/g.getExonLength(i));
			}
			else //Add -1 to make it possible to compare btwn different experiments
			    d.add(-1);
		    }
		}
	    }
	}
	return d.getArray();
    }

    public double averageExonDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++) {
		    int s = forward ? g.getExonStart(i) : g.getExonStop(i), e = forward ? g.getExonStop(i) : g.getExonStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    //System.out.println("Found " + n + " reads for exon " + s + "-" + e);
			    while (s<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				s += bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				d += n;
			    }
			    l += g.getExonLength(i)/bin;
			}
		    }
		}
	    }
	}
	//System.out.println("d = " + d + ", l = " + l);
	return d/l;
    }

    public double averageExonDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfExons(); i++) {
		int s = g.getStrand() ? g.getExonStart(i) : g.getExonStop(i), e = g.getStrand() ? g.getExonStop(i) : g.getExonStart(i);
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			//System.out.println("Found " + n + " reads for exon " + s + "-" + e);
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    s += bin;
			    for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
			    d += n;
			}
			l += g.getExonLength(i)/bin;
		    }
		}
	    }
	}
	//System.out.println("d = " + d + ", l = " + l);
	return d/l;
    }

    public double averageIntronDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    int s = forward ? g.getIntronStart(i) : g.getIntronStop(i), e = forward ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
			//System.out.println("Found " + n + " reads for intron " + s + "-" + e);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    while (s+bin<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				s += bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				d += n;
			    }
			    l += g.getIntronLength(i)/bin;
			}
		    }
		}
	    }
	}
	return d/l;
    }

    public double[] averageIntronDensities(ABIRead[] reads, boolean forward, boolean subset, int minReads) { return averageIntronDensities(reads, forward, 0, subset, minReads); }

    public double[] averageIntronDensities(ABIRead[] reads, boolean forward, int buffer, boolean subset, int minReads) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	TallyStore d = new TallyStore();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    int s = forward ? g.getIntronStart(i) : g.getIntronStop(i), e = forward ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s + buffer, e - buffer, reads);
			if (n>minReads && g.getIntronLength(i)>MIN_INTRON_LENGTH) { //exclude genes without any reads
			    d.add(n/g.getIntronLength(i));
			}
			else //Add -1 to make it possible to compare btwn different experiments
			    d.add(-1);
		    }
		}
	    }
	}
	return d.getArray();
    }

    public double averageIntronDensity(ABIRead[] reads, boolean forward, boolean subset, boolean excludeZeros) { return averageIntronDensity(reads, forward, 0, subset, excludeZeros); }

    public double averageIntronDensity(ABIRead[] reads, boolean forward, int buffer, boolean subset, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    int s = forward ? g.getIntronStart(i) : g.getIntronStop(i), e = forward ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s + buffer, e - buffer, reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    d += n;
			    l += g.getIntronLength(i);
			}
		    }
		}
	    }
	}
	return d/l;
    }


    public double averageIntronDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d = 0, l = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfIntrons(); i++) {
		int s = g.getStrand() ? g.getIntronStart(i) : g.getIntronStop(i), e = g.getStrand() ? g.getIntronStop(i) : g.getIntronStart(i);
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
		    //System.out.println("Found " + n + " reads for intron " + s + "-" + e);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    s += bin;
			    for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
			    d += n;
			}
			l += g.getIntronLength(i)/bin;
		    }
		}
	    }
	}
	return d/l;
    }

    public double averageCodingRegionDensity(ABIRead[] reads, int bin, int nLogs, boolean excludeZeros) { return averageCodingRegionDensity(reads, true, bin, nLogs, excludeZeros); }

    //////////////////////

    public double varianceCodingRegionDensity(ABIRead[] reads, boolean forward, double mean, int bin, int nLogs) { return varianceCodingRegionDensity(reads, forward, false, mean, bin, nLogs, true); }

    public double varianceCodingRegionDensity(ABIRead[] reads, boolean forward, boolean subset, double mean, int bin, int nLogs, boolean excludeZeros) {
	//Calculate the variance of the density in coding regions,
	//using a bin-size of bin. 
	double d2 = 0, len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		int s = forward ? g.getStart() : g.getEnd(), e = forward ? g.getEnd() : g.getStart();
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			    d2 += (n - mean)*(n - mean);
			    s += bin;
			}
			len += g.getLength()/bin;
		    }
		    //System.out.println("d2 = " + d2 + " bin = " + bin + " and l = " + len + " for " + g.toString());
		}
	    }
	}
	return d2/(len - 1);
    }

    public double varianceCodingRegionDensity(ABIRead[] reads, double mean, int bin, int nLogs) { return varianceCodingRegionDensity(reads, false, mean, bin, nLogs, true); }

    public double varianceCodingRegionDensity(ABIRead[] reads, boolean subset, double mean, int bin, int nLogs, boolean excludeZeros) {
	//Calculate the variance of the density in coding regions,
	//using a bin-size of bin. 
	double d2 = 0, len = 0;
	for (RefSeqGene g : genes) {
	    int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
	    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
		if (n>0 || !excludeZeros) { //exclude genes without any reads
		    while (s<e) {
			n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			d2 += (n - mean)*(n - mean);
			s += bin;
		    }
		    len += g.getLength()/bin;
		}
	    }
	}
	return d2/(len - 1);
    }

    public double varianceIntergenicDensity(ABIRead[] reads, boolean forward, boolean subset, double mean, int bin, int nLogs) {
	//Calculate the variance of the density in coding regions,
	//using a bin-size of bin. 
	double d2 = 0, len = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition() + reads[reads.length-1].getLength();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
		if (!subset || (s>=s0 && e<=e0)) {
		    len += (s - s0)/bin;
		    while (s0<s) {
			double n = (double)FindReadsAcrossFeature.numberOfReads(chr, s0, s0 + bin, reads)/bin;
			for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			d2 += (n - mean)*(n - mean);
			s0 += bin;
		    }
		    s0 = e;
		    if (s0>e0)
			break;
		}
	    }
	}
	return d2/(len - 1);
    }

    public double varianceIntergenicDensity(ABIRead[] reads, boolean subset, double mean, int bin, int nLogs) {
	//Calculate the variance of the density in coding regions,
	//using a bin-size of bin. 
	double d2 = 0, len = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition() + reads[reads.length-1].getLength();
	for (RefSeqGene g : genes) {
	    int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
	    if (!subset || (s>=s0 && e<=e0)) {
		len += (s - s0)/bin;
		while (s0<s) {
		    double n = (double)FindReadsAcrossFeature.numberOfReads(chr, s0, s0 + bin, reads)/bin;
		    for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
		    d2 += (n - mean)*(n - mean);
		    s0 += bin;
		}
		s0 = e;
		if (s0>e0)
		    break;
	    }
	}
	return d2/(len - 1);
    }

    public double varianceExonDensity(ABIRead[] reads, boolean forward, boolean subset, double mean, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d2 = 0, len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++) {
		    int s = forward ? g.getExonStart(i) : g.getExonStop(i), e = forward ? g.getExonStop(i) : g.getExonStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, forward ? g.getStart() : g.getEnd(), forward ? g.getEnd() : g.getStart(), reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    while (s<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				d2 += (n - mean)*(n - mean);
				s += bin;
			    }
			    len += g.getExonLength(i)/bin;
			}
		    }
		}
	    }
	}
	return d2/(len - 1);
    }
    
    public double varianceExonDensity(ABIRead[] reads, boolean subset, double mean, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d2 = 0, len = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfExons(); i++) {
		int s = g.getStrand() ? g.getExonStart(i) : g.getExonStop(i), e = g.getStrand() ? g.getExonStop(i) : g.getExonStart(i);
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, g.getStrand() ? g.getStart() : g.getEnd(), g.getStrand() ? g.getEnd() : g.getStart(), reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
			    d2 += (n - mean)*(n - mean);
			    s += bin;
			}
			len += g.getExonLength(i)/bin;
		    }
		}
	    }
	}
	return d2/(len - 1);
    }

    public double varianceIntronDensity(ABIRead[] reads, boolean forward, boolean subset, double mean, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d2 = 0, len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    int s = forward ? g.getIntronStart(i) : g.getIntronStop(i), e = forward ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, forward ? g.getStart() : g.getEnd(), forward ? g.getEnd() : g.getStart(), reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    while (s+bin<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				d2 += (n - mean)*(n - mean);
				s += bin;
			    }
			    len += g.getIntronLength(i)/bin;
			}
		    }
		}
	    }
	}
	return d2/(len - 1);
    }

    public double varianceIntronDensity(ABIRead[] reads, boolean subset, double mean, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads
	double d2 = 0, len = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfIntrons(); i++) {
		int s = g.getStrand() ? g.getIntronStart(i) : g.getIntronStop(i), e = g.getStrand() ? g.getIntronStop(i) : g.getIntronStart(i);
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, g.getStrand() ? g.getStart() : g.getEnd(), g.getStrand() ? g.getEnd() : g.getStart(), reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			while (s+bin<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
			    d2 += (n - mean)*(n - mean);
			    s += bin;
			}
			len += g.getIntronLength(i)/bin;
		    }
		}
	    }
	}
	return d2/(len - 1);
    }

    public TreeMap<Integer, Double> intronReadsBinned(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//if subset is true, we only consider genes that are in the
	//region covered by the reads. Key is poistion and value is
	//the number of reads.
	TreeMap<Integer, Double> bin2reads = new TreeMap<Integer, Double>();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    int s = forward ? g.getIntronStart(i) : g.getIntronStop(i), e = forward ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, forward ? g.getStart() : g.getEnd(), forward ? g.getEnd() : g.getStart(), reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    while (s+bin<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				bin2reads.put(s, n);
				s += bin;
			    }
			}
		    }
		}
	    }
	}
	return bin2reads;
    }

    public TallyStore intronReadsBinned(ABIRead[] reads, boolean forward, boolean subset, int bin, boolean excludeZeros) {
	TreeMap<Integer, Double> bin2reads = intronReadsBinned(reads, forward, subset, bin, 0, excludeZeros);
	TallyStore tally = new TallyStore();
	for (Double d : bin2reads.values())
	    tally.add(d);
	return tally;
    }

    public double[] intronReadsBinnedVec(ABIRead[] reads, boolean forward, boolean subset, int bin, boolean excludeZeros) { return intronReadsBinned(reads, forward, subset, bin, excludeZeros).getArray(); }


    //////////////////////

    //public double varianceWithinCodingRegionDensity(ABIRead[] reads, boolean forward, int bin, int nLogs) { return varianceWithinCodingRegionDensity(reads, forward, bin, nLogs, true); }

    public double varianceWithinCodingRegionDensity(ABIRead[] reads, boolean forward, int bin, int nLogs) { return varianceWithinCodingRegionDensity(reads, forward, false, bin, nLogs, true); }

    public double varianceWithinCodingRegionDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward && g.getLength()>bin) {
		int s = forward ? g.getStart() : g.getEnd(), e = forward ? g.getEnd() : g.getStart();
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			s += bin;
			double nPrev = n = (double)FindReadsAcrossFeature.numberOfReads(chr, s - bin, s, reads)/bin;
			for (int i=0; i<nLogs; i++) { nPrev = Math.log(nPrev + 1); }
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			    sigma2 += (n - nPrev)*(n - nPrev);
			    l++;
			    s += bin;
			    nPrev = n;
			}
		    }
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public double varianceWithinIntergenicDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition() + reads[reads.length-1].getLength();
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward && g.getLength()>bin) {
		int s = forward ? g.getStart() : g.getEnd(), e = forward ? g.getEnd() : g.getStart();
		if (!subset || (s>=s0 && e<=e0)) {
		    s0 += bin;
		    double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s0 - bin, s0, reads)/bin;
		    l += (s - s0)/bin;
		    while (s0<s) {
			double n = (double)FindReadsAcrossFeature.numberOfReads(chr, s0, s0 + bin, reads)/bin;
			for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			sigma2 += (n - nPrev)*(n - nPrev);
			s0 += bin;
			nPrev = n;
		    }
		    s0 = e;
		    if (s0>e0)
			break;
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public double varianceWithinIntergenicDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	int s0 = reads[0].getPosition(), e0 = reads[reads.length-1].getPosition() + reads[reads.length-1].getLength();
	for (RefSeqGene g : genes) {
	    if (g.getLength()>bin) {
		int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
		if (!subset || (s>=s0 && e<=e0)) {
		    s0 += bin;
		    double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s0 - bin, s0, reads)/bin;
		    l += (s - s0)/bin;
		    while (s0<s) {
			double n = (double)FindReadsAcrossFeature.numberOfReads(chr, s0, s0 + bin, reads)/bin;
			for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			sigma2 += (n - nPrev)*(n - nPrev);
			s0 += bin;
			nPrev = n;
		    }
		    s0 = e;
		    if (s0>e0)
			break;
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public double varianceWithinExonDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++) {
		    if (g.getExonLength(i)>bin) {
			int s = forward ? g.getExonStart(i) : g.getExonStop(i), e = forward ? g.getExonStop(i) : g.getExonStart(i);
			if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
			    if (n>0 || !excludeZeros) { //exclude genes without any reads
				s += bin;
				double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s - bin, s, reads)/bin;
				for (int j=0; j<nLogs; j++) { nPrev = Math.log(nPrev + 1); }
				while (s<e) {
				    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				    for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				    sigma2 += (n - nPrev)*(n - nPrev);
				    l++;
				    s += bin;
				    nPrev = n;
				}
				//System.out.println("Found " + l + " bins for exon " + i + " with length " + g.getExonLength(i) + " and sigma2 = " + (sigma2/(2*l - 2)));
			    }
			}
		    }
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public double varianceWithinExonDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfExons(); i++) {
		if (g.getExonLength(i)>bin) {
		    int s = g.getStrand() ? g.getExonStart(i) : g.getExonStop(i), e = g.getStrand() ? g.getExonStop(i) : g.getExonStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    s += bin;
			    double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s - bin, s, reads)/bin;
			    for (int j=0; j<nLogs; j++) { nPrev = Math.log(nPrev + 1); }
			    while (s<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				sigma2 += (n - nPrev)*(n - nPrev);
				l++;
				s += bin;
				nPrev = n;
			    }
			    //System.out.println("Found " + l + " bins for exon " + i + " with length " + g.getExonLength(i) + " and sigma2 = " + (sigma2/(2*l - 2)));
			}
		    }
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public double varianceWithinIntronDensity(ABIRead[] reads, boolean forward, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    if (g.getIntronLength(i)>bin) {
			int s = forward ? g.getIntronStart(i) : g.getIntronStop(i), e = forward ? g.getIntronStop(i) : g.getIntronStart(i);
			if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
			    if (n>0 || !excludeZeros) { //exclude genes without any reads
				s += bin;
				double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s - bin, s, reads)/bin;
				for (int j=0; j<nLogs; j++) { nPrev = Math.log(nPrev + 1); }
				while (s<e) {
				    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				    for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				    sigma2 += (n - nPrev)*(n - nPrev);
				    l++;
				    s += bin;
				    nPrev = n;
				}
			    }
			}
		    }
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public double varianceWithinIntronDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfIntrons(); i++) {
		if (g.getIntronLength(i)>bin) {
		    int s = g.getStrand() ? g.getIntronStart(i) : g.getIntronStop(i), e = g.getStrand() ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
			double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
			if (n>0 || !excludeZeros) { //exclude genes without any reads
			    s += bin;
			    double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s - bin, s, reads)/bin;
			    for (int j=0; j<nLogs; j++) { nPrev = Math.log(nPrev + 1); }
			    while (s<e) {
				n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
				for (int j=0; j<nLogs; j++) { n = Math.log(n + 1); }
				sigma2 += (n - nPrev)*(n - nPrev);
				l++;
				s += bin;
				nPrev = n;
			    }
			}
		    }
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }


    public double varianceWithinCodingRegionDensity(ABIRead[] reads, int bin, int nLogs) { return varianceWithinCodingRegionDensity(reads, false, bin, nLogs, true); }

    public double varianceWithinCodingRegionDensity(ABIRead[] reads, boolean subset, int bin, int nLogs, boolean excludeZeros) {
	//subdivide the coding regions into segments of size bin and
	//compute the variance of consequtive segments.
	double sigma2 = 0, l = 0;
	for (RefSeqGene g : genes) {
	    if (g.getLength()>bin) {
		int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
		if (!subset || (s>=reads[0].getPosition() && e<=reads[reads.length-1].getPosition())) {
		    double n = FindReadsAcrossFeature.numberOfReads(chr, s, e, reads);
		    if (n>0 || !excludeZeros) { //exclude genes without any reads
			s += bin;
			double nPrev = (double)FindReadsAcrossFeature.numberOfReads(chr, s - bin, s, reads)/bin;
			for (int i=0; i<nLogs; i++) { nPrev = Math.log(nPrev + 1); }
			while (s<e) {
			    n = (double)FindReadsAcrossFeature.numberOfReads(chr, s, s + bin, reads)/bin;
			    for (int i=0; i<nLogs; i++) { n = Math.log(n + 1); }
			    sigma2 += (n - nPrev)*(n - nPrev);
			    l++;
			    s += bin;
			    nPrev = n;
			}
		    }
		}
	    }
	}
	return sigma2/(2*(l - 1));
    }

    public int codingRegionLength(boolean forward) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) 
		len += g.getLength();
	}
	return len;
    }

    public int codingRegionLength(boolean forward, int start, int end) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		int s = forward ? g.getStart() : g.getEnd(), e = forward ? g.getEnd() : g.getStart();
		if (g.getStart()>=start && g.getEnd()<=end)
		    len += g.getLength();
	    }
	}
	return len;
    }

    public int codingRegionLength() {
	int len = 0;
	for (RefSeqGene g : genes) 
	    len += g.getLength();
	return len;
    }

    public int codingRegionLength(int start, int end) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    int s = g.getStrand() ? g.getStart() : g.getEnd(), e = g.getStrand() ? g.getEnd() : g.getStart();
	    if (g.getStart()>=start && g.getEnd()<=end)
		len += g.getLength();
	}
	return len;
    }

    public int exonLength(boolean forward) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++)
		    len += g.getExonLength(i);
	    }
	}
	return len;
    }

    public int[] getExonLengths(boolean forward) {
	int lens[] = new int[size(forward)], j = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++)
		    lens[j] += g.getExonLength(i);
		j++;
	    }
	}
	return lens;
    }

    public int[] getExonLengths() {
	int lens[] = new int[size()], j = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfExons(); i++)
		lens[j] += g.getExonLength(i);
	    j++;
	}
	return lens;
    }

    public int exonLength(boolean forward, int start, int end) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfExons(); i++) {
		    int s = g.getStrand() ? g.getExonStart(i) : g.getExonStop(i), e = g.getStrand() ? g.getExonStop(i) : g.getExonStart(i);
		    if (s>=start && e<=end)
			len += g.getExonLength(i);
		}
	    }
	}
	return len;
    }

    public int exonLength(int start, int end) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfExons(); i++) {
		int s = g.getStrand() ? g.getExonStart(i) : g.getExonStop(i), e = g.getStrand() ? g.getExonStop(i) : g.getExonStart(i);
		if (s>=start && e<=end)
		    len += g.getExonLength(i);
	    }
	}
	return len;
    }

    public int intronLength(boolean forward) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++)
		    len += g.getIntronLength(i);
	    }
	}
	return len;
    }

    public int intronLength(boolean forward, int start, int end) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    if (g.getStrand()==forward) {
		for (int i=0; i<g.numberOfIntrons(); i++) {
		    int s = g.getStrand() ? g.getIntronStart(i) : g.getIntronStop(i), e = g.getStrand() ? g.getIntronStop(i) : g.getIntronStart(i);
		    if (s>=start && e<=end)
			len += g.getIntronLength(i);
		}
	    }
	}
	return len;
    }

    public int intronLength(int start, int end) {
	int len = 0;
	for (RefSeqGene g : genes) {
	    for (int i=0; i<g.numberOfIntrons(); i++) {
		int s = g.getStrand() ? g.getIntronStart(i) : g.getIntronStop(i), e = g.getStrand() ? g.getIntronStop(i) : g.getIntronStart(i);
		if (s>=start && e<=end)
		    len += g.getIntronLength(i);
	    }
	}
	return len;
    }

}
