//Use this class to represent a graph of peaks and genes. We want to
//use it to compute the size of various subsets fulfilling certain
//criteria (eg expressed genes with nearby peaks and affected by
//RNAi). We also use it to generate a Bayesian network model for predicting if a gene will be induced or affected by 

import java.util.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;
import umontreal.iro.lecuyer.probdist.HypergeometricDist;
import org.apache.commons.math.distribution.HypergeometricDistributionImpl;
import org.apache.commons.math.distribution.AbstractDistribution;
import org.apache.commons.math.distribution.AbstractIntegerDistribution;

public class RefSeqGenesPeakGraph {

    public static final String CONTROL_RNAI_NAME_B1[] = new String[]{"AM1", "AM2", "AM3", "AM4"};
    public static final String CONTROL_RNAI_NAME_B2[] = new String[]{"AM25", "AM26", "AM27", "AM28"};
    public static final String RNAI_NAME_B1[][] = new String[][]{{"AM5", "AM6", "AM7", "AM8"}, {"AM9", "AM10", "AM11", "AM12"}, {"AM13", "AM14", "AM15", "AM16"}, {"AM17", "AM18", "AM19", "AM20"}, {"AM21", "AM22", "AM23", "AM24"}};
    public static final String RNAI_NAME_B2[][] = new String[][]{{"AM29", "AM30", "AM31", "AM32"}, {"AM33", "AM34", "AM35", "AM36"}, {"AM37", "AM38", "AM39", "AM40"}, {"AM41", "AM42", "AM43", "AM44"}, {"AM45", "AM46", "AM47", "AM48"}};

    public static final String CONTROL_NPAS4_RNAI_NAME_B1[] = new String[]{"CONTROL_0h_1", "CONTROL_1h_1", "CONTROL_3h_1", "CONTROL_6h_1"};
    public static final String CONTROL_NPAS4_RNAI_NAME_B2[] = new String[]{"CONTROL_0h_2", "CONTROL_1h_2", "CONTROL_3h_2", "CONTROL_6h_2"};
    public static final String CONTROL_NPAS4_RNAI_NAME_B3[] = new String[]{"CONTROL_0h_3", "CONTROL_1h_3", "CONTROL_3h_3", "CONTROL_6h_3"};
    public static final String NPAS4_RNAI_NAME_B1[] = new String[]{"NPAS4_0h_1", "NPAS4_1h_1", "NPAS4_3h_1", "NPAS4_6h_1"};
    public static final String NPAS4_RNAI_NAME_B2[] = new String[]{"NPAS4_0h_2", "NPAS4_1h_2", "NPAS4_3h_2", "NPAS4_6h_2"};
    public static final String NPAS4_RNAI_NAME_B3[] = new String[]{"NPAS4_0h_3", "NPAS4_1h_3", "NPAS4_3h_3", "NPAS4_6h_3"};

    public static DecimalFormat format = new DecimalFormat("###.######");

    //One graph for each chromosome
    protected UndirectedSparseGraph graph[];
    //Also keep a list of all genes for quick access
    protected RefSeqGenes refSeqGenes[];

    public RefSeqGenesPeakGraph(int nChr) {	
	graph = new UndirectedSparseGraph[nChr];
	refSeqGenes = new RefSeqGenes[nChr];
    }

    ///////////////////////////////////////////////////////

    public void addGenes(RefSeqGenes gs, int chr) {
	graph[chr-1] = new UndirectedSparseGraph();
	for (RefSeqGene g : gs.getGenes())
	    graph[chr-1].addVertex( g );
	refSeqGenes[chr-1] = gs;
    }

    public TreeSet<RefSeqGene> getGenes() { return getGenes(-1, 0, false); }

    public TreeSet<RefSeqGene> getGenes(double minExp, boolean hasRNAi) { return getGenes(-1, minExp, hasRNAi); }

    public TreeSet<RefSeqGene> getGenes(int chr, double minExp, boolean hasRNAi) {
	TreeSet<RefSeqGene> gs = new TreeSet<RefSeqGene>();
	if (minExp==0 && !hasRNAi) {
	    for (int i=0; i<refSeqGenes.length; i++) {
		if (refSeqGenes[i]!=null && (chr==-1 || chr-1==i)) 
		    gs.addAll(refSeqGenes[i].getGenes());
	    }
	}
	else if (minExp==0 && hasRNAi)
	    gs.addAll(genesWithRNAi());
	else if (minExp>0 && !hasRNAi)
	    gs.addAll(genesWithExpression(minExp, false));
	else 
	    gs.addAll(intersect(genesWithExpression(minExp, false), genesWithRNAi()));
	return gs;
    }

    public void addPeaks(int chr, TreeSet<Peak> peaks, int maxDist, int minPeakDist, boolean firstTSSOnly, boolean nearestGeneOnly) {
	//Add edges between edges and peaks that are within maxDist
	for (Peak p : peaks) {
	    //If there is another peak nearby, they should be merged
	    Peak nearby = findNearbyPeak(chr, p.getCenter(), minPeakDist);
	    if (nearby!=null) {
		//System.out.println("Merging " + nearby.toString() + " with factors " + Arrays.toString(nearby.getFactorNames(true)) + " and "  + p.toString() + " with factors " + Arrays.toString(p.getFactorNames(true)));
		nearby.merge(p);
		//System.out.println("New peak is " + nearby.toString() + " with factors " + Arrays.toString(nearby.getFactorNames(true)));
	    }
	    else if (nearestGeneOnly) {
		//Only add edge to the nearest gene for each peak. If the peak overlaps a gene connect to the TSS of that gene
		RefSeqGene gOver = refSeqGenes[chr-1].getOverlappingGene(p.getCenter());
		if (gOver!=null) {
		    graph[chr - 1].addVertex(p);
		    graph[chr - 1].addEdge(new UndirectedSparseEdge(gOver, p));
		}
		else {
		    RefSeqGene g = refSeqGenes[chr-1].getNearestTSSGene(p.getCenter(), firstTSSOnly);
		    if (Math.abs(g.getMinTSSDistance(p.getCenter()))<minPeakDist) {
			graph[chr - 1].addVertex(p);
			graph[chr - 1].addEdge(new UndirectedSparseEdge(g, p));
		    }
		}
	    }
	    else {
		//Find all the nearby genes and add edges to each one of them
		TreeSet<RefSeqGene> nearbyGenes = refSeqGenes[chr-1].getNearbyGenes(p.getCenter(), maxDist, firstTSSOnly);
		if (!nearbyGenes.isEmpty()) {
		    graph[chr - 1].addVertex(p);
		    for (RefSeqGene g : nearbyGenes)
			graph[chr - 1].addEdge(new UndirectedSparseEdge(g, p));
		}
		RefSeqGene gOver = refSeqGenes[chr-1].getOverlappingGene(p.getCenter());
		if (gOver!=null && !nearbyGenes.contains(gOver)) {
		    if (nearbyGenes.isEmpty())
			graph[chr - 1].addVertex(p);
		    graph[chr - 1].addEdge(new UndirectedSparseEdge(gOver, p));
		}
	    }
	}
    }

    public Peak findNearbyPeak(int chr, int pos, int minPeakDist) {
	Iterator it = graph[chr - 1].getVertices().iterator();
	while (it.hasNext()) {
	    Object o = it.next();
	    if (o instanceof Peak) {
		Peak p = (Peak)o;
		if (Math.abs(p.getCenter() - pos)<minPeakDist)
		    return p;
	    }
	}
	return null;
    }

    public void removeEdges() {

    }

    public TreeSet<RefSeqGene> getRandomSubset(double fraction) {
	//Get a subset of the genes for model training
	return null;
    }

    public LinkedHashSet<Peak> getPeaks() { return getPeaks(-1); }

    public LinkedHashSet<Peak> getPeaks(int chr) {
	LinkedHashSet<Peak> ps = new LinkedHashSet<Peak>();
	for (int i=0; i<graph.length; i++) {
	    if ((chr==-1 || chr==i+1) && graph[i]!=null) {
		Iterator it = graph[i].getVertices().iterator();
		while (it.hasNext()) {
		    Object o = it.next();
		    if (o instanceof Peak) 
			ps.add((Peak)o);
		}
	    }
	}
	return ps;
    }

    public TreeSet<String> getTFNames() {
	TreeSet<String> tfs = new TreeSet<String>();
	LinkedHashSet<Peak> ps = getPeaks();
	for (Peak p : ps) 
	    tfs.addAll(p.getFactors(true).keySet());
	return tfs;
    }

    //////////////////////////////////////////////////////////////////

    public LinkedHashMap<LinkedHashSet<String>, LinkedHashSet<RefSeqGene>> findPeakRNAiOverlap(double minExp, double inductionThres, boolean up, int time, int condition, int maxTFCombinations, TreeSet<String> tfNames, double pVal) {
	LinkedHashMap<LinkedHashSet<String>, LinkedHashSet<RefSeqGene>> combos2peaksMap = new LinkedHashMap<LinkedHashSet<String>, LinkedHashSet<RefSeqGene>>();
	LinkedHashSet<LinkedHashSet<String>> tfCombos = getTFCombos(tfNames, maxTFCombinations);
	LinkedHashSet<RefSeqGene> genesWithExp = genesWithExpression(minExp, true);
	LinkedHashSet<RefSeqGene> genesWithRNAi = genesWithRNAi();
	LinkedHashSet<RefSeqGene> genesWithExpAndRNAi = intersect(genesWithRNAi, genesWithExp);
	int m = genesWithExpAndRNAi.size();
	LinkedHashSet<RefSeqGene> genesWithRNAiInduction = genesWithRNAi(up, inductionThres, time, condition);
	LinkedHashSet<RefSeqGene> genesWithExpAndRNAiInduction = intersect(genesWithRNAiInduction, genesWithExp);
	int l = genesWithExpAndRNAiInduction.size();
	for (LinkedHashSet<String> tfs : tfCombos) {
	    LinkedHashSet<RefSeqGene> genesWithTFsAndExp = intersect(genesWithExpAndRNAi, genesWithPeakTFs(tfs));
	    int k = genesWithTFsAndExp.size();
	    LinkedHashSet<RefSeqGene> genesWithTFsExpAndRNAiInduction = intersect(genesWithTFsAndExp, genesWithExpAndRNAiInduction);
	    int x = genesWithTFsExpAndRNAiInduction.size();
	    //System.out.println("HypergeometricDist.barF("+m+","+l+","+k+","+x+")");
	    HypergeometricDistributionImpl hyge= new HypergeometricDistributionImpl(m, l, k);
	    double p = 1 - hyge.cumulativeProbability(x); //HypergeometricDist.barF(m, l, k, x);
	    if (p<pVal && x>0) {
		combos2peaksMap.put(tfs, genesWithTFsExpAndRNAiInduction);
		double exp = ((double)l*k)/m; //hyge.getNumericalMean();
		System.out.println("Found " + combos2peaksMap.get(tfs).size() + " at RNAi" + (up ? "_up" : "_down") + " with enrichment " + format.format((double)genesWithTFsExpAndRNAiInduction.size()/exp) + " for " + tfs.toString() + " at time " + time + " and condition " + condition);
	    }
	}
	return combos2peaksMap;
    }

    public LinkedHashSet<RefSeqGene> findInductionRNAiOverlap(double minExp, double inductionThres, boolean ieg, boolean kclUp, double inductionThresRNAi, boolean rnaiUp, int time, int condition, double pVal) {
	LinkedHashSet<RefSeqGene> genes = new LinkedHashSet<RefSeqGene>();
	LinkedHashSet<RefSeqGene> genesWithExp = genesWithExpression(minExp, true);
	LinkedHashSet<RefSeqGene> genesWithRNAi = genesWithRNAi();
	LinkedHashSet<RefSeqGene> genesWithExpAndRNAi = intersect(genesWithRNAi, genesWithExp);
	int m = genesWithExpAndRNAi.size();
	LinkedHashSet<RefSeqGene> genesWithRNAiInduction = genesWithRNAi(rnaiUp, inductionThresRNAi, time, condition);
	LinkedHashSet<RefSeqGene> genesWithExpAndRNAiInduction = intersect(genesWithRNAiInduction, genesWithExp);
	int l = genesWithExpAndRNAiInduction.size();
	LinkedHashSet<RefSeqGene> genesWithInd = genesWithInduction(inductionThres, ieg, kclUp);
	LinkedHashSet<RefSeqGene> genesWithExpAndInd = intersect(genesWithInd, genesWithExpAndRNAi);
	int k = genesWithExpAndInd.size();
	LinkedHashSet<RefSeqGene> genesWithExpAndIndAndRNAiInduction = intersect(genesWithExpAndInd, genesWithExpAndRNAiInduction);
	int x = genesWithExpAndIndAndRNAiInduction.size();
	HypergeometricDistributionImpl hyge= new HypergeometricDistributionImpl(m, l, k);
	double p = 1 - hyge.cumulativeProbability(x); //HypergeometricDist.barF(m, l, k, x);
	//System.out.println("HypergeometricDist.barF("+m+","+l+","+k+","+x+")");
	if (p<pVal && x>0) {
	    double exp = ((double)l*k)/m; //hyge.getNumericalMean();
	    System.out.println("Found " + x + " at RNAi" + (rnaiUp ? "_up" : "_down") + " with enrichment " + format.format((double)x/exp) + " for " + (ieg ? "IEG" : "LRG") + " " + (kclUp ? "up" : "down") + " at time " + time + " and condition " + condition);
	}
	return genesWithExpAndIndAndRNAiInduction;
    }

    public LinkedHashMap<LinkedHashSet<String>, LinkedHashSet<RefSeqGene>> findPeakInductionOverlap(double minExp, double inductionThres, boolean ieg, boolean up, int maxTFCombinations, TreeSet<String> tfNames, double pVal) {
	//Use this method to calculate the number of genes that have
	//at least one peak with a given combination of TFs. Loop over
	//all combinations of cardinality less than or equal to
	//maxTFCombinations.
	LinkedHashMap<LinkedHashSet<String>, LinkedHashSet<RefSeqGene>> combos2peaksMap = new LinkedHashMap<LinkedHashSet<String>, LinkedHashSet<RefSeqGene>>();
	LinkedHashSet<LinkedHashSet<String>> tfCombos = getTFCombos(tfNames, maxTFCombinations);
	LinkedHashSet<RefSeqGene> genesWithExp = genesWithExpression(minExp, true);
	LinkedHashSet<RefSeqGene> genesWithRNAi = genesWithRNAi();
	genesWithExp = intersect(genesWithRNAi, genesWithExp);
	int m = genesWithExp.size();
	LinkedHashSet<RefSeqGene> genesWithInd = genesWithInduction(inductionThres, ieg, up);
	LinkedHashSet<RefSeqGene> genesWithExpAndInd = intersect(genesWithInd, genesWithExp);
	int l = genesWithExpAndInd.size();
	for (LinkedHashSet<String> tfs : tfCombos) {
	    LinkedHashSet<RefSeqGene> genesWithTFsAndExp = intersect(genesWithExp, genesWithPeakTFs(tfs));
	    int k = genesWithTFsAndExp.size();
	    LinkedHashSet<RefSeqGene> genesWithTFsExpAndInd = intersect(genesWithTFsAndExp, genesWithExpAndInd);
	    int x = genesWithTFsExpAndInd.size();
	    //System.out.println("HypergeometricDist.barF("+m+","+l+","+k+","+x+")");
	    HypergeometricDistributionImpl hyge= new HypergeometricDistributionImpl(m, l, k);
	    double p = 1 - hyge.cumulativeProbability(x); //HypergeometricDist.barF(m, l, k, x);
	    if (p<pVal && x>0) {
		combos2peaksMap.put(tfs, genesWithTFsExpAndInd);
		double exp = ((double)l*k)/m; //hyge.getNumericalMean();
		System.out.println("Found " + combos2peaksMap.get(tfs).size() + " at " + (ieg ? "IEGs" : "LRGs")  + (up ? "_up" : "_down") + " with enrichment " + format.format((double)genesWithTFsExpAndInd.size()/exp) + " for " + tfs.toString());
	    }
	}
	return combos2peaksMap;
    }

    public LinkedHashMap<LinkedHashSet<String>, Double> peakInductionTable(double minExp, double inductionThres, boolean ieg, boolean up, TreeSet<String> tfNames) {
	//Use this method to calculate the number of genes that have
	//at least one peak with a given combination of TFs. Loop over
	//all combinations.
	LinkedHashMap<LinkedHashSet<String>, Double> combos2enrichmentMap = new LinkedHashMap<LinkedHashSet<String>, Double>();
	LinkedHashSet<LinkedHashSet<String>> tfCombos = getTFCombos(tfNames, tfNames.size());
	//Add the case of no peaks
	tfCombos.add(new LinkedHashSet<String>());
	//Get the complement
	LinkedHashSet<RefSeqGene> genesWithExp = genesWithExpression(minExp, true);
	LinkedHashSet<RefSeqGene> genesWithRNAi = genesWithRNAi();
	genesWithExp = intersect(genesWithRNAi, genesWithExp);
	int nGenes = genesWithExp.size();
	LinkedHashSet<RefSeqGene> genesWithInd = genesWithInduction(inductionThres, ieg, up);
	LinkedHashSet<RefSeqGene> genesWithExpAndInd = intersect(genesWithInd, genesWithExp);
	int nGenesInduced = genesWithExpAndInd.size();
	//Calculate what probability of getting it right we would have using the look-up table
	double pInd = ((double)nGenesInduced)/nGenes, pIndTFs = 0;
	System.out.println("Found " + nGenesInduced + " induced genes out of " + nGenes + " and baseline is " + format.format(pInd));
	for (String t : tfNames) 
	    System.out.print(t + "\t");
	System.out.println("Induced\tNotInduced\tEnrichement");
	for (LinkedHashSet<String> tfs : tfCombos) {
	    LinkedHashSet<RefSeqGene> genesWithTFsAndExp = intersect(genesWithExp, genesWithPeakTFs(tfs, getTFComboComplements(tfs, tfNames)));
	    int nGenesWithTFs = genesWithTFsAndExp.size();
	    LinkedHashSet<RefSeqGene> genesWithTFsExpAndInd = intersect(genesWithTFsAndExp, genesWithExpAndInd);
	    int nGenesInducedWithTFs = genesWithTFsExpAndInd.size();
	    //Calculate the conditional probability of being induced given this TF combination
	    double pIndGivenTFCombo = nGenesWithTFs==0 ? 0 : ((double)nGenesInducedWithTFs/nGenesWithTFs), pEnrichment = pIndGivenTFCombo/pInd;
	    //Weigh by the number of genes in this category to get the total improved
	    pIndTFs += nGenesInducedWithTFs==0 ? (double)nGenesWithTFs/nGenes : Math.max(pIndGivenTFCombo, 1/pIndGivenTFCombo)*(double)nGenesWithTFs/nGenes;
	    for (String t : tfNames) 
	    	System.out.print((tfs.contains(t) ? "T" : "F") + "\t");
	    System.out.println(nGenesInducedWithTFs + "\t" + (nGenesWithTFs - nGenesInducedWithTFs) + "\t" + format.format(pEnrichment));
	    combos2enrichmentMap.put(tfs, pEnrichment);
	}
	System.out.println("Baseline prediction rate is " + format.format(pInd) + " and using the look-up table, the performance is " + format.format(pIndTFs));
	//Do the same but for the TFs to see if there are any indications of dependencies
	for (String tf : tfNames) {
	    LinkedHashSet<RefSeqGene> genesWithTFsAndExp = intersect(genesWithExp, genesWithPeakTFs(tf));
	    int nGenesWithTF = genesWithTFsAndExp.size();
	    //Calculate what probability of getting it right we would have using the look-up table
	    pInd = ((double)nGenesWithTF)/nGenes;
	    pIndTFs = 0;
	    System.out.println("Found " + nGenesWithTF + " genes with " + tf + " out of " + nGenes + " and baseline is " + format.format(pInd));
	    for (String t : tfNames) 
		System.out.print((!t.equals(tf) ? t + "\t" : ""));
	    System.out.println("Bound\tNotBound\tEnrichement");
	    for (LinkedHashSet<String> tfs : tfCombos) {
		if (tfs.contains(tf)) {
		    tfs.remove(tf);
		    LinkedHashSet<RefSeqGene> genesWithoutTFOtherTFsAndExp = intersect(genesWithExp, genesWithPeakTFs(tfs, getTFComboComplements(tfs, tfNames)));
		    int nGenesWithoutTFs = genesWithoutTFOtherTFsAndExp.size();
		    tfs.add(tf);
		    LinkedHashSet<RefSeqGene> genesWithTFOtherTFsAndExp = intersect(genesWithExp, genesWithPeakTFs(tfs, getTFComboComplements(tfs, tfNames)));
		    int nGenesWithTFs = genesWithTFOtherTFsAndExp.size();
		    //Calculate the conditional probability of having TF given this TF combination
		    double pIndGivenTFCombo = (nGenesWithoutTFs + nGenesWithTFs)==0 ? 0 : ((double)nGenesWithTFs/(nGenesWithTFs + nGenesWithoutTFs)), pEnrichment = pIndGivenTFCombo/pInd;
		    //Weigh by the number of genes in this category to get the total improved
		    pIndTFs += nGenesWithTFs==0 ? (double)nGenesWithTFs/nGenes : Math.max(pIndGivenTFCombo, 1/pIndGivenTFCombo)*(double)nGenesWithTFs/nGenes;
		    for (String t : tfNames) 
			System.out.print((!t.equals(tf) ? (tfs.contains(t) ? "T" : "F") + "\t" : ""));
		    System.out.println(nGenesWithTFs + "\t" + nGenesWithoutTFs + "\t" + format.format(pEnrichment));
		}
	    }
	}
	//Now we go through each gene and ask what pattern of TFs it has and based on the look-up table we calculate the probability that it will be expressed
	/*double fractionOfGenes = 1.0/nGenes;
	  for (RefSeqGene g : genesWithExp) {
	  //Get all the combinations at the peaks nearby
	  LinkedHashSet<LinkedHashSet<String>> tfsAtGene = getTFs(g);
	  //Find out which one is best for predicting induction
	  double bestEnrichment = 1;
	  for (LinkedHashSet<String> tfs : tfsAtGene) {
	  if (combos2enrichmentMap.get(tfs)>bestEnrichment && combos2enrichmentMap.get(tfs)>1/bestEnrichment)
	  bestEnrichment = combos2enrichmentMap.get(tfs);
	  if (combos2enrichmentMap.get(tfs)<bestEnrichment && combos2enrichmentMap.get(tfs)<1/bestEnrichment)
	  bestEnrichment = 1/combos2enrichmentMap.get(tfs);
	  //if (g.getID()==14284)
	  //  System.out.println(format.format(bestEnrichment) + " for " + tfs.toString());
	  }
	  if (bestEnrichment==0) //We know that gene will not be active with this combo
	  pIndTFs += fractionOfGenes;
	  else
	  pIndTFs += fractionOfGenes*pInd*bestEnrichment;*/
	return combos2enrichmentMap;
    }

    public LinkedHashSet<RefSeqGene> getBestTargets(double minExp, double inductionThres, double rnaiInduction, boolean ieg, boolean kclUp, boolean rnaiUp, int time, int condition, TreeSet<String> targetTFNames) {
	//Genes that are expressed
	LinkedHashSet<RefSeqGene> genesWithExp = genesWithExpression(minExp, true);
	//Genes that have RNAi info
	LinkedHashSet<RefSeqGene> genesWithRNAi = genesWithRNAi();
	genesWithExp = intersect(genesWithRNAi, genesWithExp);
	//Genes that are induced
	LinkedHashSet<RefSeqGene> genesWithInd = genesWithInduction(inductionThres, ieg, kclUp);
	LinkedHashSet<RefSeqGene> genesWithExpAndInd = intersect(genesWithInd, genesWithExp);
	//Genes that are perturbed by RNAi
	LinkedHashSet<RefSeqGene> genesWithRNAiInduction = genesWithRNAi(rnaiUp, rnaiInduction, time, condition);
	LinkedHashSet<RefSeqGene> genesWithExpAndIndAndRNAiInduction = intersect(genesWithExpAndInd, genesWithRNAiInduction);
	//Genes that have TFs
	LinkedHashSet<RefSeqGene> genesWithTFs = genesWithPeakTFs(new LinkedHashSet<String>(targetTFNames));
	LinkedHashSet<RefSeqGene> genesWithExpAndIndAndRNAiInductionAndTFs = intersect(genesWithTFs, genesWithExpAndIndAndRNAiInduction);
	//if (genesWithExpAndIndAndRNAiInductionAndTFs.size()<5) {
	//  for (RefSeqGene g : genesWithExpAndIndAndRNAiInductionAndTFs)
	//	System.out.println(g.toString() + "\t" + format.format(g.getInductionRatio(ieg)));
	//}
	//System.out.println("Found " + genesWithExpAndIndAndRNAiInductionAndTFs.size() + " genes for time " + time + " and " + condition);
	return genesWithExpAndIndAndRNAiInductionAndTFs;
    }

    public LinkedHashSet<RefSeqGene> genesWithRNAi() {
	//Find the total number of genes for which we have RNAi info
	LinkedHashSet<RefSeqGene> gs = new LinkedHashSet<RefSeqGene>();
	for (int i=0; i<refSeqGenes.length; i++) {
	    if (refSeqGenes[i]!=null) {
		for (RefSeqGene g : refSeqGenes[i].getGenes()) {
		    if (g.hasRNAi())
			gs.add(g);
		}
	    }
	}
	//System.out.println("Found " + gs.size() + " genes with RNAi info.");
	return gs;
    }

    public LinkedHashSet<RefSeqGene> genesWithRNAi(boolean up, double inductionThres, int time, int condition, String tf) {
	if (tf.equalsIgnoreCase("Npas4"))
	    return genesWithRNAi(up, inductionThres, time, tf);
	else
	    return genesWithRNAi(up, inductionThres, time, condition);
    }

    public LinkedHashSet<RefSeqGene> genesWithRNAi(boolean up, double inductionThres, int time, int condition) {
	//Default is for Fos
	LinkedHashSet<RefSeqGene> gs = new LinkedHashSet<RefSeqGene>();
	for (int i=0; i<refSeqGenes.length; i++) {
	    if (refSeqGenes[i]!=null) {
		for (RefSeqGene g : refSeqGenes[i].getGenes()) {
		    if (g.hasRNAi(RNAI_NAME_B1[0][0])) {
			//Calculate the induction relative to the control hairpin
			Vector<Double> ctl1 = g.getRNAi(CONTROL_RNAI_NAME_B1[time]), ctl2 = g.getRNAi(CONTROL_RNAI_NAME_B2[time]);
			Vector<Double> rnai1 = g.getRNAi(RNAI_NAME_B1[condition][time]), rnai2 = g.getRNAi(RNAI_NAME_B2[condition][time]);
			for (int j=0; j<ctl1.size(); j++) {
			    //if (g.getID()==14284)
			    //	System.out.println("Found RNAi " + format.format(rnai1.elementAt(j)/ctl1.elementAt(j)) + " and " + format.format(rnai2.elementAt(j)/ctl2.elementAt(j)));
			    if (up) {
				//Make sure present in the RNAi condition
				if (rnai1.elementAt(j)>0 && rnai2.elementAt(j)>0) {
				    if (Math.abs(rnai1.elementAt(j)/ctl1.elementAt(j))>inductionThres && Math.abs(rnai2.elementAt(j)/ctl2.elementAt(j))>inductionThres) {
					gs.add(g);
					break;
				    }
				}
			    }
			    else {
				if (ctl1.elementAt(j)>0 && ctl2.elementAt(j)>0) {
				    if (Math.abs(rnai1.elementAt(j)/ctl1.elementAt(j))<1/inductionThres && Math.abs(rnai2.elementAt(j)/ctl2.elementAt(j))<1/inductionThres) {
					gs.add(g);
					break;
				    }
				}
			    }
			}
		    }
		}
	    }
	}
	//System.out.println("Found " + gs.size() + " genes for condition " + condition + " and time " + time + " with induction greater than " + inductionThres);
	return gs;
    }

    public LinkedHashSet<RefSeqGene> genesWithRNAi(boolean up, double inductionThres, int time, String tf) {
	//Use for the Npas4 data
	LinkedHashSet<RefSeqGene> gs = new LinkedHashSet<RefSeqGene>();
	for (int i=0; i<refSeqGenes.length; i++) {
	    if (refSeqGenes[i]!=null) {
		for (RefSeqGene g : refSeqGenes[i].getGenes()) {
		    if (tf.equalsIgnoreCase("Npas4") && g.hasRNAi(NPAS4_RNAI_NAME_B1[0])) {
			//Calculate the induction relative to the control hairpin
			Vector<Double> ctl1 = g.getRNAi(CONTROL_NPAS4_RNAI_NAME_B1[time]), ctl2 = g.getRNAi(CONTROL_NPAS4_RNAI_NAME_B2[time]), ctl3 = g.getRNAi(CONTROL_NPAS4_RNAI_NAME_B3[time]);
			Vector<Double> rnai1 = g.getRNAi(NPAS4_RNAI_NAME_B1[time]), rnai2 = g.getRNAi(NPAS4_RNAI_NAME_B2[time]), rnai3 = g.getRNAi(NPAS4_RNAI_NAME_B3[time]);
			if (time==3) {
			    //System.out.println(g.toString() + "\t" + g.getID() + "\t" + ctl1.toString() + "\t" + ctl2.toString() + "\t" + ctl3.toString());
			    //System.out.println(g.toString() + "\t" + g.getID() + "\t" + rnai1.toString() + "\t" + rnai2.toString() + "\t" + rnai3.toString());
			}
			for (int j=0; j<ctl1.size(); j++) {
			    if (up) {
				//Make sure present in the RNAi condition
				if (rnai1.elementAt(j)>0 && rnai2.elementAt(j)>0 && rnai3.elementAt(j)>0) {
				    if (Math.abs(rnai1.elementAt(j)/ctl1.elementAt(j))>inductionThres && Math.abs(rnai2.elementAt(j)/ctl2.elementAt(j))>inductionThres && Math.abs(rnai3.elementAt(j)/ctl3.elementAt(j))>inductionThres) {
					gs.add(g);
					break;
				    }
				}
			    }
			    else {
				if (ctl1.elementAt(j)>0 && ctl2.elementAt(j)>0 && ctl3.elementAt(j)>0) {
				    if (Math.abs(rnai1.elementAt(j)/ctl1.elementAt(j))<1/inductionThres && Math.abs(rnai2.elementAt(j)/ctl2.elementAt(j))<1/inductionThres && Math.abs(rnai3.elementAt(j)/ctl3.elementAt(j))<1/inductionThres) {
					gs.add(g);
					break;
				    }
				}
			    }
			}
		    }
		}
	    }
	}
	//System.out.println("Found " + gs.size() + " genes for condition " + condition + " and time " + time + " with induction greater than " + inductionThres);
	return gs;
    }

    ////////////////////////////////////////////
    
    public LinkedHashSet<RefSeqGene> genesWithExpression(double minExp) { return genesWithExpression(minExp, false); }

    public LinkedHashSet<RefSeqGene> genesWithExpression(double minExp, boolean codingOnly) {
	LinkedHashSet<RefSeqGene> gs = new LinkedHashSet<RefSeqGene>();
	for (int j=0; j<graph.length; j++) {
	    if (refSeqGenes[j]!=null) {
		for (RefSeqGene g : refSeqGenes[j].getGenes()) {
		    boolean exp = false;
		    for (Double d : g.getExonDensities()) 
			exp = d>minExp || exp;
		    if (exp && (!codingOnly || !g.isNoncoding())) 
			gs.add(g);
		}
	    }
	}
	//System.out.println("Found " + gs.size() + " genes with expresison greater than " + minExp);
	return gs;
    }

    public LinkedHashSet<RefSeqGene> genesWithInduction(double minInd, boolean ieg, boolean up) { return genesWithInduction(minInd, ieg, up, true); }

    public LinkedHashSet<RefSeqGene> genesWithInduction(double minInd, boolean ieg, boolean up, boolean aboveThres) {
	LinkedHashSet<RefSeqGene> gs = new LinkedHashSet<RefSeqGene>();
	for (int j=0; j<graph.length; j++) {
	    if (refSeqGenes[j]!=null) {
		for (RefSeqGene g : refSeqGenes[j].getGenes()) {
		    //For IEGs, make sure that we are above the threshold at 1h and that the max/min is at 1h
		    boolean hasInduction = up ? g.getInductionRatio(ieg)>minInd && g.getInductionRatio(true)>g.getInductionRatio(false)==ieg : g.getInductionRatio(ieg)<1/minInd && g.getInductionRatio(true)<g.getInductionRatio(false)==ieg;
		    if (aboveThres==hasInduction) 
			gs.add(g);
		}
	    }
	}
	//System.out.println("Found " + gs.size() + " genes with induction greater than " + minInd);
	return gs;
    }

    ///////////////////////////////////////////////////////

    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(String tf) {  return genesWithPeakTFs(tf, 0, Integer.MAX_VALUE); }

    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(String tf, int minDist, int maxDist) { 
	LinkedHashSet<String> tfsPresent = new LinkedHashSet<String>();
	if (tf!=null)
	    tfsPresent.add(tf);
	return genesWithPeakTFs(tfsPresent, new LinkedHashSet<String>(), minDist, maxDist); 
    }

    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(String tf, LinkedHashSet<String> tfsAbsent) { 
	LinkedHashSet<String> tfsPresent = new LinkedHashSet<String>();
	if (tf!=null)
	    tfsPresent.add(tf);
	return genesWithPeakTFs(tfsPresent, tfsAbsent, 0, Integer.MAX_VALUE); 
    }

    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(LinkedHashSet<String> tfs) { return genesWithPeakTFs(tfs, new LinkedHashSet<String>()); }
	
    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(String tfPresent, String tfAbsent, int minDist, int maxDist) { 
	LinkedHashSet<String> tfsPresent = new LinkedHashSet<String>();
	if (tfPresent!=null)
	    tfsPresent.add(tfPresent);
	LinkedHashSet<String> tfsAbsent = new LinkedHashSet<String>();
	if (tfAbsent!=null)
	    tfsAbsent.add(tfAbsent);
	return genesWithPeakTFs(tfsPresent, tfsAbsent, minDist, maxDist); 
    }

    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(LinkedHashSet<String> tfsPresent, LinkedHashSet<String> tfsAbsent) { return genesWithPeakTFs(tfsPresent, tfsAbsent, 0, Integer.MAX_VALUE); }

    public LinkedHashSet<RefSeqGene> genesWithPeakTFs(LinkedHashSet<String> tfsPresent, LinkedHashSet<String> tfsAbsent, int minDist, int maxDist) {
	LinkedHashSet<RefSeqGene> genesWithTFs = new LinkedHashSet<RefSeqGene>();
	//Go through all genes
	for (int j=0; j<graph.length; j++) {
	    if (refSeqGenes[j]!=null) {
		for (RefSeqGene g : refSeqGenes[j].getGenes()) {
		    //For each gene, find the nearby peaks
		    if (g.getNeighbors().isEmpty() && tfsPresent.isEmpty())
			genesWithTFs.add(g);
		    else {
			Iterator<UndirectedSparseVertex> it = g.getNeighbors().iterator();
			//Take the union of factors found
			LinkedHashSet<String> tfsFound = new LinkedHashSet<String>();
			while (it.hasNext()) {
			    //Find out if it is at the right distance
			    Peak p = (Peak)it.next();
			    for (Integer s : g.getStarts()) {
				if (Math.abs(s-p.getCenter())>=minDist && Math.abs(s-p.getCenter())<maxDist)
				    tfsFound.addAll(p.getFactors().keySet());
			    }
			}
			//For each gene, find if it has the required combination of TFs
			boolean foundAbsent = false, foundPresent = true;
			for (String tf : tfsFound) {
			    if (tfsAbsent.contains(tf)) {
				foundAbsent = true;
				break;
			    }
			}
			for (String tf : tfsPresent) {
			    if (!tfsFound.contains(tf)) {
				foundPresent = false;
				break;
			    }
			}
			if (foundPresent && !foundAbsent)
			    genesWithTFs.add(g);
		    }
		}
	    }
	}
	//System.out.println("Found " + genesWithTFs.size() + " genes with TFs " + tfsPresent.toString() + " present and " + tfsAbsent.toString() + " absent.");
	return genesWithTFs;
    }

    public LinkedHashSet<RefSeqGene> getGenesWithTF(String tf, int minDist, int maxDist) { return getGenesWithTF(tf, minDist, maxDist, 0, false, false, false, false, false, true); }

    public LinkedHashSet<RefSeqGene> getGenesWithTF(String tf, int minDist, int maxDist, double inductionThres, boolean iegUp, boolean lrgUp, boolean iegDown, boolean lrgDown, boolean aboveThres) {return getGenesWithTF(tf, minDist, maxDist, 0, iegUp, lrgUp, iegDown, lrgDown, aboveThres, false); }

    public LinkedHashSet<RefSeqGene> getGenesWithTF(String tf, int minDist, int maxDist, double inductionThres, boolean iegUp, boolean lrgUp, boolean iegDown, boolean lrgDown, boolean aboveThres, boolean ignoreThres) {
	LinkedHashSet<RefSeqGene> genesWithTFs = new LinkedHashSet<RefSeqGene>();
	//Go through all genes
	for (int j=0; j<graph.length; j++) {
	    if (refSeqGenes[j]!=null) {
		for (RefSeqGene g : refSeqGenes[j].getGenes()) {
		    //Check if the gene has the right induction pattern
		    if (ignoreThres || g.hasInductionPattern(inductionThres, iegUp, lrgUp, iegDown, lrgDown, aboveThres)) {
			//For each gene, find the nearby peaks
			Iterator it = g.getNeighbors().iterator();
			while (it.hasNext()) {
			    Peak p = (Peak)it.next();
			    //For each peak, find if it has the required combination of TFs
			    if (p.hasAllFactors(tf, true)) {
				//Find out if it is at the right distance
				for (Integer s : g.getStarts()) {
				    if (Math.abs(s-p.getCenter())>=minDist && Math.abs(s-p.getCenter())<maxDist)
					genesWithTFs.add(g);
				}
			    }
			}
		    }
		}
	    }
	}
	//System.out.println("Found " + genesWithTFs.size() + " genes with TFs.");
	return genesWithTFs;
    }

    public LinkedHashSet<RefSeqGene> getGenesWithTF(String tf, int minDist, int maxDist, boolean present) {
	LinkedHashSet<RefSeqGene> genesWithTFs = new LinkedHashSet<RefSeqGene>();
	//Go through all genes
	for (int j=0; j<graph.length; j++) {
	    if (refSeqGenes[j]!=null) {
		for (RefSeqGene g : refSeqGenes[j].getGenes()) {
		    //For each gene, find the nearby peaks
		    Iterator it = g.getNeighbors().iterator();
		    while (it.hasNext()) {
			Peak p = (Peak)it.next();
			//For each peak, find if it has the required combination of TFs
			if (present==p.hasAllFactors(tf, false)) {
			    //Find out if it is at the right distance
			    for (Integer s : g.getStarts()) {
				if (Math.abs(s-p.getCenter())>=minDist && Math.abs(s-p.getCenter())<maxDist)
				    genesWithTFs.add(g);
			    }
			}
		    }
		}
	    }
	}
	//System.out.println("Found " + genesWithTFs.size() + " genes with TFs.");
	return genesWithTFs;
    }

    public LinkedHashSet<LinkedHashSet<String>> getTFs(RefSeqGene g) {
	LinkedHashSet<LinkedHashSet<String>> tfs = new LinkedHashSet<LinkedHashSet<String>>();
	//For each gene, find the nearby peaks
	Iterator<UndirectedSparseVertex> it = g.getNeighbors().iterator();
	while (it.hasNext()) {
	    Peak p = (Peak)it.next();
	    LinkedHashSet<String> tf = new LinkedHashSet<String>();
	    //For each peak, find the combination of TFs
	    String ts[] = p.getFactorNames(true);
	    for (int i=0; i<ts.length; i++)
		tf.add(ts[i]);
	    tfs.add(tf);
	}
	return tfs;
    }

    public LinkedHashSet<String> getTFComboComplements(String combo, TreeSet<String> tfNames) {
	LinkedHashSet<String> tfs = new LinkedHashSet<String>();
	tfs.add(combo);
	return getTFComboComplements(tfs, tfNames);
    }

    public LinkedHashSet<String> getTFComboComplements(LinkedHashSet<String> combo, TreeSet<String> tfNames) {
	//Recursively go through TF names and extract all combinations
	LinkedHashSet<String> comp = new LinkedHashSet<String>();
	for (String tf : tfNames) {
	    if (!combo.contains(tf))
		comp.add(tf);
	}
	return comp;
    }

    public static LinkedHashSet<LinkedHashSet<String>> getTFCombos(TreeSet<String> tfNames, int maxTFCombinations) {
	//Recursively go through TF names and extract all combinations
	LinkedHashSet<LinkedHashSet<String>> combos = new LinkedHashSet<LinkedHashSet<String>>();
	for (String tf : tfNames) {
	    LinkedHashSet<String> subset = new LinkedHashSet<String>();
	    subset.add(tf);
	    combos.add(subset);
	    fillSubset(subset, tfNames, combos, maxTFCombinations - 1);
	}
	return combos;
    }
		
    protected static void fillSubset(LinkedHashSet<String> subset, TreeSet<String> tfNames, LinkedHashSet<LinkedHashSet<String>> combos, int maxTFCombinations) {
	if (maxTFCombinations>0) {
	    for (String tf : tfNames) {
		if (!subset.contains(tf)) {
		    //Create a new object
		    LinkedHashSet<String> newSubset = new LinkedHashSet<String>(subset);
		    //add the TF
		    newSubset.add(tf);
		    //Check if already exists
		    if (!combos.contains(newSubset)) {
			combos.add(newSubset);
			//continue recursing
			fillSubset(newSubset, tfNames, combos, maxTFCombinations - 1);
		    }
		}
	    }
	}
    }

    public static LinkedHashSet<RefSeqGene> intersect(LinkedHashSet<RefSeqGene> s1, LinkedHashSet<RefSeqGene> s2) {
	LinkedHashSet<RefSeqGene> s = new LinkedHashSet<RefSeqGene>();
	for (RefSeqGene g : s1) {
	    if (s2.contains(g))
		s.add(g);
	}
	return s;
    }

    

}