//use this class to represent all the RefSeq genes in a chromosome.

import java.util.*;
import java.util.regex.*;
import java.io.*;
import umontreal.iro.lecuyer.stat.*;

public class RefSeqUCSCEnsemblGenes {

    protected static final int[] CHROMOSOME_LENGTH = new int[]{197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255, 121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430, 166650296, 15902555};

    protected static final int[] CHROMOSOME_LENGTH_HS = new int[]{247249724, 242951154, 199501832, 191273068, 180857871, 170899997, 158821429, 146274831, 140273257, 135374743, 134452390, 132349540, 114142986, 106368591, 100338921, 88827260, 78774748, 76117159, 63811657, 62435970, 46944329, 49691438, 154913759, 57772959};

    protected static final int[] CHROMOSOME_LENGTH_HS_HG19 = new int[]{249250621, 243199373, 198022430, 191154276,180915260,171115067,159138663,146364022,141213431,135534747,135006516,133851895,115169878,107349540,102531392,90354753 ,81195210 ,78077248 ,59128983 ,63025520 ,48129895 ,51304566 ,155270560,59373566};

    /////////////////////////////////

    protected static final long TOTAL_CHROMOSOME_LENGTH = (long)(265489521.8*10);
    protected static final long TOTAL_CHROMOSOME_LENGTH_HS = (long)(3.080419613e+09);
    protected static final long TOTAL_CHROMOSOME_LENGTH_HS_HG19 = (long)(3.0957e+09);
    

    protected TreeSet<RefSeqUCSCEnsemblGene> genes = new TreeSet<RefSeqUCSCEnsemblGene>();
    protected int chr;
    protected String species = "Mm";
    //Keep track of the genes that have a high enough density for us
    //to be interested in (or some other criterion)
    protected int[] geneIDsKeepForward, geneIDsKeepReverse;
    protected boolean hg19 = false;

    public RefSeqUCSCEnsemblGenes(int c, boolean orchestra) {
	chr = c;
	readData(orchestra);
    }

    public RefSeqUCSCEnsemblGenes(int c, boolean orchestra, String sp) {
	chr = c;
	species = sp;
	boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	readData(orchestra);
    }

    public RefSeqUCSCEnsemblGenes(int c, boolean orchestra, String sp, boolean hg) {
	chr = c;
	species = sp;
	boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	hg19 = hg;
	readData(orchestra);
    }

    //////////////////////////////////////////////////////////////////

    protected void readData(boolean orchestra) {
	String genomeLoc = null;
	boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	genomeLoc = (orchestra ? "/groups/neuroduo/martin/Genomes/" : "/media/My Book/Genomes/") + (hs ? "Hs/v" + (hg19 ? "37" : "36") + "/" : "Mm/v9/");
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "refSeq_ucsf_ensembl_TSSs.tsv"));
	    String line = br.readLine();
	    //name    chr     tss     strand  pas  gene clusterID  type
	    Pattern mrnaPattern = Pattern.compile("^([-|\\w|\\.]+)\\s+(\\d+)\\s+(\\d+)\\s+([+|-])\\s+(\\d+)\\s+[-|\\w|\\.]+\\s+\\d*\\s*([_|\\w]+)");
	    while (line!=null) {
		Matcher m = mrnaPattern.matcher(line);
		if (m.find()) {
		    int c2;
		    c2 = Integer.parseInt(m.group(2));
		    if (chr==c2) {
			boolean strand = m.group(4).equals("+");
			int start = Integer.parseInt(m.group(3));
			int end = Integer.parseInt(m.group(5));
			//int clusterID = Integer.parseInt(m.group(6));
			boolean coding = m.group(6).contains("protein_coding") || m.group(6).equals("NA");
			//check if a version of the gene already
			//exists. If so, we merge the two versions
			//into a super-gene which is contains the
			//union of the two versions
			RefSeqUCSCEnsemblGene g = new RefSeqUCSCEnsemblGene(chr, start, end, strand, m.group(1), m.group(1).hashCode(), !coding);
			if (!genes.contains(g))
			    add(g);
		    }
		}
		//else if (!line.startsWith("#"))
		    //System.out.println("Unable to parse " + line);
		line = br.readLine();
	    }
	    br.close();
	    System.out.println("Found " + genes.size() + " genes.");
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/mrna_list.txt\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    ///////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { setFactorReads(tfName, reads, 0); }

    public void setFactorReads(String tfName, ABIRead[] reads, int strand) {
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand())) {
		int startInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), g.getStrand() ? g.getStart() : g.getEnd()));
		startInd = startInd<0 ? -startInd-1 : startInd;
		int endInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), g.getStrand() ? g.getEnd() : g.getStart()));
		endInd = endInd<0 ? -endInd-1 : endInd;
		ABIRead rs[] = new ABIRead[endInd - startInd];
		for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		g.setFactorReads(tfName, rs);
		//System.out.println(g.toString() + "\t" + rs.length + "\t" + startInd + "\t" + endInd);
	    }
	}
    }

    public void setFactorReads(String tfName, ABIRead[] reads, int windowSize, boolean tss) { setFactorReads(tfName, reads, windowSize, tss, 0); }

    public void setFactorReads(String tfName, ABIRead[] reads, int windowSize, boolean tss, int strand) {
	//Use this functions when we want to set reads found in a
	//window surrounding the TSS.
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand())) {
		int startInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), tss ? g.getStart() - windowSize : g.getEnd() - windowSize));
		startInd = startInd<0 ? -startInd-1 : startInd;
		int endInd = Arrays.binarySearch(reads, new ABIRead(chr, g.getStrand(), tss ? g.getStart() + windowSize : g.getEnd() + windowSize));
		endInd = endInd<0 ? -endInd-1 : endInd;
		ABIRead rs[] = new ABIRead[endInd - startInd];
		for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		g.setFactorReads(tfName, rs);
		//System.out.println(g.toString() + "\t" + rs.length + "\t" + startInd + "\t" + endInd);
	    }
	}
    }

    public int[] getNumberOfFactorReads(String tfName, int strand) { return getNumberOfFactorReads(tfName, strand, false); }

    public int[] getNumberOfFactorReads(String tfName, int strand, boolean codingOnly) {
	int i = 0, nGenes = strand==0 ? size() : size(strand>0);
	int nReads[] = new int[nGenes];
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (strand==0 || (strand>0 && g.getStrand()) || (strand<0 && !g.getStrand()) && (!codingOnly || !g.isNonCoding())) 
		nReads[i++] = g.getFactorReads(tfName).length;
	}
	return nReads;
    }

    public void setGeneIDsKeep(boolean forward, String name, double cutOff) { 
	Vector<Integer> ids = new Vector<Integer>();
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (g.getStrand()==forward && (double)g.getFactorReads(name).length/g.getLength()>cutOff)
		ids.addElement(g.getID());
	}
	Collections.sort(ids);
	if (forward)
	    geneIDsKeepForward = new int[ids.size()];
	else
	    geneIDsKeepReverse = new int[ids.size()];
	int j = 0;
	for (Integer i : ids) {
	    if (forward)
		geneIDsKeepForward[j++] = i;
	    else
		geneIDsKeepReverse[j++] = i;
	}
    }

    public void setGeneIDsKeep(boolean forward, int[] ids) { 
	if (forward)
	    geneIDsKeepForward = ids;
	else
	    geneIDsKeepReverse = ids;
    }

    public int[] getGeneIDsKeep(boolean forward) { return forward ? geneIDsKeepForward : geneIDsKeepReverse; }

    //////////////////////////////////////////////////////////////////

    public int getChromosome() { return chr; }

    public int size() { return genes.size(); }

    public int size(boolean forward) { 
	int n = 0;
	for (RefSeqUCSCEnsemblGene g : genes)
	    n += g.getStrand()==forward ? 1 : 0;
	return n;
    }

    public void add(RefSeqUCSCEnsemblGene m) { genes.add(m); }

    public TreeSet<RefSeqUCSCEnsemblGene> getGenes() { return genes; }

    public TreeSet<RefSeqUCSCEnsemblGene> getGenes(boolean keep) { 
	TreeSet<RefSeqUCSCEnsemblGene> gs = new TreeSet<RefSeqUCSCEnsemblGene>();
	if (keep) {
	    for (RefSeqUCSCEnsemblGene g : getGenes()) {
		if (Arrays.binarySearch(geneIDsKeepForward, g.getID())>=0 || Arrays.binarySearch(geneIDsKeepReverse, g.getID())>=0)
		    gs.add(g);
	    }
	}
	else
	    gs = genes;
	return gs;
    }

    public TreeSet<RefSeqUCSCEnsemblGene> getGenes(boolean keep, boolean forward) { 
	TreeSet<RefSeqUCSCEnsemblGene> gs = new TreeSet<RefSeqUCSCEnsemblGene>();
	int[] gk = forward ? geneIDsKeepForward : geneIDsKeepReverse;
	if (keep) {
	    for (RefSeqUCSCEnsemblGene g : getGenes()) {
		if (Arrays.binarySearch(gk, g.getID())>=0)
		    gs.add(g);
	    }
	}
	else
	    gs = genes;
	return gs;
    }

    public int[] getLengths() {
	int lens[] = new int[size()], i = 0;
	for (RefSeqUCSCEnsemblGene g : genes)
	    lens[i++] = g.getLength();
	return lens;
    }

    public int[] getLengths(boolean forward) {
	int lens[] = new int[size(forward)], i = 0;
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (g.getStrand()==forward)
		lens[i++] = g.getLength();
	}
	return lens;
    }

    public int[] getLengths(boolean forward, boolean codingOnly) {
	int lens[] = new int[size(forward)], i = 0;
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (g.getStrand()==forward && (!codingOnly && !g.isNonCoding()))
		lens[i++] = g.getLength();
	}
	return lens;
    }

    public int[] getLengths(int ids[]) {
	int lens[] = new int[ids.length], i = 0;
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (Arrays.binarySearch(ids, g.getID())>-1)
		lens[i++] = g.getLength();
	}
	return lens;
    }

    public int[] getTSSs() {
	int tsss[] = new int[size()], i = 0;
	for (RefSeqUCSCEnsemblGene g : genes)
	    tsss[i++] = g.getStart();
	return tsss;
    }

    public boolean[] getStrands() {
	boolean strands[] = new boolean[size()];
	int i = 0;
	for (RefSeqUCSCEnsemblGene g : genes) 
	    strands[i++] = g.getStrand();
	return strands;
    }

    public String[] getNames() {
	String names[] = new String[size()];
	int i = 0;
	for (RefSeqUCSCEnsemblGene g : genes)
	    names[i++] = g.getName();
	return names;
    }

    public RefSeqUCSCEnsemblGene getNearestGene(int pos) { return getNearestGene(pos, null); }

    public RefSeqUCSCEnsemblGene getNearestGene(int pos, int ids[]) {
	int nearestDist = Integer.MAX_VALUE;
	RefSeqUCSCEnsemblGene nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (ids==null || Arrays.binarySearch(ids, m.getID())>=0) {
		int dist = m.peakDistance(pos);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		    nearestGene = m;
		}
	    }
	}
	//System.out.println("Found distance " + nearestDist + " and gene " + nearestGene.toString() + " for peak " + pos);
	return nearestGene;
    }

    ///////////////////////////////////////////////////////

    public boolean nearTSS(int pos, int maxDist, boolean strand) {
	//Check if pos is within maxDist of a TSS on strand
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (g.getStrand()==strand) {
		if (pos>g.getStart() - maxDist && pos<g.getStart() + maxDist)
		    return true;
	    }
	}
	return false;
    }

    public boolean nearTSS(int pos, int maxDist) { return nearTSS(pos, maxDist, true) || nearTSS(pos, maxDist, false); }

    public boolean overlapsTSS(int start, int end) { return overlapsTSS(start, end); }

    public RefSeqUCSCEnsemblGene getOverlappingTSS(int start, int end) { 
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (start<m.getStart() && end>m.getStart())
		return m;
	}
	return null;
    }

    public int getNearestGeneDistance(int pos) {
	//Returns the distance to the nearest TSS or TES
	return Math.min(getNearestTSS(pos, true, null), getNearestTES(pos, true)); 
    }

    public int getNearestTSS(int pos) {	return getNearestTSS(pos, false, null); }

    public int getNearestTSS(int pos, int ids[]) {
	int nearestForward = getNearestTSS(pos, true, ids);
	int nearestReverse = getNearestTSS(pos, false, ids);
	return Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos) ? nearestForward : nearestReverse;
    }

    public int getNearestTSS(int pos, boolean strand) { return getNearestTSS(pos, strand, null); }

    public int getNearestTSS(int pos, boolean strand, int ids[]) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (m.getStrand()==strand && (ids==null || Arrays.binarySearch(ids, m.getID())>=0)) {
		int dist = m.getMinTSSDistance(pos);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		}
	    }
	}
	return Math.min(pos + nearestDist, Integer.MAX_VALUE);
    }

    public RefSeqUCSCEnsemblGene getNearestTSSGene(int pos) { return getNearestTSSGene(pos, null); }

    public RefSeqUCSCEnsemblGene getNearestTSSGene(int pos, int ids[]) {
	int nearestForward = getNearestTSS(pos, true, ids);
	int nearestReverse = getNearestTSS(pos, false, ids);
	boolean forward = Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos) || nearestReverse==Integer.MAX_VALUE;
	return getNearestTSSGene(pos, forward, ids);
    }

    public RefSeqUCSCEnsemblGene getNearestTSSGene(int pos, boolean strand) { return getNearestTSSGene(pos, strand, null); }

    public RefSeqUCSCEnsemblGene getNearestTSSGene(int pos, boolean strand, int ids[]) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	RefSeqUCSCEnsemblGene nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (m.getStrand()==strand && (ids==null || Arrays.binarySearch(ids, m.getID())>=0)) {
		int dist = m.getMinTSSDistance(pos);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		    nearestGene = m;
		}
	    }
	}
	return nearestGene;
    }

    /////////////////////////////////

    public boolean nearTES(int pos, int maxDist, boolean strand) {
	//Check if pos is within maxDist of a TSS on strand
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (g.getStrand()==strand) {
		if (pos>g.getEnd() - maxDist && pos<g.getEnd() + maxDist)
		    return true;
	    }
	}
	return false;
    }

    public boolean nearTES(int pos, int maxDist) { return nearTES(pos, maxDist, true) || nearTES(pos, maxDist, false); }

    public int getNearestTES(int pos) {
	int nearestForward = getNearestTES(pos, true);
	int nearestReverse = getNearestTES(pos, false);
	return Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos) ? nearestForward : nearestReverse;
    }

    public int getNearestTES(int pos, boolean strand) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (m.getStrand()==strand) {
		int dist = m.getMinTESDistance(pos);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		}
	    }
	}
	return pos + nearestDist;
    }

    public RefSeqUCSCEnsemblGene getNearestTESGene(int pos) {
	int nearestForward = getNearestTES(pos, true);
	int nearestReverse = getNearestTES(pos, false);
	return getNearestTESGene(pos, Math.abs(nearestForward - pos)<Math.abs(nearestReverse - pos));
    }

    public RefSeqUCSCEnsemblGene getNearestTESGene(int pos, boolean strand) {
	//Returns the position of the nearest TSS to pos
	int nearestDist = Integer.MAX_VALUE;
	RefSeqUCSCEnsemblGene nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (m.getStrand()==strand) {
		int dist = m.getMinTESDistance(pos);
		if (Math.abs(dist)<Math.abs(nearestDist)) {
		    nearestDist = dist;
		    nearestGene = m;
		}
	    }
	}
	return nearestGene;
    }

    /////////////////////////////////

    public RefSeqUCSCEnsemblGene getGene(String name) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    String n = m.name;
	    if (n.toLowerCase().equals(name.toLowerCase()))
		return m;
	}
	return null;
    }

    public int[] getGeneIDs() {
	int ids[] = new int[genes.size()], i = 0;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    ids[i++] = m.getID();
	}
	return ids;
    }

    public RefSeqUCSCEnsemblGene getGene(int id) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene m = (RefSeqUCSCEnsemblGene)it.next();
	    if (m.id==id)
		return m;
	}
	return null;
    }

    public RefSeqUCSCEnsemblGene getRandomGene(Random random) {
	//Use this when we are creating a null distribution where the
	//peak distances are constrained to correspond to the observed
	//ones.
	int r = random.nextInt(size()), i = 0;
	Iterator it = genes.iterator();
	RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)it.next();
	while (it.hasNext() && i++<r) { g = (RefSeqUCSCEnsemblGene)it.next(); }
	return g;
    }

    public int[] calculateNearestTSSDistances(int peakCenters[]) {
	int dists[] = new int[peakCenters.length];
	Vector tss = new Vector();
	Iterator it = genes.iterator();
	while (it.hasNext()) { tss.add(new Integer(((RefSeqUCSCEnsemblGene)it.next()).getStart())); }
	int tssLoc[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++) { tssLoc[i] = ((Integer)tss.elementAt(i)).intValue(); }
	for (int i=0; i<peakCenters.length; i++) {
	    dists[i] = Integer.MAX_VALUE;
	    for (int j=0; j<tssLoc.length; j++) 
		dists[i] = Math.min(Math.abs(peakCenters[i] - tssLoc[j]), dists[i]);
	}
	return dists;
    }

    public int[] getStarts() {
	Vector tss = new Vector();
	Iterator it = genes.iterator();
	while (it.hasNext()) { tss.addAll(new Vector(((RefSeqGene)it.next()).getStarts())); }
	int tssLoc[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++) { tssLoc[i] = ((Integer)tss.elementAt(i)).intValue(); }
	return tssLoc;
    }

    public int[] getStarts(boolean unique) {
	if (!unique) { return getStarts(); }
	//Only get the unique TSSs
	int starts[] = getStarts();
	Vector tss = new Vector();
	tss.addElement(new Integer(starts[0]));
	for (int i=1; i<starts.length; i++) {
	    if (starts[i]!=starts[i-1])
		tss.addElement(new Integer(starts[i]));
	}
	int u[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++)
	    u[i] = ((Integer)tss.elementAt(i)).intValue();
	return u;
    }

    public int[] getStarts(int[] geneIDsKeep) { return getStarts(true, geneIDsKeep); }

    public int[] getStarts(boolean unique, int[] geneIDsKeep) {
	if (geneIDsKeep==null) { return getStarts(unique); }
	//Only look at the genes which are in the list
	Vector tss = new Vector();
	Iterator it = genes.iterator();
	while (it.hasNext()) { 
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
		if (!unique)
		    tss.addAll(new Vector(g.getStarts())); 
		else {
		    tss.addElement(g.getStarts().elementAt(0));
		    for (int i=1; i<g.getStarts().size(); i++) {
			if (g.getStarts().elementAt(i-1)!=g.getStarts().elementAt(i))
			    tss.addElement(g.getStarts().elementAt(i));
		    }
		}
	    }
	}
	int u[] = new int[tss.size()];
	for (int i=0; i<tss.size(); i++)
	    u[i] = ((Integer)tss.elementAt(i)).intValue();
	return u;
    }

    public int[] getIDs() {
	int ids[] = new int[size()], i = 0;
	Iterator it = genes.iterator();
	while (it.hasNext()) { 
	    RefSeqGene g = (RefSeqGene)it.next();
	    ids[i++] = g.getID();
	}
	return ids;
    }

    public int[] getIDs(boolean internal) {
	Vector<Integer> idVec = new Vector<Integer>();
	Iterator it = genes.iterator();
	while (it.hasNext()) { 
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (g.getAlternativeStarts(true).length>0)
		idVec.addElement(g.getID());
	}
	int ids[] = new int[idVec.size()], i = 0;
	for (Integer id : idVec)
	    ids[i++] = id;
	return ids;
    }

    public int[] numberOfPeaksWithinDistance(int peakCenters[], int cutOff) {
	//Find out how many TSSs have a given number of peaks within
	//cutOff. Returns a vector where each value represents the
	//number of peaks within cutOff.
	int tssLoc[] = getStarts();
	int nPeaks[] = new int[tssLoc.length];
	for (int i=0; i<tssLoc.length; i++) {
	    nPeaks[i] = 0;
	    for (int j=0; j<peakCenters.length; j++) {
		if (Math.abs(peakCenters[j] - tssLoc[i])<cutOff)
		    nPeaks[i]++;
	    }
	}
	return nPeaks;
    }

    public int[] calculateNearestTSSDistances(boolean notSameGene) {
	//Use this method to calculate the distance to the nearest TSS
	//for all annotated TSSs.
	int dists[] = null;
	if (!notSameGene) {
	    Vector tss = new Vector();
	    Iterator it = genes.iterator();
	    while (it.hasNext()) { tss.addAll(new Vector(((RefSeqGene)it.next()).getStarts())); }
	    dists = new int[tss.size() - 1];
	    int tssLoc[] = new int[tss.size()];
	    for (int i=0; i<tss.size(); i++) { tssLoc[i] = ((Integer)tss.elementAt(i)).intValue(); }
	    dists[0] = Math.abs(tssLoc[1] - tssLoc[0]);
	    for (int i=2; i<tssLoc.length; i++) {
		int prevDist = Math.abs(tssLoc[i-1] - tssLoc[i-2]);
		int nextDist = Math.abs(tssLoc[i] - tssLoc[i-1]);
		prevDist = prevDist==0 ? Integer.MAX_VALUE : prevDist;
		nextDist = nextDist==0 ? Integer.MAX_VALUE : nextDist;
		dists[i-1] = Math.min(nextDist, prevDist);
	    }
	}
	else {
	    System.err.println("Not implemented yet!");
	}
	return dists;
    }

    public int[] interGeneDistances(boolean sameStrand) { return interGeneDistances(sameStrand, null); }

    public int[] interGeneDistances(boolean sameStrand, int ids[]) {
	//Calculate the distances between all pairs of genes. If ids[]
	//is not null, then only include genes that are contained in
	//that subset.
	Vector<Integer> dists = new Vector<Integer>();
	if (sameStrand) {
	    int prevEndForward = -1, prevStartReverse = -1;
	    for (RefSeqUCSCEnsemblGene g : genes) {
		if (g.getStrand() && (ids==null || Arrays.binarySearch(ids, g.getID())>-1)) {
		    if (prevEndForward>0 && g.getStart()>prevEndForward) {
			dists.addElement(g.getStart() - prevEndForward);
			prevEndForward = g.getEnd();
		    }
		    else
			prevEndForward = Math.max(prevEndForward, g.getEnd());
		}
		else if (ids==null || Arrays.binarySearch(ids, g.getID())>-1){
		    if (prevStartReverse>0 && g.getEnd()>prevStartReverse) {
			dists.addElement(g.getEnd() - prevStartReverse);
			prevStartReverse = g.getStart();
		    }
		    else
			prevStartReverse = Math.max(prevStartReverse, g.getStart());
		}
	    }
	}
	else {
	    int prevStartForward = -1, prevStartReverse = -1;
	    for (RefSeqUCSCEnsemblGene g : genes) {
		if (ids==null || Arrays.binarySearch(ids, g.getID())>-1)
		    dists.addElement(Math.abs(g.getStart() - getNearestTSS(g.getStart(), !g.getStrand())));
	    }
	}
	int ds[] = new int[dists.size()], i = 0;
	for (Integer d : dists)
	    ds[i++] = d;
	return ds;
    }

    public boolean overlapsGene(int start) { return overlapsGene(start, start, true) || overlapsGene(start, start, false); }

    public boolean overlapsGene(int start, int end) { return overlapsGene(start, end, true) || overlapsGene(start, end, false); }

    public boolean overlapsGene(int start, int end, boolean forward) {
	//Returns true if any part of the interval overlaps a coding region
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)it.next();
	    if (g.getStrand()==forward && g.overlaps(start, end))
		return true;
	}
	return false;
    }

    public RefSeqUCSCEnsemblGene getOverlappingGene(int center) { return getOverlappingGene(center, center); }

    public RefSeqUCSCEnsemblGene getOverlappingGene(int start, int end) { 
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (g.overlaps(start, end))
		return g;
	}
	return null;
    }

    public RefSeqUCSCEnsemblGene getOverlappingGene(int start, int end, boolean forward) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)it.next();
	    if (g.getStrand()==forward && g.overlaps(start, end))
		return g;
	}
	return null;
    }

    ///////////////////////////////////////////////////////

    public TreeSet<RefSeqUCSCEnsemblGene> getNonOverlappingGenes() { 
	TreeSet<RefSeqUCSCEnsemblGene> a = getNonOverlappingGenes(true, false);
	a.addAll(getNonOverlappingGenes(false, false)); 
	return a;
    } 

    public TreeSet<RefSeqUCSCEnsemblGene> getNonOverlappingGenes(boolean forward) { return getNonOverlappingGenes(forward, false); } 

    public TreeSet<RefSeqUCSCEnsemblGene> getNonOverlappingGenes(boolean forward, boolean codingOnly) {
	//Use this method to extract a set of genes that are
	//non-overlapping. This is a way of filtering out alternative
	//isoforms and avoid double-counting. In each cluster of
	//overlapping genes, we choose the one which is longest.
	TreeSet<RefSeqUCSCEnsemblGene> nonOverlappingGenes = new TreeSet<RefSeqUCSCEnsemblGene>(), geneSet = new TreeSet<RefSeqUCSCEnsemblGene>();
	RefSeqUCSCEnsemblGene prevGene = null;
	for (RefSeqUCSCEnsemblGene g : genes) {
	    //Genes are sorted, so we can simply loop through
	    if (g.getStrand()==forward && (!codingOnly || !g.isNonCoding())) {
		if (prevGene!=null) {
		    //Check if overlapping with previous genes
		    if (g.overlaps(geneSet))
			geneSet.add(g);
		    else {
			//Find the longest one in the previous cluster
			RefSeqUCSCEnsemblGene longestGene = geneSet.first();
			for (RefSeqUCSCEnsemblGene g2 : geneSet) 
			    longestGene = g2.getLength()>longestGene.getLength() ? g2 : longestGene;
			nonOverlappingGenes.add(longestGene);
			geneSet = new TreeSet<RefSeqUCSCEnsemblGene>();
			geneSet.add(g);
		    }
		}
		else
		    geneSet.add(g);
		prevGene = g;
	    }
	}
	//Check the last set
	RefSeqUCSCEnsemblGene longestGene = geneSet.first();
	for (RefSeqUCSCEnsemblGene g2 : geneSet) 
	    longestGene = g2.getLength()>longestGene.getLength() ? g2 : longestGene;
	nonOverlappingGenes.add(longestGene);
	return nonOverlappingGenes;
    }

    public TreeSet<RefSeqUCSCEnsemblGene> getOverlappingGenes(RefSeqUCSCEnsemblGene gene) { return getOverlappingGenes(gene, true); }

    public TreeSet<RefSeqUCSCEnsemblGene> getOverlappingGenes(RefSeqUCSCEnsemblGene gene, boolean codingOnly) {
	//Get all genes that overlap g which we assume correspond to different isoforms
	TreeSet<RefSeqUCSCEnsemblGene> overlapping = new TreeSet<RefSeqUCSCEnsemblGene>();
	overlapping.add(gene);
	for (RefSeqUCSCEnsemblGene g : genes) {
	    //Genes are sorted, so we can simply loop through
	    if (g.getStrand()==gene.getStrand() && (!codingOnly==g.isNonCoding())) {
		//Check if overlapping with previous genes
		if (g.overlaps(gene))
		    overlapping.add(g);
	    }
	}
	return overlapping;
    }

    //////////////////////////////////////////////////////////////////

    public void createRandomRegions(int[] lociToMatch, String saveFileName, int mappabilityWindowSize, double minMappability, boolean intronic, ConservationIslands conservationIslands, boolean conserved, boolean orchestra) {
	//Use this method to create a set of randomized regions
	//matching the lociToMatch vector. The random regions should
	//have the same nearest TSS distance as the lociToMatch, the
	//same gene overlap category (intragenic/extragenic) and have
	//a reasonable mappability.
	TreeMap<Integer, Integer> rand2realLoci = new TreeMap<Integer, Integer>();
	Random random = new Random(1003);
	while (rand2realLoci.size()<lociToMatch.length) {
	    for (int i=0; i<lociToMatch.length; i++) {
		if (!rand2realLoci.containsValue(lociToMatch[i])) {
		    RefSeqUCSCEnsemblGene g = getOverlappingGene(lociToMatch[i]);
		    if (g==null)
			g = getNearestTSSGene(lociToMatch[i]);
		    int tssDist = Math.abs(g.getStart() - lociToMatch[i]), randMap = 0, randLoc = -1;
		    //Find another gene
		    RefSeqUCSCEnsemblGene randGene = null;
		    if (!intronic) {
			//Put at the same distance and make sure that it is not intronic
			do {
			    randGene = getRandomGene(random);
			    randLoc = randGene.getStrand() ? randGene.getStart() - tssDist : randGene.getStart() + tssDist;
			}
			while (!validExtragenicLoc(randLoc, conserved, conservationIslands)); 
		    }
		    else {
			do {
			    randGene = getRandomGene(random);
			    randLoc = randGene.getStrand() ? randGene.getStart() + tssDist : randGene.getStart() - tssDist;
			}
			while (!validIntragenicLoc(randLoc, conserved, conservationIslands)); 
		    }
		    rand2realLoci.put(randLoc, lociToMatch[i]);
		}
	    }
	    //Also make sure that the mappability is large enough
	    System.out.println("Found sites, checking mappability....");
	    int fragLen = 35, prevLoc = 0;
	    Mappability mappability = new Mappability(orchestra, chr, mappabilityWindowSize + 2*fragLen, fragLen, species);
	    mappability.initialize();
	    TreeSet<Integer> removeKeys = new TreeSet<Integer>();
	    for (Integer randLoc : rand2realLoci.keySet()) {
		int randMap = mappability.numberOfPotentialReadsNextWindow((randLoc - prevLoc) - mappabilityWindowSize/2 - fragLen);
		//System.out.println("Found mappability " + randMap + " for " + randLoc);
		if (((double)randMap)/mappabilityWindowSize<minMappability) 
		    removeKeys.add(randLoc);
		prevLoc = randLoc;
	    }
	    for (Integer r : removeKeys)
		rand2realLoci.remove(r);
	    System.out.println("Found " + rand2realLoci.size() + " of " + lociToMatch.length);
	}
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    for (Integer r : rand2realLoci.keySet()) {
		RefSeqUCSCEnsemblGene g = getOverlappingGene(r);
		if (g!=null)
		    System.out.println("Overlapping gene is " + g.toString());
		bw.write(chr + "\t" + r.toString() + "\t" + (g==null ? "NA" : (g.getStrand() ? "TRUE" : "FALSE")) + "\t" + (conservationIslands.overlaps(chr, r) ? "TRUE" : "FALSE") + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Failed to save random region: " + e.toString());
	}
    }

    public boolean validExtragenicLoc(int randLoc, boolean conserved, ConservationIslands conservationIslands) {
	if (overlapsGene(randLoc) || randLoc==-1)
	    return false;
	return conservationIslands==null || (conserved && conservationIslands.overlaps(chr, randLoc)) || (!conserved && !conservationIslands.overlaps(chr, randLoc));
    }

    public boolean validIntragenicLoc(int randLoc, boolean conserved, ConservationIslands conservationIslands) {
	if (!overlapsGene(randLoc) || randLoc==-1)
	    return false;
	return conservationIslands==null || (conserved && conservationIslands.overlaps(chr, randLoc)) || (!conserved && !conservationIslands.overlaps(chr, randLoc));
    }

    public int getRandomRegion(int downstream, int upstream, boolean nongenic, Random random) {
	if (nongenic) {
	    //Do rejection sampling until we find a region which does
	    //not overlap any exonic or intronic region.
	    boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	    //System.out.println("Species is " + species + " and hs is " + hs);
	    int len = hs ? PeakFinder2.CHROMOSOME_LENGTH_HS[chr - 1] : PeakFinder2.CHROMOSOME_LENGTH[chr - 1];
	    int randLoc = upstream + random.nextInt(len - upstream - downstream - 1);
	    while (overlapsGene(randLoc - upstream, randLoc + downstream))
		randLoc = random.nextInt(len);
	    return randLoc;
	}
	else {
	    //Get a random gene and pull out a random part of it
	    RefSeqUCSCEnsemblGene randGene = null;
	    int len = -1, geneStart = 0, geneEnd = 0;
	    while (len<=0) {
		randGene = getRandomGene(random);
		geneStart = randGene.getStart();
		geneEnd = randGene.getEnd();
		len = randGene.getLength() - downstream - upstream - 1;
	    }
	    return randGene.getStrand() ? geneStart + upstream + random.nextInt(len) : geneStart - upstream - random.nextInt(len);
	}
    }

    public int getRandomRegion(int downstream, int upstream, Random random) {
	//Do rejection sampling until we find a region which does
	//not overlap any TSS
	boolean hs = species!=null && (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"));
	//System.out.println("Species is " + species + " and hs is " + hs);
	int len = hs ? PeakFinder2.CHROMOSOME_LENGTH_HS[chr - 1] : PeakFinder2.CHROMOSOME_LENGTH[chr - 1];
	int randLoc = upstream + random.nextInt(len - upstream - downstream - 1);
	while (overlapsTSS(randLoc - upstream, randLoc + downstream))
	  randLoc = random.nextInt(len);
	return randLoc;
    }

    //////////////////////////////////////////////////////////////////

    public static int getChromosomeLength(int chr, String species) { return getChromosomeLength(chr, species, false); }

    public static int getChromosomeLength(int chr, String species, boolean hg19) {
	if (species.toLowerCase().equals("mm"))
	    return CHROMOSOME_LENGTH[chr-1];
	else
	    return hg19 ? CHROMOSOME_LENGTH_HS_HG19[chr-1] : CHROMOSOME_LENGTH_HS[chr-1];
    }
}
