//Use this method to calculate the kernel for the MeDiChi
//algorithm. The Matlab version is ridiculously slow and this will
//hopefully be faster.

import java.util.*;

public class SidowMeDiChi {

    protected int readLength, lambda;
    protected int fragmentMeanLength, fragmentStdDev;
    protected TreeMap kernels = new TreeMap();
    protected double maxLength = -1;

    public SidowMeDiChi(){
	readLength = 25;
	lambda = 10;
    }

    public SidowMeDiChi(int lam){
	readLength = 25;
	lambda = lam;
    }


    //////////////////////////////////////////////////////////////////
    
    public void setFragmentDistribuitionParameters(double a, double b) {
	fragmentMeanLength = a;
	fragmentStdDev = b;
	maxLength = fragmentMeanLength + 2*fragmentStdDev;
	//Try to load a saved kernel which will save a lot of time.
	kernels = new TreeMap();
    }

    public double[][] calculateKernel(int startPos, int endPos){
	int len = endPos-startPos;
	double[][] kernel = new double[len][len];
	for (int j=0; j<len; j++) {
	    for (int i=0; i<len; i++)
		kernel[j][i] = fragSum(j-i);
	}
	return kernel;
    }

    public double[] calculateKernel(int startPos, int endPos, double a, double b){
	setFragmentDistribuitionParameters(a, b);
	return calculateKernel(startPos, endPos);
    }

    public double fragSum(int dx){
	if (kernelCalculated(dx)) //Check if already computed
	    return getKernel(dx);
	if (Math.abs(dx)>maxLength)
	    return 0;
	double k = 0;
	int len = 1;
	while (len<maxLength)
	    k += a*conditionalBindingProbability(dx, len++);
	//Store the result
	storeKernel(k, dx);
	//System.out.println("Calculated kernel " + k + " for dx = " + dx);
	return k;
    }

    public double conditionalBindingProbability(int dx, int len){
	//This function implements Eq (1) from the Reiss et al paper.
	double p = 0;
	if (len>2*lambda) {
	    for (int i=lambda-len; i<-lambda; i++) {
		p += phi(i, dx, len)/(double)(len - 2*lambda);
	    }
	}
	//System.out.println(p + " for dx = " + dx + " and len = " + len);
	return p;
    }
    
    public double phi(int i, int dx, int len){
	//Implementation of the hybridization probability function
	if (i>dx+probeLength || len+i<dx-probeLength) //no overlap possible
	    return 0;
	int h = 0;
	for (int j=i; j<(len + i); j++) {
	    if (j>(dx-probeLength) && j<(dx+probeLength))
		h++;
	}
	if (h==0)
	    return 0;
	else 
	    return 1/(1 + Math.exp(-(h - NU_0)*NU_S));
    }

    public boolean kernelCalculated(int dx){
	return kernels.containsKey(new Integer(dx));
    }

    public double getKernel(int dx){
	return ((Double)kernels.get(new Integer(dx))).doubleValue();
    }

    public void storeKernel(double k, int dx){
	kernels.put(new Integer(dx), new Double(k));
    }

    public boolean kernelSaturated(int dx) {
	//Use this method to find out if the kernel has saturated so
	//that we can save time by not calculating any more
	//values. The condition that we check for is that N_SATURATION
	//probes in a stretch all have the same value.
	int i = 0;
	if (dx<0) {
	    TreeMap map = new TreeMap(kernels.tailMap(new Integer(dx+1)));
	    if (map.size()>N_SATURATION) {
		//Check if the N_SATURATION keys following this one are identical
		Iterator keyIt = map.keySet().iterator();
		double prevKernel = ((Double)map.get(keyIt.next())).doubleValue();
		while (i<N_SATURATION) {
		    double kernel = ((Double)map.get(keyIt.next())).doubleValue();
		    if (kernel!=prevKernel)
			return false;
		    prevKernel = kernel;
		    i++;
		}
		//lowerKernelSaturation = prevKernel;
		return true;
	    }
	    else
		return false;
	}
	else {
	    //This case is more tricky since we must check the
	    //N_SATURATION binding sites preceeding the one where
	    //we're at.
	    TreeMap map = new TreeMap(kernels.headMap(new Integer(dx)));
	    if (map.size()>N_SATURATION) {
		int prevKey = ((Integer)map.lastKey()).intValue();
		double prevKernel = ((Double)map.get(new Integer(prevKey))).doubleValue();
		while (i<N_SATURATION) {
		    map = new TreeMap(map.headMap(new Integer(prevKey)));
		    prevKey = ((Integer)map.lastKey()).intValue();
		    double kernel = ((Double)map.get(new Integer(prevKey))).doubleValue();
		    if (kernel!=prevKernel)
			return false;
		    prevKernel = kernel;
		    i++;
		}
		//upperKernelSaturation = prevKernel;
		return true;
	    }
	    else
		return false;
	}
    }
    

}
