//This method extracts the information from the chip-seq data
//downloaded from the Sidow lab-page. It is meant to be using as
//little memory as possible. Thus, we do multiple passes over the data
//and write the intermediate results to files.

import java.util.*;
import java.util.regex.*;
import java.io.*;
//import Synergism.FastaSequence;

public class SidowParser2 {

    //Store the reverse strand by using a negative sign for the positions.
    protected int nReadsForward = 0, nReadsReverse = 0, nUniqueReadsForward = 0, nUniqueRandomReadsForward = 0, nUniqueReadsReverse = 0, nUniqueRandomReadsReverse = 0, nUniqueReads = 0, nUniqueRandomReads = 0, chr, maxPos;
    //Store the start and end points for each contig. This is useful
    //when we want to retrieve the full mapping
    protected Random random = new Random();
    protected int readLength = 25;
    protected String originalFileName, randPosReadsFileNameForward, posReadsFileNameForward, randPosReadsFileNameReverse, posReadsFileNameReverse, randPosReadsFileName, posReadsFileName;
    protected Pattern posReadPattern = Pattern.compile("([0-9]+)\t([0-9]+)");

    //////////////////////////////////////////////////////////////////
    //hard-coded information about the length of each chromosome to
    //use when generating random reads. The info is based on v36
    //(hg18) of the human genome, but it will almost certainly be ok
    //for other versions as well. In a more refined version, this info
    //should be in an external file and we should also take into
    //consideration that not the entire genome is mappable using 25 bp
    //reads, which makes the effective length shorter (around 90%
    //according to Robertson et al).
    protected int chrLengths[] = new int[]{1, 2, 3};

    //////////////////////////////////////////////////////////////////

    public SidowParser2(String fileName, int chr) {
	this.chr = chr;
	originalFileName = fileName;
	randPosReadsFileNameForward = originalFileName + "_rand_F" + (new Integer(chr)).toString();
	posReadsFileNameForward = originalFileName + "_F" + (new Integer(chr)).toString();
	randPosReadsFileNameReverse = originalFileName + "_rand_R" + (new Integer(chr)).toString();
	posReadsFileNameReverse = originalFileName + "_R" + (new Integer(chr)).toString();
	randPosReadsFileName = originalFileName + "_rand_" + (new Integer(chr)).toString();
	posReadsFileName = originalFileName + "_" + (new Integer(chr)).toString();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern sidowPattern = Pattern.compile("([A|C|G|T|\\.]) [0-9] [0-9] chr([0-9|X|Y]+):([0-9]+) ([F|R])");
	    //Go over once just to count the number of reads for this
	    //chromosome
	    while (line!=null) {
		Matcher m = sidowPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(2).equals("X"))
			c = 23;
		    else if (m.group(2).equals("Y"))
			c = 24;
		    else
			c = Integer.parseInt(m.group(2));
		    if (c==chr) {
			if (m.group(4).equals("F"))
			    nReadsForward++;
			else
			    nReadsReverse++;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    //Go through again and store them. We want to use an
	    //ordinary array to save memory, going through the file is
	    //not that slow.
	    int readsForward[] = new int[nReadsForward], readsReverse[] = new int[nReadsReverse], reads[] = new int[nReadsForward + nReadsReverse], i = 0, j = 0, k = 0;
	    br = new BufferedReader(new FileReader(fileName));
	    line = br.readLine();
	    while (line!=null) {
		Matcher m = sidowPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(2).equals("X"))
			c = 23;
		    else if (m.group(2).equals("Y"))
			c = 24;
		    else
			c = Integer.parseInt(m.group(2));
		    if (c==chr) {
			if (m.group(4).equals("F"))
			    readsForward[i++] = Integer.parseInt(m.group(3));
			else
			    readsReverse[j++] = Integer.parseInt(m.group(3));
			reads[k++] = Integer.parseInt(m.group(3));
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    //Write all the positions for the reads in an ordered list
	    nUniqueReadsForward = savePosReadsInformation(readsForward, true, false);
	    nUniqueReadsReverse = savePosReadsInformation(readsReverse, false, false);
	    nUniqueReads = savePosReadsInformation(reads, false);
	    maxPos = Math.max(readsForward[readsForward.length-1], readsReverse[readsReverse.length-1]) + readLength;	    
	}
	catch (IOException e) {
	    
	}
    }

    protected int savePosReadsInformation(int[] reads, boolean forward, boolean random) {
	int nUniqueReads = 0;
	try {
	    Arrays.sort(reads);
	    String tmpFileName = "tmpFile", posReadsFileName;
	    BufferedWriter bw = new BufferedWriter(new FileWriter(tmpFileName));
	    for (int j=0; j<reads.length; j++)
		bw.write(reads[j] + "\n");
	    bw.close();
	    if (forward && random) 
		posReadsFileName = randPosReadsFileNameForward;
	    else if (forward && !random) 
		posReadsFileName = posReadsFileNameForward;
	    else if (!forward && random)
		posReadsFileName = randPosReadsFileNameReverse;
	    else
		posReadsFileName = posReadsFileNameReverse;
	    //Now we have generated a file where the reads are
	    //sorted. Next we create a file which shows the mapping
	    //between the position and the number of reads.
	    nUniqueReads = savePositionsReads(tmpFileName, posReadsFileName);
	    //Remove the temporary file
	    (new File(tmpFileName)).delete();
	}
	catch (IOException e) {

	}
	return nUniqueReads;
    }

    protected int savePosReadsInformation(int[] reads, boolean random) {
	int nUniqueReads = 0;
	try {
	    Arrays.sort(reads);
	    String tmpFileName = "tmpFile", posReadsFileName;
	    BufferedWriter bw = new BufferedWriter(new FileWriter(tmpFileName));
	    for (int j=0; j<reads.length; j++)
		bw.write(reads[j] + "\n");
	    bw.close();
	    if (random) 
		posReadsFileName = randPosReadsFileName;
	    else
		posReadsFileName = this.posReadsFileName;
	    //Now we have generated a file where the reads are
	    //sorted. Next we create a file which shows the mapping
	    //between the position and the number of reads.
	    nUniqueReads = savePositionsReads(tmpFileName, posReadsFileName);
	    //Remove the temporary file
	    (new File(tmpFileName)).delete();
	}
	catch (IOException e) {

	}
	return nUniqueReads;
    }
	    
    protected int savePositionsReads(String tmpFileName, String posReadsFileName) {
	//Returns the number of unique positions with at least one read
	int nUnique = 0;
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(posReadsFileName));
	    BufferedReader br = new BufferedReader(new FileReader(tmpFileName));
	    String line = br.readLine();
	    //Use a sliding window to count the number of reads for
	    //each position
	    LinkedList slidingWindowReads = new LinkedList();
	    //Initialize
	    for (int i=0; i<readLength; i++) 
		slidingWindowReads.addFirst(new Integer(0));
	    int slidingWindowStart = Integer.parseInt(line), readStart = 0;
	    //Write to a new file where the first column is the
	    //position, the second column is the number of reads
	    while (line!=null) {
		readStart = Integer.parseInt(line);
		if (readStart==slidingWindowStart)  //Same pos as before
		    slidingWindowReads = addRead(slidingWindowReads);
		else if (readStart>slidingWindowStart) { //Move window forward
		    while (readStart>slidingWindowStart) {
			//Remove the first entry of the window and save it
			if (((Integer)slidingWindowReads.getFirst()).intValue()!=0) {
			    bw.write((slidingWindowStart) + "\t" + slidingWindowReads.removeFirst().toString() + "\n");
			    nUnique++;
			}
			else 
			    slidingWindowReads.removeFirst();
			slidingWindowStart++;
			//Add an empty element at the end to keep the size
			slidingWindowReads.addLast(new Integer(0));
		    }
		    //slidingWindowReads = addRead(slidingWindowReads);
		}
		else
		    System.err.println("Error! found read starting earlier than window at " + line);
		line = br.readLine();
	    }
	    while (!slidingWindowReads.isEmpty()) { //Write the last window
		if (((Integer)slidingWindowReads.getFirst()).intValue()!=0) {
		    bw.write((slidingWindowStart++) + "\t" + slidingWindowReads.removeFirst().toString() + "\n");
		    nUnique++;
		}
		else
		    slidingWindowReads.removeFirst();
	    }
	    br.close();
	    bw.close();
	}
	catch(IOException e) {

	}
	return nUnique;
    }

    protected LinkedList addRead(LinkedList l) {
	for (int i=0; i<readLength; i++)
	    l.set(i, new Integer(((Integer)l.get(i)).intValue() + 1));
	return l;
    }

    public int[] getReadsDistribution(){
	return getReadsDistribution(false);
    }

    public int[] getRandomReadsDistribution(){
	return getReadsDistribution(true);
    }

    public int[] getReadsDistribution(boolean random){
	//Returns a histogram showing the number of positions with a
	//given number of reads.
	TreeMap readsHistogram = new TreeMap();
	int readsForward[][] = random ? getRandomReads(true) : getReads(true);
	int readsReverse[][] = random ? getRandomReads(false) : getReads(false);
	for (int i=0; i<readsForward.length; i++) {
	    Integer key = new Integer(readsForward[i][1]);
	    if (readsHistogram.containsKey(key))
		readsHistogram.put(key, new Integer(((Integer)readsHistogram.get(key)).intValue() + 1));
	    else
		readsHistogram.put(key, 1);
	}
	for (int i=0; i<readsReverse.length; i++) {
	    Integer key = new Integer(readsReverse[i][1]);
	    if (readsHistogram.containsKey(key))
		readsHistogram.put(key, new Integer(((Integer)readsHistogram.get(key)).intValue() + 1));
	    else
		readsHistogram.put(key, 1);
	}
	int r[] = new int[((Integer)readsHistogram.lastKey()).intValue()];
	Iterator it = readsHistogram.keySet().iterator();
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    r[key.intValue()-1] = ((Integer)readsHistogram.get(key)).intValue();
	}
	return r;
    }

    public int calculateFDRHeight(double ratio) {
	//Returns the minimum height for a peak to be considered
	//significant given the randomized background model.
	int[] readsDist = getReadsDistribution(), randReadsDist = getRandomReadsDistribution();
	int minLen = readsDist.length<randReadsDist.length ? readsDist.length : randReadsDist.length;
	double r[] = new double[minLen];
	for (int i=0; i<minLen; i++)
	    //Calculate the ratio of expected/observed
	    r[i] = (double)randReadsDist[i]/(double)readsDist[i];
	//Find the first index of r which is lower than ratio
	for (int i=0; i<minLen; i++) {
	    if (r[i]<ratio)
		return i + 1;
	}
	return minLen + 1;
    }

    public int getNumberOfReads(){
	return nReadsForward + nReadsReverse;
    }

    public int[][] getReads(){
	//Both strands together
	return getReads(posReadsFileName, nUniqueReads);
    }

    public int[][] getReads(boolean forward){
	if (forward)
	    return getReads(posReadsFileNameForward, nUniqueReadsForward);
	return getReads(posReadsFileNameReverse, nUniqueReadsReverse);
    }

    protected int[][] getReads(String processedFileName, int nUniqueReads) {
	//Return the number of reads in a compact array. The first
	//column is the position and the second is the number of reads
	//for that position from the forward strand and the third is
	//the number of reads from the reverse strand.
	int i = 0, r[][] = new int[nUniqueReads][2];
	try {
	    BufferedReader br = new BufferedReader(new FileReader(processedFileName));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = posReadPattern.matcher(line);
		if (m.find()) {
		    r[i][0] = Integer.parseInt(m.group(1));
		    r[i++][1] = Integer.parseInt(m.group(2));
		}
		else
		    System.err.println("Error, could not parse line: " + line);
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	return r;
    }

    public void generateRandomReads() {
	//Spread out the same number of reads as in the real
	//experiment randomly.
	String randTmpFileName = "randTmp";
	int chrLength = maxPos, randStartForward[] = new int[nReadsForward], randStartReverse[] = new int[nReadsReverse], randStart[] = new int[nReadsForward + nReadsReverse];
	for (int i=0; i<nReadsForward; i++) {
	    //Get a randomized starting position
	    int r = random.nextInt(chrLength-readLength);
	    randStartForward[i] = r;
	    randStart[i] = r;
	}
	for (int i=0; i<nReadsReverse; i++) {
	    //Get a randomized starting position
	    int r = random.nextInt(chrLength-readLength);
	    randStartReverse[i] = r;
	    randStart[nReadsForward + i] = r;
	}
	nUniqueRandomReadsForward = savePosReadsInformation(randStartForward, true, true);
	nUniqueRandomReadsReverse = savePosReadsInformation(randStartReverse, false, true);
	nUniqueRandomReads = savePosReadsInformation(randStart, true);
    }

    public int[][] getRandomReads(boolean forward) {
	if (randPosReadsFileNameForward!=null && randPosReadsFileNameReverse!=null) {
	    if (forward)
		return getReads(randPosReadsFileNameForward, nUniqueRandomReadsForward);
	    return getReads(randPosReadsFileNameReverse, nUniqueRandomReadsReverse);
	}
	System.err.println("Must generate the random reads first!");
	return null;
    }

    public int[][] getRandomReads() {
	if (randPosReadsFileName!=null) 
	    return getReads(randPosReadsFileName, nUniqueRandomReads);
	System.err.println("Must generate the random reads first!");
	return null;
    }

    public int[][] findPeakPositions(double fdrRatio, int minDist) {
	//Returns the positions of the peaks with a given FDR
	//ratio. Use FastaSequence to retrieve the corresponding sequences
	int minHeight = calculateFDRHeight(fdrRatio);
	//Data struct for storing the peaks. Since these will be
	//relatively few, we can use one with objects.
	TreeMap peaks = new TreeMap();
	//Find the local maxima which exceed the minHeight. We require
	//peaks to be separated by at least minDist.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(posReadsFileNameForward));
	    String line = br.readLine();
	    Matcher m = posReadPattern.matcher(line);
	    m.find();
	    //Keep track of the last minDist positions using a queue
	    LinkedList prevReads = new LinkedList();
	    int p = Integer.parseInt(m.group(1)), reads = Integer.parseInt(m.group(2)), prevMaxRead = -2*minDist, prevMaxPos = -1;
	    while (line!=null && p<50000000) {
		m = posReadPattern.matcher(line);
		if (m.find()) {
		    int pos = Integer.parseInt(m.group(1));
		    reads = Integer.parseInt(m.group(2));
		    if ((pos - prevMaxPos)>minDist) {
			if (reads>minHeight) { //Make sure that it is larger than the last minDist points
			    if (windowMax(prevReads)<=reads) 
				peaks.put(new Integer(prevMaxPos), new Integer(prevMaxRead));
			}
			prevMaxPos = pos;
			prevMaxRead = reads;
		    }
		    else {
			if (reads>prevMaxRead) {
			    prevMaxRead = reads;
			    prevMaxPos = pos;
			}
		    }
		    //Store in the window
		    while (p<(pos-1) && p<50000000) {
			if (prevReads.size()>=minDist)
			    prevReads.removeFirst();
			prevReads.addLast(new Integer(0));
			p++;
		    }
		    if (prevReads.size()>=minDist)
			prevReads.removeFirst();
		    prevReads.addLast(new Integer(reads));
		}
		br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	int peakPos[][] = new int[peaks.size()][2], i = 0;
	Iterator it = peaks.keySet().iterator();
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    peakPos[i][0] = key.intValue();
	    peakPos[i][1] = ((Integer)peaks.get(key)).intValue();
	}
	return peakPos;
    }

    protected int windowMax(LinkedList win) {
	//Returns the largest element in a window
	ListIterator li = win.listIterator();
	int max = Integer.MIN_VALUE;
	while (li.hasNext()) {
	    int e = ((Integer)li.next()).intValue();
	    max = max>e ? max : e;
	}
	return max;
    }

}
