//Use this method to classify each gene into one of a number of
//categories based on its density. The number of categories and their
//cut-offs is adjustable and should be provided as a file.

//Example:

//java -Xmx512m SymbolicExpressionDynamics -m mRNA/hr0_20080604.unique.csfasta.ma.35.3__Genes-FEATURES-READS-r35_x500_N10M__ALL-GENES__EXN-ALL_INT-ALL_SPL_1rd.tsv -C 10

//Use split_symbolic_dynamics_file.m to post-process the output.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class SymbolicExpressionDynamics{

    public static String GENOME_LOC;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    CmdLineParser.Option cutOffsFileNameFlag = argParser.addStringOption('C', "cutOffsFileName");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option transcriptDensityFlag = argParser.addBooleanOption('x', "transcriptDensity");
	    //Use this flag to determine the minimum threshold for the densities
	    CmdLineParser.Option thresholdFlag = argParser.addDoubleOption('t', "threshold");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(cutOffsFileNameFlag);
	    String cutOffsFileName = tmp!=null ? (String)tmp : null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean transcriptDensity = argParser.getOptionValue(transcriptDensityFlag)!=null;
	    tmp = argParser.getOptionValue(thresholdFlag);
	    double threshold = tmp!=null ? ((Double)tmp).doubleValue() : 0;
	    //////////////////////////////////////////////////////////////////
	    double cutOffs[] = parseCutOffs(cutOffsFileName);
	    System.out.println("Cut offs are " + Arrays.toString(cutOffs));
	    if (chr==-1) {
		LinkedHashSet genesInGroup[] = new LinkedHashSet[3];
		RefSeqGenes annotationInfo[] = new RefSeqGenes[21];
		TreeSet rnaSeq[] = new TreeSet[3];
		for (int i=1; i<=21; i++)
		    annotationInfo[i-1] = new RefSeqGenes(i, orchestra);
		for (int j=0; j<3; j++) {
		    String mRNAFile = getMRNAFileName(mRNAFileName, j);
		    System.out.println("Opening " + mRNAFile);
		    RNASeqData rnaSeqData = new RNASeqData(mRNAFile, annotationInfo);
		    rnaSeq[j] = rnaSeqData.getRNASeqData();
		}
		rnaSeq = removeAllZeros(rnaSeq, threshold);
		for (int j=0; j<3; j++) {
		    //Find the exon densities corresponding to the various cut-offs
		    double exonDensities[] = new double[rnaSeq[j].size()];
		    Iterator it = rnaSeq[j].iterator();
		    int i = 0, nZero = 0;
		    while (it.hasNext()) {
			exonDensities[i++] = transcriptDensity ? ((RNASeqGene)it.next()).getTranscriptDensity() : ((RNASeqGene)it.next()).getExonDensity();
			if (exonDensities[i-1]==0) { nZero++; }
		    }
		    Arrays.sort(exonDensities);
		    //Find out which element corresponds to which cut-off
		    System.out.println("Found " + exonDensities.length + " exon densities and nZero = " + nZero);
		    double densityCutOffs[] = new double[cutOffs.length];
		    for (i=0; i<cutOffs.length; i++) 
			densityCutOffs[i] = exonDensities[(int)(cutOffs[i]*(double)exonDensities.length) - 1];
		    System.out.println("Cut off densities are " + Arrays.toString(densityCutOffs));
		    //Now we can find out which genes belong to each category
		    genesInGroup[j] = new LinkedHashSet();
		    double prevDens = -1;
		    for (i=0; i<cutOffs.length; i++) {
			TreeSet genes = new TreeSet();
			it = rnaSeq[j].iterator();
			while (it.hasNext()) {//Find the genes that are in this group
			    RNASeqGene rsg = (RNASeqGene)it.next();
			    if (rsg.getExonDensity()>prevDens && rsg.getExonDensity()<=densityCutOffs[i]) { genes.add(rsg); }
			}
			genesInGroup[j].add(genes);
			System.out.println("Found " + genes.size() + " for j = " + j + " and " + genesInGroup[j].size());
			prevDens = densityCutOffs[i];
		    }
		}
		String saveFileName = "mRNA/SymbolicDynamics/symbolic_dynamics" + (transcriptDensity ? "_xpt" : "") + ".txt";
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName)); //Save to file
		bw.write("#Cut-offs:");
		for (int i=0; i<cutOffs.length; i++) { bw.write("\t" + format.format(cutOffs[i])); }
		bw.write("\n#Name\tGeneID\tClass_0h\tClass_1h\tClass_6h\tExp_0h\tExp_1h\tExp_6h\n");
		Iterator it = genesInGroup[0].iterator();
		int i = 0;
		System.out.println(((TreeSet)genesInGroup[1].iterator().next()).size() + "\t" + ((TreeSet)genesInGroup[2].iterator().next()).size());
		while (it.hasNext()) {
		    TreeSet genes = (TreeSet)it.next();
		    System.out.println("Found group with " + genes.size());
		    Iterator it2 = genes.iterator();
		    while (it2.hasNext()) {
			RNASeqGene rsg = (RNASeqGene)it2.next();
			//Find out which class the gene belongs to at the other time points
			int classVec[] = new int[genesInGroup.length];
			classVec[0] = i;
			for (int j=1; j<genesInGroup.length; j++) {
			    //System.out.println(((TreeSet)genesInGroup[j].iterator().next()).size());
			    classVec[j] = getGeneClass(rsg, genesInGroup[j]);
			}
			double expVec[] = new double[genesInGroup.length]; //Find the denisities
			expVec[0] = rsg.getExonDensity();
			for (int j=1; j<genesInGroup.length; j++) 
			    expVec[j] = getExonDensity(rsg, genesInGroup[j]);
			bw.write(rsg.getAnnotationInfo().getID() + "\t" + rsg.getAnnotationInfo().getName());
			for (int j=0; j<genesInGroup.length; j++) { bw.write("\t" + classVec[j]); }
			for (int j=0; j<genesInGroup.length; j++) { bw.write("\t" + format.format(expVec[j])); }
			bw.write("\n");
		    }
		    i++;
		}
		bw.close();
	    }
	    else {

	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int getGeneClass(RNASeqGene rsg, LinkedHashSet genesInGroup) {
	//Find out which class the gene belongs to for a given time point
	int c = 0;
	Iterator it = genesInGroup.iterator();
	while (it.hasNext()) {
	    TreeSet genes = (TreeSet)it.next();
	    //System.out.println("Looking for " + rsg.toString() + " in group " + c + " containing " + genes.size());
	    if (genes.contains(rsg)) 
		return c;
	    else
		c++;
	}
	System.err.println("Error, could not find class for gene " + rsg.toString());
	return -1;
    }

    public static double getExonDensity(RNASeqGene rsg, LinkedHashSet genesInGroup) {
	Iterator it = genesInGroup.iterator();
	while (it.hasNext()) {
	    TreeSet genes = (TreeSet)it.next();
	    //System.out.println("Looking for " + rsg.toString() + " in group " + c + " containing " + genes.size());
	    if (genes.contains(rsg)) 
		return ((RNASeqGene)(new TreeSet(genes.tailSet(rsg))).first()).getExonDensity();
	}
	System.err.println("Error, could not find class for gene " + rsg.toString());
	return -1;
    }

    public static String getMRNAFileName(String mRNAFileName, int t) {
	//Find the name of the file corresponding to time-point, 0, 1 or 6 hours
	if (t==0) 
	    return mRNAFileName.replaceFirst("hr\\d", "hr0");
	else if (t==1) 
	    return mRNAFileName.replaceFirst("hr\\d", "hr1");
	else if (t==2) 
	    return mRNAFileName.replaceFirst("hr\\d", "hr6");
	System.err.println("Error, no data for this time point!");
	return mRNAFileName;
    }

    public static double[] parseCutOffs(String fileName) throws IOException {
	double cutOffs[] = null;
	try {
	    int n = Integer.parseInt(fileName);
	    cutOffs = new double[n];
	    for (int i=0; i<cutOffs.length; i++)
		cutOffs[i] = (double)(i + 1)/n;
	}
	catch (NumberFormatException e) {
	    Vector v = new Vector();
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    while (line!=null) {
		if (!line.startsWith("#"))
		    v.addElement(line);
	    }
	    br.close();
	    cutOffs = new double[v.size()];
	    for (int i=0; i<cutOffs.length; i++)
		cutOffs[i] = Double.parseDouble((String)v.elementAt(i));
	}
	return cutOffs;
    }

    public static TreeSet[] removeAllZeros(TreeSet genes[]) { return removeAllZeros(genes, 0); }

    public static TreeSet[] removeAllZeros(TreeSet genes[], double threshold) {
	//Remove the genes which are below threshold for all three time points.
	LinkedHashSet removeSet = new LinkedHashSet(), expressedSet = new LinkedHashSet();
	for (int i=0; i<genes.length; i++) {
	    Iterator it = genes[i].iterator();
	    while (it.hasNext()) {
		RNASeqGene g = (RNASeqGene)it.next();
		if (g.getExonDensity()>threshold) {
		    if (removeSet.contains(g)) {
			removeSet.remove(g); //has expression at a later time
			//System.out.println("removed " + g.toString());
		    }
		    expressedSet.add(g);
		}
		else if (g.getExonDensity()<=threshold && !expressedSet.contains(g))
		    removeSet.add(g);
	    }
	}
	System.out.println("Found " + removeSet.size() + " genes which were below the threshold.");
	Iterator it = removeSet.iterator();
	while (it.hasNext()) {
	    RNASeqGene g = (RNASeqGene)it.next();
	    for (int i=0; i<genes.length; i++)
		genes[i].remove(g);
	}
	return genes;
    }

}
