//Use this class to extract the contiguous sequence from a set of
//tiling arrays and then scan for motifs using the TFMscan software.
//
//This class is designed to be integrated into Matlab. For example:
//
//tfm_scanner = javaObject('TFMScanner', peak_file, true);
//javaMethod('addMotifMatrix', tfm_scanner, pfm_file);
//javaMethod('findMotifs', tfm_scanner, .001, dir_name);
//motif_locs = double(javaMethod('getMotifLocations', tfm_scanner, pfm_file));

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class TFMScanner {

    protected String sequences[] = null, headers[] = null;
    protected TFMParser parser;
    protected Vector<String> motifMatrices = new Vector<String>();
    //Matrix -> header -> motif pos
    protected TreeMap<String, TreeMap<String, TreeSet<Integer>>> matrixMotifMap = new TreeMap<String, TreeMap<String, TreeSet<Integer>>>();
    //Matrix -> header -> motif score
    protected TreeMap<String, double[]> matrixMotifScores = new TreeMap<String, double[]>();
    protected Random random = new Random();
    protected boolean orchestra;

    public TFMScanner(String fileName, boolean o){
	//use this when we have already extraced the peak sequences
	//and give the fasta file as an argument.
	orchestra = o;
	parseFastaFile(fileName);
    }

    public TFMScanner(String seqs[], boolean o) {
	//use this method when we have a list of sequences already parsed from file
	orchestra = o;
	sequences = seqs;
	headers = seqs;
    }

    //////////////////////////////////////////////////////////////////

    protected void parseFastaFile(String fileName) {
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    Vector seqs = new Vector(), heads = new Vector();
	    String line = br.readLine(), seq = "";
	    heads.addElement(line);
	    line = br.readLine(); //Skip forst line since it should be a comment
	    while (line!=null) {
		if (line.startsWith(">") && seq.length()>0) {
		    seqs.addElement(new String(seq));
		    seq = "";
		    heads.addElement(line);
		}
		else
		    seq += line;
		line = br.readLine();
	    }
	    br.close();
	    seqs.addElement(new String(seq));
	    sequences = new String[seqs.size()];
	    headers = new String[heads.size()];
	    for (int i=0; i<seqs.size(); i++) {
		sequences[i] = (String)seqs.elementAt(i);
		headers[i] = (String)heads.elementAt(i);
	    }
	}
	catch (IOException e) { 
	    System.err.println("Error with TFMScanner " + e.toString());
	}
    }


    //////////////////////////////////////////////////////////////////

    public void addMotifMatrix(String matrixFileName){
	motifMatrices.addElement(matrixFileName);
    }


    public Vector<String> getMotifMatrices() { return motifMatrices; }
    
    public int[] getMotifNumberDistribution(String motif) { return getMotifNumberDistribution(motif, 1); }

    public int[] getMotifNumberDistribution(String motif, int nMax) {
	//Find the number of sequences that have 0...nMax motif occurences
	int n[] = new int[nMax];
	for (String h : matrixMotifMap.get(motif).keySet()) {
	    int i = matrixMotifMap.get(motif).get(h).size();
	    //System.out.println(h + "\t" + i);
	    n[i>=nMax ? nMax-1 : i]++;
	}
	return n;
    }

    public int numberOfSequences() { return sequences.length; }

    public void findMotifs(double pVal) {
	findMotifs(pVal, "/home/mhemberg/Projects/Synergism/TFMscan1.0/Data/Mats/Transfac/");
    }

    public String createTFMMatrixFile(String matrixDir) { return createTFMMatrixFile(matrixDir, "tfm_scan_matrix_tmp.txt"); }

    public String createTFMMatrixFile(String matrixDir, String tfmMatrixFileName) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(tfmMatrixFileName));
	    bw.write(matrixDir + "\n");
	    for (String motif : motifMatrices)
		bw.write(motif + "\n");
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error writing TFM matrix file! " + e.toString());
	}
	return tfmMatrixFileName;
    }

    public void findMotifs(double pVal, String matrixDir){ findMotifs(pVal, matrixDir, "tfm_tmp.out"); }

    public void findMotifs(double pVal, String matrixDir, String tfmOutputFileName){
	//Create the file required by TFMscan when searching for motifs
	String tfmMatrixFileName = matrixDir.endsWith(".mat") ? matrixDir : createTFMMatrixFile(matrixDir);
	
	matrixMotifMap = new TreeMap<String, TreeMap<String, TreeSet<Integer>>>();
	if (matrixDir.endsWith(".mat")) {
	    try { //Parse the names of the TFs and add
		BufferedReader br = new BufferedReader(new FileReader(matrixDir));
		String line = br.readLine(); //skip first line
		line = br.readLine();
		while (line!=null) {
		    if (line.length()>0)
			motifMatrices.addElement(line);
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException ex) {
		System.out.println("Error parsing file " + ex.toString());
	    }
	    System.out.println("Found " + motifMatrices.size() + " motifs.");
	}
	for (int i=0; i<motifMatrices.size(); i++) {
	    matrixMotifMap.put((String)motifMatrices.elementAt(i), new TreeMap<String, TreeSet<Integer>>());
	    matrixMotifScores.put((String)motifMatrices.elementAt(i), new double[sequences.length]);
	}
	for (int i=0; i<sequences.length; i++) {
	    runTFMscan(sequences[i], pVal, tfmMatrixFileName);
	    //Parse the output
	    parser = new TFMParser(tfmOutputFileName);
	    for (int k=0; k<motifMatrices.size(); k++) {
		String matrix = (String)motifMatrices.elementAt(k);
		TreeSet<Integer> motifLocs = parser.findMotifLocations(matrix);
		//System.out.println("Found " + motifLocs.size() + " for " + headers[i]);
		//Put them in list
		if (!matrixMotifMap.containsKey(matrix))
		    matrixMotifMap.put(matrix, new TreeMap<String, TreeSet<Integer>>());
		if (!matrixMotifMap.get(matrix).containsKey(headers[i])) 
		    matrixMotifMap.get(matrix).put(headers[i], new TreeSet<Integer>());
		matrixMotifMap.get(matrix).get(headers[i]).addAll(motifLocs);
	    }
	    for (int k=0; k<motifMatrices.size(); k++) {
		//Get the scores for the sequence and save the largest one
		double scores[] = parser.findMotifTFMScores((String)motifMatrices.elementAt(k));
		matrixMotifScores.get((String)motifMatrices.elementAt(k))[i] = -1;
		for (int j=0; j<scores.length; j++) 
		    matrixMotifScores.get((String)motifMatrices.elementAt(k))[i] = Math.max(matrixMotifScores.get((String)motifMatrices.elementAt(k))[i], scores[j]);
	    }
	    runTFMscan(MerTree.reverseComplement(sequences[i]), pVal, tfmMatrixFileName);
	    parser = new TFMParser(tfmOutputFileName);
	    for (int k=0; k<motifMatrices.size(); k++) {
		String matrix = (String)motifMatrices.elementAt(k);
		TreeSet<Integer> motifLocs = parser.findMotifLocations(matrix, sequences[i].length());
		//System.out.println("Found " + motifLocs.size() + " for reverse " + headers[i]);
		if (!matrixMotifMap.containsKey(matrix)) 
		    matrixMotifMap.put(matrix, new TreeMap<String, TreeSet<Integer>>());
		if (!matrixMotifMap.get(matrix).containsKey(headers[i])) 
		    matrixMotifMap.get(matrix).put(headers[i], new TreeSet<Integer>());
		matrixMotifMap.get(matrix).get(headers[i]).addAll(motifLocs);
	    }
	    for (int k=0; k<motifMatrices.size(); k++) {
		//Use the highest scores from the two strands
		double scores[] = parser.findMotifTFMScores((String)motifMatrices.elementAt(k));
		for (int j=0; j<scores.length; j++) 
		    matrixMotifScores.get((String)motifMatrices.elementAt(k))[i] = Math.max(matrixMotifScores.get((String)motifMatrices.elementAt(k))[i], scores[j]);
	    }
	}
    }

    protected void runTFMscan(String seq, double pVal, String tfmMatrixFileName) { runTFMscan(seq, pVal, tfmMatrixFileName, "tfm_tmp.out"); }

    protected void runTFMscan(String seq, double pVal, String tfmMatrixFileName, String tfmOutputFileName) {
	//Write the sequence to a FASTA formatted file
	String fastaTmpFile = "tfm_scan_input_tmp.fa";
	FastaSequence2.saveSequence(fastaTmpFile, "test", seq);
	//Run TFMscan
	String tfmOutput = "", tfmCmd = "";;
	try {
	    Runtime runtime = Runtime.getRuntime();
	    String tfmLoc = orchestra ? "/home/mh168/Softwares/TFMscan1.0/TFMscan" : "/home/mhemberg/Softwares/TFMscan1.0/TFMscan";
	    tfmCmd = tfmLoc + " -p " + Math.round(-Math.log(pVal)) + " -m " + tfmMatrixFileName + " -s " + fastaTmpFile;
	    //System.out.println("Running TFMscan:\n" + tfmCmd + " > " + tfmOutputFileName);
	    Process p = runtime.exec(tfmCmd);
	    BufferedInputStream bis = new BufferedInputStream(p.getInputStream());
	    p.waitFor();
	    int b = bis.available();
	    while (b>0) {
		byte[] bytes = new byte[b];
		bis.read(bytes, 0, b);
		tfmOutput = tfmOutput.concat(new String(bytes));
		b = bis.available();
	    }
	    bis.close();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(tfmOutputFileName));
	    bw.write(tfmOutput);
	    bw.close();
	}
	catch (Exception e) {
	    System.err.println("Error using TFMscan:\n" + tfmCmd + " > " + tfmOutputFileName);
	    System.err.println(e.toString());
	}
    }

    public TreeSet<Integer> getMotifLocations(String matrix, int seq){
	//Use this method to extract the locations of the motifs
	//corresponding to a given matrix
	if (matrixMotifMap.containsKey(matrix)) {
	    TreeMap<String, TreeSet<Integer>> motifLocations = matrixMotifMap.get(matrix);
	    return motifLocations.get(headers[seq]);
	}
	System.err.println("Error, could not find any motifs for " + matrix);
	return null;
    }

    public double[][] getMotifScores() {
	//Return the score for each loci
	double scores[][] = new double[motifMatrices.size()][sequences.length];
	for (int i=0; i<motifMatrices.size(); i++) 
	    scores[i] = matrixMotifScores.get(motifMatrices.elementAt(i));
	return scores;
    }
    
}

class TFMParser {

    protected String fileName;

    public TFMParser(String f){
	fileName = f;
    }

    //public double[] findMotifTFMScores() { return findMotifTFMScores(""); }
    
    public double[] findMotifTFMScores(String matrixName) {
	Vector motifScores = new Vector();
	matrixName = matrixName.length()==0 ? null : matrixName;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern posPattern = Pattern.compile("^(\\d+)\\s+(\\w+)\\s+([\\d|\\.|-]+)");
	    while (line!=null) {
		Matcher m = posPattern.matcher(line);
		if (m.find()) {
		    if (matrixName==null || matrixName.equals(m.group(2)))
			motifScores.addElement(new Double(m.group(3)));
		    //System.out.println("Found line: " + line + ", with " + m.group(1));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	double[] motifs = new double[motifScores.size()];
	for (int i=0; i<motifScores.size(); i++)
	    motifs[i] = ((Double)motifScores.elementAt(i)).doubleValue();
	return motifs;
    }


    public TreeSet<Integer> findMotifLocations(int a) {	return findMotifLocations(""); }

    public TreeSet<Integer> findMotifLocations(String matrixName){ return findMotifLocations(matrixName, -1); }

    public TreeSet<Integer> findMotifLocations(String matrixName, int reverseLen){
	TreeSet<Integer> motifInds = new TreeSet<Integer>();
	matrixName = matrixName.length()==0 ? null : matrixName;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern posPattern = Pattern.compile("^(\\d+)\\s+(\\w+)\\s+([\\d|\\.|-]+)");
	    while (line!=null) {
		Matcher m = posPattern.matcher(line);
		if (m.find()) {
		    if (matrixName==null || matrixName.equals(m.group(2)))
			motifInds.add(reverseLen<0 ? new Integer(m.group(1)) : reverseLen - new Integer(m.group(1)));
		    //System.out.println("Found line: " + line + ", with " + m.group(1));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	return motifInds;
    }


}
