//Use this class to keep track of all the reads for a given chromosome
//and experiment. There are also methods for grouping them into
//tentative transcripts and assessing the quality of the partition.
//
//v2 is a slimmed down version where a lot of old stuff has been
//removed. It is intended to be used with FindTranscripts3.java


import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import umontreal.iro.lecuyer.probdist.*;
import umontreal.iro.lecuyer.stat.Tally;
import umontreal.iro.lecuyer.stat.TallyStore;
import org.apache.commons.math.util.MathUtils;

//import flanagan.integration.Integration;
//import flanagan.integration.IntegralFunction;



public class TranscriptFinder2 {

    public static final int MAX_ITERATIONS = 20;

    protected ABIRead reads[];
    protected int readsInd = 0, genomeInd = 0;
    protected int chr; 
    protected boolean log = true;
    private int maxReadLength = 35;
    private boolean forward = true;
    protected Tally coeffsTally[];
    protected double[] sigmaHat;
    protected String species;
    //Length-scale -> location -> coeff-val
    protected TreeMap<Integer, TreeMap<Integer, Double>> localMax = new TreeMap<Integer, TreeMap<Integer, Double>>();
    //Keep track of all the local max, this will speed up calculations
    //since we can avoid recomputing the coefficients
    


    //Keep track of the location of break points and the average
    //expression levels. Make sure to use log-read counts for stability.
    protected ReducedModel2 reducedModel = null;

    protected int readGroupInd = 0;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random(42);

    //////////////////////////////////////////////////////////////////

    public TranscriptFinder2(String fName, int c, boolean strand, int rLength, String sp, boolean l) {
	species = sp;
	reads = ABIReads.loadMRNAReadsChr(fName, strand, c, maxReadLength, false);
	chr = c;
	maxReadLength = rLength;
	forward = strand;
	log = l; //!flatten;
    }

    public TranscriptFinder2(ABIRead[] rs, int c, boolean strand, int rLength, String sp, boolean l) {
	species = sp;
	reads = rs;
	chr = c;
	maxReadLength = rLength;
	forward = strand;
	log = l; //!flatten;
    }

    protected TranscriptFinder2() { }

    //////////////////////////////////////////////////////////////////


    public void removeHaarCoefficients() {
	readGroupInd = 0;
	localMax = new TreeMap<Integer, TreeMap<Integer, Double>>();
	coeffsTally = null;
	readsInd = 0;
	genomeInd = 0;
    }

    public void calculateHaarCoefficients(int lVec[], int binSize, boolean chip) { calculateHaarCoefficients(lVec, binSize, chip, null); }

    public void calculateHaarCoefficients(int lVec[], int binSize, boolean chip, String saveInfoFileName) { calculateHaarCoefficients(lVec, binSize, null); }

    public void calculateHaarCoefficients(int lVec[], int binSize, String saveInfoFileName) {
	/*if (species.equals("sc"))
	    calculateHaarCoefficientsAll(lVec, binSize, saveInfoFileName);
	    else*/
	    calculateHaarCoefficientsBlock(lVec, binSize, saveInfoFileName);
    }
	
    public void calculateHaarCoefficientsBuffer(int lVec[]) {
	//pre-compute the normalization coefficients for the Haar wavelets
	double normCoeffs[] = new double[lVec.length];
	for (int i=0; i<lVec.length; i++)
	    normCoeffs[i] = 1/Math.sqrt(Math.pow(2, lVec[i] + 1));
	TreeMap<Double, Integer> coeffsShortestHist = new TreeMap<Double, Integer>();
	for (int i=0; i<lVec.length; i++)
	    localMax.put(lVec[i], new TreeMap<Integer, Double>());
	int winLen = (int)Math.pow(2, lVec[lVec.length - 1]);
	double readsHist[] = calculateHistogram(0, winLen);
	double w[] = new double[lVec.length], wp[] = new double[lVec.length], wpp[] = new double[lVec.length];
	//Calculate the first two coefficients
	for (int i=0; i<lVec.length; i++) 
	    wpp[i] = calculateHaarCoefficient(lVec[i], 0, readsHist, i==0 ? 0 : wpp[i-1], log, normCoeffs[i]);
	readsHist = shiftHistogram(readsHist, ++genomeInd + winLen, 2*winLen);
	for (int i=0; i<lVec.length; i++) 
	    wp[i] = calculateHaarCoefficientInc(lVec[i], 1, readsHist, wpp[i], log, normCoeffs[i]);
	readsHist = shiftHistogram(readsHist, ++genomeInd + winLen, 2*winLen);
	while (genomeInd<numberOfBasePairs(chr, species)) {
	    for (int i=0; i<lVec.length; i++) { //Find coefficients for each subband
		w[i] = calculateHaarCoefficientInc(lVec[i], Math.min(genomeInd, readsHist.length/2), readsHist, wp[i], log, normCoeffs[i]);
		//Check if the previous coeff was a local maximum
		if (i<lVec.length-1 && isLocalExtrema(w, i)) 
		    localMax.get(lVec[i]).put(genomeInd, wp[i]);
		//shift the coefficients
		wpp[i] = wp[i];
		wp[i] = w[i];
	    }
	    coeffsShortestHist.put(Math.abs(w[0]), coeffsShortestHist.containsKey(Math.abs(w[0])) ? coeffsShortestHist.get(Math.abs(w[0])) + 1 : 1);
	    readsHist = shiftHistogram(readsHist, ++genomeInd + winLen, 2*winLen);
	    //System.out.println(Arrays.toString(readsHist));
	}
	//Calculate the FDR by considering all the localMax for each band
	//double sigmaHat = estimateStdDev(coeffsShortest);
	sigmaHat[0] = estimateStdDev(coeffsShortestHist);
	//System.out.println("sigmaHat = " + format.format(sigmaHat));
    }

    public void calculateHaarCoefficientsBlock(int lVec[], int binSize) { calculateHaarCoefficientsBlock(lVec, binSize, null); }

    public void calculateHaarCoefficientsBlock(int lVec[], int binSize, String saveInfoFileName) {
	//pre-compute the normalization coefficients for the Haar wavelets
	double normCoeffs[] = new double[lVec.length];
	for (int i=0; i<lVec.length; i++)
	    normCoeffs[i] = 1/Math.sqrt(Math.pow(2, lVec[i] + 1));
	//Since the variability in terms of number of reads for a
	//given locus (and hence the magnitude of the coefficients) is
	//much smaller for ChIP than for RNA-Seq, we no longer need to
	//create bins for the histogram.
	coeffsTally = new Tally[lVec.length];
	sigmaHat = new double[lVec.length];
	try {
	    //One file for the coefficients for each length-scale
	    BufferedWriter bw[] = new BufferedWriter[lVec.length];
	    for (int i=0; i<lVec.length; i++) {
		localMax.put(lVec[i], new TreeMap<Integer, Double>());
		coeffsTally[i] = new Tally("coeffs " + lVec[i]);
		bw[i] = saveInfoFileName!=null ? new BufferedWriter(new FileWriter(saveInfoFileName + "_coeffs_" + lVec[i] + "_" + chr + ".bed")) : null;
	    }
	    int winLen = (int)Math.pow(2, lVec[lVec.length - 1]);
	    while (moreReadGroups()/* && reads[readGroupInd].getPosition()<3500000*/) {
		ABIRead[] rs = nextReadGroup(2*winLen);
		double readsHist[] = calculateHistogram(rs, maxReadLength, winLen, binSize, saveInfoFileName);
		//System.out.println("Found block with length " + readsHist.length + ", containing " + rs.length + " reads starting at " + (rs[0].getPosition() - maxReadLength));
		double wp = 0;
		int histStart = rs[0].getPosition() - winLen;
		for (int i=0; i<lVec.length; i++) { //Find coefficients for each subband
		    double w[] = new double[readsHist.length + 2*winLen];
		    //Calculate the first coefficient
		    w[0] = i==0 ? calculateHaarCoefficient(lVec[i] - binSize, 0, readsHist, log, normCoeffs[i], reducedModel, histStart, binSize) : calculateHaarCoefficient(lVec[i] - binSize, 0, readsHist, wp, log, normCoeffs[i], reducedModel, histStart, binSize);
		    for (int j=1; j<readsHist.length; j++)  //subsequent coefficients can be calculated incrementally and hence significantly faster
			w[j] = calculateHaarCoefficientInc(lVec[i] - binSize, j, readsHist, w[j-1], log, normCoeffs[i], reducedModel, histStart, binSize);
		    localMax.get(lVec[i]).putAll(findLocalMax(w, rs[0].getPosition() - Math.min(winLen, rs[0].getPosition()), binSize));
		    wp = w[0]; //to speed up calculation of first coeff at the next length scale
		    int b = (int)Math.pow(2, binSize);
		    for (int j=0; j<w.length; j++) {
			if (j>0 && j<w.length-1 && isLocalExtrema(w, j))
			    coeffsTally[i].add(w[j]);
			if (bw[i]!=null && (rs[0].getPosition() - winLen/b + (j + 1)*b)<PeakFinder2.CHROMOSOME_LENGTH[chr-1]) 
			    bw[i].write("chr" + chr + "\t" + (rs[0].getPosition() - winLen + j*b) + "\t" + (rs[0].getPosition() - winLen + (j + 1)*b) + "\t" + format.format(w[j]) + "\n"); 
		    }
		}
	    }
	    for (int i=0; i<coeffsTally.length; i++) {
		sigmaHat[i] = coeffsTally[i].numberObs()>1 ? coeffsTally[i].standardDeviation() : 0;
		//System.out.println("sigmaHat = " + format.format(coeffsTally[i].standardDeviation()) + " for mean " + format.format(coeffsTally[i].average()) + " from " + coeffsTally[i].numberObs());
	    }
	    for (int i=0; i<lVec.length; i++) {
		if (bw[i]!=null) { 
		    bw[i].close(); 
		    String name = saveInfoFileName.substring(saveInfoFileName.lastIndexOf("/") + 1) + "_coeffs_" + lVec[i];
		    String colorStr = CreateWigFiles.getColorString(name);
		    int priority = CreateWigFiles.getPriority(name);
		    String trackInfoStr = "track type=bigWig name=\"" + name + "_" + chr + "\" description=\"" + name + "_" + chr + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + FindTranscripts.URL_NAME + name + "_" + chr + ".bw";
		    //CreateWigFiles.addTrackInfo(trackInfoStr, FindTranscripts.TRACK_INFO_FILE_NAME);
		    //CreateWigFiles.createWWWLink(saveInfoFileName + "_coeffs_" + lVec[i] + "_" + chr + ".bed", "/www/" + FindTranscripts.WWW_DIR_NAME, species, hg19);
		}
	    }
	}
	catch (IOException e) { 

	}
    }
    
    public void calculateHaarCoefficientsAll(int lVec[], int binSize) { calculateHaarCoefficientsAll(lVec, binSize, null); }

    public void calculateHaarCoefficientsAll(int lVec[], int binSize, String saveInfoFileName) {
	//pre-compute the normalization coefficients for the Haar wavelets
	double normCoeffs[] = new double[lVec.length];
	Tally coeffsTally[] = new Tally[lVec.length];
	sigmaHat = new double[lVec.length];	
	for (int i=0; i<lVec.length; i++) {
	    normCoeffs[i] = 1/Math.sqrt(Math.pow(2, lVec[i] + 1));
	    localMax.put(lVec[i], new TreeMap<Integer, Double>());
	    coeffsTally[i] = new Tally();
	}
	int winLen = (int)Math.pow(2, lVec[lVec.length - 1]);
	double readsHist[] = calculateHistogram(reads, maxReadLength, winLen, binSize);
	System.out.println("Found histogram with length " + readsHist.length + " starting at " + reads[0].getPosition());
	double wp = 0;
	for (int i=0; i<lVec.length; i++) { //Find coefficients for each subband
	    double w[] = new double[readsHist.length + 2*winLen];
	    //Calculate the first coefficient
	    w[0] = i==0 ? calculateHaarCoefficient(lVec[i], 0, readsHist, log, normCoeffs[i]) : calculateHaarCoefficient(lVec[i], 0, readsHist, wp, log, normCoeffs[i]);
	    for (int j=1; j<readsHist.length; j++)  //subsequent coefficients can be calculated incrementally and hence significantly faster
		w[j] = calculateHaarCoefficientInc(lVec[i], j, readsHist, w[j], log, normCoeffs[i]);
	    //w[j] = calculateHaarCoefficientInc(lVec[i], j, readsHist, w[j-1], normCoeffs[i]);
	    //We do not need to store all coefficients. We can
	    //save memory by only saving the ones for the
	    //shortest scale (needed for estimating sigmaHat)
	    //and the local maxima with corresponding
	    //locations (the only ones that may get selected
	    //later on).
	    localMax.get(lVec[i]).putAll(findLocalMax(w, reads[0].getPosition() - winLen, binSize));
	    wp = w[0]; //to speed up calculation of first coeff at the next length scale
	    for (int j=0; j<w.length; j++)
		coeffsTally[i].add(w[j]);
	}
	for (int i=0; i<coeffsTally.length; i++) 
	    sigmaHat[i] = coeffsTally[i].numberObs()>1 ? coeffsTally[i].standardDeviation() : 0; 
	//System.out.println("sigmaHat = " + format.format(sigmaHat));
    }

    ///////////////////////////////////////////////////////

    public int findRefinedReducedModel(int lMin, int lMax, int binSize, double fdr, boolean mergeSegments, boolean gmm, String saveInfoFileName, RefSeqUCSCEnsemblGenes geneInfo, Vector<Peak> breakPoints, Vector<Peak> peaks, int maxDivergentTSSDistance, String costFunction) { return findRefinedReducedModel(lMin, lMax, binSize, fdr, mergeSegments, gmm, saveInfoFileName, geneInfo, breakPoints, peaks, maxDivergentTSSDistance, costFunction, -1); }

    public int findRefinedReducedModel(int lMin, int lMax, int binSize, double fdr, boolean mergeSegments, boolean gmm, String saveInfoFileName, RefSeqUCSCEnsemblGenes geneInfo, Vector<Peak> breakPointPeaks, Vector<Peak> peaks, int maxDivergentTSSDistance, String costFunction, int nIterationsMax) { 
	//Use this method to find the best model, by applying the
	//wavelet method iteratively
	double cost = reducedModel==null ? 0 : evaluateCostFunction(costFunction, reducedModel, geneInfo.getGeneIDsKeep(forward), peaks, maxDivergentTSSDistance, !forward), prevCost = 0;
	if (cost>=prevCost) {
	    int[] lVec = FindTranscripts.getLengthVec(lMin, lMax, 1);
	    Vector<Integer> breakPoints = findBreakpoints(lVec, fdr, null, breakPointPeaks, saveInfoFileName);
	    if (breakPoints.size()>0) {
		reducedModel = findReducedModel(reducedModel, 0, lVec, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName);
		//reducedModel.refineTranscribedRegionBoundaries(getReads(), binSize, lMin, maxReadLength, log);
		reducedModel.compare2Annotation(maxDivergentTSSDistance, geneInfo);
		reducedModel.categorizeTranscribedRegions();
		//Find out how many of the transcripts have a peak near their start
		TreeMap<TranscribedRegion2, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion2, Peak>() : reducedModel.compare2Peaks(peaks, maxDivergentTSSDistance);
		//FindTranscripts.printStats(fdr, reducedModel, geneIDsKeepS, transcriptsWithPeaks, peaks);
		//Iterative refinement
		cost = evaluateCostFunction(costFunction, reducedModel, geneInfo.getGeneIDsKeep(forward), peaks, maxDivergentTSSDistance, !forward);
		int nIterations = 1, lmi = lMin, lma = lMax;
		reducedModel.removeAll();
		//ReducedModel2 prevModel = (ReducedModel2)reducedModel.clone();
		while (nIterations<MAX_ITERATIONS && (nIterations<nIterationsMax || nIterationsMax<0)) { //Try again to see if we can find more interesting stuff the second time aroun
		    nIterations++;
		    int lVecMore[] = FindTranscripts.getLengthVec(lmi, lma, 1);
		    removeHaarCoefficients();
		    calculateHaarCoefficients(lVecMore, binSize, null);
		    breakPoints = findBreakpoints(lVecMore, fdr, reducedModel, peaks, null);
		    ReducedModel2 r = findReducedModel((ReducedModel2)reducedModel.clone(), nIterations, lVecMore, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName);
		    r.removeShortSegments();
		    //r.refineTranscribedRegionBoundaries(getReads(), binSize, lmi, maxReadLength, log);
		    r.compare2Annotation(maxDivergentTSSDistance, geneInfo);
		    r.categorizeTranscribedRegions();
		    //FindTranscripts.printTypeStats(r, getReads().length);
		    //Find out how many of the transcripts have a peak near their start
		    transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion2, Peak>() : r.compare2Peaks(peaks, maxDivergentTSSDistance);
		    //FindTranscripts.printStats(fdr, r, geneIDsKeepS, transcriptsWithPeaks, peaks);
		    prevCost = cost;
		    cost = evaluateCostFunction(costFunction, r, geneInfo.getGeneIDsKeep(forward), peaks, maxDivergentTSSDistance, !forward);
		    r.removeAll();
		    System.out.println("Cost is " + format.format(cost) + " and previous round had " + format.format(prevCost));
		    if (prevCost<cost || nIterationsMax>0) 
			reducedModel = r;
		    else 
			break;
		}
		return nIterations;
	    }
	}
	return 0;
    }

    public ReducedModel2 getReducedModel() { return reducedModel; }

    ///////////////////////////////////////////////////////

    public Vector<Integer> findBreakpointsInput(int lVec[], double fdr) { return null; } //ugly solution
	

    public Vector<Integer> findBreakpoints(int lVec[], double fdr) { return findBreakpoints(lVec, fdr, null); }

    public Vector<Integer> findBreakpoints(int lVec[], double fdr, ReducedModel2 reducedModel) { return findBreakpoints(lVec, fdr, reducedModel, null, null); }
	
    public Vector<Integer> findBreakpoints(int lVec[], double fdr, ReducedModel2 reducedModel, Vector<Peak> peaks, String saveInfoFileName) {
	//Keep track of breakpoints stemming from local maxima and
	//local minima separately. In particular, we want to allow for
	//the situation where breakpoints are placed closed to one
	//another if they correspond to different types of
	//extrema. This should allow us to smooth over a longer length
	//scale and still be able to catch short transcribed regions.
	Vector<Integer> breakPoints = new Vector<Integer>(), breakPointsMax = new Vector<Integer>(), breakPointsMin = new Vector<Integer>();
	//for (int i=0; i<lVec.length; i++) { 
	for (int i=lVec.length-1; i>=0; i--) { 
	    //System.out.println("Found " + breakPoints.size() + " break points for scale " + lVec[i]);
	    if (reducedModel!=null) {
		//Remove the coefficients that are located in segments
		coeffsTally[i] = new Tally();
		if (localMax.containsKey(lVec[i])) {
		    for (Integer pos : localMax.get(lVec[i]).keySet()) {
			if (!reducedModel.segmentOverlaps(pos))
			    coeffsTally[i].add(localMax.get(lVec[i]).get(pos));
		    }
		}
	    }
	    //Invert the mapping so that we go from abs locMax -> position
	    TreeMap<Double, Integer> maxLoc = new TreeMap<Double, Integer>();
	    for (Integer pos : localMax.get(lVec[i]).keySet()) {
		//remove the break points that are sitting on top of regions that are already known to be segments
		if (reducedModel==null || !reducedModel.segmentOverlaps(pos)) {
		    double v = localMax.get(lVec[i]).get(pos);
		    if (!maxLoc.containsKey(v))
			maxLoc.put(v, pos);
		    else { //collision
			while (maxLoc.containsKey(v))
			    v += random.nextGaussian()*.00000001;
			maxLoc.put(v, pos);
		    }
		}
	    }
	    if (coeffsTally[i].numberObs()>1) {
		double zThreshold = calculateThreshold(coeffsTally[i], fdr);
		double zThresholdUp = calculateThreshold(coeffsTally[i], fdr, true);
		double zThresholdLow = calculateThreshold(coeffsTally[i], fdr, false);
		//System.out.println("Threshold is " + format.format(zThresholdLow) + " and " + format.format(zThresholdUp) + " for " + lVec[i]);
		//determine which coefficients to keep, we want to avoid using overlapping ones.
		Vector<Double> maxs = new Vector<Double>(maxLoc.keySet());
		for (int j=maxs.size()-1; j>=0; j--) {
		    double m = maxs.elementAt(j);
		    int pos = maxLoc.get(m);
		    if (m>0 && m>zThresholdUp && !nearbyBreakPoint(breakPoints, reducedModel, pos, lVec[i])) {
			breakPoints.addElement(pos);
			//System.out.println("Found breakpoint " + pos + " with maxima " + format.format(m) + " at scale " + lVec[i]);
		    }
		    else if (m<0 && m<zThresholdLow && !nearbyBreakPoint(breakPoints, reducedModel, pos, lVec[i])) {
			breakPoints.addElement(pos);
			//System.out.println("Found breakpoint " + pos + " with minima " + format.format(m) + " at scale " + lVec[i]);
		    }
		}
	    }
	}
	if (reducedModel!=null) {
	    for (TranscribedRegion2 t : reducedModel.getSegments()) { //Add breakspoints corresponding to the start and end of each existing segment
		breakPoints.add(t.getStart());
		breakPoints.add(t.getEnd());
	    }
	}
	if (peaks!=null) { //Finally, add breakpoints corresponding to Pol2 peaks
	    for (Peak p : peaks) {
		//if (!nearbyBreakPoint(breakPoints, reducedModel, p.getCenter(), lVec[0])) {
		//breakPoints.add(p.getStart());
		//breakPoints.add(p.getEnd());
		//breakPoints.add(p.getCenter());
		int peakWidth = p.getEnd() - p.getStart();
		breakPoints.add(p.getCenter() - Math.min(ReducedModel2.MIN_TRANSCRIPT_LENGTH/2 - 5, peakWidth/4));
		breakPoints.add(p.getCenter() + Math.min(ReducedModel2.MIN_TRANSCRIPT_LENGTH/2 - 5, peakWidth/4));
		    //}
	    }
	    //System.out.println("Added break-points for " + peaks.size());
	}
	if (breakPoints.size()>0) {
	    breakPoints.setElementAt(0, 0);
	    Collections.sort(breakPoints);
	}
	//System.out.println("Found a total of " + breakPoints.size() + " break points.");
	if (saveInfoFileName!=null) {
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveInfoFileName + "_breakPoints_" + chr + ".bed"));
		for (Integer i : breakPoints)
		    bw.write("chr" + chr + "\t" + i + "\t" + (i + 1) + "\t2\n"); 
		bw.close(); 
		String name = saveInfoFileName.substring(saveInfoFileName.lastIndexOf("/") + 1) + "_breakPoints";
		String colorStr = CreateWigFiles.getColorString(name);
		int priority = CreateWigFiles.getPriority(name);
		String trackInfoStr = "track type=bigWig name=\"" + name + "_" + chr + "\" description=\"" + name + "_" + chr + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + FindTranscripts.URL_NAME + name + "_" + chr + ".bw";
		//CreateWigFiles.addTrackInfo(trackInfoStr, FindTranscripts3.BREAK_POINTS_FILE_NAME);
		//CreateWigFiles.createWWWLink(saveInfoFileName + "_breakPoints_" + chr + ".bed", "/www/" + FindTranscripts.WWW_DIR_NAME);
	    }
	    catch (IOException e) {
		System.err.println("Error saving break points " + e.toString());
	    }
	}
	return breakPoints;
    }

    public ReducedModel2 findReducedModel(boolean merge) { 
	if (merge)
	    reducedModel.merge(reads, maxReadLength, log);
	return reducedModel; 
    }

    public ReducedModel2 findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, Vector<Integer> breakPointsInput, boolean mergeSegments, boolean gmm, int binSize) { return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, null); }

    public ReducedModel2 findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, boolean mergeSegments, boolean gmm, int binSize) { return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, null); }

    public ReducedModel2 findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, Vector<Integer> breakPointsInput, boolean mergeSegments, boolean gmm, int binSize, String saveInfoFileName) { return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName); }

    public ReducedModel2 findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, boolean mergeSegments, boolean gmm, int binSize, String saveInfoFileName) {
	//Now we can reconstruct the signal by averaging between the
	//breakpoints. We need to add a L_1 penalty to make sure that
	//not all the segments get values.
	//We need to find out which density corresponds to a given FDR
	//for each length scale
	if (gmm)
	    reducedModel = new ReducedModel2(reads, breakPoints, chr, maxReadLength, lVec, mergeSegments, log, binSize, species, false, fdr, saveInfoFileName);
	else {	    
	    double cutOffDens[] = calculateReadLevelCutOffs(lVec, fdr);
	    reducedModel = new ReducedModel2(reads, breakPoints, chr, maxReadLength, lVec, cutOffDens, mergeSegments, log, species, false);
	}
	return reducedModel;
    }

    public ReducedModel2 findReducedModel(ReducedModel2 reducedModel, int nIterations, int lVec[], double fdr, Vector<Integer> breakPoints, boolean mergeSegments, boolean gmm, int binSize, String saveInfoFileName) {
	//Use this method when we have a model which we would like to refine and find additional segments for.
	if (reducedModel==null)
	    return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName);
	if (gmm)
	    reducedModel.createMoreSegments(nIterations, reads, breakPoints, maxReadLength, lVec, mergeSegments, log, binSize, fdr, saveInfoFileName);
	//System.out.println(reducedModel.toString());
	return reducedModel;
    }

    public double[] calculateReadLevelCutOffs(int[] lVec, double fdr) {
	double cutOffs[] = new double[lVec.length];
	for (int i=0; i<lVec.length; i++) { 
	    TreeMap<Double, Integer> densHist = ABIReads.calculateReadDensity(chr, reads, (int)Math.pow(2, lVec[i]), ABIReads.meanReadLength(reads), maxReadLength, species, false, log);
	    TreeMap<Double, Integer> densHistVar = ABIReads.calculateReadDensityVariance(chr, reads, (int)Math.pow(2, lVec[i]), ABIReads.meanReadLength(reads), maxReadLength, species, false, log);
	    int totBins = numberOfBins((int)Math.pow(2, lVec[i])), nBins = 0;
	    Iterator it = densHist.keySet().iterator();
	    double keyDens = 0;
	    //System.out.println("There were " + totBins + " bins in total and threshold is " + format.format((1-fdr)*(double)totBins));
	    //for (Double k : densHistVar.keySet())
	    //System.out.println(format.format(k) + "\t" + format.format(densHistVar.get(k)));
	    while (nBins<(int)Math.round((1-fdr)*(double)totBins) && it.hasNext()) {
		keyDens = (Double)it.next();
		nBins += densHist.get(keyDens);
		//System.out.println(format.format(keyDens) + "\t" + densHist.get(keyDens));
	    }
	    //System.out.println("Cut-off density is " + format.format(keyDens) + " for length " + lVec[i] + " with " + nBins);
	    cutOffs[i] = keyDens;
	}
	return cutOffs;
    }

    public boolean nearbyBreakPoint(Vector<Integer> breakPoints, int pos, int l) { return nearbyBreakPoint(breakPoints, null, pos, l); }

    public boolean nearbyBreakPoint(Vector<Integer> breakPoints, ReducedModel2 reducedModel, int pos, int l) {
	for (Integer b : breakPoints) {
	    if (b + Math.pow(2, l)>pos && b - Math.pow(2, l)<pos)
		return true;
	}
	if (reducedModel!=null) {
	    for (TranscribedRegion2 t : reducedModel.getSegments()) {
		if ((t.getStart() + Math.pow(2, l)>pos && t.getStart() - Math.pow(2, l)<pos) || (t.getEnd() + Math.pow(2, l)>pos && t.getEnd() - Math.pow(2, l)<pos))
		    return true;
	    }
	}
	return false;
    }

    public int numberOfBins(int lengthScale) { return numberOfBins(lengthScale, chr, species); }

    public static int numberOfBins(int lengthScale, int chr, String species) {
	if (species.equals("mm"))
	    return RefSeqUCSCEnsemblGenes.CHROMOSOME_LENGTH[chr - 1]/lengthScale;
	System.err.println("Warning, no genome length information for " + species);
	return 0;
    }

    public static int numberOfBasePairs(int chr, String species) { return numberOfBins(1, chr, species); }

    ////////////////////////////////////////////

    //This is a simpler version than the one in the original paper by
    //BenYaacov et al. Just invert the normal distribution specified
    //by sigmaHat and then find the cut-off required for the desired FDR.
    public static double calculateThreshold(Tally coeffsTally, double fdr, boolean upper) { return NormalDist.inverseF(coeffsTally.average(), coeffsTally.numberObs()>1 ? coeffsTally.standardDeviation() : 0, upper ? 1-fdr/2 : fdr/2); }

    //Assume that it is centered
    public static double calculateThreshold(Tally coeffsTally, double fdr) { return NormalDist.inverseF(0, coeffsTally.numberObs()>1 ? coeffsTally.standardDeviation() : 0, 1-fdr/2); }

    public static double calculateThreshold(double locMax[], double sigmaHat, double fdr) { return calculateThreshold(locMax, sigmaHat, fdr, null); }

    

    public static double calculateThreshold(double locMax[], double sigmaHat, double fdr, String saveFileName) {
	//Returns the threshold for which we need to go above for a
	//coefficient to be included.
	Arrays.sort(locMax); //ascending order
	double locMaxRev[] = new double[locMax.length];
	for (int i=0; i<locMax.length; i++) { locMaxRev[i] = locMax[locMax.length - i - 1]; }
	double thres = fdr/locMaxRev.length;
	if (saveFileName!=null) {
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		for (int j=0; j<locMax.length; j++) 
		    bw.write(format.format(locMax[j]) + "\n");
		bw.close();
	    }
	    catch (IOException e) { }	    
	}
	for (int i=0; i<locMaxRev.length; i++) {
	    double pVal = 2*NormalDist.barF01(locMaxRev[i]/sigmaHat);
	    if (pVal>(i+1)*thres) { //over the threshold, take the previous value
		int thresInd = i==0 ? i : i-1;
		//System.out.println("Found threshold " + format.format(locMaxRev[thresInd]) + " with thres " + (i*thres) + " at " + thresInd + " with p-val " + pVal + " for " + locMax.length);
		return locMaxRev[thresInd];
	    }
	}
	return locMaxRev.length>0 ? locMaxRev[0] + 1 : 1;
    }

    public static double estimateStdDev(Vector<double[]> dss) {
	//Concatenate the first line of all the vectors first
	int n = 0, i = 0;
	for (double[] ds : dss)
	    n += ds.length;
	double d[] = new double[n];
	for (double[] ds : dss) {
	    for (int j=0; j<ds.length; j++)
		d[i++] = Math.abs(ds[j]);
	}
	return estimateStdDev(d);
    }

    public static double estimateStdDev(double d[]) {
	//use the median absolute deviation
	double c[] = new double[d.length];
	Tally tl = new Tally("medians, log");
	for (int i=0; i<d.length; i++) { 
	    c[i] = Math.abs(d[i]); 
	    if (c[i]>0)
		tl.add(Math.log(c[i]));
	    else
		tl.add(Math.log(.0000001));
	}
	Arrays.sort(c);
	//System.out.println(tl.shortReportHeader());
	//System.out.println(tl.reportAndCIStudent(.95));
	//System.out.println("log-normal median = " + Math.exp(tl.average()));
	//System.out.println("exact median = " + format.format(c[c.length/2]));
	return c[c.length/2]/.6745; //constant taken from paper
    }

    public static double estimateStdDev(TreeMap<Integer, Integer> coeffHist, double binSize) {
	//use the median absolute deviation
	int n = 0, m = 0;
	for (Integer k : coeffHist.keySet())
	    n += coeffHist.get(k);
	for (Integer k : coeffHist.keySet()) {
	    m += coeffHist.get(k);
	    if (m>n/2)
		return binSize*((double)k)/.6745; //constant taken from paper
	}
	System.err.println("Error, could not find median!");
	return 0;
    }

    public static double estimateStdDev(TreeMap<Double, Integer> coeffHist) {
	//use the median absolute deviation
	int n = 0, m = 0;
	for (Double k : coeffHist.keySet())
	    n += coeffHist.get(k);
	for (Double k : coeffHist.keySet()) {
	    m += coeffHist.get(k);
	    if (m>n/2)
		return k/.6745; //constant taken from paper
	}
	System.err.println("Error, could not find median!");
	return 0;
    }

    public static double estimateStdDev(Tally t) {
	//System.out.println(t.shortReportHeader());
	//System.out.println(t.reportAndCIStudent(.95));
	double m = t.average();
	//System.out.println("approx median = " + format.format(m));
	return m/.6745; //constant taken from paper
    }


    ///////////////////////////////////////////////////////

    public static TreeMap<Integer, Double> findLocalMax(double w[], int startPos, int binSize) { return findLocalMax(w, startPos, binSize, 1); }

    public static TreeMap<Integer, Double> findLocalMax(double w[], int startPos, int binSize, double var) {
	//Return all the local maxima found in the vector w. Try
	//dividing the square of the coefficients with the estimated
	//variance to clean up the signal (not likely to help much
	//since we are looking for local maxima).
	TreeMap<Integer, Double> locValMap = new TreeMap<Integer, Double>();
	int n = 0, j = 0, bin = (int)Math.pow(2, binSize);
	for (int i=1; i<w.length-1; i++) {
	    if (isLocalExtrema(w, i))
		locValMap.put(startPos + i*bin, w[i]);
	}
	return locValMap;
    }

    public static boolean isLocalExtrema(double w[], int i) { 
	//This function makes sure that we are not at a plateau that
	//will go further up or down once it has ended
	if (Math.abs(w[i-1])<Math.abs(w[i]) && Math.abs(w[i+1])<Math.abs(w[i])) //True extrema
	    return true;
	else if (Math.abs(w[i-1])==Math.abs(w[i]) && Math.abs(w[i+1])<Math.abs(w[i])) {
	    int j = i-1;
	    while (j>0 && w[j]==w[j+1])
		j--;
	    return Math.abs(w[j])<Math.abs(w[i]);
	}
	else if (Math.abs(w[i-1])<Math.abs(w[i]) && Math.abs(w[i+1])==Math.abs(w[i])) { //start or end of plateau, could be considered an extrema
	    int j = i+1;
	    while (j<w.length && w[j]==w[j-1])
		j++;
	    return Math.abs(w[j])<Math.abs(w[i]);
	}
	return false;
    }

    ////////////////////////////////////////////

    public double[] calculateHistogram(int start, int end) { return calculateHistogram(start, end, this.reads, this.chr, this.maxReadLength); }
    
    public static double[] calculateHistogram(int start, int end, ABIRead[] reads, int chr, int maxReadLength) { return calculateHistogram(start, end, reads, chr, maxReadLength, false); }

    public static double[] calculateHistogram(int start, int end, ABIRead[] reads, int chr, int maxReadLength, boolean log) {
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), start - maxReadLength));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), end));
	endInd = endInd<0 ? -endInd-1 : endInd;	
	//System.out.println("StartInd is " + startInd + " for " + start + " and endInd is " + endInd + " for " + end);
	double hist[] = new double[end - start];
	for (int i=startInd; i<endInd; i++) {
	    for (int j=0; j<reads[i].getLength(); j++) { 
		if (reads[i].getPosition() - start + j>=0 && reads[i].getPosition() - start + j<hist.length)
		    hist[reads[i].getPosition() - start + j]++; 
	    }
	}
	if (log) {
	    for (int i=0; i<hist.length; i++)
		hist[i] = Math.log(hist[i] + 1);
	}
	return hist;
    }

    public double[] calculateHistogram(ABIRead rs[]) { return calculateHistogram(rs, maxReadLength, forward); }

    public static double[] calculateHistogram(ABIRead rs[], int maxReadLen, boolean strand) {
	//Unfold all the reads and create a histogram. This may be a
	//very long array if the number of reads is large. 
	Arrays.sort(rs);
	double hist[] = new double[rs[rs.length-1].getPosition() - rs[0].getPosition() + 2*maxReadLen];
	int r0 = rs[0].getPosition();
	Arrays.fill(hist, 0);
	for (int i=0; i<rs.length; i++) {
	    for (int j=0; j<rs[i].getLength(); j++) { hist[rs[i].getPosition() - r0 + j + maxReadLen]++; }
	}
	return hist;
    }

    public static double[] calculateHistogram(ABIRead rs[], int maxReadLen, int pad, int binSize) { return calculateHistogram(rs, maxReadLen, pad, binSize, null); }

    public static double[] calculateHistogram(ABIRead rs[], int maxReadLen, int pad, int binSize, String saveInfoFileName) {
	//pad the histogram with pad zeros before and after
	Arrays.sort(rs);
	int bin = (int)Math.pow(2, binSize), padBefore = Math.min(pad, rs[0].getPosition()), len = rs[rs.length-1].getPosition() - rs[0].getPosition() + padBefore + pad + maxReadLen;
	double hist[] = new double[len/bin+1];
	int r0 = rs[0].getPosition();
	Arrays.fill(hist, 0);
	for (int i=0; i<rs.length; i++) {
	    for (int j=0; j<rs[i].getLength(); j++) { hist[(rs[i].getPosition() - r0 + j + padBefore)/bin]++; }
	}
	if (saveInfoFileName!=null) {
	    String name = saveInfoFileName + "_hist_" + rs[0].getChromosome() + ".txt";
	    System.out.println("Saving histograms to " + name);
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(name));
		for (int i=0; i<hist.length; i++)
		    bw.write((r0 + i*bin) + "\t" + hist[i] + "\n");
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving histogram to " + name);
	    }
	}
	return hist;
    }

    public static double calculateVariance(double hist[], boolean log) {
	double v = 0, m = TranscribedRegion2.calculateAverageReadLevel(hist, log);
	for (int i=0; i<hist.length; i++)
	    v += log ? (Math.log(hist[i] + 1) - m)*(Math.log(hist[i] + 1) - m) : (hist[i] - m)*(hist[i] - m);
	return v/(hist.length - 1);
    }

    /////////////////////////////////

    public static double calculateHaarCoefficientNReads(int l, int n, int maxReadLen, boolean log, double normFactor) {
	//Calculate the coefficient that one would obtain if there were n reads in the window of length 2^l - 1.
	double hist[] = new double[(int)Math.pow(2, l) - 1], r = (double)(maxReadLen*n)/hist.length;
	if (r>1)
	    Arrays.fill(hist, r);
	else {
	    for (int i=0; i<maxReadLen*n; i++)
		hist[i] = 1;
	}
	return calculateHaarCoefficient(l, 0, hist, log, normFactor);
    }

    public static double calculateHaarCoefficient(int l, int ind, double hist[], boolean log, double normFactor) {
	//Use this when calculating the coefficient for the first
	//location and length scale, before we can use previously
	//computed results.
	double w = 0;
	for (int i=ind; i<Math.min(hist.length, ind + Math.pow(2, l) - 1); i++)
	    w += log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	for (int i=Math.max(0, ind - (int)Math.pow(2, l)); i<ind - 1; i++)
	    w -= log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	return w;
    }

    public static double calculateHaarCoefficient(int l, int ind, double hist[], boolean log, double normFactor, ReducedModel2 reducedModel, int histStart, int binSize) {
	//Use this when calculating the coefficient for the first
	//location and length scale, before we can use previously
	//computed results.
	if (reducedModel==null)
	    return calculateHaarCoefficient(l, ind, hist, log, normFactor);	
	double w = 0;
	for (int i=ind; i<Math.min(hist.length, ind + Math.pow(2, l) - 1); i++) {
	    if (!reducedModel.segmentOverlaps(histStart + (i - 1)*binSize, histStart + (i)*binSize))
		w += log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	}
	for (int i=Math.max(0, ind - (int)Math.pow(2, l)); i<ind - 1; i++) {
	    if (!reducedModel.segmentOverlaps(histStart + (i - 1)*binSize, histStart + (i)*binSize))
		w -= log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	}
	return w;
    }

    public static double calculateHaarCoefficient(int l, int ind, double hist[], double wpl, boolean log, double normFactor) {
	//Since we have already split the set into groups where we
	//know that the bases before and after the first and last read
	//are zero, we can speed up the calculation by ignoring those
	//bases when looping.
	double w = wpl/Math.sqrt(2);
	for (int i=ind + (int)Math.pow(2, l - 1); i<Math.min(hist.length, ind + Math.pow(2, l) - 1); i++)
	    w += log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	for (int i=Math.max(0, ind - (int)Math.pow(2, l)); i<ind - (int)Math.pow(2, l - 1) - 1; i++)
	    w -= log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	return w;
    }

    public static double calculateHaarCoefficient(int l, int ind, double hist[], double wpl, boolean log, double normFactor, ReducedModel2 reducedModel, int histStart, int binSize) {
	//Since we have already split the set into groups where we
	//know that the bases before and after the first and last read
	//are zero, we can speed up the calculation by ignoring those
	//bases when looping.
	if (reducedModel==null)
	    return calculateHaarCoefficient(l, ind, hist, wpl, log, normFactor);
	double w = wpl/Math.sqrt(2);
	for (int i=ind + (int)Math.pow(2, l - 1); i<Math.min(hist.length, ind + Math.pow(2, l) - 1); i++) {
	    if (!reducedModel.segmentOverlaps(histStart + (i - 1)*binSize, histStart + (i)*binSize))
		w += log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	}
	for (int i=Math.max(0, ind - (int)Math.pow(2, l)); i<ind - (int)Math.pow(2, l - 1) - 1; i++) {
	    if (!reducedModel.segmentOverlaps(histStart + (i - 1)*binSize, histStart + (i)*binSize))
		w -= log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	}
	return w;
    }
    
    public static double calculateHaarCoefficientInc(int l, int ind, double hist[], double wpp, boolean log, double normFactor) {
	//Save time by only doing the changes compared to the previous
	//coefficient:
	//w_n = w_(n-1) + y(n+2^l-1) + y(n-2^l-1) - 2*y(n-1)
	int l2 = (int)Math.pow(2, l);
	double yForwardAdd = ind + l2<hist.length ? (log ? Math.log(hist[ind + l2] + 1)*normFactor : (double)hist[ind + l2]*normFactor) : 0;
	double yBackwardAdd = ind - 1 - l2>0 ? (log ? Math.log(hist[ind - 1 - l2] + 1)*normFactor : (double)hist[ind - 1 - l2]*normFactor) : 0;
	double ySub = ind>0 ? (log ? Math.log(hist[ind - 1] + 1)*normFactor : (double)hist[ind - 1]*normFactor) : 0;
	return wpp - 2*ySub + yForwardAdd + yBackwardAdd;
    }

    public static double calculateHaarCoefficientInc(int l, int ind, double hist[], double wpp, boolean log, double normFactor, ReducedModel2 reducedModel, int histStart, int binSize) {
	//Save time by only doing the changes compared to the previous
	//coefficient:
	//w_n = w_(n-1) + y(n+2^l-1) + y(n-2^l-1) - 2*y(n-1)
	if (reducedModel==null)
	    return calculateHaarCoefficientInc(l, ind, hist, wpp, log, normFactor);
	//Also, ignore regions that are already covered by a transcribed region
	int l2 = (int)Math.pow(2, l);
	double yForwardAdd = 0, yBackwardAdd = 0, ySub = 0;
	if (ind + l2<hist.length && !reducedModel.segmentOverlaps(histStart + (ind + l2 - 1)*binSize, histStart + (ind + l2)*binSize))
	    yForwardAdd = log ? Math.log(hist[ind + l2] + 1)*normFactor : (double)hist[ind + l2]*normFactor;
	if (ind - 1 - l2>0 && !reducedModel.segmentOverlaps(histStart + (ind - l2 - 1)*binSize, histStart + (ind - l2)*binSize))
	    yBackwardAdd = log ? Math.log(hist[ind - 1 - l2] + 1)*normFactor : (double)hist[ind - 1 - l2]*normFactor;
	if (ind>0 && !reducedModel.segmentOverlaps(histStart + (ind - 1)*binSize, histStart + (ind)*binSize))
	    ySub = log ? Math.log(hist[ind - 1] + 1)*normFactor : (double)hist[ind - 1]*normFactor;
	return wpp - 2*ySub + yForwardAdd + yBackwardAdd;
    }

    //////////////////////////////////////////////////////////////////
    //For calculating statistics about the density of the coverage

    public ABIRead[] getReads() { return reads; }

    public double[] shiftHistogram(double[] hist, int newLastWinInd, int maxLen) {
	double h[] = new double[Math.min(maxLen, hist.length + 1)];
	if (h.length==maxLen) { //move all the stuff one bp
	    for (int i=0; i<hist.length-1; i++)
		h[i] = hist[i+1];
	}
	else {
	    for (int i=0; i<hist.length; i++)
		h[i] = hist[i];
	}
	int i = 0; //check if there is anything to be added in the last slot
	while (readsInd+i<reads.length && reads[readsInd+i].getPosition()<newLastWinInd) {
	    if (reads[readsInd+i].overlaps(newLastWinInd)) //reads may have different length and hence we cannot be sure that there will be overlap
		h[h.length - 1]++;
	    i++;
	}
	//step readsInd to ignore reads which may no longer contribute
	while (readsInd<reads.length && reads[readsInd].getPosition()<newLastWinInd && !reads[readsInd].overlaps(newLastWinInd)) { 
	    readsInd++; 
	    //System.out.println(readsInd + "\t" + genomeInd + "\t" + newLastWinInd);
	}
	return h;
    }

    public boolean moreReadGroups() { return readGroupInd<reads.length; }

    public ABIRead[] nextReadGroup(int sepDist) {
	int startInd = readGroupInd, endInd;
	while (readGroupInd<reads.length-1 && reads[readGroupInd+1].getPosition() - reads[readGroupInd].getPosition()<sepDist) 
	    readGroupInd++;
	endInd = ++readGroupInd;
	ABIRead rs[] = new ABIRead[endInd - startInd];
	for (int i=startInd; i<endInd; i++)
	    rs[i-startInd] = reads[i];
	return rs;
    }

    public static double evaluateCostFunction(String costFunction, ReducedModel2 r, int geneIDsKeep[], Vector<Peak> peaks, int maxDivergentTSSDistance, boolean reverseStrand) {
	double cov = 0;
	if (costFunction.equalsIgnoreCase("genes"))
	    //cov = (double)r.numberOfGenesWithOneTranscript()/geneIDsKeep.length;
	    cov = (double)r.numberOfGenesWithOneTranscript()/geneIDsKeep.length - (double)r.numberOfGenesWithNTranscripts(2, true, false)/geneIDsKeep.length;
	//else if (costFunction.equalsIgnoreCase("coverage"))
	//  cov = r.proportionOfReads() - r.proportionTranscribed();
	else if (costFunction.equalsIgnoreCase("ppv"))
	    cov = r.getPositivePredictiveValue();
	else if (costFunction.equalsIgnoreCase("sensitivity"))
	    cov = r.getSensitivity(geneIDsKeep);
	else if (costFunction.equalsIgnoreCase("peaks")) {
	    TreeMap<TranscribedRegion2, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion2, Peak>() : r.compare2Peaks(peaks, maxDivergentTSSDistance);
	    cov =  (double)transcriptsWithPeaks.size()/(peaks.size() - transcriptsWithPeaks.size() + r.numberOfTranscribedRegions());
	}
	return cov;
    }

}