//Use this class when we want to analyze the wider peaks formed by the
//elongated forms of Pol2 or histones. A slightly different strategy
//is employed here, which is more direcly related to the empirical
//Bayes strategy suggested by Efron since we no longer have a
//theoretical null model to go by. The statistic which we use is the
//difference in z-scores between the ChIP and the input
//distributions. These scores can either be pre-calculated for the
//annotated regions, or they can be

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
//import org.apache.commons.math.stat.regression.OLSMultipleLinearRegression;
import Jama.Matrix;


public class WiderPeakLocFDR {

    public static final String Z_VAL_SUFFIX = "_genes_reads_";

    public static DecimalFormat format = new DecimalFormat("###.#######");

    //////////////////////////////////////////////////////////////////

    protected Vector zDiff = new Vector();
    protected TreeSet zDiffNorm;
    protected TreeMap hist = new TreeMap();
    //The coefficients for the estimate of the empirical
    protected double[] coeffs;
    //These are required to transform a z-value 
    protected double mean, stdDev, binWidth;
    //Paramterers for the null dist
    protected double p0, mu0, sigma0;
    protected int order = 7;
    //Keep track of the data which we have recovered for each gene
    protected TreeMap geneData = new TreeMap();

    //////////////////////////////////////////////////////////////////

    public WiderPeakLocFDR(String fileName, double b, int windowSize, int windowStep, int nMappabilityBins, int nGCBins, boolean orchestra) {
	for (int i=19; i<=19; i++) { //Find the z-diff values for all chromosomes
	    RefSeqGenes refSeqGenes = new RefSeqGenes(i, orchestra);
	    Vector zd = loadZValues(i, fileName, refSeqGenes);
	    if (zd!=null)
		zDiff.addAll(zd);
	    else
		zDiff.addAll(getZValues(i, fileName, refSeqGenes, windowSize, windowStep, nMappabilityBins, nGCBins));
	}
	binWidth = b;
	//Normalize by subtracting the mean and dividing by the std
	//dev. This is equivalent to zscore() in Matlab
	zDiffNorm = new TreeSet(zScore(zDiff));
	    System.out.println("Found " + zDiffNorm.size() + " z-values.");
	//Bin to create a histogram. Note that the histograms contain
	//the raw counts for each bin and that they are not
	//normalized.
	hist = binData(zDiffNorm, binWidth);
	//Use the Poisson regression method suggested by Efron to find
	//a fit. The values in the vector are coefficients for Poisson
	//GLM.
	coeffs = findPoissonGLMFit(hist, binWidth);
	//Find a quadratic approximation to estimate the
	//null-distribution. Do as in Efron and fit a quadratic
	//function using OLS to central 1/3 range of the z-values. The
	//values in the vector are p0, mean and std dev for the
	//approximate normal distribution.
	findEmpiricalNullDistribution();
    }

    //////////////////////////////////////////////////////////////////

    public double[] findEmpiricalNullDistribution() {
	//Find the central 1/3 of the range
	double maxZ = binWidth*(double)((Integer)hist.lastKey()).intValue(), minZ = binWidth*(double)((Integer)hist.firstKey()).intValue();
	double centerStart = -(maxZ - minZ)/6, centerEnd = (maxZ - minZ)/6;
	//Regress a normal dist with a weight based on the central values
	int keyStart = (int)Math.round(Math.floor((-(maxZ - minZ)/6)/binWidth)), keyEnd = (int)Math.round(Math.floor(((maxZ - minZ)/6)/binWidth));
	int k = keyStart;
	double normConst = binWidth*(double)zDiffNorm.size();
	//Find the number of bins in this range
	double y[][] = new double[keyEnd-keyStart][1], x[][] = new double[keyEnd-keyStart][3];
	while (k<keyEnd) {
	    double z = k*binWidth + binWidth/2; //mid-point of the bin
	    Integer key = new Integer(k);
	    if (hist.containsKey(key)) {
		//y[k-keyStart][0] = Math.log((double)((Integer)hist.get(key)).intValue()/normConst);
		//Fit to the Poisson-approximation
		y[k-keyStart][0] = Math.log(empiricalDensity(z));
		x[k-keyStart][0] = z*z;
		x[k-keyStart][1] = z;
		x[k-keyStart][2] = 1;
	    }
	    k++;
	}
	Matrix yMatrix = new Matrix(y), xMatrix = new Matrix(x);
	Matrix b = xMatrix.solve(yMatrix);
	double cs[] = b.getRowPackedCopy();
	//Obtain the parameters for the null-distribution from the regression
	sigma0 = Math.pow(-2*cs[0], -.5);
	mu0 = cs[1]*sigma0*sigma0;
	p0 = Math.exp(cs[2] + (mu0*mu0/(sigma0*sigma0) + Math.log(2*Math.PI*sigma0*sigma0))/2);
	System.out.println("p0 = " + p0 + ", mu0 = " + mu0 + ", sigma0 = " + sigma0);
	return new double[]{p0, mu0, sigma0};
    }

    public double[] findPoissonGLMFit(TreeMap hist, double binWidth) { return findPoissonGLMFit(hist, binWidth, 7); }

    public double[] findPoissonGLMFit(TreeMap hist, double binWidth, int order) {
	//This methdo returns the regression coefficients for the
	//Poisson-fit of the normalized GLM for the histogram.
	double normConst = binWidth*(double)zDiffNorm.size();
	double x[][] = new double[hist.size()][order+1], y[][] = new double[hist.size()][1];
	Iterator it = hist.keySet().iterator();
	int j = 0;
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    double z = ((double)key.intValue())*binWidth + binWidth/2; //mid-point of the bin
	    for (int i=0; i<=order; i++) { x[j][i] = Math.pow(z, i); }
	    y[j++][0] = Math.log((double)(((Integer)hist.get(key)).intValue())/normConst);
	    //y[j++][0] = Math.log((double)(((Integer)hist.get(key)).intValue()));
	}
	Matrix yMatrix = new Matrix(y), xMatrix = new Matrix(x);
	Matrix b = xMatrix.solve(yMatrix);
	//Determine the final coefficient so that the we integrate to 1
	double ysum = 0, xsum = 0;
	it = hist.keySet().iterator();
	j = 0;
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    double z = ((double)key.intValue())*binWidth + binWidth/2; //mid-point of the bin
	    ysum += y[j][0];
	    double xs = 0;
	    for (int i=1; i<=order; i++)
		xs += b.get(i, 0)*Math.pow(z, i);
	    xsum += Math.exp(xs);
	}
	b.set(0, 0, -Math.log(xsum));
	b.transpose().print(8, 6);
	return b.getRowPackedCopy();
    }

    public TreeMap binData(TreeSet data, double binWidth) {
	//We typically assume that the data has been normalized using
	//zScore. This means that we can always use the same binWidth
	//without further worries. When choosing the bins, we make
	//sure that one of them starts at zero. First we create the
	//bins which we will be needing. The keys are integers,
	//corresponding to multiples of binWidth for the end of the
	//bin furthest from zero.
	Iterator it = data.iterator(); //Now we can add the data to the histogram
	while (it.hasNext()) {
	    double v = ((Double)it.next()).doubleValue();
	    Integer key = null;
	    key = new Integer((int)Math.round(Math.floor(v/binWidth)));
	    Integer val = hist.containsKey(key) ? (Integer)hist.get(key) : new Integer(0);
	    hist.put(key, new Integer(val.intValue() + 1));	    
	}
	it = hist.keySet().iterator(); 
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    //System.out.println(key.toString() + "\t" + (((Integer)hist.get(key)).toString()));
	}
	return hist;
    }
    
    public Vector zScore(Vector vec) {
	double s = 0, s2 = 0;
	for (int i=0; i<vec.size(); i++) 
	    s += ((Double)vec.elementAt(i)).doubleValue();
	mean = s/vec.size();
	for (int i=0; i<vec.size(); i++) 
	    s2 += Math.pow(((Double)vec.elementAt(i)).doubleValue() - mean, 2);
	stdDev = Math.sqrt(s2/(vec.size() - 1));
	double minZ = 0, maxZ = 0;
	Vector v = new Vector(vec.size());
	for (int i=0; i<vec.size(); i++) {
	    double z = ((((Double)vec.elementAt(i)).doubleValue() - mean)/stdDev);
	    while (v.contains(new Double(z))) //Make sure to avoid doubles
		z += 1e-13;
	    v.addElement(new Double(z));
	    minZ = minZ<z ? minZ : z;
	    maxZ = maxZ>z ? maxZ : z;
	}
	System.out.println(mean + "\t" + stdDev + "\t" + v.size() + "\t" + minZ + "\t" + maxZ + "\n");
	return v;
    }

    public static Vector getZValues(int chr, String fileName, RefSeqGenes refSeqGenes, int windowSize, int windowStep, int nMappabilityBins, int nGCBins) {
	Vector zDiffs = new Vector();
	WindowFinder windowFinder = null;
	try {
	    windowFinder = new WindowFinder(fileName, chr, true, windowSize, windowStep, nMappabilityBins, nGCBins);
	}
	catch (IOException e) {
	    System.err.println("Error reading from " + fileName);
	    return zDiffs;
	}
	try {
	    BufferedWriter zValWriter = new BufferedWriter(new FileWriter(fileName + Z_VAL_SUFFIX + chr + ".txt"));
	    if (chr==1) { zValWriter.write("#name\tid\treads\tinput\treadsPoisson\tinputPoisson\tzdiff\tlength\n"); }
	    Iterator it = refSeqGenes.getGenes().iterator();
	    while (it.hasNext()) {
		RefSeqGene gene = (RefSeqGene)it.next();
		//Find the degree of Pol2-enrichment for the coding
		//region of this gene
		Window w = windowFinder.findReadsInCodingRegion(gene, null);
		//Calculate the enrichment of the reads over the
		//enrichment of the input. We want to compare with
		//the local expected enrichment rather than the
		//global.
		if (w!=null) {
		    double len = (double)gene.getLength()/windowSize, lambda = w.getMeanReadsEnrichment()*len;
		    double zd = (w.numberOfReads() - lambda)/Math.sqrt(lambda) - (w.numberOfInput() - lambda)/Math.sqrt(lambda);
		    zDiffs.addElement(new Double(zd));
		    zValWriter.write(gene.getName() + "\t" + gene.getID() + "\t" + w.numberOfReads() + "\t" + w.numberOfInput() + "\t" + format.format(w.getMeanReadsEnrichment()) + "\t" + format.format(w.getMeanInputEnrichment()) + "\t" + format.format(zd) + "\t" + gene.getLength() + "\n");
		}
		else {
		    System.out.println("Could not find any reads for " + gene.toString());
		    zValWriter.write(gene.getName() + "\t" + gene.getID() + "\t0\t0\t0\t0\t0\t" + gene.getLength() + "\n");
		}
	    }
	    zValWriter.close();
	    System.out.println("Saved to " + fileName);
	}
	catch (IOException e) {
	    System.err.println("Error reading from " + fileName);
	}
	return zDiffs;
    }
	
    public Vector loadZValues(int chr, String fileName, RefSeqGenes refSeqGenes) {
	Vector zDiffs = new Vector();
	try {
	    System.out.println("Trying to load from " + fileName + Z_VAL_SUFFIX + chr + ".txt");
	    BufferedReader zValReader = new BufferedReader(new FileReader(fileName + Z_VAL_SUFFIX + chr + ".txt"));
	    //name id reads input readsPoisson inputPoisson zdiff length 
	    Pattern zValPattern = Pattern.compile("(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+([\\.|\\d])+\\s+([\\.|\\d])+\\s+([-|\\.|\\d]+)\\s+(\\d+)");
	    String line = zValReader.readLine();
	    while (line!=null) {
		Matcher m = zValPattern.matcher(line);
		if (m.find()) {
		    zDiffs.addElement(new Double(m.group(7)));
		    RefSeqGene gene = refSeqGenes.getGene(m.group(1));
		    Window w = new Window(chr, gene.getStart(), gene.getEnd(), Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4)), Double.parseDouble(m.group(5)), Double.parseDouble(m.group(6)));
		    geneData.put(gene, w);
		}
		line = zValReader.readLine();
	    }
	    zValReader.close();
	}
	catch (IOException e) {
	    System.err.println("Error loading from " + fileName);
	    //zValReader.close();
	    return null;
	}
	return zDiffs;
    }

    public static TreeMap loadZValues(String fileName, RefSeqGenes refSeqGenes) {
	//Key is gene and value is z-score
	TreeMap zDiffs = new TreeMap();
	try {
	    BufferedReader zValReader = new BufferedReader(new FileReader(fileName + Z_VAL_SUFFIX + refSeqGenes.getChromosome() + ".txt"));
	    //name id reads input readsPoisson inputPoisson zdiff length 
	    Pattern zValPattern = Pattern.compile("(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+([\\.|\\d])+\\s+([\\.|\\d])+\\s+([-|\\.|\\d]+)\\s+(\\d+)");
	    String line = zValReader.readLine();
	    while (line!=null) {
		Matcher m = zValPattern.matcher(line);
		if (m.find()) {
		    RefSeqGene gene = refSeqGenes.getGene(m.group(1));
		    zDiffs.put(gene, new Double(m.group(7)));
		}
		line = zValReader.readLine();
	    }
	    zValReader.close();
	}
	catch (IOException e) {
	    System.err.println("Error loading from " + fileName);
	    //zValReader.close();
	    return null;
	}
	return zDiffs;
    }

    //////////////////////////////////////////////////////////////////

    public double getFDRThreshold(double z) {
	//Calculate the corresponding FDR value for this z-value.
	//First we need to normalize the z-value
	double zNorm = (z - mean)/stdDev;
	return p0*normalDensity(zNorm, mu0, sigma0)/empiricalDensity(zNorm);
    }

    public double empiricalDensity(double z) {
	//Evaluate the Poisson regression model
	double f = 0;
	for (int i=0; i<coeffs.length; i++) 
	    f += coeffs[i]*Math.pow(z, i);
	return Math.exp(f)/binWidth;
    }

    public TreeMap getGeneData() { return geneData; }

    public double[] getBinCenters() {
	double c[] = new double[hist.size()];
	int i = 0;
	Iterator it = hist.keySet().iterator();
	while (it.hasNext()) 
	    c[i++] = binWidth*(double)((Integer)it.next()).intValue() + binWidth/2;
	return c;
    }

    public int[] getBinCounts() {
	int c[] = new int[hist.size()], i = 0;
	Iterator it = hist.keySet().iterator();
	while (it.hasNext()) 
	    c[i++] = ((Integer)hist.get((Integer)it.next())).intValue();
	return c;
    }

    public double getP0() { return p0; }

    public double getMu0() { return mu0; }

    public double getSigma0() { return sigma0; }

    public static double normalDensity(double x, double mu, double sigma) { return normalDensity((x-mu)/sigma)/sigma; }

    public static double normalDensity(double x) { return Math.exp(-x*x/2)/Math.sqrt(2*Math.PI); }

}
