import org.apache.commons.math.stat.descriptive.SummaryStatistics;
import java.text.DecimalFormat;
import java.util.*;

public class Window {
    
    public static DecimalFormat format = new DecimalFormat("###.##");

    //Use this class to represent a window of a fixed length, which
    //can be extended by merging nearby windows.

    protected int chr, start, end, nReads, nInput, windowSize;
    //Keep track of the location of the peak maximum
    protected int max = 0, maxLoc;
    protected SummaryStatistics readsPoisson = new SummaryStatistics(), inputPoisson = new SummaryStatistics(), gcContent = new SummaryStatistics();
    //Use to keep track of the proportion of reads which are above the detection limit. This information is used when computing the saturation
    protected SummaryStatistics proportionAbove = new SummaryStatistics();
    //use this string to keep track of the bed-file representation of the window
    protected String bedString = "";

    //////////////////////////////////////////////////////////////////

    public Window(int c, int s, int e, int n, int i, double rp, double gc) {
	chr = c;
	start = s;
	end = e;
	windowSize = e - s;;
	nReads = n;
	nInput = i;
	max = n;
	maxLoc = (s + e)/2;
	readsPoisson.addValue(rp);
	gcContent.addValue(gc);
    }

    public Window(int c, int s, int e, int n, int i, double rp, double gc, int dist) {
	chr = c;
	start = s;
	end = e;
	windowSize = e - s;;
	nReads = n;
	nInput = i;
	max = n;
	maxLoc = (s + e)/2;
	readsPoisson.addValue(rp);
	gcContent.addValue(gc);
	proportionAbove.addValue((double)dist/nReads);
    }

    public Window(int c, int s, int e, int n, int i, double rp, double ip, double gc) {
	chr = c;
	start = s;
	end = e;
	windowSize = e - s;;
	nReads = n;
	nInput = i;
	max = n;
	maxLoc = (s + e)/2;
	readsPoisson.addValue(rp);
	inputPoisson.addValue(ip);
	gcContent.addValue(gc);
    }

    protected Window() { } //For the sub-class DifferentialWindow

    //////////////////////////////////////////////////////////////////

    public void setCenter(int n, int loc) {
	if (n>max) {
	    maxLoc = loc;
	    max = n;
	}
    }

    public void addReads(int n) { nReads += n; }

    public void addInput(int n) { nInput += n; }

    public void addReadsPoisson(double p) { readsPoisson.addValue(p); }

    public void addInputPoisson(double p) { inputPoisson.addValue(p); }

    public int numberOfReads() { return nReads; }

    public int numberOfInput() { return nInput; }

    //public double numberOfReads(boolean normalized) { return normalized ? numberOfReads() : numberOfNormalizedReads(); }

    public void setEnd(int e) { end = e; }

    public int getStart() { return start; }
    
    public int getEnd() { return end; }

    public int getCenter() { return maxLoc; }

    public int getLength() { return Math.abs(getEnd() - getStart()); }

    public int getChromosome() { return chr; }

    public double readsEnrichment() { return (double)nReads/(readsPoisson.getMean()*((double)(end - start))/windowSize); }

    public double getMeanReadsEnrichment() { return readsPoisson.getMean(); }

    public double getMeanInputEnrichment() { return inputPoisson.getMean(); }

    //calculate the expected number of reads in the window
    public double getReadsLambda() { return readsPoisson.getMean()*((double)(end - start))/windowSize; }

    public double getInputLambda() { return inputPoisson.getMean()*((double)(end - start))/windowSize; }

    public double getReadsZScore() { return ((double)nReads - getReadsLambda())/Math.sqrt(getReadsLambda()); }

    public double getInputZScore() { return ((double)nInput - getInputLambda())/Math.sqrt(getInputLambda()); }

    public double getZScoreDiff() { return getReadsZScore() - getInputZScore(); }

    public String getBedString() { return bedString; }

    public void appendBedString(String s) { bedString += s; }

    public double getGCContent() { return gcContent.getMean(); }

    public void setDetectionDistance(int d, int nReads) { proportionAbove.addValue((double)d/nReads); }

    public double getProportionAboveLimit() { return proportionAbove.getMean(); }

    public TreeSet<OverlapRegion> findOverlappingRegions(TreeSet<OverlapRegion> rs) {
	TreeSet<OverlapRegion> over = new TreeSet<OverlapRegion>();
	for (OverlapRegion r : rs) {
	    if (r.overlaps(this))
		over.add(r);
	}
	return over;
    }    

    //////////////////////////////////////////////////////////////////

    public String toString() { return chr + "\t" + start + "\t" + end + "\t" + format.format(getGCContent()) + "\t" + nReads + "\t" + nInput + "\t" + format.format(readsEnrichment()); }

    public boolean equals(Object o) {
	if (!(o instanceof Window) )
	    System.err.println("Error comparing " + o.toString());
	Window w = (Window)o;
	return w.chr==chr && w.start==start && w.end==end && w.nReads==nReads && w.nInput==nInput; 
    }
	

}
